Package: rsample
Title: General Resampling Infrastructure
Version: 1.0.0
Authors@R: c(
    person("Julia", "Silge", , "julia.silge@rstudio.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-3671-836X")),
    person("Fanny", "Chow", , "fannybchow@gmail.com", role = "aut"),
    person("Max", "Kuhn", , "max@rstudio.com", role = "aut"),
    person("Hadley", "Wickham", , "hadley@rstudio.com", role = "aut"),
    person("RStudio", role = c("cph", "fnd"))
  )
Maintainer: Julia Silge <julia.silge@rstudio.com>
Description: Classes and functions to create and summarize different types
    of resampling objects (e.g. bootstrap, cross-validation).
License: MIT + file LICENSE
URL: https://rsample.tidymodels.org,
        https://github.com/tidymodels/rsample
BugReports: https://github.com/tidymodels/rsample/issues
Depends: R (>= 3.4)
Imports: dplyr (>= 1.0.0), ellipsis, furrr, generics, glue, methods,
        pillar, purrr, rlang (>= 0.4.10), slider (>= 0.1.5), tibble,
        tidyr, tidyselect, vctrs (>= 0.3.0)
Suggests: broom, covr, ggplot2, knitr, modeldata, recipes (>= 0.1.4),
        rmarkdown, stats, testthat (>= 3.0.0), utils, xml2
VignetteBuilder: knitr
Config/Needs/website: GGally, nlstools, tidymodels,
        tidyverse/tidytemplate
Encoding: UTF-8
RoxygenNote: 7.2.0
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2022-06-24 17:08:48 UTC; juliasilge
Author: Julia Silge [aut, cre] (<https://orcid.org/0000-0002-3671-836X>),
  Fanny Chow [aut],
  Max Kuhn [aut],
  Hadley Wickham [aut],
  RStudio [cph, fnd]
Repository: CRAN
Date/Publication: 2022-06-24 18:10:02 UTC

Package: rsample
Title: General Resampling Infrastructure
Version: 1.3.2
Authors@R: c(
    person("Hannah", "Frick", , "hannah@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-6049-5258")),
    person("Fanny", "Chow", , "fannybchow@gmail.com", role = "aut"),
    person("Max", "Kuhn", , "max@posit.co", role = "aut"),
    person("Michael", "Mahoney", , "mike.mahoney.218@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0003-2402-304X")),
    person("Julia", "Silge", , "julia.silge@posit.co", role = "aut",
           comment = c(ORCID = "0000-0002-3671-836X")),
    person("Hadley", "Wickham", , "hadley@posit.co", role = "aut"),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Classes and functions to create and summarize different types
    of resampling objects (e.g. bootstrap, cross-validation).
License: MIT + file LICENSE
URL: https://rsample.tidymodels.org,
        https://github.com/tidymodels/rsample
BugReports: https://github.com/tidymodels/rsample/issues
Depends: R (>= 4.1)
Imports: cli, dplyr (>= 1.1.1), furrr, generics, glue, lifecycle,
        methods, pillar, purrr (>= 1.0.0), rlang (>= 1.1.0), slider (>=
        0.1.5), tibble, tidyr, tidyselect, vctrs (>= 0.5.0)
Suggests: broom, covr, ggplot2, knitr, modeldata, recipes (>= 0.1.4),
        rmarkdown, stats, testthat (>= 3.0.0), utils, whisker, withr,
        xml2
VignetteBuilder: knitr
Config/Needs/website: GGally, nlstools, tidymodels,
        tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-04-23
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-01-29 18:32:44 UTC; hannah
Author: Hannah Frick [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6049-5258>),
  Fanny Chow [aut],
  Max Kuhn [aut],
  Michael Mahoney [aut] (ORCID: <https://orcid.org/0000-0003-2402-304X>),
  Julia Silge [aut] (ORCID: <https://orcid.org/0000-0002-3671-836X>),
  Hadley Wickham [aut],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Hannah Frick <hannah@posit.co>
Repository: CRAN
Date/Publication: 2026-01-30 06:10:15 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",rset)
S3method("names<-",rset)
S3method(.get_fingerprint,default)
S3method(.get_fingerprint,rset)
S3method(analysis,default)
S3method(analysis,initial_validation_split)
S3method(analysis,rsplit)
S3method(analysis,val_split)
S3method(as.data.frame,rsplit)
S3method(as.integer,rsplit)
S3method(assessment,default)
S3method(assessment,initial_validation_split)
S3method(assessment,internal_calibration_split)
S3method(assessment,rsplit)
S3method(assessment,val_split)
S3method(calibration,default)
S3method(calibration,internal_calibration_split)
S3method(complement,apparent_split)
S3method(complement,default)
S3method(complement,rof_split)
S3method(complement,rsplit)
S3method(complement,sliding_index_split)
S3method(complement,sliding_period_split)
S3method(complement,sliding_window_split)
S3method(dim,initial_validation_split)
S3method(dim,rsplit)
S3method(get_rsplit,default)
S3method(get_rsplit,rset)
S3method(int_bca,bootstraps)
S3method(int_bca,default)
S3method(int_pctl,bootstraps)
S3method(int_pctl,default)
S3method(int_t,bootstraps)
S3method(int_t,default)
S3method(internal_calibration_split,apparent_split)
S3method(internal_calibration_split,boot_split)
S3method(internal_calibration_split,clustering_split)
S3method(internal_calibration_split,default)
S3method(internal_calibration_split,group_boot_split)
S3method(internal_calibration_split,group_initial_validation_split)
S3method(internal_calibration_split,group_mc_split)
S3method(internal_calibration_split,group_val_split)
S3method(internal_calibration_split,group_vfold_split)
S3method(internal_calibration_split,initial_time_split)
S3method(internal_calibration_split,initial_validation_split)
S3method(internal_calibration_split,initial_validation_time_split)
S3method(internal_calibration_split,mc_split)
S3method(internal_calibration_split,sliding_index_split)
S3method(internal_calibration_split,sliding_period_split)
S3method(internal_calibration_split,sliding_window_split)
S3method(internal_calibration_split,time_val_split)
S3method(internal_calibration_split,val_split)
S3method(internal_calibration_split,vfold_split)
S3method(labels,rset)
S3method(labels,rsplit)
S3method(labels,vfold_cv)
S3method(make_splits,data.frame)
S3method(make_splits,default)
S3method(make_splits,list)
S3method(obj_sum,rsplit)
S3method(populate,rset)
S3method(populate,rsplit)
S3method(pretty,apparent)
S3method(pretty,bootstraps)
S3method(pretty,clustering_cv)
S3method(pretty,group_bootstraps)
S3method(pretty,group_mc_cv)
S3method(pretty,group_validation_split)
S3method(pretty,group_vfold_cv)
S3method(pretty,loo_cv)
S3method(pretty,manual_rset)
S3method(pretty,mc_cv)
S3method(pretty,nested_cv)
S3method(pretty,permutations)
S3method(pretty,rolling_origin)
S3method(pretty,sliding_index)
S3method(pretty,sliding_period)
S3method(pretty,sliding_window)
S3method(pretty,validation_set)
S3method(pretty,validation_split)
S3method(pretty,vfold_cv)
S3method(print,apparent)
S3method(print,bootstraps)
S3method(print,clustering_cv)
S3method(print,group_bootstraps)
S3method(print,group_mc_cv)
S3method(print,group_validation_split)
S3method(print,group_vfold_cv)
S3method(print,initial_split)
S3method(print,initial_validation_split)
S3method(print,internal_calibration_split)
S3method(print,loo_cv)
S3method(print,manual_rset)
S3method(print,mc_cv)
S3method(print,nested_cv)
S3method(print,permutations)
S3method(print,rolling_origin)
S3method(print,rsplit)
S3method(print,sliding_index)
S3method(print,sliding_period)
S3method(print,sliding_window)
S3method(print,val_split)
S3method(print,validation_split)
S3method(print,vfold_cv)
S3method(reverse_splits,default)
S3method(reverse_splits,perm_split)
S3method(reverse_splits,permutations)
S3method(reverse_splits,rset)
S3method(reverse_splits,rsplit)
S3method(testing,default)
S3method(testing,initial_validation_split)
S3method(testing,rsplit)
S3method(testing,val_split)
S3method(tidy,nested_cv)
S3method(tidy,rset)
S3method(tidy,rsplit)
S3method(tidy,vfold_cv)
S3method(training,default)
S3method(training,initial_validation_split)
S3method(training,rsplit)
S3method(training,val_split)
S3method(type_sum,rsplit)
S3method(validation,default)
S3method(validation,initial_validation_split)
S3method(validation,val_split)
S3method(vec_cast,apparent.apparent)
S3method(vec_cast,apparent.data.frame)
S3method(vec_cast,apparent.tbl_df)
S3method(vec_cast,bootstraps.bootstraps)
S3method(vec_cast,bootstraps.data.frame)
S3method(vec_cast,bootstraps.tbl_df)
S3method(vec_cast,clustering_cv.clustering_cv)
S3method(vec_cast,clustering_cv.data.frame)
S3method(vec_cast,clustering_cv.tbl_df)
S3method(vec_cast,data.frame.apparent)
S3method(vec_cast,data.frame.bootstraps)
S3method(vec_cast,data.frame.clustering_cv)
S3method(vec_cast,data.frame.group_bootstraps)
S3method(vec_cast,data.frame.group_mc_cv)
S3method(vec_cast,data.frame.group_validation_split)
S3method(vec_cast,data.frame.group_vfold_cv)
S3method(vec_cast,data.frame.loo_cv)
S3method(vec_cast,data.frame.manual_rset)
S3method(vec_cast,data.frame.mc_cv)
S3method(vec_cast,data.frame.nested_cv)
S3method(vec_cast,data.frame.permutations)
S3method(vec_cast,data.frame.rolling_origin)
S3method(vec_cast,data.frame.sliding_index)
S3method(vec_cast,data.frame.sliding_period)
S3method(vec_cast,data.frame.sliding_window)
S3method(vec_cast,data.frame.validation_set)
S3method(vec_cast,data.frame.validation_split)
S3method(vec_cast,data.frame.validation_time_split)
S3method(vec_cast,data.frame.vfold_cv)
S3method(vec_cast,group_bootstraps.data.frame)
S3method(vec_cast,group_bootstraps.group_bootstraps)
S3method(vec_cast,group_bootstraps.tbl_df)
S3method(vec_cast,group_mc_cv.data.frame)
S3method(vec_cast,group_mc_cv.group_mc_cv)
S3method(vec_cast,group_mc_cv.tbl_df)
S3method(vec_cast,group_validation_split.data.frame)
S3method(vec_cast,group_validation_split.group_validation_split)
S3method(vec_cast,group_validation_split.tbl_df)
S3method(vec_cast,group_vfold_cv.data.frame)
S3method(vec_cast,group_vfold_cv.group_vfold_cv)
S3method(vec_cast,group_vfold_cv.tbl_df)
S3method(vec_cast,loo_cv.data.frame)
S3method(vec_cast,loo_cv.loo_cv)
S3method(vec_cast,loo_cv.tbl_df)
S3method(vec_cast,manual_rset.data.frame)
S3method(vec_cast,manual_rset.manual_rset)
S3method(vec_cast,manual_rset.tbl_df)
S3method(vec_cast,mc_cv.data.frame)
S3method(vec_cast,mc_cv.mc_cv)
S3method(vec_cast,mc_cv.tbl_df)
S3method(vec_cast,nested_cv.data.frame)
S3method(vec_cast,nested_cv.nested_cv)
S3method(vec_cast,nested_cv.tbl_df)
S3method(vec_cast,permutations.data.frame)
S3method(vec_cast,permutations.permutations)
S3method(vec_cast,permutations.tbl_df)
S3method(vec_cast,rolling_origin.data.frame)
S3method(vec_cast,rolling_origin.rolling_origin)
S3method(vec_cast,rolling_origin.tbl_df)
S3method(vec_cast,sliding_index.data.frame)
S3method(vec_cast,sliding_index.sliding_index)
S3method(vec_cast,sliding_index.tbl_df)
S3method(vec_cast,sliding_period.data.frame)
S3method(vec_cast,sliding_period.sliding_period)
S3method(vec_cast,sliding_period.tbl_df)
S3method(vec_cast,sliding_window.data.frame)
S3method(vec_cast,sliding_window.sliding_window)
S3method(vec_cast,sliding_window.tbl_df)
S3method(vec_cast,tbl_df.apparent)
S3method(vec_cast,tbl_df.bootstraps)
S3method(vec_cast,tbl_df.clustering_cv)
S3method(vec_cast,tbl_df.group_bootstraps)
S3method(vec_cast,tbl_df.group_mc_cv)
S3method(vec_cast,tbl_df.group_validation_split)
S3method(vec_cast,tbl_df.group_vfold_cv)
S3method(vec_cast,tbl_df.loo_cv)
S3method(vec_cast,tbl_df.manual_rset)
S3method(vec_cast,tbl_df.mc_cv)
S3method(vec_cast,tbl_df.nested_cv)
S3method(vec_cast,tbl_df.permutations)
S3method(vec_cast,tbl_df.rolling_origin)
S3method(vec_cast,tbl_df.sliding_index)
S3method(vec_cast,tbl_df.sliding_period)
S3method(vec_cast,tbl_df.sliding_window)
S3method(vec_cast,tbl_df.validation_set)
S3method(vec_cast,tbl_df.validation_split)
S3method(vec_cast,tbl_df.validation_time_split)
S3method(vec_cast,tbl_df.vfold_cv)
S3method(vec_cast,validation_set.data.frame)
S3method(vec_cast,validation_set.tbl_df)
S3method(vec_cast,validation_set.validation_set)
S3method(vec_cast,validation_split.data.frame)
S3method(vec_cast,validation_split.tbl_df)
S3method(vec_cast,validation_split.validation_split)
S3method(vec_cast,validation_time_split.data.frame)
S3method(vec_cast,validation_time_split.tbl_df)
S3method(vec_cast,validation_time_split.validation_time_split)
S3method(vec_cast,vfold_cv.data.frame)
S3method(vec_cast,vfold_cv.tbl_df)
S3method(vec_cast,vfold_cv.vfold_cv)
S3method(vec_ptype2,apparent.apparent)
S3method(vec_ptype2,apparent.data.frame)
S3method(vec_ptype2,apparent.tbl_df)
S3method(vec_ptype2,bootstraps.bootstraps)
S3method(vec_ptype2,bootstraps.data.frame)
S3method(vec_ptype2,bootstraps.tbl_df)
S3method(vec_ptype2,clustering_cv.clustering_cv)
S3method(vec_ptype2,clustering_cv.data.frame)
S3method(vec_ptype2,clustering_cv.tbl_df)
S3method(vec_ptype2,data.frame.apparent)
S3method(vec_ptype2,data.frame.bootstraps)
S3method(vec_ptype2,data.frame.clustering_cv)
S3method(vec_ptype2,data.frame.group_bootstraps)
S3method(vec_ptype2,data.frame.group_mc_cv)
S3method(vec_ptype2,data.frame.group_validation_split)
S3method(vec_ptype2,data.frame.group_vfold_cv)
S3method(vec_ptype2,data.frame.loo_cv)
S3method(vec_ptype2,data.frame.manual_rset)
S3method(vec_ptype2,data.frame.mc_cv)
S3method(vec_ptype2,data.frame.nested_cv)
S3method(vec_ptype2,data.frame.permutations)
S3method(vec_ptype2,data.frame.rolling_origin)
S3method(vec_ptype2,data.frame.sliding_index)
S3method(vec_ptype2,data.frame.sliding_period)
S3method(vec_ptype2,data.frame.sliding_window)
S3method(vec_ptype2,data.frame.validation_set)
S3method(vec_ptype2,data.frame.validation_split)
S3method(vec_ptype2,data.frame.validation_time_split)
S3method(vec_ptype2,data.frame.vfold_cv)
S3method(vec_ptype2,group_bootstraps.data.frame)
S3method(vec_ptype2,group_bootstraps.group_bootstraps)
S3method(vec_ptype2,group_bootstraps.tbl_df)
S3method(vec_ptype2,group_mc_cv.data.frame)
S3method(vec_ptype2,group_mc_cv.group_mc_cv)
S3method(vec_ptype2,group_mc_cv.tbl_df)
S3method(vec_ptype2,group_validation_split.data.frame)
S3method(vec_ptype2,group_validation_split.group_validation_split)
S3method(vec_ptype2,group_validation_split.tbl_df)
S3method(vec_ptype2,group_vfold_cv.data.frame)
S3method(vec_ptype2,group_vfold_cv.group_vfold_cv)
S3method(vec_ptype2,group_vfold_cv.tbl_df)
S3method(vec_ptype2,loo_cv.data.frame)
S3method(vec_ptype2,loo_cv.loo_cv)
S3method(vec_ptype2,loo_cv.tbl_df)
S3method(vec_ptype2,manual_rset.data.frame)
S3method(vec_ptype2,manual_rset.manual_rset)
S3method(vec_ptype2,manual_rset.tbl_df)
S3method(vec_ptype2,mc_cv.data.frame)
S3method(vec_ptype2,mc_cv.mc_cv)
S3method(vec_ptype2,mc_cv.tbl_df)
S3method(vec_ptype2,nested_cv.data.frame)
S3method(vec_ptype2,nested_cv.nested_cv)
S3method(vec_ptype2,nested_cv.tbl_df)
S3method(vec_ptype2,permutations.data.frame)
S3method(vec_ptype2,permutations.permutations)
S3method(vec_ptype2,permutations.tbl_df)
S3method(vec_ptype2,rolling_origin.data.frame)
S3method(vec_ptype2,rolling_origin.rolling_origin)
S3method(vec_ptype2,rolling_origin.tbl_df)
S3method(vec_ptype2,sliding_index.data.frame)
S3method(vec_ptype2,sliding_index.sliding_index)
S3method(vec_ptype2,sliding_index.tbl_df)
S3method(vec_ptype2,sliding_period.data.frame)
S3method(vec_ptype2,sliding_period.sliding_period)
S3method(vec_ptype2,sliding_period.tbl_df)
S3method(vec_ptype2,sliding_window.data.frame)
S3method(vec_ptype2,sliding_window.sliding_window)
S3method(vec_ptype2,sliding_window.tbl_df)
S3method(vec_ptype2,tbl_df.apparent)
S3method(vec_ptype2,tbl_df.bootstraps)
S3method(vec_ptype2,tbl_df.clustering_cv)
S3method(vec_ptype2,tbl_df.group_bootstraps)
S3method(vec_ptype2,tbl_df.group_mc_cv)
S3method(vec_ptype2,tbl_df.group_validation_split)
S3method(vec_ptype2,tbl_df.group_vfold_cv)
S3method(vec_ptype2,tbl_df.loo_cv)
S3method(vec_ptype2,tbl_df.manual_rset)
S3method(vec_ptype2,tbl_df.mc_cv)
S3method(vec_ptype2,tbl_df.nested_cv)
S3method(vec_ptype2,tbl_df.permutations)
S3method(vec_ptype2,tbl_df.rolling_origin)
S3method(vec_ptype2,tbl_df.sliding_index)
S3method(vec_ptype2,tbl_df.sliding_period)
S3method(vec_ptype2,tbl_df.sliding_window)
S3method(vec_ptype2,tbl_df.validation_set)
S3method(vec_ptype2,tbl_df.validation_split)
S3method(vec_ptype2,tbl_df.validation_time_split)
S3method(vec_ptype2,tbl_df.vfold_cv)
S3method(vec_ptype2,validation_set.data.frame)
S3method(vec_ptype2,validation_set.tbl_df)
S3method(vec_ptype2,validation_set.validation_set)
S3method(vec_ptype2,validation_split.data.frame)
S3method(vec_ptype2,validation_split.tbl_df)
S3method(vec_ptype2,validation_split.validation_split)
S3method(vec_ptype2,validation_time_split.data.frame)
S3method(vec_ptype2,validation_time_split.tbl_df)
S3method(vec_ptype2,validation_time_split.validation_time_split)
S3method(vec_ptype2,vfold_cv.data.frame)
S3method(vec_ptype2,vfold_cv.tbl_df)
S3method(vec_ptype2,vfold_cv.vfold_cv)
S3method(vec_ptype_abbr,apparent)
S3method(vec_ptype_abbr,bootstraps)
S3method(vec_ptype_abbr,clustering_cv)
S3method(vec_ptype_abbr,group_bootstraps)
S3method(vec_ptype_abbr,group_initial_split)
S3method(vec_ptype_abbr,group_initial_validation_split)
S3method(vec_ptype_abbr,group_mc_cv)
S3method(vec_ptype_abbr,group_validation_split)
S3method(vec_ptype_abbr,group_vfold_cv)
S3method(vec_ptype_abbr,initial_split)
S3method(vec_ptype_abbr,initial_time_split)
S3method(vec_ptype_abbr,initial_validation_split)
S3method(vec_ptype_abbr,manual_rset)
S3method(vec_ptype_abbr,nested_cv)
S3method(vec_ptype_abbr,permutations)
S3method(vec_ptype_abbr,rolling_origin)
S3method(vec_ptype_abbr,sliding_index)
S3method(vec_ptype_abbr,sliding_period)
S3method(vec_ptype_abbr,sliding_window)
S3method(vec_ptype_abbr,validation_set)
S3method(vec_ptype_abbr,validation_split)
S3method(vec_ptype_abbr,validation_time_split)
S3method(vec_ptype_abbr,vfold_cv)
S3method(vec_restore,apparent)
S3method(vec_restore,bootstraps)
S3method(vec_restore,clustering_cv)
S3method(vec_restore,group_bootstraps)
S3method(vec_restore,group_mc_cv)
S3method(vec_restore,group_validation_split)
S3method(vec_restore,group_vfold_cv)
S3method(vec_restore,loo_cv)
S3method(vec_restore,manual_rset)
S3method(vec_restore,mc_cv)
S3method(vec_restore,nested_cv)
S3method(vec_restore,permutations)
S3method(vec_restore,rolling_origin)
S3method(vec_restore,sliding_index)
S3method(vec_restore,sliding_period)
S3method(vec_restore,sliding_window)
S3method(vec_restore,validation_set)
S3method(vec_restore,validation_split)
S3method(vec_restore,validation_time_split)
S3method(vec_restore,vfold_cv)
export(.get_fingerprint)
export(.get_split_args)
export(add_resample_id)
export(all_of)
export(analysis)
export(any_of)
export(apparent)
export(assessment)
export(bootstraps)
export(calibration)
export(caret2rsample)
export(clustering_cv)
export(complement)
export(contains)
export(ends_with)
export(everything)
export(form_pred)
export(get_rsplit)
export(group_bootstraps)
export(group_initial_split)
export(group_initial_validation_split)
export(group_mc_cv)
export(group_validation_split)
export(group_vfold_cv)
export(initial_split)
export(initial_time_split)
export(initial_validation_split)
export(initial_validation_time_split)
export(int_bca)
export(int_pctl)
export(int_t)
export(internal_calibration_split)
export(last_col)
export(loo_cv)
export(make_splits)
export(make_strata)
export(manual_rset)
export(matches)
export(mc_cv)
export(nested_cv)
export(new_rset)
export(num_range)
export(permutations)
export(populate)
export(reg_intervals)
export(reshuffle_rset)
export(reverse_splits)
export(rolling_origin)
export(rsample2caret)
export(rset_reconstruct)
export(sliding_index)
export(sliding_period)
export(sliding_window)
export(starts_with)
export(testing)
export(tidy)
export(training)
export(validation)
export(validation_set)
export(validation_split)
export(validation_time_split)
export(vfold_cv)
import(rlang)
import(vctrs)
importFrom(cli,cli_abort)
importFrom(cli,cli_text)
importFrom(cli,cli_warn)
importFrom(dplyr,arrange)
importFrom(dplyr,arrange_)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,case_when)
importFrom(dplyr,do)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,last)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(dplyr,summarize)
importFrom(dplyr,ungroup)
importFrom(furrr,future_map)
importFrom(generics,tidy)
importFrom(lifecycle,deprecated)
importFrom(methods,formalArgs)
importFrom(pillar,type_sum)
importFrom(purrr,list_rbind)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_dbl)
importFrom(purrr,map_lgl)
importFrom(purrr,pluck)
importFrom(stats,dist)
importFrom(stats,pnorm)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,setNames)
importFrom(stats,terms)
importFrom(tibble,as_tibble)
importFrom(tibble,is_tibble)
importFrom(tibble,obj_sum)
importFrom(tibble,tibble)
importFrom(tidyr,unnest)
importFrom(tidyselect,all_of)
importFrom(tidyselect,any_of)
importFrom(tidyselect,contains)
importFrom(tidyselect,ends_with)
importFrom(tidyselect,everything)
importFrom(tidyselect,last_col)
importFrom(tidyselect,matches)
importFrom(tidyselect,num_range)
importFrom(tidyselect,one_of)
importFrom(tidyselect,starts_with)
importFrom(tidyselect,vars_select)
importFrom(utils,globalVariables)
importFrom(vctrs,vec_count)
importFrom(vctrs,vec_in)
importFrom(vctrs,vec_ptype_abbr)
importFrom(vctrs,vec_slice)
importFrom(vctrs,vec_unique_count)

# rsample <a href='https://rsample.tidymodels.org/'><img src='man/figures/logo.png' align="right" height="139" alt="a boot on a green background" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/tidymodels/rsample/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidymodels/rsample/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/tidymodels/rsample/graph/badge.svg)](https://app.codecov.io/gh/tidymodels/rsample)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/rsample)](https://cran.r-project.org/package=rsample)
[![Downloads](https://cranlogs.r-pkg.org/badges/rsample)](https://cran.r-project.org/package=rsample)
[![lifecycle](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
<!-- badges: end -->

## Overview

The rsample package provides functions to create different types of
resamples and corresponding classes for their analysis. The goal is to
have a modular set of methods that can be used for:

- resampling for estimating the sampling distribution of a statistic
- estimating model performance using a holdout set

The scope of rsample is to provide the basic building blocks for
creating and analyzing resamples of a data set, but this package does
not include code for modeling or calculating statistics. The [Working
with Resample
Sets](https://rsample.tidymodels.org/articles/Working_with_rsets.html)
vignette gives a demonstration of how rsample tools can be used when
building models.

Note that resampled data sets created by rsample are directly accessible
in a resampling object but do not contain much overhead in memory. Since
the original data is not modified, R does not make an automatic copy.

For example, creating 50 bootstraps of a data set does not create an
object that is 50-fold larger in memory:

``` r
library(rsample)
library(mlbench)

data(LetterRecognition)
lobstr::obj_size(LetterRecognition)
#> 2,644,640 B

set.seed(35222)
boots <- bootstraps(LetterRecognition, times = 50)
lobstr::obj_size(boots)
#> 6,686,776 B

# Object size per resample
lobstr::obj_size(boots)/nrow(boots)
#> 133,735.5 B

# Fold increase is <<< 50
as.numeric(lobstr::obj_size(boots)/lobstr::obj_size(LetterRecognition))
#> [1] 2.528426
```

<sup>Created on 2022-02-28 by the [reprex
package](https://reprex.tidyverse.org) (v2.0.1)</sup>

The memory usage for 50 bootstrap samples is less than 3-fold more than
the original data set.

## Installation

To install it, use:

``` r
install.packages("rsample")
```

And the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("rsample")
```

## Contributing

This project is released with a [Contributor Code of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

- For questions and discussions about tidymodels packages, modeling, and
  machine learning, please [post on Posit
  Community](https://forum.posit.co/new-topic?category_id=15&tags=tidymodels,question).

- If you think you have encountered a bug, please [submit an
  issue](https://github.com/tidymodels/rsample/issues).

- Either way, learn how to create and share a
  [reprex](https://reprex.tidyverse.org/articles/articles/learn-reprex.html)
  (a minimal, reproducible example), to clearly communicate about your
  code.

- Check out further details on [contributing guidelines for tidymodels
  packages](https://www.tidymodels.org/contribute/) and [how to get
  help](https://www.tidymodels.org/help/).

# ------------------------------------------------------------------------------
# {{{name}}}

# `vec_restore()`
#
# Called at the end of `vec_slice()` and `vec_ptype()` after all slicing has
# been done on the proxy object.
#
# If no changes have been made to the row/column structure of rset specific
# columns, then we can return an rset subclass. Otherwise, the resulting
# object is no longer guaranteed to return a valid rset, and we have to
# fallback to a bare tibble.
#
# It is very important that the result of `vec_ptype()` is a bare tibble.
# This ensures that the `vec_ptype2.<rset-subclass>.<rset-subclass>()` methods
# never get called. `vec_ptype()` is able to return a bare tibble because it
# essentially takes a 0-row slice of the rset, and then calls `vec_restore()`.
# Because the row structure has been modified, we return a bare tibble from
# `vec_restore.<rset-subclass>()`.
#
# Currently `vec_restore()` uses inheritance in vctrs, but I don't expect this
# to be the case in the future. For that reason, I use explicit methods for
# each individual rset subclass, rather than implementing `vec_restore.rset()`.
#' @export
vec_restore.{{{name}}} <- function(x, to, ...) {
  rset_reconstruct(x, to)
}

# `vec_ptype2()`
#
# The purpose of `vec_ptype2()` methods is generally to determine the type
# of the output in operations like `vec_c()` or `vec_rbind()`. However, this
# implementation does not use any custom `vec_ptype2()` methods at all. This is
# explained below.
#
# `vec_ptype2()` internally works by calling `vec_ptype()` on both `x` and `y`,
# and then looking up the common type of those two ptypes. Generally, the
# ptype of a vector `x` returned from `vec_ptype()` has the same class as `x`.
# However, for rsample objects it makes more sense for the ptype of an rset to
# be a bare tibble. The reason for this is because a ptype of a data frame is
# generally a 0-row slice. However for rsample rsets this doesn't make sense
# (you can't have 0 rows in a 10-fold cv rset), so instead we return a bare
# tibble as the ptype.
# So when `vec_ptype2()` is called on an rset, that rset is downgraded to a
# bare tibble _before_ the search for a `vec_ptype2()` method occurs. This
# means that it will never find a ptype2 method such as
# `vec_ptype2.bootstraps.tbl_df()`, because the <bootstraps> will become a
# tbl_df first. This means that we rely entirely on the `tbl_df` ptype2 methods,
# which are already implemented in vctrs to return another tbl_df.
#
# The implications of this are that whenever a rset subclass is combined with
# another rset subclass or bare tibble, the resulting common type is always
# another bare tibble. So if you `vec_c(<bootstraps>, <tbl_df>)` the result
# will always be a tibble, never a bootstraps object. This makes sense because
# you might be adding rows, which would invalidate the structure of the
# bootstraps object.

#' @export
vec_ptype2.{{{name}}}.{{{name}}} <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.{{{name}}}.{{{name}}}")
}
#' @export
vec_ptype2.{{{name}}}.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.{{{name}}}.tbl_df")
}
#' @export
vec_ptype2.tbl_df.{{{name}}} <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.{{{name}}}")
}
#' @export
vec_ptype2.{{{name}}}.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.{{{name}}}.data.frame")
}
#' @export
vec_ptype2.data.frame.{{{name}}} <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.{{{name}}}")
}

# `vec_cast()`
#
# The `vec_cast()` methods for rset objects really only serve 1 purpose. They
# cast an rset subclass to a tibble or data frame. The cast to tibble is most
# useful. Most of the operations in vctrs work by finding a common type
# with `vec_ptype2()`, and then casting all of the inputs to that common type.
# Because `vec_ptype2()` returns a bare tibble anytime a rset-subclass is
# involved, we will always be casting the rset subclass to a tibble.
# The cast method uses `vctrs::tib_cast()`, which always returns a bare tibble
# with all of the data in `x` cast to the type of `to`.

#' @export
vec_cast.{{{name}}}.{{{name}}} <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.{{{name}}}.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.{{{name}}} <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.{{{name}}}.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.{{{name}}} <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

#' Reverse the analysis and assessment sets
#'
#' This functions "swaps" the analysis and assessment sets of either a single
#' `rsplit` or all `rsplit`s in the `splits` column of an `rset` object.
#'
#' @param x An `rset` or `rsplit` object.
#' @param ... Not currently used.
#'
#' @return An object of the same class as `x`
#'
#' @examples
#' set.seed(123)
#' starting_splits <- vfold_cv(mtcars, v = 3)
#' reverse_splits(starting_splits)
#' reverse_splits(starting_splits$splits[[1]])
#'
#' @rdname reverse_splits
#' @export
reverse_splits <- function(x, ...) {
  UseMethod("reverse_splits")
}

#' @rdname reverse_splits
#' @export
reverse_splits.default <- function(x, ...) {
  cli_abort(
    "{.arg x} must be either an {.cls rsplit} or an {.cls rset} object."
  )
}

#' @rdname reverse_splits
#' @export
reverse_splits.permutations <- function(x, ...) {
  cli_abort(
    "Permutations cannot have their splits reversed."
  )
}

#' @rdname reverse_splits
#' @export
reverse_splits.perm_split <- reverse_splits.permutations

#' @rdname reverse_splits
#' @export
reverse_splits.rsplit <- function(x, ...) {
  rlang::check_dots_empty()

  out_splits <- list(
    analysis = as.integer(x, data = "assessment"),
    assessment = as.integer(x, data = "analysis")
  )
  out_splits <- make_splits(out_splits, x$data)
  class(out_splits) <- class(x)
  out_splits
}

#' @rdname reverse_splits
#' @export
reverse_splits.rset <- function(x, ...) {
  rlang::check_dots_empty()

  x$splits <- purrr::map(x$splits, reverse_splits)

  x
}

#' Simple Training/Test Set Splitting
#'
#' `initial_split()` creates a single binary split of the data into a training
#'  set and testing set. `initial_time_split()` does the same, but takes the
#'  _first_ `prop` samples for training, instead of a random selection.
#'  `group_initial_split()` creates splits of the data based
#'  on some grouping variable, so that all data in a "group" is assigned to
#'  the same split.
#'
#' @details `training()` and `testing()` are used to extract the resulting data.
#'
#' To avoid data leakage when using lagged variables, lag the predictors before
#' the initial split.
#'
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @export
#' @return An `rsplit` object that can be used with the `training()` and `testing()`
#'  functions to extract the data in each split.
#' @examplesIf rlang::is_installed("modeldata")
#' set.seed(1353)
#' car_split <- initial_split(mtcars)
#' train_data <- training(car_split)
#' test_data <- testing(car_split)
#'
#' data(drinks, package = "modeldata")
#' drinks_split <- initial_time_split(drinks)
#' train_data <- training(drinks_split)
#' test_data <- testing(drinks_split)
#' c(max(train_data$date), min(test_data$date))
#'
#' set.seed(1353)
#' car_split <- group_initial_split(mtcars, cyl)
#' train_data <- training(car_split)
#' test_data <- testing(car_split)
#'
#' @export
#'
initial_split <- function(
  data,
  prop = 3 / 4,
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  ...
) {
  check_dots_empty()
  check_prop(prop)

  res <-
    mc_cv(
      data = data,
      prop = prop,
      strata = {{ strata }},
      breaks = breaks,
      pool = pool,
      times = 1
    )
  attrib <- .get_split_args(res, allow_strata_false = TRUE)

  res <- res$splits[[1]]

  attrib$times <- NULL
  for (i in names(attrib)) {
    attr(res, i) <- attrib[[i]]
  }

  class(res) <- c("initial_split", class(res))

  res
}

#' @rdname initial_split
#' @param lag `r lifecycle::badge("deprecated")` This is deprecated, please lag
#' your predictors prior to splitting the dataset.
#' @export
initial_time_split <- function(
  data,
  prop = 3 / 4,
  lag = lifecycle::deprecated(),
  ...
) {
  check_dots_empty()
  check_prop(prop)

  if (lifecycle::is_present(lag)) {
    lifecycle::deprecate_soft(
      when = "1.3.1.9000",
      what = "initial_time_split(lag)",
      details = "Please lag your predictors prior to splitting the dataset."
    )
  } else {
    lag <- 0
  }

  if (!is.numeric(lag) | !(lag %% 1 == 0)) {
    cli_abort("{.arg lag} must be a whole number.")
  }

  n_train <- floor(nrow(data) * prop)

  if (lag > n_train) {
    cli_abort(
      "{.arg lag} must be less than or equal to the number of training observations."
    )
  }

  split <- rsplit(data, 1:n_train, (n_train + 1 - lag):nrow(data))
  splits <- list(split)
  ids <- "Resample1"
  rset <- new_rset(splits, ids)

  res <- rset$splits[[1]]

  attrib <- list(
    prop = prop,
    lag = lag
  )
  for (i in names(attrib)) {
    attr(res, i) <- attrib[[i]]
  }

  class(res) <- c("initial_time_split", "initial_split", class(res))
  res
}

#' @rdname initial_split
#' @export
#' @param x An `rsplit` object produced by `initial_split()` or
#'  `initial_time_split()`.
training <- function(x, ...) {
  UseMethod("training")
}

#' @export
#' @rdname initial_split
training.default <- function(x, ...) {
  cls <- class(x)
  cli::cli_abort(
    "No method for objects of class{?es}: {cls}"
  )
}

#' @rdname initial_split
#' @export
training.rsplit <- function(x, ...) {
  analysis(x, ...)
}

#' @rdname initial_split
#' @export
testing <- function(x, ...) {
  UseMethod("testing")
}

#' @export
#' @rdname initial_split
testing.default <- function(x, ...) {
  cls <- class(x)
  cli::cli_abort(
    "No method for objects of class{?es}: {cls}"
  )
}

#' @rdname initial_split
#' @export
testing.rsplit <- function(x, ...) {
  assessment(x, ...)
}

#' @inheritParams make_groups
#' @rdname initial_split
#' @export
group_initial_split <- function(
  data,
  group,
  prop = 3 / 4,
  ...,
  strata = NULL,
  pool = 0.1
) {
  check_dots_empty()
  check_prop(prop)

  if (missing(strata)) {
    res <- group_mc_cv(
      data = data,
      group = {{ group }},
      prop = prop,
      times = 1
    )
  } else {
    res <- group_mc_cv(
      data = data,
      group = {{ group }},
      prop = prop,
      times = 1,
      strata = {{ strata }},
      pool = pool
    )
  }

  attrib <- .get_split_args(res, allow_strata_false = TRUE)

  res <- res$splits[[1]]

  attrib$times <- NULL
  for (i in names(attrib)) {
    attr(res, i) <- attrib[[i]]
  }
  class(res) <- c("group_initial_split", "initial_split", class(res))

  res
}

test_that("informative error messages", {
  expect_snapshot(error = TRUE, {
    internal_calibration_split(NULL)
  })

  # we don't know how manual rsets are constructed,
  # thus we can't provide an internal_calibration_split method
  df <- data.frame(x = c(1, 2, 3, 4, 5, 6))
  indices <- list(
    list(analysis = c(1L, 2L), assessment = 3L),
    list(analysis = c(4L, 5L), assessment = 6L)
  )
  splits <- lapply(indices, make_splits, data = df)
  m_rset <- manual_rset(splits, c("Split 1", "Split 2"))

  expect_snapshot(error = TRUE, {
    internal_calibration_split(m_rset)
  })
})

# mc ---------------------------------------------------------------------

test_that("mc_split", {
  set.seed(11)
  r_set <- mc_cv(warpbreaks)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c("mc_split_cal", "internal_calibration_split", "mc_split", "rsplit"),
    exact = TRUE
  )
})

test_that("mc_split can create mock split", {
  dat <- data.frame(x = 1:2, y = 1:2)

  set.seed(11)
  r_set <- mc_cv(dat, prop = 1 / 2, times = 1)
  split_args <- .get_split_args(r_set)
  # analysis set only contains 1 row, thus can't split further
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c("mc_split_cal", "internal_calibration_split", "mc_split", "rsplit"),
    exact = TRUE
  )
})

test_that("group_mc_split", {
  skip_if_not_installed("modeldata")

  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(11)
  r_set <- group_mc_cv(ames, "MS_SubClass")
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "group_mc_split_cal",
      "internal_calibration_split",
      "group_mc_split",
      "mc_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("group_mc_split can create mock split", {
  dat <- data.frame(x = 1:2, y = 1:2, group = c("A", "B"))

  set.seed(11)
  r_set <- group_mc_cv(dat, group = "group", prop = 1 / 2, times = 1)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "group_mc_split_cal",
      "internal_calibration_split",
      "group_mc_split",
      "mc_split",
      "rsplit"
    ),
    exact = TRUE
  )
})


# vfold ------------------------------------------------------------------

test_that("vfold_split", {
  set.seed(11)
  r_set <- vfold_cv(warpbreaks, v = 5)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "vfold_split_cal",
      "internal_calibration_split",
      "vfold_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("vfold_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3)

  set.seed(11)
  r_set <- vfold_cv(dat, v = 2)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "vfold_split_cal",
      "internal_calibration_split",
      "vfold_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("group_vfold_split", {
  skip_if_not_installed("modeldata")

  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(11)
  r_set <- group_vfold_cv(ames, "MS_SubClass")
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "group_vfold_split_cal",
      "internal_calibration_split",
      "group_vfold_split",
      "vfold_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("group_vfold_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3, group = c("A", "B", "B"))

  set.seed(11)
  r_set <- group_vfold_cv(dat, group = "group", v = 2)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "group_vfold_split_cal",
      "internal_calibration_split",
      "group_vfold_split",
      "vfold_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

# bootstrap --------------------------------------------------------------

test_that("boot_split", {
  set.seed(11)
  r_set <- bootstraps(warpbreaks, times = 2)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_lte(
    nrow(split_cal$data),
    analysis(r_split) |> nrow()
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[complement(split_cal), ],
    ignore_attr = "row.names"
  )
  expect_s3_class(
    split_cal,
    c("boot_split_cal", "internal_calibration_split", "boot_split", "rsplit"),
    exact = TRUE
  )
})

test_that("boot_split can create mock split", {
  dat <- data.frame(x = 1, y = 1)

  set.seed(11)
  r_set <- bootstraps(dat, times = 1) |> suppressWarnings()
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c("boot_split_cal", "internal_calibration_split", "boot_split", "rsplit"),
    exact = TRUE
  )
})

test_that("group_boot_split", {
  skip_if_not_installed("modeldata")

  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(11)
  r_set <- group_bootstraps(ames, group = "MS_SubClass", times = 2)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_lte(
    nrow(split_cal$data),
    analysis(r_split) |> nrow()
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[complement(split_cal), ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "group_boot_split_cal",
      "internal_calibration_split",
      "group_boot_split",
      "boot_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("group_boot_split can create mock split", {
  dat <- data.frame(x = 1:2, y = 1:2, group = c("A", "B"))

  set.seed(11)
  r_set <- group_bootstraps(dat, group = "group", times = 1)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "group_boot_split_cal",
      "internal_calibration_split",
      "group_boot_split",
      "boot_split",
      "rsplit"
    ),
    exact = TRUE
  )
})


# validation set ---------------------------------------------------------

test_that("initial_validation_split", {
  set.seed(11)
  initial_vsplit <- initial_validation_split(
    warpbreaks,
    prop = c(0.6, 0.2)
  )
  r_set <- validation_set(initial_vsplit)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c("val_split_cal", "internal_calibration_split", "val_split", "rsplit"),
    exact = TRUE
  )
})

test_that("val_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3)

  set.seed(11)
  initial_vsplit <- initial_validation_split(dat, prop = c(0.4, 0.33))
  r_set <- validation_set(initial_vsplit)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c("val_split_cal", "internal_calibration_split", "val_split", "rsplit"),
    exact = TRUE
  )
})

test_that("group_initial_validation_split", {
  skip_if_not_installed("modeldata")

  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(11)
  initial_vsplit <- group_initial_validation_split(
    ames,
    group = "MS_SubClass",
    prop = c(0.7, 0.2)
  )
  r_set <- validation_set(initial_vsplit)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "group_val_split_cal",
      "internal_calibration_split",
      "group_val_split",
      "val_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("group_val_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3, group = c("A", "B", "C"))

  set.seed(11)
  initial_vsplit <- group_initial_validation_split(
    dat,
    group = "group",
    prop = c(0.4, 0.33)
  )
  r_set <- validation_set(initial_vsplit)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "group_val_split_cal",
      "internal_calibration_split",
      "group_val_split",
      "val_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("initial_validation_time_split", {
  set.seed(11)
  initial_vsplit <- initial_validation_time_split(
    warpbreaks,
    prop = c(0.6, 0.2)
  )
  r_set <- validation_set(initial_vsplit)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "time_val_split_cal",
      "internal_calibration_split",
      "time_val_split",
      "val_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("time_val_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3)

  set.seed(11)
  initial_vsplit <- initial_validation_time_split(dat, prop = c(0.4, 0.33))
  r_set <- validation_set(initial_vsplit)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "time_val_split_cal",
      "internal_calibration_split",
      "time_val_split",
      "val_split",
      "rsplit"
    ),
    exact = TRUE
  )
})


# clustering -------------------------------------------------------------

test_that("clustering_split", {
  set.seed(11)
  r_set <- clustering_cv(warpbreaks, vars = breaks, v = 5)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[-split_cal$in_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "clustering_split_cal",
      "internal_calibration_split",
      "clustering_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("clustering_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3)

  set.seed(11)
  r_set <- clustering_cv(dat, vars = x, v = 2)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "clustering_split_cal",
      "internal_calibration_split",
      "clustering_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

# apparent ---------------------------------------------------------------

test_that("apparent_split", {
  set.seed(11)
  r_set <- apparent(warpbreaks)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split)

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )
  expect_identical(
    calibration(split_cal),
    analysis(r_split)
  )

  expect_s3_class(
    split_cal,
    c(
      "apparent_split_cal",
      "internal_calibration_split",
      "apparent_split",
      "rsplit"
    ),
    exact = TRUE
  )
})


# slide ------------------------------------------------------------------

test_that("sliding_window_split", {
  df <- data.frame(x = 1:10)
  r_set <- sliding_window(df, lookback = 4, assess_start = 3, assess_stop = 5)

  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$in_id,
    1:3
  )
  expect_identical(
    split_cal$out_id,
    4:5
  )

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, , drop = FALSE],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, , drop = FALSE],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "sliding_window_split_cal",
      "internal_calibration_split",
      "sliding_window_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("sliding_window_split needs at least 2 observations", {
  df <- data.frame(x = 1:10)
  r_set <- sliding_window(df, lookback = 0)

  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "sliding_window_split_cal",
      "internal_calibration_split",
      "sliding_window_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("sliding_window_split with incomplete sets", {
  df <- data.frame(x = 1:10)
  r_set <- sliding_window(
    df,
    lookback = 4,
    assess_start = 3,
    assess_stop = 5,
    complete = FALSE
  )

  split_args <- .get_split_args(r_set)

  # not enough observations for a full calibration set and (inner) analysis set
  r_split <- get_rsplit(r_set, 2)
  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  r_split <- get_rsplit(r_set, 3)
  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$in_id,
    1L
  )
  expect_identical(
    split_cal$out_id,
    2:3
  )

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, , drop = FALSE],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, , drop = FALSE],
    ignore_attr = "row.names"
  )
})

test_that("sliding_index_split", {
  df <- data.frame(x = 1:10)
  r_set <- sliding_index(df, x, lookback = 4, assess_start = 3, assess_stop = 5)

  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$in_id,
    1:3
  )
  expect_identical(
    split_cal$out_id,
    4:5
  )

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, , drop = FALSE],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, , drop = FALSE],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "sliding_index_split_cal",
      "internal_calibration_split",
      "sliding_index_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("sliding_index_split can lookback over irregular index", {
  df <- data.frame(x = c(1, 3, 4, 5, 6:10))
  r_set <- sliding_index(df, x, lookback = 4, assess_start = 3, assess_stop = 5)

  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$in_id,
    1:2
  )
  expect_identical(
    split_cal$out_id,
    3:4
  )
})

test_that("sliding_index_split can make calibration set relative to irregular index", {
  df <- data.frame(x = c(1:7, 9:20))
  r_set <- sliding_index(
    df,
    x,
    lookback = 9,
    assess_start = 3,
    assess_stop = 10
  )

  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$in_id,
    1:5
  )
  expect_identical(
    split_cal$out_id,
    7:9
  )
})

test_that("sliding_index_split needs at least 2 observations", {
  df <- data.frame(x = 1:10)
  r_set <- sliding_index(df, x, lookback = 0)

  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "sliding_index_split_cal",
      "internal_calibration_split",
      "sliding_index_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("sliding_index_split with incomplete sets", {
  df <- data.frame(x = 1:10)
  r_set <- sliding_index(
    df,
    x,
    lookback = 4,
    assess_start = 3,
    assess_stop = 5,
    complete = FALSE
  )

  split_args <- .get_split_args(r_set)

  # not enough observations for a full calibration set and (inner) analysis set
  r_split <- get_rsplit(r_set, 2)
  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  r_split <- get_rsplit(r_set, 3)
  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$in_id,
    1L
  )
  expect_identical(
    split_cal$out_id,
    2:3
  )

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, , drop = FALSE],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, , drop = FALSE],
    ignore_attr = "row.names"
  )
})

test_that("sliding_period_split", {
  index <- vctrs::new_date(c(-18, -14, -2, 0, 1, 3:10, 31))
  df <- data.frame(index = index)

  r_set <- sliding_period(df, index, period = "month", lookback = 1)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(
    split_cal$in_id,
    1:3
  )
  expect_identical(
    split_cal$out_id,
    4:13
  )

  expect_identical(
    split_cal$data,
    analysis(r_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, , drop = FALSE],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, , drop = FALSE],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "sliding_period_split_cal",
      "internal_calibration_split",
      "sliding_period_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("sliding_period_split when looking back over multiple periods, only complete ones are used", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31))
  df <- data.frame(index = index)

  r_set <- sliding_period(df, index, period = "month", lookback = 1)

  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  split_cal <- internal_calibration_split(r_split, split_args)

  expect_identical(split_cal$in_id, 1L)
  expect_identical(split_cal$out_id, 2L)
})

test_that("sliding_period_split needs at least 2 observations", {
  index <- vctrs::new_date(c(-60, 0, 32))
  df <- data.frame(index = index)

  r_set <- sliding_period(df, index, period = "month", lookback = 1)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "sliding_period_split_cal",
      "internal_calibration_split",
      "sliding_period_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("sliding_period_split needs observations in at least 2 periods", {
  index <- vctrs::new_date(c(-60, 0, 1, 2, 32))
  df <- data.frame(index = index)

  r_set <- sliding_period(df, index, period = "month", lookback = 1)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "sliding_period_split_cal",
      "internal_calibration_split",
      "sliding_period_split",
      "rsplit"
    ),
    exact = TRUE
  )

  index <- vctrs::new_date(c(0, 1, 2, 32))
  df <- data.frame(index = index)

  r_set <- sliding_period(df, index, period = "month", lookback = 0)
  split_args <- .get_split_args(r_set)
  r_split <- get_rsplit(r_set, 1)

  expect_snapshot({
    split_cal <- internal_calibration_split(r_split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    analysis(r_split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "sliding_period_split_cal",
      "internal_calibration_split",
      "sliding_period_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

# initial split ----------------------------------------------------------

test_that("initial_split", {
  set.seed(11)
  car_split <- initial_split(mtcars)
  split_cal <- internal_calibration_split(car_split, .get_split_args(car_split))

  expect_identical(
    split_cal$data,
    training(car_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "mc_split_cal",
      "internal_calibration_split",
      "initial_split",
      "mc_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("group_initial_split", {
  set.seed(11)
  car_split <- group_initial_split(mtcars, cyl)
  split_cal <- internal_calibration_split(car_split, .get_split_args(car_split))

  expect_identical(
    split_cal$data,
    training(car_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "group_mc_split_cal",
      "internal_calibration_split",
      "group_initial_split",
      "initial_split",
      "group_mc_split",
      "mc_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("initial_time_split", {
  set.seed(11)
  car_split <- initial_time_split(mtcars)
  split_cal <- internal_calibration_split(car_split, .get_split_args(car_split))

  expect_identical(
    split_cal$data,
    training(car_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "initial_time_split_cal",
      "internal_calibration_split",
      "initial_time_split",
      "initial_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("initial_time_split can create mock split", {
  dat <- data.frame(x = 1:2, y = 1:2)

  split <- initial_time_split(dat, prop = 0.5)
  split_args <- .get_split_args(split)

  expect_snapshot({
    split_cal <- internal_calibration_split(split, split_args)
  })

  expect_identical(
    analysis(split_cal),
    training(split)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "initial_time_split_cal",
      "internal_calibration_split",
      "initial_time_split",
      "initial_split",
      "rsplit"
    ),
    exact = TRUE
  )
})


# initial validation split -----------------------------------------------

test_that("initial_validation_split", {
  set.seed(11)
  car_split <- initial_validation_split(mtcars)
  split_cal <- internal_calibration_split(car_split, .get_split_args(car_split))

  expect_identical(
    split_cal$data,
    training(car_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "initial_validation_split_cal",
      "internal_calibration_split",
      "mc_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("initial_validation_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3)

  set.seed(11)
  initial_vsplit <- initial_validation_split(dat, prop = c(0.4, 0.33))
  split_args <- .get_split_args(initial_vsplit)

  expect_snapshot({
    split_cal <- internal_calibration_split(initial_vsplit, split_args)
  })

  expect_identical(
    analysis(split_cal),
    training(initial_vsplit)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "initial_validation_split_cal",
      "internal_calibration_split",
      "mc_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("group_initial_validation_split", {
  skip_if_not_installed("modeldata")
  data(ames, package = "modeldata", envir = rlang::current_env())

  set.seed(12)
  ames_split <- group_initial_validation_split(ames, group = Neighborhood)
  split_cal <- internal_calibration_split(
    ames_split,
    .get_split_args(ames_split)
  )

  expect_identical(
    split_cal$data,
    training(ames_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "group_initial_validation_split_cal",
      "internal_calibration_split",
      "group_mc_split",
      "mc_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("group_initial_validation_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3, group = c("A", "B", "C"))

  set.seed(11)
  initial_vsplit <- group_initial_validation_split(
    dat,
    group = "group",
    prop = c(0.4, 0.33)
  )
  split_args <- .get_split_args(initial_vsplit)

  expect_snapshot({
    split_cal <- internal_calibration_split(initial_vsplit, split_args)
  })

  expect_identical(
    analysis(split_cal),
    training(initial_vsplit)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "group_initial_validation_split_cal",
      "internal_calibration_split",
      "group_mc_split",
      "mc_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("initial_validation_time_split", {
  skip_if_not_installed("modeldata")
  data(drinks, package = "modeldata", envir = rlang::current_env())

  set.seed(12)
  drinks_split <- initial_validation_time_split(drinks)
  split_cal <- internal_calibration_split(
    drinks_split,
    .get_split_args(drinks_split)
  )

  expect_identical(
    split_cal$data,
    training(drinks_split)
  )

  expect_identical(
    analysis(split_cal),
    split_cal$data[split_cal$in_id, ],
    ignore_attr = "row.names"
  )
  expect_identical(
    calibration(split_cal),
    split_cal$data[split_cal$out_id, ],
    ignore_attr = "row.names"
  )

  expect_s3_class(
    split_cal,
    c(
      "initial_validation_time_split_cal",
      "internal_calibration_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

test_that("initial_validation_time_split can create mock split", {
  dat <- data.frame(x = 1:3, y = 1:3)

  set.seed(11)
  initial_vsplit <- initial_validation_time_split(dat, prop = c(0.4, 0.33))
  split_args <- .get_split_args(initial_vsplit)

  expect_snapshot({
    split_cal <- internal_calibration_split(initial_vsplit, split_args)
  })

  expect_identical(
    analysis(split_cal),
    training(initial_vsplit)
  )

  expect_identical(
    nrow(calibration(split_cal)),
    0L
  )

  expect_s3_class(
    split_cal,
    c(
      "initial_validation_time_split_cal",
      "internal_calibration_split",
      "rsplit"
    ),
    exact = TRUE
  )
})

# mock split -------------------------------------------------------------

test_that("can create a mock split", {
  mock_split <- internal_calibration_split_mock(
    mtcars,
    class = "internal_calibration_split"
  )
  mock_analysis <- analysis(mock_split)
  mock_calibration <- calibration(mock_split)

  expect_identical(mock_analysis, mtcars)
  expect_identical(nrow(mock_calibration), 0L)

  expect_s3_class(mock_split, "rsplit")

  mock_split <- internal_calibration_split_mock(mtcars, class = "mock_class")
  expect_s3_class(mock_split, "mock_class")
})

#   ----------------------------------------------------------------------

test_that("assessment() fails", {
  internal_calibration_split <- internal_calibration_split_mock(
    mtcars,
    class = "internal_calibration_split"
  )

  expect_snapshot(error = TRUE, {
    assessment(internal_calibration_split)
  })
})

test_that("print()", {
  internal_calibration_split <- internal_calibration_split_mock(
    mtcars,
    class = "internal_calibration_split"
  )

  expect_snapshot(
    print(internal_calibration_split)
  )
})

test_that("basic set", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))

  set.seed(11)
  initial_val_split <- initial_validation_split(dat1, prop = c(0.6, 0.2))
  val_set <- validation_set(initial_val_split)
  train_and_val <- dplyr::bind_rows(
    training(initial_val_split),
    validation(initial_val_split)
  )

  expect_s3_class(val_set, "rset")
  expect_s3_class(val_set, "validation_set")

  set_sizes <- dim_rset(val_set)
  expect_true(all(set_sizes$analysis == 12))
  expect_true(all(set_sizes$assessment == 4))

  expect_equal(length(val_set$splits), 1)
  expect_equal(
    val_set$splits[[1]]$data,
    train_and_val,
    ignore_attr = "row.names"
  )
})

test_that("accessor functions for `val_split`", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))

  set.seed(11)
  initial_val_split <- initial_validation_split(dat1, prop = c(0.6, 0.2))
  val_split <- validation_set(initial_val_split) |> get_rsplit(1)

  expect_equal(analysis(val_split), training(initial_val_split))
  expect_equal(training(val_split), training(initial_val_split))
  expect_equal(assessment(val_split), validation(initial_val_split))
  expect_equal(validation(val_split), validation(initial_val_split))
  expect_snapshot(error = TRUE, testing(val_split))
})


test_that("working with Surv objects - issue #443", {
  srv <-
    list(
      age = c(74, 68, 56, 57, 60, 74, 76, 77, 39, 75, 66, 58),
      sex = c(1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 2),
      surv_obj = structure(
        c(
          306,
          455,
          1010,
          210,
          883,
          1022,
          116,
          188,
          191,
          105,
          174,
          177,
          1,
          1,
          0,
          1,
          1,
          0,
          1,
          0,
          0,
          0,
          0,
          0
        ),
        dim = c(12L, 2L),
        dimnames = list(NULL, c("time", "status")),
        type = "right",
        class = "Surv"
      )
    )
  surv_df <-
    structure(
      srv,
      row.names = paste(1:12),
      class = "data.frame"
    )

  surv_tbl <- dplyr::as_tibble(surv_df)

  # data frame input

  set.seed(472)
  surv_split_df <- initial_validation_split(surv_df, prop = c(.3, .3))

  expect_s3_class(surv_split_df$data$surv_obj, "Surv")

  expect_s3_class(training(surv_split_df)$surv_obj, "Surv")
  expect_s3_class(testing(surv_split_df)$surv_obj, "Surv")
  expect_s3_class(validation(surv_split_df)$surv_obj, "Surv")

  surv_rs_df <- validation_set(surv_split_df)
  expect_s3_class(surv_rs_df$splits[[1]]$data$surv_obj, "Surv")

  expect_s3_class(training(surv_rs_df$splits[[1]])$surv_obj, "Surv")
  expect_s3_class(validation(surv_rs_df$splits[[1]])$surv_obj, "Surv")
  expect_s3_class(analysis(surv_rs_df$splits[[1]])$surv_obj, "Surv")
  expect_s3_class(assessment(surv_rs_df$splits[[1]])$surv_obj, "Surv")

  # tibble input

  set.seed(472)
  surv_split_tbl <- initial_validation_split(surv_tbl, prop = c(.3, .3))

  expect_s3_class(surv_split_tbl$data$surv_obj, "Surv")

  expect_s3_class(training(surv_split_tbl)$surv_obj, "Surv")
  expect_s3_class(testing(surv_split_tbl)$surv_obj, "Surv")
  expect_s3_class(validation(surv_split_tbl)$surv_obj, "Surv")

  surv_rs_tbl <- validation_set(surv_split_tbl)
  expect_s3_class(surv_rs_tbl$splits[[1]]$data$surv_obj, "Surv")

  expect_s3_class(training(surv_rs_tbl$splits[[1]])$surv_obj, "Surv")
  expect_s3_class(validation(surv_rs_tbl$splits[[1]])$surv_obj, "Surv")
  expect_s3_class(analysis(surv_rs_tbl$splits[[1]])$surv_obj, "Surv")
  expect_s3_class(assessment(surv_rs_tbl$splits[[1]])$surv_obj, "Surv")
})

test_that("default param", {
  set.seed(11)
  rs1 <- initial_split(dat1)
  expect_equal(class(rs1), c("initial_split", "mc_split", "rsplit"))
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), nrow(dat1) * 3 / 4)
  expect_equal(nrow(ts1), nrow(dat1) / 4)
})

test_that("default time param", {
  rs1 <- initial_time_split(dat1)
  expect_equal(class(rs1), c("initial_split", "rsplit"))
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3 / 4))
  expect_equal(nrow(ts1), ceiling(nrow(dat1) / 4))
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3 / 4)))
})

test_that("default time param with lag", {
  rs1 <- initial_time_split(dat1, lag = 5)
  expect_equal(class(rs1), c("initial_split", "rsplit"))
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3 / 4))
  expect_equal(nrow(ts1), ceiling(nrow(dat1) / 4) + 5)
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3 / 4)))
  expect_equal(ts1, dat1[(floor(nrow(dat1) * 3 / 4) + 1 - 5):nrow(dat1), ])

  data(drinks, package = "modeldata")

  # Whole numbers only
  expect_snapshot(
    initial_time_split(drinks, lag = 12.5),
    error = TRUE
  )
  # Lag must be less than number of training observations
  expect_snapshot(
    initial_time_split(drinks, lag = 500),
    error = TRUE
  )
})

test_that("`prop` computes the proportion for analysis (#217)", {
  set.seed(11)

  props <- c(.1, .9)

  for (prop in props) {
    # Not stratified
    split <- initial_split(airquality, prop = prop)
    actual <- nrow(analysis(split))
    expect <- as.integer(floor(nrow(airquality) * prop))

    expect_identical(actual, expect)

    # Stratified
    split <- initial_split(airquality, prop = prop, strata = Month)
    actual <- nrow(analysis(split))
    expect <- as.integer(sum(floor(table(airquality$Month) * prop)))

    expect_identical(actual, expect)
  }
})


test_that("printing initial split objects", {
  expect_snapshot(initial_split(mtcars))
  expect_snapshot(initial_time_split(mtcars))
})

test_that("default param", {
  set.seed(11)
  rs1 <- initial_split(dat1)
  expect_equal(class(rs1), c("initial_split", "mc_split", "rsplit"))
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), nrow(dat1) * 3 / 4)
  expect_equal(nrow(ts1), nrow(dat1) / 4)
})

test_that("default time param", {
  rs1 <- initial_time_split(dat1)
  expect_equal(class(rs1), c("initial_time_split", "initial_split", "rsplit"))
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3 / 4))
  expect_equal(nrow(ts1), ceiling(nrow(dat1) / 4))
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3 / 4)))
})

test_that("`lag` arg to `initial_time_split()` is deprecated", {
  expect_snapshot({
    initial_time_split(mtcars, lag = 2)
  })
})

test_that("default time param with lag", {
  withr::local_options(lifecycle_verbosity = "quiet")

  rs1 <- initial_time_split(dat1, lag = 5)
  expect_equal(class(rs1), c("initial_time_split", "initial_split", "rsplit"))
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3 / 4))
  expect_equal(nrow(ts1), ceiling(nrow(dat1) / 4) + 5)
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3 / 4)))
  expect_equal(
    ts1,
    dat1[(floor(nrow(dat1) * 3 / 4) + 1 - 5):nrow(dat1), ],
    ignore_attr = "row.names"
  )
})

test_that("`initial_time_split()` error messages", {
  skip_if_not_installed("modeldata")
  data(drinks, package = "modeldata", envir = rlang::current_env())

  expect_snapshot(error = TRUE, {
    initial_time_split(drinks, prop = 2)
  })

  withr::local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(error = TRUE, {
    initial_time_split(drinks, lag = 12.5)
  })

  expect_snapshot(error = TRUE, {
    initial_time_split(drinks, lag = nrow(drinks) + 1)
  })
})

test_that("default group param", {
  rs1 <- group_initial_split(dat1, c)
  expect_equal(
    class(rs1),
    c(
      "group_initial_split",
      "initial_split",
      "group_mc_split",
      "mc_split",
      "rsplit"
    )
  )
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), nrow(dat1) * 3 / 4)
  expect_equal(nrow(ts1), nrow(dat1) / 4)
})

test_that("`prop` computes the proportion for analysis (#217)", {
  set.seed(11)

  props <- c(.1, .9)

  for (prop in props) {
    # Not stratified
    split <- initial_split(airquality, prop = prop)
    actual <- nrow(analysis(split))
    expect <- as.integer(floor(nrow(airquality) * prop))

    expect_identical(actual, expect)

    # Stratified
    split <- initial_split(airquality, prop = prop, strata = Month)
    actual <- nrow(analysis(split))
    expect <- as.integer(sum(floor(table(airquality$Month) * prop)))

    expect_identical(actual, expect)
  }
})

test_that("grouping -- strata", {
  set.seed(11)

  n_common_class <- 70
  n_rare_class <- 30

  group_table <- tibble(
    group = 1:100,
    outcome = sample(c(rep(0, n_common_class), rep(1, n_rare_class)))
  )
  observation_table <- tibble(
    group = sample(1:100, 5e4, replace = TRUE),
    observation = 1:5e4
  )
  sample_data <- dplyr::full_join(
    group_table,
    observation_table,
    by = "group",
    multiple = "all"
  )
  rs4 <- group_initial_split(sample_data, group, strata = outcome)
  expect_equal(mean(as.data.frame(rs4)$outcome == 1), 0.3, tolerance = 1e-2)

  expect_identical(length(intersect(rs4$in_ind, rs4$out_id)), 0L)
})

test_that("`prop` computes the proportion for group analysis", {
  rs1 <- group_initial_split(dat1, c, prop = 1 / 2)
  expect_equal(
    class(rs1),
    c(
      "group_initial_split",
      "initial_split",
      "group_mc_split",
      "mc_split",
      "rsplit"
    )
  )
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), nrow(dat1) * 1 / 2)
  expect_equal(nrow(ts1), nrow(dat1) / 2)
  expect_equal(nrow(tr1), nrow(ts1))
})

test_that("printing initial split objects", {
  expect_snapshot(initial_split(mtcars))
  expect_snapshot(initial_time_split(mtcars))
})

test_that("prop is checked", {
  expect_snapshot(error = TRUE, {
    initial_split(mtcars, prop = 1)
  })
  expect_snapshot(error = TRUE, {
    initial_time_split(mtcars, prop = 1)
  })
  expect_snapshot(error = TRUE, {
    group_initial_split(mtcars, group = "cyl", prop = 1)
  })
})

context("Initial splitting")

library(testthat)
library(rsample)
library(purrr)

dat1 <- data.frame(a = 1:20, b = letters[1:20])

test_that('default param', {
  set.seed(11)
  rs1 <- initial_split(dat1)
  expect_equal(class(rs1), c("rsplit", "mc_split"))
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), nrow(dat1)*3/4)
  expect_equal(nrow(ts1), nrow(dat1)/4)
})

test_that('default time param', {
  rs1 <- initial_time_split(dat1)
  expect_equal(class(rs1), "rsplit")
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3/4))
  expect_equal(nrow(ts1), ceiling(nrow(dat1) / 4))
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3/4)))
})

test_that('default time param with lag', {
  rs1 <- initial_time_split(dat1, lag = 5)
  expect_equal(class(rs1), "rsplit")
  tr1 <- training(rs1)
  ts1 <- testing(rs1)
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3/4))
  expect_equal(nrow(ts1), ceiling(nrow(dat1) / 4) + 5)
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3/4)) )
  expect_equal(ts1, dat1[(floor(nrow(dat1) * 3/4) + 1 - 5):nrow(dat1),] )

  expect_error(initial_time_split(drinks, lag = 12.5)) # Whole numbers only
  expect_error(initial_time_split(drinks, lag = 500))  # Lag must be less than number of training observations

})

test_that("grouped variants have the same classes as nongrouped outputs", {
  skip_if_not_installed("withr")
  # for `validation_split()` and variants
  withr::local_options(lifecycle_verbosity = "quiet")

  grouped_variants <- grep("^group_", names(rset_subclasses), value = TRUE)

  for (x in grouped_variants) {
    grouped_variant <- do.call(x, list(data = test_data(), group = "y"))

    ungrouped_variant <- gsub("^group_", "", x)
    ungrouped_variant <- do.call(ungrouped_variant, list(data = test_data()))

    expect_true(
      all(class(ungrouped_variant) %in% class(grouped_variant))
    )

    expect_true(
      all(
        class(ungrouped_variant$splits[[1]]) %in%
          class(grouped_variant$splits[[1]])
      )
    )
  }

  # Test initial_split() and initial_validation_split() separately,
  # as they don't return an rset
  grouped_variant <- group_initial_split(test_data(), y)
  ungrouped_variant <- initial_split(test_data())

  expect_true(
    all(class(ungrouped_variant) %in% class(grouped_variant))
  )

  grouped_variant <- group_initial_validation_split(test_data(), y)
  ungrouped_variant <- initial_validation_split(test_data())

  expect_true(
    all(class(ungrouped_variant) %in% class(grouped_variant))
  )
})

test_that("grouped variants are consistent across R sessions", {
  skip_if_not_installed("withr")
  # for `validation_split()` and variants
  withr::local_options(lifecycle_verbosity = "quiet")

  grouped_variants <- grep("^group_", names(rset_subclasses), value = TRUE)

  for (x in grouped_variants) {
    withr::with_seed(
      123,
      rs <- do.call(x, list(data = dplyr::starwars, group = "name"))
    )
    expect_snapshot(
      analysis(rs$splits[[1]])
    )
  }
})

#' Monte Carlo Cross-Validation
#'
#' One resample of Monte Carlo cross-validation takes a random sample (without
#'  replacement) of the original data set to be used for analysis. All other
#'  data points are added to the assessment set.
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @param times The number of times to repeat the sampling.
#' @export
#' @return An tibble with classes `mc_cv`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and a
#'  column called `id` that has a character string with the resample identifier.
#' @examplesIf rlang::is_installed("modeldata")
#' mc_cv(mtcars, times = 2)
#' mc_cv(mtcars, prop = .5, times = 2)
#'
#' library(purrr)
#' data(wa_churn, package = "modeldata")
#'
#' set.seed(13)
#' resample1 <- mc_cv(wa_churn, times = 3, prop = .5)
#' map_dbl(
#'   resample1$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' resample2 <- mc_cv(wa_churn, strata = churn, times = 3, prop = .5)
#' map_dbl(
#'   resample2$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' resample3 <- mc_cv(wa_churn, strata = tenure, breaks = 6, times = 3, prop = .5)
#' map_dbl(
#'   resample3$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#' @export
mc_cv <- function(
  data,
  prop = 3 / 4,
  times = 25,
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  ...
) {
  check_dots_empty()
  check_prop(prop)

  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) strata <- NULL
  }

  check_strata(strata, data)

  split_objs <-
    mc_splits(
      data = data,
      prop = prop,
      times = times,
      strata = strata,
      breaks = breaks,
      pool = pool
    )

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  if (!is.null(strata)) {
    names(strata) <- NULL
  }
  mc_att <- list(
    prop = prop,
    times = times,
    strata = strata,
    breaks = breaks,
    pool = pool
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = mc_att,
    subclass = c("mc_cv", "rset")
  )
}

# Get the indices of the assessment set from the analysis set
mc_complement <- function(ind, n) {
  list(
    analysis = ind,
    assessment = setdiff(1:n, ind)
  )
}


mc_splits <- function(
  data,
  prop = 3 / 4,
  times = 25,
  strata = NULL,
  breaks = 4,
  pool = 0.1
) {
  n <- nrow(data)
  if (is.null(strata)) {
    indices <- purrr::map(rep(n, times), sample, size = floor(n * prop))
  } else {
    stratas <- tibble::tibble(
      idx = 1:n,
      strata = make_strata(
        getElement(data, strata),
        breaks = breaks,
        pool = pool
      )
    )
    stratas <- split_unnamed(stratas, stratas$strata)
    stratas <-
      purrr::map(stratas, strat_sample, prop = prop, times = times) |>
      list_rbind()
    indices <- split_unnamed(stratas$idx, stratas$rs_id)
  }
  indices <- lapply(indices, mc_complement, n = n)
  split_objs <-
    purrr::map(indices, make_splits, data = data, class = "mc_split")
  list(
    splits = split_objs,
    id = names0(length(split_objs), "Resample")
  )
}

strat_sample <- function(x, prop, times, ...) {
  n <- nrow(x)
  idx <- purrr::map(rep(n, times), sample, size = floor(n * prop), ...)
  out <- purrr::map(idx, function(ind, x) x[sort(ind), "idx"], x = x) |>
    list_rbind()
  out$rs_id <- rep(1:times, each = floor(n * prop))
  out
}

#' Group Monte Carlo Cross-Validation
#'
#' Group Monte Carlo cross-validation creates splits of the data based
#'  on some grouping variable (which may have more than a single row
#'  associated with it). One resample of Monte Carlo cross-validation takes a
#'  random sample (without replacement) of groups in the original data set to be
#'  used for analysis. All other data points are added to the assessment set.
#'  A common use of this kind of resampling is when you have
#'  repeated measures of the same subject.
#'
#' @inheritParams mc_cv
#' @inheritParams make_groups
#' @export
#' @return A tibble with classes `group_mc_cv`,
#'  `rset`, `tbl_df`, `tbl`, and `data.frame`.
#'  The results include a column for the data split objects and an
#'  identification variable.
#' @examplesIf rlang::is_installed("modeldata")
#' data(ames, package = "modeldata")
#'
#' set.seed(123)
#' group_mc_cv(ames, group = Neighborhood, times = 5)
#'
#' @export
group_mc_cv <- function(
  data,
  group,
  prop = 3 / 4,
  times = 25,
  ...,
  strata = NULL,
  pool = 0.1
) {
  check_dots_empty()
  check_prop(prop)

  group <- validate_group({{ group }}, data)

  if (!missing(strata)) {
    strata <- check_grouped_strata({{ group }}, {{ strata }}, pool, data)
  }

  split_objs <-
    group_mc_splits(
      data = data,
      group = group,
      prop = prop,
      times = times,
      strata = strata,
      pool = pool
    )

  # This is needed for printing checks; strata can't be missing for mc_cv
  if (is.null(strata)) {
    strata <- FALSE
  }

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.
  split_objs$splits <- map(split_objs$splits, rm_out)

  mc_att <- list(
    group = group,
    prop = prop,
    times = times,
    balance = "prop",
    strata = strata,
    pool = pool
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = mc_att,
    subclass = c("group_mc_cv", "mc_cv", "group_rset", "rset")
  )
}

group_mc_splits <- function(
  data,
  group,
  prop = 3 / 4,
  times = 25,
  strata = NULL,
  pool = 0.1
) {
  group <- getElement(data, group)
  if (!is.null(strata)) {
    strata <- getElement(data, strata)
    strata <- as.character(strata)
    strata <- make_strata(strata, pool = pool)
  }

  n <- nrow(data)

  indices <- make_groups(
    data,
    group,
    times,
    balance = "prop",
    prop = prop,
    replace = FALSE,
    strata = strata
  )

  indices <- lapply(indices, mc_complement, n = n)

  split_objs <-
    purrr::map(
      indices,
      make_splits,
      data = data,
      class = c("group_mc_split", "mc_split")
    )

  all_assessable <- purrr::map(split_objs, function(x) nrow(assessment(x)))
  if (any(all_assessable == 0)) {
    cli_abort(
      c(
        "Some assessment sets contained zero rows",
        i = "Consider using a non-grouped resampling method"
      ),
      call = rlang::caller_env()
    )
  }

  list(
    splits = split_objs,
    id = names0(length(split_objs), "Resample")
  )
}

context("Monte Carlo CV")

library(testthat)
library(rsample)
library(purrr)

dat1 <- data.frame(a = 1:20, b = letters[1:20])

test_that('default param', {
  set.seed(11)
  rs1 <- mc_cv(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 15))
  expect_true(all(sizes1$assessment == 5))
  same_data <-
    map_lgl(rs1$splits, function(x)
      all.equal(x$data, dat1))
  expect_true(all(same_data))

  good_holdout <- map_lgl(rs1$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))
})

test_that('different percent', {
  set.seed(11)
  rs2 <- mc_cv(dat1, prop = .5)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == 10))
  expect_true(all(sizes2$assessment == 10))
  same_data <-
    map_lgl(rs2$splits, function(x)
      all.equal(x$data, dat1))
  expect_true(all(same_data))

  good_holdout <- map_lgl(rs2$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))
})

test_that('strata', {
  set.seed(11)
  rs3 <- mc_cv(warpbreaks, strata = "tension")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis == 42))
  expect_true(all(sizes3$assessment == 12))

  rate <- map_dbl(rs3$splits,
                  function(x) {
                    dat <- as.data.frame(x)$tension
                    mean(dat == "M")
                  })
  expect_true(length(unique(rate)) == 1)

  good_holdout <- map_lgl(rs3$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))
})


test_that('bad args', {
  expect_error(mc_cv(warpbreaks, strata = warpbreaks$tension))
  expect_error(mc_cv(warpbreaks, strata = c("tension", "wool")))
})


test_that('printing', {
  expect_output(print(mc_cv(warpbreaks)))
})


test_that('rsplit labels', {
  rs <- mc_cv(mtcars)
  all_labs <- map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("default param", {
  set.seed(11)
  rs1 <- mc_cv(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 15))
  expect_true(all(sizes1$assessment == 5))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("different percent", {
  set.seed(11)
  rs2 <- mc_cv(dat1, prop = .5)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == 10))
  expect_true(all(sizes2$assessment == 10))
  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("strata", {
  set.seed(11)

  rs3 <- mc_cv(warpbreaks, strata = "tension")
  sizes3 <- dim_rset(rs3)

  # sum(floor(table(warpbreaks$tension) * prop)) = 39
  expect_true(all(sizes3$analysis == 39))
  expect_true(all(sizes3$assessment == 15))

  rate <- purrr::map_dbl(
    rs3$splits,
    function(x) {
      dat <- as.data.frame(x)$tension
      mean(dat == "M")
    }
  )
  expect_true(length(unique(rate)) == 1)

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})


test_that("bad args", {
  expect_snapshot(error = TRUE, {
    mc_cv(mtcars, prop = -1)
  })
  expect_snapshot(error = TRUE, {
    mc_cv(mtcars, prop = 1)
  })
  expect_snapshot(error = TRUE, {
    mc_cv(warpbreaks, strata = warpbreaks$tension)
  })
  expect_snapshot(error = TRUE, {
    mc_cv(warpbreaks, strata = c("tension", "wool"))
  })
})


test_that("printing", {
  expect_snapshot(mc_cv(warpbreaks))
})


test_that("rsplit labels", {
  rs <- mc_cv(mtcars)
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("grouping - bad args", {
  expect_snapshot(error = TRUE, {
    group_mc_cv(warpbreaks, group = warpbreaks$tension)
  })
  expect_snapshot(error = TRUE, {
    group_mc_cv(warpbreaks, group = c("tension", "wool"))
  })
  expect_snapshot(error = TRUE, {
    group_mc_cv(warpbreaks, group = "tensio")
  })
  expect_snapshot(error = TRUE, {
    group_mc_cv(warpbreaks)
  })
  expect_snapshot(error = TRUE, {
    group_mc_cv(mtcars, group = "cyl", prop = 1)
  })
  expect_snapshot(error = TRUE, {
    group_mc_cv(warpbreaks, group = "tension", balance = "groups")
  })
  set.seed(1)
  expect_snapshot(error = TRUE, {
    group_mc_cv(warpbreaks, group = "tension", prop = 0.99)
  })
})

test_that("grouping - default param", {
  set.seed(11)
  rs1 <- group_mc_cv(warpbreaks, "tension")
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 36))
  expect_true(all(sizes1$assessment == 18))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, warpbreaks)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("grouping -- strata", {
  set.seed(11)

  n_common_class <- 70
  n_rare_class <- 30

  group_table <- tibble(
    group = 1:100,
    outcome = sample(c(rep(0, n_common_class), rep(1, n_rare_class)))
  )
  observation_table <- tibble(
    group = sample(1:100, 5e4, replace = TRUE),
    observation = 1:5e4
  )
  sample_data <- dplyr::full_join(
    group_table,
    observation_table,
    by = "group",
    multiple = "all"
  )
  rs4 <- group_mc_cv(sample_data, group, times = 5, strata = outcome)
  sizes4 <- dim_rset(rs4)
  expect_snapshot(sizes4)

  rate <- purrr::map_dbl(
    rs4$splits,
    function(x) {
      dat <- as.data.frame(x)$outcome
      mean(dat == "1")
    }
  )
  expect_equal(mean(rate), 0.3, tolerance = 1e-2)

  good_holdout <- purrr::map_lgl(
    rs4$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("grouping - tibble input", {
  warpbreaks2 <- tibble::as_tibble(warpbreaks)
  set.seed(11)
  rs3 <- group_mc_cv(warpbreaks2, "tension")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis == 36))
  expect_true(all(sizes3$assessment == 18))
  same_data <-
    purrr::map_lgl(rs3$splits, function(x) {
      all.equal(x$data, warpbreaks2)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("grouping with times = 1 works", {
  set.seed(11)
  rs3 <- group_mc_cv(warpbreaks, "tension", times = 1)
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis == 36))
  expect_true(all(sizes3$assessment == 18))
  same_data <-
    purrr::map_lgl(rs3$splits, function(x) {
      all.equal(x$data, warpbreaks)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("grouping - printing", {
  expect_snapshot(group_mc_cv(warpbreaks, "tension"))
})

test_that("grouping - printing with ...", {
  expect_snapshot(
    print(group_mc_cv(warpbreaks, "tension"), n = 2)
  )
})

test_that("grouping - rsplit labels", {
  rs <- group_mc_cv(warpbreaks, "tension")
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

#' Obtain a identifier for the resamples
#'
#' This function returns a hash (or NA) for an attribute that is created when
#' the `rset` was initially constructed. This can be used to compare with other
#' resampling objects to see if they are the same.
#' @param x An `rset` or `tune_results` object.
#' @param ... Not currently used.
#' @return A character value or `NA_character_` if the object was created prior
#' to rsample version 0.1.0.
#' @rdname get_fingerprint
#' @aliases .get_fingerprint
#' @examples
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(2)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars, repeats = 2))
#' @export
.get_fingerprint <- function(x, ...) {
  UseMethod(".get_fingerprint")
}

#' @export
#' @rdname get_fingerprint
.get_fingerprint.default <- function(x, ...) {
  cls <- class(x)
  cli_abort("No method for objects of class{?es}: {cls}")
}

#' @export
#' @rdname get_fingerprint
.get_fingerprint.rset <- function(x, ...) {
  check_dots_empty()
  att <- attributes(x)
  if (any(names(att) == "fingerprint")) {
    res <- att$fingerprint
  } else {
    res <- NA_character_
  }
  res
}

test_that("fingerprinting", {
  set.seed(1)
  rs_1 <- vfold_cv(mtcars)
  fp_1 <- .get_fingerprint(rs_1)

  set.seed(1)
  fp_2 <- .get_fingerprint(vfold_cv(mtcars))

  set.seed(1)
  fp_3 <- .get_fingerprint(vfold_cv(mtcars, repeats = 2))

  expect_true(class(fp_1) == "character")
  expect_true(class(fp_2) == "character")
  expect_true(class(fp_3) == "character")
  expect_equal(fp_1, fp_2)
  expect_false(fp_1 == fp_3)
})

rsplit <- function(data, in_id, out_id) {
  if (!is.data.frame(data) & !is.matrix(data)) {
    cli_abort("{.arg data} must be a data frame.")
  }

  if (!is.integer(in_id) | any(in_id < 1)) {
    cli_abort("{.arg in_id} must be a positive integer vector.")
  }

  if (!all(is.na(out_id))) {
    if (!is.integer(out_id) | any(out_id < 1)) {
      cli_abort("{.arg out_id} must be a positive integer vector.")
    }
  }

  if (length(in_id) == 0) {
    cli_abort("At least one row should be selected for the analysis set.")
  }

  structure(
    list(
      data = data,
      in_id = in_id,
      out_id = out_id
    ),
    class = "rsplit"
  )
}

#' @export
print.rsplit <- function(x, ...) {
  out_char <-
    if (is_missing_out_id(x)) {
      paste(length(complement(x)))
    } else {
      paste(length(x$out_id))
    }

  cat("<Analysis/Assess/Total>\n")
  cat("<", length(x$in_id), "/", out_char, "/", nrow(x$data), ">\n", sep = "")
}

#' @export
as.integer.rsplit <-
  function(x, data = c("analysis", "assessment"), ...) {
    data <- match.arg(data)
    if (data == "analysis") {
      out <- x$in_id
    } else {
      out <- if (is_missing_out_id(x)) {
        complement(x)
      } else {
        x$out_id
      }
    }
    out
  }


#' Convert an `rsplit` object to a data frame
#'
#' The analysis or assessment code can be returned as a data
#'   frame (as dictated by the `data` argument) using
#'   `as.data.frame.rsplit()`. `analysis()` and
#'   `assessment()` are shortcuts.
#' @param x An `rsplit` object.
#' @param row.names `NULL` or a character vector giving the row names for the data frame. Missing values are not allowed.
#' @param optional A logical: should the column names of the data be checked for legality?
#' @param data Either `"analysis"` or `"assessment"` to specify which data are returned.
#' @param ... Not currently used.
#' @examples
#' library(dplyr)
#' set.seed(104)
#' folds <- vfold_cv(mtcars)
#'
#' model_data_1 <- folds$splits[[1]] |> analysis()
#' holdout_data_1 <- folds$splits[[1]] |> assessment()
#' @export
as.data.frame.rsplit <-
  function(x, row.names = NULL, optional = FALSE, data = "analysis", ...) {
    if (!is.null(row.names)) {
      cli::cli_warn(
        "{.arg row.names} is kept for consistency with the underlying class but 
        non-NULL values will be ignored."
      )
    }
    if (optional) {
      cli::cli_warn(
        "{.arg optional} is kept for consistency with the underlying class but
        TRUE values will be ignored."
      )
    }
    if (!is.null(x$col_id)) {
      if (identical(data, "assessment")) {
        rsplit_class <- class(x)[[1]]
        cli_abort(
          "There is no assessment data set for an {.arg rsplit} object
          with class {.cls {rsplit_class}}."
        )
      }
      ind <- as.integer(x, data = data, ...)
      permuted_col <- vctrs::vec_slice(x$data, ind) |>
        dplyr::select(x$col_id)
      x$data[, x$col_id] <- permuted_col
      return(x$data)
    }
    row_ind <- as.integer(x, data = data, ...)
    vctrs::vec_slice(x$data, row_ind)
  }

#' @rdname as.data.frame.rsplit
#' @export
analysis <- function(x, ...) {
  UseMethod("analysis")
}

#' @export
#' @rdname as.data.frame.rsplit
analysis.default <- function(x, ...) {
  cls <- class(x)
  cli::cli_abort(
    "No method for objects of class{?es}: {cls}"
  )
}

#' @export
#' @rdname as.data.frame.rsplit
analysis.rsplit <- function(x, ...) {
  as.data.frame(x, data = "analysis", ...)
}
#' @rdname as.data.frame.rsplit
#' @export
assessment <- function(x, ...) {
  UseMethod("assessment")
}

#' @export
#' @rdname as.data.frame.rsplit
assessment.default <- function(x, ...) {
  cls <- class(x)
  cli::cli_abort(
    "No method for objects of class{?es}: {cls}"
  )
}

#' @rdname as.data.frame.rsplit
#' @export
assessment.rsplit <- function(x, ...) {
  as.data.frame(x, data = "assessment", ...)
}

#' @export
dim.rsplit <- function(x, ...) {
  c(
    analysis = length(x$in_id),
    assessment = length(complement(x)),
    n = nrow(x$data),
    p = ncol(x$data)
  )
}

#' @method obj_sum rsplit
#' @export
obj_sum.rsplit <- function(x, ...) {
  out_char <-
    if (is_missing_out_id(x)) {
      paste(length(complement(x)))
    } else {
      paste(length(x$out_id))
    }

  paste0(
    "split [",
    length(x$in_id),
    "/",
    out_char,
    "]"
  )
}


#' @method type_sum rsplit
#' @export
type_sum.rsplit <- function(x, ...) {
  out_char <-
    if (is_missing_out_id(x)) {
      format_n(length(complement(x)))
    } else {
      format_n(length(x$out_id))
    }

  paste0(
    "split [",
    format_n(length(x$in_id)),
    "/",
    out_char,
    "]"
  )
}


format_n <- function(x, digits = 1) {
  case_when(
    log10(x) < 3 ~ paste(x),
    log10(x) >= 3 & log10(x) < 6 ~
      paste0(round(x / 1000, digits = digits), "K"),
    TRUE ~ paste0(round(x / 1000000, digits = digits), "M"),
  )
}

is_rsplit <- function(x) {
  inherits(x, "rsplit")
}

context("Rsplit constructor")

library(testthat)
library(rsample)

dat1 <- data.frame(a = 1:100, b = 101:200)
size1 <- object.size(dat1)

dat2 <- as.matrix(dat1)

test_that('simple rsplit', {
  rs1 <- rsplit(dat1, 1:2, 4:5)
  expect_equal(rs1$data, dat1)
  expect_equal(rs1$in_id, 1:2)
  expect_equal(rs1$out_id, 4:5)
})

test_that('simple rsplit with matrices', {
  rs2 <- rsplit(dat2, 1:2, 4:5)
  expect_equal(rs2$data, dat2)
  expect_equal(rs2$in_id, 1:2)
  expect_equal(rs2$out_id, 4:5)
})

test_that('bad inputs', {
  expect_error(rsplit(as.list(dat1), 1:2, 4:5))
  expect_error(rsplit(dat1, letters[1:2], 4:5))
  expect_error(rsplit(as.list(dat1), 1:2, letters[4:5]))
  expect_error(rsplit(as.list(dat1), -1:2, 4:5))
  expect_error(rsplit(as.list(dat1), 1:2, -4:5))
  expect_error(rsplit(as.list(dat1), integer(0), 4:5))
})

test_that('as.data.frame', {
  rs3 <- rsplit(dat1, 1:2, 4:5)
  expect_equal(as.data.frame(rs3), dat1[1:2,])
  expect_equal(as.data.frame(rs3, data = "assessment"), dat1[4:5,])

  rs4 <- rsplit(dat1, rep(1:2, each = 3), rep(4:5, c(2, 1)))
  expect_equal(as.data.frame(rs4), dat1[c(1, 1, 1, 2, 2, 2),])
  expect_equal(as.data.frame(rs4, data = "assessment"), dat1[c(4, 4, 5),])
})


test_that('print methods', {
  verify_output(test_path("print_test_output", "rsplit"), {
    set.seed(233)
    print(vfold_cv(mtcars)$splits[[1]])
  })
  verify_output(test_path("print_test_output", "val_split"), {
    set.seed(233)
    print(validation_split(mtcars)$splits[[1]])
  })
})

context("Group resampling")

library(testthat)
library(rsample)
library(purrr)
library(tibble)

warpbreaks2 <- as_tibble(warpbreaks)

get_id_left_out <- function(x)
     unique(as.character(assessment(x)$tension))

test_that('bad args', {
  expect_error(group_vfold_cv(warpbreaks, group = warpbreaks$tension))
  expect_error(group_vfold_cv(warpbreaks, group = c("tension", "wool")))
  expect_error(group_vfold_cv(warpbreaks, group = "tensio"))
  expect_error(group_vfold_cv(warpbreaks))
  expect_error(group_vfold_cv(warpbreaks, group = "tension", v = 10))
})


test_that('default param', {
  set.seed(11)
  rs1 <- group_vfold_cv(warpbreaks, "tension")
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 36))
  expect_true(all(sizes1$assessment == 18))
  same_data <-
    map_lgl(rs1$splits, function(x)
      all.equal(x$data, warpbreaks))
  expect_true(all(same_data))

  good_holdout <- map_lgl(rs1$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))

  sp_out <- map_chr(rs1$splits, get_id_left_out)
  expect_true(all(table(sp_out) == 1))
})


test_that('v < max v', {
  set.seed(11)
  rs2 <- group_vfold_cv(warpbreaks, "tension", v = 2)
  sizes2 <- dim_rset(rs2)

  expect_true(!all(sizes2$analysis == 36))
  expect_true(!all(sizes2$assessment == 18))
  same_data <-
    map_lgl(rs2$splits, function(x)
      all.equal(x$data, warpbreaks))
  expect_true(all(same_data))

  good_holdout <- map_lgl(rs2$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))

  sp_out <- map(rs2$splits, get_id_left_out)
  expect_true(all(table(unlist(sp_out)) == 1))
})

test_that('tibble input', {
  set.seed(11)
  rs3 <- group_vfold_cv(warpbreaks2, "tension")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis == 36))
  expect_true(all(sizes3$assessment == 18))
  same_data <-
    map_lgl(rs3$splits, function(x)
      all.equal(x$data, warpbreaks2))
  expect_true(all(same_data))

  good_holdout <- map_lgl(rs3$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))

  sp_out <- map_chr(rs3$splits, get_id_left_out)
  expect_true(all(table(sp_out) == 1))
})


test_that('printing', {
  expect_output(print(group_vfold_cv(warpbreaks, "tension")))
})

test_that('printing with ...', {
  verify_output(test_path("test-print-groups.txt"), {
    print(group_vfold_cv(warpbreaks, "tension"), n = 2)
  })
})

test_that('rsplit labels', {
  rs <- group_vfold_cv(warpbreaks, "tension")
  all_labs <- map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

context("Rset constructor")

library(testthat)
library(rsample)

cars_10fold <- vfold_cv(mtcars)

test_that('bad args', {
  expect_error(
    new_rset(cars_10fold$splits[1:2], cars_10fold$id)
  )
  expect_error(
    new_rset(cars_10fold$splits, cars_10fold[ "splits"])
  )
  expect_error(
    new_rset(cars_10fold$splits, cars_10fold$splits)
  )
  expect_error(
    new_rset(list(1), "x"),
    "must be an `rsplit` object"
  )
  args <- list(a = 1, b = 2, 3)
  expect_error(
    new_rset(
      cars_10fold$splits,
      cars_10fold$id,
      attrib = args
      )
  )
})

test_that('rset with attributes', {
  args <- list(value = "potato")
  res3 <- new_rset(
    cars_10fold$splits,
    cars_10fold$id,
    attrib = args
  )
  expect_equal(sort(names(attributes(res3))),
               c("class", "names", "row.names", "value"))
  expect_equal(attr(res3, "value"), "potato")
})

test_that('rset with additional classes', {
  res4 <- new_rset(
    cars_10fold$splits,
    cars_10fold$id,
    subclass = "potato"
  )
  expect_equal(class(res4),
               c("potato", "tbl_df", "tbl", "data.frame"))
})

test_that('not an rsplit', {
  folds <- vfold_cv(mtcars)
  expect_error(analysis(folds$splits[1]))
  expect_error(assessment(folds$splits[1]))
})

test_that("can create a split with an empty assessment set (#188)", {
  df <- data.frame(x = c(1, 2, 3, 4))
  indices <- list(analysis = 1:4, assessment = integer())

  split <- make_splits(indices, df)

  expect_identical(split$out_id, integer())
  expect_identical(assessment(split), df[0, , drop = FALSE])
})

test_that("cannot create a split with an empty analysis set", {
  df <- data.frame(x = c(1, 2, 3, 4))
  indices <- list(analysis = integer(), assessment = 1:4)

  expect_error(make_splits(indices, df), "At least one row")
})

test_that("create a split from training and testing dataframes", {
  training <- tibble(x = c(1, 2, 3, 4))
  testing <- tibble(x = c(5, 6))

  split <- make_splits(training, testing)
  expect_identical(analysis(split), training)
  expect_identical(assessment(split), testing)
})

test_that("can create a split from empty testing dataframe", {
  training <- tibble(x = c(1, 2, 3, 4))
  testing <- tibble()

  split <- make_splits(training, testing)
  expect_identical(split$out_id, integer())
  expect_identical(analysis(split), training)
})

test_that("cannot create a split from empty training dataframe", {
  training <- tibble()
  testing <- tibble(x = c(5, 6))

  expect_error(
    make_splits(training, testing),
    "The analysis set must contain at least one row."
  )
})

test_that("cannot create a split from dataframes with different columns", {
  training <- tibble(x = c(1, 2, 3, 4))
  testing <- tibble(y = c(5, 6))

  expect_error(
    make_splits(training, testing),
    "The analysis and assessment sets must have"
  )
})

test_that("improper argument", {
  expect_error(make_splits("potato"), "There is no method available to")
})

test_that("simple rsplit", {
  rs1 <- rsplit(dat1, 1:2, 4:5)
  expect_equal(rs1$data, dat1)
  expect_equal(rs1$in_id, 1:2)
  expect_equal(rs1$out_id, 4:5)
})

test_that("simple rsplit with matrices", {
  dat2 <- as.matrix(dat1)
  rs2 <- rsplit(dat2, 1:2, 4:5)
  expect_equal(rs2$data, dat2)
  expect_equal(rs2$in_id, 1:2)
  expect_equal(rs2$out_id, 4:5)
})

test_that("bad inputs", {
  expect_snapshot(error = TRUE, {
    rsplit(as.list(dat1), 1:2, 4:5)
  })
  expect_snapshot(error = TRUE, {
    rsplit(dat1, letters[1:2], 4:5)
  })
  expect_snapshot(error = TRUE, {
    rsplit(as.list(dat1), 1:2, letters[4:5])
  })
  expect_snapshot(error = TRUE, {
    rsplit(as.list(dat1), -1:2, 4:5)
  })
  expect_snapshot(error = TRUE, {
    rsplit(as.list(dat1), 1:2, -4:5)
  })
  expect_snapshot(error = TRUE, {
    rsplit(as.list(dat1), integer(0), 4:5)
  })
})

test_that("as.data.frame", {
  rs3 <- rsplit(dat1, 1:2, 4:5)
  expect_equal(as.data.frame(rs3), dat1[1:2, ])
  expect_equal(
    as.data.frame(rs3, data = "assessment"),
    dat1[4:5, ],
    ignore_attr = "row.names"
  )

  rs4 <- rsplit(dat1, rep(1:2, each = 3), rep(4:5, c(2, 1)))
  expect_equal(
    as.data.frame(rs4),
    dat1[c(1, 1, 1, 2, 2, 2), ],
    ignore_attr = "row.names"
  )
  expect_equal(
    as.data.frame(rs4, data = "assessment"),
    dat1[c(4, 4, 5), ],
    ignore_attr = "row.names"
  )
})


test_that("print methods", {
  expect_snapshot({
    set.seed(233)
    vfold_cv(mtcars)$splits[[1]]
  })

  skip_if_not_installed("withr")
  withr::local_options(lifecycle_verbosity = "quiet")

  expect_snapshot({
    set.seed(233)
    validation_split(mtcars)$splits[[1]]
  })
  expect_snapshot({
    set.seed(233)
    validation_split(mtcars)
  })
})

test_that("`complement()` error messages", {
  fake_rsplit <- 1
  class(fake_rsplit) <- c("not_an_rsplit")
  expect_snapshot(error = TRUE, {
    complement(fake_rsplit)
  })
  class(fake_rsplit) <- c("not_an_rsplit", "really_not_an_rsplit")
  expect_snapshot(error = TRUE, {
    complement(fake_rsplit)
  })
  expect_snapshot(error = TRUE, {
    get_stored_out_id(list(out_id = NA))
  })
})

test_that("as.data.frame() works for permutations with Surv object without the survival package loaded - issue #443", {
  srv <-
    list(
      age = c(74, 68, 56, 57, 60, 74, 76, 77, 39, 75, 66, 58),
      sex = c(1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 2),
      surv_obj = structure(
        c(
          306,
          455,
          1010,
          210,
          883,
          1022,
          116,
          188,
          191,
          105,
          174,
          177,
          1,
          1,
          0,
          1,
          1,
          0,
          1,
          0,
          0,
          0,
          0,
          0
        ),
        dim = c(12L, 2L),
        dimnames = list(NULL, c("time", "status")),
        type = "right",
        class = "Surv"
      )
    )
  surv_df <-
    structure(
      srv,
      row.names = paste(1:12),
      class = "data.frame"
    )

  set.seed(472)
  surv_permutation_df <- permutations(
    surv_df,
    permute = c(age, surv_obj),
    times = 1
  ) |>
    get_rsplit(1)

  expect_s3_class(surv_permutation_df$data$surv_obj, "Surv")
  expect_s3_class(analysis(surv_permutation_df)$surv_obj, "Surv")
})

test_that("default param", {
  set.seed(11)
  rs1 <- vfold_cv(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 18))
  expect_true(all(sizes1$assessment == 2))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("repeated", {
  set.seed(11)
  rs2 <- vfold_cv(dat1, repeats = 4)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == 18))
  expect_true(all(sizes2$assessment == 2))
  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("strata", {
  set.seed(11)
  skip_if_not_installed("modeldata")
  data("mlc_churn", package = "modeldata")
  rs3 <- vfold_cv(mlc_churn, repeats = 2, strata = "voice_mail_plan")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis %in% 4499:4501))
  expect_true(all(sizes3$assessment %in% 499:501))

  rate <- purrr::map_dbl(
    rs3$splits,
    function(x) {
      dat <- as.data.frame(x)$voice_mail_plan
      mean(dat == "yes")
    }
  )
  expect_equal(mean(unique(rate)), 0.2645925848)

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  expect_snapshot(
    rs4 <- vfold_cv(mlc_churn, strata = state, pool = 0.01)
  )
})

test_that("strata arg is checked", {
  expect_snapshot(error = TRUE, {
    vfold_cv(iris, strata = iris$Species)
  })

  # errors from `check_strata()`
  expect_snapshot(error = TRUE, {
    vfold_cv(iris, strata = c("Species", "Sepal.Width"))
  })

  expect_snapshot(error = TRUE, {
    vfold_cv(iris, strata = NA)
  })

  # make Surv object without a dependeny on the survival package
  surv_obj <- structure(
    c(306, 455, 1010, 210, 883, 1, 1, 0, 1, 1),
    dim = c(5L, 2L),
    dimnames = list(NULL, c("time", "status")),
    type = "right",
    class = "Surv"
  )
  dat <- data.frame(a = 1:5)
  # add Surv object like this for older R versions (<= 4.2.3)
  dat$b <- surv_obj
  expect_snapshot(error = TRUE, {
    vfold_cv(dat, strata = b)
  })
})

test_that("v arg is checked", {
  expect_snapshot(error = TRUE, {
    vfold_cv(iris, v = -500)
  })
  expect_snapshot(error = TRUE, {
    vfold_cv(iris, v = 1)
  })
  expect_snapshot(error = TRUE, {
    vfold_cv(iris, v = NULL)
  })
  expect_snapshot(error = TRUE, {
    vfold_cv(iris, v = 500)
  })
  expect_snapshot(error = TRUE, {
    vfold_cv(mtcars, v = nrow(mtcars))
  })
})

test_that("repeats arg is checked", {
  expect_snapshot(error = TRUE, {
    vfold_cv(iris, v = 150, repeats = 2)
  })
  expect_snapshot(error = TRUE, {
    vfold_cv(Orange, repeats = 0)
  })
  expect_snapshot(error = TRUE, {
    vfold_cv(Orange, repeats = NULL)
  })
})

test_that("printing", {
  expect_snapshot(vfold_cv(mtcars))
})


test_that("rsplit labels", {
  rs <- vfold_cv(mtcars)
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)

  rs2 <- vfold_cv(mtcars, repeats = 4)
  all_labs2 <- purrr::map(rs2$splits, labels) |>
    list_rbind()
  original_id2 <- rs2[, grepl("^id", names(rs2))]
  expect_equal(all_labs2, original_id2)
})

test_that("grouping -- bad args", {
  expect_snapshot(error = TRUE, {
    group_vfold_cv(warpbreaks, group = warpbreaks$tension)
  })
  expect_snapshot(error = TRUE, {
    group_vfold_cv(warpbreaks, group = c("tension", "wool"))
  })
  expect_snapshot(error = TRUE, {
    group_vfold_cv(warpbreaks, group = "tensio")
  })
  expect_snapshot(error = TRUE, {
    group_vfold_cv(warpbreaks)
  })
  expect_snapshot(error = TRUE, {
    group_vfold_cv(warpbreaks, group = "tension", v = 10)
  })
  expect_snapshot(error = TRUE, {
    group_vfold_cv(dat1, c, v = 4, repeats = 4)
  })
  expect_snapshot(error = TRUE, {
    group_vfold_cv(dat1, c, repeats = 4)
  })
  expect_snapshot(error = TRUE, {
    group_vfold_cv(Orange, v = 1, group = "Tree")
  })
})


test_that("grouping -- default param", {
  set.seed(11)
  rs1 <- group_vfold_cv(warpbreaks, "tension")
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 36))
  expect_true(all(sizes1$assessment == 18))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, warpbreaks)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  sp_out <- purrr::map_chr(rs1$splits, get_id_left_out)
  expect_true(all(table(sp_out) == 1))
})

test_that("grouping works with non-missing strata = NULL", {
  set.seed(11)
  rset_strata_missing <- group_vfold_cv(warpbreaks, "tension")

  expect_no_error({
    set.seed(11)
    rset_strata_null <- group_vfold_cv(warpbreaks, "tension", strata = NULL)
  })

  expect_identical(rset_strata_null, rset_strata_missing)
})

test_that("grouping -- v < max v", {
  set.seed(11)
  rs2 <- group_vfold_cv(warpbreaks, "tension", v = 2)
  sizes2 <- dim_rset(rs2)

  expect_true(!all(sizes2$analysis == 36))
  expect_true(!all(sizes2$assessment == 18))
  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, warpbreaks)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  sp_out <- purrr::map(rs2$splits, get_id_left_out)
  expect_true(all(table(unlist(sp_out)) == 1))
})

test_that("grouping -- tibble input", {
  warpbreaks2 <- tibble::as_tibble(warpbreaks)
  set.seed(11)
  rs3 <- group_vfold_cv(warpbreaks2, "tension")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis == 36))
  expect_true(all(sizes3$assessment == 18))
  same_data <-
    purrr::map_lgl(rs3$splits, function(x) {
      all.equal(x$data, warpbreaks2)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  sp_out <- purrr::map_chr(rs3$splits, get_id_left_out)
  expect_true(all(table(sp_out) == 1))
})

test_that("grouping -- other balance methods", {
  skip_if_not_installed("modeldata")
  data(ames, package = "modeldata")
  set.seed(11)
  rs1 <- group_vfold_cv(
    ames,
    "Neighborhood",
    balance = "observations",
    v = 5
  )
  expect_snapshot(rs1)

  sizes1 <- dim_rset(rs1)
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, ames)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  expect_true(
    !any(
      unique(as.character(assessment(rs1$splits[[1]])$Neighborhood)) %in%
        unique(as.character(analysis(rs1$splits[[1]])$Neighborhood))
    )
  )
})

test_that("grouping -- strata", {
  set.seed(11)

  n_common_class <- 70
  n_rare_class <- 30

  group_table <- tibble(
    group = 1:100,
    outcome = sample(c(rep(0, n_common_class), rep(1, n_rare_class)))
  )
  observation_table <- tibble(
    group = sample(1:100, 1e5, replace = TRUE),
    observation = 1:1e5
  )
  sample_data <- dplyr::full_join(
    group_table,
    observation_table,
    by = "group",
    multiple = "all"
  )
  rs4 <- group_vfold_cv(sample_data, group, v = 5, strata = outcome)
  sizes4 <- dim_rset(rs4)
  expect_snapshot(sizes4)

  rate <- purrr::map_dbl(
    rs4$splits,
    function(x) {
      dat <- as.data.frame(x)$outcome
      mean(dat == "1")
    }
  )
  expect_equal(mean(rate), 0.3, tolerance = 1e-2)

  good_holdout <- purrr::map_lgl(
    rs4$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  expect_snapshot(
    group_vfold_cv(sample_data, group, strata = outcome)
  )

  expect_equal(
    nrow(
      suppressWarnings(
        group_vfold_cv(sample_data, group, strata = outcome)
      )
    ),
    n_rare_class
  )

  rs5 <- group_vfold_cv(
    sample_data,
    group,
    v = 5,
    strata = outcome,
    balance = "observations"
  )
  sizes5 <- dim_rset(rs5)
  expect_snapshot(sizes5)

  rate <- purrr::map_dbl(
    rs5$splits,
    function(x) {
      dat <- as.data.frame(x)$outcome
      mean(dat == "1")
    }
  )
  expect_equal(mean(rate), 0.3, tolerance = 1e-2)

  good_holdout <- purrr::map_lgl(
    rs5$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  expect_snapshot(
    group_vfold_cv(sample_data, group, strata = outcome)
  )

  expect_equal(
    nrow(
      suppressWarnings(
        group_vfold_cv(
          sample_data,
          group,
          strata = outcome,
          balance = "observations"
        )
      )
    ),
    n_rare_class
  )
})

test_that("grouping fails for strata not constant across group members", {
  set.seed(11)

  n_common_class <- 70
  n_rare_class <- 30

  group_table <- tibble(
    group = 1:100,
    outcome = sample(c(rep(0, n_common_class), rep(1, n_rare_class)))
  )
  observation_table <- tibble(
    group = sample(1:100, 1e5, replace = TRUE),
    observation = 1:1e5
  )
  sample_data <- dplyr::full_join(
    group_table,
    observation_table,
    by = "group",
    multiple = "all"
  )

  # violate requirement
  sample_data$outcome[1] <- ifelse(sample_data$outcome[1], 0, 1)

  expect_snapshot(error = TRUE, {
    group_vfold_cv(sample_data, group, v = 5, strata = outcome)
  })
})

test_that("grouping -- repeated", {
  set.seed(11)
  rs2 <- group_vfold_cv(dat1, c, v = 3, repeats = 4)
  sizes2 <- dim_rset(rs2)

  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("grouping -- printing", {
  expect_snapshot(group_vfold_cv(warpbreaks, "tension"))
})

test_that("grouping -- printing with ...", {
  expect_snapshot(
    print(group_vfold_cv(warpbreaks, "tension"), n = 2)
  )
})

test_that("grouping -- rsplit labels", {
  rs <- group_vfold_cv(warpbreaks, "tension")
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("simple rsplit", {
  rs1 <- rsplit(dat1, 1:2, 4:5)
  expect_equal(rs1$data, dat1)
  expect_equal(rs1$in_id, 1:2)
  expect_equal(rs1$out_id, 4:5)
})

test_that("simple rsplit with matrices", {
  dat2 <- as.matrix(dat1)
  rs2 <- rsplit(dat2, 1:2, 4:5)
  expect_equal(rs2$data, dat2)
  expect_equal(rs2$in_id, 1:2)
  expect_equal(rs2$out_id, 4:5)
})

test_that("bad inputs", {
  expect_error(rsplit(as.list(dat1), 1:2, 4:5))
  expect_error(rsplit(dat1, letters[1:2], 4:5))
  expect_error(rsplit(as.list(dat1), 1:2, letters[4:5]))
  expect_error(rsplit(as.list(dat1), -1:2, 4:5))
  expect_error(rsplit(as.list(dat1), 1:2, -4:5))
  expect_error(rsplit(as.list(dat1), integer(0), 4:5))
})

test_that("as.data.frame", {
  rs3 <- rsplit(dat1, 1:2, 4:5)
  expect_equal(as.data.frame(rs3), dat1[1:2, ])
  expect_equal(as.data.frame(rs3, data = "assessment"), dat1[4:5, ])

  rs4 <- rsplit(dat1, rep(1:2, each = 3), rep(4:5, c(2, 1)))
  expect_equal(as.data.frame(rs4), dat1[c(1, 1, 1, 2, 2, 2), ])
  expect_equal(as.data.frame(rs4, data = "assessment"), dat1[c(4, 4, 5), ])
})


test_that("print methods", {
  expect_snapshot({
    set.seed(233)
    vfold_cv(mtcars)$splits[[1]]
  })
  expect_snapshot({
    set.seed(233)
    validation_split(mtcars)$splits[[1]]
  })
  expect_snapshot({
    set.seed(233)
    validation_split(mtcars)
  })
})

test_that("default complement method errors", {
  expect_snapshot(
    complement("a string"),
    error = TRUE
  )
})

test_that("can create a manual rset", {
  df <- data.frame(x = c(1, 2, 3, 4))

  indices <- list(
    list(analysis = 1L, assessment = 2L),
    list(analysis = 3L, assessment = 4L)
  )

  splits <- lapply(indices, make_splits, data = df)

  rset <- manual_rset(splits, c("Split 1", "Split 2"))

  expect_s3_class(rset, c("manual_rset", "rset"))
  expect_identical(rset$id, c("Split 1", "Split 2"))
})

test_that("can use analysis/assessment functions", {
  df <- data.frame(x = c(1, 2, 3))
  indices <- list(analysis = c(1L, 2L), assessment = 3L)
  splits <- list(make_splits(indices, df))

  rset <- manual_rset(splits, "Split 1")

  expect_identical(analysis(rset$splits[[1]]), df[1:2, 1, drop = FALSE])
  expect_identical(assessment(rset$splits[[1]]), df[3, 1, drop = FALSE])
})

test_that("`pretty()` works", {
  df <- data.frame(x = c(1, 2, 3))
  indices <- list(analysis = c(1L, 2L), assessment = 3L)
  splits <- list(make_splits(indices, df))

  rset <- manual_rset(splits, "Split 1")

  expect_identical(pretty(rset), "Manual resampling")
})

library(testthat)
library(rsample)

context("Labels")

test_that('basic cv', {
  cv_obj <- vfold_cv(mtcars)
  expect_equal(cv_obj$id, labels(cv_obj))
  expect_is(labels(cv_obj), "character")
  expect_s3_class(labels(cv_obj, TRUE), "factor")
})

test_that('repeated cv', {
  rcv_obj <- vfold_cv(mtcars, repeats = 3)
  expect_equal(paste(rcv_obj$id, rcv_obj$id2, sep = "."),
               labels(rcv_obj))
  expect_is(labels(rcv_obj), "character")
  expect_s3_class(labels(rcv_obj, TRUE), "factor")
})

test_that('nested cv', {
  expect_error(
    labels(
      nested_cv(mtcars,
                outside = vfold_cv(v = 3),
                inside = bootstraps(times = 5)
                )
      )
    )
})

test_that('adding labels', {
  set.seed(363)
  car_folds <- vfold_cv(mtcars, repeats = 3)

  res <-
    analysis(car_folds$splits[[1]]) %>%
    add_resample_id(car_folds$splits[[1]])
  expect_equal(colnames(res), c(colnames(mtcars), "id", "id2"))

  car_bt <- bootstraps(mtcars)

  res <- analysis(car_bt$splits[[1]]) %>%
    add_resample_id(car_bt$splits[[1]])

  expect_equal(colnames(res), c(colnames(mtcars), "id"))

  res <- analysis(car_bt$splits[[1]]) %>%
    add_resample_id(car_bt$splits[[1]], TRUE)

  expect_equal(colnames(res), c(colnames(mtcars), ".id"))

  expect_error(
    analysis(car_folds$splits[[1]]) %>%
      add_resample_id(car_folds$splits[[1]], 7)
  )
  expect_error(
    analysis(car_folds$splits[[1]]) %>%
      add_resample_id(car_folds$splits[[1]], c(TRUE, TRUE))
  )

  expect_error(
    analysis(car_folds$splits[[1]]) %>%
      add_resample_id(car_folds$splits)
  )

  expect_error(
    analysis(car_folds$splits[[1]]) %>%
      as.matrix() %>%
      add_resample_id(car_folds$splits[[1]])
  )
})

test_that("default param", {
  set.seed(11)
  rs1 <- permutations(mtcars, 1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == nrow(mtcars)))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, mtcars)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("apparent", {
  rs2 <- permutations(mtcars, 1, apparent = TRUE)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == nrow(mtcars)))
  expect_true(all(sizes2$assessment[nrow(sizes2)] == nrow(mtcars)))
  expect_equal(sizes2$assessment[sizes2$id == "Apparent"], nrow(mtcars))
})

test_that("no assessment set", {
  xx <- permutations(mtcars, 1)
  expect_snapshot(assessment(xx$splits[[1]]), error = TRUE)
})

test_that("bad args", {
  expect_error(permutations(mtcars)) # no columns specified
  expect_error(permutations(mtcars, foo)) # column doesn't exist
  expect_error(permutations(mtcars, start_with("z"))) # column doesn't exist
  expect_error(permutations(mtcars, everything())) # all columns
})

test_that("printing", {
  expect_snapshot(permutations(mtcars, 1))
})

test_that("rsplit labels", {
  rs <- permutations(mtcars, 1)
  all_labs <- purrr::map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("filtering/slicing rows", {
  x <- permutations(mtcars, 1:3)
  xf <- dplyr::filter(x, id == "Permutations01")
  xs <- dplyr::slice(x, 1)
  expect_identical(class(xf), c("tbl_df", "tbl", "data.frame"))
  expect_identical(class(xs), c("tbl_df", "tbl", "data.frame"))
})

test_that("column binding", {
  x <- permutations(mtcars, 1:3)
  xcb1 <- bind_cols(x, y = LETTERS[1:nrow(x)])
  xcb2 <- bind_cols(x, mtcars = tidyr::nest(mtcars, data = everything()))
  xcb3 <- bind_cols(y = LETTERS[1:nrow(x)], x)
  expect_identical(class(xcb1), class(x))
  expect_identical(class(xcb2), class(x))
  expect_false(identical(class(xcb3), class(x)))
})

test_that("can create a manual rset", {
  df <- data.frame(x = c(1, 2, 3, 4))

  indices <- list(
    list(analysis = 1L, assessment = 2L),
    list(analysis = 3L, assessment = 4L)
  )

  splits <- lapply(indices, make_splits, data = df)

  rset <- manual_rset(splits, c("Split 1", "Split 2"))

  expect_s3_class(rset, c("manual_rset", "rset"))
  expect_identical(rset$id, c("Split 1", "Split 2"))
})

test_that("can use analysis/assessment functions", {
  df <- data.frame(x = c(1, 2, 3))
  indices <- list(analysis = c(1L, 2L), assessment = 3L)
  splits <- list(make_splits(indices, df))

  rset <- manual_rset(splits, "Split 1")

  expect_identical(analysis(rset$splits[[1]]), df[1:2, 1, drop = FALSE])
  expect_identical(
    assessment(rset$splits[[1]]),
    df[3, 1, drop = FALSE],
    ignore_attr = "row.names"
  )
})

test_that("`pretty()` works", {
  df <- data.frame(x = c(1, 2, 3))
  indices <- list(analysis = c(1L, 2L), assessment = 3L)
  splits <- list(make_splits(indices, df))

  rset <- manual_rset(splits, "Split 1")

  expect_identical(pretty(rset), "Manual resampling")
})

test_that("basic cv", {
  cv_obj <- vfold_cv(mtcars)
  expect_equal(cv_obj$id, labels(cv_obj))
  expect_type(labels(cv_obj), "character")
  expect_s3_class(labels(cv_obj, TRUE), "factor")
})

test_that("repeated cv", {
  rcv_obj <- vfold_cv(mtcars, repeats = 3)
  expect_equal(
    paste(rcv_obj$id, rcv_obj$id2, sep = "."),
    labels(rcv_obj)
  )
  expect_type(labels(rcv_obj), "character")
  expect_s3_class(labels(rcv_obj, TRUE), "factor")
})

test_that("nested cv", {
  expect_snapshot(
    labels(
      nested_cv(mtcars,
        outside = vfold_cv(v = 3),
        inside = bootstraps(times = 5)
      )
    ),
    error = TRUE
  )
})

test_that("adding labels", {
  set.seed(363)
  car_folds <- vfold_cv(mtcars, repeats = 3)

  res <-
    analysis(car_folds$splits[[1]]) %>%
    add_resample_id(car_folds$splits[[1]])
  expect_equal(colnames(res), c(colnames(mtcars), "id", "id2"))

  car_bt <- bootstraps(mtcars)

  res <- analysis(car_bt$splits[[1]]) %>%
    add_resample_id(car_bt$splits[[1]])

  expect_equal(colnames(res), c(colnames(mtcars), "id"))

  res <- analysis(car_bt$splits[[1]]) %>%
    add_resample_id(car_bt$splits[[1]], TRUE)

  expect_equal(colnames(res), c(colnames(mtcars), ".id"))

  expect_error(
    analysis(car_folds$splits[[1]]) %>%
      add_resample_id(car_folds$splits[[1]], 7)
  )
  expect_error(
    analysis(car_folds$splits[[1]]) %>%
      add_resample_id(car_folds$splits[[1]], c(TRUE, TRUE))
  )

  expect_error(
    analysis(car_folds$splits[[1]]) %>%
      add_resample_id(car_folds$splits)
  )

  expect_error(
    analysis(car_folds$splits[[1]]) %>%
      as.matrix() %>%
      add_resample_id(car_folds$splits[[1]])
  )
})

test_that("bad args", {
  expect_error(
    new_rset(car_folds$splits[1:2], car_folds$id)
  )
  expect_error(
    new_rset(car_folds$splits, car_folds["splits"])
  )
  expect_error(
    new_rset(car_folds$splits, car_folds$splits)
  )
  expect_error(
    new_rset(list(1), "x"),
    "must be an `rsplit` object"
  )
  args <- list(a = 1, b = 2, 3)
  expect_error(
    new_rset(
      car_folds$splits,
      car_folds$id,
      attrib = args
    )
  )
})

test_that("rset with attributes", {
  args <- list(value = "potato")
  res3 <- new_rset(
    car_folds$splits,
    car_folds$id,
    attrib = args
  )
  expect_equal(
    sort(names(attributes(res3))),
    c("class", "fingerprint", "names", "row.names", "value")
  )
  expect_equal(attr(res3, "value"), "potato")
})

test_that("rset with additional classes", {
  res4 <- new_rset(
    car_folds$splits,
    car_folds$id,
    subclass = "potato"
  )
  expect_equal(
    class(res4),
    c("potato", "tbl_df", "tbl", "data.frame")
  )
})

test_that("not an rsplit", {
  folds <- vfold_cv(mtcars)
  expect_error(analysis(folds$splits[1]))
  expect_error(assessment(folds$splits[1]))
})

test_that("reverse_splits is working", {
  skip_if_not_installed("withr")

  reversable_subclasses <- setdiff(names(rset_subclasses), "permutations")
  reversable_subclasses <- rset_subclasses[reversable_subclasses]
  for (x in reversable_subclasses) {
    set.seed(123)
    rev_x <- reverse_splits(x)
    expect_identical(analysis(x$splits[[1]]), assessment(rev_x$splits[[1]]))
    expect_identical(assessment(x$splits[[1]]), analysis(rev_x$splits[[1]]))
    expect_identical(class(x), class(rev_x))
    expect_identical(class(x$splits[[1]]), class(rev_x$splits[[1]]))
  }

  expect_snapshot(
    reverse_splits(1),
    error = TRUE
  )

  permutes <- permutations(mtcars, cyl)

  expect_snapshot(
    reverse_splits(permutes),
    error = TRUE
  )

  expect_snapshot(
    reverse_splits(permutes$splits[[1]]),
    error = TRUE
  )
})

test_that("simple rsplit", {
  rs1 <- rsplit(dat1, 1:2, 4:5)
  expect_equal(rs1$data, dat1)
  expect_equal(rs1$in_id, 1:2)
  expect_equal(rs1$out_id, 4:5)
})

test_that("simple rsplit with matrices", {
  dat2 <- as.matrix(dat1)
  rs2 <- rsplit(dat2, 1:2, 4:5)
  expect_equal(rs2$data, dat2)
  expect_equal(rs2$in_id, 1:2)
  expect_equal(rs2$out_id, 4:5)
})

test_that("bad inputs", {
  expect_error(rsplit(as.list(dat1), 1:2, 4:5))
  expect_error(rsplit(dat1, letters[1:2], 4:5))
  expect_error(rsplit(as.list(dat1), 1:2, letters[4:5]))
  expect_error(rsplit(as.list(dat1), -1:2, 4:5))
  expect_error(rsplit(as.list(dat1), 1:2, -4:5))
  expect_error(rsplit(as.list(dat1), integer(0), 4:5))
})

test_that("as.data.frame", {
  rs3 <- rsplit(dat1, 1:2, 4:5)
  expect_equal(as.data.frame(rs3), dat1[1:2, ])
  expect_equal(as.data.frame(rs3, data = "assessment"), dat1[4:5, ], ignore_attr = "row.names")

  rs4 <- rsplit(dat1, rep(1:2, each = 3), rep(4:5, c(2, 1)))
  expect_equal(as.data.frame(rs4), dat1[c(1, 1, 1, 2, 2, 2), ], ignore_attr = "row.names")
  expect_equal(as.data.frame(rs4, data = "assessment"), dat1[c(4, 4, 5), ], ignore_attr = "row.names")
})


test_that("print methods", {
  expect_snapshot({
    set.seed(233)
    vfold_cv(mtcars)$splits[[1]]
  })

  skip_if_not_installed("withr")
  withr::local_options(lifecycle_verbosity = "quiet")

  expect_snapshot({
    set.seed(233)
    validation_split(mtcars)$splits[[1]]
  })
  expect_snapshot({
    set.seed(233)
    validation_split(mtcars)
  })
})

test_that("default complement method errors", {
  expect_snapshot(
    complement("a string"),
    error = TRUE
  )
})

test_that("as.data.frame() works for permutations with Surv object without the survival package loaded - issue #443", {
  srv <-
    list(
      age = c(74, 68, 56, 57, 60, 74, 76, 77, 39, 75, 66, 58),
      sex = c(1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 2),
      surv_obj = structure(
        c(306, 455, 1010, 210, 883, 1022, 116, 188,  191, 105, 174,  177,
          1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0),
        dim = c(12L, 2L),
        dimnames = list(NULL, c("time", "status")),
        type = "right",
        class = "Surv"))
  surv_df <-
    structure(
      srv,
      row.names = paste(1:12),
      class = "data.frame")

  set.seed(472)
  surv_permutation_df <- permutations(surv_df, permute = c(age, surv_obj), times = 1) %>%
    get_rsplit(1)

  expect_s3_class(surv_permutation_df$data$surv_obj, "Surv")
  expect_s3_class(analysis(surv_permutation_df)$surv_obj, "Surv")
})

test_that("default param", {
  set.seed(11)
  rs1 <- permutations(mtcars, 1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == nrow(mtcars)))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, mtcars)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("apparent", {
  rs2 <- permutations(mtcars, 1, apparent = TRUE)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == nrow(mtcars)))
  expect_true(all(sizes2$assessment[nrow(sizes2)] == nrow(mtcars)))
  expect_equal(sizes2$assessment[sizes2$id == "Apparent"], nrow(mtcars))
})

test_that("no assessment set", {
  xx <- permutations(mtcars, 1)
  expect_snapshot(assessment(xx$splits[[1]]), error = TRUE)
})

test_that("bad args", {
  expect_snapshot(error = TRUE, {
    permutations(mtcars)
  })
  expect_snapshot(error = TRUE, {
    permutations(mtcars, foo)
  })
  expect_snapshot(error = TRUE, {
    permutations(mtcars, start_with("z"))
  })
  expect_snapshot(error = TRUE, {
    permutations(mtcars, everything())
  })
})

test_that("printing", {
  expect_snapshot(permutations(mtcars, 1))
})

test_that("rsplit labels", {
  rs <- permutations(mtcars, 1)
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("filtering/slicing rows", {
  x <- permutations(mtcars, 1:3)
  xf <- dplyr::filter(x, id == "Permutations01")
  xs <- dplyr::slice(x, 1)
  expect_identical(class(xf), c("tbl_df", "tbl", "data.frame"))
  expect_identical(class(xs), c("tbl_df", "tbl", "data.frame"))
})

test_that("column binding", {
  x <- permutations(mtcars, 1:3)
  xcb1 <- bind_cols(x, y = LETTERS[1:nrow(x)])
  xcb2 <- bind_cols(x, mtcars = tidyr::nest(mtcars, data = everything()))
  xcb3 <- bind_cols(y = LETTERS[1:nrow(x)], x)
  expect_identical(class(xcb1), class(x))
  expect_identical(class(xcb2), class(x))
  expect_false(identical(class(xcb3), class(x)))
})

test_that("basic cv", {
  cv_obj <- vfold_cv(mtcars)
  expect_equal(cv_obj$id, labels(cv_obj))
  expect_type(labels(cv_obj), "character")
  expect_s3_class(labels(cv_obj, TRUE), "factor")
})

test_that("repeated cv", {
  rcv_obj <- vfold_cv(mtcars, repeats = 3)
  expect_equal(
    paste(rcv_obj$id, rcv_obj$id2, sep = "."),
    labels(rcv_obj)
  )
  expect_type(labels(rcv_obj), "character")
  expect_s3_class(labels(rcv_obj, TRUE), "factor")
})

test_that("nested cv", {
  expect_snapshot(
    labels(
      nested_cv(
        mtcars,
        outside = vfold_cv(v = 3),
        inside = bootstraps(times = 5)
      )
    ),
    error = TRUE
  )
})

test_that("adding labels", {
  set.seed(363)
  car_folds <- vfold_cv(mtcars, repeats = 3)

  res <-
    analysis(car_folds$splits[[1]]) |>
    add_resample_id(car_folds$splits[[1]])
  expect_equal(colnames(res), c(colnames(mtcars), "id", "id2"))

  car_bt <- bootstraps(mtcars)

  res <- analysis(car_bt$splits[[1]]) |>
    add_resample_id(car_bt$splits[[1]])

  expect_equal(colnames(res), c(colnames(mtcars), "id"))

  res <- analysis(car_bt$splits[[1]]) |>
    add_resample_id(car_bt$splits[[1]], TRUE)

  expect_equal(colnames(res), c(colnames(mtcars), ".id"))

  expect_snapshot(error = TRUE, {
    analysis(car_folds$splits[[1]]) |>
      add_resample_id(car_folds$splits[[1]], 7)
  })
  expect_snapshot(error = TRUE, {
    analysis(car_folds$splits[[1]]) |>
      add_resample_id(car_folds$splits[[1]], c(TRUE, TRUE))
  })

  expect_snapshot(error = TRUE, {
    analysis(car_folds$splits[[1]]) |>
      add_resample_id(car_folds$splits)
  })

  expect_snapshot(error = TRUE, {
    analysis(car_folds$splits[[1]]) |>
      as.matrix() |>
      add_resample_id(car_folds$splits[[1]])
  })
})

test_that("bad args", {
  expect_snapshot(error = TRUE, {
    new_rset(car_folds$splits[1:2], car_folds$id)
  })
  expect_snapshot(error = TRUE, {
    new_rset(car_folds$splits, car_folds["splits"])
  })
  expect_snapshot(error = TRUE, {
    new_rset(car_folds$splits, car_folds$splits)
  })
  expect_snapshot(error = TRUE, {
    new_rset(list(1), "x")
  })
  args <- list(a = 1, b = 2, 3)
  expect_snapshot(error = TRUE, {
    new_rset(
      car_folds$splits,
      car_folds$id,
      attrib = args
    )
  })
})

test_that("rset with attributes", {
  args <- list(value = "potato")
  res3 <- new_rset(
    car_folds$splits,
    car_folds$id,
    attrib = args
  )
  expect_equal(
    sort(names(attributes(res3))),
    c("class", "fingerprint", "names", "row.names", "value")
  )
  expect_equal(attr(res3, "value"), "potato")
})

test_that("rset with additional classes", {
  res4 <- new_rset(
    car_folds$splits,
    car_folds$id,
    subclass = "potato"
  )
  expect_equal(
    class(res4),
    c("potato", "tbl_df", "tbl", "data.frame")
  )
})

test_that("not an rsplit", {
  folds <- vfold_cv(mtcars)
  expect_snapshot(error = TRUE, {
    analysis(folds$splits[1])
  })
  expect_snapshot(error = TRUE, {
    assessment(folds$splits[1])
  })
})

# nocov start

.onLoad <- function(libname, pkgname) {
  vctrs::s3_register(
    "dplyr::dplyr_reconstruct",
    "rset",
    method = dplyr_reconstruct_rset
  )
}

release_questions <- function() {
  c(
    "Have you added all new classes to `rset_subclasses` (in R/compat-vctrs-helpers.R)?",
    "Have you re-generated vctrs compatibility functions (using inst/generate_vctrs.R) and document()ed?"
  )
}

# nocov end

#' @import vctrs
NULL

# ------------------------------------------------------------------------------

stop_incompatible_cast_rset <- function(x, to, ..., x_arg, to_arg) {
  details <- "Can't cast to an rset because attributes are likely incompatible."
  stop_incompatible_cast(
    x,
    to,
    x_arg = x_arg,
    to_arg = to_arg,
    details = details
  )
}

stop_never_called <- function(fn) {
  rlang::abort(paste0("Internal error: `", fn, "()` should never be called."))
}

#' @export
vec_ptype_abbr.initial_split <- function(x, ...) {
  "i_splt"
}

#' @export
vec_ptype_abbr.initial_time_split <- function(x, ...) {
  "i_t_splt"
}

#' @export
vec_ptype_abbr.group_initial_split <- function(x, ...) {
  "g_i_splt"
}

#' @export
vec_ptype_abbr.initial_validation_split <- function(x, ...) {
  "i_v_splt"
}

#' @export
vec_ptype_abbr.group_initial_validation_split <- function(x, ...) {
  "giv_splt"
}

#' @export
vec_ptype_abbr.validation_set <- function(x, ...) {
  "v_set"
}

#' @export
vec_ptype_abbr.validation_split <- function(x, ...) {
  "v_splt"
}

#' @export
vec_ptype_abbr.validation_time_split <- function(x, ...) {
  "v_t_splt"
}

#' @export
vec_ptype_abbr.group_validation_split <- function(x, ...) {
  "g_v_splt"
}

#' @export
vec_ptype_abbr.bootstraps <- function(x, ...) {
  "boot"
}

#' @export
vec_ptype_abbr.group_bootstraps <- function(x, ...) {
  "g_boot"
}

#' @export
vec_ptype_abbr.vfold_cv <- function(x, ...) {
  "vfold"
}

#' @export
vec_ptype_abbr.group_vfold_cv <- function(x, ...) {
  "g_vfold"
}

#' @export
vec_ptype_abbr.group_mc_cv <- function(x, ...) {
  "g_mc_cv"
}

#' @export
vec_ptype_abbr.clustering_cv <- function(x, ...) {
  "clt_fold"
}

#' @export
vec_ptype_abbr.rolling_origin <- function(x, ...) {
  "roll_ori"
}

#' @export
vec_ptype_abbr.sliding_window <- function(x, ...) {
  "slide_w"
}

#' @export
vec_ptype_abbr.sliding_index <- function(x, ...) {
  "slide_i"
}

#' @export
vec_ptype_abbr.sliding_period <- function(x, ...) {
  "slide_p"
}

#' @export
vec_ptype_abbr.nested_cv <- function(x, ...) {
  "nest_cv"
}

#' @export
vec_ptype_abbr.apparent <- function(x, ...) {
  "appr"
}

#' @export
vec_ptype_abbr.manual_rset <- function(x, ...) {
  "m_rset"
}

#' @export
vec_ptype_abbr.permutations <- function(x, ...) {
  "perm"
}

#' Constructors for split objects
#' @export
#' @examples
#' df <- data.frame(
#'   year = 1900:1999,
#'   value = 10 + 8*1900:1999 + runif(100L, 0, 100)
#' )
#' split_from_indices <- make_splits(
#'   x = list(analysis = which(df$year <= 1980),
#'            assessment = which(df$year > 1980)),
#'   data = df
#' )
#' split_from_data_frame <- make_splits(
#'   x = df[df$year <= 1980,],
#'   assessment = df[df$year > 1980,]
#' )
#' identical(split_from_indices, split_from_data_frame)
make_splits <- function(x, ...) {
  UseMethod("make_splits")
}

#' @rdname make_splits
#' @param x A list of integers with names "analysis" and "assessment", or a
#' data frame of analysis or training data.
#' @export
make_splits.default <- function(x, ...) {
  cls <- class(x)
  cli_abort("No method for objects of class{?es}: {cls}")
}

#' @rdname make_splits
#' @param data A data frame.
#' @param class An optional class to give the object.
#' @param ... Not currently used.
#' @export
make_splits.list <- function(x, data, class = NULL, ...) {
  rlang::check_dots_empty()
  res <- rsplit(data, x$analysis, x$assessment)
  if (!is.null(class)) {
    res <- add_class(res, class)
  }
  res
}

#' @rdname make_splits
#' @param assessment A data frame of assessment or testing data, which can be empty.
#' @export
make_splits.data.frame <- function(x, assessment, class = NULL, ...) {
  rlang::check_dots_empty()
  if (nrow(x) == 0) {
    cli_abort("The analysis set must contain at least one row.")
  }

  ind_analysis <- seq_len(nrow(x))
  if (nrow(assessment) == 0) {
    ind_assessment <- integer()
  } else {
    if (!identical(colnames(x), colnames(assessment))) {
      cli_abort("The analysis and assessment sets must have the same columns.")
    }
    ind_assessment <- nrow(x) + seq_len(nrow(assessment))
  }

  data <- bind_rows(x, assessment)
  ind <- list(
    analysis = ind_analysis,
    assessment = ind_assessment
  )

  make_splits(ind, data, class = class)
}

merge_lists <- function(a, b) list(analysis = a, assessment = b)

dim_rset <- function(x, ...) {
  check_dots_empty()
  dims <- purrr::map(x$splits, dim)
  dims <- do.call("rbind", dims)
  dims <- tibble::as_tibble(dims)
  id_cols <- grep("(^id$)|(^id[1-9]$)", colnames(x), value = TRUE)
  for (i in seq_along(id_cols)) {
    dims[id_cols[i]] <- getElement(x, id_cols[i])
  }
  dims
}

names0 <- function(num, prefix = "x") {
  if (num == 0L) {
    return(character())
  }
  ind <- format(1:num)
  ind <- gsub(" ", "0", ind)
  paste0(prefix, ind)
}

add_class <- function(x, cls) {
  class(x) <- c(cls, class(x))
  x
}

check_prop <- function(prop, call = caller_env()) {
  check_number_decimal(prop, call = call)
  if (!(prop > 0)) {
    cli_abort("{.arg prop} must be greater than 0.", call = call)
  }
  if (!(prop < 1)) {
    cli_abort("{.arg prop} must be less than 1.", call = call)
  }
  invisible(NULL)
}

check_strata <- function(strata, data, call = caller_env()) {
  check_string(strata, allow_null = TRUE, call = call)

  if (!is.null(strata)) {
    if (inherits(data[, strata], "Surv")) {
      cli_abort(
        c(
          "{.field strata} cannot be a {.cls Surv} object.",
          "i" = "Use the time or event variable directly."
        ),
        call = call
      )
    }
  }
  invisible(NULL)
}

split_unnamed <- function(x, f) {
  out <- split(x, f)
  unname(out)
}

#' Get the split arguments from an rset
#' @param x An `rset` or `initial_split` object.
#' @param allow_strata_false A logical to specify which value to use if no
#' stratification was specified. The default is to use `strata = NULL`, the
#' alternative is `strata = FALSE`.
#' @return A list of arguments used to create the rset.
#' @keywords internal
#' @export
.get_split_args <- function(x, allow_strata_false = FALSE) {
  all_attributes <- attributes(x)
  function_name_used_to_create <- switch(
    all_attributes$class[[1]],
    "validation_set" = "initial_validation_split",
    "group_validation_set" = "group_initial_validation_split",
    "time_validation_set" = "initial_validation_time_split",
    all_attributes$class[[1]]
  )
  function_used_to_create <- .find_resampling_function(
    function_name_used_to_create
  )
  args <- names(formals(function_used_to_create))

  if ("initial_validation_split" %in% all_attributes$class) {
    split_args <- all_attributes$val_att
  } else {
    split_args <- all_attributes[args]
  }
  split_args <- split_args[!is.na(names(split_args))]

  if (identical(split_args$strata, FALSE) && !allow_strata_false) {
    split_args$strata <- NULL
  }
  split_args
}

.find_resampling_function <- function(fn_name, ..., call = caller_env()) {
  rlang::check_dots_empty()

  # looks in rsample, the global environment, and search path
  fn <- tryCatch(
    get(fn_name, mode = "function"),
    error = function(e) NULL
  )
  if (!is.null(fn)) {
    return(fn)
  }

  # looks in spatialsample, even when it's not attached/in the search path
  fn <- tryCatch(
    getExportedValue("spatialsample", fn_name),
    error = function(e) NULL
  )
  if (!is.null(fn)) {
    return(fn)
  }

  cli_abort("Could not find function {.fn {fn_name}}.", call = call)
}

#' Retrieve individual rsplits objects from an rset
#'
#' @param x The `rset` object to retrieve an rsplit from.
#' @param index An integer indicating which rsplit to retrieve: `1` for the
#' rsplit in the first row of the rset, `2` for the second, and so on.
#' @param ... Not currently used.
#'
#' @return The rsplit object in row `index` of `rset`
#'
#' @examples
#' set.seed(123)
#' (starting_splits <- group_vfold_cv(mtcars, cyl, v = 3))
#' get_rsplit(starting_splits, 1)
#'
#' @rdname get_rsplit
#' @export
get_rsplit <- function(x, index, ...) {
  UseMethod("get_rsplit")
}

#' @rdname get_rsplit
#' @export
get_rsplit.rset <- function(x, index, ...) {
  rlang::check_dots_empty()

  n_rows <- nrow(x)

  acceptable_index <- length(index) == 1 &&
    rlang::is_integerish(index) &&
    index > 0 &&
    index <= n_rows

  if (!acceptable_index) {
    msg <- ifelse(
      length(index) != 1,
      glue::glue("Index was of length {length(index)}."),
      glue::glue("A value of {index} was provided.")
    )

    cli_abort(
      c(
        "{.arg index} must be a length-1 integer between 1 and {n_rows}.",
        "*" = msg
      )
    )
  }

  x$splits[[index]]
}

#' @rdname get_rsplit
#' @export
get_rsplit.default <- function(x, index, ...) {
  cls <- class(x)
  cli_abort("No method for objects of class{?es}: {cls}")
}

context("Rolling window resampling")

library(testthat)
library(rsample)
library(purrr)

dat1 <- data.frame(a = 1:20, b = letters[1:20])

test_that('default param', {
  rs1 <- rolling_origin(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$assessment == 1))
  expect_true(all(sizes1$analysis == 5:19))
  same_data <-
    map_lgl(rs1$splits, function(x)
      all.equal(x$data, dat1))
  expect_true(all(same_data))

  for (i in 1:nrow(rs1)) {
    expect_equal(rs1$splits[[i]]$in_id,
                 1:(i + attr(rs1, "initial") - 1))
    expect_equal(rs1$splits[[i]]$out_id,
                 i + attr(rs1, "initial"))
  }

})

test_that('larger holdout', {
  rs2 <- rolling_origin(dat1, assess = 3)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$assessment == 3))
  expect_true(all(sizes2$analysis == 5:17))

  for (i in 1:nrow(rs2)) {
    expect_equal(rs2$splits[[i]]$in_id,
                 1:(i + attr(rs2, "initial") - 1))
    expect_equal(rs2$splits[[i]]$out_id,
                 (i + attr(rs2, "initial")):
                   (i + attr(rs2, "initial") + attr(rs2, "assess") - 1))
  }

})

test_that('fixed analysis size', {
  rs3 <- rolling_origin(dat1, cumulative = FALSE)
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$assessment == 1))
  expect_true(all(sizes3$analysis == 5))

  for (i in 1:nrow(rs3)) {
    expect_equal(rs3$splits[[i]]$in_id,
                 i:(i + attr(rs3, "initial") - 1))
    expect_equal(rs3$splits[[i]]$out_id,
                 i + attr(rs3, "initial"))
  }

})


test_that('skipping', {
  rs4 <- rolling_origin(dat1, cumulative = FALSE, skip = 2)
  sizes4 <- dim_rset(rs4)

  expect_true(all(sizes4$assessment == 1))
  expect_true(all(sizes4$analysis == 5))

  for (i in 1:nrow(rs4)) {
    expect_equal(rs4$splits[[i]]$in_id,
                 (i + attr(rs4, "skip")*(i-1)):
                   (i + attr(rs4, "skip")*(i-1) + attr(rs4, "initial") -1))
    expect_equal(rs4$splits[[i]]$out_id,
                 i + attr(rs4, "skip")*(i-1) + attr(rs4, "initial"))
  }

})

test_that('lag', {
  rs5 <- rolling_origin(dat1, initial = 5, assess = 1, cumulative = FALSE, skip = 0, lag = 3)
  sizes5 <- dim_rset(rs5)

  expect_true(all(sizes5$assessment == attr(rs5, "assess") + attr(rs5, "lag")))
  expect_true(all(sizes5$analysis == attr(rs5, "initial")))

  for (i in 1:nrow(rs5)) {
    expect_equal(rs5$splits[[i]]$in_id,
                 i:(i + attr(rs5, "initial") - 1))
    expect_equal(rs5$splits[[i]]$out_id,
                 (i + attr(rs5, "initial") - attr(rs5, "lag")):(i + attr(rs5, "initial") + attr(rs5, "assess") - 1))
  }

  expect_error(rolling_origin(drinks, initial = 5, lag = 6)) # lag must be less than training observations
  expect_error(olling_origin(drinks, lag = 2.1))             # lag must be whole number
})

test_that('rsplit labels', {
  rs <- rolling_origin(dat1)
  all_labs <- map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("default param", {
  set.seed(11)
  rs1 <- clustering_cv(dat1, c, v = 2)

  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("repeated", {
  set.seed(11)
  rs2 <- clustering_cv(dat1, c, v = 2, repeats = 4)
  sizes2 <- dim_rset(rs2)

  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("bad args", {
  expect_snapshot(error = TRUE, {
    clustering_cv(dat1)
  })
  expect_snapshot(error = TRUE, {
    clustering_cv(iris, Sepal.Length, v = -500)
  })
  expect_snapshot(error = TRUE, {
    clustering_cv(iris, Sepal.Length, v = 500)
  })
  expect_snapshot(error = TRUE, {
    clustering_cv(iris, Sepal.Length, cluster_function = "not an option")
  })
  expect_snapshot(error = TRUE, {
    clustering_cv(Orange, v = 1, vars = "Tree")
  })
  expect_snapshot(error = TRUE, {
    clustering_cv(Orange, repeats = 0)
  })
  expect_snapshot(error = TRUE, {
    clustering_cv(Orange, repeats = NULL)
  })
  expect_snapshot(error = TRUE, {
    clustering_cv(mtcars, mpg, v = nrow(mtcars))
  })
})

test_that("printing", {
  set.seed(11)
  expect_snapshot(clustering_cv(dat1, c, v = 2))
})

test_that("rsplit labels", {
  rs <- clustering_cv(dat1, c, v = 2)
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

context("Tidy methods")

library(testthat)
library(rsample)
library(purrr)

check_ind <- function(x, tdat) {
  in_dat <- subset(tdat, Data == "Analysis")
  in_check <- all(sort(in_dat$Row) == x$in_ind)
  out_dat <- subset(tdat, Data == "Analysis")
  out_check <- all(sort(out_dat$Row) == x$out_ind)
  in_check & out_check
}

dat1 <- data.frame(a = 1:20, b = letters[1:20])

test_that('simple boot', {
  set.seed(11)
  rs1 <- bootstraps(dat1)
  td1 <- tidy(rs1, unique_ind = FALSE)
  expect_gt(nrow(td1), nrow(dat1))

  name_vals <- names0(nrow(rs1), "Bootstrap")
  for(i in 1:nrow(rs1)) {
    expect_true(
      check_ind(rs1$splits[[i]],
                subset(td1, Resample == name_vals[i])
                )
      )
  }
})


test_that('vfold', {
  set.seed(11)
  rs2 <- vfold_cv(dat1)
  td2 <- tidy(rs2, unique_ind = FALSE)

  for(i in 1:nrow(rs2)) {
    expect_true(
      check_ind(rs2$splits[[i]],
                subset(td2, Fold == rs2$id[i])
      )
    )
  }
})

test_that('vfold with repeats', {
  set.seed(11)
  rs3 <- vfold_cv(dat1, repeats = 2)
  td3 <- tidy(rs3, unique_ind = FALSE)

  for(i in 1:nrow(rs3)) {
    expect_true(
      check_ind(rs3$splits[[i]],
                subset(td3, Fold == rs3$id2[i] & Repeat == rs3$id[i])
      )
    )
  }
})

test_that("can create a split with an empty assessment set (#188)", {
  df <- data.frame(x = c(1, 2, 3, 4))
  indices <- list(analysis = 1:4, assessment = integer())

  split <- make_splits(indices, df)

  expect_identical(split$out_id, integer())
  expect_identical(assessment(split), df[0, , drop = FALSE])
})

test_that("cannot create a split with an empty analysis set", {
  df <- data.frame(x = c(1, 2, 3, 4))
  indices <- list(analysis = integer(), assessment = 1:4)

  expect_snapshot(error = TRUE, {
    make_splits(indices, df)
  })
})

test_that("create a split from training and testing dataframes", {
  training <- tibble(x = c(1, 2, 3, 4))
  testing <- tibble(x = c(5, 6))

  split <- make_splits(training, testing)
  expect_identical(analysis(split), training)
  expect_identical(assessment(split), testing)
})

test_that("can create a split from empty testing dataframe", {
  training <- tibble(x = c(1, 2, 3, 4))
  testing <- tibble()

  split <- make_splits(training, testing)
  expect_identical(split$out_id, integer())
  expect_identical(analysis(split), training)
})

test_that("cannot create a split from empty training dataframe", {
  training <- tibble()
  testing <- tibble(x = c(5, 6))

  expect_snapshot(error = TRUE, {
    make_splits(training, testing)
  })
})

test_that("cannot create a split from dataframes with different columns", {
  training <- tibble(x = c(1, 2, 3, 4))
  testing <- tibble(y = c(5, 6))

  expect_snapshot(error = TRUE, {
    make_splits(training, testing)
  })
})

test_that("improper argument", {
  expect_snapshot(error = TRUE, {
    make_splits("potato")
  })
})

test_that("basic naming sequences", {
  expect_equal(names0(2), c("x1", "x2"))
  expect_equal(names0(2, "y"), c("y1", "y2"))
  expect_equal(
    names0(10),
    c(paste0("x0", 1:9), "x10")
  )
})

test_that("get_rsplit()", {
  skip_if_not_installed("withr")
  # for `validation_split()` and variants
  withr::local_options(lifecycle_verbosity = "quiet")

  val <- withr::with_seed(
    11,
    validation_split(warpbreaks)
  )

  expect_identical(val$splits[[1]], get_rsplit(val, 1))

  expect_snapshot(error = TRUE, {
    get_rsplit(val, 3)
  })

  expect_snapshot(error = TRUE, {
    get_rsplit(val, c(1, 2))
  })

  expect_snapshot(error = TRUE, {
    get_rsplit(val, 1.5)
  })

  expect_snapshot(error = TRUE, {
    get_rsplit(warpbreaks, 1)
  })
})

test_that("simple boot", {
  set.seed(11)
  rs1 <- bootstraps(dat1)
  td1 <- tidy(rs1, unique_ind = FALSE)
  expect_gt(nrow(td1), nrow(dat1))

  name_vals <- names0(nrow(rs1), "Bootstrap")
  for (i in 1:nrow(rs1)) {
    expect_true(
      check_ind(
        rs1$splits[[i]],
        subset(td1, Resample == name_vals[i])
      )
    )
  }
})


test_that("vfold", {
  set.seed(11)
  rs2 <- vfold_cv(dat1)
  td2 <- tidy(rs2)

  for (i in 1:nrow(rs2)) {
    expect_true(
      check_ind(
        rs2$splits[[i]],
        subset(td2, Fold == rs2$id[i])
      )
    )
  }
})

test_that("vfold with repeats", {
  set.seed(11)
  rs3 <- vfold_cv(dat1, repeats = 2)
  td3 <- tidy(rs3)

  for (i in 1:nrow(rs3)) {
    expect_true(
      check_ind(
        rs3$splits[[i]],
        subset(td3, Fold == rs3$id2[i] & Repeat == rs3$id[i])
      )
    )
  }
})

test_that("bad args", {
  expect_error(group_vfold_cv(warpbreaks, group = warpbreaks$tension))
  expect_error(group_vfold_cv(warpbreaks, group = c("tension", "wool")))
  expect_error(group_vfold_cv(warpbreaks, group = "tensio"))
  expect_error(group_vfold_cv(warpbreaks))
  expect_error(group_vfold_cv(warpbreaks, group = "tension", v = 10))
})


test_that("default param", {
  set.seed(11)
  rs1 <- group_vfold_cv(warpbreaks, "tension")
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 36))
  expect_true(all(sizes1$assessment == 18))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, warpbreaks)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  sp_out <- purrr::map_chr(rs1$splits, get_id_left_out)
  expect_true(all(table(sp_out) == 1))
})


test_that("v < max v", {
  set.seed(11)
  rs2 <- group_vfold_cv(warpbreaks, "tension", v = 2)
  sizes2 <- dim_rset(rs2)

  expect_true(!all(sizes2$analysis == 36))
  expect_true(!all(sizes2$assessment == 18))
  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, warpbreaks)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  sp_out <- purrr::map(rs2$splits, get_id_left_out)
  expect_true(all(table(unlist(sp_out)) == 1))
})

test_that("tibble input", {
  warpbreaks2 <- tibble::as_tibble(warpbreaks)
  set.seed(11)
  rs3 <- group_vfold_cv(warpbreaks2, "tension")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis == 36))
  expect_true(all(sizes3$assessment == 18))
  same_data <-
    purrr::map_lgl(rs3$splits, function(x) {
      all.equal(x$data, warpbreaks2)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  sp_out <- purrr::map_chr(rs3$splits, get_id_left_out)
  expect_true(all(table(sp_out) == 1))
})

test_that("printing", {
  expect_snapshot(group_vfold_cv(warpbreaks, "tension"))
})

test_that("printing with ...", {
  expect_snapshot(
    print(group_vfold_cv(warpbreaks, "tension"), n = 2)
  )
})

test_that("rsplit labels", {
  rs <- group_vfold_cv(warpbreaks, "tension")
  all_labs <- purrr::map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

context("Naming functions")

library(testthat)
library(rsample)

test_that('basic naming sequences', {
  expect_equal(names0(2), c("x1", "x2"))
  expect_equal(names0(2, "y"), c("y1", "y2"))
  expect_equal(names0(10),
               c(paste0("x0", 1:9), "x10"))
})

test_that("default param", {
  set.seed(11)
  rs1 <- vfold_cv(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 18))
  expect_true(all(sizes1$assessment == 2))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("repeated", {
  set.seed(11)
  rs2 <- vfold_cv(dat1, repeats = 4)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == 18))
  expect_true(all(sizes2$assessment == 2))
  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("strata", {
  set.seed(11)
  data("mlc_churn", package = "modeldata")
  rs3 <- vfold_cv(mlc_churn, repeats = 2, strata = "voice_mail_plan")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis %in% 4499:4501))
  expect_true(all(sizes3$assessment %in% 499:501))

  rate <- purrr::map_dbl(
    rs3$splits,
    function(x) {
      dat <- as.data.frame(x)$voice_mail_plan
      mean(dat == "yes")
    }
  )
  expect_equal(mean(unique(rate)), 0.2645925848)

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  expect_snapshot(
    rs4 <- vfold_cv(mlc_churn, strata = state, pool = 0.01)
  )
})


test_that("bad args", {
  expect_error(vfold_cv(iris, strata = iris$Species))
  expect_error(vfold_cv(iris, strata = c("Species", "Sepal.Width")))
  expect_snapshot_error(vfold_cv(iris, v = -500))
  expect_snapshot_error(vfold_cv(iris, v = 500))
})

test_that("printing", {
  expect_snapshot(vfold_cv(mtcars))
})


test_that("rsplit labels", {
  rs <- vfold_cv(mtcars)
  all_labs <- purrr::map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)

  rs2 <- vfold_cv(mtcars, repeats = 4)
  all_labs2 <- purrr::map_df(rs2$splits, labels)
  original_id2 <- rs2[, grepl("^id", names(rs2))]
  expect_equal(all_labs2, original_id2)
})

context("V-fold CV")

library(testthat)
library(rsample)
library(purrr)
library(modeldata)

dat1 <- data.frame(a = 1:20, b = letters[1:20])

test_that('default param', {
  set.seed(11)
  rs1 <- vfold_cv(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 18))
  expect_true(all(sizes1$assessment == 2))
  same_data <-
    map_lgl(rs1$splits, function(x)
      all.equal(x$data, dat1))
  expect_true(all(same_data))

  good_holdout <- map_lgl(rs1$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))
})

test_that('repeated', {
  set.seed(11)
  rs2 <- vfold_cv(dat1, repeats = 4)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == 18))
  expect_true(all(sizes2$assessment == 2))
  same_data <-
    map_lgl(rs2$splits, function(x)
      all.equal(x$data, dat1))
  expect_true(all(same_data))

  good_holdout <- map_lgl(rs2$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))
})

test_that('strata', {
  set.seed(11)
  data("mlc_churn", package = "modeldata")
  rs3 <- vfold_cv(mlc_churn, repeats = 2, strata = "voice_mail_plan")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis %in% 4499:4501))
  expect_true(all(sizes3$assessment %in% 499:501))

  rate <- map_dbl(rs3$splits,
                  function(x) {
                    dat <- as.data.frame(x)$voice_mail_plan
                    mean(dat == "yes")
                  })
  expect_equal(mean(unique(rate)), 0.2645925848)

  good_holdout <- map_lgl(rs3$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))
})


test_that('bad args', {
  expect_error(vfold_cv(iris, strata = iris$Species))
  expect_error(vfold_cv(iris, strata = c("Species", "Sepal.Width")))
})

test_that('printing', {
  expect_output(print(vfold_cv(mtcars)))
})


test_that('rsplit labels', {
  rs <- vfold_cv(mtcars)
  all_labs <- map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)

  rs2 <- vfold_cv(mtcars, repeats = 4)
  all_labs2 <- map_df(rs2$splits, labels)
  original_id2 <- rs2[, grepl("^id", names(rs2))]
  expect_equal(all_labs2, original_id2)
})

test_that("simple numerics", {
  set.seed(13333)
  x1 <- rnorm(1000)
  str1a <- make_strata(x1)
  tab1a <- table(str1a)
  expect_equal(as.vector(tab1a), rep(250, 4))

  expect_snapshot(str1b <- make_strata(x1, depth = 500))
  tab1b <- table(str1b)
  expect_equal(as.vector(tab1b), rep(500, 2))

  str1c <- make_strata(c(NA, x1[1:999]))
  tab1c <- table(str1c)
  expect_true(all(as.vector(tab1c) %in% 249:251))
})

test_that("simple character", {
  x2 <- factor(rep(LETTERS[1:12], each = 20))
  expect_snapshot({
    str2a <- make_strata(x2, pool = 0.05)
  })
  expect_equal(table(str2a, dnn = ""), table(x2, dnn = ""))

  x2[5] <- NA
  expect_snapshot({
    str2b <- make_strata(x2, pool = 0.05)
  })
  expect_true(all(as.vector(table(str2b, dnn = "")) %in% 19:21))
})

test_that("bad data", {
  x3 <- factor(rep(LETTERS[1:15], each = 50))
  expect_snapshot(s0 <- make_strata(x3))
  expect_snapshot(s1 <- make_strata(x3, pool = 0.06))
  expect_snapshot(s2 <- make_strata(mtcars$mpg))
  expect_snapshot(s3 <- make_strata(seq_len(50), breaks = -1))
})


# check_strata() ----------------------------------------------------------

test_that("don't stratify on Surv objects", {
  df <- data.frame(
    time = c(85, 79, 70, 6, 32, 8, 17, 93, 81, 76),
    event = c(0, 0, 1, 0, 0, 0, 1, 1, 1, 1)
  )
  df$surv <- structure(
    c(
      85,
      79,
      70,
      6,
      32,
      8,
      17,
      93,
      81,
      76,
      0,
      0,
      1,
      0,
      0,
      0,
      1,
      1,
      1,
      1
    ),
    .Dim = c(10L, 2L),
    .Dimnames = list(NULL, c("time", "status")),
    type = "right",
    class = "Surv"
  )

  expect_snapshot(error = TRUE, {
    check_strata("surv", df)
  })
})

context("Bootstrapping")

library(testthat)
library(rsample)
library(purrr)

dat1 <- data.frame(a = 1:20, b = letters[1:20])

test_that('default param', {
  set.seed(11)
  rs1 <- bootstraps(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == nrow(dat1)))
  same_data <-
    map_lgl(rs1$splits, function(x)
      all.equal(x$data, dat1))
  expect_true(all(same_data))

  good_holdout <- map_lgl(rs1$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))
})

test_that('apparent', {
  rs2 <- bootstraps(dat1, apparent = TRUE)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == nrow(dat1)))
  expect_true(all(sizes2$assessment[nrow(sizes2)] == nrow(dat1)))
  expect_equal(sizes2$assessment[sizes2$id == "Apparent"], nrow(dat1))
  res2 <-
    as.data.frame(rs2$splits[[nrow(sizes2)]], data = "assessment")
  expect_equal(res2, dat1)
})


test_that('strata', {
  set.seed(11)
  rs4 <- bootstraps(warpbreaks,  strata = "tension")
  sizes4 <- dim_rset(rs4)

  expect_true(all(sizes4$analysis == nrow(warpbreaks)))

  rate <- map_dbl(rs4$splits,
                  function(x) {
                    dat <- as.data.frame(x)$tension
                    mean(dat == "M")
                  })
  expect_true(length(unique(rate)) == 1)

  good_holdout <- map_lgl(rs4$splits,
                          function(x) {
                            length(intersect(x$in_ind, x$out_id)) == 0
                          })
  expect_true(all(good_holdout))

  rs5 <- bootstraps(warpbreaks, apparent = TRUE, strata = "tension")
  sizes5 <- dim_rset(rs5)

  expect_true(all(sizes5$analysis == nrow(warpbreaks)))
  expect_true(all(sizes5$assessment[nrow(sizes5)] == nrow(warpbreaks)))
  expect_equal(sizes5$assessment[sizes5$id == "Apparent"], nrow(warpbreaks))
  res5 <-
    as.data.frame(rs5$splits[[nrow(sizes5)]], data = "assessment")
  expect_equal(res5, warpbreaks)

})


test_that('bad args', {
  expect_error(bootstraps(warpbreaks, strata = warpbreaks$tension))
  expect_error(bootstraps(warpbreaks, strata = c("tension", "wool")))
})


test_that('printing', {
  expect_output(print(bootstraps(warpbreaks)))
})


test_that('rsplit labels', {
  rs <- bootstraps(warpbreaks)
  all_labs <- map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("default param", {
  rs1 <- rolling_origin(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$assessment == 1))
  expect_true(all(sizes1$analysis == 5:19))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  for (i in 1:nrow(rs1)) {
    expect_equal(
      rs1$splits[[i]]$in_id,
      1:(i + attr(rs1, "initial") - 1)
    )
    expect_equal(
      rs1$splits[[i]]$out_id,
      i + attr(rs1, "initial")
    )
  }
})

test_that("larger holdout", {
  rs2 <- rolling_origin(dat1, assess = 3)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$assessment == 3))
  expect_true(all(sizes2$analysis == 5:17))

  for (i in 1:nrow(rs2)) {
    expect_equal(
      rs2$splits[[i]]$in_id,
      1:(i + attr(rs2, "initial") - 1)
    )
    expect_equal(
      rs2$splits[[i]]$out_id,
      # fmt: skip
      (i + attr(rs2, "initial")):
      (i + attr(rs2, "initial") + attr(rs2, "assess") - 1)
      # fmt: skip
    )
  }
})

test_that("fixed analysis size", {
  rs3 <- rolling_origin(dat1, cumulative = FALSE)
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$assessment == 1))
  expect_true(all(sizes3$analysis == 5))

  for (i in 1:nrow(rs3)) {
    expect_equal(
      rs3$splits[[i]]$in_id,
      i:(i + attr(rs3, "initial") - 1)
    )
    expect_equal(
      rs3$splits[[i]]$out_id,
      i + attr(rs3, "initial")
    )
  }
})


test_that("skipping", {
  rs4 <- rolling_origin(dat1, cumulative = FALSE, skip = 2)
  sizes4 <- dim_rset(rs4)

  expect_true(all(sizes4$assessment == 1))
  expect_true(all(sizes4$analysis == 5))

  for (i in 1:nrow(rs4)) {
    expect_equal(
      rs4$splits[[i]]$in_id,
      # fmt: skip
      (i + attr(rs4, "skip") * (i - 1)):
      (i + attr(rs4, "skip") * (i - 1) + attr(rs4, "initial") - 1)
      # fmt: skip
    )
    expect_equal(
      rs4$splits[[i]]$out_id,
      i + attr(rs4, "skip") * (i - 1) + attr(rs4, "initial")
    )
  }
})

test_that("lag", {
  rs5 <- rolling_origin(
    dat1,
    initial = 5,
    assess = 1,
    cumulative = FALSE,
    skip = 0,
    lag = 3
  )
  sizes5 <- dim_rset(rs5)

  expect_true(all(sizes5$assessment == attr(rs5, "assess") + attr(rs5, "lag")))
  expect_true(all(sizes5$analysis == attr(rs5, "initial")))

  for (i in 1:nrow(rs5)) {
    expect_equal(
      rs5$splits[[i]]$in_id,
      i:(i + attr(rs5, "initial") - 1)
    )
    expect_equal(
      rs5$splits[[i]]$out_id,
      # fmt: skip
      (i + attr(rs5, "initial") - attr(rs5, "lag")):
      (i + attr(rs5, "initial") + attr(rs5, "assess") - 1)
      # fmt: skip
    )
  }

  skip_if_not_installed("modeldata")
  data("drinks", package = "modeldata", envir = rlang::current_env())
  expect_snapshot(error = TRUE, {
    # lag must be less than the number of training observations
    rolling_origin(drinks, initial = 5, lag = 6)
  })
  expect_snapshot(error = TRUE, {
    # lag must be a whole number
    rolling_origin(drinks, lag = 2.1)
  })
})

test_that("rsplit labels", {
  rs <- rolling_origin(dat1)
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

#' Find Labels from rset Object
#'
#' Produce a vector of resampling labels (e.g. "Fold1") from
#'  an `rset` object. Currently, `nested_cv`
#'  is not supported.
#'
#' @param object An `rset` object
#' @param make_factor A logical for whether the results should be
#'  a character or a factor.
#' @param ... Not currently used.
#' @return A single character or factor vector.
#' @export
#' @examples
#' labels(vfold_cv(mtcars))
labels.rset <- function(object, make_factor = FALSE, ...) {
  if (inherits(object, "nested_cv"))
    stop("`labels` not implemented for nested resampling",
         call. = FALSE)
  if (make_factor)
    as.factor(object$id)
  else
    as.character(object$id)
}

#' @rdname labels.rset
#' @export
labels.vfold_cv <- function(object, make_factor = FALSE, ...) {
  if (inherits(object, "nested_cv"))
    stop("`labels` not implemented for nested resampling",
         call. = FALSE)
  is_repeated <- attr(object, "repeats") > 1
  if (is_repeated) {
    out <- as.character(paste(object$id, object$id2, sep = "."))
  } else
    out <- as.character(object$id)
  if (make_factor)
    out <- as.factor(out)
  out
}

#' Find Labels from rsplit Object
#'
#' Produce a tibble of identification variables so that single
#'  splits can be linked to a particular resample.
#'
#' @param object An `rsplit` object
#' @param ... Not currently used.
#' @return A tibble.
#' @seealso add_resample_id
#' @export
#' @examples
#' cv_splits <- vfold_cv(mtcars)
#' labels(cv_splits$splits[[1]])
labels.rsplit <- function(object, ...) {
  out <- if ("id" %in% names(object))
    object$id
  else
    tibble()
  out
}

## The `pretty` methods below are good for when you need to
## textually describe the resampling procedure. Note that they
## can have more than one element (in the case of nesting)


#' Short Descriptions of rsets
#'
#' Produce a character vector describing the resampling method.
#'
#' @param x An `rset` object
#' @param ... Not currently used.
#' @return A character vector.
#' @export pretty.vfold_cv
#' @export
#' @method pretty vfold_cv
#' @keywords internal
pretty.vfold_cv <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(details$v, "-fold cross-validation")
  if (details$repeats > 1)
    res <- paste(res, "repeated", details$repeats, "times")
  if (details$strata)
    res <- paste(res, "using stratification")
  res
}

#' @export pretty.loo_cv
#' @export
#' @method pretty loo_cv
#' @rdname pretty.vfold_cv
pretty.loo_cv <- function(x, ...)
  "Leave-one-out cross-validation"

#' @export pretty.apparent
#' @export
#' @method pretty apparent
#' @rdname pretty.vfold_cv
pretty.apparent <- function(x, ...)
  "Apparent sampling"

#' @export pretty.rolling_origin
#' @export
#' @method pretty rolling_origin
#' @rdname pretty.vfold_cv
pretty.rolling_origin <- function(x, ...)
  "Rolling origin forecast resampling"

#' @export pretty.sliding_window
#' @export
#' @method pretty sliding_window
#' @rdname pretty.vfold_cv
pretty.sliding_window <- function(x, ...)
  "Sliding window resampling"

#' @export pretty.sliding_index
#' @export
#' @method pretty sliding_index
#' @rdname pretty.vfold_cv
pretty.sliding_index <- function(x, ...)
  "Sliding index resampling"

#' @export pretty.sliding_period
#' @export
#' @method pretty sliding_period
#' @rdname pretty.vfold_cv
pretty.sliding_period <- function(x, ...)
  "Sliding period resampling"

#' @export pretty.mc_cv
#' @export
#' @method pretty mc_cv
#' @rdname pretty.vfold_cv
pretty.mc_cv <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Monte Carlo cross-validation (",
    signif(details$prop, 2),
    "/",
    signif(1 - details$prop, 2),
    ") with ",
    details$times,
    " resamples "
  )
  if (details$strata)
    res <- paste(res, "using stratification")
  res
}

#' @export pretty.validation_split
#' @export
#' @method pretty validation_split
#' @rdname pretty.vfold_cv
pretty.validation_split <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Validation Set Split (",
    signif(details$prop, 2),
    "/",
    signif(1 - details$prop, 2),
    ") "
  )
  if (details$strata)
    res <- paste(res, "using stratification")
  res
}

#' @export pretty.nested_cv
#' @export
#' @method pretty nested_cv
#' @rdname pretty.vfold_cv
pretty.nested_cv <- function(x, ...) {
  details <- attributes(x)

  if (is_call(details$outside)) {
    class(x) <- class(x)[!(class(x) == "nested_cv")]
    outer_label <- pretty(x)
  } else {
    outer_label <- paste0("`", deparse(details$outside), "`")
  }

  inner_label <- if (is_call(details$inside))
    pretty(x$inner_resamples[[1]])
  else
    paste0("`", deparse(details$inside), "`")

  res <- c("Nested resampling:",
           paste(" outer:", outer_label),
           paste(" inner:", inner_label))
  res
}

#' @export pretty.bootstraps
#' @export
#' @method pretty bootstraps
#' @rdname pretty.vfold_cv
pretty.bootstraps <- function(x, ...) {
  details <- attributes(x)
  res <- "Bootstrap sampling"
  if (details$strata)
    res <- paste(res, "using stratification")
  if (details$apparent)
    res <- paste(res, "with apparent sample")
  res
}


#' @export pretty.group_vfold_cv
#' @export
#' @method pretty group_vfold_cv
#' @rdname pretty.vfold_cv
pretty.group_vfold_cv  <- function(x, ...) {
  details <- attributes(x)
  paste0("Group ", details$v, "-fold cross-validation")
}

#' @export pretty.manual_rset
#' @export
#' @method pretty manual_rset
#' @rdname pretty.vfold_cv
pretty.manual_rset <- function(x, ...) {
  "Manual resampling"
}


#' Augment a data set with resampling identifiers
#'
#' For a data set, `add_resample_id()` will add at least one new column that
#'  identifies which resample that the data came from. In most cases, a single
#'  column is added but for some resampling methods, two or more are added.
#' @param .data A data frame
#' @param split A single `rset` object.
#' @param dots A single logical: should the id columns be prefixed with a "."
#'  to avoid name conflicts with `.data`?
#' @return An updated data frame.
#' @examples
#' library(dplyr)
#'
#' set.seed(363)
#' car_folds <- vfold_cv(mtcars, repeats = 3)
#'
#' analysis(car_folds$splits[[1]]) %>%
#'   add_resample_id(car_folds$splits[[1]]) %>%
#'   head()
#'
#' car_bt <- bootstraps(mtcars)
#'
#' analysis(car_bt$splits[[1]]) %>%
#'   add_resample_id(car_bt$splits[[1]]) %>%
#'   head()
#' @seealso labels.rsplit
#' @export
add_resample_id <- function(.data, split, dots = FALSE) {
  if (!inherits(dots, "logical") || length(dots) > 1) {
    stop("`dots` should be a single logical.", call. = FALSE)
  }
  if (!inherits(.data, "data.frame")) {
    stop("`.data` should be a data frame.", call. = FALSE)
  }
  if (!inherits(split, "rsplit")) {
    stop("`split` should be a single 'rset' object.", call. = FALSE)
  }
  labs <- labels(split)

  if (!tibble::is_tibble(labs) && nrow(labs) == 1) {
    stop("`split` should be a single 'rset' object.", call. = FALSE)
  }

  if (dots) {
    colnames(labs) <- paste0(".", colnames(labs))
  }

  cbind(.data, labs)
}

#' Monte Carlo Cross-Validation
#'
#' One resample of Monte Carlo cross-validation takes a random sample (without
#'  replacement) of the original data set to be used for analysis. All other
#'  data points are added to the assessment set.
#' @details The `strata` argument causes the random sampling to be conducted
#'  *within the stratification variable*. This can help ensure that the number of
#'  data points in the analysis data is equivalent to the proportions in the
#'  original data set. (Strata below 10% of the total are pooled together.)
#' @inheritParams vfold_cv
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @param times The number of times to repeat the sampling.
#' @param strata A variable that is used to conduct stratified sampling to
#'  create the resamples. This could be a single character value or a variable
#'  name that corresponds to a variable that exists in the data frame.
#' @param breaks A single number giving the number of bins desired to stratify
#'  a numeric stratification variable.
#' @export
#' @return An tibble with classes `mc_cv`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and a
#'  column called `id` that has a character string with the resample identifier.
#' @examples
#' mc_cv(mtcars, times = 2)
#' mc_cv(mtcars, prop = .5, times = 2)
#'
#' library(purrr)
#' data(wa_churn, package = "modeldata")
#'
#' set.seed(13)
#' resample1 <- mc_cv(wa_churn, times = 3, prop = .5)
#' map_dbl(resample1$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#'
#' set.seed(13)
#' resample2 <- mc_cv(wa_churn, strata = "churn", times = 3, prop = .5)
#' map_dbl(resample2$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#'
#' set.seed(13)
#' resample3 <- mc_cv(wa_churn, strata = "tenure", breaks = 6, times = 3, prop = .5)
#' map_dbl(resample3$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#' @export
mc_cv <- function(data, prop = 3/4, times = 25, strata = NULL, breaks = 4, ...) {

  if(!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if(length(strata) == 0) strata <- NULL
  }

  strata_check(strata, names(data))

  split_objs <-
    mc_splits(data = data,
              prop = 1 - prop,
              times = times,
              strata = strata,
              breaks = breaks)

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  mc_att <- list(prop = prop,
                 times = times,
                 strata = !is.null(strata))

  new_rset(splits = split_objs$splits,
           ids = split_objs$id,
           attrib = mc_att,
           subclass = c("mc_cv", "rset"))
}

# Get the indices of the analysis set from the assessment set
mc_complement <- function(ind, n) {
  list(analysis = setdiff(1:n, ind),
       assessment = ind)
}


mc_splits <- function(data, prop = 3/4, times = 25, strata = NULL, breaks = 4) {
  if (!is.numeric(prop) | prop >= 1 | prop <= 0)
    stop("`prop` must be a number on (0, 1).", call. = FALSE)

  n <- nrow(data)
  if (is.null(strata)) {
    indices <- purrr::map(rep(n, times), sample, size = floor(n * prop))
  } else {
    stratas <- tibble::tibble(idx = 1:n,
                              strata = make_strata(getElement(data, strata),
                                                   breaks = breaks))
    stratas <- split_unnamed(stratas, stratas$strata)
    stratas <-
      purrr::map_df(stratas, strat_sample, prop = prop, times = times)
    indices <- split_unnamed(stratas$idx, stratas$rs_id)
  }
  indices <- lapply(indices, mc_complement, n = n)
  split_objs <-
    purrr::map(indices, make_splits, data = data, class = "mc_split")
  list(splits = split_objs,
       id = names0(length(split_objs), "Resample"))
}

strat_sample <- function(x, prop, times, ...) {
  n <- nrow(x)
  idx <- purrr::map(rep(n, times), sample, size = floor(n*prop), ...)
  out <- purrr::map_df(idx, function(ind, x) x[sort(ind), "idx"], x = x)
  out$rs_id <- rep(1:times, each = floor(n*prop))
  out
}

#' @export
print.mc_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("mc_cv", "rset"))]
  print(x, ...)
}

rsplit <- function(data, in_id, out_id) {
  if (!is.data.frame(data) & !is.matrix(data))
    stop("`data` must be a data frame.", call. = FALSE)

  if (!is.integer(in_id) | any(in_id < 1))
    stop("`in_id` must be a positive integer vector.", call. = FALSE)

  if(!all(is.na(out_id))) {
    if (!is.integer(out_id) | any(out_id < 1))
      stop("`out_id` must be a positive integer vector.", call. = FALSE)
  }

  if (length(in_id) == 0)
    stop("At least one row should be selected for the analysis set.",
         call. = FALSE)

  structure(
    list(
      data = data,
      in_id = in_id,
      out_id = out_id
    ),
    class = "rsplit"
  )
}

#' @export
print.rsplit <- function(x, ...) {
  out_char <-
    if (all(is.na(x$out_id)))
      paste(length(complement(x)))
  else
    paste(length(x$out_id))

  cat("<Analysis/Assess/Total>\n")
  cat("<",
      length(x$in_id), "/",
      out_char, "/",
      nrow(x$data), ">\n",
      sep = "")
}

#' @export
as.integer.rsplit <-
  function(x, data = c("analysis", "assessment"), ...) {
    data <- match.arg(data)
    if (data == "analysis")
      out <- x$in_id
    else {
      out <- if (all(is.na(x$out_id)))
        complement(x)
      else
        x$out_id
    }
    out
  }


#' Convert an `rsplit` object to a data frame
#'
#' The analysis or assessment code can be returned as a data
#'   frame (as dictated by the `data` argument) using
#'   `as.data.frame.rsplit`. `analysis` and
#'   `assessment` are shortcuts.
#' @param x An `rsplit` object.
#' @param row.names `NULL` or a character vector giving the row names for the data frame. Missing values are not allowed.
#' @param optional A logical: should the column names of the data be checked for legality?
#' @param data Either "analysis" or "assessment" to specify which data are returned.
#' @param ...	Additional arguments to be passed to or from methods. Not currently used.
#' @examples
#' library(dplyr)
#' set.seed(104)
#' folds <- vfold_cv(mtcars)
#'
#' model_data_1 <- folds$splits[[1]] %>% analysis()
#' holdout_data_1 <- folds$splits[[1]] %>% assessment()
#' @export
as.data.frame.rsplit <-
  function(x,
           row.names = NULL,
           optional = FALSE,
           data = "analysis",
           ...) {

  if (!is.null(row.names))
    warning( "`row.names` is kept for consistency with the ",
             "underlying class but non-NULL values will be ",
             "ignored.", call. = FALSE)
  if (optional)
    warning( "`optional` is kept for consistency with the ",
             "underlying class but TRUE values will be ",
             "ignored.", call. = FALSE)
  x$data[as.integer(x, data = data, ...), , drop = FALSE]
}

#' @rdname as.data.frame.rsplit
#' @export
analysis <- function(x, ...) {
  if (!inherits(x, "rsplit"))
    stop("`x` should be an `rsplit` object", call. = FALSE)
  as.data.frame(x, data = "analysis", ...)
}
#' @rdname as.data.frame.rsplit
#' @export
assessment <- function(x, ...){
  if (!inherits(x, "rsplit"))
    stop("`x` should be an `rsplit` object", call. = FALSE)
  as.data.frame(x, data = "assessment", ...)
}

#' @export
dim.rsplit <- function(x, ...) {
  c(
    analysis = length(x$in_id),
    assessment = length(complement(x)),
    n = nrow(x$data),
    p = ncol(x$data)
  )
}

#' @method obj_sum rsplit
#' @export
obj_sum.rsplit <- function(x, ...) {
  out_char <-
    if (all(is.na(x$out_id)))
      paste(length(complement(x)))
  else
    paste(length(x$out_id))

  paste0("rsplit [",
         length(x$in_id), "/",
         out_char, "]")
}


#' @method type_sum rsplit
#' @export
type_sum.rsplit <- function(x, ...) {
  out_char <-
    if (all(is.na(x$out_id)))
      format_n(length(complement(x)))
  else
    format_n(length(x$out_id))

  paste0(
    "split [",
    format_n(length(x$in_id)), "/",
    out_char, "]"
  )
}


format_n <- function(x, digits = 1) {
  case_when(
    log10(x) < 3  ~ paste(x),
    log10(x) >= 3 & log10(x) < 6 ~ paste0(round(x/1000, digits = digits), "K"),
    TRUE ~ paste0(round(x/1000000, digits = digits), "M"),
  )
}

is_rsplit <- function(x) {
  inherits(x, "rsplit")
}

# nocov start

.onLoad <- function(libname, pkgname) {
  if (dplyr_pre_1.0.0()) {
    vctrs::s3_register("dplyr::mutate", "rset", method = mutate_rset)
    vctrs::s3_register("dplyr::arrange", "rset", method = arrange_rset)
    vctrs::s3_register("dplyr::filter", "rset", method = filter_rset)
    vctrs::s3_register("dplyr::rename", "rset", method = rename_rset)
    vctrs::s3_register("dplyr::select", "rset", method = select_rset)
    vctrs::s3_register("dplyr::slice", "rset", method = slice_rset)
  } else {
    vctrs::s3_register("dplyr::dplyr_reconstruct", "rset", method = dplyr_reconstruct_rset)
  }
}

# nocov end

## adapted from
## https://github.com/hadley/dtplyr/blob/2308ff25e88bb81fe84f9051e37ddd9d572189ee/R/compat-dplyr-0.6.0.R
## and based on 
## https://github.com/tidyverse/googledrive/commit/95455812d2e0d6bdf92b5f6728e3265bf65d8467#diff-ba61d4f2ccd992868e27305a9ab68a3c

## function is called in .onLoad()
#' @importFrom rlang is_string
register_s3_method <- function(pkg, generic, class, fun = NULL) { # nocov start
  stopifnot(is_string(pkg))
  envir <- asNamespace(pkg)

  stopifnot(is_string(generic))
  stopifnot(is_string(class))
  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  }
  stopifnot(is.function(fun))

  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = envir)
  }

  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = envir)
    }
  )
} # nocov end

reset_rset <- function(x) {
  stopifnot(inherits(x, "data.frame"))
  structure(x, class = c("rset", "tbl_df", "tbl", "data.frame"))
}

#' @importFrom tibble is_tibble
is_rset <- function(x) {
  is_tibble(x) && 
    all("splits" %in% names(x)) &&
    length(grepl("^id", names(x))) > 0
}

#' @export
`[.rset` <- function(x, i, j, drop = FALSE) {
  maybe_rset(NextMethod())
}

# A list of attribute names in the various resampling functions. These
# could get stripped off by dplyr operations
rsample_att <- c("times", "apparent", "strata", "v", "repeats", 
                 "group", "prop", "outside", "inside", 
                 "initial", "assess", "cumulative", "skip"
                 )

#' @importFrom dplyr as_tibble
maybe_rset <- function(x, extras = NULL, att = NULL) {
  if (is_rset(x)) {
    x <- reset_rset(x)
    
    ## possibly reset attributes that dplyr methods removed
    att <- att[names(att) %in% rsample_att]
    if (length(att) > 0) {
      missing_att <- setdiff(names(att), attributes(x))
      if (length(missing_att) > 0) {
        for (i in missing_att)
          attr(x, i) <- att[[i]]
      }
    }
    
    ## Add an missing classes
    if(length(extras) > 0)
      class(x) <- c(extras, class(x))
  } else {
    x <- as_tibble(x)
  }
  x
}

## rsample does not import any generics from dplyr,
## but if dplyr is loaded and main verbs are used on a 
## `rset` object generated from the various resampling 
## functions, we want to retain the `rset`` class (and 
## any others) if it is proper to do so therefore these 
## S3 methods are registered manually in .onLoad()

base_classes <- c("rset", class(tibble()))

arrange.rset <- function(.data, ...) {
  # Find out if there are extra classes beyond `rset` such 
  # as `vfold_cv` or `bootstraps` and add them in after
  # the `dplyr` method is executed. 
  extra_classes <- setdiff(class(.data), base_classes)
  orig_att <- attributes(.data)
  maybe_rset(NextMethod(), extras = extra_classes, att = orig_att)
}

filter.rset <- function(.data, ...) {
  extra_classes <- setdiff(class(.data), base_classes)
  orig_att <- attributes(.data)
  maybe_rset(NextMethod(), extras = extra_classes, att = orig_att)
}

# `mutate` appears to add rownames but remove other attributes. We'll
# add them back in. 
mutate.rset <- function(.data, ...) {
  extra_classes <- setdiff(class(.data), base_classes)
  orig_att <- attributes(.data)
  maybe_rset(NextMethod(), extras = extra_classes, att = orig_att)
}

rename.rset <- function(.data, ...) {
  extra_classes <- setdiff(class(.data), base_classes)
  orig_att <- attributes(.data)
  maybe_rset(NextMethod(), extras = extra_classes, att = orig_att)
}

select.rset <- function(.data, ...) {
  extra_classes <- setdiff(class(.data), base_classes)
  orig_att <- attributes(.data)
  maybe_rset(NextMethod(), extras = extra_classes, att = orig_att)
}

slice.rset <- function(.data, ...) {
  extra_classes <- setdiff(class(.data), base_classes)
  orig_att <- attributes(.data)
  maybe_rset(NextMethod(), extras = extra_classes, att = orig_att)
}

#' rsample: General Resampling Infrastructure for R
#'
#'\pkg{rsample} has functions to create variations of a data set
#'  that can be used to evaluate models or to estimate the
#'  sampling distribution of some statistic.
#'
#' @section Terminology:
#'\itemize{
#'  \item A **resample** is the result of a two-way split of a
#'  data set. For example, when bootstrapping, one part of the
#'  resample is a sample with replacement of the original data.
#'  The other part of the split contains the instances that were
#'  not contained in the bootstrap sample. The data structure
#'  `rsplit` is used to store a single resample. 
#'  \item When the data are split in two, the portion that is
#'  used to estimate the model or calculate the statistic is
#'  called the **analysis** set here. In machine learning this
#'  is sometimes called the "training set" but this would be
#'  poorly named since it might conflict with any initial split
#'  of the original data.
#'  \item Conversely, the other data in the split are called the
#'     **assessment** data. In bootstrapping, these data are
#'     often called the "out-of-bag" samples. 
#'  \item A collection of resamples is contained in an 
#'  `rset` object.
#'}
#'
#' @section Basic Functions:
#' The main resampling functions are: [vfold_cv()], 
#'   [bootstraps()], [mc_cv()], 
#'   [rolling_origin()], and [nested_cv()].
#' @docType package
#' @name rsample
NULL

#' Simple Training/Test Set Splitting
#'
#' `initial_split` creates a single binary split of the data into a training
#'  set and testing set. `initial_time_split` does the same, but takes the
#'  _first_ `prop` samples for training, instead of a random selection.
#'  `training` and `testing` are used to extract the resulting data.
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @export
#' @return An `rsplit` object that can be used with the `training` and `testing`
#'  functions to extract the data in each split.
#' @examples
#' set.seed(1353)
#' car_split <- initial_split(mtcars)
#' train_data <- training(car_split)
#' test_data <- testing(car_split)
#'
#' data(drinks, package = "modeldata")
#' drinks_split <- initial_time_split(drinks)
#' train_data <- training(drinks_split)
#' test_data <- testing(drinks_split)
#' c(max(train_data$date), min(test_data$date)) # no lag
#'
#' # With 12 period lag
#' drinks_lag_split <- initial_time_split(drinks, lag = 12)
#' train_data <- training(drinks_lag_split)
#' test_data <- testing(drinks_lag_split)
#' c(max(train_data$date), min(test_data$date)) # 12 period lag
#' @export
#'
initial_split <- function(data, prop = 3 / 4,
                          strata = NULL, breaks = 4, pool = 0.1, ...) {
  res <-
    mc_cv(
      data = data,
      prop = prop,
      strata = {{ strata }},
      breaks = breaks,
      pool = pool,
      times = 1,
      ...
    )
  res <- res$splits[[1]]
  class(res) <- c("initial_split", class(res))
  res
}

#' @rdname initial_split
#' @param lag A value to include a lag between the assessment
#'  and analysis set. This is useful if lagged predictors will be used
#'  during training and testing.
#' @export
initial_time_split <- function(data, prop = 3 / 4, lag = 0, ...) {
  if (!is.numeric(prop) | prop >= 1 | prop <= 0) {
    rlang::abort("`prop` must be a number on (0, 1).")
  }

  if (!is.numeric(lag) | !(lag %% 1 == 0)) {
    rlang::abort("`lag` must be a whole number.")
  }

  n_train <- floor(nrow(data) * prop)

  if (lag > n_train) {
    rlang::abort("`lag` must be less than or equal to the number of training observations.")
  }

  split <- rsplit(data, 1:n_train, (n_train + 1 - lag):nrow(data))
  splits <- list(split)
  ids <- "Resample1"
  rset <- new_rset(splits, ids)

  res <- rset$splits[[1]]
  class(res) <- c("initial_split", class(res))
  res
}

#' @rdname initial_split
#' @export
#' @param x An `rsplit` object produced by `initial_split()` or
#'  `initial_time_split()`.
training <- function(x) analysis(x)
#' @rdname initial_split
#' @export
testing <- function(x) assessment(x)

#' @import vctrs
NULL

# `vec_restore()`
#
# Called at the end of `vec_slice()` and `vec_ptype()` after all slicing has
# been done on the proxy object.
#
# If no changes have been made to the row/column structure of rset specific
# columns, then we can return an rset subclass. Otherwise, the resulting
# object is no longer guaranteed to return a valid rset, and we have to
# fallback to a bare tibble.
#
# It is very important that the result of `vec_ptype()` is a bare tibble.
# This ensures that the `vec_ptype2.<rset-subclass>.<rset-subclass>()` methods
# never get called. `vec_ptype()` is able to return a bare tibble because it
# essentially takes a 0-row slice of the rset, and then calls `vec_restore()`.
# Because the row structure has been modified, we return a bare tibble from
# `vec_restore.<rset-subclass>()`.
#
# Currently `vec_restore()` uses inheritance in vctrs, but I don't expect this
# to be the case in the future. For that reason, I use explicit methods for
# each individual rset subclass, rather than implementing `vec_restore.rset()`.


# `vec_ptype2()`
#
# The purpose of `vec_ptype2()` methods is generally to determine the type
# of the output in operations like `vec_c()` or `vec_rbind()`. However, this
# implementation does not use any custom `vec_ptype2()` methods at all. This is
# explained below.
#
# `vec_ptype2()` internally works by calling `vec_ptype()` on both `x` and `y`,
# and then looking up the common type of those two ptypes. Generally, the
# ptype of a vector `x` returned from `vec_ptype()` has the same class as `x`.
# However, for rsample objects it makes more sense for the ptype of an rset to
# be a bare tibble. The reason for this is because a ptype of a data frame is
# generally a 0-row slice. However for rsample rsets this doesn't make sense
# (you can't have 0 rows in a 10-fold cv rset), so instead we return a bare
# tibble as the ptype.
# So when `vec_ptype2()` is called on an rset, that rset is downgraded to a
# bare tibble _before_ the search for a `vec_ptype2()` method occurs. This
# means that it will never find a ptype2 method such as
# `vec_ptype2.bootstraps.tbl_df()`, because the <bootstraps> will become a
# tbl_df first. This means that we rely entirely on the `tbl_df` ptype2 methods,
# which are already implemented in vctrs to return another tbl_df.
#
# The implications of this are that whenever a rset subclass is combined with
# another rset subclass or bare tibble, the resulting common type is always
# another bare tibble. So if you `vec_c(<bootstraps>, <tbl_df>)` the result
# will always be a tibble, never a bootstraps object. This makes sense because
# you might be adding rows, which would invalidate the structure of the
# bootstraps object.


# `vec_cast()`
#
# The `vec_cast()` methods for rset objects really only serve 1 purpose. They
# cast an rset subclass to a tibble or data frame. The cast to tibble is most
# useful. Most of the operations in vctrs work by finding a common type
# with `vec_ptype2()`, and then casting all of the inputs to that common type.
# Because `vec_ptype2()` returns a bare tibble anytime a rset-subclass is
# involved, we will always be casting the rset subclass to a tibble.
# The cast method uses `vctrs::tib_cast()`, which always returns a bare tibble
# with all of the data in `x` cast to the type of `to`.

# ------------------------------------------------------------------------------
# bootstraps

#' @export
vec_restore.bootstraps <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.bootstraps.bootstraps <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.bootstraps.bootstraps")
}
#' @export
vec_ptype2.bootstraps.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.bootstraps.tbl_df")
}
#' @export
vec_ptype2.tbl_df.bootstraps <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.bootstraps")
}
#' @export
vec_ptype2.bootstraps.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.bootstraps.data.frame")
}
#' @export
vec_ptype2.data.frame.bootstraps <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.bootstraps")
}


#' @export
vec_cast.bootstraps.bootstraps <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.bootstraps.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.bootstraps <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.bootstraps.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.bootstraps <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# vfold_cv

#' @export
vec_restore.vfold_cv <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.vfold_cv.vfold_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.vfold_cv.vfold_cv")
}
#' @export
vec_ptype2.vfold_cv.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.vfold_cv.tbl_df")
}
#' @export
vec_ptype2.tbl_df.vfold_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.vfold_cv")
}
#' @export
vec_ptype2.vfold_cv.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.vfold_cv.data.frame")
}
#' @export
vec_ptype2.data.frame.vfold_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.vfold_cv")
}


#' @export
vec_cast.vfold_cv.vfold_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.vfold_cv.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.vfold_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.vfold_cv.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.vfold_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# group_vfold_cv

#' @export
vec_restore.group_vfold_cv <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.group_vfold_cv.group_vfold_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.group_vfold_cv.group_vfold_cv")
}
#' @export
vec_ptype2.group_vfold_cv.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.group_vfold_cv.tbl_df")
}
#' @export
vec_ptype2.tbl_df.group_vfold_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.group_vfold_cv")
}
#' @export
vec_ptype2.group_vfold_cv.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.group_vfold_cv.data.frame")
}
#' @export
vec_ptype2.data.frame.group_vfold_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.group_vfold_cv")
}


#' @export
vec_cast.group_vfold_cv.group_vfold_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.group_vfold_cv.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.group_vfold_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.group_vfold_cv.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.group_vfold_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# loo_cv

#' @export
vec_restore.loo_cv <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.loo_cv.loo_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.loo_cv.loo_cv")
}
#' @export
vec_ptype2.loo_cv.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.loo_cv.tbl_df")
}
#' @export
vec_ptype2.tbl_df.loo_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.loo_cv")
}
#' @export
vec_ptype2.loo_cv.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.loo_cv.data.frame")
}
#' @export
vec_ptype2.data.frame.loo_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.loo_cv")
}


#' @export
vec_cast.loo_cv.loo_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.loo_cv.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.loo_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.loo_cv.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.loo_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# mc_cv

#' @export
vec_restore.mc_cv <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.mc_cv.mc_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.mc_cv.mc_cv")
}
#' @export
vec_ptype2.mc_cv.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.mc_cv.tbl_df")
}
#' @export
vec_ptype2.tbl_df.mc_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.mc_cv")
}
#' @export
vec_ptype2.mc_cv.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.mc_cv.data.frame")
}
#' @export
vec_ptype2.data.frame.mc_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.mc_cv")
}


#' @export
vec_cast.mc_cv.mc_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.mc_cv.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.mc_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.mc_cv.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.mc_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# nested_cv

#' @export
vec_restore.nested_cv <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.nested_cv.nested_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.nested_cv.nested_cv")
}
#' @export
vec_ptype2.nested_cv.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.nested_cv.tbl_df")
}
#' @export
vec_ptype2.tbl_df.nested_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.nested_cv")
}
#' @export
vec_ptype2.nested_cv.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.nested_cv.data.frame")
}
#' @export
vec_ptype2.data.frame.nested_cv <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.nested_cv")
}


#' @export
vec_cast.nested_cv.nested_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.nested_cv.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.nested_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.nested_cv.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.nested_cv <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# validation_split

#' @export
vec_restore.validation_split <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.validation_split.validation_split <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.validation_split.validation_split")
}
#' @export
vec_ptype2.validation_split.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.validation_split.tbl_df")
}
#' @export
vec_ptype2.tbl_df.validation_split <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.validation_split")
}
#' @export
vec_ptype2.validation_split.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.validation_split.data.frame")
}
#' @export
vec_ptype2.data.frame.validation_split <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.validation_split")
}


#' @export
vec_cast.validation_split.validation_split <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.validation_split.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.validation_split <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.validation_split.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.validation_split <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# rolling_origin

#' @export
vec_restore.rolling_origin <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.rolling_origin.rolling_origin <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.rolling_origin.rolling_origin")
}
#' @export
vec_ptype2.rolling_origin.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.rolling_origin.tbl_df")
}
#' @export
vec_ptype2.tbl_df.rolling_origin <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.rolling_origin")
}
#' @export
vec_ptype2.rolling_origin.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.rolling_origin.data.frame")
}
#' @export
vec_ptype2.data.frame.rolling_origin <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.rolling_origin")
}


#' @export
vec_cast.rolling_origin.rolling_origin <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.rolling_origin.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.rolling_origin <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.rolling_origin.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.rolling_origin <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# sliding_window

#' @export
vec_restore.sliding_window <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.sliding_window.sliding_window <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_window.sliding_window")
}
#' @export
vec_ptype2.sliding_window.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_window.tbl_df")
}
#' @export
vec_ptype2.tbl_df.sliding_window <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.sliding_window")
}
#' @export
vec_ptype2.sliding_window.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_window.data.frame")
}
#' @export
vec_ptype2.data.frame.sliding_window <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.sliding_window")
}


#' @export
vec_cast.sliding_window.sliding_window <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.sliding_window.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.sliding_window <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.sliding_window.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.sliding_window <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# sliding_index

#' @export
vec_restore.sliding_index <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.sliding_index.sliding_index <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_index.sliding_index")
}
#' @export
vec_ptype2.sliding_index.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_index.tbl_df")
}
#' @export
vec_ptype2.tbl_df.sliding_index <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.sliding_index")
}
#' @export
vec_ptype2.sliding_index.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_index.data.frame")
}
#' @export
vec_ptype2.data.frame.sliding_index <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.sliding_index")
}


#' @export
vec_cast.sliding_index.sliding_index <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.sliding_index.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.sliding_index <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.sliding_index.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.sliding_index <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# sliding_period

#' @export
vec_restore.sliding_period <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.sliding_period.sliding_period <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_period.sliding_period")
}
#' @export
vec_ptype2.sliding_period.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_period.tbl_df")
}
#' @export
vec_ptype2.tbl_df.sliding_period <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.sliding_period")
}
#' @export
vec_ptype2.sliding_period.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.sliding_period.data.frame")
}
#' @export
vec_ptype2.data.frame.sliding_period <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.sliding_period")
}


#' @export
vec_cast.sliding_period.sliding_period <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.sliding_period.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.sliding_period <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.sliding_period.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.sliding_period <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# manual_rset

#' @export
vec_restore.manual_rset <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.manual_rset.manual_rset <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.manual_rset.manual_rset")
}
#' @export
vec_ptype2.manual_rset.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.manual_rset.tbl_df")
}
#' @export
vec_ptype2.tbl_df.manual_rset <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.manual_rset")
}
#' @export
vec_ptype2.manual_rset.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.manual_rset.data.frame")
}
#' @export
vec_ptype2.data.frame.manual_rset <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.manual_rset")
}


#' @export
vec_cast.manual_rset.manual_rset <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.manual_rset.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.manual_rset <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.manual_rset.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.manual_rset <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------
# apparent

#' @export
vec_restore.apparent <- function(x, to, ...) {
  rset_reconstruct(x, to)
}


#' @export
vec_ptype2.apparent.apparent <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.apparent.apparent")
}
#' @export
vec_ptype2.apparent.tbl_df <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.apparent.tbl_df")
}
#' @export
vec_ptype2.tbl_df.apparent <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.tbl_df.apparent")
}
#' @export
vec_ptype2.apparent.data.frame <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.apparent.data.frame")
}
#' @export
vec_ptype2.data.frame.apparent <- function(x, y, ..., x_arg = "", y_arg = "") {
  stop_never_called("vec_ptype2.data.frame.apparent")
}


#' @export
vec_cast.apparent.apparent <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.apparent.tbl_df <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.tbl_df.apparent <- function(x, to, ..., x_arg = "", to_arg = "") {
  tib_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.apparent.data.frame <- function(x, to, ..., x_arg = "", to_arg = "") {
  stop_incompatible_cast_rset(x, to, x_arg = x_arg, to_arg = to_arg)
}
#' @export
vec_cast.data.frame.apparent <- function(x, to, ..., x_arg = "", to_arg = "") {
  df_cast(x, to, ..., x_arg = x_arg, to_arg = to_arg)
}

# ------------------------------------------------------------------------------

stop_incompatible_cast_rset <- function(x, to, ..., x_arg, to_arg) {
  details <- "Can't cast to an rset because attributes are likely incompatible."
  stop_incompatible_cast(x, to, x_arg = x_arg, to_arg = to_arg, details = details)
}

stop_never_called <- function(fn) {
  rlang::abort(paste0("Internal error: `", fn, "()` should never be called."))
}

#' Manual resampling
#'
#' `manual_rset()` is used for constructing the most minimal rset possible. It
#' can be useful when you have custom rsplit objects built from
#' [make_splits()], or when you want to create a new rset from splits
#' contained within an existing rset.
#'
#' @param splits A list of `"rsplit"` objects. It is easiest to create these
#'   using [make_splits()].
#'
#' @param ids A character vector of ids. The length of `ids` must be the same
#'   as the length of `splits`.
#'
#' @export
#' @examples
#' df <- data.frame(x = c(1, 2, 3, 4, 5, 6))
#'
#' # Create an rset from custom indices
#' indices <- list(
#'   list(analysis = c(1L, 2L), assessment = 3L),
#'   list(analysis = c(4L, 5L), assessment = 6L)
#' )
#'
#' splits <- lapply(indices, make_splits, data = df)
#'
#' manual_rset(splits, c("Split 1", "Split 2"))
#'
#' # You can also use this to create an rset from a subset of an
#' # existing rset
#' resamples <- vfold_cv(mtcars)
#' best_split <- resamples[5, ]
#' manual_rset(best_split$splits, best_split$id)
manual_rset <- function(splits, ids) {
  new_manual_rset(splits, ids)
}

new_manual_rset <- function(splits, ids) {
  new_rset(splits, ids, subclass = c("manual_rset", "rset"))
}

dat1 <- data.frame(a = 1:20, b = letters[1:20])
car_folds <- vfold_cv(mtcars)

new_rng_snapshots <- utils::compareVersion("3.6.0", as.character(getRversion())) > 0

expect_s3_class_rset <- function(x) {
  expect_s3_class(x, "rset")
}

expect_s3_class_bare_tibble <- function(x) {
  expect_s3_class(x, c("tbl_df", "tbl", "data.frame"), exact = TRUE)
}

get_stats <- function(split, ...) {
  dat <- analysis(split)
  x <- dat[[1]]
  tibble(
    term = "mean",
    estimate = mean(x, na.rm = TRUE),
    std.error = sqrt(var(x, na.rm = TRUE) / sum(!is.na(x)))
  )
}

check_ind <- function(x, tdat) {
  in_dat <- subset(tdat, Data == "Analysis")
  in_check <- all(sort(in_dat$Row) == x$in_ind)
  out_dat <- subset(tdat, Data == "Analysis")
  out_check <- all(sort(out_dat$Row) == x$out_ind)
  in_check & out_check
}

get_id_left_out <- function(x) {
  unique(as.character(assessment(x)$tension))
}

# These tests should pass on all supported versions of dplyr. Both pre and
# post dplyr 1.0.0 should work.
# When `compat-dplyr-old.R` is removed and support for dplyr < 1.0.0 is
# deprecated, these tests should move to `test-compat-dplyr.R` instead.
# Do not just delete them, as they are important tests and are not repeated in
# `compat-dplyr.R`.

library(dplyr)

# ------------------------------------------------------------------------------
# mutate()

test_that("mutate() can keep rset class", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(mutate(x, x = 1))
    expect_identical(mutate(x, x = 1)$x, rep(1, vec_size(x)))
  }
})

test_that("mutate() drops rset class if any rset columns are touched", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(mutate(x, splits = 1))
    expect_s3_class_bare_tibble(mutate(x, id = 1))

    expect_identical(mutate(x, splits = 1)$splits, rep(1, vec_size(x)))
    expect_identical(mutate(x, id = 1)$id, rep(1, vec_size(x)))
  }
})

test_that("mutate() keeps rset class if replacement rset column is same as original", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(mutate(x, splits = splits))
    expect_s3_class_rset(mutate(x, id = id))
  }
})

test_that("adding a column that looks like an `id` drops the class", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(mutate(x, id9 = 1))
  }
})

# ------------------------------------------------------------------------------
# arrange()

test_that("arrange() keeps rset class when row order is modified", {
  for (x in rset_subclasses) {
    x <- mutate(x, rn = row_number())
    expect_s3_class_rset(arrange(x, desc(rn)))
  }
})

test_that("arrange() keeps rset class when row order is untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(arrange(x))

    x <- mutate(x, rn = row_number())
    expect_s3_class_rset(arrange(x, rn))
  }
})

# ------------------------------------------------------------------------------
# filter()

test_that("filter() drops rset class when rows are modified", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(filter(x, 0 == 1))
    expect_s3_class_bare_tibble(filter(x, is.numeric(id)))
  }
})

test_that("filter() keeps rset class if row order is untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(filter(x))
    expect_s3_class_rset(filter(x, is.character(id)))
  }
})

# ------------------------------------------------------------------------------
# rename()

test_that("renaming can keep the rset class", {
  for (x in rset_subclasses) {
    x <- mutate(x, a = 1)
    x <- rename(x, b = a)
    expect_s3_class_rset(x)
  }
})

test_that("renaming `id` at all drops the rset class", {
  for (x in rset_subclasses) {
    x <- rename(x, id9 = id)
    expect_s3_class_bare_tibble(x)
  }
})

test_that("renaming `id` to a non-id name drops the rset class", {
  for (x in rset_subclasses) {
    x <- rename(x, stuff = id)
    expect_s3_class_bare_tibble(x)
  }
})

test_that("for nested_cv, renaming `inner_resamples` drops the rset class", {
  x <- rset_subclasses$nested_cv
  x <- rename(x, inner_stuff = inner_resamples)
  expect_s3_class_bare_tibble(x)
})

# ------------------------------------------------------------------------------
# select()

test_that("select() can keep rset class", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(select(x, everything()))
  }
})

test_that("select() drops rset class if any rset columns aren't selected", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(select(x, id))
    expect_s3_class_bare_tibble(select(x, splits))
  }
})

# ------------------------------------------------------------------------------
# slice()

test_that("slice() drops rset class when rows are modified", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(slice(x, 0))
  }
})

test_that("slice() keeps rset class when rows are untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(slice(x))
    expect_s3_class_rset(slice(x, seq_len(nrow(x))))
  }
})

context("Nested CV")

library(testthat)
library(rsample)
library(purrr)

test_that('default param', {
  set.seed(11)
  rs1 <- nested_cv(mtcars[1:30,],
                   outside = vfold_cv(v = 10),
                   inside = vfold_cv(v = 3))
  sizes1 <- dim_rset(rs1)
  expect_true(all(sizes1$analysis == 27))
  expect_true(all(sizes1$assessment == 3))
  subsizes1 <- map(rs1$inner_resamples, dim_rset)
  subsizes1 <- do.call("rbind", subsizes1)
  expect_true(all(subsizes1$analysis == 18))
  expect_true(all(subsizes1$assessment == 9))

  set.seed(11)
  rs2 <- nested_cv(mtcars[1:30,],
                   outside = vfold_cv(v = 10),
                   inside = bootstraps(times = 3))
  sizes2 <- dim_rset(rs2)
  expect_true(all(sizes2$analysis == 27))
  expect_true(all(sizes2$assessment == 3))
  subsizes2 <- map(rs2$inner_resamples, dim_rset)
  subsizes2 <- do.call("rbind", subsizes2)
  expect_true(all(subsizes2$analysis == 27))

  set.seed(11)
  rs3 <- nested_cv(mtcars[1:30,],
                   outside = vfold_cv(v = 10),
                   inside = mc_cv(prop = 2/3, times = 3))
  sizes3 <- dim_rset(rs3)
  expect_true(all(sizes3$analysis == 27))
  expect_true(all(sizes3$assessment == 3))
  subsizes3 <- map(rs3$inner_resamples, dim_rset)
  subsizes3 <- do.call("rbind", subsizes3)
  expect_true(all(subsizes3$analysis == 18))
  expect_true(all(subsizes3$assessment == 9))
})

test_that('bad args', {
  expect_warning(
    nested_cv(mtcars,
              outside = bootstraps(times = 5),
              inside = vfold_cv(V = 3))
  )
  folds <- vfold_cv(mtcars)
  expect_error(
    nested_cv(mtcars,
              outside = vfold_cv(),
              inside = folds)
  )
})

test_that('printing', {
  rs1 <- nested_cv(mtcars[1:30,],
                   outside = vfold_cv(v = 10),
                   inside = vfold_cv(v = 3))
  expect_output(print(rs1))
})

test_that('rsplit labels', {
  rs <- nested_cv(mtcars[1:30,],
                  outside = vfold_cv(v = 10),
                  inside = vfold_cv(v = 3))
  all_labs <- map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

# ------------------------------------------------------------------------------
# `[`

test_that("can keep the rset class", {
  x <- rset_subclasses$nested_cv
  loc <- seq_len(ncol(x))
  expect_s3_class_rset(x[loc])
})

test_that("drops the rset class if missing `inner_resamples`", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names <- names[names != "inner_resamples"]

  expect_s3_class_bare_tibble(x[names])
})

test_that("drops the rset class if duplicating `inner_resamples`", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names <- c(names, "inner_resamples")

  expect_s3_class_bare_tibble(x[names])
})

# ------------------------------------------------------------------------------
# `names<-`

test_that("can keep the rset subclass when renaming doesn't touch rset columns", {
  x <- rset_subclasses$nested_cv
  x <- mutate(x, a = 1)

  names <- names(x)
  names[names == "a"] <- "b"

  names(x) <- names

  expect_s3_class_rset(x)
})

test_that("drops the rset class if `inner_resamples` is renamed", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names[names == "inner_resamples"] <- "inner_things"

  names(x) <- names

  expect_s3_class_bare_tibble(x)
})

test_that("drops the rset class if `inner_resamples` is moved", {
  x <- rset_subclasses$nested_cv
  x <- mutate(x, a = 1)

  names <- names(x)
  new_names <- names
  new_names[names == "inner_resamples"] <- "a"
  new_names[names == "a"] <- "inner_resamples"

  names(x) <- new_names

  expect_s3_class_bare_tibble(x)
})

test_that("default param", {
  set.seed(11)
  rs1 <- nested_cv(mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = vfold_cv(v = 3)
  )
  sizes1 <- dim_rset(rs1)
  expect_true(all(sizes1$analysis == 27))
  expect_true(all(sizes1$assessment == 3))
  subsizes1 <- purrr::map(rs1$inner_resamples, dim_rset)
  subsizes1 <- do.call("rbind", subsizes1)
  expect_true(all(subsizes1$analysis == 18))
  expect_true(all(subsizes1$assessment == 9))

  set.seed(11)
  rs2 <- nested_cv(mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = bootstraps(times = 3)
  )
  sizes2 <- dim_rset(rs2)
  expect_true(all(sizes2$analysis == 27))
  expect_true(all(sizes2$assessment == 3))
  subsizes2 <- purrr::map(rs2$inner_resamples, dim_rset)
  subsizes2 <- do.call("rbind", subsizes2)
  expect_true(all(subsizes2$analysis == 27))

  set.seed(11)
  rs3 <- nested_cv(mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = mc_cv(prop = 2 / 3, times = 3)
  )
  sizes3 <- dim_rset(rs3)
  expect_true(all(sizes3$analysis == 27))
  expect_true(all(sizes3$assessment == 3))
  subsizes3 <- purrr::map(rs3$inner_resamples, dim_rset)
  subsizes3 <- do.call("rbind", subsizes3)
  expect_true(all(subsizes3$analysis == 18))
  expect_true(all(subsizes3$assessment == 9))
})

test_that("bad args", {
  expect_snapshot({
    skip_if(new_rng_snapshots)
    set.seed(123)
    nested_cv(mtcars,
      outside = bootstraps(times = 5),
      inside = vfold_cv(V = 3)
    )
  })
  folds <- vfold_cv(mtcars)
  expect_snapshot(
    nested_cv(mtcars,
      outside = vfold_cv(),
      inside = folds
    ),
    error = TRUE
  )
})

test_that("can pass in variables", {
  make_folds <- function(df) {
    outer_cv <- 5
    inner_cv <- 4
    nested_cv(df,
      outside = vfold_cv(v = outer_cv),
      inside = vfold_cv(v = inner_cv)
    )
  }

  rs1 <- make_folds(mtcars[1:30, ])
  sizes1 <- dim_rset(rs1)
  expect_true(all(sizes1$analysis == 24))
  expect_true(all(sizes1$assessment == 6))
})

test_that("printing", {
  rs1 <- nested_cv(mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = vfold_cv(v = 3)
  )
  expect_snapshot(rs1)
})

test_that("rsplit labels", {
  rs <- nested_cv(mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = vfold_cv(v = 3)
  )
  all_labs <- purrr::map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

# ------------------------------------------------------------------------------
# `[`

test_that("can keep the rset class", {
  x <- rset_subclasses$nested_cv
  loc <- seq_len(ncol(x))
  expect_s3_class_rset(x[loc])
})

test_that("drops the rset class if missing `inner_resamples`", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names <- names[names != "inner_resamples"]

  expect_s3_class_bare_tibble(x[names])
})

test_that("drops the rset class if duplicating `inner_resamples`", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names <- c(names, "inner_resamples")

  expect_s3_class_bare_tibble(x[names])
})

# ------------------------------------------------------------------------------
# `names<-`

test_that("can keep the rset subclass when renaming doesn't touch rset columns", {
  x <- rset_subclasses$nested_cv
  x <- mutate(x, a = 1)

  names <- names(x)
  names[names == "a"] <- "b"

  names(x) <- names

  expect_s3_class_rset(x)
})

test_that("drops the rset class if `inner_resamples` is renamed", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names[names == "inner_resamples"] <- "inner_things"

  names(x) <- names

  expect_s3_class_bare_tibble(x)
})

test_that("drops the rset class if `inner_resamples` is moved", {
  x <- rset_subclasses$nested_cv
  x <- mutate(x, a = 1)

  names <- names(x)
  new_names <- names
  new_names[names == "inner_resamples"] <- "a"
  new_names[names == "a"] <- "inner_resamples"

  names(x) <- new_names

  expect_s3_class_bare_tibble(x)
})

test_that("default param", {
  set.seed(11)
  rs1 <- nested_cv(
    mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = vfold_cv(v = 3)
  )
  sizes1 <- dim_rset(rs1)
  expect_true(all(sizes1$analysis == 27))
  expect_true(all(sizes1$assessment == 3))
  subsizes1 <- purrr::map(rs1$inner_resamples, dim_rset)
  subsizes1 <- do.call("rbind", subsizes1)
  expect_true(all(subsizes1$analysis == 18))
  expect_true(all(subsizes1$assessment == 9))

  set.seed(11)
  rs2 <- nested_cv(
    mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = bootstraps(times = 3)
  )
  sizes2 <- dim_rset(rs2)
  expect_true(all(sizes2$analysis == 27))
  expect_true(all(sizes2$assessment == 3))
  subsizes2 <- purrr::map(rs2$inner_resamples, dim_rset)
  subsizes2 <- do.call("rbind", subsizes2)
  expect_true(all(subsizes2$analysis == 27))

  set.seed(11)
  rs3 <- nested_cv(
    mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = mc_cv(prop = 2 / 3, times = 3)
  )
  sizes3 <- dim_rset(rs3)
  expect_true(all(sizes3$analysis == 27))
  expect_true(all(sizes3$assessment == 3))
  subsizes3 <- purrr::map(rs3$inner_resamples, dim_rset)
  subsizes3 <- do.call("rbind", subsizes3)
  expect_true(all(subsizes3$analysis == 18))
  expect_true(all(subsizes3$assessment == 9))
})

test_that("bad args", {
  expect_snapshot({
    skip_if(new_rng_snapshots)
    set.seed(123)
    nested_cv(mtcars, outside = bootstraps(times = 5), inside = vfold_cv(v = 3))
  })
  folds <- vfold_cv(mtcars)
  expect_snapshot(
    nested_cv(mtcars, outside = vfold_cv(), inside = folds),
    error = TRUE
  )
})

test_that("can pass in variables", {
  make_folds <- function(df) {
    outer_cv <- 5
    inner_cv <- 4
    nested_cv(
      df,
      outside = vfold_cv(v = outer_cv),
      inside = vfold_cv(v = inner_cv)
    )
  }

  rs1 <- make_folds(mtcars[1:30, ])
  sizes1 <- dim_rset(rs1)
  expect_true(all(sizes1$analysis == 24))
  expect_true(all(sizes1$assessment == 6))
})

test_that("printing", {
  rs1 <- nested_cv(
    mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = vfold_cv(v = 3)
  )
  expect_snapshot(rs1)
})

test_that("rsplit labels", {
  rs <- nested_cv(
    mtcars[1:30, ],
    outside = vfold_cv(v = 10),
    inside = vfold_cv(v = 3)
  )
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("long calls don't error", {
  skip_if_not_installed("modeldata")
  expect_no_error(
    nested_cv(
      modeldata::Chicago,
      outside = sliding_period(
        index = date,
        period = "month",
        origin = modeldata::Chicago$date[1]
      ),
      inside = vfold_cv(v = 4)
    )
  )
})

# ------------------------------------------------------------------------------
# `[`
skip_if_not_installed("withr")
test_that("can keep the rset class", {
  x <- rset_subclasses$nested_cv
  loc <- seq_len(ncol(x))
  expect_s3_class_rset(x[loc])
})

test_that("drops the rset class if missing `inner_resamples`", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names <- names[names != "inner_resamples"]

  expect_s3_class_bare_tibble(x[names])
})

test_that("drops the rset class if duplicating `inner_resamples`", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names <- c(names, "inner_resamples")

  expect_s3_class_bare_tibble(x[names])
})

# ------------------------------------------------------------------------------
# `names<-`

test_that("can keep the rset subclass when renaming doesn't touch rset columns", {
  x <- rset_subclasses$nested_cv
  x <- mutate(x, a = 1)

  names <- names(x)
  names[names == "a"] <- "b"

  names(x) <- names

  expect_s3_class_rset(x)
})

test_that("drops the rset class if `inner_resamples` is renamed", {
  x <- rset_subclasses$nested_cv

  names <- names(x)
  names[names == "inner_resamples"] <- "inner_things"

  names(x) <- names

  expect_s3_class_bare_tibble(x)
})

test_that("drops the rset class if `inner_resamples` is moved", {
  x <- rset_subclasses$nested_cv
  x <- mutate(x, a = 1)

  names <- names(x)
  new_names <- names
  new_names[names == "inner_resamples"] <- "a"
  new_names[names == "a"] <- "inner_resamples"

  names(x) <- new_names

  expect_s3_class_bare_tibble(x)
})

# ------------------------------------------------------------------------------
# vec_restore()

test_that("vec_restore() returns an rset subclass if `x` retains rset structure", {
  for (x in rset_subclasses) {
    expect_identical(vec_restore(x, x), x)
    expect_s3_class_rset(vec_restore(x, x))
  }
})

test_that("vec_restore() returns bare tibble if `x` loses rset structure", {
  for (x in rset_subclasses) {
    col <- x[1]
    row <- x[0,]

    expect_s3_class_bare_tibble(vec_restore(col, x))
    expect_s3_class_bare_tibble(vec_restore(row, x))
  }
})

test_that("vec_restore() retains extra attributes of `to` when not falling back", {
  for (x in rset_subclasses) {
    to <- x
    attr(to, "foo") <- "bar"

    x_tbl <- x[1]

    expect_identical(attr(vec_restore(x, to), "foo"), "bar")
    expect_identical(attr(vec_restore(x_tbl, to), "foo"), NULL)

    expect_s3_class_rset(vec_restore(x, to))
    expect_s3_class_bare_tibble(vec_restore(x_tbl, to))
  }
})

# ------------------------------------------------------------------------------
# vec_ptype2()

test_that("vec_ptype2() is working", {
  for (x in rset_subclasses) {
    tbl <- tibble::tibble(x = 1)
    df <- data.frame(x = 1)

    # rset-rset
    expect_identical(vec_ptype2(x, x), vec_ptype2(tib_upcast(x), tib_upcast(x)))

    # rset-tbl_df
    expect_identical(vec_ptype2(x, tbl), vec_ptype2(tib_upcast(x), tbl))
    expect_identical(vec_ptype2(tbl, x), vec_ptype2(tbl, tib_upcast(x)))

    # rset-df
    expect_identical(vec_ptype2(x, df), vec_ptype2(tib_upcast(x), df))
    expect_identical(vec_ptype2(df, x), vec_ptype2(df, tib_upcast(x)))
  }
})

# ------------------------------------------------------------------------------
# vec_cast()

test_that("vec_cast() is working", {
  for (x in rset_subclasses) {
    tbl <- tib_upcast(x)
    df <- as.data.frame(tbl)

    # rset-rset
    expect_error(vec_cast(x, x), class = "vctrs_error_incompatible_type")

    # rset-tbl_df
    expect_identical(vec_cast(x, tbl), tbl)
    expect_error(vec_cast(tbl, x), class = "vctrs_error_incompatible_type")

    # rset-df
    expect_identical(vec_cast(x, df), df)
    expect_error(vec_cast(df, x), class = "vctrs_error_incompatible_type")
  }
})

# ------------------------------------------------------------------------------
# vctrs methods

test_that("vec_ptype() returns a bare tibble", {
  for (x in rset_subclasses) {
    expect_identical(vec_ptype(x), vec_ptype(tib_upcast(x)))
    expect_s3_class_bare_tibble(vec_ptype(x))
  }
})

test_that("vec_slice() generally returns a bare tibble", {
  for (x in rset_subclasses) {
    expect_identical(vec_slice(x, 0), vec_slice(tib_upcast(x), 0))
    expect_s3_class_bare_tibble(vec_slice(x, 0))
  }
})

test_that("vec_slice() can return an rset if all rows are selected", {
  for (x in rset_subclasses) {
    expect_identical(vec_slice(x, TRUE), x)
    expect_s3_class_rset(vec_slice(x, TRUE))
  }
})

test_that("vec_c() returns a bare tibble", {
  for (x in rset_subclasses) {
    tbl <- tib_upcast(x)

    expect_identical(vec_c(x), vec_c(tbl))
    expect_identical(vec_c(x, x), vec_c(tbl, tbl))
    expect_identical(vec_c(x, tbl), vec_c(tbl, tbl))

    expect_s3_class_bare_tibble(vec_c(x))
    expect_s3_class_bare_tibble(vec_c(x, x))
  }
})

test_that("vec_rbind() returns a bare tibble", {
  for (x in rset_subclasses) {
    tbl <- tib_upcast(x)

    expect_identical(vec_rbind(x), vec_rbind(tbl))
    expect_identical(vec_rbind(x, x), vec_rbind(tbl, tbl))
    expect_identical(vec_rbind(x, tbl), vec_rbind(tbl, tbl))

    expect_s3_class_bare_tibble(vec_rbind(x))
    expect_s3_class_bare_tibble(vec_cbind(x, x))
  }
})

test_that("vec_cbind() returns a bare tibble", {
  for (x in rset_subclasses) {
    tbl <- tib_upcast(x)

    expect_identical(vec_cbind(x), vec_cbind(tbl))
    expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))

    expect_s3_class_bare_tibble(vec_cbind(x))
    expect_s3_class_bare_tibble(vec_cbind(x, x))
  }
})

library(dplyr)

# ------------------------------------------------------------------------------
# dplyr_reconstruct()

test_that("dplyr_reconstruct() returns an rset subclass if `x` retains rset structure", {
  for (x in rset_subclasses) {
    expect_identical(dplyr_reconstruct(x, x), x)
    expect_s3_class_rset(dplyr_reconstruct(x, x))
  }
})

test_that("dplyr_reconstruct() returns bare tibble if `x` loses rset structure", {
  for (x in rset_subclasses) {
    col <- x[1]
    row <- x[0, ]

    expect_s3_class_bare_tibble(dplyr_reconstruct(col, x))
    expect_s3_class_bare_tibble(dplyr_reconstruct(row, x))
  }
})

test_that("dplyr_reconstruct() retains extra attributes of `to` when not falling back", {
  for (x in rset_subclasses) {
    to <- x
    attr(to, "foo") <- "bar"

    x_tbl <- x[1]

    expect_identical(attr(dplyr_reconstruct(x, to), "foo"), "bar")
    expect_identical(attr(dplyr_reconstruct(x_tbl, to), "foo"), NULL)

    expect_s3_class_rset(dplyr_reconstruct(x, to))
    expect_s3_class_bare_tibble(dplyr_reconstruct(x_tbl, to))
  }
})

# ------------------------------------------------------------------------------
# dplyr_col_modify()

test_that("can add columns and retain rset class", {
  for (x in rset_subclasses) {
    cols <- list(x = rep(1, vec_size(x)))

    result <- dplyr_col_modify(x, cols)

    expect_s3_class_rset(result)
    expect_identical(result$x, cols$x)
  }
})

test_that("modifying rset columns removes rset class", {
  for (x in rset_subclasses) {
    cols <- list(splits = rep(1, vec_size(x)))

    result <- dplyr_col_modify(x, cols)

    expect_s3_class_bare_tibble(result)
    expect_identical(result$splits, cols$splits)
  }

  for (x in rset_subclasses) {
    cols <- list(id = rep(1, vec_size(x)))

    result <- dplyr_col_modify(x, cols)

    expect_s3_class_bare_tibble(result)
    expect_identical(result$id, cols$id)
  }
})

test_that("replacing rset columns with the exact same column retains rset class", {
  for (x in rset_subclasses) {
    cols <- list(splits = x$splits)

    result <- dplyr_col_modify(x, cols)

    expect_s3_class_rset(result)
    expect_identical(result, x)
  }
})

test_that("for nested_cv, `inner_resamples` is also a protected column", {
  x <- rset_subclasses$nested_cv
  cols <- list(inner_resamples = rep(1, vec_size(x)))
  expect_s3_class_bare_tibble(dplyr_col_modify(x, cols))
})

# ------------------------------------------------------------------------------
# dplyr_row_slice()

test_that("row slicing generally removes the rset subclass", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(dplyr_row_slice(x, 0))
  }
})

test_that("row slicing and duplicating any rows removes the rset subclass", {
  # Remove rsets with only 1 row
  subclasses <- rset_subclasses
  subclasses$apparent <- NULL
  subclasses$validation_split <- NULL

  for (x in subclasses) {
    loc <- seq_len(nrow(x))
    loc[length(loc)] <- 1L
    expect_s3_class_bare_tibble(dplyr_row_slice(x, loc))
  }
})

test_that("row slicing and selecting everything keeps the rset subclass", {
  for (x in rset_subclasses) {
    loc <- seq_len(nrow(x))
    expect_s3_class_rset(dplyr_row_slice(x, loc))
  }
})

test_that("rset subclass is kept if row order is changed but all rows are present", {
  for (x in rset_subclasses) {
    loc <- rev(seq_len(nrow(x)))
    expect_s3_class_rset(dplyr_row_slice(x, loc))
  }
})

# ------------------------------------------------------------------------------
# mutate()

test_that("mutate() can keep rset class", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(mutate(x, x = 1))
    expect_identical(mutate(x, x = 1)$x, rep(1, vec_size(x)))
  }
})

test_that("mutate() drops rset class if any rset columns are touched", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(mutate(x, splits = 1))
    expect_s3_class_bare_tibble(mutate(x, id = 1))

    expect_identical(mutate(x, splits = 1)$splits, rep(1, vec_size(x)))
    expect_identical(mutate(x, id = 1)$id, rep(1, vec_size(x)))
  }
})

test_that("mutate() keeps rset class if replacement rset column is same as original", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(mutate(x, splits = splits))
    expect_s3_class_rset(mutate(x, id = id))
  }
})

test_that("adding a column that looks like an `id` drops the class", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(mutate(x, id9 = 1))
  }
})

# ------------------------------------------------------------------------------
# arrange()

test_that("arrange() keeps rset class when row order is modified", {
  for (x in rset_subclasses) {
    x <- mutate(x, rn = row_number())
    expect_s3_class_rset(arrange(x, desc(rn)))
  }
})

test_that("arrange() keeps rset class when row order is untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(arrange(x))

    x <- mutate(x, rn = row_number())
    expect_s3_class_rset(arrange(x, rn))
  }
})

# ------------------------------------------------------------------------------
# filter()

test_that("filter() drops rset class when rows are modified", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(filter(x, 0 == 1))
    expect_s3_class_bare_tibble(filter(x, is.numeric(id)))
  }
})

test_that("filter() keeps rset class if row order is untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(filter(x))
    expect_s3_class_rset(filter(x, is.character(id)))
  }
})

# ------------------------------------------------------------------------------
# rename()

test_that("renaming can keep the rset class", {
  for (x in rset_subclasses) {
    x <- mutate(x, a = 1)
    x <- rename(x, b = a)
    expect_s3_class_rset(x)
  }
})

test_that("renaming `id` at all drops the rset class", {
  for (x in rset_subclasses) {
    x <- rename(x, id9 = id)
    expect_s3_class_bare_tibble(x)
  }
})

test_that("renaming `id` to a non-id name drops the rset class", {
  for (x in rset_subclasses) {
    x <- rename(x, stuff = id)
    expect_s3_class_bare_tibble(x)
  }
})

test_that("for nested_cv, renaming `inner_resamples` drops the rset class", {
  x <- rset_subclasses$nested_cv
  x <- rename(x, inner_stuff = inner_resamples)
  expect_s3_class_bare_tibble(x)
})

# ------------------------------------------------------------------------------
# select()

test_that("select() can keep rset class", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(select(x, everything()))
  }
})

test_that("select() drops rset class if any rset columns aren't selected", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(select(x, id))
    expect_s3_class_bare_tibble(select(x, splits))
  }
})

# ------------------------------------------------------------------------------
# slice()

test_that("slice() drops rset class when rows are modified", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(slice(x, 0))
  }
})

test_that("slice() keeps rset class when rows are untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(slice(x))
    expect_s3_class_rset(slice(x, seq_len(nrow(x))))
  }
})

# ------------------------------------------------------------------------------
# summarise()

test_that("summarise() always drops the rset class", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(summarise(x, y = 1))
    expect_s3_class_bare_tibble(summarise(x, splits = splits[1], id = id[1]))
  }
})

# ------------------------------------------------------------------------------
# group_by()

test_that("group_by() always returns a bare grouped-df or bare tibble", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(group_by(x))
    expect_s3_class(group_by(x, splits), c("grouped_df", "tbl_df", "tbl", "data.frame"), exact = TRUE)
  }
})

# ------------------------------------------------------------------------------
# ungroup()

test_that("ungroup() returns a rset", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(ungroup(x))
  }
})

# ------------------------------------------------------------------------------
# relocate()

test_that("can relocate() and keep the class", {
  for (x in rset_subclasses) {
    x <- relocate(x, id)
    expect_s3_class_rset(x)
  }
})

# ------------------------------------------------------------------------------
# distinct()

test_that("distinct() keeps the class if everything is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(distinct(x))
  }
})

test_that("distinct() drops the class if any rset columns are lost", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(distinct(x, splits))
  }
})

# ------------------------------------------------------------------------------
# left_join()

test_that("left_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(left_join(x, x, by = names(x)))

    y <- tibble(id = x$id[[1]], x = 1)
    expect_s3_class_rset(left_join(x, y, by = "id"))
  }
})

test_that("left_join() can lose rset class if rows are added", {
  for (x in rset_subclasses) {
    y <- tibble(id = x$id[[1]], x = 1:2)
    expect_s3_class_bare_tibble(left_join(x, y, by = "id"))
  }
})

# ------------------------------------------------------------------------------
# right_join()

test_that("right_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(right_join(x, x, by = names(x)))

    x_names <- names(x)
    id_names <- x_names[col_starts_with_id(x_names)]

    y <- mutate(select(x, all_of(id_names)), x = 1)
    expect_s3_class_rset(right_join(x, y, by = id_names))
  }
})

test_that("right_join() can lose rset class if rows are added", {
  for (x in rset_subclasses) {
    y <- tibble(id = x$id[[1]], x = 1:2)
    expect_s3_class_bare_tibble(right_join(x, y, by = "id"))
  }
})

test_that("right_join() restores to the type of first input", {
  for (x in rset_subclasses) {
    y <- tibble(id = x$id[[1]], x = 1)
    # technically rset structure is intact, but `y` is a bare tibble!
    expect_s3_class_bare_tibble(right_join(y, x, by = "id"))
  }
})

# ------------------------------------------------------------------------------
# full_join()

test_that("full_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(full_join(x, x, by = names(x)))
  }
})

test_that("full_join() can lose rset class if rows are added", {
  for (x in rset_subclasses) {
    y <- tibble(id = "foo", x = 1)
    expect_s3_class_bare_tibble(full_join(x, y, by = "id"))
  }
})

# ------------------------------------------------------------------------------
# anti_join()

test_that("anti_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    y <- tibble(id = "foo")
    expect_s3_class_rset(anti_join(x, y, by = "id"))
  }
})

test_that("anti_join() can lose rset class if rows are removed", {
  for (x in rset_subclasses) {
    y <- tibble(id = x$id[[1]], x = 1)
    expect_s3_class_bare_tibble(anti_join(x, y, by = "id"))
  }
})

# ------------------------------------------------------------------------------
# semi_join()

test_that("semi_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(semi_join(x, x, by = names(x)))
  }
})

test_that("semi_join() can lose rset class if rows are removed", {
  for (x in rset_subclasses) {
    y <- tibble(id = "foo", x = 1)
    expect_s3_class_bare_tibble(semi_join(x, y, by = "id"))
  }
})

# ------------------------------------------------------------------------------
# nest_join()

test_that("nest_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    y <- mutate(x, foo = "bar")
    expect_s3_class_rset(nest_join(x, y, by = names(x)))
  }
})

# ------------------------------------------------------------------------------
# bind_rows()

test_that("bind_rows() keeps the class if there are no new rows/cols and the first object is an rset subclass", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(bind_rows(x))
    expect_s3_class_rset(bind_rows(x, tibble()))
    expect_s3_class_bare_tibble(bind_rows(tibble(), x))
  }
})

test_that("bind_rows() drops the class with new rows", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(bind_rows(x, x))
  }
})

# ------------------------------------------------------------------------------
# bind_cols()

test_that("bind_cols() keeps the class if there are no new rows and the first object is an rset subclass", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(bind_cols(x))
    expect_s3_class_rset(bind_cols(x, tibble(x = 1)))
    expect_s3_class_bare_tibble(bind_cols(tibble(x = 1), x))
  }
})

test_that("bind_cols() drops the class with new rows", {
  # Use rset subclass with 1 row, these get recycled
  x <- rset_subclasses$apparent
  expect_s3_class_bare_tibble(bind_cols(x, tibble(x = 1:2)))
})

skip_if_not(rlang::is_installed("withr"))
# ------------------------------------------------------------------------------
# vec_restore()

test_that("vec_restore() returns an rset subclass if `x` retains rset structure", {
  for (x in rset_subclasses) {
    expect_identical(vec_restore(x, x), x)
    expect_s3_class_rset(vec_restore(x, x))
  }
})

test_that("vec_restore() returns bare tibble if `x` loses rset structure", {
  for (x in rset_subclasses) {
    col <- x[1]
    row <- x[0, ]

    expect_s3_class_bare_tibble(vec_restore(col, x))
    expect_s3_class_bare_tibble(vec_restore(row, x))
  }
})

test_that("vec_restore() retains extra attributes of `to` when not falling back", {
  for (x in rset_subclasses) {
    to <- x
    attr(to, "foo") <- "bar"

    x_tbl <- x[1]

    expect_identical(attr(vec_restore(x, to), "foo"), "bar")
    expect_identical(attr(vec_restore(x_tbl, to), "foo"), NULL)

    expect_s3_class_rset(vec_restore(x, to))
    expect_s3_class_bare_tibble(vec_restore(x_tbl, to))
  }
})

# ------------------------------------------------------------------------------
# vec_ptype2()

test_that("vec_ptype2() is working", {
  for (x in rset_subclasses) {
    tbl <- tibble::tibble(x = 1)
    df <- data.frame(x = 1)

    # rset-rset
    expect_identical(vec_ptype2(x, x), vec_ptype2(tib_upcast(x), tib_upcast(x)))

    # rset-tbl_df
    expect_identical(vec_ptype2(x, tbl), vec_ptype2(tib_upcast(x), tbl))
    expect_identical(vec_ptype2(tbl, x), vec_ptype2(tbl, tib_upcast(x)))

    # rset-df
    expect_identical(vec_ptype2(x, df), vec_ptype2(tib_upcast(x), df))
    expect_identical(vec_ptype2(df, x), vec_ptype2(df, tib_upcast(x)))
  }
})

# ------------------------------------------------------------------------------
# vec_cast()

test_that("vec_cast() is working", {
  for (x in rset_subclasses) {
    tbl <- tib_upcast(x)
    df <- as.data.frame(tbl)

    # rset-rset
    expect_error(vec_cast(x, x), class = "vctrs_error_incompatible_type")

    # rset-tbl_df
    expect_identical(vec_cast(x, tbl), tbl)
    expect_error(vec_cast(tbl, x), class = "vctrs_error_incompatible_type")

    # rset-df
    expect_identical(vec_cast(x, df), df)
    expect_error(vec_cast(df, x), class = "vctrs_error_incompatible_type")
  }
})

# ------------------------------------------------------------------------------
# vctrs methods

test_that("vec_ptype() returns a bare tibble", {
  for (x in rset_subclasses) {
    expect_identical(vec_ptype(x), vec_ptype(tib_upcast(x)))
    expect_s3_class_bare_tibble(vec_ptype(x))
  }
})

test_that("vec_slice() generally returns a bare tibble", {
  for (x in rset_subclasses) {
    expect_identical(vec_slice(x, 0), vec_slice(tib_upcast(x), 0))
    expect_s3_class_bare_tibble(vec_slice(x, 0))
  }
})

test_that("vec_slice() can return an rset if all rows are selected", {
  for (x in rset_subclasses) {
    expect_identical(vec_slice(x, TRUE), x)
    expect_s3_class_rset(vec_slice(x, TRUE))
  }
})

test_that("vec_c() returns a bare tibble", {
  for (x in rset_subclasses) {
    tbl <- tib_upcast(x)

    expect_identical(vec_c(x), vec_c(tbl))
    expect_identical(vec_c(x, x), vec_c(tbl, tbl))
    expect_identical(vec_c(x, tbl), vec_c(tbl, tbl))

    expect_s3_class_bare_tibble(vec_c(x))
    expect_s3_class_bare_tibble(vec_c(x, x))
  }
})

test_that("vec_rbind() returns a bare tibble", {
  for (x in rset_subclasses) {
    tbl <- tib_upcast(x)

    expect_identical(vec_rbind(x), vec_rbind(tbl))
    expect_identical(vec_rbind(x, x), vec_rbind(tbl, tbl))
    expect_identical(vec_rbind(x, tbl), vec_rbind(tbl, tbl))

    expect_s3_class_bare_tibble(vec_rbind(x))
    expect_snapshot(expect_s3_class_bare_tibble(vec_cbind(x, x)))
  }
})

test_that("vec_cbind() returns a bare tibble", {
  for (x in rset_subclasses) {
    tbl <- tib_upcast(x)

    expect_identical(vec_cbind(x), vec_cbind(tbl))
    expect_snapshot(expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl)))
    expect_snapshot(expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl)))

    expect_s3_class_bare_tibble(vec_cbind(x))
    expect_snapshot(expect_s3_class_bare_tibble(vec_cbind(x, x)))
  }
})

library(dplyr)
skip_if_not_installed("withr")
# ------------------------------------------------------------------------------
# dplyr_reconstruct()

test_that("dplyr_reconstruct() returns an rset subclass if `x` retains rset structure", {
  for (x in rset_subclasses) {
    expect_identical(dplyr_reconstruct(x, x), x)
    expect_s3_class_rset(dplyr_reconstruct(x, x))
  }
})

test_that("dplyr_reconstruct() returns bare tibble if `x` loses rset structure", {
  for (x in rset_subclasses) {
    col <- x[1]
    row <- x[0, ]

    expect_s3_class_bare_tibble(dplyr_reconstruct(col, x))
    expect_s3_class_bare_tibble(dplyr_reconstruct(row, x))
  }
})

test_that("dplyr_reconstruct() retains extra attributes of `to` when not falling back", {
  for (x in rset_subclasses) {
    to <- x
    attr(to, "foo") <- "bar"

    x_tbl <- x[1]

    expect_identical(attr(dplyr_reconstruct(x, to), "foo"), "bar")
    expect_identical(attr(dplyr_reconstruct(x_tbl, to), "foo"), NULL)

    expect_s3_class_rset(dplyr_reconstruct(x, to))
    expect_s3_class_bare_tibble(dplyr_reconstruct(x_tbl, to))
  }
})

# ------------------------------------------------------------------------------
# dplyr_col_modify()

test_that("can add columns and retain rset class", {
  for (x in rset_subclasses) {
    cols <- list(x = rep(1, vec_size(x)))

    result <- dplyr_col_modify(x, cols)

    expect_s3_class_rset(result)
    expect_identical(result$x, cols$x)
  }
})

test_that("modifying rset columns removes rset class", {
  for (x in rset_subclasses) {
    cols <- list(splits = rep(1, vec_size(x)))

    result <- dplyr_col_modify(x, cols)

    expect_s3_class_bare_tibble(result)
    expect_identical(result$splits, cols$splits)
  }

  for (x in rset_subclasses) {
    cols <- list(id = rep(1, vec_size(x)))

    result <- dplyr_col_modify(x, cols)

    expect_s3_class_bare_tibble(result)
    expect_identical(result$id, cols$id)
  }
})

test_that("replacing rset columns with the exact same column retains rset class", {
  for (x in rset_subclasses) {
    cols <- list(splits = x$splits)

    result <- dplyr_col_modify(x, cols)

    expect_s3_class_rset(result)
    expect_identical(result, x)
  }
})

test_that("for nested_cv, `inner_resamples` is also a protected column", {
  x <- rset_subclasses$nested_cv
  cols <- list(inner_resamples = rep(1, vec_size(x)))
  expect_s3_class_bare_tibble(dplyr_col_modify(x, cols))
})

# ------------------------------------------------------------------------------
# dplyr_row_slice()

test_that("row slicing generally removes the rset subclass", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(dplyr_row_slice(x, 0))
  }
})

test_that("row slicing and duplicating any rows removes the rset subclass", {
  # Remove rsets with only 1 row
  subclasses <- rset_subclasses
  subclasses$apparent <- NULL
  subclasses$validation_split <- NULL
  subclasses$validation_time_split <- NULL
  subclasses$group_validation_split <- NULL
  subclasses$validation_set <- NULL

  for (x in subclasses) {
    loc <- seq_len(nrow(x))
    loc[length(loc)] <- 1L
    expect_s3_class_bare_tibble(dplyr_row_slice(x, loc))
  }
})

test_that("row slicing and selecting everything keeps the rset subclass", {
  for (x in rset_subclasses) {
    loc <- seq_len(nrow(x))
    expect_s3_class_rset(dplyr_row_slice(x, loc))
  }
})

test_that("rset subclass is kept if row order is changed but all rows are present", {
  for (x in rset_subclasses) {
    loc <- rev(seq_len(nrow(x)))
    expect_s3_class_rset(dplyr_row_slice(x, loc))
  }
})

# ------------------------------------------------------------------------------
# mutate()

test_that("mutate() can keep rset class", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(mutate(x, x = 1))
    expect_identical(mutate(x, x = 1)$x, rep(1, vec_size(x)))
  }
})

test_that("mutate() drops rset class if any rset columns are touched", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(mutate(x, splits = 1))
    expect_s3_class_bare_tibble(mutate(x, id = 1))

    expect_identical(mutate(x, splits = 1)$splits, rep(1, vec_size(x)))
    expect_identical(mutate(x, id = 1)$id, rep(1, vec_size(x)))
  }
})

test_that("mutate() keeps rset class if replacement rset column is same as original", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(mutate(x, splits = splits))
    expect_s3_class_rset(mutate(x, id = id))
  }
})

test_that("adding a column that looks like an `id` drops the class", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(mutate(x, id9 = 1))
  }
})

# ------------------------------------------------------------------------------
# arrange()

test_that("arrange() keeps rset class when row order is modified", {
  for (x in rset_subclasses) {
    x <- mutate(x, rn = row_number())
    expect_s3_class_rset(arrange(x, desc(rn)))
  }
})

test_that("arrange() keeps rset class when row order is untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(arrange(x))

    x <- mutate(x, rn = row_number())
    expect_s3_class_rset(arrange(x, rn))
  }
})

# ------------------------------------------------------------------------------
# filter()

test_that("filter() drops rset class when rows are modified", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(filter(x, 0 == 1))
    expect_s3_class_bare_tibble(filter(x, is.numeric(id)))
  }
})

test_that("filter() keeps rset class if row order is untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(filter(x))
    expect_s3_class_rset(filter(x, is.character(id)))
  }
})

# ------------------------------------------------------------------------------
# rename()

test_that("renaming can keep the rset class", {
  for (x in rset_subclasses) {
    x <- mutate(x, a = 1)
    x <- rename(x, b = a)
    expect_s3_class_rset(x)
  }
})

test_that("renaming `id` at all drops the rset class", {
  for (x in rset_subclasses) {
    x <- rename(x, id9 = id)
    expect_s3_class_bare_tibble(x)
  }
})

test_that("renaming `id` to a non-id name drops the rset class", {
  for (x in rset_subclasses) {
    x <- rename(x, stuff = id)
    expect_s3_class_bare_tibble(x)
  }
})

test_that("for nested_cv, renaming `inner_resamples` drops the rset class", {
  x <- rset_subclasses$nested_cv
  x <- rename(x, inner_stuff = inner_resamples)
  expect_s3_class_bare_tibble(x)
})

# ------------------------------------------------------------------------------
# select()

test_that("select() can keep rset class", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(select(x, everything()))
  }
})

test_that("select() drops rset class if any rset columns aren't selected", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(select(x, id))
    expect_s3_class_bare_tibble(select(x, splits))
  }
})

# ------------------------------------------------------------------------------
# slice()

test_that("slice() drops rset class when rows are modified", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(slice(x, 0))
  }
})

test_that("slice() keeps rset class when rows are untouched", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(slice(x, seq_len(nrow(x))))
  }
})

# ------------------------------------------------------------------------------
# summarise()

test_that("summarise() always drops the rset class", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(summarise(x, y = 1))
    expect_s3_class_bare_tibble(summarise(x, splits = splits[1], id = id[1]))
  }
})

# ------------------------------------------------------------------------------
# group_by()

test_that("group_by() always returns a bare grouped-df or bare tibble", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(group_by(x))
    expect_s3_class(
      group_by(x, splits),
      c("grouped_df", "tbl_df", "tbl", "data.frame"),
      exact = TRUE
    )
  }
})

# ------------------------------------------------------------------------------
# ungroup()

test_that("ungroup() returns a rset", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(ungroup(x))
  }
})

# ------------------------------------------------------------------------------
# relocate()

test_that("can relocate() and keep the class", {
  for (x in rset_subclasses) {
    x <- relocate(x, id)
    expect_s3_class_rset(x)
  }
})

# ------------------------------------------------------------------------------
# distinct()

test_that("distinct() keeps the class if everything is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(distinct(x))
  }
})

test_that("distinct() drops the class if any rset columns are lost", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(distinct(x, splits))
  }
})

# ------------------------------------------------------------------------------
# left_join()

test_that("left_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(left_join(x, x, by = names(x)))

    y <- tibble(id = x$id[[1]], x = 1)
    expect_s3_class_rset(left_join(x, y, by = "id"))
  }
})

test_that("left_join() can lose rset class if rows are added", {
  for (x in rset_subclasses) {
    y <- tibble(id = x$id[[1]], x = 1:2)
    expect_s3_class_bare_tibble(left_join(
      x,
      y,
      by = "id",
      multiple = "all",
      relationship = "many-to-many"
    ))
  }
})

# ------------------------------------------------------------------------------
# right_join()

test_that("right_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(right_join(x, x, by = names(x)))

    x_names <- names(x)
    id_names <- x_names[col_starts_with_id(x_names)]

    y <- mutate(select(x, all_of(id_names)), x = 1)
    expect_s3_class_rset(right_join(x, y, by = id_names))
  }
})

test_that("right_join() can lose rset class if rows are added", {
  for (x in rset_subclasses) {
    y <- tibble(id = x$id[[1]], x = 1:2)
    expect_s3_class_bare_tibble(right_join(
      x,
      y,
      by = "id",
      multiple = "all",
      relationship = "many-to-many"
    ))
  }
})

test_that("right_join() restores to the type of first input", {
  for (x in rset_subclasses) {
    y <- tibble(id = x$id[[1]], x = 1)
    # technically rset structure is intact, but `y` is a bare tibble!
    expect_s3_class_bare_tibble(right_join(y, x, by = "id", multiple = "all"))
  }
})

# ------------------------------------------------------------------------------
# full_join()

test_that("full_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(full_join(x, x, by = names(x)))
  }
})

test_that("full_join() can lose rset class if rows are added", {
  for (x in rset_subclasses) {
    y <- tibble(id = "foo", x = 1)
    expect_s3_class_bare_tibble(full_join(x, y, by = "id"))
  }
})

# ------------------------------------------------------------------------------
# anti_join()

test_that("anti_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    y <- tibble(id = "foo")
    expect_s3_class_rset(anti_join(x, y, by = "id"))
  }
})

test_that("anti_join() can lose rset class if rows are removed", {
  for (x in rset_subclasses) {
    y <- tibble(id = x$id[[1]], x = 1)
    expect_s3_class_bare_tibble(anti_join(x, y, by = "id"))
  }
})

# ------------------------------------------------------------------------------
# semi_join()

test_that("semi_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(semi_join(x, x, by = names(x)))
  }
})

test_that("semi_join() can lose rset class if rows are removed", {
  for (x in rset_subclasses) {
    y <- tibble(id = "foo", x = 1)
    expect_s3_class_bare_tibble(semi_join(x, y, by = "id"))
  }
})

# ------------------------------------------------------------------------------
# nest_join()

test_that("nest_join() can keep rset class if rset structure is intact", {
  for (x in rset_subclasses) {
    y <- mutate(x, foo = "bar")
    expect_s3_class_rset(nest_join(x, y, by = names(x)))
  }
})

# ------------------------------------------------------------------------------
# bind_rows()

test_that("bind_rows() keeps the class if there are no new rows/cols and the first object is an rset subclass", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(bind_rows(x))
    expect_s3_class_rset(bind_rows(x, tibble()))
    expect_s3_class_bare_tibble(bind_rows(tibble(), x))
  }
})

test_that("bind_rows() drops the class with new rows", {
  for (x in rset_subclasses) {
    expect_s3_class_bare_tibble(bind_rows(x, x))
  }
})

# ------------------------------------------------------------------------------
# bind_cols()

test_that("bind_cols() keeps the class if there are no new rows and the first object is an rset subclass", {
  for (x in rset_subclasses) {
    expect_s3_class_rset(bind_cols(x))
    expect_s3_class_rset(bind_cols(x, tibble(x = 1)))
    expect_s3_class_bare_tibble(bind_cols(tibble(x = 1), x))
  }
})

test_that("bind_cols() drops the class with new rows", {
  # Use rset subclass with 1 row, these get recycled
  x <- rset_subclasses$apparent
  expect_s3_class_bare_tibble(bind_cols(x, tibble(x = 1:2)))
})

# Alias required for help links in downstream packages
#' @aliases select_helpers
#' @importFrom tidyselect contains
#' @export
tidyselect::contains
#' @importFrom tidyselect ends_with
#' @export
tidyselect::ends_with
#' @importFrom tidyselect everything
#' @export
tidyselect::everything
#' @importFrom tidyselect matches
#' @export
tidyselect::matches
#' @importFrom tidyselect num_range
#' @export
tidyselect::num_range
#' @importFrom tidyselect starts_with
#' @export
tidyselect::starts_with
#' @importFrom tidyselect last_col
#' @export
tidyselect::last_col
#' @importFrom tidyselect any_of
#' @export
tidyselect::any_of
#' @importFrom tidyselect all_of
#' @export
tidyselect::all_of

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import rlang
#' @importFrom lifecycle deprecated
#' @importFrom cli cli_abort cli_warn cli_text
#' @importFrom utils globalVariables
#' @importFrom purrr map map2 map_dbl pluck map_lgl list_rbind
#' @importFrom tibble tibble is_tibble as_tibble obj_sum
#' @importFrom pillar type_sum
#' @importFrom tidyr unnest
#' @importFrom dplyr select bind_cols bind_rows arrange_ arrange full_join
#' @importFrom dplyr mutate last ungroup group_by inner_join summarize do case_when
#' @importFrom vctrs vec_unique_count vec_count vec_slice vec_in vec_ptype_abbr
#' @importFrom methods formalArgs
#' @importFrom stats quantile setNames qnorm pnorm dist
#' @importFrom tidyselect vars_select one_of
#' @importFrom furrr future_map
## usethis namespace: end
NULL

#------------------------------------------------------------------------------#

utils::globalVariables(
  c(
    "model",
    "splits",
    "statistic",
    "Data",
    "Row",
    "id",
    ".",
    ".estimate",
    ".lower",
    ".upper",
    "Z0",
    "Za",
    "Zl",
    "Zu",
    "a",
    "cubed",
    "estimate",
    "orig",
    "p0",
    "squared",
    "term",
    "theta_0",
    "loo",
    "n",
    "..index",
    "models",
    paste0("predictor_", gsub(" ", "0", format(1:20))),
    "linear_pred",
    "non_linear_1",
    "non_linear_2",
    "non_linear_3",
    "outcome",
    "true_prob",
    "two_factor_1",
    "two_factor_2",
    "rand"
  )
)

#' Monte Carlo Cross-Validation
#'
#' One resample of Monte Carlo cross-validation takes a random sample (without
#'  replacement) of the original data set to be used for analysis. All other
#'  data points are added to the assessment set.
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @param times The number of times to repeat the sampling.
#' @export
#' @return An tibble with classes `mc_cv`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and a
#'  column called `id` that has a character string with the resample identifier.
#' @examples
#' mc_cv(mtcars, times = 2)
#' mc_cv(mtcars, prop = .5, times = 2)
#'
#' library(purrr)
#' data(wa_churn, package = "modeldata")
#'
#' set.seed(13)
#' resample1 <- mc_cv(wa_churn, times = 3, prop = .5)
#' map_dbl(
#'   resample1$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' resample2 <- mc_cv(wa_churn, strata = churn, times = 3, prop = .5)
#' map_dbl(
#'   resample2$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' resample3 <- mc_cv(wa_churn, strata = tenure, breaks = 6, times = 3, prop = .5)
#' map_dbl(
#'   resample3$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#' @export
mc_cv <- function(data, prop = 3 / 4, times = 25,
                  strata = NULL, breaks = 4, pool = 0.1, ...) {
  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) strata <- NULL
  }

  strata_check(strata, data)

  split_objs <-
    mc_splits(
      data = data,
      prop = prop,
      times = times,
      strata = strata,
      breaks = breaks,
      pool = pool
    )

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  mc_att <- list(
    prop = prop,
    times = times,
    strata = !is.null(strata)
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = mc_att,
    subclass = c("mc_cv", "rset")
  )
}

# Get the indices of the assessment set from the analysis set
mc_complement <- function(ind, n) {
  list(
    analysis = ind,
    assessment = setdiff(1:n, ind)
  )
}


mc_splits <- function(data, prop = 3 / 4, times = 25,
                      strata = NULL, breaks = 4, pool = 0.1) {
  if (!is.numeric(prop) | prop >= 1 | prop <= 0) {
    rlang::abort("`prop` must be a number on (0, 1).")
  }

  n <- nrow(data)
  if (is.null(strata)) {
    indices <- purrr::map(rep(n, times), sample, size = floor(n * prop))
  } else {
    stratas <- tibble::tibble(
      idx = 1:n,
      strata = make_strata(getElement(data, strata),
        breaks = breaks,
        pool = pool
      )
    )
    stratas <- split_unnamed(stratas, stratas$strata)
    stratas <-
      purrr::map_df(stratas, strat_sample, prop = prop, times = times)
    indices <- split_unnamed(stratas$idx, stratas$rs_id)
  }
  indices <- lapply(indices, mc_complement, n = n)
  split_objs <-
    purrr::map(indices, make_splits, data = data, class = "mc_split")
  list(
    splits = split_objs,
    id = names0(length(split_objs), "Resample")
  )
}

strat_sample <- function(x, prop, times, ...) {
  n <- nrow(x)
  idx <- purrr::map(rep(n, times), sample, size = floor(n * prop), ...)
  out <- purrr::map_df(idx, function(ind, x) x[sort(ind), "idx"], x = x)
  out$rs_id <- rep(1:times, each = floor(n * prop))
  out
}

# These compat methods are for dplyr < 1.0.0.
# They are conditionally registered in `zzz.R` if an old version of dplyr is
# being used. Otherwise the methods in `compat-dplyr.R` are registered.
# Eventually these should be deprecated and dplyr 1.0.0 should be required.
# Only verbs that were previously supported in rsample are continually
# supported now. This means that there is more complete support when a user
# upgrades to dplyr >= 1.0.0. When support for dplyr < 1.0.0 is deprecated,
# move the tests in `compact-dplyr-old.R` to `compat-dplyr.R`. The methods
# are named with underscores between the generic and the class, not with dots,
# because when working interactively with `load_all()`, just having
# a function named `mutate.rset` will register a dplyr method internally in
# the package, even if you are on dplyr 1.0.0, which is highly confusing and
# undesirable. So we work around that by naming them in such a way that they
# are not picked up by S3 registration, and then we manually register them
# with the `method` argument of `s3_register()`.

dplyr_pre_1.0.0 <- function() {
  utils::packageVersion("dplyr") <= "0.8.5"
}

# Registered in `.onLoad()`
mutate_rset <- function(.data, ...) {
  out <- NextMethod()
  rset_reconstruct(out, .data)
}

# Registered in `.onLoad()`
arrange_rset <- function(.data, ...) {
  out <- NextMethod()
  rset_reconstruct(out, .data)
}

# Registered in `.onLoad()`
filter_rset <- function(.data, ...) {
  out <- NextMethod()
  rset_reconstruct(out, .data)
}

# Registered in `.onLoad()`
rename_rset <- function(.data, ...) {
  out <- NextMethod()
  rset_reconstruct(out, .data)
}

# Registered in `.onLoad()`
select_rset <- function(.data, ...) {
  out <- NextMethod()
  rset_reconstruct(out, .data)
}

# Registered in `.onLoad()`
slice_rset <- function(.data, ...) {
  out <- NextMethod()
  rset_reconstruct(out, .data)
}

#' Create a Validation Split for Tuning
#'
#' `validation_set()` creates a the validation split for model tuning.
#'
#' @param split An object of class `initial_validation_split`, such as resulting
#' from [initial_validation_split()] or [group_initial_validation_split()].
#' @param x An `rsplit` object produced by `validation_set()`.
#' @inheritParams rlang::args_dots_empty
#'
#' @return An tibble with classes `validation_set`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split object and a
#'  column called `id` that has a character string with the resample identifier.
#' @export
#'
#' @examples
#' set.seed(1353)
#' car_split <- initial_validation_split(mtcars)
#' car_set <- validation_set(car_split)
validation_set <- function(split, ...) {
  rlang::check_dots_empty()

  train_and_val <- dplyr::bind_rows(
    training(split),
    validation(split)
  )
  n_train <- length(split$train_id)

  val_split <- rsplit(
    data = train_and_val,
    in_id = seq_len(n_train),
    out_id = NA
  )

  split_type <- switch(
    class(split)[1],
    "group_initial_validation_split" = "group",
    "initial_validation_time_split" = "time",
    "initial_validation_split" = NULL
  )

  # "val_split" is same class as via the alternative `validation_split()`
  if (is.null(split_type)) {
    class(val_split) <- c("val_split", "rsplit")
  } else {
    class(val_split) <- c(
      paste0(split_type, "_val_split"),
      "val_split",
      "rsplit"
    )
  }

  val_att <- attr(split, "val_att")
  val_att[["origin_3way"]] <- TRUE

  if (is.null(split_type)) {
    rset_classes <- c("validation_set", "rset")
  } else {
    rset_classes <- c(
      paste0(split_type, "_validation_set"),
      "validation_set",
      "rset"
    )
  }

  new_rset(
    splits = list(val_split),
    ids = "validation",
    attrib = val_att,
    subclass = rset_classes
  )
}

# accessor methods for the rsplit -----------------------------------------

#' @rdname validation_set
#' @export
analysis.val_split <- function(x, ...) {
  as.data.frame(x, data = "analysis", ...)
}

#' @rdname validation_set
#' @export
assessment.val_split <- function(x, ...) {
  as.data.frame(x, data = "assessment", ...)
}

#' @rdname validation_set
#' @export
training.val_split <- function(x, ...) {
  analysis(x, ...)
}

#' @rdname validation_set
#' @export
validation.val_split <- function(x, ...) {
  assessment(x, ...)
}

#' @rdname validation_set
#' @export
testing.val_split <- function(x, ...) {
  cli_abort(c(
    "The testing data is not part of the validation set object.",
    "i" = "It is part of the result of the initial 3-way split, e.g., with {.fun initial_validation_split}."
  ))
}

#' Find Labels from rset Object
#'
#' Produce a vector of resampling labels (e.g. "Fold1") from
#'  an `rset` object. Currently, `nested_cv`
#'  is not supported.
#'
#' @param object An `rset` object
#' @param make_factor A logical for whether the results should be
#'  a character or a factor.
#' @param ... Not currently used.
#' @return A single character or factor vector.
#' @export
#' @examples
#' labels(vfold_cv(mtcars))
labels.rset <- function(object, make_factor = FALSE, ...) {
  if (inherits(object, "nested_cv")) {
    rlang::abort("`labels` not implemented for nested resampling")
  }
  if (make_factor) {
    as.factor(object$id)
  } else {
    as.character(object$id)
  }
}

#' @rdname labels.rset
#' @export
labels.vfold_cv <- function(object, make_factor = FALSE, ...) {
  if (inherits(object, "nested_cv")) {
    rlang::abort("`labels` not implemented for nested resampling")
  }
  is_repeated <- attr(object, "repeats") > 1
  if (is_repeated) {
    out <- as.character(paste(object$id, object$id2, sep = "."))
  } else {
    out <- as.character(object$id)
  }
  if (make_factor) {
    out <- as.factor(out)
  }
  out
}

#' Find Labels from rsplit Object
#'
#' Produce a tibble of identification variables so that single
#'  splits can be linked to a particular resample.
#'
#' @param object An `rsplit` object
#' @param ... Not currently used.
#' @return A tibble.
#' @seealso add_resample_id
#' @export
#' @examples
#' cv_splits <- vfold_cv(mtcars)
#' labels(cv_splits$splits[[1]])
labels.rsplit <- function(object, ...) {
  out <- if ("id" %in% names(object)) {
    object$id
  } else {
    tibble()
  }
  out
}

#' Augment a data set with resampling identifiers
#'
#' For a data set, `add_resample_id()` will add at least one new column that
#'  identifies which resample that the data came from. In most cases, a single
#'  column is added but for some resampling methods, two or more are added.
#' @param .data A data frame
#' @param split A single `rset` object.
#' @param dots A single logical: should the id columns be prefixed with a "."
#'  to avoid name conflicts with `.data`?
#' @return An updated data frame.
#' @examples
#' library(dplyr)
#'
#' set.seed(363)
#' car_folds <- vfold_cv(mtcars, repeats = 3)
#'
#' analysis(car_folds$splits[[1]]) %>%
#'   add_resample_id(car_folds$splits[[1]]) %>%
#'   head()
#'
#' car_bt <- bootstraps(mtcars)
#'
#' analysis(car_bt$splits[[1]]) %>%
#'   add_resample_id(car_bt$splits[[1]]) %>%
#'   head()
#' @seealso labels.rsplit
#' @export
add_resample_id <- function(.data, split, dots = FALSE) {
  if (!inherits(dots, "logical") || length(dots) > 1) {
    rlang::abort("`dots` should be a single logical.")
  }
  if (!inherits(.data, "data.frame")) {
    rlang::abort("`.data` should be a data frame.")
  }
  if (!inherits(split, "rsplit")) {
    rlang::abort("`split` should be a single 'rset' object.")
  }
  labs <- labels(split)

  if (!tibble::is_tibble(labs) && nrow(labs) == 1) {
    rlang::abort("`split` should be a single 'rset' object.")
  }

  if (dots) {
    colnames(labs) <- paste0(".", colnames(labs))
  }

  cbind(.data, labs)
}

#' V-Fold Cross-Validation
#'
#' V-fold cross-validation (also known as k-fold cross-validation) randomly
#'  splits the data into V groups of roughly equal size (called "folds"). A
#'  resample of the analysis data consists of V-1 of the folds while the
#'  assessment set contains the final fold. In basic V-fold cross-validation
#'  (i.e. no repeats), the number of resamples is equal to V.
#' @details With more than one repeat, the basic V-fold cross-validation is
#'  conducted each time. For example, if three repeats are used with `v = 10`,
#'  there are a total of 30 splits: three groups of 10 that are generated
#'  separately.
#' @template strata_details
#' @inheritParams make_strata
#' @param data A data frame.
#' @param v The number of partitions of the data set.
#' @param repeats The number of times to repeat the V-fold partitioning.
#' @param strata A variable in `data` (single character or name) used to conduct
#'  stratified sampling. When not `NULL`, each resample is created within the
#'  stratification variable. Numeric `strata` are binned into quartiles.
#' @inheritParams rlang::args_dots_empty
#' @export
#' @return A tibble with classes `vfold_cv`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and
#'  one or more identification variables. For a single repeat, there will be
#'  one column called `id` that has a character string with the fold identifier.
#'  For repeats, `id` is the repeat number and an additional column called `id2`
#'  that contains the fold information (within repeat).
#'
#' @examplesIf rlang::is_installed("modeldata")
#' vfold_cv(mtcars, v = 10)
#' vfold_cv(mtcars, v = 10, repeats = 2)
#'
#' library(purrr)
#' data(wa_churn, package = "modeldata")
#'
#' set.seed(13)
#' folds1 <- vfold_cv(wa_churn, v = 5)
#' map_dbl(
#'   folds1$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' folds2 <- vfold_cv(wa_churn, strata = churn, v = 5)
#' map_dbl(
#'   folds2$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' folds3 <- vfold_cv(wa_churn, strata = tenure, breaks = 6, v = 5)
#' map_dbl(
#'   folds3$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#' @export
vfold_cv <- function(
  data,
  v = 10,
  repeats = 1,
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  ...
) {
  check_dots_empty()
  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) strata <- NULL
  }

  check_strata(strata, data)
  check_number_whole(repeats, min = 1)

  if (repeats == 1) {
    split_objs <- vfold_splits(
      data = data,
      v = v,
      strata = strata,
      breaks = breaks,
      pool = pool
    )
  } else {
    if (v == nrow(data)) {
      cli_abort(
        "Repeated resampling when {.arg v} is {v} would create identical resamples."
      )
    }
    for (i in 1:repeats) {
      tmp <- vfold_splits(
        data = data,
        v = v,
        strata = strata,
        breaks = breaks,
        pool = pool
      )
      tmp$id2 <- tmp$id
      tmp$id <- names0(repeats, "Repeat")[i]
      split_objs <- if (i == 1) {
        tmp
      } else {
        rbind(split_objs, tmp)
      }
    }
  }

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  ## Save some overall information

  if (!is.null(strata)) {
    names(strata) <- NULL
  }
  cv_att <- list(
    v = v,
    repeats = repeats,
    strata = strata,
    breaks = breaks,
    pool = pool
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs[, grepl("^id", names(split_objs))],
    attrib = cv_att,
    subclass = c("vfold_cv", "rset")
  )
}


vfold_splits <- function(
  data,
  v = 10,
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  prevent_loo = TRUE
) {
  n <- nrow(data)
  check_v(v, n, prevent_loo = prevent_loo, call = rlang::caller_env())

  if (is.null(strata)) {
    folds <- sample(rep(1:v, length.out = n))
    idx <- seq_len(n)
    indices <- split_unnamed(idx, folds)
  } else {
    stratas <- tibble::tibble(
      idx = 1:n,
      strata = make_strata(
        getElement(data, strata),
        breaks = breaks,
        pool = pool
      )
    )
    stratas <- split_unnamed(stratas, stratas$strata)
    stratas <- purrr::map(stratas, add_vfolds, v = v)
    stratas <- dplyr::bind_rows(stratas)
    indices <- split_unnamed(stratas$idx, stratas$folds)
  }

  indices <- lapply(indices, default_complement, n = n)

  split_objs <- purrr::map(
    indices,
    make_splits,
    data = data,
    class = "vfold_split"
  )
  tibble::tibble(
    splits = split_objs,
    id = names0(length(split_objs), "Fold")
  )
}

#' Group V-Fold Cross-Validation
#'
#' Group V-fold cross-validation creates splits of the data based
#'  on some grouping variable (which may have more than a single row
#'  associated with it). The function can create as many splits as
#'  there are unique values of the grouping variable or it can
#'  create a smaller set of splits where more than one group is left
#'  out at a time. A common use of this kind of resampling is when you have
#'  repeated measures of the same subject.
#'
#' @inheritParams vfold_cv
#' @param v The number of partitions of the data set. If left as `NULL` (the
#'   default), `v` will be set to the number of unique values in the grouping
#'   variable, creating "leave-one-group-out" splits.
#' @param balance If `v` is less than the number of unique groups, how should
#'  groups be combined into folds? Should be one of
#'  `"groups"`, which will assign roughly the same number of groups to each
#'  fold, or `"observations"`, which will assign roughly the same number of
#'  observations to each fold.
#' @inheritParams make_groups
#'
#' @export
#' @return A tibble with classes `group_vfold_cv`,
#'  `rset`, `tbl_df`, `tbl`, and `data.frame`.
#'  The results include a column for the data split objects and an
#'  identification variable.
#' @examplesIf rlang::is_installed("modeldata")
#' data(ames, package = "modeldata")
#'
#' set.seed(123)
#' group_vfold_cv(ames, group = Neighborhood, v = 5)
#' group_vfold_cv(
#'   ames,
#'   group = Neighborhood,
#'   v = 5,
#'   balance = "observations"
#' )
#' group_vfold_cv(ames, group = Neighborhood, v = 5, repeats = 2)
#'
#' # Leave-one-group-out CV
#' group_vfold_cv(ames, group = Neighborhood)
#'
#' library(dplyr)
#' data(Sacramento, package = "modeldata")
#'
#' city_strata <- Sacramento |>
#'   group_by(city) |>
#'   summarize(strata = mean(price)) |>
#'   reframe(city = city,
#'           strata = cut(strata, quantile(strata), include.lowest = TRUE))
#'
#' sacramento_data <- Sacramento |>
#'   full_join(city_strata, by = "city")
#'
#' group_vfold_cv(sacramento_data, city, strata = strata)
#'
#' @export
group_vfold_cv <- function(
  data,
  group = NULL,
  v = NULL,
  repeats = 1,
  balance = c("groups", "observations"),
  ...,
  strata = NULL,
  pool = 0.1
) {
  check_dots_empty()
  check_number_whole(repeats, min = 1)
  group <- validate_group({{ group }}, data)
  balance <- rlang::arg_match(balance)

  if (!missing(strata)) {
    strata <- check_grouped_strata({{ group }}, {{ strata }}, pool, data)
  }

  if (repeats == 1) {
    split_objs <- group_vfold_splits(
      data = data,
      group = group,
      v = v,
      balance = balance,
      strata = strata,
      pool = pool
    )
  } else {
    if (is.null(v)) {
      cli_abort(
        "Repeated resampling when {.arg v} is {.val NULL} would create identical resamples."
      )
    }
    if (v == length(unique(getElement(data, group)))) {
      cli_abort(
        "Repeated resampling when {.arg v} is {.val {v}} would create identical resamples."
      )
    }
    for (i in 1:repeats) {
      tmp <- group_vfold_splits(
        data = data,
        group = group,
        v = v,
        balance = balance,
        strata = strata,
        pool = pool
      )
      tmp$id2 <- tmp$id
      tmp$id <- names0(repeats, "Repeat")[i]
      split_objs <- if (i == 1) {
        tmp
      } else {
        rbind(split_objs, tmp)
      }
    }
  }

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  # Update `v` if not supplied directly
  if (is.null(v)) {
    v <- length(split_objs$splits)
  }

  ## Save some overall information

  cv_att <- list(
    v = v,
    group = group,
    balance = balance,
    repeats = 1,
    strata = strata,
    pool = pool
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs[, grepl("^id", names(split_objs))],
    attrib = cv_att,
    subclass = c("group_vfold_cv", "vfold_cv", "group_rset", "rset")
  )
}

group_vfold_splits <- function(
  data,
  group,
  v = NULL,
  balance,
  strata = NULL,
  pool = 0.1
) {
  group <- getElement(data, group)
  max_v <- length(unique(group))
  if (!is.null(strata)) {
    strata <- getElement(data, strata)
    strata <- as.character(strata)
    strata <- make_strata(strata, pool = pool)

    if (is.null(v)) {
      # Set max_v to be the lowest number of groups in a single strata
      # to ensure that all folds get each strata
      max_v <- min(
        vec_count(
          vec_unique(
            data.frame(group, strata)
          )$strata
        )$count
      )
      message <- c(
        "Leaving {.code v = NULL} while using stratification will set {.arg v} to the number of groups present in the least common stratum."
      )

      if (max_v < 5) {
        cli_abort(
          c(
            message,
            "*" = "The least common stratum only had {.val {max_v}} groups, which may not be enough for cross-validation.",
            "i" = "Set {.arg v} explicitly to override this error."
          ),
          call = rlang::caller_env()
        )
      }

      cli_warn(
        c(
          message,
          i = "Set {.arg v} explicitly to override this warning."
        ),
        call = rlang::caller_env()
      )
    }
  }

  if (is.null(v)) {
    v <- max_v
  }
  check_v(
    v = v,
    max_v = max_v,
    rows = "groups",
    prevent_loo = FALSE,
    call = rlang::caller_env()
  )

  indices <- make_groups(data, group, v, balance, strata)
  indices <- lapply(indices, default_complement, n = nrow(data))
  split_objs <-
    purrr::map(
      indices,
      make_splits,
      data = data,
      class = c("group_vfold_split", "vfold_split")
    )
  tibble::tibble(
    splits = split_objs,
    id = names0(length(split_objs), "Resample")
  )
}

add_vfolds <- function(x, v) {
  x$folds <- sample(rep(1:v, length.out = nrow(x)))
  x
}

check_v <- function(
  v,
  max_v,
  rows = "rows",
  prevent_loo = TRUE,
  call = rlang::caller_env()
) {
  check_number_whole(v, min = 2, call = call)

  if (v > max_v) {
    cli_abort(
      "The number of {rows} is less than {.arg v} = {.val {v}}.",
      call = call
    )
  }
  if (prevent_loo && isTRUE(v == max_v)) {
    cli_abort(
      c(
        "Leave-one-out cross-validation is not supported by this function.",
        "x" = "You set {.arg v} to {.code nrow(data)}, which would result in a leave-one-out cross-validation.",
        "i" = "Use {.fn loo_cv} in this case."
      ),
      call = call
    )
  }
}

check_grouped_strata <- function(
  group,
  strata,
  pool,
  data,
  call = caller_env()
) {
  strata <- tidyselect::vars_select(names(data), !!enquo(strata))

  # if strata was NULL this is empty, thus return NULL
  if (length(strata) < 1) {
    return(NULL)
  }

  grouped_table <- tibble(
    group = getElement(data, group),
    strata = getElement(data, strata)
  )

  if (
    nrow(vctrs::vec_unique(grouped_table)) !=
      nrow(vctrs::vec_unique(grouped_table["group"]))
  ) {
    cli_abort(
      "{.field strata} must be constant across all members of each {.field group}.",
      call = call
    )
  }

  strata
}

test_that("reshuffle_rset is working", {
  skip_if_not_installed("withr")
  # for `validation_split()` and variants
  withr::local_options(lifecycle_verbosity = "quiet")

  supported_subclasses <- rset_subclasses[
    setdiff(names(rset_subclasses), c("manual_rset"))
  ]

  # Reshuffling with the same seed, in the same order,
  # should recreate the same objects
  out <- withr::with_seed(
    123,
    lapply(
      supported_subclasses,
      function(x) suppressWarnings(reshuffle_rset(x))
    )
  )

  for (i in seq_along(supported_subclasses)) {
    expect_identical(
      out[[i]],
      supported_subclasses[[i]]
    )
  }

  # Check to make sure that stratification,
  # with non-default arguments,
  # is supported by reshuffled_resample

  # Select any non-grouped function in rset_subclasses with a strata argument:
  supports_strata <- purrr::map_lgl(
    names(supported_subclasses),
    \(.x) {
      any(names(formals(.x)) == "strata") && !any(names(formals(.x)) == "group")
    }
  )
  supports_strata <- names(supported_subclasses)[supports_strata]

  for (i in seq_along(supports_strata)) {
    # Fit those functions with non-default arguments:
    set.seed(123)
    resample <- do.call(
      supports_strata[i],
      list(
        data = test_data(),
        strata = "y",
        breaks = 2,
        pool = 0.2
      )
    )
    # Reshuffle them under the same seed to ensure they're identical
    set.seed(123)
    reshuffled_resample <- reshuffle_rset(resample)
    expect_identical(resample, reshuffled_resample)
  }

  # Select any grouped function in rset_subclasses with a strata argument:
  grouped_strata <- purrr::map_lgl(
    names(supported_subclasses),
    \(.x) {
      any(names(formals(.x)) == "strata") && any(names(formals(.x)) == "group")
    }
  )
  grouped_strata <- names(supported_subclasses)[grouped_strata]

  set.seed(11)

  group_table <- tibble::tibble(
    group = 1:100,
    outcome = sample(c(rep(0, 89), rep(1, 11)))
  )
  observation_table <- tibble::tibble(
    group = sample(1:100, 5e4, replace = TRUE),
    observation = 1:5e4
  )
  sample_data <- dplyr::full_join(
    group_table,
    observation_table,
    by = "group",
    multiple = "all"
  )

  for (i in seq_along(grouped_strata)) {
    # Fit those functions with non-default arguments:
    set.seed(123)
    resample <- suppressWarnings(
      do.call(
        grouped_strata[i],
        list(
          data = sample_data,
          group = "group",
          strata = "outcome",
          pool = 0.2
        )
      )
    )
    # Reshuffle them under the same seed to ensure they're identical
    set.seed(123)
    reshuffled_resample <- reshuffle_rset(resample)
    expect_identical(resample, reshuffled_resample)
  }

  for (i in seq_along(non_random_classes)) {
    expect_snapshot(
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    )
  }

  resample <- vfold_cv(mtcars, strata = cyl)
  attr(resample, "strata") <- TRUE

  expect_snapshot(error = TRUE, {
    reshuffle_rset(resample)
  })

  expect_snapshot(error = TRUE, {
    reshuffle_rset(rset_subclasses[["manual_rset"]])
  })

  expect_snapshot(error = TRUE, {
    reshuffle_rset(rset_subclasses[["manual_rset"]]$splits[[1]])
  })
})

context("Compatibility with dplyr")

library(rsample)
library(testthat)
library(dplyr)

###################################################################

obj_1 <- vfold_cv(mtcars)
obj_2 <- bootstraps(mtcars)
obj_3 <- rolling_origin(mtcars)
obj_4 <- nested_cv(mtcars, obj_1, inside = bootstraps(times = 5))
obj_5 <- mc_cv(mtcars)
obj_6 <- loo_cv(mtcars)
obj_7 <- group_vfold_cv(mtcars, group = "am")

###################################################################

check_att <- function(x, y)
  length(setdiff(names(attributes(x)), names(attributes(x)))) == 0

###################################################################

test_that('object types', {
  expect_true(is_rset(obj_1))
  expect_false(is_rset(obj_1[, -1]))
})

###################################################################

test_that('dplyr ops', {
  expect_true(
    is_rset(
      obj_2 %>% filter(id == "Bootstrap02")
    )
  )
  expect_true(
    is_rset(
      obj_3 %>% mutate(blah = substr(id, 1, 3))
    )
  )
  expect_true(
    is_rset(
      obj_4 %>% select(splits, id)
    )
  )
  expect_true(
    is_rset(
      obj_1 %>% arrange(id)
    )
  )
  expect_true(
    is_rset(
      obj_1 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah)
    )
  )
  expect_true(
    check_att(
      obj_1 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah),
      obj_1
    )
  )
  expect_true(
    is_rset(
      obj_2 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah)
    )
  )
  expect_true(
    check_att(
      obj_2 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah),
      obj_2
    )
  )
  expect_true(
    is_rset(
      obj_3 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah)
    )
  )
  expect_true(
    check_att(
      obj_3 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah),
      obj_3
    )
  )
  expect_true(
    is_rset(
      obj_4 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah)
    )
  )
  expect_true(
    check_att(
      obj_4 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah),
      obj_4
    )
  )
  expect_true(
    is_rset(
      obj_5 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah)
    )
  )
  expect_true(
    check_att(
      obj_5 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah),
      obj_5
    )
  )
  expect_true(
    is_rset(
      obj_6 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah)
    )
  )
  expect_true(
    check_att(
      obj_6 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah),
      obj_6
    )
  )
  expect_true(
    is_rset(
      obj_7 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah)
    )
  )
  expect_true(
    check_att(
      obj_7 %>% mutate(blah = substr(id, 1, 3)) %>% rename(newer = blah),
      obj_7
    )
  )
  expect_true(
    is_rset(
      obj_3 %>% slice(1L)
    )
  )
})

context("Gather from tidyr")

library(testthat)
library(rsample)
library(tidyr)

cvs <- vfold_cv(mtcars)
cvs$one <- 1
cvs$two <- 2

expt <- cvs %>% gather(model, statistic, -id)


test_that('basics', {
  res_1 <- gather(cvs)
  expect_equal(res_1, expt)
})


test_that('extra args ignored', {
  res_2 <- gather(cvs, contains("o"))
  expect_equal(res_2, expt)
  res_2 <- gather(cvs, key = ignored)
  expect_equal(res_2, expt)
  res_3 <- gather(cvs, ignored)
  expect_equal(res_2, expt)
})

test_that('no extra cols', {
  expect_error(gather(vfold_cv(mtcars)))
})

test_that('no splits', {
  cvs2 <- cvs
  cvs2$splits <- NULL
  res_4 <- gather(cvs2)
  expect_equal(res_4, expt)
})

#' Constructors for split objects
#' @export
make_splits <- function(x, ...) {
  UseMethod("make_splits")
}

#' @rdname make_splits
#' @param x A list of integers with names "analysis" and "assessment", or a
#' data frame of analysis or training data.
#' @export
make_splits.default <- function(x, ...) {
  rlang::abort("There is no method available to make an rsplit from `x`.")
}

#' @rdname make_splits
#' @param data A data frame.
#' @param class An optional class to give the object.
#' @param ... Further arguments passed to or from other methods (not currently
#' used).
#' @export
make_splits.list <- function(x, data, class = NULL, ...) {
  ellipsis::check_dots_empty()
  res <- rsplit(data, x$analysis, x$assessment)
  if (!is.null(class)) {
    res <- add_class(res, class)
  }
  res
}

#' @rdname make_splits
#' @param assessment A data frame of assessment or testing data, which can be empty.
#' @export
make_splits.data.frame <- function(x, assessment, ...) {
  ellipsis::check_dots_empty()
  if (nrow(x) == 0) {
    rlang::abort("The analysis set must contain at least one row.")
  }

  ind_analysis <- 1:nrow(x)
  if (nrow(assessment) == 0) {
    ind_assessment <- integer()
  } else {
    if (!identical(colnames(x), colnames(assessment))) {
      rlang::abort("The analysis and assessment sets must have the same columns.")
    }
    ind_assessment <- nrow(x) + 1:nrow(assessment)
  }

  data <- bind_rows(x, assessment)
  ind <- list(
    analysis = ind_analysis,
    assessment = ind_assessment
  )

  make_splits(ind, data)
}

merge_lists <- function(a, b) list(analysis = a, assessment = b)

dim_rset <- function(x, ...) {
  dims <- purrr::map(x$splits, dim)
  dims <- do.call("rbind", dims)
  dims <- tibble::as_tibble(dims)
  id_cols <- grep("(^id$)|(^id[1-9]$)", colnames(x), value = TRUE)
  for (i in seq_along(id_cols)) {
    dims[id_cols[i]] <- getElement(x, id_cols[i])
  }
  dims
}

names0 <- function(num, prefix = "x") {
  if (num == 0L) {
    return(character())
  }
  ind <- format(1:num)
  ind <- gsub(" ", "0", ind)
  paste0(prefix, ind)
}

add_class <- function(x, cls) {
  class(x) <- c(cls, class(x))
  x
}

strata_check <- function(strata, data) {
  if (!is.null(strata)) {
    if (!is.character(strata) | length(strata) != 1) {
      rlang::abort("`strata` should be a single name or character value.")
    }
    if (inherits(data[, strata], "Surv")) {
      rlang::abort("`strata` cannot be a `Surv` object. Use the time or event variable directly.")
    }
    if (!(strata %in% names(data))) {
      rlang::abort(strata, " is not in `data`.")
    }
  }
  invisible(NULL)
}

split_unnamed <- function(x, f) {
  out <- split(x, f)
  unname(out)
}

#' Obtain a identifier for the resamples
#'
#' This function returns a hash (or NA) for an attribute that is created when
#' the `rset` was initially constructed. This can be used to compare with other
#' resampling objects to see if they are the same.
#' @param x An `rset` or `tune_results` object.
#' @param ... Not currently used.
#' @return A character value or `NA_character_` if the object was created prior
#' to `rsample` version 0.1.0.
#' @rdname get_fingerprint
#' @aliases .get_fingerprint
#' @examples
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(2)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars, repeats = 2))
#' @export
.get_fingerprint <- function(x, ...) {
  UseMethod(".get_fingerprint")
}

#' @export
#' @rdname get_fingerprint
.get_fingerprint.default <- function(x, ...) {
  cls <- paste0("'", class(x), "'", collapse = ", ")
  rlang::abort(
    paste("No `.get_fingerprint()` method for this class(es)", cls)
  )
}

#' @export
#' @rdname get_fingerprint
.get_fingerprint.rset <- function(x, ...) {
  att <- attributes(x)
  if (any(names(att) == "fingerprint")) {
    res <- att$fingerprint
  } else {
    res <- NA_character_
  }
  res
}

test_that("simple numerics", {
  set.seed(13333)
  x1 <- rnorm(1000)
  str1a <- make_strata(x1)
  tab1a <- table(str1a)
  expect_equal(as.vector(tab1a), rep(250, 4))

  expect_warning(str1b <- make_strata(x1, depth = 500), "2 breaks instead")
  tab1b <- table(str1b)
  expect_equal(as.vector(tab1b), rep(500, 2))

  str1c <- make_strata(c(NA, x1[1:999]))
  tab1c <- table(str1c)
  expect_true(all(as.vector(tab1c) %in% 249:251))
})

test_that("simple character", {
  x2 <- factor(rep(LETTERS[1:12], each = 20))
  expect_warning(
    str2a <- make_strata(x2, pool = 0.05),
    "Stratifying groups that make up 5%"
  )
  expect_equal(table(str2a, dnn = ""), table(x2, dnn = ""))


  x2[5] <- NA
  expect_warning(
    str2b <- make_strata(x2, pool = 0.05),
    "Stratifying groups that make up 5%"
  )
  expect_true(all(as.vector(table(str2b, dnn = "")) %in% 19:21))
})

test_that("bad data", {
  x3 <- factor(rep(LETTERS[1:15], each = 50))
  expect_warning(make_strata(x3), "Too little data")
  expect_snapshot(s1 <- make_strata(x3, pool = 0.06))
  expect_snapshot(s2 <- make_strata(mtcars$mpg))
  expect_snapshot(s3 <- make_strata(seq_len(50), breaks = -1))
})



# strata_check() ----------------------------------------------------------

test_that("don't stratify on Surv objects", {
  df <- data.frame(
    time = c(85, 79, 70, 6, 32, 8, 17, 93, 81, 76),
    event = c(0, 0, 1, 0, 0, 0, 1, 1, 1, 1)
  )
  df$surv <- structure(
    c(
      85, 79, 70, 6, 32, 8, 17, 93, 81, 76,
      0, 0, 1, 0, 0, 0, 1, 1, 1, 1
    ),
    .Dim = c(10L, 2L),
    .Dimnames = list(NULL, c("time", "status")),
    type = "right",
    class = "Surv"
  )

  expect_error(strata_check("surv", df))
})

#' Time-based Resampling
#'
#' @description
#' These resampling functions are focused on various forms of _time series_
#' resampling.
#'
#' - `sliding_window()` uses the row number when computing the resampling
#'   indices. It is independent of any time index, but is useful with
#'   completely regular series.
#'
#' - `sliding_index()` computes resampling indices relative to the `index`
#'   column. This is often a Date or POSIXct column, but doesn't have to be.
#'   This is useful when resampling irregular series, or for using irregular
#'   lookback periods such as `lookback = lubridate::years(1)` with daily
#'   data (where the number of days in a year may vary).
#'
#' - `sliding_period()` first breaks up the `index` into less granular groups
#'   based on `period`, and then uses that to construct the resampling indices.
#'   This is extremely useful for constructing rolling monthly or yearly
#'   windows from daily data.
#'
#' @inheritParams ellipsis::dots_empty
#'
#' @param data A data frame.
#'
#' @param index The index to compute resampling indices relative to, specified
#'   as a bare column name. This must be an existing column in `data`.
#'
#'   - For `sliding_index()`, this is commonly a date vector, but is not
#'     required.
#'
#'   - For `sliding_period()`, it is required that this is a Date or POSIXct
#'     vector.
#'
#'   The `index` must be an _increasing_ vector, but duplicate values are
#'   allowed. Additionally, the index cannot contain any missing values.
#'
#' @param period The period to group the `index` by. This is specified as a
#'   single string, such as `"year"` or `"month"`. See the `.period` argument
#'   of [slider::slide_index()] for the full list of options and further
#'   explanation.
#'
#' @param lookback The number of elements to look back from the current element
#'   when computing the resampling indices of the analysis set. The current
#'   row is always included in the analysis set.
#'
#'   - For `sliding_window()`, a single integer defining the number of rows to
#'     look back from the current row.
#'
#'   - For `sliding_index()`, a single object that will be subtracted from the
#'     `index` as `index - lookback` to define the boundary of where to start
#'     searching for rows to include in the current resample. This is often
#'     an integer value corresponding to the number of days to look back,
#'     or a lubridate Period object.
#'
#'   - For `sliding_period()`, a single integer defining the number of groups
#'     to look back from the current group, where the groups were defined from
#'     breaking up the `index` according to the `period`.
#'
#'   In all cases, `Inf` is also allowed to force an expanding window.
#'
#' @param assess_start,assess_stop This combination of arguments determines
#'   how far into the future to look when constructing the assessment set.
#'   Together they construct a range of
#'   `[index + assess_start, index + assess_stop]` to search for rows to include
#'   in the assessment set.
#'
#'   Generally, `assess_start` will always be `1` to indicate that the first
#'   value to potentially include in the assessment set should start one element
#'   after the current row, but it can be increased to a larger value to
#'   create "gaps" between the analysis and assessment set if you are worried
#'   about high levels of correlation in short term forecasting.
#'
#'   - For `sliding_window()`, these are both single integers defining the
#'     number of rows to look forward from the current row.
#'
#'   - For `sliding_index()`, these are single objects that will be added
#'     to the `index` to compute the range to search for rows to include
#'     in the assessment set. This is often an integer value corresponding to
#'     the number of days to look forward, or a lubridate Period object.
#'
#'   - For `sliding_period()`, these are both single integers defining the
#'     number of groups to look forward from the current group, where the groups
#'     were defined from breaking up the `index` according to the `period`.
#'
#' @param complete A single logical. When using `lookback` to compute the
#'   analysis sets, should only complete windows be considered? If set to
#'   `FALSE`, partial windows will be used until it is possible to create
#'   a complete window (based on `lookback`). This is a way to use an
#'   expanding window up to a certain point, and then switch to a sliding
#'   window.
#'
#' @param step A single positive integer. After computing the resampling
#'   indices, `step` is used to thin out the results by selecting every
#'   `step`-th result by subsetting the indices with
#'   `seq(1L, n_indices, by = step)`. `step` is applied after `skip`.
#'   Note that `step` is independent of any time `index` used.
#'
#' @param skip A single positive integer, or zero. After computing the
#'   resampling indices, the first `skip` results will be dropped by subsetting
#'   the indices with `seq(skip + 1L, n_indices)`. This can be especially
#'   useful when combined with `lookback = Inf`, which creates an expanding
#'   window starting from the first row. By skipping forward, you can drop
#'   the first few windows that have very few data points. `skip` is
#'   applied before `step`. Note that `skip` is independent of any time
#'   `index` used.
#'
#' @param every A single positive integer. The number of periods to group
#'   together.
#'
#'   For example, if the `period` was set to `"year"` with an `every`
#'   value of 2, then the years 1970 and 1971 would be placed in the same
#'   group.
#'
#' @param origin The reference date time value. The default when left
#'   as `NULL` is the epoch time of `1970-01-01 00:00:00`,
#'   _in the time zone of the index_.
#'
#'   This is generally used to define the anchor time to count from,
#'   which is relevant when the `every` value is `> 1`.
#'
#' @seealso
#' [rolling_origin()]
#'
#' [slider::slide()], [slider::slide_index()], and [slider::slide_period()],
#' which power these resamplers.
#'
#' @name slide-resampling
#'
#' @examples
#' library(vctrs)
#' library(tibble)
#' library(modeldata)
#' data("Chicago")
#'
#' index <- new_date(c(1, 3, 4, 7, 8, 9, 13, 15, 16, 17))
#' df <- tibble(x = 1:10, index = index)
#' df
#'
#' # Look back two rows beyond the current row, for a total of three rows
#' # in each analysis set. Each assessment set is composed of the two rows after
#' # the current row.
#' sliding_window(df, lookback = 2, assess_stop = 2)
#'
#' # Same as before, but step forward by 3 rows between each resampling slice,
#' # rather than just by 1.
#' rset <- sliding_window(df, lookback = 2, assess_stop = 2, step = 3)
#' rset
#'
#' analysis(rset$splits[[1]])
#' analysis(rset$splits[[2]])
#'
#' # Now slide relative to the `index` column in `df`. This time we look back
#' # 2 days from the current row's `index` value, and 2 days forward from
#' # it to construct the assessment set. Note that this series is irregular,
#' # so it produces different results than `sliding_window()`. Additionally,
#' # note that it is entirely possible for the assessment set to contain no
#' # data if you have a highly irregular series and "look forward" into a
#' # date range where no data points actually exist!
#' sliding_index(df, index, lookback = 2, assess_stop = 2)
#'
#' # With `sliding_period()`, we can break up our date index into less granular
#' # chunks, and slide over them instead of the index directly. Here we'll use
#' # the Chicago data, which contains daily data spanning 16 years, and we'll
#' # break it up into rolling yearly chunks. Three years worth of data will
#' # be used for the analysis set, and one years worth of data will be held out
#' # for performance assessment.
#' sliding_period(
#'   Chicago,
#'   date,
#'   "year",
#'   lookback = 2,
#'   assess_stop = 1
#' )
#'
#' # Because `lookback = 2`, three years are required to form a "complete"
#' # window of data. To allow partial windows, set `complete = FALSE`.
#' # Here that first constructs two expanding windows until a complete three
#' # year window can be formed, at which point we switch to a sliding window.
#' sliding_period(
#'   Chicago,
#'   date,
#'   "year",
#'   lookback = 2,
#'   assess_stop = 1,
#'   complete = FALSE
#' )
#'
#' # Alternatively, you could break the resamples up by month. Here we'll
#' # use an expanding monthly window by setting `lookback = Inf`, and each
#' # assessment set will contain two months of data. To ensure that we have
#' # enough data to fit our models, we'll `skip` the first 4 expanding windows.
#' # Finally, to thin out the results, we'll `step` forward by 2 between
#' # each resample.
#' sliding_period(
#'   Chicago,
#'   date,
#'   "month",
#'   lookback = Inf,
#'   assess_stop = 2,
#'   skip = 4,
#'   step = 2
#' )
NULL

#' @export
#' @rdname slide-resampling
sliding_window <- function(data,
                           ...,
                           lookback = 0L,
                           assess_start = 1L,
                           assess_stop = 1L,
                           complete = TRUE,
                           step = 1L,
                           skip = 0L) {
  ellipsis::check_dots_empty()

  if (!is.data.frame(data)) {
    rlang::abort("`data` must be a data frame.")
  }

  lookback <- check_lookback(lookback)
  assess_start <- check_assess(assess_start, "assess_start")
  assess_stop <- check_assess(assess_stop, "assess_stop")
  step <- check_step(step)
  skip <- check_skip(skip)

  if (assess_start > assess_stop) {
    rlang::abort("`assess_start` must be less than or equal to `assess_stop`.")
  }

  seq <- vctrs::vec_seq_along(data)

  id_in <- slider::slide(
    .x = seq,
    .f = identity,
    .before = lookback,
    .after = 0L,
    .step = 1L,
    .complete = complete
  )

  id_out <- slider::slide(
    .x = seq,
    .f = identity,
    .before = -assess_start,
    .after = assess_stop,
    .step = 1L,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (!identical(skip, 0L)) {
    indices <- slice_skip(indices, skip)
  }

  if (!identical(step, 1L)) {
    indices <- slice_step(indices, step)
  }

  splits <- purrr::map(
    indices,
    ~ make_splits(.x, data = data, class = "sliding_window_split")
  )

  ids <- names0(length(indices), prefix = "Slice")

  attrib <- list(
    lookback = lookback,
    assess_start = assess_start,
    assess_stop = assess_stop,
    complete = complete,
    step = step,
    skip = skip
  )

  new_rset(
    splits = splits,
    ids = ids,
    attrib = attrib,
    subclass = c("sliding_window", "rset")
  )
}

# ------------------------------------------------------------------------------

#' @export
#' @rdname slide-resampling
sliding_index <- function(data,
                          index,
                          ...,
                          lookback = 0L,
                          assess_start = 1L,
                          assess_stop = 1L,
                          complete = TRUE,
                          step = 1L,
                          skip = 0L) {
  ellipsis::check_dots_empty()

  if (!is.data.frame(data)) {
    rlang::abort("`data` must be a data frame.")
  }

  step <- check_step(step)
  skip <- check_skip(skip)

  index <- rlang::enexpr(index)
  loc <- tidyselect::eval_select(index, data)

  if (length(loc) != 1L) {
    rlang::abort("`index` must specify exactly one column in `data`.")
  }

  index <- data[[loc]]

  seq <- vctrs::vec_seq_along(data)

  id_in <- slider::slide_index(
    .x = seq,
    .i = index,
    .f = identity,
    .before = lookback,
    .after = 0L,
    .complete = complete
  )

  id_out <- slider::slide_index(
    .x = seq,
    .i = index,
    .f = identity,
    .before = -assess_start,
    .after = assess_stop,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (!identical(skip, 0L)) {
    indices <- slice_skip(indices, skip)
  }

  if (!identical(step, 1L)) {
    indices <- slice_step(indices, step)
  }

  splits <- purrr::map(
    indices,
    ~ make_splits(.x, data = data, class = "sliding_index_split")
  )

  ids <- names0(length(indices), prefix = "Slice")

  attrib <- list(
    lookback = lookback,
    assess_start = assess_start,
    assess_stop = assess_stop,
    complete = complete,
    step = step,
    skip = skip
  )

  new_rset(
    splits = splits,
    ids = ids,
    attrib = attrib,
    subclass = c("sliding_index", "rset")
  )
}

# ------------------------------------------------------------------------------

#' @export
#' @rdname slide-resampling
sliding_period <- function(data,
                           index,
                           period,
                           ...,
                           lookback = 0L,
                           assess_start = 1L,
                           assess_stop = 1L,
                           complete = TRUE,
                           step = 1L,
                           skip = 0L,
                           every = 1L,
                           origin = NULL) {
  ellipsis::check_dots_empty()

  if (!is.data.frame(data)) {
    rlang::abort("`data` must be a data frame.")
  }

  lookback <- check_lookback(lookback)
  assess_start <- check_assess(assess_start, "assess_start")
  assess_stop <- check_assess(assess_stop, "assess_stop")
  step <- check_step(step)

  if (assess_start > assess_stop) {
    rlang::abort("`assess_start` must be less than or equal to `assess_stop`.")
  }

  index <- rlang::enexpr(index)
  loc <- tidyselect::eval_select(index, data)

  if (length(loc) != 1L) {
    rlang::abort("`index` must specify exactly one column in `data`.")
  }

  index <- data[[loc]]

  seq <- vctrs::vec_seq_along(data)

  id_in <- slider::slide_period(
    .x = seq,
    .i = index,
    .period = period,
    .f = identity,
    .every = every,
    .origin = origin,
    .before = lookback,
    .after = 0L,
    .complete = complete
  )

  id_out <- slider::slide_period(
    .x = seq,
    .i = index,
    .period = period,
    .f = identity,
    .every = every,
    .origin = origin,
    .before = -assess_start,
    .after = assess_stop,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (!identical(skip, 0L)) {
    indices <- slice_skip(indices, skip)
  }

  if (!identical(step, 1L)) {
    indices <- slice_step(indices, step)
  }

  splits <- purrr::map(
    indices,
    ~ make_splits(.x, data = data, class = "sliding_period_split")
  )

  ids <- names0(length(indices), prefix = "Slice")

  attrib <- list(
    period = period,
    lookback = lookback,
    assess_start = assess_start,
    assess_stop = assess_stop,
    complete = complete,
    step = step,
    skip = skip,
    every = every,
    origin = origin
  )

  new_rset(
    splits = splits,
    ids = ids,
    attrib = attrib,
    subclass = c("sliding_period", "rset")
  )
}

# ------------------------------------------------------------------------------

check_lookback <- function(x) {
  if (vctrs::vec_size(x) != 1L) {
    rlang::abort(paste0("`lookback` must have size 1."))
  }

  if (identical(x, Inf)) {
    return(x)
  }

  if (!rlang::is_integerish(x, finite = TRUE)) {
    rlang::abort(paste0("`lookback` must be an integer of size 1, or `Inf`."))
  }

  if (x < 0L) {
    rlang::abort(paste0("`lookback` must be positive, or zero."))
  }

  vctrs::vec_cast(x, integer(), x_arg = "lookback")
}

check_assess <- function(x, arg) {
  if (vctrs::vec_size(x) != 1L) {
    rlang::abort(paste0("`", arg, "` must have size 1."))
  }

  if (identical(x, Inf)) {
    return(x)
  }

  if (!rlang::is_integerish(x, finite = TRUE)) {
    rlang::abort(paste0("`", arg, "` must be an integer of size 1, or `Inf`."))
  }

  if (x <= 0L) {
    rlang::abort(paste0("`", arg, "` must be positive."))
  }

  vctrs::vec_cast(x, integer(), x_arg = arg)
}

check_step <- function(x) {
  if (vctrs::vec_size(x) != 1L) {
    rlang::abort(paste0("`step` must have size 1."))
  }

  if (!rlang::is_integerish(x, finite = TRUE)) {
    rlang::abort(paste0("`step` must be an integer of size 1."))
  }

  if (x <= 0L) {
    rlang::abort(paste0("`step` must be positive."))
  }

  vctrs::vec_cast(x, integer(), x_arg = "step")
}

check_skip <- function(x) {
  if (vctrs::vec_size(x) != 1L) {
    rlang::abort(paste0("`skip` must have size 1."))
  }

  if (!rlang::is_integerish(x, finite = TRUE)) {
    rlang::abort(paste0("`skip` must be an integer of size 1."))
  }

  if (x < 0L) {
    rlang::abort(paste0("`skip` must be positive, or zero."))
  }

  vctrs::vec_cast(x, integer(), x_arg = "skip")
}

compute_complete_indices <- function(id_in, id_out) {
  # Remove where either list has a `NULL` element.
  # These are incomplete windows.
  id_in_na <- vctrs::vec_equal_na(id_in)
  id_out_na <- vctrs::vec_equal_na(id_out)

  id_either_na <- id_in_na | id_out_na

  id_in <- vctrs::vec_slice(id_in, !id_either_na)
  id_out <- vctrs::vec_slice(id_out, !id_either_na)

  purrr::map2(id_in, id_out, merge_lists)
}

slice_skip <- function(indices, skip) {
  n_indices <- length(indices)
  slicer <- rlang::seq2(skip + 1L, n_indices)
  vctrs::vec_slice(indices, slicer)
}

slice_step <- function(indices, step) {
  n_indices <- length(indices)
  slicer <- seq2_by(1L, n_indices, by = step)
  indices <- vctrs::vec_slice(indices, slicer)
}

seq2_by <- function(from, to, by) {
  if (length(from) != 1) {
    rlang::abort("`from` must be length one")
  }
  if (length(to) != 1) {
    rlang::abort("`to` must be length one")
  }

  by <- as.integer(by)

  if (length(by) != 1) {
    rlang::abort("`by` must be length one")
  }
  if (by <= 0L) {
    rlang::abort("`by` must be positive")
  }

  if (from > to) {
    integer()
  } else {
    seq.int(from, to, by = by)
  }
}

# ------------------------------------------------------------------------------
# `sliding_window()`

test_that("defaults work", {
  df <- data.frame(x = 1:3)
  x <- sliding_window(df)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 2L)
  expect_identical(split2[["out_id"]], 3L)
})

test_that("lookback always uses complete windows", {
  df <- data.frame(x = 1:4)
  x <- sliding_window(df, lookback = 1)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:2)
  expect_identical(split1[["out_id"]], 3L)

  expect_identical(split2[["in_id"]], 2:3)
  expect_identical(split2[["out_id"]], 4L)
})

test_that("can step forward between slices", {
  df <- data.frame(x = 1:6)
  x <- sliding_window(df, lookback = 1, step = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:2)
  expect_identical(split1[["out_id"]], 3L)

  expect_identical(split2[["in_id"]], 3:4)
  expect_identical(split2[["out_id"]], 5L)
})

test_that("can generate assessment slices", {
  df <- data.frame(x = 1:4)
  x <- sliding_window(df, assess_stop = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2:3)

  expect_identical(split2[["in_id"]], 2L)
  expect_identical(split2[["out_id"]], 3:4)

  expect_identical(nrow(x), 2L)
})

test_that("can add analysis / assessment gaps", {
  df <- data.frame(x = 1:7)
  x <- sliding_window(df, lookback = 1, assess_start = 3, assess_stop = 4)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:2)
  expect_identical(split1[["out_id"]], 5:6)

  expect_identical(split2[["in_id"]], 2:3)
  expect_identical(split2[["out_id"]], 6:7)

  expect_identical(nrow(x), 2L)
})

test_that("can create an expanding window", {
  df <- data.frame(x = 1:4)
  x <- sliding_window(df, lookback = Inf)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]
  split3 <- x[["splits"]][[3]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 1:2)
  expect_identical(split2[["out_id"]], 3L)

  expect_identical(split3[["in_id"]], 1:3)
  expect_identical(split3[["out_id"]], 4L)

  expect_identical(nrow(x), 3L)
})

test_that("can skip first few resampling slices", {
  df <- data.frame(x = 1:8)
  x <- sliding_window(df, lookback = 1, skip = 3)

  split1 <- x[["splits"]][[1]]

  expect_identical(split1[["in_id"]], 4:5)
  expect_identical(split1[["out_id"]], 6L)

  expect_identical(nrow(x), 3L)
})

test_that("`skip` is applied before `step`", {
  df <- data.frame(x = 1:8)
  x <- sliding_window(df, lookback = 1, skip = 3, step = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 4:5)
  expect_identical(split1[["out_id"]], 6L)

  expect_identical(split2[["in_id"]], 6:7)
  expect_identical(split2[["out_id"]], 8L)

  expect_identical(nrow(x), 2L)
})

test_that("can use incomplete windows at the beginning", {
  df <- data.frame(x = 1:5)
  x <- sliding_window(df, lookback = 2, complete = FALSE)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]
  split3 <- x[["splits"]][[3]]
  split4 <- x[["splits"]][[4]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 1:2)
  expect_identical(split2[["out_id"]], 3L)

  expect_identical(split3[["in_id"]], 1:3)
  expect_identical(split3[["out_id"]], 4L)

  expect_identical(split4[["in_id"]], 2:4)
  expect_identical(split4[["out_id"]], 5L)

  expect_identical(nrow(x), 4L)
})

test_that("`data` is validated", {
  expect_snapshot(error = TRUE, {
    sliding_window(1)
  })
})

test_that("`lookback` is validated", {
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), lookback = -1)
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), lookback = "a")
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), lookback = c(1, 2))
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), lookback = NA)
  })
})

test_that("`assess_start` is validated", {
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_start = -1)
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_start = "a")
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_start = c(1, 2))
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_start = NA)
  })
})

test_that("`assess_stop` is validated", {
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_stop = -1)
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_stop = "a")
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_stop = c(1, 2))
  })
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_stop = NA)
  })
})

test_that("`assess_start` must be before or equal to `assess_stop`", {
  expect_snapshot(error = TRUE, {
    sliding_window(data.frame(), assess_start = 2, assess_stop = 1)
  })
})

# ------------------------------------------------------------------------------
# `sliding_index()`

test_that("defaults works", {
  df <- data.frame(x = 1:3)
  x <- sliding_index(df, x)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1L)
  expect_identical(split1$out_id, 2L)

  expect_identical(split2$in_id, 2L)
  expect_identical(split2$out_id, 3L)

  expect_identical(nrow(x), 2L)
})

test_that("can lookback over irregular index", {
  df <- data.frame(x = c(1, 3, 4, 5))
  x <- sliding_index(df, x, lookback = 1)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 2L)
  expect_identical(split1$out_id, 3L)

  expect_identical(split2$in_id, 2:3)
  expect_identical(split2$out_id, 4L)

  expect_identical(nrow(x), 2L)
})

test_that("can compute assessment indices relative to irregular index", {
  df <- data.frame(x = c(1, 3, 4, 5, 7, 8))
  x <- sliding_index(df, x, lookback = 1, assess_stop = 2)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]
  split3 <- x$splits[[3]]

  expect_identical(split1$in_id, 2L)
  expect_identical(split1$out_id, 3:4)

  expect_identical(split2$in_id, 2:3)
  expect_identical(split2$out_id, 4L)

  expect_identical(split3$in_id, 3:4)
  expect_identical(split3$out_id, 5L)

  expect_identical(nrow(x), 3L)
})

test_that("it is possible to create empty assessment sets", {
  # Look forward 1->2 values from `5`, so creates a window with range of [6, 7].
  # But no `x` values fall in this range. However, in theory it is "possible"
  # to make a complete window starting at `5`, which is why `complete = TRUE`
  # didn't remove it.
  df <- data.frame(x = c(1, 3, 4, 5, 8, 9))
  x <- sliding_index(df, x, lookback = 1, assess_stop = 2)

  split3 <- x$splits[[3]]

  expect_identical(split3$in_id, 3:4)
  expect_identical(split3$out_id, integer())

  expect_identical(nrow(x), 3L)
})

test_that("can add a gap between the analysis and assessment set", {
  df <- data.frame(x = c(1, 3, 4, 5, 7, 8))
  x <- sliding_index(df, x, lookback = 2, assess_start = 2, assess_stop = 3)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]
  split3 <- x$splits[[3]]

  expect_identical(split1$in_id, 1:2)
  expect_identical(split1$out_id, 4L)

  expect_identical(split2$in_id, 2:3)
  expect_identical(split2$out_id, 5L)

  expect_identical(split3$in_id, 2:4)
  expect_identical(split3$out_id, 5:6)

  expect_identical(nrow(x), 3L)
})

test_that("can use `step` to thin results after calling `slide_index()`", {
  df <- data.frame(x = c(1, 3, 4, 6, 7, 10))
  x <- sliding_index(df, x, lookback = 2, assess_stop = 2, step = 2)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1:2)
  expect_identical(split1$out_id, 3L)

  expect_identical(split2$in_id, 3:4)
  expect_identical(split2$out_id, 5L)

  expect_identical(nrow(x), 2L)
})

test_that("can skip first few resampling slices", {
  df <- data.frame(x = c(1, 3, 4, 6, 7, 10))
  x <- sliding_index(df, x, lookback = 1, skip = 2)

  split1 <- x[["splits"]][[1]]

  expect_identical(split1[["in_id"]], 4L)
  expect_identical(split1[["out_id"]], 5L)

  expect_identical(nrow(x), 2L)
})

test_that("`skip` is applied before `step`", {
  df <- data.frame(x = c(1, 3, 4, 6, 7, 9, 11, 13, 14))
  x <- sliding_index(df, x, lookback = 1, skip = 3, step = 2, assess_stop = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 4:5)
  expect_identical(split1[["out_id"]], 6L)

  expect_identical(split2[["in_id"]], 7L)
  expect_identical(split2[["out_id"]], 8L)

  expect_identical(nrow(x), 2L)
})

test_that("can use incomplete windows at the beginning", {
  df <- data.frame(x = c(1, 3, 4, 5, 7))
  x <- sliding_index(df, x, lookback = 2, complete = FALSE, assess_stop = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]
  split3 <- x[["splits"]][[3]]
  split4 <- x[["splits"]][[4]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 1:2)
  expect_identical(split2[["out_id"]], 3:4)

  expect_identical(split3[["in_id"]], 2:3)
  expect_identical(split3[["out_id"]], 4L)

  expect_identical(split4[["in_id"]], 2:4)
  expect_identical(split4[["out_id"]], 5L)

  expect_identical(nrow(x), 4L)
})

test_that("`data` is validated", {
  expect_snapshot(error = TRUE, {
    sliding_index(1)
  })
})

test_that("`index` is validated", {
  df <- data.frame(x = 1:2)
  expect_snapshot(error = TRUE, {
    sliding_index(df, y)
  })
})

# ------------------------------------------------------------------------------
# `sliding_period()`

test_that("can group by month", {
  index <- vctrs::new_date(c(-1, 0, 1, 31))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, period = "month")

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1L)
  expect_identical(split1$out_id, 2:3)

  expect_identical(split2$in_id, 2:3)
  expect_identical(split2$out_id, 4L)

  expect_identical(nrow(x), 2L)
})

test_that("can group by year", {
  index <- vctrs::new_date(c(-1, 0, 1, 31))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, period = "year")

  split1 <- x$splits[[1]]

  expect_identical(split1$in_id, 1L)
  expect_identical(split1$out_id, 2:4)

  expect_identical(nrow(x), 1L)
})

test_that("when looking back over multiple periods, only complete ones are used", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, period = "month", lookback = 1)

  split1 <- x$splits[[1]]

  expect_identical(split1$in_id, 1:2)
  expect_identical(split1$out_id, 3:4)

  expect_identical(nrow(x), 2L)
})

test_that("can look forward to assess over multiple periods", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, period = "month", assess_stop = 2)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1L)
  expect_identical(split1$out_id, 2:4)

  expect_identical(split2$in_id, 2L)
  expect_identical(split2$out_id, 3:5)

  expect_identical(nrow(x), 2L)
})

test_that("can use `step` to thin results after calling `slide_period()`", {
  df <- data.frame(x = vctrs::new_date(c(1, 3, 4, 6, 7, 10)))
  x <- sliding_period(df, x, "day", lookback = 2, assess_stop = 2, step = 2)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1:2)
  expect_identical(split1$out_id, 3L)

  expect_identical(split2$in_id, 3:4)
  expect_identical(split2$out_id, 5L)

  expect_identical(nrow(x), 2L)
})

test_that("can skip first few resampling slices", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31, 59))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, "month", lookback = 1, skip = 2)

  split1 <- x[["splits"]][[1]]

  expect_identical(split1[["in_id"]], 3:5)
  expect_identical(split1[["out_id"]], 6L)

  expect_identical(nrow(x), 1L)
})

test_that("can skip with expanding window", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31, 59))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, "month", lookback = Inf, skip = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:4)
  expect_identical(split1[["out_id"]], 5L)

  expect_identical(split2[["in_id"]], 1:5)
  expect_identical(split2[["out_id"]], 6L)

  expect_identical(nrow(x), 2L)
})

test_that("`skip` is applied before `step`", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31, 59, 90))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, "month", lookback = Inf, skip = 2, step = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:4)
  expect_identical(split1[["out_id"]], 5L)

  expect_identical(split2[["in_id"]], 1:6)
  expect_identical(split2[["out_id"]], 7L)

  expect_identical(nrow(x), 2L)
})

test_that("can use incomplete windows at the beginning", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 59, 90))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, "month", lookback = 2, complete = FALSE)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]
  split3 <- x[["splits"]][[3]]
  split4 <- x[["splits"]][[4]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 1:2)
  expect_identical(split2[["out_id"]], 3:4)

  expect_identical(split3[["in_id"]], 1:4)
  expect_identical(split3[["out_id"]], integer())

  expect_identical(split4[["in_id"]], 3:5)
  expect_identical(split4[["out_id"]], 6L)

  expect_identical(nrow(x), 4L)
})

test_that("`data` is validated", {
  expect_snapshot(error = TRUE, {
    sliding_period(1)
  })
})

test_that("`index` is validated", {
  df <- data.frame(x = 1:2)
  expect_snapshot(error = TRUE, {
    sliding_period(df, y)
  })
})

# ------------------------------------------------------------------------------
# `sliding_window()`

test_that("defaults work", {
  df <- data.frame(x = 1:3)
  x <- sliding_window(df)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 2L)
  expect_identical(split2[["out_id"]], 3L)
})

test_that("lookback always uses complete windows", {
  df <- data.frame(x = 1:4)
  x <- sliding_window(df, lookback = 1)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:2)
  expect_identical(split1[["out_id"]], 3L)

  expect_identical(split2[["in_id"]], 2:3)
  expect_identical(split2[["out_id"]], 4L)
})

test_that("can step forward between slices", {
  df <- data.frame(x = 1:6)
  x <- sliding_window(df, lookback = 1, step = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:2)
  expect_identical(split1[["out_id"]], 3L)

  expect_identical(split2[["in_id"]], 3:4)
  expect_identical(split2[["out_id"]], 5L)
})

test_that("can generate assessment slices", {
  df <- data.frame(x = 1:4)
  x <- sliding_window(df, assess_stop = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2:3)

  expect_identical(split2[["in_id"]], 2L)
  expect_identical(split2[["out_id"]], 3:4)

  expect_identical(nrow(x), 2L)
})

test_that("can add analysis / assessment gaps", {
  df <- data.frame(x = 1:7)
  x <- sliding_window(df, lookback = 1, assess_start = 3, assess_stop = 4)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:2)
  expect_identical(split1[["out_id"]], 5:6)

  expect_identical(split2[["in_id"]], 2:3)
  expect_identical(split2[["out_id"]], 6:7)

  expect_identical(nrow(x), 2L)
})

test_that("can create an expanding window", {
  df <- data.frame(x = 1:4)
  x <- sliding_window(df, lookback = Inf)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]
  split3 <- x[["splits"]][[3]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 1:2)
  expect_identical(split2[["out_id"]], 3L)

  expect_identical(split3[["in_id"]], 1:3)
  expect_identical(split3[["out_id"]], 4L)

  expect_identical(nrow(x), 3L)
})

test_that("can skip first few resampling slices", {
  df <- data.frame(x = 1:8)
  x <- sliding_window(df, lookback = 1, skip = 3)

  split1 <- x[["splits"]][[1]]

  expect_identical(split1[["in_id"]], 4:5)
  expect_identical(split1[["out_id"]], 6L)

  expect_identical(nrow(x), 3L)
})

test_that("`skip` is applied before `step`", {
  df <- data.frame(x = 1:8)
  x <- sliding_window(df, lookback = 1, skip = 3, step = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 4:5)
  expect_identical(split1[["out_id"]], 6L)

  expect_identical(split2[["in_id"]], 6:7)
  expect_identical(split2[["out_id"]], 8L)

  expect_identical(nrow(x), 2L)
})

test_that("can use incomplete windows at the beginning", {
  df <- data.frame(x = 1:5)
  x <- sliding_window(df, lookback = 2, complete = FALSE)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]
  split3 <- x[["splits"]][[3]]
  split4 <- x[["splits"]][[4]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 1:2)
  expect_identical(split2[["out_id"]], 3L)

  expect_identical(split3[["in_id"]], 1:3)
  expect_identical(split3[["out_id"]], 4L)

  expect_identical(split4[["in_id"]], 2:4)
  expect_identical(split4[["out_id"]], 5L)

  expect_identical(nrow(x), 4L)
})

test_that("`data` is validated", {
  expect_error(sliding_window(1), "`data` must be a data frame")
})

test_that("`lookback` is validated", {
  expect_error(sliding_window(data.frame(), lookback = -1), "`lookback` must be positive, or zero")
  expect_error(sliding_window(data.frame(), lookback = "a"), "`lookback` must be an integer")
  expect_error(sliding_window(data.frame(), lookback = c(1, 2)), "`lookback` must have size 1")
  expect_error(sliding_window(data.frame(), lookback = NA), "`lookback` must be an integer")
})

test_that("`assess_start` is validated", {
  expect_error(sliding_window(data.frame(), assess_start = -1), "`assess_start` must be positive")
  expect_error(sliding_window(data.frame(), assess_start = "a"), "`assess_start` must be an integer")
  expect_error(sliding_window(data.frame(), assess_start = c(1, 2)), "`assess_start` must have size 1")
  expect_error(sliding_window(data.frame(), assess_start = NA), "`assess_start` must be an integer")
})

test_that("`assess_stop` is validated", {
  expect_error(sliding_window(data.frame(), assess_stop = -1), "`assess_stop` must be positive")
  expect_error(sliding_window(data.frame(), assess_stop = "a"), "`assess_stop` must be an integer")
  expect_error(sliding_window(data.frame(), assess_stop = c(1, 2)), "`assess_stop` must have size 1")
  expect_error(sliding_window(data.frame(), assess_stop = NA), "`assess_stop` must be an integer")
})

test_that("`assess_start` must be before or equal to `assess_stop`", {
  expect_error(sliding_window(data.frame(), assess_start = 2, assess_stop = 1), "less than or equal to")
})

# ------------------------------------------------------------------------------
# `sliding_index()`

test_that("defaults works", {
  df <- data.frame(x = 1:3)
  x <- sliding_index(df, x)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1L)
  expect_identical(split1$out_id, 2L)

  expect_identical(split2$in_id, 2L)
  expect_identical(split2$out_id, 3L)

  expect_identical(nrow(x), 2L)
})

test_that("can lookback over irregular index", {
  df <- data.frame(x = c(1, 3, 4, 5))
  x <- sliding_index(df, x, lookback = 1)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 2L)
  expect_identical(split1$out_id, 3L)

  expect_identical(split2$in_id, 2:3)
  expect_identical(split2$out_id, 4L)

  expect_identical(nrow(x), 2L)
})

test_that("can compute assessment indices relative to irregular index", {
  df <- data.frame(x = c(1, 3, 4, 5, 7, 8))
  x <- sliding_index(df, x, lookback = 1, assess_stop = 2)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]
  split3 <- x$splits[[3]]

  expect_identical(split1$in_id, 2L)
  expect_identical(split1$out_id, 3:4)

  expect_identical(split2$in_id, 2:3)
  expect_identical(split2$out_id, 4L)

  expect_identical(split3$in_id, 3:4)
  expect_identical(split3$out_id, 5L)

  expect_identical(nrow(x), 3L)
})

test_that("it is possible to create empty assessment sets", {
  # Look forward 1->2 values from `5`, so creates a window with range of [6, 7].
  # But no `x` values fall in this range. However, in theory it is "possible"
  # to make a complete window starting at `5`, which is why `complete = TRUE`
  # didn't remove it.
  df <- data.frame(x = c(1, 3, 4, 5, 8, 9))
  x <- sliding_index(df, x, lookback = 1, assess_stop = 2)

  split3 <- x$splits[[3]]

  expect_identical(split3$in_id, 3:4)
  expect_identical(split3$out_id, integer())

  expect_identical(nrow(x), 3L)
})

test_that("can add a gap between the analysis and assessment set", {
  df <- data.frame(x = c(1, 3, 4, 5, 7, 8))
  x <- sliding_index(df, x, lookback = 2, assess_start = 2, assess_stop = 3)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]
  split3 <- x$splits[[3]]

  expect_identical(split1$in_id, 1:2)
  expect_identical(split1$out_id, 4L)

  expect_identical(split2$in_id, 2:3)
  expect_identical(split2$out_id, 5L)

  expect_identical(split3$in_id, 2:4)
  expect_identical(split3$out_id, 5:6)

  expect_identical(nrow(x), 3L)
})

test_that("can use `step` to thin results after calling `slide_index()`", {
  df <- data.frame(x = c(1, 3, 4, 6, 7, 10))
  x <- sliding_index(df, x, lookback = 2, assess_stop = 2, step = 2)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1:2)
  expect_identical(split1$out_id, 3L)

  expect_identical(split2$in_id, 3:4)
  expect_identical(split2$out_id, 5L)

  expect_identical(nrow(x), 2L)
})

test_that("can skip first few resampling slices", {
  df <- data.frame(x = c(1, 3, 4, 6, 7, 10))
  x <- sliding_index(df, x, lookback = 1, skip = 2)

  split1 <- x[["splits"]][[1]]

  expect_identical(split1[["in_id"]], 4L)
  expect_identical(split1[["out_id"]], 5L)

  expect_identical(nrow(x), 2L)
})

test_that("`skip` is applied before `step`", {
  df <- data.frame(x = c(1, 3, 4, 6, 7, 9, 11, 13, 14))
  x <- sliding_index(df, x, lookback = 1, skip = 3, step = 2, assess_stop = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 4:5)
  expect_identical(split1[["out_id"]], 6L)

  expect_identical(split2[["in_id"]], 7L)
  expect_identical(split2[["out_id"]], 8L)

  expect_identical(nrow(x), 2L)
})

test_that("can use incomplete windows at the beginning", {
  df <- data.frame(x = c(1, 3, 4, 5, 7))
  x <- sliding_index(df, x, lookback = 2, complete = FALSE, assess_stop = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]
  split3 <- x[["splits"]][[3]]
  split4 <- x[["splits"]][[4]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 1:2)
  expect_identical(split2[["out_id"]], 3:4)

  expect_identical(split3[["in_id"]], 2:3)
  expect_identical(split3[["out_id"]], 4L)

  expect_identical(split4[["in_id"]], 2:4)
  expect_identical(split4[["out_id"]], 5L)

  expect_identical(nrow(x), 4L)
})

test_that("`data` is validated", {
  expect_error(sliding_index(1), "`data` must be a data frame")
})

test_that("`index` is validated", {
  df <- data.frame(x = 1:2)
  expect_error(sliding_index(df, y))
})

# ------------------------------------------------------------------------------
# `sliding_period()`

test_that("can group by month", {
  index <- vctrs::new_date(c(-1, 0, 1, 31))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, period = "month")

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1L)
  expect_identical(split1$out_id, 2:3)

  expect_identical(split2$in_id, 2:3)
  expect_identical(split2$out_id, 4L)

  expect_identical(nrow(x), 2L)
})

test_that("can group by year", {
  index <- vctrs::new_date(c(-1, 0, 1, 31))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, period = "year")

  split1 <- x$splits[[1]]

  expect_identical(split1$in_id, 1L)
  expect_identical(split1$out_id, 2:4)

  expect_identical(nrow(x), 1L)
})

test_that("when looking back over multiple periods, only complete ones are used", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, period = "month", lookback = 1)

  split1 <- x$splits[[1]]

  expect_identical(split1$in_id, 1:2)
  expect_identical(split1$out_id, 3:4)

  expect_identical(nrow(x), 2L)
})

test_that("can look forward to assess over multiple periods", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, period = "month", assess_stop = 2)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1L)
  expect_identical(split1$out_id, 2:4)

  expect_identical(split2$in_id, 2L)
  expect_identical(split2$out_id, 3:5)

  expect_identical(nrow(x), 2L)
})

test_that("can use `step` to thin results after calling `slide_period()`", {
  df <- data.frame(x = vctrs::new_date(c(1, 3, 4, 6, 7, 10)))
  x <- sliding_period(df, x, "day", lookback = 2, assess_stop = 2, step = 2)

  split1 <- x$splits[[1]]
  split2 <- x$splits[[2]]

  expect_identical(split1$in_id, 1:2)
  expect_identical(split1$out_id, 3L)

  expect_identical(split2$in_id, 3:4)
  expect_identical(split2$out_id, 5L)

  expect_identical(nrow(x), 2L)
})

test_that("can skip first few resampling slices", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31, 59))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, "month", lookback = 1, skip = 2)

  split1 <- x[["splits"]][[1]]

  expect_identical(split1[["in_id"]], 3:5)
  expect_identical(split1[["out_id"]], 6L)

  expect_identical(nrow(x), 1L)
})

test_that("can skip with expanding window", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31, 59))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, "month", lookback = Inf, skip = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:4)
  expect_identical(split1[["out_id"]], 5L)

  expect_identical(split2[["in_id"]], 1:5)
  expect_identical(split2[["out_id"]], 6L)

  expect_identical(nrow(x), 2L)
})

test_that("`skip` is applied before `step`", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 31, 59, 90))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, "month", lookback = Inf, skip = 2, step = 2)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]

  expect_identical(split1[["in_id"]], 1:4)
  expect_identical(split1[["out_id"]], 5L)

  expect_identical(split2[["in_id"]], 1:6)
  expect_identical(split2[["out_id"]], 7L)

  expect_identical(nrow(x), 2L)
})

test_that("can use incomplete windows at the beginning", {
  index <- vctrs::new_date(c(-32, -1, 0, 1, 59, 90))
  df <- data.frame(index = index)

  x <- sliding_period(df, index, "month", lookback = 2, complete = FALSE)

  split1 <- x[["splits"]][[1]]
  split2 <- x[["splits"]][[2]]
  split3 <- x[["splits"]][[3]]
  split4 <- x[["splits"]][[4]]

  expect_identical(split1[["in_id"]], 1L)
  expect_identical(split1[["out_id"]], 2L)

  expect_identical(split2[["in_id"]], 1:2)
  expect_identical(split2[["out_id"]], 3:4)

  expect_identical(split3[["in_id"]], 1:4)
  expect_identical(split3[["out_id"]], integer())

  expect_identical(split4[["in_id"]], 3:5)
  expect_identical(split4[["out_id"]], 6L)

  expect_identical(nrow(x), 4L)
})

test_that("`data` is validated", {
  expect_error(sliding_period(1), "`data` must be a data frame")
})

test_that("`index` is validated", {
  df <- data.frame(x = 1:2)
  expect_error(sliding_period(df, y))
})

library(testthat)
library(rsample)

if (requireNamespace("xml2")) {
  test_check("rsample", reporter = MultiReporter$new(reporters = list(JunitReporter$new(file = "test-results.xml"), CheckReporter$new())))
} else {
  test_check("rsample")
}

#' Create an Initial Train/Validation/Test Split
#'
#' `initial_validation_split()` creates a random three-way split of the data
#' into a training set, a validation set, and a testing set.
#' `initial_validation_time_split()` does the same, but instead of a random
#' selection the training, validation, and testing set are in order of the full
#' data set, with the first observations being put into the training set.
#' `group_initial_validation_split()` creates similar random splits of the data
#' based on some grouping variable, so that all data in a "group" are assigned
#' to the same partition.
#'
#' @details [training()], [validation()], and [testing()] can be used to extract the
#' resulting data sets.
#' Use [validation_set()] to create an `rset` object for use with functions from
#' the tune package such as `tune::tune_grid()`.
#'
#' @template strata_details
#'
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param prop A length-2 vector of proportions of data to be retained for training and
#' validation data, respectively.
#' @inheritParams rlang::args_dots_empty
#' @param x An object of class `initial_validation_split`.
#'
#' @return An `initial_validation_split` object that can be used with the
#' [training()], [validation()], and [testing()] functions to extract the data
#' in each split.
#'
#' @seealso [validation_set()]
#'
#' @export
#' @examplesIf rlang::is_installed("modeldata")
#' set.seed(1353)
#' car_split <- initial_validation_split(mtcars)
#' train_data <- training(car_split)
#' validation_data <- validation(car_split)
#' test_data <- testing(car_split)
#'
#' data(drinks, package = "modeldata")
#' drinks_split <- initial_validation_time_split(drinks)
#' train_data <- training(drinks_split)
#' validation_data <- validation(drinks_split)
#' c(max(train_data$date), min(validation_data$date))
#'
#' data(ames, package = "modeldata")
#' set.seed(1353)
#' ames_split <- group_initial_validation_split(ames, group = Neighborhood)
#' train_data <- training(ames_split)
#' validation_data <- validation(ames_split)
#' test_data <- testing(ames_split)
initial_validation_split <- function(
  data,
  prop = c(0.6, 0.2),
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  ...
) {
  rlang::check_dots_empty()

  check_prop_3(prop)
  prop_train <- prop[1]
  prop_val <- prop[2] / (1 - prop_train)

  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) {
      strata <- NULL
    }
  }
  check_strata(strata, data)

  split_train <- mc_cv(
    data = data,
    prop = prop_train,
    strata = {{ strata }},
    breaks = breaks,
    pool = pool,
    times = 1,
    ...
  )
  split_train <- split_train$splits[[1]]
  train_id <- split_train$in_id

  val_and_test <- assessment(split_train)
  val_and_test_id <- complement(split_train)

  split_val <- mc_cv(
    data = val_and_test,
    prop = prop_val,
    strata = {{ strata }},
    breaks = breaks,
    pool = pool / (1 - prop_train),
    times = 1,
    ...
  )
  split_val <- split_val$splits[[1]]
  val_id <- val_and_test_id[split_val$in_id]

  res <- list(
    data = data,
    train_id = train_id,
    val_id = val_id,
    test_id = NA,
    id = "split"
  )

  # include those so that they can be attached to the `rset` later in `validation_set()`
  if (!is.null(strata)) {
    names(strata) <- NULL
  }
  val_att <- list(
    prop = prop,
    strata = strata,
    breaks = breaks,
    pool = pool
  )
  attr(res, "val_att") <- val_att

  class(res) <- c("initial_validation_split", "three_way_split")
  res
}

check_prop_3 <- function(prop, call = rlang::caller_env()) {
  if (!is.numeric(prop)) {
    cli_abort("{.arg prop} needs to be numeric.", call = call)
  }
  if (any(is.na(prop))) {
    cli_abort("{.arg prop} cannot contain `NA`.", call = call)
  }
  if (any(is.null(prop))) {
    cli_abort("{.arg prop} cannot contain `NULL`.", call = call)
  }
  if (length(prop) != 2L) {
    cli_abort(
      "{.arg prop} needs to contain the proportions for training and validation.",
      call = call
    )
  }
  if (any(!(prop > 0)) | any(!(prop < 1))) {
    cli_abort("Elements of {.arg prop} need to be in (0, 1).", call = call)
  }
  if (!(sum(prop) > 0) | !(sum(prop) < 1)) {
    cli_abort(
      "The sum of the proportions in {.arg prop} needs to be in (0, 1).",
      call = call
    )
  }
  invisible(prop)
}

#' @rdname initial_validation_split
#' @export
initial_validation_time_split <- function(data, prop = c(0.6, 0.2), ...) {
  rlang::check_dots_empty()

  check_prop_3(prop)
  prop_train <- prop[1]
  prop_val <- prop[2] / (1 - prop_train)

  n_train <- floor(nrow(data) * prop_train)
  n_val <- floor((nrow(data) - n_train) * prop_val)

  train_id <- seq(1, n_train, by = 1)
  val_id <- seq(n_train + 1, n_train + n_val, by = 1)

  res <- list(
    data = data,
    train_id = train_id,
    val_id = val_id,
    test_id = NA,
    id = "split"
  )

  # include those so that they can be attached to the `rset` later in `validation_set()`
  val_att <- list(
    prop = prop
  )
  attr(res, "val_att") <- val_att

  class(res) <- c(
    "initial_validation_time_split",
    "initial_validation_split",
    "three_way_split"
  )
  res
}

#' @inheritParams make_groups
#' @rdname initial_validation_split
#' @export
group_initial_validation_split <- function(
  data,
  group,
  prop = c(0.6, 0.2),
  ...,
  strata = NULL,
  pool = 0.1
) {
  rlang::check_dots_empty()

  group <- validate_group({{ group }}, data)

  check_prop_3(prop)
  prop_train <- prop[1]
  prop_val <- prop[2] / (1 - prop_train)

  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) {
      strata <- NULL
    }
  }
  check_strata(strata, data)

  if (missing(strata)) {
    split_train <- group_mc_cv(
      data = data,
      group = {{ group }},
      prop = prop_train,
      times = 1
    )
  } else {
    split_train <- group_mc_cv(
      data = data,
      group = {{ group }},
      prop = prop_train,
      times = 1,
      strata = {{ strata }},
      pool = pool
    )
  }
  split_train <- split_train$splits[[1]]
  train_id <- split_train$in_id

  val_and_test <- assessment(split_train)
  val_and_test_id <- complement(split_train)

  if (missing(strata)) {
    split_val <- group_mc_cv(
      data = val_and_test,
      group = {{ group }},
      prop = prop_val,
      times = 1
    )
  } else {
    split_val <- group_mc_cv(
      data = val_and_test,
      group = {{ group }},
      prop = prop_val,
      times = 1,
      strata = {{ strata }},
      pool = pool
    )
  }
  split_val <- split_val$splits[[1]]
  val_id <- val_and_test_id[split_val$in_id]

  res <- list(
    data = data,
    train_id = train_id,
    val_id = val_id,
    test_id = NA,
    id = "split"
  )

  # include those so that they can be attached to the `rset` later in `validation_set()`
  if (!is.null(strata)) {
    names(strata) <- NULL
  }
  val_att <- list(
    group = group,
    prop = prop,
    strata = strata,
    pool = pool
  )
  attr(res, "val_att") <- val_att

  class(res) <- c(
    "group_initial_validation_split",
    "initial_validation_split",
    "three_way_split"
  )
  res
}

#' @export
#' @rdname initial_validation_split
training.initial_validation_split <- function(x, ...) {
  check_dots_empty()
  ind <- sort(x$train_id)
  vctrs::vec_slice(x$data, ind)
}

#' @export
#' @rdname initial_validation_split
testing.initial_validation_split <- function(x, ...) {
  check_dots_empty()
  ind <- -sort(c(x$train_id, x$val_id))
  vctrs::vec_slice(x$data, ind)
}

#' @export
#' @rdname initial_validation_split
validation <- function(x, ...) {
  UseMethod("validation")
}

#' @export
#' @rdname initial_validation_split
validation.default <- function(x, ...) {
  cls <- class(x)
  cli_abort(
    "No method for objects of class{?es}: {.cls {cls}}"
  )
}

#' @export
#' @rdname initial_validation_split
validation.initial_validation_split <- function(x, ...) {
  check_dots_empty()

  ind <- sort(x$val_id)
  vctrs::vec_slice(x$data, ind)
}


#' @export
#' @keywords internal
analysis.initial_validation_split <- function(x, ...) {
  cli_abort(
    c(
      "The initial validation split does not contain an analysis set.",
      "i" = "You can access the training data with {.fun training}."
    )
  )
}

#' @export
#' @keywords internal
assessment.initial_validation_split <- function(x, ...) {
  cli_abort(
    c(
      "The initial validation split does not contain an assessment set.",
      "i" = "You can access the testing data with {.fun testing}."
    )
  )
}

#' @export
print.initial_validation_split <- function(x, ...) {
  n_test <- nrow(x$data) - length(x$train_id) - length(x$val_id)
  n_test_char <- paste(n_test)

  cat("<Training/Validation/Testing/Total>\n")
  cat(
    "<",
    length(x$train_id),
    "/",
    length(x$val_id),
    "/",
    n_test_char,
    "/",
    nrow(x$data),
    ">\n",
    sep = ""
  )
}

#' @export
dim.initial_validation_split <- function(x, ...) {
  n_test <- nrow(x$data) - length(x$train_id) - length(x$val_id)
  c(
    training = length(x$train_id),
    validation = length(x$val_id),
    testing = length(n_test),
    n = nrow(x$data),
    p = ncol(x$data)
  )
}

test_that("basic split", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))

  set.seed(11)
  rs1 <- initial_validation_split(dat1, prop = c(0.6, 0.2))

  expect_equal(class(rs1), c("initial_validation_split", "three_way_split"))

  size_train <- length(rs1$train_id)
  size_val <- length(rs1$val_id)

  expect_equal(size_train, 12)
  expect_equal(size_val, 4)
  expect_equal(rs1$test_id, NA)

  expect_equal(rs1$data, dat1)

  good_holdout <- length(intersect(rs1$train_id, rs1$val_id))
  expect_equal(good_holdout, 0)
})

test_that("basic split - accessor functions", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))

  set.seed(11)
  val_split <- initial_validation_split(dat1, prop = c(0.6, 0.2))

  dat_train <- training(val_split)
  dat_val <- validation(val_split)
  dat_test <- testing(val_split)

  expect_equal(nrow(dat_train), 12)
  expect_equal(nrow(dat_val), 4)
  expect_equal(nrow(dat_test), 4)

  expect_snapshot(error = TRUE, {
    analysis(val_split)
  })
  expect_snapshot(error = TRUE, {
    assessment(val_split)
  })
})

test_that("basic split stratified", {
  dat <- data.frame(
    id = 1:100,
    f = factor(rep(letters[1:3], times = c(50, 25, 25)))
  )

  val_split <- initial_validation_split(dat, strata = f)

  # distribution of strata should be similar in all 3 data sets
  expected <- c(50, 25, 25) / 100
  actual_train <- dat[val_split$train_id, "f"] |> table() |> prop.table()
  expect_equal(as.vector(actual_train), expected)

  actual_val <- dat[val_split$val_id, "f"] |> table() |> prop.table()
  expect_equal(as.vector(actual_val), expected)

  actual_test <- dat[-c(val_split$train_id, val_split$val_id), "f"] |>
    table() |>
    prop.table()
  expect_equal(as.vector(actual_test), expected)

  # bad args for `strata`
  expect_snapshot(error = TRUE, {
    initial_validation_split(dat, strata = does_not_exist)
  })
  expect_snapshot(error = TRUE, {
    initial_validation_split(dat, strata = c(x, f))
  })
  expect_snapshot(error = TRUE, {
    initial_validation_split(dat, strata = rep(1:3, times = c(50, 25, 25)))
  })
})

test_that("time split", {
  dat1 <- data.frame(a = 1:109)

  set.seed(11)
  rs1 <- initial_validation_time_split(dat1, prop = c(0.6, 0.2))

  expect_s3_class(
    rs1,
    c(
      "initial_validation_time_split",
      "initial_validation_split",
      "three_way_split"
    )
  )

  exp_size_train <- floor(nrow(dat1) * 0.6)
  exp_size_val <- floor((nrow(dat1) - exp_size_train) * 0.2 / (1 - 0.6))

  expect_equal(rs1$train_id, seq(1, exp_size_train))
  expect_equal(
    rs1$val_id,
    seq(exp_size_train + 1, exp_size_train + exp_size_val)
  )
  expect_equal(rs1$test_id, NA)

  expect_equal(rs1$data, dat1)

  good_val <- length(intersect(rs1$train_id, rs1$val_id))
  expect_equal(good_val, 0)
  good_test <- length(intersect(rs1$val_id, rs1$test_id))
  expect_equal(good_test, 0)
})

test_that("grouped split", {
  # all observations of each group should be in only one of the 3 data sets
  # = all obs in the same group and no intersection in the groups
  #   from the 3 data sets

  dat <- data.frame(
    id = 1:100,
    g = rep(1:5, each = 20)
  )

  val_split <- group_initial_validation_split(
    dat,
    group = g,
    prop = c(0.6, 0.2)
  )

  dat_train <- dat[val_split$train_id, ]
  dat_val <- dat[val_split$val_id, ]
  dat_test <- dat[-c(val_split$train_id, val_split$val_id), ]

  expect_equal(nrow(dat_train), 60)
  expect_equal(nrow(dat_val), 20)
  expect_equal(nrow(dat_test), 20)

  g_train <- dat_train |> dplyr::count(g)
  g_val <- dat_val |> dplyr::count(g)
  g_test <- dat_test |> dplyr::count(g)

  # all obs of the chosen groups are here
  # which also means there are none elsewhere
  expect_true(all(g_train$n == 20))
  expect_true(all(g_val$n == 20))
  expect_true(all(g_test$n == 20))

  intersect_train_val <- intersect(g_train$id, g_val$id)
  intersect_train_test <- intersect(g_train$id, g_test$id)
  intersect_val_test <- intersect(g_val$id, g_test$id)
  expect_equal(length(intersect_train_val), 0)
  expect_equal(length(intersect_train_test), 0)
  expect_equal(length(intersect_val_test), 0)
})

test_that("grouped split stratified", {
  set.seed(11)

  n_common_class <- 70
  n_rare_class <- 30

  group_table <- tibble(
    group = 1:100,
    outcome = sample(c(rep(0, n_common_class), rep(1, n_rare_class)))
  )
  observation_table <- tibble(
    group = sample(1:100, 5e4, replace = TRUE),
    observation = 1:5e4
  )
  sample_data <- dplyr::full_join(
    group_table,
    observation_table,
    by = "group",
    multiple = "all"
  )

  val_split <- group_initial_validation_split(
    sample_data,
    group = "group",
    prop = c(0.6, 0.2),
    strata = outcome
  )

  dat_train <- sample_data[val_split$train_id, ]
  dat_val <- sample_data[val_split$val_id, ]
  dat_test <- sample_data[-c(val_split$train_id, val_split$val_id), ]

  expect_equal(mean(dat_train$outcome == 1), 0.3, tolerance = 1e-2)
  expect_equal(mean(dat_val$outcome == 1), 0.3, tolerance = 1e-2)
  expect_equal(mean(dat_test$outcome == 1), 0.3, tolerance = 1e-1)

  intersect_train_val <- intersect(val_split$train_id, val_split$val_id)
  expect_equal(length(intersect_train_val), 0)
})

test_that("grouped split - accessor functions", {
  dat <- data.frame(
    id = 1:100,
    g = rep(1:5, each = 20)
  )

  set.seed(1)
  val_split <- group_initial_validation_split(
    dat,
    group = g,
    prop = c(0.6, 0.2)
  )

  dat_train <- training(val_split)
  dat_val <- validation(val_split)
  dat_test <- testing(val_split)

  expect_equal(nrow(dat_train), 60)
  expect_equal(nrow(dat_val), 20)
  expect_equal(nrow(dat_test), 20)

  expect_snapshot(error = TRUE, {
    analysis(val_split)
  })
  expect_snapshot(error = TRUE, {
    assessment(val_split)
  })
})

test_that("check_prop_3() works", {
  expect_snapshot(error = TRUE, check_prop_3(0.3))
  expect_snapshot(error = TRUE, check_prop_3("zero"))
  expect_snapshot(error = TRUE, check_prop_3(NULL))
  expect_snapshot(error = TRUE, check_prop_3(NA))
  expect_snapshot(error = TRUE, check_prop_3(0))

  expect_snapshot(error = TRUE, check_prop_3(c(0.3, NA)))
  expect_snapshot(error = TRUE, check_prop_3(c(0.3, NULL)))
  expect_snapshot(error = TRUE, check_prop_3(c(0.3, 1)))
  expect_snapshot(error = TRUE, check_prop_3(c(0.3, 0.7)))
})

.onLoad <- function(libname, pkgname) {

  if (requireNamespace("dplyr", quietly = TRUE)) {
    register_s3_method("dplyr", "arrange", "rset")
    register_s3_method("dplyr", "filter", "rset")
    register_s3_method("dplyr", "mutate", "rset")
    register_s3_method("dplyr", "rename", "rset")
    register_s3_method("dplyr", "select", "rset")
    register_s3_method("dplyr", "slice", "rset")
  }
  
  invisible()
  
}

## ----ex_setup, include=FALSE------------------------------------------------------------
knitr::opts_chunk$set(
  message = FALSE,
  digits = 3,
  collapse = TRUE,
  comment = "#>"
  )
options(digits = 3, width = 90)
library(ggplot2)
theme_set(theme_bw())

## ----attrition, message=FALSE-----------------------------------------------------------
library(rsample)
data("attrition", package = "modeldata")
names(attrition)
table(attrition$Attrition)

## ----form, message=FALSE----------------------------------------------------------------
mod_form <- as.formula(Attrition ~ JobSatisfaction + Gender + MonthlyIncome)

## ----model_vfold, message=FALSE---------------------------------------------------------
library(rsample)
set.seed(4622)
rs_obj <- vfold_cv(attrition, v = 10, repeats = 10)
rs_obj

## ----lm_func----------------------------------------------------------------------------
## splits will be the `rsplit` object with the 90/10 partition
holdout_results <- function(splits, ...) {
  # Fit the model to the 90%
  mod <- glm(..., data = analysis(splits), family = binomial)
  # Save the 10%
  holdout <- assessment(splits)
  # `augment` will save the predictions with the holdout data set
  res <- broom::augment(mod, newdata = holdout)
  # Class predictions on the assessment set from class probs
  lvls <- levels(holdout$Attrition)
  predictions <- factor(ifelse(res$.fitted > 0, lvls[2], lvls[1]),
                        levels = lvls)
  # Calculate whether the prediction was correct
  res$correct <- predictions == holdout$Attrition
  # Return the assessment data set with the additional columns
  res
}

## ----onefold, warning = FALSE-----------------------------------------------------------
example <- holdout_results(rs_obj$splits[[1]],  mod_form)
dim(example)
dim(assessment(rs_obj$splits[[1]]))
## newly added columns:
example[1:10, setdiff(names(example), names(attrition))]

## ----model_purrr, warning=FALSE---------------------------------------------------------
library(purrr)
rs_obj$results <- map(rs_obj$splits,
                      holdout_results,
                      mod_form)
rs_obj

## ----model_acc--------------------------------------------------------------------------
rs_obj$accuracy <- map_dbl(rs_obj$results, function(x) mean(x$correct))
summary(rs_obj$accuracy)

## ----type_plot--------------------------------------------------------------------------
ggplot(attrition, aes(x = Gender, y = MonthlyIncome)) + 
  geom_boxplot() + 
  scale_y_log10()

## ----mean_diff--------------------------------------------------------------------------
median_diff <- function(splits) {
  x <- analysis(splits)
  median(x$MonthlyIncome[x$Gender == "Female"]) - 
      median(x$MonthlyIncome[x$Gender == "Male"])     
}

## ----boot_mean_diff---------------------------------------------------------------------
set.seed(353)
bt_resamples <- bootstraps(attrition, times = 500)

## ----stats------------------------------------------------------------------------------
bt_resamples$wage_diff <- map_dbl(bt_resamples$splits, median_diff)

## ----stats_plot-------------------------------------------------------------------------
ggplot(bt_resamples, aes(x = wage_diff)) + 
  geom_line(stat = "density", adjust = 1.25) + 
  xlab("Difference in Median Monthly Income (Female - Male)")

## ----ci---------------------------------------------------------------------------------
quantile(bt_resamples$wage_diff, 
         probs = c(0.025, 0.975))

## ----coefs------------------------------------------------------------------------------
glm_coefs <- function(splits, ...) {
  ## use `analysis` or `as.data.frame` to get the analysis data
  mod <- glm(..., data = analysis(splits), family = binomial)
  as.data.frame(t(coef(mod)))
}
bt_resamples$betas <- map(.x = bt_resamples$splits, 
                          .f = glm_coefs, 
                          mod_form)
bt_resamples
bt_resamples$betas[[1]]

## ----tidy_rsplit------------------------------------------------------------------------
first_resample <- bt_resamples$splits[[1]]
class(first_resample)
tidy(first_resample)

## ----tidy_rset--------------------------------------------------------------------------
class(bt_resamples)
tidy(bt_resamples)

#' Cluster Cross-Validation
#'
#' Cluster cross-validation splits the data into V groups of
#'  disjointed sets using k-means clustering of some variables.
#'  A resample of the analysis data consists of V-1 of the
#'  folds/clusters while the assessment set contains the final fold/cluster. In
#'  basic cross-validation (i.e. no repeats), the number of resamples
#'  is equal to V.
#'
#' @details
#' The variables in the `vars` argument are used for k-means clustering of
#'  the data into disjointed sets or for hierarchical clustering of the data.
#'  These clusters are used as the folds for cross-validation. Depending on how
#'  the data are distributed, there may not be an equal number of points
#'  in each fold.
#'
#' You can optionally provide a custom function to `distance_function`. The
#' function should take a data frame (as created via `data[vars]`) and return
#' a [stats::dist()] object with distances between data points.
#'
#' You can optionally provide a custom function to `cluster_function`. The
#' function must take three arguments:
#' - `dists`, a [stats::dist()] object with distances between data points
#' - `v`, a length-1 numeric for the number of folds to create
#' - `...`, to pass any additional named arguments to your function
#'
#' The function should return a vector of cluster assignments of length
#' `nrow(data)`, with each element of the vector corresponding to the matching
#' row of the data frame.
#'
#' @inheritParams vfold_cv
#' @param vars A vector of bare variable names to use to cluster the data.
#' @param repeats The number of times to repeat the clustered partitioning.
#' @param distance_function Which function should be used for distance calculations?
#' Defaults to [stats::dist()]. You can also provide your own
#' function; see `Details`.
#' @param cluster_function Which function should be used for clustering?
#' Options are either `"kmeans"` (to use [stats::kmeans()])
#' or `"hclust"` (to use [stats::hclust()]). You can also provide your own
#' function; see `Details`.
#' @param ... Extra arguments passed on to `cluster_function`.
#'
#' @return A tibble with classes `rset`, `tbl_df`, `tbl`, and `data.frame`.
#'  The results include a column for the data split objects and
#'  an identification variable `id`.
#'
#' @examplesIf rlang::is_installed("modeldata")
#' data(ames, package = "modeldata")
#' clustering_cv(ames, vars = c(Sale_Price, First_Flr_SF, Second_Flr_SF), v = 2)
#'
#' @rdname clustering_cv
#' @export
clustering_cv <- function(
  data,
  vars,
  v = 10,
  repeats = 1,
  distance_function = "dist",
  cluster_function = c("kmeans", "hclust"),
  ...
) {
  check_number_whole(repeats, min = 1)

  if (!rlang::is_function(cluster_function)) {
    cluster_function <- rlang::arg_match(cluster_function)
  }

  vars <- tidyselect::eval_select(rlang::enquo(vars), data = data)
  if (rlang::is_empty(vars)) {
    cli_abort(
      "{.arg vars} is required and must contain at least one variable in {.arg data}."
    )
  }
  vars <- data[vars]

  if (repeats == 1) {
    dists <- rlang::exec(distance_function, vars)
    split_objs <- clustering_splits(
      data = data,
      dists = dists,
      v = v,
      cluster_function = cluster_function,
      ...
    )
  } else {
    for (i in 1:repeats) {
      dists <- rlang::exec(distance_function, vars)
      tmp <- clustering_splits(
        data = data,
        dists = dists,
        v = v,
        cluster_function = cluster_function,
        ...
      )
      tmp$id2 <- tmp$id
      tmp$id <- names0(repeats, "Repeat")[i]
      split_objs <- if (i == 1) {
        tmp
      } else {
        rbind(split_objs, tmp)
      }
    }
  }

  split_objs$splits <- map(split_objs$splits, rm_out)

  ## Save some overall information

  cv_att <- list(
    v = v,
    vars = names(vars),
    repeats = repeats,
    distance_function = distance_function,
    cluster_function = cluster_function
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs[, grepl("^id", names(split_objs))],
    attrib = cv_att,
    subclass = c("clustering_cv", "rset")
  )
}

clustering_splits <- function(
  data,
  dists,
  v = 10,
  cluster_function = c("kmeans", "hclust"),
  ...
) {
  if (!rlang::is_function(cluster_function)) {
    cluster_function <- rlang::arg_match(cluster_function)
  }

  check_v(v, nrow(data), "rows", call = rlang::caller_env())
  n <- nrow(data)

  clusterer <- ifelse(
    rlang::is_function(cluster_function),
    "custom",
    cluster_function
  )
  folds <- switch(
    clusterer,
    "kmeans" = {
      clusters <- stats::kmeans(dists, centers = v, ...)
      clusters$cluster
    },
    "hclust" = {
      clusters <- stats::hclust(dists, ...)
      stats::cutree(clusters, k = v)
    },
    do.call(cluster_function, list(dists = dists, v = v, ...))
  )

  idx <- seq_len(n)
  indices <- split_unnamed(idx, folds)
  indices <- lapply(indices, default_complement, n = n)

  split_objs <- purrr::map(
    indices,
    make_splits,
    data = data,
    class = c("clustering_split")
  )
  tibble::tibble(
    splits = split_objs,
    id = names0(length(split_objs), "Fold")
  )
}

#' Manual resampling
#'
#' `manual_rset()` is used for constructing the most minimal rset possible. It
#' can be useful when you have custom rsplit objects built from
#' [make_splits()], or when you want to create a new rset from splits
#' contained within an existing rset.
#'
#' @param splits A list of `"rsplit"` objects. It is easiest to create these
#'   using [make_splits()].
#'
#' @param ids A character vector of ids. The length of `ids` must be the same
#'   as the length of `splits`.
#'
#' @export
#' @examples
#' df <- data.frame(x = c(1, 2, 3, 4, 5, 6))
#'
#' # Create an rset from custom indices
#' indices <- list(
#'   list(analysis = c(1L, 2L), assessment = 3L),
#'   list(analysis = c(4L, 5L), assessment = 6L)
#' )
#'
#' splits <- lapply(indices, make_splits, data = df)
#'
#' manual_rset(splits, c("Split 1", "Split 2"))
#'
#' # You can also use this to create an rset from a subset of an
#' # existing rset
#' resamples <- vfold_cv(mtcars)
#' best_split <- resamples[5,]
#' manual_rset(best_split$splits, best_split$id)
manual_rset <- function(splits, ids) {
  new_manual_rset(splits, ids)
}

new_manual_rset <- function(splits, ids) {
  new_rset(splits, ids, subclass = c("manual_rset", "rset"))
}

#' @export
print.manual_rset <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("manual_rset", "rset"))]
  print(x, ...)
}

#' @importFrom utils globalVariables
#' @importFrom purrr map map_df map2 map_dfr map_dbl pluck map_lgl
#' @importFrom tibble tibble is_tibble as_tibble obj_sum type_sum
#' @importFrom tidyr gather unnest
#' @importFrom dplyr select %>% bind_cols bind_rows arrange_ arrange full_join
#' @importFrom dplyr mutate last ungroup group_by inner_join summarize do case_when
#' @importFrom rlang !! is_call is_string enquo quos exec is_list abort warn
#' @importFrom methods formalArgs
#' @importFrom stats quantile setNames qnorm pnorm
#' @importFrom tidyselect vars_select one_of
#' @importFrom furrr future_map_dfr
#' @importFrom tidyr gather



# ------------------------------------------------------------------------------

utils::globalVariables(
  c("model", "splits", "statistic", "Data", "Row", "id", ".", ".estimate",
    ".lower", ".upper", "Z0", "Za", "Zl", "Zu", "a", "cubed", "estimate",
    "orig", "p0", "squared", "term", "theta_0", "loo", "n", "..index"
    )
  )

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

#------------------------------------------------------------------------------#

#' @importFrom utils globalVariables
#' @importFrom purrr map map_df map2 map_dfr map_dbl pluck map_lgl
#' @importFrom tibble tibble is_tibble as_tibble obj_sum
#' @importFrom pillar type_sum
#' @importFrom tidyr gather unnest
#' @importFrom dplyr select %>% bind_cols bind_rows arrange_ arrange full_join
#' @importFrom dplyr mutate last ungroup group_by inner_join summarize do case_when
#' @importFrom vctrs vec_unique_count vec_count vec_slice vec_in
#' @importFrom rlang !! is_call is_string enquo quos exec is_list abort warn
#' @importFrom methods formalArgs
#' @importFrom stats quantile setNames qnorm pnorm
#' @importFrom tidyselect vars_select one_of
#' @importFrom furrr future_map_dfr
#' @importFrom tidyr gather

#------------------------------------------------------------------------------#

utils::globalVariables(
  c(
    "model", "splits", "statistic", "Data", "Row", "id", ".", ".estimate",
    ".lower", ".upper", "Z0", "Za", "Zl", "Zu", "a", "cubed", "estimate",
    "orig", "p0", "squared", "term", "theta_0", "loo", "n", "..index", "models",
    paste0("predictor_", gsub(" ", "0", format(1:20))), "linear_pred",
    "non_linear_1", "non_linear_2", "non_linear_3", "outcome", "true_prob",
    "two_factor_1", "two_factor_2", "rand"
  )
)

#' Simple Training/Test Set Splitting
#'
#' `initial_split` creates a single binary split of the data into a training
#'  set and testing set. `initial_time_split` does the same, but takes the
#'  _first_ `prop` samples for training, instead of a random selection.
#'  `group_initial_split` creates splits of the data based
#'  on some grouping variable, so that all data in a "group" is assigned to
#'  the same split.
#'  `training` and `testing` are used to extract the resulting data.
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @export
#' @return An `rsplit` object that can be used with the `training` and `testing`
#'  functions to extract the data in each split.
#' @examplesIf rlang::is_installed("modeldata")
#' set.seed(1353)
#' car_split <- initial_split(mtcars)
#' train_data <- training(car_split)
#' test_data <- testing(car_split)
#'
#' data(drinks, package = "modeldata")
#' drinks_split <- initial_time_split(drinks)
#' train_data <- training(drinks_split)
#' test_data <- testing(drinks_split)
#' c(max(train_data$date), min(test_data$date)) # no lag
#'
#' # With 12 period lag
#' drinks_lag_split <- initial_time_split(drinks, lag = 12)
#' train_data <- training(drinks_lag_split)
#' test_data <- testing(drinks_lag_split)
#' c(max(train_data$date), min(test_data$date)) # 12 period lag
#'
#' set.seed(1353)
#' car_split <- group_initial_split(mtcars, cyl)
#' train_data <- training(car_split)
#' test_data <- testing(car_split)
#'
#' @export
#'
initial_split <- function(data, prop = 3 / 4,
                          strata = NULL, breaks = 4, pool = 0.1, ...) {
  check_dots_empty()
  res <-
    mc_cv(
      data = data,
      prop = prop,
      strata = {{ strata }},
      breaks = breaks,
      pool = pool,
      times = 1
    )
  res <- res$splits[[1]]
  class(res) <- c("initial_split", class(res))
  res
}

#' @rdname initial_split
#' @param lag A value to include a lag between the assessment
#'  and analysis set. This is useful if lagged predictors will be used
#'  during training and testing.
#' @export
initial_time_split <- function(data, prop = 3 / 4, lag = 0, ...) {
  check_dots_empty()
  if (!is.numeric(prop) | prop >= 1 | prop <= 0) {
    rlang::abort("`prop` must be a number on (0, 1).")
  }

  if (!is.numeric(lag) | !(lag %% 1 == 0)) {
    rlang::abort("`lag` must be a whole number.")
  }

  n_train <- floor(nrow(data) * prop)

  if (lag > n_train) {
    rlang::abort("`lag` must be less than or equal to the number of training observations.")
  }

  split <- rsplit(data, 1:n_train, (n_train + 1 - lag):nrow(data))
  splits <- list(split)
  ids <- "Resample1"
  rset <- new_rset(splits, ids)

  res <- rset$splits[[1]]
  class(res) <- c("initial_time_split", "initial_split", class(res))
  res
}

#' @rdname initial_split
#' @export
#' @param x An `rsplit` object produced by `initial_split()` or
#'  `initial_time_split()`.
training <- function(x, ...) {
  UseMethod("training")
}

#' @export
#' @rdname initial_split
training.default <- function(x, ...) {
  cls <- class(x)
  cli::cli_abort(
    "No method for objects of class{?es}: {cls}"
  )
}

#' @rdname initial_split
#' @export
training.rsplit <- function(x, ...) {
  analysis(x, ...)
}

#' @rdname initial_split
#' @export
testing <- function(x, ...) {
  UseMethod("testing")
}

#' @export
#' @rdname initial_split
testing.default <- function(x, ...) {
  cls <- class(x)
  cli::cli_abort(
    "No method for objects of class{?es}: {cls}"
  )
}

#' @rdname initial_split
#' @export
testing.rsplit <- function(x, ...) {
  assessment(x, ...)
}

#' @inheritParams make_groups
#' @rdname initial_split
#' @export
group_initial_split <- function(data, group, prop = 3 / 4, ..., strata = NULL, pool = 0.1) {
  check_dots_empty()

  if (missing(strata)) {
    res <- group_mc_cv(
        data = data,
        group = {{ group }},
        prop = prop,
        times = 1
      )
  } else {
    res <- group_mc_cv(
        data = data,
        group = {{ group }},
        prop = prop,
        times = 1,
        strata = {{ strata }},
        pool = pool
      )
  }
  res <- res$splits[[1]]
  class(res) <- c("group_initial_split", "initial_split", class(res))
  res
}

#' Group V-Fold Cross-Validation
#'
#' Group V-fold cross-validation creates splits of the data based
#'  on some grouping variable (which may have more than a single row
#'  associated with it). The function can create as many splits as
#'  there are unique values of the grouping variable or it can
#'  create a smaller set of splits where more than one value is left
#'  out at a time.
#'
#' @param data A data frame.
#' @param group This could be a single character value or a variable
#'  name that corresponds to a variable that exists in the data frame.
#' @param v The number of partitions of the data set. If let
#'  `NULL`, `v` will be set to the number of unique values
#'  in the group.
#' @param ... Not currently used.
#' @export
#' @return A tibble with classes `group_vfold_cv`,
#'  `rset`, `tbl_df`, `tbl`, and `data.frame`.
#'  The results include a column for the data split objects and an
#'  identification variable.
#' @examples
#' set.seed(3527)
#' test_data <- data.frame(id = sort(sample(1:20, size = 80, replace = TRUE)))
#' test_data$dat <- runif(nrow(test_data))
#'
#' set.seed(5144)
#' split_by_id <- group_vfold_cv(test_data, group = "id")
#'
#' get_id_left_out <- function(x)
#'   unique(assessment(x)$id)
#'
#' library(purrr)
#' table(map_int(split_by_id$splits, get_id_left_out))
#'
#' set.seed(5144)
#' split_by_some_id <- group_vfold_cv(test_data, group = "id", v = 7)
#' held_out <- map(split_by_some_id$splits, get_id_left_out)
#' table(unlist(held_out))
#' # number held out per resample:
#' map_int(held_out, length)
#' @export
group_vfold_cv <- function(data, group = NULL, v = NULL, ...) {

  if(!missing(group)) {
    group <- tidyselect::vars_select(names(data), !!enquo(group))
    if(length(group) == 0) {
      group <- NULL
    }
  }

  if (is.null(group) || !is.character(group) || length(group) != 1)
    stop(
      "`group` should be a single character value for the column ",
      "that will be used for splitting.",
      call. = FALSE
    )
  if (!any(names(data) == group))
    stop("`group` should be a column in `data`.", call. = FALSE)

  split_objs <- group_vfold_splits(data = data, group = group, v = v)

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  # Update `v` if not supplied directly
  if (is.null(v)) {
    v <- length(split_objs$splits)
  }

  ## Save some overall information

  cv_att <- list(v = v, group = group)

  new_rset(splits = split_objs$splits,
           ids = split_objs[, grepl("^id", names(split_objs))],
           attrib = cv_att,
           subclass = c("group_vfold_cv", "rset"))
}

group_vfold_splits <- function(data, group, v = NULL) {
  uni_groups <- unique(getElement(data, group))
  max_v <- length(uni_groups)

  if (is.null(v)) {
    v <- max_v
  } else {
    if (v > max_v)
      stop("`v` should be less than ", max_v, call. = FALSE)
  }
  data_ind <- data.frame(..index = 1:nrow(data), ..group = getElement(data, group))
  keys <- data.frame(..group = uni_groups)

  n <- nrow(keys)
  keys$..folds <- sample(rep(1:v, length.out = n))
  data_ind <- data_ind %>%
    full_join(keys, by = "..group") %>%
    arrange(..index)
  indices <- split_unnamed(data_ind$..index, data_ind$..folds)
  indices <- lapply(indices, vfold_complement, n = nrow(data))
  split_objs <-
    purrr::map(indices,
               make_splits,
               data = data,
               class = "group_vfold_split")
  tibble::tibble(splits = split_objs,
                 id = names0(length(split_objs), "Resample"))
}

#' @export
print.group_vfold_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("group_vfold_cv", "rset"))]
  print(x, ...)
}

#' Group V-Fold Cross-Validation
#'
#' Group V-fold cross-validation creates splits of the data based
#'  on some grouping variable (which may have more than a single row
#'  associated with it). The function can create as many splits as
#'  there are unique values of the grouping variable or it can
#'  create a smaller set of splits where more than one value is left
#'  out at a time. A common use of this kind of resampling is when you have
#'  repeated measures of the same subject.
#'
#' @inheritParams vfold_cv
#' @param v The number of partitions of the data set. If left as `NULL`, `v`
#'  will be set to the number of unique values in the group.
#' @inheritParams make_groups
#' @export
#' @return A tibble with classes `group_vfold_cv`,
#'  `rset`, `tbl_df`, `tbl`, and `data.frame`.
#'  The results include a column for the data split objects and an
#'  identification variable.
#' @examplesIf rlang::is_installed("modeldata")
#' data(Sacramento, package = "modeldata")
#'
#' set.seed(123)
#' group_vfold_cv(Sacramento, group = city, v = 5)
#'
#' @export
group_vfold_cv <- function(data, group = NULL, v = NULL, ...) {
  if (!missing(group)) {
    group <- tidyselect::vars_select(names(data), !!enquo(group))
    if (length(group) == 0) {
      group <- NULL
    }
  }

  if (is.null(group) || !is.character(group) || length(group) != 1) {
    rlang::abort(
      "`group` should be a single character value for the column that will be used for splitting."
    )
  }
  if (!any(names(data) == group)) {
    rlang::abort("`group` should be a column in `data`.")
  }

  split_objs <- group_vfold_splits(data = data, group = group, v = v)

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  # Update `v` if not supplied directly
  if (is.null(v)) {
    v <- length(split_objs$splits)
  }

  ## Save some overall information

  cv_att <- list(v = v, group = group)

  new_rset(
    splits = split_objs$splits,
    ids = split_objs[, grepl("^id", names(split_objs))],
    attrib = cv_att,
    subclass = c("group_vfold_cv", "rset")
  )
}

group_vfold_splits <- function(data, group, v = NULL) {

  group <- getElement(data, group)
  max_v <- length(unique(group))

  if (is.null(v)) {
    v <- max_v
  } else {
    check_v(v = v, max_v = max_v, rows = "rows", call = rlang::caller_env())
  }

  indices <- make_groups(data, group, v)
  indices <- lapply(indices, default_complement, n = nrow(data))
  split_objs <-
    purrr::map(indices,
      make_splits,
      data = data,
      class = "group_vfold_split"
    )
  tibble::tibble(
    splits = split_objs,
    id = names0(length(split_objs), "Resample")
  )
}

#' Leave-One-Out Cross-Validation
#'
#' Leave-one-out (LOO) cross-validation uses one data point in the original
#'  set as the assessment data and all other data points as the analysis set. A
#'  LOO resampling set has as many resamples as rows in the original data set.
#' @inheritParams vfold_cv
#' @return An tibble with classes `loo_cv`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and
#'  one column called `id` that has a character string with the resample
#'  identifier.
#' @examples
#' loo_cv(mtcars)
#' @export
loo_cv <- function(data, ...) {
  check_dots_empty()
  split_objs <- vfold_splits(data = data, v = nrow(data), prevent_loo = FALSE)
  split_objs <-
    list(
      splits = map(split_objs$splits, change_class),
      id = paste0("Resample", seq_along(split_objs$id))
    )

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    subclass = c("loo_cv", "rset")
  )
}

change_class <- function(x) {
  class(x) <- c("rsplit", "loo_split")
  x
}

test_that("Loooooo", {
  loo1 <- loo_cv(dat1)
  expect_equal(nrow(loo1), nrow(dat1))

  same_data <-
    purrr::map_lgl(loo1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  holdouts <-
    purrr::map_lgl(loo1$splits, function(x) {
      length(x$out_id) == 1
    })
  expect_true(all(holdouts))

  retained <-
    purrr::map_lgl(loo1$splits, function(x) {
      length(x$in_id) == (nrow(dat1) - 1)
    })
  expect_true(all(retained))
})

test_that("printing", {
  expect_snapshot(loo_cv(dat1))
})

test_that("rsplit labels", {
  rs <- loo_cv(mtcars)
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

context("Leave-one-out CV")

library(testthat)
library(rsample)
library(purrr)

dat1 <- data.frame(a = 1:10, b = letters[1:10])

test_that('Loooooo', {

  loo1 <- loo_cv(dat1)
  expect_equal(nrow(loo1), nrow(dat1))

  same_data <-
    map_lgl(loo1$splits, function(x)
      all.equal(x$data, dat1))
  expect_true(all(same_data))

  holdouts <-
    map_lgl(loo1$splits, function(x)
      length(x$out_id) == 1)
  expect_true(all(holdouts))

  retained <-
    map_lgl(loo1$splits, function(x)
      length(x$in_id) == (nrow(dat1) - 1))
  expect_true(all(retained))
})

test_that('printing', {
  expect_output(print(loo_cv(dat1)))
})

test_that('rsplit labels', {
  rs <- loo_cv(mtcars)
  all_labs <- map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

#' V-Fold Cross-Validation
#'
#' V-fold cross-validation (also known as k-fold cross-validation) randomly
#'  splits the data into V groups of roughly equal size (called "folds"). A
#'  resample of the analysis data consists of V-1 of the folds while the
#'  assessment set contains the final fold. In basic V-fold cross-validation
#'  (i.e. no repeats), the number of resamples is equal to V.
#' @details With more than one repeat, the basic V-fold cross-validation is
#'  conducted each time. For example, if three repeats are used with `v = 10`,
#'  there are a total of 30 splits: three groups of 10 that are generated
#'  separately.
#' @template strata_details
#' @inheritParams make_strata
#' @param data A data frame.
#' @param v The number of partitions of the data set.
#' @param repeats The number of times to repeat the V-fold partitioning.
#' @param strata A variable in `data` (single character or name) used to conduct
#'  stratified sampling. When not `NULL`, each resample is created within the
#'  stratification variable. Numeric `strata` are binned into quartiles.
#' @param ... Not currently used.
#' @export
#' @return A tibble with classes `vfold_cv`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and
#'  one or more identification variables. For a single repeat, there will be
#'  one column called `id` that has a character string with the fold identifier.
#'  For repeats, `id` is the repeat number and an additional column called `id2`
#'  that contains the fold information (within repeat).

#' @examples
#' vfold_cv(mtcars, v = 10)
#' vfold_cv(mtcars, v = 10, repeats = 2)
#'
#' library(purrr)
#' data(wa_churn, package = "modeldata")
#'
#' set.seed(13)
#' folds1 <- vfold_cv(wa_churn, v = 5)
#' map_dbl(
#'   folds1$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' folds2 <- vfold_cv(wa_churn, strata = churn, v = 5)
#' map_dbl(
#'   folds2$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' folds3 <- vfold_cv(wa_churn, strata = tenure, breaks = 6, v = 5)
#' map_dbl(
#'   folds3$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#' @export
vfold_cv <- function(data, v = 10, repeats = 1,
                     strata = NULL, breaks = 4, pool = 0.1, ...) {
  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) strata <- NULL
  }

  strata_check(strata, data)

  if (repeats == 1) {
    split_objs <- vfold_splits(
      data = data, v = v,
      strata = strata, breaks = breaks, pool = pool
    )
  } else {
    for (i in 1:repeats) {
      tmp <- vfold_splits(data = data, v = v, strata = strata, pool = pool)
      tmp$id2 <- tmp$id
      tmp$id <- names0(repeats, "Repeat")[i]
      split_objs <- if (i == 1) {
        tmp
      } else {
        rbind(split_objs, tmp)
      }
    }
  }

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  ## Save some overall information

  cv_att <- list(v = v, repeats = repeats, strata = !is.null(strata))

  new_rset(
    splits = split_objs$splits,
    ids = split_objs[, grepl("^id", names(split_objs))],
    attrib = cv_att,
    subclass = c("vfold_cv", "rset")
  )
}


vfold_splits <- function(data, v = 10, strata = NULL, breaks = 4, pool = 0.1) {

  n <- nrow(data)
  check_v(v, n, call = rlang::caller_env())

  if (is.null(strata)) {
    folds <- sample(rep(1:v, length.out = n))
    idx <- seq_len(n)
    indices <- split_unnamed(idx, folds)
  } else {
    stratas <- tibble::tibble(
      idx = 1:n,
      strata = make_strata(getElement(data, strata),
        breaks = breaks,
        pool = pool
      )
    )
    stratas <- split_unnamed(stratas, stratas$strata)
    stratas <- purrr::map(stratas, add_vfolds, v = v)
    stratas <- dplyr::bind_rows(stratas)
    indices <- split_unnamed(stratas$idx, stratas$folds)
  }

  indices <- lapply(indices, default_complement, n = n)

  split_objs <- purrr::map(indices, make_splits, data = data, class = "vfold_split")
  tibble::tibble(
    splits = split_objs,
    id = names0(length(split_objs), "Fold")
  )
}

add_vfolds <- function(x, v) {
  x$folds <- sample(rep(1:v, length.out = nrow(x)))
  x
}

check_v <- function(v, max_v, rows = "rows", call = rlang::caller_env()) {
  if (!is.numeric(v) || length(v) != 1 || v < 0) {
    rlang::abort("`v` must be a single positive integer", call = call)
  } else if (v > max_v) {
    rlang::abort(
      glue::glue("The number of {rows} is less than `v = {v}`"), call = call
    )
  }
}

#' V-Fold Cross-Validation
#'
#' V-fold cross-validation randomly splits the data into V groups of roughly
#'  equal size (called "folds"). A resample of the analysis data consisted of
#'  V-1 of the folds while the assessment set contains the final fold. In basic
#'  V-fold cross-validation (i.e. no repeats), the number of resamples is equal
#'  to V.
#' @details
#' The `strata` argument causes the random sampling to be conducted *within
#'  the stratification variable*. This can help ensure that the number of data
#'  points in the analysis data is equivalent to the proportions in the original
#'  data set. (Strata below 10% of the total are pooled together.)
#' When more than one repeat is requested, the basic V-fold cross-validation
#'  is conducted each time. For example, if three repeats are used with `v =
#'  10`, there are a total of 30 splits which as three groups of 10 that are
#'  generated separately.
#' @param data A data frame.
#' @param v The number of partitions of the data set.
#' @param repeats The number of times to repeat the V-fold partitioning.
#' @param strata A variable that is used to conduct stratified sampling to
#'  create the folds. This could be a single character value or a variable name
#'  that corresponds to a variable that exists in the data frame.
#' @param breaks A single number giving the number of bins desired to stratify
#'  a numeric stratification variable.
#' @param ... Not currently used.
#' @export
#' @return A tibble with classes `vfold_cv`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and
#'  one or more identification variables. For a single repeat, there will be
#'  one column called `id` that has a character string with the fold identifier.
#'  For repeats, `id` is the repeat number and an additional column called `id2`
#'  that contains the fold information (within repeat).

#' @examples
#' vfold_cv(mtcars, v = 10)
#' vfold_cv(mtcars, v = 10, repeats = 2)
#'
#' library(purrr)
#' data(wa_churn, package = "modeldata")
#'
#' set.seed(13)
#' folds1 <- vfold_cv(wa_churn, v = 5)
#' map_dbl(folds1$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#'
#' set.seed(13)
#' folds2 <- vfold_cv(wa_churn, strata = "churn", v = 5)
#' map_dbl(folds2$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#'
#' set.seed(13)
#' folds3 <- vfold_cv(wa_churn, strata = "tenure", breaks = 6, v = 5)
#' map_dbl(folds3$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#' @export
vfold_cv <- function(data, v = 10, repeats = 1, strata = NULL, breaks = 4, ...) {

  if(!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if(length(strata) == 0) strata <- NULL
  }

  strata_check(strata, names(data))

  if (repeats == 1) {
    split_objs <- vfold_splits(data = data, v = v, strata = strata, breaks = breaks)
  } else {
    for (i in 1:repeats) {
      tmp <- vfold_splits(data = data, v = v, strata = strata)
      tmp$id2 <- tmp$id
      tmp$id <- names0(repeats, "Repeat")[i]
      split_objs <- if (i == 1)
        tmp
      else
        rbind(split_objs, tmp)
    }
  }

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  ## Save some overall information

  cv_att <- list(v = v, repeats = repeats, strata = !is.null(strata))

  new_rset(splits = split_objs$splits,
           ids = split_objs[, grepl("^id", names(split_objs))],
           attrib = cv_att,
           subclass = c("vfold_cv", "rset"))
}

# Get the indices of the analysis set from the assessment set
vfold_complement <- function(ind, n) {
  list(analysis = setdiff(1:n, ind),
       assessment = ind)
}

vfold_splits <- function(data, v = 10, strata = NULL, breaks = 4) {
  if (!is.numeric(v) || length(v) != 1)
    stop("`v` must be a single integer.", call. = FALSE)

  n <- nrow(data)
  if (is.null(strata)) {
    folds <- sample(rep(1:v, length.out = n))
    idx <- seq_len(n)
    indices <- split_unnamed(idx, folds)
  } else {
    stratas <- tibble::tibble(idx = 1:n,
                              strata = make_strata(getElement(data, strata),
                                                   breaks = breaks))
    stratas <- split_unnamed(stratas, stratas$strata)
    stratas <- purrr::map(stratas, add_vfolds, v = v)
    stratas <- dplyr::bind_rows(stratas)
    indices <- split_unnamed(stratas$idx, stratas$folds)
  }

  indices <- lapply(indices, vfold_complement, n = n)

  split_objs <- purrr::map(indices, make_splits, data = data, class = "vfold_split")
  tibble::tibble(splits = split_objs,
                 id = names0(length(split_objs), "Fold"))
}

add_vfolds <- function(x, v) {
  x$folds <- sample(rep(1:v, length.out = nrow(x)))
  x
}

#' @export
print.vfold_cv <- function(x, ...) {
  cat("# ", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("vfold_cv", "rset"))]
  print(x, ...)
}

#' Rolling Origin Forecast Resampling
#'
#' This resampling method is useful when the data set has a strong time
#'  component. The resamples are not random and contain data points that are
#'  consecutive values. The function assumes that the original data set are
#'  sorted in time order.
#' @details The main options, `initial` and `assess`, control the number of
#'  data points from the original data that are in the analysis and assessment
#'  set, respectively. When `cumulative = TRUE`, the analysis set will grow as
#'  resampling continues while the assessment set size will always remain
#'  static.
#' `skip` enables the function to not use every data point in the resamples.
#'  When `skip = 0`, the resampling data sets will increment by one position.
#'  Suppose that the rows of a data set are consecutive days. Using `skip = 6`
#'  will make the analysis data set to operate on *weeks* instead of days. The
#'  assessment set size is not affected by this option.
#' @seealso
#' [sliding_window()], [sliding_index()], and [sliding_period()] for additional
#' time based resampling functions.
#' @inheritParams vfold_cv
#' @param initial The number of samples used for analysis/modeling in the
#'  initial resample.
#' @param assess The number of samples used for each assessment resample.
#' @param cumulative A logical. Should the analysis resample grow beyond the
#'  size specified by `initial` at each resample?.
#' @param skip A integer indicating how many (if any) _additional_ resamples
#'  to skip to thin the total amount of data points in the analysis resample.
#' See the example below.
#' @param lag A value to include a lag between the assessment
#'  and analysis set. This is useful if lagged predictors will be used
#'  during training and testing.
#' @export
#' @return An tibble with classes `rolling_origin`, `rset`, `tbl_df`, `tbl`,
#'  and `data.frame`. The results include a column for the data split objects
#'  and a column called `id` that has a character string with the resample
#'  identifier.
#' @examples
#' set.seed(1131)
#' ex_data <- data.frame(row = 1:20, some_var = rnorm(20))
#' dim(rolling_origin(ex_data))
#' dim(rolling_origin(ex_data, skip = 2))
#' dim(rolling_origin(ex_data, skip = 2, cumulative = FALSE))
#'
#' # You can also roll over calendar periods by first nesting by that period,
#' # which is especially useful for irregular series where a fixed window
#' # is not useful. This example slides over 5 years at a time.
#' library(dplyr)
#' library(tidyr)
#' data(drinks, package = "modeldata")
#'
#' drinks_annual <- drinks %>%
#'   mutate(year = as.POSIXlt(date)$year + 1900) %>%
#'   nest(data = c(-year))
#'
#' multi_year_roll <- rolling_origin(drinks_annual, cumulative = FALSE)
#'
#' analysis(multi_year_roll$splits[[1]])
#' assessment(multi_year_roll$splits[[1]])
#' @export
rolling_origin <- function(data, initial = 5, assess = 1,
                           cumulative = TRUE, skip = 0, lag = 0, ...) {
  n <- nrow(data)

  if (n < initial + assess) {
    rlang::abort(
      "There should be at least ", initial + assess, " rows in `data`."
    )
  }

  if (!is.numeric(lag) | !(lag %% 1 == 0)) {
    rlang::abort("`lag` must be a whole number.")
  }

  if (lag > initial) {
    rlang::abort(
      "`lag` must be less than or equal to the number of training observations."
    )
  }

  stops <- seq(initial, (n - assess), by = skip + 1)
  starts <- if (!cumulative) {
    stops - initial + 1
  } else {
    starts <- rep(1, length(stops))
  }

  in_ind <- mapply(seq, starts, stops, SIMPLIFY = FALSE)
  out_ind <-
    mapply(seq, stops + 1 - lag, stops + assess, SIMPLIFY = FALSE)
  indices <- mapply(merge_lists, in_ind, out_ind, SIMPLIFY = FALSE)
  split_objs <-
    purrr::map(indices, make_splits, data = data, class = "rof_split")
  split_objs <- list(
    splits = split_objs,
    id = names0(length(split_objs), "Slice")
  )

  roll_att <- list(
    initial = initial,
    assess = assess,
    cumulative = cumulative,
    skip = skip,
    lag = lag
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = roll_att,
    subclass = c("rolling_origin", "rset")
  )
}

#' Leave-One-Out Cross-Validation
#'
#' Leave-one-out (LOO) cross-validation uses one data point in the original
#'  set as the assessment data and all other data points as the analysis set. A
#'  LOO resampling set has as many resamples as rows in the original data set.
#' @inheritParams vfold_cv
#' @return An tibble with classes `loo_cv`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and
#'  one column called `id` that has a character string with the resample
#'  identifier.
#' @examples
#' loo_cv(mtcars)
#' @export
loo_cv <- function(data, ...) {
  split_objs <- vfold_splits(data = data, v = nrow(data))
  split_objs <-
    list(splits = map(split_objs$splits, change_class),
         id = paste0("Resample", seq_along(split_objs$id)))

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)

  new_rset(splits = split_objs$splits,
           ids = split_objs$id,
           subclass = c("loo_cv", "rset"))
}

#' @export
print.loo_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("loo_cv", "rset"))]
  print(x, ...)
}

change_class <- function(x) {
  class(x) <- c("rsplit", "loo_split")
  x
}

#' Create or Modify Stratification Variables
#'
#' This function can create strata from numeric data and make non-numeric data
#'  more conducive for stratification.
#'
#' @details
#' For numeric data, if the number of unique levels is less than
#' `nunique`, the data are treated as categorical data.
#'
#' For categorical inputs, the function will find levels of `x` than
#'   occur in the data with percentage less than `pool`. The values from
#'   these groups will be randomly assigned to the remaining strata (as will
#'   data points that have missing values in `x`).
#'
#' For numeric data with more unique values than `nunique`, the data
#'   will be converted to being categorical based on percentiles of the data.
#'   The percentile groups will have no more than 20 percent of the data in
#'   each group. Again, missing values in `x` are randomly assigned
#'   to groups.
#'
#' @param x An input vector.
#' @param breaks A single number giving the number of bins desired to stratify a
#'   numeric stratification variable.
#' @param nunique An integer for the number of unique value threshold in the
#'   algorithm.
#' @param pool A proportion of data used to determine if a particular group is
#'   too small and should be pooled into another group. We do not recommend
#'   decreasing this argument below its default of 0.1 because of the dangers
#'   of stratifying groups that are too small.
#' @param depth An integer that is used to determine the best number of
#'   percentiles that should be used. The number of bins are based on
#'   `min(5, floor(n / depth))` where `n = length(x)`.
#'   If `x` is numeric, there must be at least 40 rows in the data set
#'   (when `depth = 20`) to conduct stratified sampling.
#'
#' @export
#' @return  A factor vector.
#' @examples
#' set.seed(61)
#' x1 <- rpois(100, lambda = 5)
#' table(x1)
#' table(make_strata(x1))
#'
#' set.seed(554)
#' x2 <- rpois(100, lambda = 1)
#' table(x2)
#' table(make_strata(x2))
#'
#' # small groups are randomly assigned
#' x3 <- factor(x2)
#' table(x3)
#' table(make_strata(x3))
#'
#' x4 <- rep(LETTERS[1:7], c(37, 26, 3, 7, 11, 10, 2))
#' table(x4)
#' table(make_strata(x4))
#' table(make_strata(x4, pool = 0.1))
#' table(make_strata(x4, pool = 0.0))
#'
#' # not enough data to stratify
#' x5 <- rnorm(20)
#' table(make_strata(x5))
#'
#' set.seed(483)
#' x6 <- rnorm(200)
#' quantile(x6, probs = (0:10) / 10)
#' table(make_strata(x6, breaks = 10))
#' @export
make_strata <- function(x, breaks = 4, nunique = 5, pool = .1, depth = 20) {
  default_pool <- 0.1
  num_vals <- unique(stats::na.omit(x))
  n <- length(x)
  if (length(num_vals) <= nunique | is.character(x) | is.factor(x)) {
    x <- factor(x)
    xtab <- sort(table(x))
    pcts <- xtab / n

    ## This should really be based on some combo of rate and number.
    if (all(pcts < pool)) {
      cli_warn(c(
        "Too little data to stratify.",
        "*" = "Resampling will be unstratified."
      ))
      return(factor(rep("strata1", n)))
    }

    if (pool < default_pool & any(pcts < default_pool)) {
      cli_warn(c(
        "Stratifying groups that make up {round(100 * pool)}% of the data may 
        be statistically risky.",
        i = "Consider increasing {.arg pool} to at least 0.1."
      ))
    }

    ## Small groups will be randomly allocated to stratas at end
    ## These should probably go into adjacent groups but this works for now
    if (any(pcts < pool)) {
      x[x %in% names(pcts)[pcts < pool]] <- NA
    }
    ## The next line will also relevel the data if `x` was a factor
    out <- factor(as.character(x))
  } else {
    if (breaks < 2) {
      cli_warn(c(
        "The bins specified by {.arg breaks} must be >=2.",
        "*" = "Resampling will be unstratified."
      ))
      return(factor(rep("strata1", n)))
    } else if (floor(n / breaks) < depth) {
      cli_warn(c(
        "The number of observations in each quantile is below the recommended 
        threshold of {depth}.",
        "*" = "Stratification will use {floor(n / depth)} breaks instead."
      ))
    }
    breaks <- min(breaks, floor(n / depth))
    if (breaks < 2) {
      cli_warn(c(
        "Too little data to stratify.",
        "*" = "Resampling will be unstratified."
      ))
      return(factor(rep("strata1", n)))
    }
    pctls <- quantile(x, probs = (0:breaks) / breaks, na.rm = TRUE)
    pctls <- unique(pctls)
    out <- cut(x, breaks = pctls, include.lowest = TRUE)
  }

  num_miss <- sum(is.na(x))
  if (num_miss > 0) {
    out[is.na(x)] <- sample(levels(out), size = num_miss, replace = TRUE)
  }

  out
}

context("Strata constructor")

library(testthat)
library(rsample)
library(purrr)

test_that('simple numerics', {
  set.seed(13333)
  x1 <- rnorm(1000)
  str1a <- make_strata(x1)
  tab1a <- table(str1a)
  expect_equal(as.vector(tab1a), rep(250, 4))

  str1b <- expect_warning(make_strata(x1, depth = 500), "2 breaks instead")
  tab1b <- table(str1b)
  expect_equal(as.vector(tab1b), rep(500, 2))
})

test_that('simple character', {
  x2 <- factor(rep(LETTERS[1:5], each = 50))
  str2a <- make_strata(x2)
  expect_equal(table(str2a, dnn = ""), table(x2, dnn = ""))
})

test_that('bad data', {
  x3 <- factor(rep(LETTERS[1:15], each = 50))
  expect_warning(make_strata(x3))
  expect_warning(make_strata(mtcars$mpg))
})

#' Determine the Assessment Samples
#'
#' This method and function help find which data belong in the analysis and
#' assessment sets.
#'
#' Given an `rsplit` object, `complement()` will determine which
#'   of the data rows are contained in the assessment set. To save space,
#'   many of the `rsplit` objects will not contain indices for the
#'   assessment split.
#'
#' @param x An `rsplit` object
#' @param ... Not currently used
#' @return A integer vector.
#' @seealso [populate()]
#' @examples
#' set.seed(28432)
#' fold_rs <- vfold_cv(mtcars)
#' head(fold_rs$splits[[1]]$in_id)
#' fold_rs$splits[[1]]$out_id
#' complement(fold_rs$splits[[1]])
#' @export
complement <- function(x, ...) {
  UseMethod("complement")
}

#' @export
#' @rdname complement
complement.rsplit <- function(x, ...) {
  if (!is_missing_out_id(x)) {
    return(x$out_id)
  } else {
    (1:nrow(x$data))[-unique(x$in_id)]
  }
}
#' @export
#' @rdname complement
complement.rof_split <- function(x, ...) {
  get_stored_out_id(x)
}
#' @export
#' @rdname complement
complement.sliding_window_split <- function(x, ...) {
  get_stored_out_id(x)
}
#' @export
#' @rdname complement
complement.sliding_index_split <- function(x, ...) {
  get_stored_out_id(x)
}
#' @export
#' @rdname complement
complement.sliding_period_split <- function(x, ...) {
  get_stored_out_id(x)
}

get_stored_out_id <- function(x) {
  out_id <- x$out_id

  if (length(out_id) == 0L) {
    return(out_id)
  }

  if (all(is.na(out_id))) {
    rlang::abort("Cannot derive the assessment set for this type of resampling.")
  }

  out_id
}

#' @export
#' @rdname complement
complement.apparent_split <- function(x, ...) {
  if (!is_missing_out_id(x)) {
    return(x$out_id)
  } else {
    1:nrow(x$data)
  }
}

#' @export
complement.default <- function(x, ...) {
  cls <- paste0("'", class(x), "'", collapse = ", ")
  rlang::abort(
    paste("No `complement()` method for this class(es)", cls)
  )
}

# Get the indices of the analysis set from the assessment set
default_complement <- function(ind, n) {
  list(
    analysis = setdiff(1:n, ind),
    assessment = unique(ind)
  )
}


#' Add Assessment Indices
#'
#' Many `rsplit` and `rset` objects do not contain indicators for
#'   the assessment samples. `populate()` can be used to fill the slot
#'   for the appropriate indices.
#' @param x A `rsplit` and `rset` object.
#' @param ... Not currently used
#' @return An object of the same kind with the integer indices.
#' @examples
#' set.seed(28432)
#' fold_rs <- vfold_cv(mtcars)
#'
#' fold_rs$splits[[1]]$out_id
#' complement(fold_rs$splits[[1]])
#'
#' populate(fold_rs$splits[[1]])$out_id
#'
#' fold_rs_all <- populate(fold_rs)
#' fold_rs_all$splits[[1]]$out_id
#' @export
populate <- function(x, ...) UseMethod("populate")

#' @export
populate.rsplit <- function(x, ...) {
  x$out_id <- complement(x, ...)
  x
}

#' @export
populate.rset <- function(x, ...) {
  x$splits <- map(x$splits, populate)
  x
}

## This will remove the assessment indices from an rsplit object
rm_out <- function(x) {
  x$out_id <- NA
  x
}

is_missing_out_id <- function(x) {
  identical(x$out_id, NA)
}

#' Constructors for split objects
#' @export
#' @examples
#' df <- data.frame(
#'   year = 1900:1999,
#'   value = 10 + 8*1900:1999 + runif(100L, 0, 100)
#' )
#' split_from_indices <- make_splits(
#'   x = list(analysis = which(df$year <= 1980),
#'            assessment = which(df$year > 1980)),
#'   data = df
#' )
#' split_from_data_frame <- make_splits(
#'   x = df[df$year <= 1980,],
#'   assessment = df[df$year > 1980,]
#' )
#' identical(split_from_indices, split_from_data_frame)
make_splits <- function(x, ...) {
  UseMethod("make_splits")
}

#' @rdname make_splits
#' @param x A list of integers with names "analysis" and "assessment", or a
#' data frame of analysis or training data.
#' @export
make_splits.default <- function(x, ...) {
  rlang::abort("There is no method available to make an rsplit from `x`.")
}

#' @rdname make_splits
#' @param data A data frame.
#' @param class An optional class to give the object.
#' @param ... Not currently used.
#' @export
make_splits.list <- function(x, data, class = NULL, ...) {
  rlang::check_dots_empty()
  res <- rsplit(data, x$analysis, x$assessment)
  if (!is.null(class)) {
    res <- add_class(res, class)
  }
  res
}

#' @rdname make_splits
#' @param assessment A data frame of assessment or testing data, which can be empty.
#' @export
make_splits.data.frame <- function(x, assessment, ...) {
  rlang::check_dots_empty()
  if (nrow(x) == 0) {
    rlang::abort("The analysis set must contain at least one row.")
  }

  ind_analysis <- seq_len(nrow(x))
  if (nrow(assessment) == 0) {
    ind_assessment <- integer()
  } else {
    if (!identical(colnames(x), colnames(assessment))) {
      rlang::abort("The analysis and assessment sets must have the same columns.")
    }
    ind_assessment <- nrow(x) + seq_len(nrow(assessment))
  }

  data <- bind_rows(x, assessment)
  ind <- list(
    analysis = ind_analysis,
    assessment = ind_assessment
  )

  make_splits(ind, data)
}

merge_lists <- function(a, b) list(analysis = a, assessment = b)

dim_rset <- function(x, ...) {
  check_dots_empty()
  dims <- purrr::map(x$splits, dim)
  dims <- do.call("rbind", dims)
  dims <- tibble::as_tibble(dims)
  id_cols <- grep("(^id$)|(^id[1-9]$)", colnames(x), value = TRUE)
  for (i in seq_along(id_cols)) {
    dims[id_cols[i]] <- getElement(x, id_cols[i])
  }
  dims
}

names0 <- function(num, prefix = "x") {
  if (num == 0L) {
    return(character())
  }
  ind <- format(1:num)
  ind <- gsub(" ", "0", ind)
  paste0(prefix, ind)
}

add_class <- function(x, cls) {
  class(x) <- c(cls, class(x))
  x
}

strata_check <- function(strata, data) {
  if (!is.null(strata)) {
    if (!is.character(strata) | length(strata) != 1) {
      rlang::abort("`strata` should be a single name or character value.")
    }
    if (inherits(data[, strata], "Surv")) {
      rlang::abort("`strata` cannot be a `Surv` object. Use the time or event variable directly.")
    }
    if (!(strata %in% names(data))) {
      rlang::abort(strata, " is not in `data`.")
    }
  }
  invisible(NULL)
}

split_unnamed <- function(x, f) {
  out <- split(x, f)
  unname(out)
}

#' Obtain a identifier for the resamples
#'
#' This function returns a hash (or NA) for an attribute that is created when
#' the `rset` was initially constructed. This can be used to compare with other
#' resampling objects to see if they are the same.
#' @param x An `rset` or `tune_results` object.
#' @param ... Not currently used.
#' @return A character value or `NA_character_` if the object was created prior
#' to `rsample` version 0.1.0.
#' @rdname get_fingerprint
#' @aliases .get_fingerprint
#' @examples
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(2)
#' .get_fingerprint(vfold_cv(mtcars))
#'
#' set.seed(1)
#' .get_fingerprint(vfold_cv(mtcars, repeats = 2))
#' @export
.get_fingerprint <- function(x, ...) {
  UseMethod(".get_fingerprint")
}

#' @export
#' @rdname get_fingerprint
.get_fingerprint.default <- function(x, ...) {
  cls <- paste0("'", class(x), "'", collapse = ", ")
  rlang::abort(
    paste("No `.get_fingerprint()` method for this class(es)", cls)
  )
}

#' @export
#' @rdname get_fingerprint
.get_fingerprint.rset <- function(x, ...) {
  check_dots_empty()
  att <- attributes(x)
  if (any(names(att) == "fingerprint")) {
    res <- att$fingerprint
  } else {
    res <- NA_character_
  }
  res
}

#' Reverse the analysis and assessment sets
#'
#' This functions "swaps" the analysis and assessment sets of either a single
#' `rsplit` or all `rsplit`s in the `splits` column of an `rset` object.
#'
#' @param x An `rset` or `rsplit` object.
#' @param ... Not currently used.
#'
#' @return An object of the same class as `x`
#'
#' @examples
#' set.seed(123)
#' starting_splits <- vfold_cv(mtcars, v = 3)
#' reverse_splits(starting_splits)
#' reverse_splits(starting_splits$splits[[1]])
#'
#' @rdname reverse_splits
#' @export
reverse_splits <- function(x, ...) {
  UseMethod("reverse_splits")
}

#' @rdname reverse_splits
#' @export
reverse_splits.default <- function(x, ...) {
  rlang::abort(
    "`x` must be either an `rsplit` or an `rset` object"
  )
}

#' @rdname reverse_splits
#' @export
reverse_splits.permutations <- function(x, ...) {
  rlang::abort(
    "Permutations cannot have their splits reversed"
  )
}

#' @rdname reverse_splits
#' @export
reverse_splits.perm_split <- reverse_splits.permutations

#' @rdname reverse_splits
#' @export
reverse_splits.rsplit <- function(x, ...) {

  rlang::check_dots_empty()

  out_splits <- list(
    analysis = as.integer(x, data = "assessment"),
    assessment = as.integer(x, data = "analysis")
  )
  out_splits <- make_splits(out_splits, x$data)
  class(out_splits) <- class(x)
  out_splits

}

#' @rdname reverse_splits
#' @export
reverse_splits.rset <- function(x, ...) {

  rlang::check_dots_empty()

  x$splits <- purrr::map(x$splits, reverse_splits)

  x
}

#' "Reshuffle" an rset to re-generate a new rset with the same parameters
#'
#' This function re-generates an rset object, using the same arguments used
#' to generate the original.
#'
#' @param rset The `rset` object to be reshuffled
#'
#' @return An rset of the same class as `rset`.
#'
#' @examples
#' set.seed(123)
#' (starting_splits <- group_vfold_cv(mtcars, cyl, v = 3))
#' reshuffle_rset(starting_splits)
#'
#' @export
reshuffle_rset <- function(rset) {
  if (!inherits(rset, "rset")) {
    rlang::abort("`rset` must be an rset object")
  }

  if (inherits(rset, "manual_rset")) {
    rlang::abort("`manual_rset` objects cannot be reshuffled")
  }

  # non-random classes is defined below
  if (any(non_random_classes %in% class(rset))) {
    cls <- class(rset)[[1]]
    rlang::warn(
      glue::glue("`reshuffle_rset()` will return an identical rset when called on {cls} objects")
    )
    if ("validation_set" %in% class(rset)) {
      return(rset)
    }
  }

  arguments <- attributes(rset)
  useful_arguments <- names(formals(arguments$class[[1]]))
  useful_arguments <- arguments[useful_arguments]
  useful_arguments <- useful_arguments[!is.na(names(useful_arguments))]
  if (identical(useful_arguments$strata, FALSE)) {
    useful_arguments$strata <- NULL
  } else if (identical(useful_arguments$strata, TRUE)) {
    rlang::abort(
      "Cannot reshuffle this rset (`attr(rset, 'strata')` is `TRUE`, not a column identifier)",
      i = "If the original object was created with an older version of rsample, try recreating it with the newest version of the package"
    )
  }

  do.call(
    arguments$class[[1]],
    c(list(data = rset$splits[[1]]$data), useful_arguments)
  )
}

non_random_classes <- c(
  "sliding_index",
  "sliding_period",
  "sliding_window",
  "rolling_origin",
  "validation_time_split",
  "validation_set"
)

#' Retrieve individual rsplits objects from an rset
#'
#' @param x The `rset` object to retrieve an rsplit from.
#' @param index An integer indicating which rsplit to retrieve: `1` for the
#' rsplit in the first row of the rset, `2` for the second, and so on.
#' @param ... Not currently used.
#'
#' @return The rsplit object in row `index` of `rset`
#'
#' @examples
#' set.seed(123)
#' (starting_splits <- group_vfold_cv(mtcars, cyl, v = 3))
#' get_rsplit(starting_splits, 1)
#'
#' @rdname get_rsplit
#' @export
get_rsplit <- function(x, index, ...) {
  UseMethod("get_rsplit")
}

#' @rdname get_rsplit
#' @export
get_rsplit.rset <- function(x, index, ...) {
  rlang::check_dots_empty()

  n_rows <- nrow(x)

  acceptable_index <- length(index) == 1 &&
    rlang::is_integerish(index) &&
    index > 0 &&
    index <= n_rows

  if (!acceptable_index) {
    msg <- ifelse(
      length(index) != 1,
      glue::glue("Index was of length {length(index)}."),
      glue::glue("A value of {index} was provided.")
      )

    rlang::abort(
      c(
        glue::glue("`index` must be a length-1 integer between 1 and {n_rows}."),
        x = msg
      )
    )
  }

  x$splits[[index]]
}

#' @rdname get_rsplit
#' @export
get_rsplit.default <- function(x, index, ...) {
  cls <- paste0("'", class(x), "'", collapse = ", ")
  rlang::abort(
    paste("No `get_rsplit()` method for this class(es)", cls)
  )
}

#' Create or Modify Stratification Variables
#'
#' For stratified resampling, this function can create strata from numeric data
#'   and also make non-numeric data more conducive to be used for
#'   stratification.

#' @details
#' For numeric data, if the number of unique levels is less than
#' `nunique`, the data are treated as categorical data.
#'
#' For categorical inputs, the function will find levels of `x` than
#'   occur in the data with percentage less than `pool`. The values from
#'   these groups will be randomly assigned to the remaining strata (as will
#'   data points that have missing values in `x`).
#'
#' For numeric data with more unique values than `nunique`, the data
#'   will be converted to being categorical based on percentiles of the data.
#'   The percentile groups will have no more than 20 percent of the data in
#'   each group. Again, missing values in `x` are randomly assigned
#'   to groups.
#'
#' @param x An input vector.
#' @param breaks A single number giving the number of bins desired to stratify a
#'   numeric stratification variable.
#' @param nunique An integer for the number of unique value threshold in the
#'   algorithm.
#' @param pool A proportion of data used to determine if a particular group is
#'   too small and should be pooled into another group.
#' @param depth An integer that is used to determine the best number of
#'   percentiles that should be used. The number of bins are based on
#'   `min(5, floor(n / depth))` where `n = length(x)`.
#'   If `x` is numeric, there must be at least 40 rows in the data set
#'   (when `depth = 20`) to conduct stratified sampling.
#'
#' @export
#' @return  A factor vector.
#' @examples
#' set.seed(61)
#' x1 <- rpois(100, lambda = 5)
#' table(x1)
#' table(make_strata(x1))
#'
#' set.seed(554)
#' x2 <- rpois(100, lambda = 1)
#' table(x2)
#' table(make_strata(x2))
#'
#' # small groups are randomly assigned
#' x3 <- factor(x2)
#' table(x3)
#' table(make_strata(x3))
#'
#' # `oilType` data from `caret`
#' x4 <- rep(LETTERS[1:7], c(37, 26, 3, 7, 11, 10, 2))
#' table(x4)
#' table(make_strata(x4))
#' table(make_strata(x4, pool = 0.1))
#' table(make_strata(x4, pool = 0.0))
#'
#' # not enough data to stratify
#' x5 <- rnorm(20)
#' table(make_strata(x5))
#'
#' set.seed(483)
#' x6 <- rnorm(200)
#' quantile(x6, probs = (0:10)/10)
#' table(make_strata(x6, breaks = 10))
#' @export
make_strata <- function(x, breaks = 4, nunique = 5, pool = .1, depth = 20) {
  num_vals <- unique(x)
  n <- length(x)
  num_miss <- sum(is.na(x))
  if (length(num_vals) <= nunique | is.character(x) | is.factor(x)) {
    x <- factor(x)
    xtab <- sort(table(x))
    pcts <- xtab / n

    ## This should really be based on some combo of rate and number.
    if (all(pcts < pool)) {
      warning("Too little data to stratify. Unstratified resampling ",
              "will be used.",
              call. = FALSE)
      return(factor(rep("strata1", n)))
    }
    ## Small groups will be randomly allocated to stratas at end
    ## These should probably go into adjacent groups but this works for now
    if (any(pcts < pool))
      x[x %in% names(pcts)[pcts < pool]] <- NA
    ## The next line will also relevel the data if `x` was a factor
    out <- factor(as.character(x))
  } else {
    if (floor(n / breaks) < depth) {
      warning(paste0("The number of observations in each quantile is ",
              "below the recommended threshold of ", depth, ". Stratification ",
              "will be done with ", floor(n/depth), " breaks instead."),
              call. = FALSE)
    }
    breaks <- min(breaks, floor(n/depth))
    if (breaks < 2) {
      warning("Too little data to stratify. Unstratified resampling ",
              "will be used.",
              call. = FALSE)
      return(factor(rep("strata1", n)))
    }
    pctls <- quantile(x, probs = (0:breaks) / breaks)
    pctls <- unique(pctls)
    out <- cut(x, breaks = pctls, include.lowest = TRUE)
  }

  num_miss <- sum(is.na(x))
  if (num_miss > 0)
    out[is.na(x)] <- sample(levels(out), size = num_miss, replace = TRUE)

  out
}

# Bootstrap confidence interval code

# ------------------------------------------------------------------------------
# helpers


check_rset <- function(x, app = TRUE) {
  if (!inherits(x, "bootstraps"))
    stop("`.data` should be an `rset` object generated from `bootstraps()`",
         call. = FALSE)

  if (app) {
    if(x %>% dplyr::filter(id == "Apparent") %>% nrow() != 1)
      stop("Please set `apparent = TRUE` in `bootstraps()` function",
           call. = FALSE)
  }
  invisible(NULL)
}


stat_fmt_err <- paste("`statistics` should select a list column of tidy results.")
stat_nm_err <- paste("The tibble in `statistics` should have columns for",
                     "'estimate' and 'term`")
std_exp <- c("std.error", "robust.se")

check_tidy_names <- function(x, std_col) {
  # check for proper columns
  if (sum(colnames(x) == "estimate") != 1) {
    stop(stat_nm_err, call. = FALSE)
  }
  if (sum(colnames(x) == "term") != 1) {
    stop(stat_nm_err, call. = FALSE)
  }
  if (std_col) {
    std_candidates <- colnames(x) %in% std_exp
    if (sum(std_candidates) != 1) {
      stop("`statistics` should select a single column for the standard ",
           "error.", call. = FALSE)
    }
  }
  invisible(TRUE)
}

check_tidy <- function(x, std_col = FALSE) {
  if (!is.list(x)) {
    stop(stat_fmt_err, call. = FALSE)
  }

  # convert to data frame from list
  has_id <- any(names(x) == "id")
  if (has_id) {
    list_cols <- names(x)[map_lgl(x, is_list)]
    x <- try(tidyr::unnest(x, cols = list_cols), silent = TRUE)
  } else {
    x <- try(map_dfr(x, ~ .x), silent = TRUE)
  }

  if (inherits(x, "try-error")) {
    stop(stat_fmt_err, call. = FALSE)
  }

  check_tidy_names(x, std_col)

  if (std_col) {
    std_candidates <- colnames(x) %in% std_exp
    std_candidates <- colnames(x)[std_candidates]
    if (has_id) {
      x <-
        dplyr::select(x, term, estimate, id, tidyselect::one_of(std_candidates)) %>%
        mutate(id = (id == "Apparent")) %>%
        setNames(c("term", "estimate", "orig", "std_err"))
    } else {
      x <-
        dplyr::select(x, term, estimate, tidyselect::one_of(std_candidates)) %>%
        setNames(c("term", "estimate", "std_err"))
    }

  } else {
    if (has_id) {
      x <-
        dplyr::select(x, term, estimate, id) %>%
        mutate(orig = (id == "Apparent")) %>%
        dplyr::select(-id)
    } else {
      x <- dplyr::select(x, term, estimate)
    }
  }

  x
}


get_p0 <- function(x, alpha = 0.05) {
  orig <- x %>%
    group_by(term) %>%
    dplyr::filter(orig) %>%
    dplyr::select(term, theta_0 = estimate) %>%
    ungroup()
  x %>%
    dplyr::filter(!orig) %>%
    inner_join(orig, by = "term") %>%
    group_by(term) %>%
    summarize(p0 = mean(estimate <= theta_0, na.rm = TRUE)) %>%
    mutate(Z0 = stats::qnorm(p0),
           Za = stats::qnorm(1 - alpha / 2))
}

new_stats <- function(x, lo, hi) {
  res <- as.numeric(quantile(x, probs = c(lo, hi), na.rm = TRUE))
  tibble(.lower = min(res), .estimate = mean(x, na.rm = TRUE), .upper = max(res))
}

has_dots <- function(x) {
  nms <- names(formals(x))
  if (!any(nms == "...")) {
    stop("`.fn` must have an argument `...`.", call. = FALSE)
  }
  invisible(NULL)
}

check_num_resamples <- function(x, B = 1000) {
  x <-
    x %>%
    dplyr::group_by(term) %>%
    dplyr::summarize(n = sum(!is.na(estimate))) %>%
    dplyr::filter(n < B)

  if (nrow(x) > 0) {
    terms <- paste0("`", x$term, "`")
    msg <-
      paste0(
        "Recommend at least ", B, " non-missing bootstrap resamples for ",
        ifelse(length(terms) > 1, "terms: ", "term "),
        paste0(terms, collapse = ", "),
        "."
      )
    warning(msg, call. = FALSE)
  }
  invisible(NULL)
}

# ------------------------------------------------------------------------------
# percentile code


pctl_single <- function(stats, alpha = 0.05) {

  if (all(is.na(stats)))
    stop("All statistics have missing values..", call. = FALSE)

  if (!is.numeric(stats))
    stop("`stats` must be a numeric vector.", call. = FALSE)

  # stats is a numeric vector of values
  ci <- stats %>% quantile(probs = c(alpha / 2, 1 - alpha / 2), na.rm = TRUE)

  # return a tibble with .lower, .estimate, .upper
  res <- tibble(
    .lower = min(ci),
    .estimate = mean(stats, na.rm = TRUE),
    .upper = max(ci),
    .alpha = alpha,
    .method = "percentile"
  )
  res
}

#' Bootstrap confidence intervals
#' @description
#' Calculate bootstrap confidence intervals using various methods.
#' @param .data A data frame containing the bootstrap resamples created using
#'  `bootstraps()`. For t- and BCa-intervals, the `apparent` argument
#'  should be set to `TRUE`. Even if the `apparent` argument is set to
#'  `TRUE` for the percentile method, the apparent data is never used in calculating
#'  the percentile confidence interval.
#' @param statistics An unquoted column name or `dplyr` selector that identifies
#'  a single column in the data set that contains the individual bootstrap
#'  estimates. This can be a list column of tidy tibbles (that contains columns
#'  `term` and `estimate`) or a simple numeric column. For t-intervals, a
#'  standard tidy column (usually called `std.err`) is required.
#'  See the examples below.
#' @param alpha Level of significance
#' @return Each function returns a tibble with columns `.lower`,
#'  `.estimate`, `.upper`, `.alpha`, `.method`, and `term`.
#'  `.method` is the type of interval (eg. "percentile",
#'  "student-t", or "BCa"). `term` is the name of the estimate. Note
#'  the `.estimate` returned from `int_pctl()`
#'  is the mean of the estimates from the bootstrap resamples
#'  and not the estimate from the apparent model.
#' @details Percentile intervals are the standard method of
#'  obtaining confidence intervals but require thousands of
#'  resamples to be accurate. T-intervals may need fewer
#'  resamples but require a corresponding variance estimate.
#'  Bias-corrected and accelerated intervals require the original function
#'  that was used to create the statistics of interest and are
#'  computationally taxing.
#'
#' @references Davison, A., & Hinkley, D. (1997). _Bootstrap Methods and their
#'  Application_. Cambridge: Cambridge University Press.
#'  doi:10.1017/CBO9780511802843
#'
#' @examples
#' \donttest{
#' library(broom)
#' library(dplyr)
#' library(purrr)
#' library(tibble)
#'
#' lm_est <- function(split, ...) {
#'   lm(mpg ~ disp + hp, data = analysis(split)) %>%
#'     tidy()
#' }
#'
#' set.seed(52156)
#' car_rs <-
#'   bootstraps(mtcars, 500, apparent = TRUE) %>%
#'   mutate(results = map(splits, lm_est))
#'
#' int_pctl(car_rs, results)
#' int_t(car_rs, results)
#' int_bca(car_rs, results, .fn = lm_est)
#'
#' # putting results into a tidy format
#' rank_corr <- function(split) {
#'   dat <- analysis(split)
#'   tibble(
#'     term = "corr",
#'     estimate = cor(dat$sqft, dat$price, method = "spearman"),
#'     # don't know the analytical std.err so no t-intervals
#'     std.err = NA_real_
#'   )
#' }
#'
#' set.seed(69325)
#' data(Sacramento, package = "modeldata")
#' bootstraps(Sacramento, 1000, apparent = TRUE) %>%
#'   mutate(correlations = map(splits, rank_corr)) %>%
#'   int_pctl(correlations)
#' }
#' @export
int_pctl <- function(.data, statistics, alpha = 0.05) {

  check_rset(.data, app = FALSE)
  if (length(alpha) != 1 || !is.numeric(alpha)) {
    abort("`alpha` must be a single numeric value.")
  }

  .data <- .data %>% dplyr::filter(id != "Apparent")

  column_name <- tidyselect::vars_select(names(.data), !!rlang::enquo(statistics))
  if (length(column_name) != 1) {
    stop(stat_fmt_err, call. = FALSE)
  }
  stats <- .data[[column_name]]
  stats <- check_tidy(stats, std_col = FALSE)

  check_num_resamples(stats, B = 1000)

  vals <- stats %>%
    dplyr::group_by(term) %>%
    dplyr::do(pctl_single(.$estimate, alpha = alpha)) %>%
    dplyr::ungroup()
  vals

}

# ------------------------------------------------------------------------------
# t interval code

t_single <- function(stats, std_err, is_orig, alpha = 0.05) {
  # stats is a numeric vector of values
  # vars is a numeric vector of variances
  # return a tibble with .lower, .estimate, .upper
  # which_orig is the index of stats and std_err that has the original result

  if (all(is.na(stats)))
    stop("All statistics have missing values.", call. = FALSE)

  if (!is.logical(is_orig) || any(is.na(is_orig))) {
    stop("`is_orig` should be a logical column the same length as `stats` ",
         "with no missing values.", call. = FALSE)
  }
  if (length(stats) != length(std_err) && length(stats) != length(is_orig)) {
    stop("`stats`, `std_err`, and `is_orig` should have the same length.",
         call. = FALSE)
  }
  if (sum(is_orig) != 1) {
    stop("The original statistic must be in a single row.", call. = FALSE)
  }

  theta_obs   <- stats[is_orig]
  std_err_obs <- std_err[is_orig]

  stats   <- stats[!is_orig]
  std_err <- std_err[!is_orig]

  z_dist <-
    (stats - theta_obs) / std_err

  z_pntl <-
    quantile(z_dist, probs = c(alpha / 2, 1 - (alpha) / 2), na.rm = TRUE)

  ci <- theta_obs - z_pntl * std_err_obs

  tibble(
    .lower = min(ci),
    .estimate = mean(stats, na.rm = TRUE),
    .upper = max(ci),
    .alpha = alpha,
    .method = "student-t"
  )
}


#' @rdname int_pctl
#' @export
int_t <- function(.data, statistics, alpha = 0.05) {

  check_rset(.data)
  if (length(alpha) != 1 || !is.numeric(alpha)) {
    abort("`alpha` must be a single numeric value.")
  }

  column_name <- tidyselect::vars_select(names(.data), !!enquo(statistics))
  if (length(column_name) != 1) {
    stop(stat_fmt_err, call. = FALSE)
  }
  stats <- .data %>% dplyr::select(!!column_name, id)
  stats <- check_tidy(stats, std_col = TRUE)

  check_num_resamples(stats, B = 500)

  vals <-
    stats %>%
    dplyr::group_by(term) %>%
    dplyr::do(t_single(.$estimate, .$std_err, .$orig, alpha = alpha)) %>%
    dplyr::ungroup()
  vals
}


# ----------------------------------------------------------------

bca_calc <- function(stats, orig_data, alpha = 0.05, .fn, ...) {

  # TODO check per term
  if (all(is.na(stats$estimate))) {
    stop("All statistics have missing values.", call. = FALSE)
  }

  ### Estimating Z0 bias-correction
  bias_corr_stats <- get_p0(stats, alpha = alpha)

  # need the original data frame here
  loo_rs <- loo_cv(orig_data)

  # We can't be sure what we will get back from the analysis function.
  # To test, we run on the first LOO data set and see if it is a vector or df
  loo_test <- try(rlang::exec(.fn, loo_rs$splits[[1]], ...), silent = TRUE)
  if (inherits(loo_test, "try-error")) {
    cat("Running `.fn` on the LOO resamples produced an error:\n")
    print(loo_test)
    stop("`.fn` failed.", call. = FALSE)
  }

  loo_res <- furrr::future_map_dfr(loo_rs$splits, .fn, ...)

  loo_estimate <-
    loo_res %>%
    dplyr::group_by(term) %>%
    dplyr::summarize(loo = mean(estimate, na.rm = TRUE)) %>%
    dplyr::inner_join(loo_res, by = "term")  %>%
    dplyr::group_by(term) %>%
    dplyr::summarize(
      cubed = sum((loo - estimate)^3),
      squared = sum((loo - estimate)^2)
    ) %>%
    dplyr::ungroup() %>%
    dplyr::inner_join(bias_corr_stats, by = "term") %>%
    dplyr::mutate(
      a = cubed/(6 * (squared^(3 / 2))),
      Zu = (Z0 + Za) / ( 1 - a * (Z0 + Za)) + Z0,
      Zl = (Z0 - Za) / (1 - a * (Z0 - Za)) + Z0,
      lo = stats::pnorm(Zl, lower.tail = TRUE),
      hi = stats::pnorm(Zu, lower.tail = TRUE)
    )

  terms <- loo_estimate$term
  stats <- stats %>% dplyr::filter(!orig)
  for (i in seq_along(terms)) {
    tmp <- new_stats(stats$estimate[ stats$term == terms[i] ],
                     lo = loo_estimate$lo[i],
                     hi = loo_estimate$hi[i])
    tmp$term <- terms[i]
    if (i == 1) {
      ci_bca <- tmp
    } else {
      ci_bca <- bind_rows(ci_bca, tmp)
    }
  }
  ci_bca <-
    ci_bca %>%
    dplyr::select(term, .lower, .estimate, .upper) %>%
    dplyr::mutate(
      .alpha = alpha,
      .method = "BCa"
    )
}


#' @rdname int_pctl
#' @param .fn A function to calculate statistic of interest. The
#'  function should take an `rsplit` as the first argument and the `...` are
#'  required.
#' @param ... Arguments to pass to `.fn`.
#' @references \url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}
#' @export
int_bca <- function(.data, statistics, alpha = 0.05, .fn, ...) {

  check_rset(.data)
  if (length(alpha) != 1 || !is.numeric(alpha)) {
    abort("`alpha` must be a single numeric value.")
  }

  has_dots(.fn)

  column_name <- tidyselect::vars_select(names(.data), !!enquo(statistics))
  if (length(column_name) != 1) {
    stop(stat_fmt_err, call. = FALSE)
  }
  stats <- .data %>% dplyr::select(!!column_name, id)
  stats <- check_tidy(stats)

  check_num_resamples(stats, B = 1000)

  vals <- bca_calc(stats, .data$splits[[1]]$data, alpha = alpha, .fn = .fn, ...)
  vals
}

library(rsample)
library(testthat)
library(purrr)
library(tibble)
library(dplyr)
library(broom)

data("attrition", package = "modeldata")

context("Bootstrap intervals")

# ------------------------------------------------------------------------------

get_stats <- function(split, ...) {
  dat <- analysis(split)
  x <- dat[[1]]
  tibble(
    term = "mean",
    estimate = mean(x, na.rm = TRUE),
    std.error = sqrt(var(x, na.rm = TRUE)/sum(!is.na(x)))
  )
}

# ------------------------------------------------------------------------------

n <- 1000
mu <- 10
sigma <- 1

set.seed(888)
rand_nums <- rnorm(n, mu, sigma)
ttest <- tidy(t.test(rand_nums))
ttest_lower_conf <- tidy(t.test(rand_nums, conf.level = 0.8))
dat <- data.frame(x = rand_nums)

set.seed(456765)
bt_norm <-
  bootstraps(dat, times = 1000, apparent = TRUE) %>%
  dplyr::mutate(
    stats = map(splits,  ~ get_stats(.x))
  )

test_that('Bootstrap estimate of mean is close to estimate of mean from normal distribution',{
  skip_on_cran()
  single_pct_res <- int_pctl(bt_norm, stats)

  single_t_res <- int_t(bt_norm, stats)

  single_bca_res <- int_bca(bt_norm, stats, .fn = get_stats)

  single_bca_res_lower_conf <- int_bca(bt_norm, stats, .fn = get_stats, alpha = 0.2)

  expect_equal(ttest$conf.low,
               single_pct_res$.lower,
               tolerance = 0.001)
  expect_equal(unname(ttest$estimate),
               single_pct_res$.estimate,
               tolerance = 0.001)
  expect_equal(ttest$conf.high,
               single_pct_res$.upper,
               tolerance = 0.001)

  expect_equal(ttest$conf.low,
               single_t_res$.lower,
               tolerance = 0.001)
  expect_equal(unname(ttest$estimate),
               single_t_res$.estimate,
               tolerance = 0.001)
  expect_equal(ttest$conf.high,
               single_pct_res$.upper,
               tolerance = 0.001)

  expect_equal(ttest$conf.low,
               single_bca_res$.lower,
               tolerance = 0.001)
  expect_equal(unname(ttest$estimate),
               single_bca_res$.estimate,
               tolerance = 0.001)
  expect_equal(ttest$conf.high,
               single_bca_res$.upper,
               tolerance = 0.001)

  expect_equal(ttest_lower_conf$conf.low,
               single_bca_res_lower_conf$.lower,
               tolerance = 0.001)
  expect_equal(unname(ttest_lower_conf$estimate),
               single_bca_res_lower_conf$.estimate,
               tolerance = 0.001)
  expect_equal(ttest_lower_conf$conf.high,
               single_bca_res_lower_conf$.upper,
               tolerance = 0.001)
})

# ------------------------------------------------------------------------------

context("Wrapper Functions")

test_that("Wrappers -- selection of multiple variables works", {

  func <- function(split, ...) {
    lm(Age ~ HourlyRate + DistanceFromHome, data = analysis(split)) %>% tidy()
  }

  # generate boostrap resamples
  set.seed(888)
  bt_resamples <- bootstraps(attrition, times = 1000, apparent = TRUE) %>%
    mutate(res = map(splits, func))

  attrit_tidy <-
    lm(Age ~ HourlyRate + DistanceFromHome, data = attrition) %>%
    tidy(conf.int = TRUE) %>%
    dplyr::arrange(term)

  pct_res <-
    int_pctl(bt_resamples, res) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(pct_res$conf.low,  pct_res$.lower, tolerance = .01)
  expect_equal(pct_res$conf.high, pct_res$.upper, tolerance = .01)


  t_res <-
    int_t(bt_resamples, res) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(t_res$conf.low,  t_res$.lower, tolerance = .01)
  expect_equal(t_res$conf.high, t_res$.upper, tolerance = .01)


  bca_res <-
    int_bca(bt_resamples, res, .fn = func) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(bca_res$conf.low,  bca_res$.lower, tolerance = .01)
  expect_equal(bca_res$conf.high, bca_res$.upper, tolerance = .01)

})

# ------------------------------------------------------------------------------

context("boot_ci() Prompt Errors: Too Many NAs")

test_that('Upper & lower confidence interval does not contain NA', {

  bad_stats <- function(split, ...) {
    tibble(
      term = "mean",
      estimate = NA_real_,
      std.error = runif(1)
    )
  }

  set.seed(888)
  bt_resamples <- bootstraps(data.frame(x = 1:100), times = 1000, apparent = TRUE) %>%
    mutate(res = map(splits, bad_stats))

  expect_error(
    expect_warning(
      int_pctl(bt_resamples, res),
      "at least 1000 non-missing"
    ),
    "missing values"
  )

  expect_error(
    expect_warning(
      int_t(bt_resamples, res),
      "at least 1000 non-missing"
    ),
    "missing values"
  )

  expect_error(
    expect_warning(
      int_bca(bt_resamples, res, .fn = bad_stats),
      "at least 1000 non-missing"
    ),
    "missing values"
  )

})

# ------------------------------------------------------------------------------

context("boot_ci() Insufficient Number of Bootstrap Resamples")

set.seed(456765)
bt_small <-
  bootstraps(dat, times = 10, apparent = TRUE) %>%
  dplyr::mutate(
    stats = map(splits,  ~ get_stats(.x)),
    junk = 1:11
  )

test_that(
  "Sufficient replications needed to sufficiently reduce Monte Carlo sampling Error for BCa method", {

    expect_warning(int_pctl(bt_small, stats))
    expect_warning(int_t(bt_small, stats))
    expect_warning(int_bca(bt_small, stats, .fn = get_stats))

  }
)


context("boot_ci() Input Validation")


test_that("bad input", {
  expect_error(int_pctl(bt_small, id))
  expect_error(int_pctl(bt_small, junk))
  expect_error(int_pctl(bt_small, stats, alpha = c(0.05, 0.2)))
  expect_error(int_t(bt_small, stats, alpha = "potato"))
  expect_error(int_bca(bt_small, stats, alpha = 1:2, .fn = get_stats))


  bad_bt_norm <-
    bt_norm %>%
    mutate(stats = map(stats, ~ .x[, 1:2]))
  expect_error(int_t(bad_bt_norm, stats))

  expect_error(int_bca(bad_bt_norm, stats))

  no_dots <- function(split) {
    dat <- analysis(split)
    x <- dat[[1]]
    tibble(
      term = "mean",
      estimate = mean(x, na.rm = TRUE),
      std.error = sqrt(var(x, na.rm = TRUE)/sum(!is.na(x)))
    )
  }
  expect_error(
    int_bca(bt_norm, stats, .fn = no_dots),
    "must have an argument"
  )

  expect_error(int_pctl(as.data.frame(bt_norm), stats))
  expect_error(int_t(as.data.frame(bt_norm), stats))
  expect_error(int_bca(as.data.frame(bt_norm), stats, .fn = get_stats))

  expect_error(
    int_t(bt_norm %>% dplyr::filter(id != "Apparent"), stats)
    )
  expect_error(
    int_bca(bt_norm %>% dplyr::filter(id != "Apparent"), stats, .fn = get_stats)
  )

  poo <- function(x) {
    x$estimate <- "a"
    x
  }
  badder_bt_norm <-
    bt_norm %>%
    mutate(
      bad_term = map(stats, ~ .x %>% setNames(c("a", "estimate", "std.err"))),
      bad_est = map(stats, ~ .x %>% setNames(c("term", "b", "std.err"))),
      bad_err = map(stats, ~ .x %>% setNames(c("term", "estimate", "c"))),
      bad_num = map(stats, ~ poo(.x))
      )
  expect_error(int_pctl(badder_bt_norm, bad_term))
  expect_error(int_t(badder_bt_norm, bad_err))
  expect_error(int_bca(badder_bt_norm, bad_est, .fn = get_stats))
  expect_error(int_pctl(badder_bt_norm, bad_num))

})

test_that("Bootstrap estimate of mean is close to estimate of mean from normal distribution", {
  skip_if_not_installed("broom")
  skip_on_cran()

  set.seed(888)
  rand_nums <- rnorm(n = 1000, mean = 10, sd = 1)
  dat <- data.frame(x = rand_nums)

  set.seed(456765)
  bt_norm <- bootstraps(dat, times = 1000, apparent = TRUE) |>
    dplyr::mutate(
      stats = purrr::map(splits, \(.x) get_stats(.x))
    )

  ttest <- broom::tidy(t.test(rand_nums))
  ttest_lower_conf <- broom::tidy(t.test(rand_nums, conf.level = 0.8))
  single_pct_res <- int_pctl(bt_norm, stats)

  single_t_res <- int_t(bt_norm, stats)

  single_bca_res <- int_bca(bt_norm, stats, .fn = get_stats)

  single_bca_res_lower_conf <- int_bca(
    bt_norm,
    stats,
    .fn = get_stats,
    alpha = 0.2
  )

  expect_equal(ttest$conf.low, single_pct_res$.lower, tolerance = 0.001)
  expect_equal(
    unname(ttest$estimate),
    single_pct_res$.estimate,
    tolerance = 0.001
  )
  expect_equal(ttest$conf.high, single_pct_res$.upper, tolerance = 0.001)

  expect_equal(ttest$conf.low, single_t_res$.lower, tolerance = 0.001)
  expect_equal(
    unname(ttest$estimate),
    single_t_res$.estimate,
    tolerance = 0.001
  )
  expect_equal(ttest$conf.high, single_pct_res$.upper, tolerance = 0.001)

  expect_equal(ttest$conf.low, single_bca_res$.lower, tolerance = 0.001)
  expect_equal(
    unname(ttest$estimate),
    single_bca_res$.estimate,
    tolerance = 0.001
  )
  expect_equal(ttest$conf.high, single_bca_res$.upper, tolerance = 0.001)

  expect_equal(
    ttest_lower_conf$conf.low,
    single_bca_res_lower_conf$.lower,
    tolerance = 0.001
  )
  expect_equal(
    unname(ttest_lower_conf$estimate),
    single_bca_res_lower_conf$.estimate,
    tolerance = 0.001
  )
  expect_equal(
    ttest_lower_conf$conf.high,
    single_bca_res_lower_conf$.upper,
    tolerance = 0.001
  )
})

# ------------------------------------------------------------------------------

test_that("Wrappers -- selection of multiple variables works", {
  skip_if_not_installed("broom")
  skip_if_not_installed("modeldata")
  data("attrition", package = "modeldata")
  func <- function(split, ...) {
    lm(Age ~ HourlyRate + DistanceFromHome, data = analysis(split)) |> tidy()
  }

  # generate boostrap resamples
  set.seed(888)
  bt_resamples <- bootstraps(attrition, times = 1000, apparent = TRUE) |>
    mutate(res = purrr::map(splits, func))

  attrit_tidy <-
    lm(Age ~ HourlyRate + DistanceFromHome, data = attrition) |>
    tidy(conf.int = TRUE) |>
    dplyr::arrange(term)

  pct_res <-
    int_pctl(bt_resamples, res) |>
    inner_join(attrit_tidy, by = "term")
  expect_equal(pct_res$conf.low, pct_res$.lower, tolerance = .01)
  expect_equal(pct_res$conf.high, pct_res$.upper, tolerance = .01)

  t_res <-
    int_t(bt_resamples, res) |>
    inner_join(attrit_tidy, by = "term")
  expect_equal(t_res$conf.low, t_res$.lower, tolerance = .01)
  expect_equal(t_res$conf.high, t_res$.upper, tolerance = .01)

  bca_res <-
    int_bca(bt_resamples, res, .fn = func) |>
    inner_join(attrit_tidy, by = "term")
  expect_equal(bca_res$conf.low, bca_res$.lower, tolerance = .01)
  expect_equal(bca_res$conf.high, bca_res$.upper, tolerance = .01)
})

# ------------------------------------------------------------------------------

test_that("Upper & lower confidence interval does not contain NA", {
  bad_stats <- function(split, ...) {
    tibble(
      term = "mean",
      estimate = NA_real_,
      std.error = runif(1)
    )
  }

  set.seed(888)
  bt_resamples <- bootstraps(
    data.frame(x = 1:100),
    times = 1000,
    apparent = TRUE
  ) |>
    mutate(res = purrr::map(splits, bad_stats))

  expect_snapshot(
    int_pctl(bt_resamples, res),
    error = TRUE
  )

  expect_snapshot(
    int_t(bt_resamples, res),
    error = TRUE
  )

  expect_snapshot(
    int_bca(bt_resamples, res, .fn = bad_stats),
    error = TRUE
  )
})

# ------------------------------------------------------------------------------

test_that("Sufficient replications needed to sufficiently reduce Monte Carlo sampling Error for BCa method", {
  set.seed(888)
  rand_nums <- rnorm(n = 1000, mean = 10, sd = 1)
  dat <- data.frame(x = rand_nums)
  set.seed(456765)
  bt_small <-
    bootstraps(dat, times = 10, apparent = TRUE) |>
    dplyr::mutate(
      stats = purrr::map(splits, \(.x) get_stats(.x)),
      junk = 1:11
    )

  expect_snapshot(int_pctl(bt_small, stats))
  expect_snapshot(int_t(bt_small, stats))
})

test_that("Sufficient replications needed to sufficiently reduce Monte Carlo sampling Error for BCa method", {
  skip("#539 message about loading purrr in the snapshot in R CMD check hard")
  # unskip this by moving the expectation back into the test_that block above

  set.seed(888)
  rand_nums <- rnorm(n = 1000, mean = 10, sd = 1)
  dat <- data.frame(x = rand_nums)
  set.seed(456765)
  bt_small <-
    bootstraps(dat, times = 10, apparent = TRUE) |>
    dplyr::mutate(
      stats = purrr::map(splits, \(.x) get_stats(.x)),
      junk = 1:11
    )

  expect_snapshot(int_bca(bt_small, stats, .fn = get_stats))
})

test_that("bad input", {
  set.seed(888)
  rand_nums <- rnorm(n = 1000, mean = 10, sd = 1)
  dat <- data.frame(x = rand_nums)
  set.seed(456765)
  bt_small <-
    bootstraps(dat, times = 10, apparent = TRUE) |>
    dplyr::mutate(
      stats = purrr::map(splits, \(.x) get_stats(.x)),
      junk = 1:11
    )

  expect_snapshot(error = TRUE, {
    int_pctl(bt_small, id)
  })
  expect_snapshot(error = TRUE, {
    int_pctl(bt_small, junk)
  })
  expect_snapshot(error = TRUE, {
    int_pctl(bt_small, stats, alpha = c(0.05, 0.2))
  })
  expect_snapshot(error = TRUE, {
    int_t(bt_small, stats, alpha = "potato")
  })
  expect_snapshot(error = TRUE, {
    int_bca(bt_small, stats, alpha = 1:2, .fn = get_stats)
  })
  expect_snapshot(error = TRUE, {
    int_pctl(vfold_cv(mtcars))
  })
  expect_snapshot(error = TRUE, {
    int_t(vfold_cv(mtcars))
  })
  expect_snapshot(error = TRUE, {
    int_bca(vfold_cv(mtcars))
  })

  set.seed(888)
  rand_nums <- rnorm(n = 1000, mean = 10, sd = 1)
  dat <- data.frame(x = rand_nums)

  set.seed(456765)
  bt_norm <- bootstraps(dat, times = 1000, apparent = TRUE) |>
    dplyr::mutate(
      stats = purrr::map(splits, \(.x) get_stats(.x))
    )

  bad_bt_norm <-
    bt_norm |>
    mutate(stats = purrr::map(stats, \(.x) .x[, 1:2]))
  expect_snapshot(error = TRUE, {
    int_t(bad_bt_norm, stats)
  })

  no_dots <- function(split) {
    dat <- analysis(split)
    x <- dat[[1]]
    tibble(
      term = "mean",
      estimate = mean(x, na.rm = TRUE),
      std.error = sqrt(var(x, na.rm = TRUE) / sum(!is.na(x)))
    )
  }
  expect_snapshot(error = TRUE, {
    int_bca(bt_norm, stats, .fn = no_dots)
  })

  expect_snapshot(error = TRUE, {
    int_pctl(as.data.frame(bt_norm), stats)
  })
  expect_snapshot(error = TRUE, {
    int_t(as.data.frame(bt_norm), stats)
  })

  expect_snapshot(error = TRUE, {
    int_bca(as.data.frame(bt_norm), stats, .fn = get_stats)
  })

  poo <- function(x) {
    x$estimate <- "a"
    x
  }
  badder_bt_norm <-
    bt_norm |>
    mutate(
      bad_term = purrr::map(
        stats,
        \(.x) .x |> setNames(c("a", "estimate", "std.err"))
      ),
      bad_est = purrr::map(
        stats,
        \(.x) .x |> setNames(c("term", "b", "std.err"))
      ),
      bad_err = purrr::map(
        stats,
        \(.x) .x |> setNames(c("term", "estimate", "c"))
      ),
      bad_num = purrr::map(stats, \(.x) poo(.x))
    )
  expect_snapshot(error = TRUE, {
    int_pctl(badder_bt_norm, bad_term)
  })
  expect_snapshot(error = TRUE, {
    int_t(badder_bt_norm, bad_err)
  })
  expect_snapshot(error = TRUE, {
    int_bca(badder_bt_norm, bad_est, .fn = get_stats)
  })
  expect_snapshot(error = TRUE, {
    int_pctl(badder_bt_norm, bad_num)
  })
})

test_that("checks for apparent bootstrap", {
  rs_boot <- bootstraps(mtcars, times = 10, apparent = FALSE)
  expect_snapshot(error = TRUE, {
    int_t(rs_boot)
  })
  expect_snapshot(error = TRUE, {
    int_bca(rs_boot)
  })
})

test_that("checks input for statistics", {
  dat <- data.frame(x = rnorm(n = 1000, mean = 10, sd = 1))
  rs_boot <- bootstraps(dat, times = 10, apparent = TRUE)

  rs_boot_missing_term <- rs_boot |>
    dplyr::mutate(
      stats = purrr::map(1:11, \(.x) tibble(estimate = 1))
    )
  expect_snapshot(error = TRUE, {
    int_t(rs_boot_missing_term, stats)
  })

  rs_boot_missing_estimate <- rs_boot |>
    dplyr::mutate(
      stats = purrr::map(1:11, \(.x) tibble(term = 1))
    )
  expect_snapshot(error = TRUE, {
    int_t(rs_boot_missing_estimate, stats)
  })

  rs_boot_missing_std_err <- rs_boot |>
    dplyr::mutate(
      stats = purrr::map(1:11, \(.x) tibble(term = 1, estimate = 2))
    )
  expect_snapshot(error = TRUE, {
    int_t(rs_boot_missing_std_err, stats)
  })
})

# ------------------------------------------------------------------------------

test_that("compute intervals with additional grouping terms", {
  skip_if_not_installed("broom")

  lm_coefs <- function(dat) {
    mod <- lm(mpg ~ I(1 / disp), data = dat)
    tidy(mod)
  }

  coef_by_engine_shape <- function(split, ...) {
    split |>
      analysis() |>
      dplyr::rename(.vs = vs) |>
      tidyr::nest(.by = .vs) |>
      dplyr::mutate(coefs = map(data, lm_coefs)) |>
      dplyr::select(-data) |>
      tidyr::unnest(coefs)
  }

  set.seed(270)
  boot_rs <-
    bootstraps(mtcars, 1000, apparent = TRUE) |>
    dplyr::mutate(results = purrr::map(splits, coef_by_engine_shape))

  pctl_res <- int_pctl(boot_rs, results)
  t_res <- int_t(boot_rs, results)
  bca_res <- int_bca(boot_rs, results, .fn = coef_by_engine_shape)

  exp_ptype <-
    tibble::tibble(
      term = character(0),
      .vs = numeric(0),
      .lower = numeric(0),
      .estimate = numeric(0),
      .upper = numeric(0),
      .alpha = numeric(0),
      .method = character(0)
    )

  expect_equal(pctl_res[0, ], exp_ptype)
  expect_equal(t_res[0, ], exp_ptype)
  expect_equal(bca_res[0, ], exp_ptype)

  exp_combos <-
    tibble::tribble(
      ~term,         ~.vs,
      "(Intercept)",    0,
      "(Intercept)",    1,
      "I(1/disp)",      0,
      "I(1/disp)",      1
    )

  group_patterns <- function(x) {
    dplyr::distinct(x, term, .vs) |>
      dplyr::arrange(term, .vs)
  }

  expect_equal(group_patterns(pctl_res), exp_combos)
  expect_equal(group_patterns(t_res), exp_combos)
  expect_equal(group_patterns(bca_res), exp_combos)
})

n <- 1000
mu <- 10
sigma <- 1

set.seed(888)
rand_nums <- rnorm(n, mu, sigma)
ttest <- broom::tidy(t.test(rand_nums))
ttest_lower_conf <- broom::tidy(t.test(rand_nums, conf.level = 0.8))
dat <- data.frame(x = rand_nums)

set.seed(456765)
bt_norm <-
  bootstraps(dat, times = 1000, apparent = TRUE) %>%
  dplyr::mutate(
    stats = purrr::map(splits, ~ get_stats(.x))
  )

test_that("Bootstrap estimate of mean is close to estimate of mean from normal distribution", {
  skip_on_cran()
  single_pct_res <- int_pctl(bt_norm, stats)

  single_t_res <- int_t(bt_norm, stats)

  single_bca_res <- int_bca(bt_norm, stats, .fn = get_stats)

  single_bca_res_lower_conf <- int_bca(bt_norm, stats, .fn = get_stats, alpha = 0.2)

  expect_equal(ttest$conf.low,
               single_pct_res$.lower,
               tolerance = 0.001
  )
  expect_equal(unname(ttest$estimate),
               single_pct_res$.estimate,
               tolerance = 0.001
  )
  expect_equal(ttest$conf.high,
               single_pct_res$.upper,
               tolerance = 0.001
  )

  expect_equal(ttest$conf.low,
               single_t_res$.lower,
               tolerance = 0.001
  )
  expect_equal(unname(ttest$estimate),
               single_t_res$.estimate,
               tolerance = 0.001
  )
  expect_equal(ttest$conf.high,
               single_pct_res$.upper,
               tolerance = 0.001
  )

  expect_equal(ttest$conf.low,
               single_bca_res$.lower,
               tolerance = 0.001
  )
  expect_equal(unname(ttest$estimate),
               single_bca_res$.estimate,
               tolerance = 0.001
  )
  expect_equal(ttest$conf.high,
               single_bca_res$.upper,
               tolerance = 0.001
  )

  expect_equal(ttest_lower_conf$conf.low,
               single_bca_res_lower_conf$.lower,
               tolerance = 0.001
  )
  expect_equal(unname(ttest_lower_conf$estimate),
               single_bca_res_lower_conf$.estimate,
               tolerance = 0.001
  )
  expect_equal(ttest_lower_conf$conf.high,
               single_bca_res_lower_conf$.upper,
               tolerance = 0.001
  )
})

# ------------------------------------------------------------------------------

test_that("Wrappers -- selection of multiple variables works", {
  data("attrition", package = "modeldata")
  func <- function(split, ...) {
    lm(Age ~ HourlyRate + DistanceFromHome, data = analysis(split)) %>% tidy()
  }

  # generate boostrap resamples
  set.seed(888)
  bt_resamples <- bootstraps(attrition, times = 1000, apparent = TRUE) %>%
    mutate(res = purrr::map(splits, func))

  attrit_tidy <-
    lm(Age ~ HourlyRate + DistanceFromHome, data = attrition) %>%
    tidy(conf.int = TRUE) %>%
    dplyr::arrange(term)

  pct_res <-
    int_pctl(bt_resamples, res) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(pct_res$conf.low, pct_res$.lower, tolerance = .01)
  expect_equal(pct_res$conf.high, pct_res$.upper, tolerance = .01)


  t_res <-
    int_t(bt_resamples, res) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(t_res$conf.low, t_res$.lower, tolerance = .01)
  expect_equal(t_res$conf.high, t_res$.upper, tolerance = .01)


  bca_res <-
    int_bca(bt_resamples, res, .fn = func) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(bca_res$conf.low, bca_res$.lower, tolerance = .01)
  expect_equal(bca_res$conf.high, bca_res$.upper, tolerance = .01)
})

# ------------------------------------------------------------------------------

test_that("Upper & lower confidence interval does not contain NA", {
  bad_stats <- function(split, ...) {
    tibble(
      term = "mean",
      estimate = NA_real_,
      std.error = runif(1)
    )
  }

  set.seed(888)
  bt_resamples <- bootstraps(data.frame(x = 1:100), times = 1000, apparent = TRUE) %>%
    mutate(res = purrr::map(splits, bad_stats))

  expect_snapshot(
    int_pctl(bt_resamples, res),
    error = TRUE
  )

  expect_snapshot(
    int_t(bt_resamples, res),
    error = TRUE
  )

  expect_snapshot(
    int_bca(bt_resamples, res, .fn = bad_stats),
    error = TRUE
  )
})

# ------------------------------------------------------------------------------

set.seed(456765)
bt_small <-
  bootstraps(dat, times = 10, apparent = TRUE) %>%
  dplyr::mutate(
    stats = purrr::map(splits, ~ get_stats(.x)),
    junk = 1:11
  )

test_that(
  "Sufficient replications needed to sufficiently reduce Monte Carlo sampling Error for BCa method",
  {
    expect_warning(int_pctl(bt_small, stats))
    expect_warning(int_t(bt_small, stats))
    expect_warning(int_bca(bt_small, stats, .fn = get_stats))
  }
)


test_that("bad input", {
  expect_error(int_pctl(bt_small, id))
  expect_error(int_pctl(bt_small, junk))
  expect_error(int_pctl(bt_small, stats, alpha = c(0.05, 0.2)))
  expect_error(int_t(bt_small, stats, alpha = "potato"))
  expect_error(int_bca(bt_small, stats, alpha = 1:2, .fn = get_stats))


  bad_bt_norm <-
    bt_norm %>%
    mutate(stats = purrr::map(stats, ~ .x[, 1:2]))
  expect_error(int_t(bad_bt_norm, stats))

  expect_error(int_bca(bad_bt_norm, stats))

  no_dots <- function(split) {
    dat <- analysis(split)
    x <- dat[[1]]
    tibble(
      term = "mean",
      estimate = mean(x, na.rm = TRUE),
      std.error = sqrt(var(x, na.rm = TRUE) / sum(!is.na(x)))
    )
  }
  expect_error(
    int_bca(bt_norm, stats, .fn = no_dots),
    "must have an argument"
  )

  expect_error(int_pctl(as.data.frame(bt_norm), stats))
  expect_error(int_t(as.data.frame(bt_norm), stats))
  expect_error(int_bca(as.data.frame(bt_norm), stats, .fn = get_stats))

  expect_error(
    int_t(bt_norm %>% dplyr::filter(id != "Apparent"), stats)
  )
  expect_error(
    int_bca(bt_norm %>% dplyr::filter(id != "Apparent"), stats, .fn = get_stats)
  )

  poo <- function(x) {
    x$estimate <- "a"
    x
  }
  badder_bt_norm <-
    bt_norm %>%
    mutate(
      bad_term = purrr::map(stats, ~ .x %>% setNames(c("a", "estimate", "std.err"))),
      bad_est = purrr::map(stats, ~ .x %>% setNames(c("term", "b", "std.err"))),
      bad_err = purrr::map(stats, ~ .x %>% setNames(c("term", "estimate", "c"))),
      bad_num = purrr::map(stats, ~ poo(.x))
    )
  expect_error(int_pctl(badder_bt_norm, bad_term))
  expect_error(int_t(badder_bt_norm, bad_err))
  expect_error(int_bca(badder_bt_norm, bad_est, .fn = get_stats))
  expect_error(int_pctl(badder_bt_norm, bad_num))
})


# ------------------------------------------------------------------------------

test_that("regression intervals", {
  skip_on_cran()

  expect_error(
    {
      set.seed(1)
      int_1 <- reg_intervals(mpg ~ disp + wt, data = mtcars)
    },
    regexp = NA
  )

  expect_equal(
    names(int_1),
    c("term", ".lower", ".estimate", ".upper", ".alpha", ".method")
  )

  expect_snapshot({
    skip_if(new_rng_snapshots)
    set.seed(123)
    int_2 <- reg_intervals(
      mpg ~ disp + wt,
      data = mtcars,
      filter = term == "wt",
      model_fn = "glm",
      keep_reps = TRUE
    )
    int_2
  })

  expect_equal(
    names(int_2),
    c("term", ".lower", ".estimate", ".upper", ".alpha", ".method", ".replicates")
  )
  expect_true(nrow(int_2) == 1)
  expect_true(all(int_2$term == "wt"))


  expect_error(
    reg_intervals(mpg ~ disp + wt, data = mtcars, model_fn = "potato"),
    "`model_fn` must be one of"
  )
  expect_error(
    reg_intervals(mpg ~ disp + wt, data = mtcars, type = "random"),
    "`type` must be one of"
  )
  expect_error(
    reg_intervals(mpg ~ disp + wt, data = mtcars, alpha = "a"),
    "must be a single numeric value"
  )
})

n <- 1000
mu <- 10
sigma <- 1

set.seed(888)
rand_nums <- rnorm(n, mu, sigma)
dat <- data.frame(x = rand_nums)

set.seed(456765)
bt_norm <-
  bootstraps(dat, times = 1000, apparent = TRUE) %>%
  dplyr::mutate(
    stats = purrr::map(splits, ~ get_stats(.x))
  )

test_that("Bootstrap estimate of mean is close to estimate of mean from normal distribution", {
  skip_if_not_installed("broom")
  skip_on_cran()
  ttest <- broom::tidy(t.test(rand_nums))
  ttest_lower_conf <- broom::tidy(t.test(rand_nums, conf.level = 0.8))
  single_pct_res <- int_pctl(bt_norm, stats)

  single_t_res <- int_t(bt_norm, stats)

  single_bca_res <- int_bca(bt_norm, stats, .fn = get_stats)

  single_bca_res_lower_conf <- int_bca(bt_norm, stats, .fn = get_stats, alpha = 0.2)

  expect_equal(ttest$conf.low,
               single_pct_res$.lower,
               tolerance = 0.001
  )
  expect_equal(unname(ttest$estimate),
               single_pct_res$.estimate,
               tolerance = 0.001
  )
  expect_equal(ttest$conf.high,
               single_pct_res$.upper,
               tolerance = 0.001
  )

  expect_equal(ttest$conf.low,
               single_t_res$.lower,
               tolerance = 0.001
  )
  expect_equal(unname(ttest$estimate),
               single_t_res$.estimate,
               tolerance = 0.001
  )
  expect_equal(ttest$conf.high,
               single_pct_res$.upper,
               tolerance = 0.001
  )

  expect_equal(ttest$conf.low,
               single_bca_res$.lower,
               tolerance = 0.001
  )
  expect_equal(unname(ttest$estimate),
               single_bca_res$.estimate,
               tolerance = 0.001
  )
  expect_equal(ttest$conf.high,
               single_bca_res$.upper,
               tolerance = 0.001
  )

  expect_equal(ttest_lower_conf$conf.low,
               single_bca_res_lower_conf$.lower,
               tolerance = 0.001
  )
  expect_equal(unname(ttest_lower_conf$estimate),
               single_bca_res_lower_conf$.estimate,
               tolerance = 0.001
  )
  expect_equal(ttest_lower_conf$conf.high,
               single_bca_res_lower_conf$.upper,
               tolerance = 0.001
  )
})

# ------------------------------------------------------------------------------

test_that("Wrappers -- selection of multiple variables works", {
  skip_if_not_installed("broom")
  skip_if_not_installed("modeldata")
  data("attrition", package = "modeldata")
  func <- function(split, ...) {
    lm(Age ~ HourlyRate + DistanceFromHome, data = analysis(split)) %>% tidy()
  }

  # generate boostrap resamples
  set.seed(888)
  bt_resamples <- bootstraps(attrition, times = 1000, apparent = TRUE) %>%
    mutate(res = purrr::map(splits, func))

  attrit_tidy <-
    lm(Age ~ HourlyRate + DistanceFromHome, data = attrition) %>%
    tidy(conf.int = TRUE) %>%
    dplyr::arrange(term)

  pct_res <-
    int_pctl(bt_resamples, res) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(pct_res$conf.low, pct_res$.lower, tolerance = .01)
  expect_equal(pct_res$conf.high, pct_res$.upper, tolerance = .01)


  t_res <-
    int_t(bt_resamples, res) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(t_res$conf.low, t_res$.lower, tolerance = .01)
  expect_equal(t_res$conf.high, t_res$.upper, tolerance = .01)


  bca_res <-
    int_bca(bt_resamples, res, .fn = func) %>%
    inner_join(attrit_tidy, by = "term")
  expect_equal(bca_res$conf.low, bca_res$.lower, tolerance = .01)
  expect_equal(bca_res$conf.high, bca_res$.upper, tolerance = .01)
})

# ------------------------------------------------------------------------------

test_that("Upper & lower confidence interval does not contain NA", {
  bad_stats <- function(split, ...) {
    tibble(
      term = "mean",
      estimate = NA_real_,
      std.error = runif(1)
    )
  }

  set.seed(888)
  bt_resamples <- bootstraps(data.frame(x = 1:100), times = 1000, apparent = TRUE) %>%
    mutate(res = purrr::map(splits, bad_stats))

  expect_snapshot(
    int_pctl(bt_resamples, res),
    error = TRUE
  )

  expect_snapshot(
    int_t(bt_resamples, res),
    error = TRUE
  )

  expect_snapshot(
    int_bca(bt_resamples, res, .fn = bad_stats),
    error = TRUE
  )
})

# ------------------------------------------------------------------------------

set.seed(456765)
bt_small <-
  bootstraps(dat, times = 10, apparent = TRUE) %>%
  dplyr::mutate(
    stats = purrr::map(splits, ~ get_stats(.x)),
    junk = 1:11
  )

test_that(
  "Sufficient replications needed to sufficiently reduce Monte Carlo sampling Error for BCa method",
  {
    expect_warning(int_pctl(bt_small, stats))
    expect_warning(int_t(bt_small, stats))
    expect_warning(int_bca(bt_small, stats, .fn = get_stats))
  }
)


test_that("bad input", {
  expect_error(int_pctl(bt_small, id))
  expect_error(int_pctl(bt_small, junk))
  expect_error(int_pctl(bt_small, stats, alpha = c(0.05, 0.2)))
  expect_error(int_t(bt_small, stats, alpha = "potato"))
  expect_error(int_bca(bt_small, stats, alpha = 1:2, .fn = get_stats))
  expect_error(int_pctl(vfold_cv(mtcars)))
  expect_error(int_t(vfold_cv(mtcars)))
  expect_error(int_bca(vfold_cv(mtcars)))

  bad_bt_norm <-
    bt_norm %>%
    mutate(stats = purrr::map(stats, ~ .x[, 1:2]))
  expect_error(int_t(bad_bt_norm, stats))

  expect_error(int_bca(bad_bt_norm, stats))

  no_dots <- function(split) {
    dat <- analysis(split)
    x <- dat[[1]]
    tibble(
      term = "mean",
      estimate = mean(x, na.rm = TRUE),
      std.error = sqrt(var(x, na.rm = TRUE) / sum(!is.na(x)))
    )
  }
  expect_error(
    int_bca(bt_norm, stats, .fn = no_dots),
    "must have an argument"
  )

  expect_error(int_pctl(as.data.frame(bt_norm), stats))
  expect_error(int_t(as.data.frame(bt_norm), stats))
  expect_error(int_bca(as.data.frame(bt_norm), stats, .fn = get_stats))

  expect_error(
    int_t(bt_norm %>% dplyr::filter(id != "Apparent"), stats)
  )
  expect_error(
    int_bca(bt_norm %>% dplyr::filter(id != "Apparent"), stats, .fn = get_stats)
  )

  poo <- function(x) {
    x$estimate <- "a"
    x
  }
  badder_bt_norm <-
    bt_norm %>%
    mutate(
      bad_term = purrr::map(stats, ~ .x %>% setNames(c("a", "estimate", "std.err"))),
      bad_est = purrr::map(stats, ~ .x %>% setNames(c("term", "b", "std.err"))),
      bad_err = purrr::map(stats, ~ .x %>% setNames(c("term", "estimate", "c"))),
      bad_num = purrr::map(stats, ~ poo(.x))
    )
  expect_error(int_pctl(badder_bt_norm, bad_term))
  expect_error(int_t(badder_bt_norm, bad_err))
  expect_error(int_bca(badder_bt_norm, bad_est, .fn = get_stats))
  expect_error(int_pctl(badder_bt_norm, bad_num))
})


# ------------------------------------------------------------------------------

test_that("regression intervals", {
  skip_if_not_installed("broom")
  skip_on_cran()

  expect_error(
    {
      set.seed(1)
      int_1 <- reg_intervals(mpg ~ disp + wt, data = mtcars)
    },
    regexp = NA
  )

  expect_equal(
    names(int_1),
    c("term", ".lower", ".estimate", ".upper", ".alpha", ".method")
  )

  expect_snapshot({
    skip_if(new_rng_snapshots)
    set.seed(123)
    int_2 <- reg_intervals(
      mpg ~ disp + wt,
      data = mtcars,
      filter = term == "wt",
      model_fn = "glm",
      keep_reps = TRUE
    )
    int_2
  })

  expect_equal(
    names(int_2),
    c("term", ".lower", ".estimate", ".upper", ".alpha", ".method", ".replicates")
  )
  expect_true(nrow(int_2) == 1)
  expect_true(all(int_2$term == "wt"))


  expect_error(
    reg_intervals(mpg ~ disp + wt, data = mtcars, model_fn = "potato"),
    "`model_fn` must be one of"
  )
  expect_error(
    reg_intervals(mpg ~ disp + wt, data = mtcars, type = "random"),
    "`type` must be one of"
  )
  expect_error(
    reg_intervals(mpg ~ disp + wt, data = mtcars, alpha = "a"),
    "must be a single numeric value"
  )
})

# Bootstrap confidence interval code

# ------------------------------------------------------------------------------
# helpers

check_includes_apparent <- function(x, call = caller_env()) {
  if (x |> dplyr::filter(id == "Apparent") |> nrow() != 1) {
    cli_abort(
      c(
        "The bootstrap resamples must include an apparent sample.",
        i = "Please set {.code apparent = TRUE} in the {.fn bootstraps} function."
      ),
      call = call
    )
  }
  invisible(NULL)
}


statistics_format_error <- cli::format_inline(
  "{.arg statistics} should select a list column of tidy results."
)
std_exp <- c("std.error", "robust.se")

check_statistics_names <- function(x, std_col, call = caller_env()) {
  # check for proper columns
  if (sum(colnames(x) == "estimate") != 1) {
    cli_abort(
      "The tibble in {.arg statistics} must have a column for 'estimate'.",
      call = call
    )
  }
  if (sum(colnames(x) == "term") != 1) {
    cli_abort(
      "The tibble in {.arg statistics} must have a column for 'term'.",
      call = call
    )
  }
  if (std_col) {
    std_candidates <- colnames(x) %in% std_exp
    if (sum(std_candidates) != 1) {
      cli_abort(
        "{.arg statistics} should select a single column for the standard error.",
        call = call
      )
    }
  }
  invisible(TRUE)
}

check_statistics <- function(x, std_col = FALSE, call = caller_env()) {
  if (!is.list(x)) {
    cli_abort(statistics_format_error, call = call)
  }

  # convert to data frame from list
  has_id <- any(names(x) == "id")
  if (has_id) {
    list_cols <- names(x)[map_lgl(x, is_list)]
    x <- try(tidyr::unnest(x, cols = all_of(list_cols)), silent = TRUE)
  } else {
    x <- try(map(x, \(.x) .x) |> list_rbind(), silent = TRUE)
  }

  if (inherits(x, "try-error")) {
    cli_abort(statistics_format_error, call = call)
  }

  check_statistics_names(x, std_col, call = call)

  if (std_col) {
    std_candidates <- colnames(x) %in% std_exp
    std_candidates <- colnames(x)[std_candidates]
    re_name <- list(std_err = std_candidates)
    if (has_id) {
      x <-
        dplyr::select(
          x,
          term,
          estimate,
          id,
          tidyselect::one_of(std_candidates),
          dplyr::starts_with(".")
        ) |>
        mutate(orig = (id == "Apparent")) |>
        dplyr::rename(!!!re_name)
    } else {
      x <-
        dplyr::select(
          x,
          term,
          estimate,
          tidyselect::one_of(std_candidates),
          dplyr::starts_with(".")
        ) |>
        dplyr::rename(!!!re_name)
    }
  } else {
    if (has_id) {
      x <-
        dplyr::select(x, term, estimate, id, dplyr::starts_with(".")) |>
        mutate(orig = (id == "Apparent")) |>
        dplyr::select(-id)
    } else {
      x <- dplyr::select(x, term, estimate, dplyr::starts_with("."))
    }
  }

  x
}


get_p0 <- function(x, alpha = 0.05, groups) {
  group_sym <- rlang::syms(groups)

  orig <- x |>
    group_by(!!!group_sym) |>
    dplyr::filter(orig) |>
    dplyr::select(!!!group_sym, theta_0 = estimate) |>
    ungroup()
  x |>
    dplyr::filter(!orig) |>
    inner_join(orig, by = groups) |>
    group_by(!!!group_sym) |>
    summarize(p0 = mean(estimate <= theta_0, na.rm = TRUE)) |>
    mutate(
      Z0 = stats::qnorm(p0),
      Za = stats::qnorm(1 - alpha / 2)
    )
}

new_stats <- function(x, lo, hi) {
  res <- as.numeric(quantile(x, probs = c(lo, hi), na.rm = TRUE))
  tibble(
    .lower = min(res),
    .estimate = mean(x, na.rm = TRUE),
    .upper = max(res)
  )
}

check_has_dots <- function(x, call = caller_env()) {
  nms <- names(formals(x))
  if (!any(nms == "...")) {
    cli_abort("{.arg .fn} must have an argument {.arg ...}.", call = call)
  }
  invisible(NULL)
}

check_num_resamples <- function(x, B = 1000, call = caller_env()) {
  x <-
    x |>
    dplyr::group_by(term) |>
    dplyr::summarize(n = sum(!is.na(estimate))) |>
    dplyr::filter(n < B)

  if (nrow(x) > 0) {
    terms <- as.character(x$term)
    cli_warn(
      "Recommend at least {B} non-missing bootstrap resamples for {cli::qty(terms)} term{?s} {.code {terms}}.",
      call = call
    )
  }
  invisible(NULL)
}

# ------------------------------------------------------------------------------
# percentile code

pctl_single <- function(stats, alpha = 0.05) {
  if (all(is.na(stats))) {
    cli_abort("All statistics have missing values.", call = call2("int_pctl"))
  }

  if (!is.numeric(stats)) {
    cli_abort("All statistics must be numeric.", call = call2("int_pctl"))
  }

  # stats is a numeric vector of values
  ci <- stats |> quantile(probs = c(alpha / 2, 1 - alpha / 2), na.rm = TRUE)

  # return a tibble with .lower, .estimate, .upper
  res <- tibble(
    .lower = min(ci),
    .estimate = mean(stats, na.rm = TRUE),
    .upper = max(ci),
    .alpha = alpha,
    .method = "percentile"
  )
  res
}

#' Bootstrap confidence intervals
#' @description
#' Calculate bootstrap confidence intervals using various methods.
#' @param .data A object containing the bootstrap resamples, created using
#'  [bootstraps()]. For t- and BCa-intervals, the `apparent` argument
#'  should be set to `TRUE`. Even if the `apparent` argument is set to
#'  `TRUE` for the percentile method, the apparent data is never used in calculating
#'  the percentile confidence interval.
#' @param statistics An unquoted column name or `dplyr` selector that identifies
#'  a single column in the data set containing the individual bootstrap
#'  estimates. This must be a list column of tidy tibbles (with columns
#'  `term` and `estimate`). Optionally, users can include columns whose names
#'  begin with a period and the intervals will be created for each combination
#'  of these variables and the `term` column. For t-intervals, a standard tidy
#'  column (usually called `std.error`) is required. See the examples below.
#' @param alpha Level of significance.
#' @param .fn A function to calculate statistic of interest. The
#'  function should take an `rsplit` as the first argument and the `...` are
#'  required.
#' @param ... Arguments to pass to `.fn` (`int_bca()` only).
#' @references \url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}
#' @return Each function returns a tibble with columns `.lower`,
#'  `.estimate`, `.upper`, `.alpha`, `.method`, and `term`.
#'  `.method` is the type of interval (eg. "percentile",
#'  "student-t", or "BCa"). `term` is the name of the estimate. Note
#'  the `.estimate` returned from `int_pctl()`
#'  is the mean of the estimates from the bootstrap resamples
#'  and not the estimate from the apparent model.
#' @details Percentile intervals are the standard method of
#'  obtaining confidence intervals but require thousands of
#'  resamples to be accurate. T-intervals may need fewer
#'  resamples but require a corresponding variance estimate.
#'  Bias-corrected and accelerated intervals require the original function
#'  that was used to create the statistics of interest and are
#'  computationally taxing.
#' @seealso [reg_intervals()]
#' @references Davison, A., & Hinkley, D. (1997). _Bootstrap Methods and their
#'  Application_. Cambridge: Cambridge University Press.
#'  doi:10.1017/CBO9780511802843
#'
#' @examplesIf rlang::is_installed("broom") & rlang::is_installed("modeldata")
#' \donttest{
#' library(broom)
#' library(dplyr)
#' library(purrr)
#' library(tibble)
#' library(tidyr)
#'
#' # ------------------------------------------------------------------------------
#'
#' lm_est <- function(split, ...) {
#'   lm(mpg ~ disp + hp, data = analysis(split)) |>
#'     tidy()
#' }
#'
#' set.seed(52156)
#' car_rs <-
#'   bootstraps(mtcars, 500, apparent = TRUE) |>
#'   mutate(results = map(splits, lm_est))
#'
#' int_pctl(car_rs, results)
#' int_t(car_rs, results)
#' int_bca(car_rs, results, .fn = lm_est)
#'
#' # ------------------------------------------------------------------------------
#'
#' # putting results into a tidy format
#' rank_corr <- function(split) {
#'   dat <- analysis(split)
#'   tibble(
#'     term = "corr",
#'     estimate = cor(dat$sqft, dat$price, method = "spearman"),
#'     # don't know the analytical std.error so no t-intervals
#'     std.error = NA_real_
#'   )
#' }
#'
#' set.seed(69325)
#' data(Sacramento, package = "modeldata")
#' bootstraps(Sacramento, 1000, apparent = TRUE) |>
#'   mutate(correlations = map(splits, rank_corr)) |>
#'   int_pctl(correlations)
#'
#' # ------------------------------------------------------------------------------
#' # An example of computing the interval for each value of a custom grouping
#' # factor (type of house in this example)
#'
#' # Get regression estimates for each house type
#' lm_est <- function(split, ...) {
#'   analysis(split) |>
#'     tidyr::nest(.by = c(type)) |>
#'     # Compute regression estimates for each house type
#'     mutate(
#'       betas = purrr::map(data, \(.x) lm(log10(price) ~ sqft, data = .x) |> tidy())
#'     ) |>
#'     # Convert the column name to begin with a period
#'     rename(.type = type) |>
#'     select(.type, betas) |>
#'     unnest(cols = betas)
#' }
#'
#' set.seed(52156)
#' house_rs <-
#'   bootstraps(Sacramento, 1000, apparent = TRUE) |>
#'   mutate(results = map(splits, lm_est))
#'
#' int_pctl(house_rs, results)
#' }
#' @export
int_pctl <- function(.data, ...) {
  UseMethod("int_pctl")
}

#' @export
#' @rdname int_pctl
int_pctl.default <- function(.data, ...) {
  cls <- class(.data)
  cli_abort(
    "No method for objects of class{?es}: {.cls {cls}}"
  )
}

#' @export
#' @rdname int_pctl
int_pctl.bootstraps <- function(.data, statistics, alpha = 0.05, ...) {
  check_dots_empty()
  check_number_decimal(alpha, min = 0, max = 1)

  .data <- .data |> dplyr::filter(id != "Apparent")

  column_name <- tidyselect::vars_select(
    names(.data),
    !!rlang::enquo(statistics)
  )
  if (length(column_name) != 1) {
    cli_abort(statistics_format_error)
  }
  stats <- .data[[column_name]]
  stats <- check_statistics(stats, std_col = FALSE)

  check_num_resamples(stats, B = 1000)

  stat_groups <- c("term", grep("^\\.", names(stats), value = TRUE))
  stat_groups <- rlang::syms(stat_groups)

  vals <- stats |>
    dplyr::group_by(!!!stat_groups) |>
    dplyr::do(pctl_single(.$estimate, alpha = alpha)) |>
    dplyr::ungroup()
  vals
}

# ------------------------------------------------------------------------------
# t interval code

t_single <- function(stats, std_err, is_orig, alpha = 0.05) {
  # stats is a numeric vector of values
  # vars is a numeric vector of variances
  # return a tibble with .lower, .estimate, .upper
  # which_orig is the index of stats and std_err that has the original result

  if (all(is.na(stats))) {
    cli_abort("All statistics have missing values.", call = call2("int_t"))
  }

  if (!is.logical(is_orig) || any(is.na(is_orig))) {
    cli_abort(
      "{.arg is_orig} should be a logical column the same length as {.arg stats} with no missing values."
    )
  }
  if (length(stats) != length(std_err) && length(stats) != length(is_orig)) {
    function_args <- c('stats', 'std_err', 'is_orig')
    cli_abort("{.arg {function_args}} should have the same length.")
  }
  if (sum(is_orig) != 1) {
    cli_abort("The original statistic must be in a single row.")
  }

  theta_obs <- stats[is_orig]
  std_err_obs <- std_err[is_orig]

  stats <- stats[!is_orig]
  std_err <- std_err[!is_orig]

  z_dist <-
    (stats - theta_obs) / std_err

  z_pntl <-
    quantile(z_dist, probs = c(alpha / 2, 1 - (alpha) / 2), na.rm = TRUE)

  ci <- theta_obs - z_pntl * std_err_obs

  tibble(
    .lower = min(ci),
    .estimate = mean(stats, na.rm = TRUE),
    .upper = max(ci),
    .alpha = alpha,
    .method = "student-t"
  )
}

#' @rdname int_pctl
#' @export
int_t <- function(.data, ...) {
  UseMethod("int_t")
}

#' @export
#' @rdname int_pctl
int_t.default <- function(.data, ...) {
  cls <- class(.data)
  cli_abort(
    "No method for objects of class{?es}: {.cls {cls}}"
  )
}

#' @rdname int_pctl
#' @export
int_t.bootstraps <- function(.data, statistics, alpha = 0.05, ...) {
  check_dots_empty()
  check_includes_apparent(.data)
  check_number_decimal(alpha, min = 0, max = 1)

  column_name <- tidyselect::vars_select(names(.data), !!enquo(statistics))
  if (length(column_name) != 1) {
    cli_abort(statistics_format_error)
  }
  stats <- .data |> dplyr::select(!!column_name, id)
  stats <- check_statistics(stats, std_col = TRUE)

  check_num_resamples(stats, B = 500)

  stat_groups <- c("term", grep("^\\.", names(stats), value = TRUE))
  stat_groups <- rlang::syms(stat_groups)
  vals <- stats |>
    dplyr::group_by(!!!stat_groups) |>
    dplyr::do(t_single(.$estimate, .$std_err, .$orig, alpha = alpha)) |>
    dplyr::ungroup()
  vals
}


# ----------------------------------------------------------------

bca_calc <- function(
  stats,
  orig_data,
  alpha = 0.05,
  .fn,
  ...,
  call = caller_env()
) {
  # TODO check per term
  if (all(is.na(stats$estimate))) {
    cli_abort("All statistics have missing values.", call = call)
  }

  stat_groups_chr <- c("term", grep("^\\.", names(stats), value = TRUE))
  stat_groups_sym <- rlang::syms(stat_groups_chr)

  ### Estimating Z0 bias-correction
  bias_corr_stats <- get_p0(stats, alpha = alpha, groups = stat_groups_chr)

  # need the original data frame here
  loo_rs <- loo_cv(orig_data)

  # We can't be sure what we will get back from the analysis function.
  # To test, we run on the first LOO data set and see if it is a vector or df
  loo_test <- try(rlang::exec(.fn, loo_rs$splits[[1]], ...), silent = TRUE)
  if (inherits(loo_test, "try-error")) {
    cli_text("Running {.fn .fn} on the LOO resamples produced an error:")
    print(loo_test)
    cli_abort("{.arg .fn} failed.", call = call)
  }

  loo_res <- furrr::future_map(loo_rs$splits, .fn, ...) |> list_rbind()

  loo_estimate <-
    loo_res |>
    dplyr::group_by(!!!stat_groups_sym) |>
    dplyr::summarize(loo = mean(estimate, na.rm = TRUE)) |>
    dplyr::inner_join(loo_res, by = stat_groups_chr, multiple = "all") |>
    dplyr::group_by(!!!stat_groups_sym) |>
    dplyr::summarize(
      cubed = sum((loo - estimate)^3),
      squared = sum((loo - estimate)^2)
    ) |>
    dplyr::ungroup() |>
    dplyr::inner_join(bias_corr_stats, by = stat_groups_chr) |>
    dplyr::mutate(
      a = cubed / (6 * (squared^(3 / 2))),
      Zu = (Z0 + Za) / (1 - a * (Z0 + Za)) + Z0,
      Zl = (Z0 - Za) / (1 - a * (Z0 - Za)) + Z0,
      lo = stats::pnorm(Zl, lower.tail = TRUE),
      hi = stats::pnorm(Zu, lower.tail = TRUE)
    )

  terms <- loo_estimate$term
  stats <- stats |> dplyr::filter(!orig)

  keys <- stats |> dplyr::distinct(!!!stat_groups_sym)
  for (i in seq_len(nrow(keys))) {
    tmp_stats <- dplyr::inner_join(stats, keys[i, ], by = stat_groups_chr)
    tmp_loo <- dplyr::inner_join(loo_estimate, keys[i, ], by = stat_groups_chr)

    tmp <- new_stats(tmp_stats$estimate, lo = tmp_loo$lo, hi = tmp_loo$hi)
    tmp <- dplyr::bind_cols(tmp, keys[i, ])
    if (i == 1) {
      ci_bca <- tmp
    } else {
      ci_bca <- dplyr::bind_rows(ci_bca, tmp)
    }
  }
  ci_bca <-
    ci_bca |>
    dplyr::select(!!!stat_groups_sym, .lower, .estimate, .upper) |>
    dplyr::mutate(
      .alpha = alpha,
      .method = "BCa"
    )
}

#' @rdname int_pctl
#' @export
int_bca <- function(.data, ...) {
  UseMethod("int_bca")
}

#' @export
#' @rdname int_pctl
int_bca.default <- function(.data, ...) {
  cls <- class(.data)
  cli_abort(
    "No method for objects of class{?es}: {.cls {cls}}"
  )
}

#' @rdname int_pctl
#' @export
int_bca.bootstraps <- function(.data, statistics, alpha = 0.05, .fn, ...) {
  check_includes_apparent(.data)
  check_number_decimal(alpha, min = 0, max = 1)

  check_has_dots(.fn)

  column_name <- tidyselect::vars_select(names(.data), !!enquo(statistics))
  if (length(column_name) != 1) {
    cli_abort(statistics_format_error)
  }
  stats <- .data |> dplyr::select(!!column_name, id, dplyr::starts_with("."))
  stats <- check_statistics(stats)

  check_num_resamples(stats, B = 1000)

  vals <- bca_calc(stats, .data$splits[[1]]$data, alpha = alpha, .fn = .fn, ...)
  vals
}

#' Create a Validation Set
#'
#' `validation_split()` takes a single random sample (without replacement) of
#'  the original data set to be used for analysis. All other data points are
#'  added to the assessment set (to be used as the validation set).
#' @details The `strata` argument causes the random sampling to be conducted
#'  *within the stratification variable*. This can help ensure that the number of
#'  data points in the analysis data is equivalent to the proportions in the
#'  original data set. (Strata below 10% of the total are pooled together.)
#' @inheritParams vfold_cv
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @param strata A variable that is used to conduct stratified sampling to
#'  create the resamples. This could be a single character value or a variable
#'  name that corresponds to a variable that exists in the data frame.
#' @param breaks A single number giving the number of bins desired to stratify
#'  a numeric stratification variable.
#' @export
#' @return An tibble with classes `validation_split`, `rset`, `tbl_df`, `tbl`,
#'  and `data.frame`. The results include a column for the data split objects
#'  and a column called `id` that has a character string with the resample
#'  identifier.
#' @examples
#' validation_split(mtcars, prop = .9)
#' @export
validation_split <- function(data, prop = 3/4, strata = NULL, breaks = 4, ...) {

  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) {
      strata <- NULL
    }
  }

  strata_check(strata, names(data))

  split_objs <-
    mc_splits(data = data,
              prop = 1 - prop,
              times = 1,
              strata = strata,
              breaks = breaks)

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)
  class(split_objs$splits[[1]]) <- c("val_split", "rsplit")

  val_att <- list(prop = prop,
                 strata = !is.null(strata))

  new_rset(splits = split_objs$splits,
           ids = "validation",
           attrib = val_att,
           subclass = c("validation_split", "rset"))
}

#' @export
print.validation_split <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("validation_split", "rset"))]
  print(x, ...)
}


#' @export
print.val_split<- function(x, ...) {

  if (all(is.na(x$out_id))) {
    out_char <- paste(length(complement(x)))
  } else {
    out_char <- paste(length(x$out_id))
  }

  cat("<Training/Validation/Total>\n")
  cat("<",
      length(x$in_id), "/",
      out_char, "/",
      nrow(x$data), ">\n",
      sep = "")
}

#' Gather an `rset` Object
#'
#' This method uses `gather` on an `rset` object to stack all of
#'  the non-ID or split columns in the data and is useful for
#'  stacking model evaluation statistics. The resulting data frame
#'  has a column based on the column names of `data` and another for
#'  the values.
#'
#' @param data An `rset` object.
#' @param key,value,... Not specified in this method and will be
#'  ignored. Note that this means that selectors are ignored if
#'  they are passed to the function.
#' @param na.rm If `TRUE`, will remove rows from output where the
#'  value column in `NA`.
#' @param convert If `TRUE` will automatically run
#'  `type.convert()` on the key column. This is useful if the column
#'  names are actually numeric, integer, or logical.
#' @param factor_key If FALSE, the default, the key values will be
#'  stored as a character vector. If `TRUE`, will be stored as a
#'  factor, which preserves the original ordering of the columns.
#' @return A data frame with the ID columns, a column called
#'  `model` (with the previous column names), and a column called
#'  `statistic` (with the values).
#' @examples
#' library(rsample)
#' cv_obj <- vfold_cv(mtcars, v = 10)
#' cv_obj$lm_rmse <- rnorm(10, mean = 2)
#' cv_obj$nnet_rmse <- rnorm(10, mean = 1)
#' gather(cv_obj)
#' @export gather.rset
#' @export
#' @method gather rset
gather.rset <- function(data, key = NULL, value = NULL, ..., na.rm = TRUE,
                        convert = FALSE, factor_key = TRUE) {
  if (any(names(data) == "splits")) {
    data <- data %>% dplyr::select(-splits)
  }

  data <- as.data.frame(data)

  id_vars <- grep("^id", names(data), value = TRUE)

  other_vars <- names(data)[!(names(data) %in% id_vars)]
  if (length(other_vars) < 2) {
    rlang::abort(
      paste0(
        "There should be at least two other columns ",
        "(besides `id` variables) in the data set to ",
        "use `gather.rset()`."
      )
    )
  }

  tidyr::gather(
    data,
    key = model,
    value = statistic,
    - !!id_vars,
    na.rm = na.rm,
    convert = convert,
    factor_key = factor_key
  )
}

#' Create a Validation Set
#'
#' @description
#' `r lifecycle::badge('deprecated')`
#'
#' This function is deprecated because it's part of an approach to constructing
#' a training, validation, and testing set by doing a sequence of two binary
#' splits: testing / not-testing (with [initial_split()] or one of its variants)
#' and then not-testing split into training/validation with `validation_split()`.
#' Instead, now use [initial_validation_split()] or one if its variants to
#' construct the three sets via one 3-way split.
#'
#' `validation_split()` takes a single random sample (without replacement) of
#'  the original data set to be used for analysis. All other data points are
#'  added to the assessment set (to be used as the validation set).
#'  `validation_time_split()` does the same, but takes the _first_ `prop` samples
#'  for training, instead of a random selection.
#'  `group_validation_split()` creates splits of the data based
#'  on some grouping variable, so that all data in a "group" is assigned to
#'  the same split.
#'
#'  Note that the input `data` to `validation_split()`, `validation_time_split()`,
#'  and `group_validation_split()` should _not_ contain the testing data. To
#'  create a three-way split directly of the entire data set, use
#'  [initial_validation_split()].
#'
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @export
#' @return An tibble with classes `validation_split`, `rset`, `tbl_df`, `tbl`,
#'  and `data.frame`. The results include a column for the data split objects
#'  and a column called `id` that has a character string with the resample
#'  identifier.
#'
#' @seealso [initial_validation_split()], [group_initial_validation_split()],
#' [validation_set()]
#'
#' @keywords internal
#' @examplesIf rlang::is_installed("modeldata")
#' cars_split <- initial_split(mtcars)
#' cars_not_testing <- training(cars_split)
#' validation_split(cars_not_testing, prop = .9)
#' group_validation_split(cars_not_testing, cyl)
#'
#' data(drinks, package = "modeldata")
#' validation_time_split(drinks[1:200,])
#'
#' # Alternative
#' cars_split_3 <- initial_validation_split(mtcars)
#' validation_set(cars_split_3)
validation_split <- function(
  data,
  prop = 3 / 4,
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  ...
) {
  lifecycle::deprecate_warn(
    "1.2.0",
    "validation_split()",
    "initial_validation_split()"
  )

  check_dots_empty()
  check_prop(prop)

  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) {
      strata <- NULL
    }
  }

  check_strata(strata, data)

  split_objs <-
    mc_splits(
      data = data,
      prop = prop,
      times = 1,
      strata = strata,
      breaks = breaks,
      pool = pool
    )

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)
  class(split_objs$splits[[1]]) <- c("val_split", "rsplit")

  if (!is.null(strata)) {
    names(strata) <- NULL
  }
  val_att <- list(
    prop = prop,
    strata = strata,
    breaks = breaks,
    pool = pool
  )

  new_rset(
    splits = split_objs$splits,
    ids = "validation",
    attrib = val_att,
    subclass = c("validation_split", "rset")
  )
}

#' @rdname validation_split
#' @inheritParams vfold_cv
#' @inheritParams initial_time_split
#' @param lag A value to include a lag between the assessment and analysis set.
#' This is useful if lagged predictors will be used during training and
#' testing.
#' @export
validation_time_split <- function(data, prop = 3 / 4, lag = 0, ...) {
  lifecycle::deprecate_warn(
    "1.2.0",
    "validation_time_split()",
    "initial_validation_time_split()"
  )

  check_dots_empty()

  check_prop(prop)
  if (!is.numeric(prop) | prop >= 1 | prop <= 0) {
    rlang::abort("`prop` must be a number on (0, 1).")
  }

  if (!is.numeric(lag) | !(lag %% 1 == 0)) {
    rlang::abort("`lag` must be a whole number.")
  }

  n_train <- floor(nrow(data) * prop)

  if (lag > n_train) {
    rlang::abort(
      "`lag` must be less than or equal to the number of training observations."
    )
  }

  split <- rsplit(data, 1:n_train, (n_train + 1 - lag):nrow(data))
  class(split) <- c("val_time_split", "val_split", "rsplit")
  splits <- list(split)

  val_att <- list(prop = prop, strata = FALSE)

  new_rset(
    splits = splits,
    ids = "validation",
    attrib = val_att,
    subclass = c("validation_time_split", "validation_split", "rset")
  )
}

#' @rdname validation_split
#' @inheritParams group_initial_split
#' @export
group_validation_split <- function(
  data,
  group,
  prop = 3 / 4,
  ...,
  strata = NULL,
  pool = 0.1
) {
  lifecycle::deprecate_warn(
    "1.2.0",
    "group_validation_split()",
    "group_initial_validation_split()"
  )

  check_dots_empty()

  group <- validate_group({{ group }}, data)
  check_prop(prop)

  if (!missing(strata)) {
    strata <- check_grouped_strata({{ group }}, {{ strata }}, pool, data)
  }

  split_objs <-
    group_mc_splits(
      data = data,
      group = {{ group }},
      prop = prop,
      times = 1,
      strata = {{ strata }},
      pool = pool
    )

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)
  class(split_objs$splits[[1]]) <- c("group_val_split", "val_split", "rsplit")

  # This is needed for printing -- strata cannot be missing
  if (is.null(strata)) {
    strata <- FALSE
  }
  val_att <- list(
    prop = prop,
    group = group,
    strata = strata,
    pool = pool
  )

  new_rset(
    splits = split_objs$splits,
    ids = "validation",
    attrib = val_att,
    subclass = c(
      "group_validation_split",
      "validation_split",
      "group_rset",
      "rset"
    )
  )
}

test_that("`validation_split()` is deprecated", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))

  expect_snapshot({
    set.seed(11)
    rs1 <- validation_split(dat1)
    sizes1 <- dim_rset(rs1)

    expect_true(all(sizes1$analysis == 15))
    expect_true(all(sizes1$assessment == 5))
  })
})

test_that("default param", {
  withr::local_options(lifecycle_verbosity = "quiet")

  set.seed(11)
  rs1 <- validation_split(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 15))
  expect_true(all(sizes1$assessment == 5))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("`validation_time_split()` is deprecated", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))

  expect_snapshot({
    set.seed(11)
    rs1 <- validation_time_split(dat1)
    sizes1 <- dim_rset(rs1)

    expect_true(all(sizes1$analysis == 15))
    expect_true(all(sizes1$assessment == 5))
  })
})

test_that("default time param", {
  withr::local_options(lifecycle_verbosity = "quiet")

  set.seed(11)
  rs1 <- validation_time_split(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 15))
  expect_true(all(sizes1$assessment == 5))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
  tr1 <- training(rs1$splits[[1]])
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3 / 4))
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3 / 4)))
})

test_that("default time param with lag", {
  withr::local_options(lifecycle_verbosity = "quiet")

  skip_if_not_installed("modeldata")
  data(drinks, package = "modeldata")

  rs1 <- validation_time_split(dat1, lag = 5)
  expect_s3_class(rs1, "validation_split")
  tr1 <- training(rs1$splits[[1]])
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3 / 4))
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3 / 4)))

  expect_snapshot(validation_time_split(drinks))
  expect_snapshot(validation_time_split(drinks, lag = 12.5), error = TRUE)
  expect_snapshot(validation_time_split(drinks, lag = 500), error = TRUE)
})

test_that("assessment set of time split includes the lag (#376)", {
  withr::local_options(lifecycle_verbosity = "quiet")

  toy_data <- data.frame(id = 1:100)
  val_rset <- validation_time_split(toy_data, prop = 0.75, lag = 2)
  dat_assess <- assessment(val_rset$splits[[1]])
  expect_equal(dat_assess$id, 74:100)
})

test_that("`group_validation_split()` is deprecated", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))

  expect_snapshot({
    set.seed(11)
    rs1 <- group_validation_split(dat1, c)
    sizes1 <- dim_rset(rs1)

    expect_true(all(sizes1$analysis == 15))
    expect_true(all(sizes1$assessment == 5))
  })
})

test_that("default group param", {
  withr::local_options(lifecycle_verbosity = "quiet")

  set.seed(11)
  rs1 <- group_validation_split(dat1, c)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 15))
  expect_true(all(sizes1$assessment == 5))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("grouping -- strata", {
  withr::local_options(lifecycle_verbosity = "quiet")

  set.seed(11)

  n_common_class <- 70
  n_rare_class <- 30

  group_table <- tibble(
    group = 1:100,
    outcome = sample(c(rep(0, n_common_class), rep(1, n_rare_class)))
  )
  observation_table <- tibble(
    group = sample(1:100, 5e4, replace = TRUE),
    observation = 1:5e4
  )
  sample_data <- dplyr::full_join(
    group_table,
    observation_table,
    by = "group",
    multiple = "all"
  )
  rs4 <- group_validation_split(sample_data, group, strata = outcome)
  sizes4 <- dim_rset(rs4)
  expect_snapshot(sizes4)

  rate <- purrr::map_dbl(
    rs4$splits,
    function(x) {
      dat <- as.data.frame(x)$outcome
      mean(dat == "1")
    }
  )
  expect_equal(mean(rate), 0.3, tolerance = 1e-2)

  good_holdout <- purrr::map_lgl(
    rs4$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("different percent", {
  withr::local_options(lifecycle_verbosity = "quiet")

  set.seed(11)
  rs2 <- validation_split(dat1, prop = .5)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == 10))
  expect_true(all(sizes2$assessment == 10))
  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  set.seed(11)
  rs2_group <- group_validation_split(dat1, c, prop = .5)
  sizes2_group <- dim_rset(rs2_group)

  expect_true(all(sizes2_group$analysis == 10))
  expect_true(all(sizes2_group$assessment == 10))
  same_data <-
    purrr::map_lgl(rs2_group$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2_group$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("strata", {
  withr::local_options(lifecycle_verbosity = "quiet")

  set.seed(11)
  rs3 <- validation_split(warpbreaks, strata = "tension")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis == 39))
  expect_true(all(sizes3$assessment == 15))

  rate <- purrr::map_dbl(
    rs3$splits,
    function(x) {
      dat <- as.data.frame(x)$tension
      mean(dat == "M")
    }
  )
  expect_true(length(unique(rate)) == 1)

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("bad args", {
  withr::local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(error = TRUE, {
    validation_split(mtcars, prop = 1)
  })
  expect_snapshot(error = TRUE, {
    validation_time_split(mtcars, prop = 1)
  })
  expect_snapshot(error = TRUE, {
    group_validation_split(mtcars, group = "cyl", prop = 1)
  })

  expect_snapshot(error = TRUE, {
    validation_split(warpbreaks, strata = warpbreaks$tension)
  })
  expect_snapshot(error = TRUE, {
    validation_split(warpbreaks, strata = c("tension", "wool"))
  })
})

test_that("printing", {
  withr::local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(validation_split(warpbreaks))
  expect_snapshot(validation_split(warpbreaks)$splits[[1]])
})

test_that("rsplit labels", {
  withr::local_options(lifecycle_verbosity = "quiet")

  rs <- validation_split(mtcars)
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

test_that("default param", {
  set.seed(11)
  rs1 <- validation_split(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 15))
  expect_true(all(sizes1$assessment == 5))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("default time param", {
  set.seed(11)
  rs1 <- validation_time_split(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == 15))
  expect_true(all(sizes1$assessment == 5))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
  tr1 <- training(rs1$splits[[1]])
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3 / 4))
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3 / 4)))
})

test_that("default time param with lag", {
  data(drinks, package = "modeldata")

  rs1 <- validation_time_split(dat1, lag = 5)
  expect_s3_class(rs1, "validation_split")
  tr1 <- training(rs1$splits[[1]])
  expect_equal(nrow(tr1), floor(nrow(dat1) * 3 / 4))
  expect_equal(tr1, dplyr::slice(dat1, 1:floor(nrow(dat1) * 3 / 4)))

  expect_snapshot(validation_time_split(drinks))
  expect_snapshot(validation_time_split(drinks, lag = 12.5), error = TRUE)
  expect_snapshot(validation_time_split(drinks, lag = 500), error = TRUE)
})


test_that("different percent", {
  set.seed(11)
  rs2 <- validation_split(dat1, prop = .5)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == 10))
  expect_true(all(sizes2$assessment == 10))
  same_data <-
    purrr::map_lgl(rs2$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs2$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("strata", {
  set.seed(11)
  rs3 <- validation_split(warpbreaks, strata = "tension")
  sizes3 <- dim_rset(rs3)

  expect_true(all(sizes3$analysis == 39))
  expect_true(all(sizes3$assessment == 15))

  rate <- purrr::map_dbl(
    rs3$splits,
    function(x) {
      dat <- as.data.frame(x)$tension
      mean(dat == "M")
    }
  )
  expect_true(length(unique(rate)) == 1)

  good_holdout <- purrr::map_lgl(
    rs3$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})


test_that("bad args", {
  expect_error(validation_split(warpbreaks, strata = warpbreaks$tension))
  expect_error(validation_split(warpbreaks, strata = c("tension", "wool")))
})


test_that("printing", {
  expect_snapshot(validation_split(warpbreaks))
  expect_snapshot(validation_split(warpbreaks)$splits[[1]])
})

test_that("rsplit labels", {
  rs <- validation_split(mtcars)
  all_labs <- purrr::map_df(rs$splits, labels)
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

#' Bootstrap Sampling
#'
#' A bootstrap sample is a sample that is the same size as the original data
#'  set that is made using replacement. This results in analysis samples that
#'  have multiple replicates of some of the original rows of the data. The
#'  assessment set is defined as the rows of the original data that were not
#'  included in the bootstrap sample. This is often referred to as the
#'  "out-of-bag" (OOB) sample.
#' @details The argument `apparent` enables the option of an additional
#'  "resample" where the analysis and assessment data sets are the same as the
#'  original data set. This can be required for some types of analysis of the
#'  bootstrap results.
#' The `strata` argument is based on a similar argument in the random forest
#'  package were the bootstrap samples are conducted *within the stratification
#'  variable*. This can help ensure that the number of data points in the
#'  bootstrap sample is equivalent to the proportions in the original data set.
#'  (Strata below 10% of the total are pooled together.)
#' @inheritParams vfold_cv
#' @param times The number of bootstrap samples.
#' @param strata A variable that is used to conduct stratified sampling. When
#'  not `NULL`, each bootstrap sample is created within the stratification
#'  variable. This could be a single character value or a variable name that
#'  corresponds to a variable that exists in the data frame.
#' @param breaks A single number giving the number of bins desired to stratify
#'  a numeric stratification variable.
#' @param apparent A logical. Should an extra resample be added where the
#'  analysis and holdout subset are the entire data set. This is required for
#'  some estimators used by the `summary` function that require the apparent
#'  error rate.
#' @export
#' @return An tibble with classes `bootstraps`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and a
#'  column called `id` that has a character string with the resample identifier.
#' @examples
#' bootstraps(mtcars, times = 2)
#' bootstraps(mtcars, times = 2, apparent = TRUE)
#'
#' library(purrr)
#' data(wa_churn, package = "modeldata")
#'
#' set.seed(13)
#' resample1 <- bootstraps(wa_churn, times = 3)
#' map_dbl(resample1$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#'
#' set.seed(13)
#' resample2 <- bootstraps(wa_churn, strata = "churn", times = 3)
#' map_dbl(resample2$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#'
#' set.seed(13)
#' resample3 <- bootstraps(wa_churn, strata = "tenure", breaks = 6, times = 3)
#' map_dbl(resample3$splits,
#'         function(x) {
#'           dat <- as.data.frame(x)$churn
#'           mean(dat == "Yes")
#'         })
#' @export
bootstraps <-
  function(data,
           times = 25,
           strata = NULL,
           breaks = 4,
           apparent = FALSE,
           ...) {

  if(!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if(length(strata) == 0) strata <- NULL
  }

  strata_check(strata, names(data))

  split_objs <-
    boot_splits(
      data = data,
      times = times,
      strata = strata,
      breaks = breaks
    )
  if(apparent)
    split_objs <- bind_rows(split_objs, apparent(data))

  boot_att <- list(times = times,
                   apparent = apparent,
                   strata = !is.null(strata))

  new_rset(splits = split_objs$splits,
           ids = split_objs$id,
           attrib = boot_att,
           subclass = c("bootstraps", "rset"))

}

# Get the indices of the analysis set from the analysis set (= bootstrap sample)
boot_complement <- function(ind, n) {
  list(analysis = ind, assessment = NA)
}

boot_splits <-
  function(data,
           times = 25,
           strata = NULL,
           breaks = 4) {

  n <- nrow(data)

  if (is.null(strata)) {
    indices <- purrr::map(rep(n, times), sample, replace = TRUE)
  } else {
    stratas <- tibble::tibble(idx = 1:n,
                              strata = make_strata(getElement(data, strata),
                                                   breaks = breaks))
    stratas <- split_unnamed(stratas, stratas$strata)
    stratas <-
      purrr::map_df(
        stratas,
        strat_sample,
        prop = 1,
        times = times,
        replace = TRUE
      )
    indices <- split_unnamed(stratas$idx, stratas$rs_id)
  }

  indices <- lapply(indices, boot_complement, n = n)

  split_objs <-
    purrr::map(indices, make_splits, data = data, class = "boot_split")
  list(splits = split_objs,
       id = names0(length(split_objs), "Bootstrap"))
}

#' @export
print.bootstraps <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("bootstraps", "rset"))]
  print(x, ...)
}

# Bootstrap confidence interval code

# ------------------------------------------------------------------------------
# helpers


check_rset <- function(x, app = TRUE) {
  if (!inherits(x, "bootstraps")) {
    rlang::abort("`.data` should be an `rset` object generated from `bootstraps()`")
  }

  if (app) {
    if (x %>% dplyr::filter(id == "Apparent") %>% nrow() != 1) {
      rlang::abort("Please set `apparent = TRUE` in `bootstraps()` function")
    }
  }
  invisible(NULL)
}


stat_fmt_err <- paste("`statistics` should select a list column of tidy results.")
stat_nm_err <- paste(
  "The tibble in `statistics` should have columns for",
  "'estimate' and 'term`"
)
std_exp <- c("std.error", "robust.se")

check_tidy_names <- function(x, std_col) {
  # check for proper columns
  if (sum(colnames(x) == "estimate") != 1) {
    rlang::abort(stat_nm_err)
  }
  if (sum(colnames(x) == "term") != 1) {
    rlang::abort(stat_nm_err)
  }
  if (std_col) {
    std_candidates <- colnames(x) %in% std_exp
    if (sum(std_candidates) != 1) {
      rlang::abort(
        "`statistics` should select a single column for the standard error."
      )
    }
  }
  invisible(TRUE)
}

check_tidy <- function(x, std_col = FALSE) {
  if (!is.list(x)) {
    rlang::abort(stat_fmt_err)
  }

  # convert to data frame from list
  has_id <- any(names(x) == "id")
  if (has_id) {
    list_cols <- names(x)[map_lgl(x, is_list)]
    x <- try(tidyr::unnest(x, cols = all_of(list_cols)), silent = TRUE)
  } else {
    x <- try(map_dfr(x, ~.x), silent = TRUE)
  }

  if (inherits(x, "try-error")) {
    rlang::abort(stat_fmt_err)
  }

  check_tidy_names(x, std_col)

  if (std_col) {
    std_candidates <- colnames(x) %in% std_exp
    std_candidates <- colnames(x)[std_candidates]
    if (has_id) {
      x <-
        dplyr::select(x, term, estimate, id, tidyselect::one_of(std_candidates)) %>%
        mutate(id = (id == "Apparent")) %>%
        setNames(c("term", "estimate", "orig", "std_err"))
    } else {
      x <-
        dplyr::select(x, term, estimate, tidyselect::one_of(std_candidates)) %>%
        setNames(c("term", "estimate", "std_err"))
    }
  } else {
    if (has_id) {
      x <-
        dplyr::select(x, term, estimate, id) %>%
        mutate(orig = (id == "Apparent")) %>%
        dplyr::select(-id)
    } else {
      x <- dplyr::select(x, term, estimate)
    }
  }

  x
}


get_p0 <- function(x, alpha = 0.05) {
  orig <- x %>%
    group_by(term) %>%
    dplyr::filter(orig) %>%
    dplyr::select(term, theta_0 = estimate) %>%
    ungroup()
  x %>%
    dplyr::filter(!orig) %>%
    inner_join(orig, by = "term") %>%
    group_by(term) %>%
    summarize(p0 = mean(estimate <= theta_0, na.rm = TRUE)) %>%
    mutate(
      Z0 = stats::qnorm(p0),
      Za = stats::qnorm(1 - alpha / 2)
    )
}

new_stats <- function(x, lo, hi) {
  res <- as.numeric(quantile(x, probs = c(lo, hi), na.rm = TRUE))
  tibble(.lower = min(res), .estimate = mean(x, na.rm = TRUE), .upper = max(res))
}

has_dots <- function(x) {
  nms <- names(formals(x))
  if (!any(nms == "...")) {
    rlang::abort("`.fn` must have an argument `...`.")
  }
  invisible(NULL)
}

check_num_resamples <- function(x, B = 1000) {
  x <-
    x %>%
    dplyr::group_by(term) %>%
    dplyr::summarize(n = sum(!is.na(estimate))) %>%
    dplyr::filter(n < B)

  if (nrow(x) > 0) {
    terms <- paste0("`", x$term, "`")
    msg <-
      paste0(
        "Recommend at least ", B, " non-missing bootstrap resamples for ",
        ifelse(length(terms) > 1, "terms: ", "term "),
        paste0(terms, collapse = ", "),
        "."
      )
    rlang::warn(msg)
  }
  invisible(NULL)
}

# ------------------------------------------------------------------------------
# percentile code


pctl_single <- function(stats, alpha = 0.05) {
  if (all(is.na(stats))) {
    rlang::abort("All statistics have missing values..")
  }

  if (!is.numeric(stats)) {
    rlang::abort("`stats` must be a numeric vector.")
  }

  # stats is a numeric vector of values
  ci <- stats %>% quantile(probs = c(alpha / 2, 1 - alpha / 2), na.rm = TRUE)

  # return a tibble with .lower, .estimate, .upper
  res <- tibble(
    .lower = min(ci),
    .estimate = mean(stats, na.rm = TRUE),
    .upper = max(ci),
    .alpha = alpha,
    .method = "percentile"
  )
  res
}

#' Bootstrap confidence intervals
#' @description
#' Calculate bootstrap confidence intervals using various methods.
#' @param .data A data frame containing the bootstrap resamples created using
#'  `bootstraps()`. For t- and BCa-intervals, the `apparent` argument
#'  should be set to `TRUE`. Even if the `apparent` argument is set to
#'  `TRUE` for the percentile method, the apparent data is never used in calculating
#'  the percentile confidence interval.
#' @param statistics An unquoted column name or `dplyr` selector that identifies
#'  a single column in the data set containing the individual bootstrap
#'  estimates. This must be a list column of tidy tibbles (with columns
#'  `term` and `estimate`). For t-intervals, a
#'  standard tidy column (usually called `std.err`) is required.
#'  See the examples below.
#' @param alpha Level of significance
#' @return Each function returns a tibble with columns `.lower`,
#'  `.estimate`, `.upper`, `.alpha`, `.method`, and `term`.
#'  `.method` is the type of interval (eg. "percentile",
#'  "student-t", or "BCa"). `term` is the name of the estimate. Note
#'  the `.estimate` returned from `int_pctl()`
#'  is the mean of the estimates from the bootstrap resamples
#'  and not the estimate from the apparent model.
#' @details Percentile intervals are the standard method of
#'  obtaining confidence intervals but require thousands of
#'  resamples to be accurate. T-intervals may need fewer
#'  resamples but require a corresponding variance estimate.
#'  Bias-corrected and accelerated intervals require the original function
#'  that was used to create the statistics of interest and are
#'  computationally taxing.
#' @seealso [reg_intervals()]
#' @references Davison, A., & Hinkley, D. (1997). _Bootstrap Methods and their
#'  Application_. Cambridge: Cambridge University Press.
#'  doi:10.1017/CBO9780511802843
#'
#' @examples
#' \donttest{
#' library(broom)
#' library(dplyr)
#' library(purrr)
#' library(tibble)
#'
#' lm_est <- function(split, ...) {
#'   lm(mpg ~ disp + hp, data = analysis(split)) %>%
#'     tidy()
#' }
#'
#' set.seed(52156)
#' car_rs <-
#'   bootstraps(mtcars, 500, apparent = TRUE) %>%
#'   mutate(results = map(splits, lm_est))
#'
#' int_pctl(car_rs, results)
#' int_t(car_rs, results)
#' int_bca(car_rs, results, .fn = lm_est)
#'
#' # putting results into a tidy format
#' rank_corr <- function(split) {
#'   dat <- analysis(split)
#'   tibble(
#'     term = "corr",
#'     estimate = cor(dat$sqft, dat$price, method = "spearman"),
#'     # don't know the analytical std.err so no t-intervals
#'     std.err = NA_real_
#'   )
#' }
#'
#' set.seed(69325)
#' data(Sacramento, package = "modeldata")
#' bootstraps(Sacramento, 1000, apparent = TRUE) %>%
#'   mutate(correlations = map(splits, rank_corr)) %>%
#'   int_pctl(correlations)
#' }
#' @export
int_pctl <- function(.data, statistics, alpha = 0.05) {
  check_rset(.data, app = FALSE)
  if (length(alpha) != 1 || !is.numeric(alpha)) {
    abort("`alpha` must be a single numeric value.")
  }

  .data <- .data %>% dplyr::filter(id != "Apparent")

  column_name <- tidyselect::vars_select(names(.data), !!rlang::enquo(statistics))
  if (length(column_name) != 1) {
    rlang::abort(stat_fmt_err)
  }
  stats <- .data[[column_name]]
  stats <- check_tidy(stats, std_col = FALSE)

  check_num_resamples(stats, B = 1000)

  vals <- stats %>%
    dplyr::group_by(term) %>%
    dplyr::do(pctl_single(.$estimate, alpha = alpha)) %>%
    dplyr::ungroup()
  vals
}

# ------------------------------------------------------------------------------
# t interval code

t_single <- function(stats, std_err, is_orig, alpha = 0.05) {
  # stats is a numeric vector of values
  # vars is a numeric vector of variances
  # return a tibble with .lower, .estimate, .upper
  # which_orig is the index of stats and std_err that has the original result

  if (all(is.na(stats))) {
    rlang::abort("All statistics have missing values.")
  }

  if (!is.logical(is_orig) || any(is.na(is_orig))) {
    rlang::abort(
      "`is_orig` should be a logical column the same length as `stats` with no missing values."
    )
  }
  if (length(stats) != length(std_err) && length(stats) != length(is_orig)) {
    rlang::abort("`stats`, `std_err`, and `is_orig` should have the same length.")
  }
  if (sum(is_orig) != 1) {
    rlang::abort("The original statistic must be in a single row.")
  }

  theta_obs <- stats[is_orig]
  std_err_obs <- std_err[is_orig]

  stats <- stats[!is_orig]
  std_err <- std_err[!is_orig]

  z_dist <-
    (stats - theta_obs) / std_err

  z_pntl <-
    quantile(z_dist, probs = c(alpha / 2, 1 - (alpha) / 2), na.rm = TRUE)

  ci <- theta_obs - z_pntl * std_err_obs

  tibble(
    .lower = min(ci),
    .estimate = mean(stats, na.rm = TRUE),
    .upper = max(ci),
    .alpha = alpha,
    .method = "student-t"
  )
}


#' @rdname int_pctl
#' @export
int_t <- function(.data, statistics, alpha = 0.05) {
  check_rset(.data)
  if (length(alpha) != 1 || !is.numeric(alpha)) {
    abort("`alpha` must be a single numeric value.")
  }

  column_name <- tidyselect::vars_select(names(.data), !!enquo(statistics))
  if (length(column_name) != 1) {
    rlang::abort(stat_fmt_err)
  }
  stats <- .data %>% dplyr::select(!!column_name, id)
  stats <- check_tidy(stats, std_col = TRUE)

  check_num_resamples(stats, B = 500)

  vals <-
    stats %>%
    dplyr::group_by(term) %>%
    dplyr::do(t_single(.$estimate, .$std_err, .$orig, alpha = alpha)) %>%
    dplyr::ungroup()
  vals
}


# ----------------------------------------------------------------

bca_calc <- function(stats, orig_data, alpha = 0.05, .fn, ...) {

  # TODO check per term
  if (all(is.na(stats$estimate))) {
    rlang::abort("All statistics have missing values.")
  }

  ### Estimating Z0 bias-correction
  bias_corr_stats <- get_p0(stats, alpha = alpha)

  # need the original data frame here
  loo_rs <- loo_cv(orig_data)

  # We can't be sure what we will get back from the analysis function.
  # To test, we run on the first LOO data set and see if it is a vector or df
  loo_test <- try(rlang::exec(.fn, loo_rs$splits[[1]], ...), silent = TRUE)
  if (inherits(loo_test, "try-error")) {
    cat("Running `.fn` on the LOO resamples produced an error:\n")
    print(loo_test)
    rlang::abort("`.fn` failed.")
  }

  loo_res <- furrr::future_map_dfr(loo_rs$splits, .fn, ...)

  loo_estimate <-
    loo_res %>%
    dplyr::group_by(term) %>%
    dplyr::summarize(loo = mean(estimate, na.rm = TRUE)) %>%
    dplyr::inner_join(loo_res, by = "term") %>%
    dplyr::group_by(term) %>%
    dplyr::summarize(
      cubed = sum((loo - estimate)^3),
      squared = sum((loo - estimate)^2)
    ) %>%
    dplyr::ungroup() %>%
    dplyr::inner_join(bias_corr_stats, by = "term") %>%
    dplyr::mutate(
      a = cubed / (6 * (squared^(3 / 2))),
      Zu = (Z0 + Za) / (1 - a * (Z0 + Za)) + Z0,
      Zl = (Z0 - Za) / (1 - a * (Z0 - Za)) + Z0,
      lo = stats::pnorm(Zl, lower.tail = TRUE),
      hi = stats::pnorm(Zu, lower.tail = TRUE)
    )

  terms <- loo_estimate$term
  stats <- stats %>% dplyr::filter(!orig)
  for (i in seq_along(terms)) {
    tmp <- new_stats(stats$estimate[stats$term == terms[i]],
                     lo = loo_estimate$lo[i],
                     hi = loo_estimate$hi[i]
    )
    tmp$term <- terms[i]
    if (i == 1) {
      ci_bca <- tmp
    } else {
      ci_bca <- bind_rows(ci_bca, tmp)
    }
  }
  ci_bca <-
    ci_bca %>%
    dplyr::select(term, .lower, .estimate, .upper) %>%
    dplyr::mutate(
      .alpha = alpha,
      .method = "BCa"
    )
}


#' @rdname int_pctl
#' @param .fn A function to calculate statistic of interest. The
#'  function should take an `rsplit` as the first argument and the `...` are
#'  required.
#' @param ... Arguments to pass to `.fn`.
#' @references \url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}
#' @export
int_bca <- function(.data, statistics, alpha = 0.05, .fn, ...) {
  check_rset(.data)
  if (length(alpha) != 1 || !is.numeric(alpha)) {
    abort("`alpha` must be a single numeric value.")
  }

  has_dots(.fn)

  column_name <- tidyselect::vars_select(names(.data), !!enquo(statistics))
  if (length(column_name) != 1) {
    rlang::abort(stat_fmt_err)
  }
  stats <- .data %>% dplyr::select(!!column_name, id)
  stats <- check_tidy(stats)

  check_num_resamples(stats, B = 1000)

  vals <- bca_calc(stats, .data$splits[[1]]$data, alpha = alpha, .fn = .fn, ...)
  vals
}

dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))
car_folds <- vfold_cv(mtcars)

new_rng_snapshots <- utils::compareVersion(
  "3.6.0",
  as.character(getRversion())
) >
  0

expect_s3_class_rset <- function(x) {
  expect_s3_class(x, "rset")
}

expect_s3_class_bare_tibble <- function(x) {
  expect_s3_class(x, c("tbl_df", "tbl", "data.frame"), exact = TRUE)
}

get_stats <- function(split, ...) {
  dat <- analysis(split)
  x <- dat[[1]]
  tibble(
    term = "mean",
    estimate = mean(x, na.rm = TRUE),
    std.error = sqrt(var(x, na.rm = TRUE) / sum(!is.na(x)))
  )
}

check_ind <- function(x, tdat) {
  in_dat <- subset(tdat, Data == "Analysis")
  in_check <- all(sort(in_dat$Row) == x$in_ind)
  out_dat <- subset(tdat, Data == "Analysis")
  out_check <- all(sort(out_dat$Row) == x$out_ind)
  in_check & out_check
}

get_id_left_out <- function(x) {
  unique(as.character(assessment(x)$tension))
}

#' Job Attrition
#'
#' @details These data are from the IBM Watson Analytics Lab.
#'  The website describes the data with \dQuote{Uncover the
#'  factors that lead to employee attrition and explore important
#'  questions such as \sQuote{show me a breakdown of distance
#'  from home by job role and attrition} or \sQuote{compare
#'  average monthly income by education and attrition}. This is a
#'  fictional data set created by IBM data scientists.}. There
#'  are 1470 rows.
#'
#' These data are now in the `modeldata` package.
#'
#' @name attrition
NULL


#' Two Class Data
#'
#' @details There are artificial data with two predictors (`A` and `B`) and
#'  a factor outcome variable (`Class`).
#'
#' These data are now in the `modeldata` package.
#'
#' @name two_class_dat
NULL

#' Sample Time Series Data
#'
#' @details Drink sales. The exact name of the series from FRED is:
#' "Merchant Wholesalers, Except Manufacturers' Sales Branches and Offices
#' Sales: Nondurable Goods: Beer, Wine, and Distilled Alcoholic Beverages Sales"
#'
#' These data are now in the `modeldata` package.
#' @name drinks
NULL

#' Tidy Resampling Object
#'
#' The `tidy` function from the \pkg{broom} package can be used on `rset` and
#'  `rsplit` objects to generate tibbles with which rows are in the analysis and
#'  assessment sets.
#' @param x A  `rset` or  `rsplit` object
#' @param unique_ind Should unique row identifiers be returned? For example,
#'  if `FALSE` then bootstrapping results will include multiple rows in the
#'  sample for the same row in the original data.
#' @param ... Not currently used.
#' @return A tibble with columns `Row` and `Data`. The latter has possible
#'  values "Analysis" or "Assessment". For `rset` inputs, identification columns
#'  are also returned but their names and values depend on the type of
#'  resampling. `vfold_cv` contains a column "Fold" and, if repeats are used,
#'  another called "Repeats". `bootstraps` and `mc_cv` use the column
#'  "Resample".
#' @details Note that for nested resampling, the rows of the inner resample,
#'  named `inner_Row`, are *relative* row indices and do not correspond to the
#'  rows in the original data set.
#' @examples
#' library(ggplot2)
#' theme_set(theme_bw())
#'
#' set.seed(4121)
#' cv <- tidy(vfold_cv(mtcars, v = 5))
#' ggplot(cv, aes(x = Fold, y = Row, fill = Data)) +
#'   geom_tile() +
#'   scale_fill_brewer()
#'
#' set.seed(4121)
#' rcv <- tidy(vfold_cv(mtcars, v = 5, repeats = 2))
#' ggplot(rcv, aes(x = Fold, y = Row, fill = Data)) +
#'   geom_tile() +
#'   facet_wrap(~Repeat) +
#'   scale_fill_brewer()
#'
#' set.seed(4121)
#' mccv <- tidy(mc_cv(mtcars, times = 5))
#' ggplot(mccv, aes(x = Resample, y = Row, fill = Data)) +
#'   geom_tile() +
#'   scale_fill_brewer()
#'
#' set.seed(4121)
#' bt <- tidy(bootstraps(mtcars, time = 5))
#' ggplot(bt, aes(x = Resample, y = Row, fill = Data)) +
#'   geom_tile() +
#'   scale_fill_brewer()
#'
#' dat <- data.frame(day = 1:30)
#' # Resample by week instead of day
#' ts_cv <- rolling_origin(dat,
#'   initial = 7, assess = 7,
#'   skip = 6, cumulative = FALSE
#' )
#' ts_cv <- tidy(ts_cv)
#' ggplot(ts_cv, aes(x = Resample, y = factor(Row), fill = Data)) +
#'   geom_tile() +
#'   scale_fill_brewer()
#' @export
tidy.rsplit <- function(x, unique_ind = TRUE, ...) {
  if (unique_ind) x$in_id <- unique(x$in_id)
  out <- tibble(
    Row = c(x$in_id, complement(x)),
    Data = rep(
      c("Analysis", "Assessment"),
      c(length(x$in_id), length(complement(x)))
    )
  )
  out <- dplyr::arrange(.data = out, Data, Row)
  out
}

#' @rdname tidy.rsplit
#' @export
tidy.rset <- function(x, ...) {
  dots <- list(...)
  unique_ind <- dplyr::if_else(is.null(dots$unique_ind),
    TRUE,
    dots$unique_ind
  )
  stacked <- purrr::map(x$splits, tidy, unique_ind = unique_ind)
  for (i in seq_along(stacked)) {
    stacked[[i]]$Resample <- x$id[i]
  }
  stacked <- dplyr::bind_rows(stacked)
  stacked <- dplyr::arrange(.data = stacked, Data, Row)
  stacked
}
#' @rdname tidy.rsplit
#' @export
tidy.vfold_cv <- function(x, ...) {
  stacked <- purrr::map(x$splits, tidy)
  for (i in seq_along(stacked)) {
    if (attr(x, "repeats") > 1) {
      stacked[[i]]$Repeat <- x$id[i]
      stacked[[i]]$Fold <- x$id2[i]
    } else {
      stacked[[i]]$Fold <- x$id[i]
    }
  }
  stacked <- dplyr::bind_rows(stacked)
  stacked <- dplyr::arrange(.data = stacked, Data, Row)
  stacked
}

#' @rdname tidy.rsplit
#' @export
tidy.nested_cv <- function(x, ...) {
  x$inner_tidy <- purrr::map(x$inner_resamples, tidy_wrap)
  inner_tidy <- tidyr::unnest(x, inner_tidy)
  class(x) <- class(x)[class(x) != "nested_cv"]
  outer_tidy <- tidy(x)
  id_cols <- names(outer_tidy)
  id_cols <- id_cols[!(id_cols %in% c("Row", "Data"))]

  inner_id <- grep("^id", names(inner_tidy))
  if (length(inner_id) != length(id_cols)) {
    rlang::abort("Cannot merge tidy data sets")
  }
  names(inner_tidy)[inner_id] <- id_cols
  full_join(outer_tidy, inner_tidy, by = id_cols)
}

tidy_wrap <- function(x) {
  x <- tidy(x)
  names(x) <- paste0("inner_", names(x))
  x
}

#' Bootstrap Sampling
#'
#' A bootstrap sample is a sample that is the same size as the original data
#'  set that is made using replacement. This results in analysis samples that
#'  have multiple replicates of some of the original rows of the data. The
#'  assessment set is defined as the rows of the original data that were not
#'  included in the bootstrap sample. This is often referred to as the
#'  "out-of-bag" (OOB) sample.
#' @details The argument `apparent` enables the option of an additional
#'  "resample" where the analysis and assessment data sets are the same as the
#'  original data set. This can be required for some types of analysis of the
#'  bootstrap results.
#'
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param times The number of bootstrap samples.
#' @param apparent A logical. Should an extra resample be added where the
#'  analysis and holdout subset are the entire data set. This is required for
#'  some estimators used by the [summary()] function that require the apparent
#'  error rate.
#' @export
#' @return A tibble with classes `bootstraps`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and a
#'  column called `id` that has a character string with the resample identifier.
#' @examplesIf rlang::is_installed("modeldata")
#' bootstraps(mtcars, times = 2)
#' bootstraps(mtcars, times = 2, apparent = TRUE)
#'
#' library(purrr)
#' library(modeldata)
#' data(wa_churn)
#'
#' set.seed(13)
#' resample1 <- bootstraps(wa_churn, times = 3)
#' map_dbl(
#'   resample1$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' resample2 <- bootstraps(wa_churn, strata = churn, times = 3)
#' map_dbl(
#'   resample2$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' resample3 <- bootstraps(wa_churn, strata = tenure, breaks = 6, times = 3)
#' map_dbl(
#'   resample3$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#' @export
bootstraps <- function(
  data,
  times = 25,
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  apparent = FALSE,
  ...
) {
  check_dots_empty()

  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) strata <- NULL
  }

  check_strata(strata, data)

  split_objs <-
    boot_splits(
      data = data,
      times = times,
      strata = strata,
      breaks = breaks,
      pool = pool
    )

  if (apparent) {
    split_objs <- bind_rows(split_objs, apparent(data))
  }

  if (!is.null(strata)) {
    names(strata) <- NULL
  }
  boot_att <- list(
    times = times,
    apparent = apparent,
    strata = strata,
    breaks = breaks,
    pool = pool
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = boot_att,
    subclass = c("bootstraps", "rset")
  )
}

# Get the indices of the analysis set from the analysis set (= bootstrap sample)
boot_complement <- function(ind, n) {
  list(analysis = ind, assessment = NA)
}

boot_splits <- function(
  data,
  times = 25,
  strata = NULL,
  breaks = 4,
  pool = 0.1
) {
  n <- nrow(data)

  if (is.null(strata)) {
    indices <- purrr::map(rep(n, times), sample, replace = TRUE)
  } else {
    stratas <- tibble::tibble(
      idx = 1:n,
      strata = make_strata(
        getElement(data, strata),
        breaks = breaks,
        pool = pool
      )
    )
    stratas <- split_unnamed(stratas, stratas$strata)
    stratas <-
      purrr::map(
        stratas,
        strat_sample,
        prop = 1,
        times = times,
        replace = TRUE
      ) |>
      list_rbind()
    indices <- split_unnamed(stratas$idx, stratas$rs_id)
  }

  indices <- lapply(indices, boot_complement, n = n)

  split_objs <-
    purrr::map(indices, make_splits, data = data, class = "boot_split")

  all_assessable <- purrr::map(split_objs, function(x) nrow(assessment(x)))
  if (any(all_assessable == 0)) {
    cli_warn(
      "Some assessment sets contained zero rows.",
      class = "rsample_bootstrap_empty_assessment",
      call = rlang::caller_env()
    )
  }

  list(
    splits = split_objs,
    id = names0(length(split_objs), "Bootstrap")
  )
}

#' Group Bootstraps
#'
#' Group bootstrapping creates splits of the data based
#'  on some grouping variable (which may have more than a single row
#'  associated with it). A common use of this kind of resampling is when you
#'  have repeated measures of the same subject.
#'  A bootstrap sample is a sample that is the same size as the original data
#'  set that is made using replacement. This results in analysis samples that
#'  have multiple replicates of some of the original rows of the data. The
#'  assessment set is defined as the rows of the original data that were not
#'  included in the bootstrap sample. This is often referred to as the
#'  "out-of-bag" (OOB) sample.
#' @details The argument `apparent` enables the option of an additional
#'  "resample" where the analysis and assessment data sets are the same as the
#'  original data set. This can be required for some types of analysis of the
#'  bootstrap results.
#'
#' @inheritParams bootstraps
#' @inheritParams make_groups
#' @export
#' @return An tibble with classes `group_bootstraps` `bootstraps`, `rset`,
#'  `tbl_df`, `tbl`, and `data.frame`. The results include a column for the data
#'  split objects and a column called `id` that has a character string with the
#'  resample identifier.
#' @examplesIf rlang::is_installed("modeldata")
#' data(ames, package = "modeldata")
#'
#' set.seed(13)
#' group_bootstraps(ames, Neighborhood, times = 3)
#' group_bootstraps(ames, Neighborhood, times = 3, apparent = TRUE)
#'
#' @export
group_bootstraps <- function(
  data,
  group,
  times = 25,
  apparent = FALSE,
  ...,
  strata = NULL,
  pool = 0.1
) {
  check_dots_empty()

  group <- validate_group({{ group }}, data)

  if (!missing(strata)) {
    strata <- check_grouped_strata({{ group }}, {{ strata }}, pool, data)
  }

  split_objs <-
    group_boot_splits(
      data = data,
      group = group,
      times = times,
      strata = strata,
      pool = pool
    )

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.
  split_objs$splits <- map(split_objs$splits, rm_out)

  if (apparent) {
    split_objs <- bind_rows(split_objs, apparent(data))
  }

  # This is needed for printing checks; strata can't be missing
  if (is.null(strata)) {
    strata <- FALSE
  }
  boot_att <- list(
    times = times,
    apparent = apparent,
    strata = strata,
    pool = pool,
    group = group
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = boot_att,
    subclass = c("group_bootstraps", "bootstraps", "group_rset", "rset")
  )
}

group_boot_splits <- function(
  data,
  group,
  times = 25,
  strata = NULL,
  pool = 0.1
) {
  group <- getElement(data, group)
  if (!is.null(strata)) {
    strata <- getElement(data, strata)
    strata <- as.character(strata)
    strata <- make_strata(strata, pool = pool)
  }

  n <- nrow(data)

  indices <- make_groups(
    data,
    group,
    times,
    balance = "prop",
    prop = 1,
    replace = TRUE,
    strata = strata
  )

  indices <- lapply(indices, boot_complement, n = n)
  split_objs <- purrr::map(
    indices,
    make_splits,
    data = data,
    class = c("group_boot_split", "boot_split")
  )

  all_assessable <- purrr::map(split_objs, function(x) nrow(assessment(x)))
  if (any(all_assessable == 0)) {
    cli_warn(
      c(
        "Some assessment sets contained zero rows.",
        i = "Consider using a non-grouped resampling method."
      ),
      call = rlang::caller_env()
    )
  }

  list(
    splits = split_objs,
    id = names0(length(split_objs), "Bootstrap")
  )
}

test_that("default param", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20])
  set.seed(11)
  rs1 <- bootstraps(dat1)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == nrow(dat1)))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})

test_that("apparent", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20])
  rs2 <- bootstraps(dat1, apparent = TRUE)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == nrow(dat1)))
  expect_true(all(sizes2$assessment[nrow(sizes2)] == nrow(dat1)))
  expect_equal(sizes2$assessment[sizes2$id == "Apparent"], nrow(dat1))
  res2 <-
    as.data.frame(rs2$splits[[nrow(sizes2)]], data = "assessment")
  expect_equal(res2, dat1)
})

test_that("groups", {
  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:10, 2))
  set.seed(11)
  rs1 <- group_bootstraps(dat1, c)
  sizes1 <- dim_rset(rs1)

  expect_true(all(sizes1$analysis == nrow(dat1)))
  same_data <-
    purrr::map_lgl(rs1$splits, function(x) {
      all.equal(x$data, dat1)
    })
  expect_true(all(same_data))

  good_holdout <- purrr::map_lgl(
    rs1$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  dat1 <- data.frame(a = 1:20, b = letters[1:20], c = rep(1:4, 5))
  rs2 <- bootstraps(dat1, apparent = TRUE)
  sizes2 <- dim_rset(rs2)

  expect_true(all(sizes2$analysis == nrow(dat1)))
  expect_true(all(sizes2$assessment[nrow(sizes2)] == nrow(dat1)))
  expect_equal(sizes2$assessment[sizes2$id == "Apparent"], nrow(dat1))
  res2 <-
    as.data.frame(rs2$splits[[nrow(sizes2)]], data = "assessment")
  expect_equal(res2, dat1)
})

test_that("strata", {
  set.seed(11)
  rs4 <- bootstraps(warpbreaks, strata = "tension")
  sizes4 <- dim_rset(rs4)

  expect_true(all(sizes4$analysis == nrow(warpbreaks)))

  rate <- purrr::map_dbl(
    rs4$splits,
    function(x) {
      dat <- as.data.frame(x)$tension
      mean(dat == "M")
    }
  )
  expect_true(length(unique(rate)) == 1)

  good_holdout <- purrr::map_lgl(
    rs4$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))

  rs5 <- bootstraps(warpbreaks, apparent = TRUE, strata = "tension")
  sizes5 <- dim_rset(rs5)

  expect_true(all(sizes5$analysis == nrow(warpbreaks)))
  expect_true(all(sizes5$assessment[nrow(sizes5)] == nrow(warpbreaks)))
  expect_equal(sizes5$assessment[sizes5$id == "Apparent"], nrow(warpbreaks))
  res5 <-
    as.data.frame(rs5$splits[[nrow(sizes5)]], data = "assessment")
  expect_equal(res5, warpbreaks)
})

test_that("grouping -- strata", {
  set.seed(11)

  n_common_class <- 70
  n_rare_class <- 30

  group_table <- tibble(
    group = 1:100,
    outcome = sample(c(rep(0, n_common_class), rep(1, n_rare_class)))
  )
  observation_table <- tibble(
    group = sample(1:100, 5e4, replace = TRUE),
    observation = 1:5e4
  )
  sample_data <- dplyr::full_join(
    group_table,
    observation_table,
    by = "group",
    multiple = "all"
  )
  rs4 <- group_bootstraps(sample_data, group, times = 5, strata = outcome)
  sizes4 <- dim_rset(rs4)
  expect_snapshot(sizes4)

  rate <- purrr::map_dbl(
    rs4$splits,
    function(x) {
      dat <- as.data.frame(x)$outcome
      mean(dat == "1")
    }
  )
  expect_equal(mean(rate), 0.3, tolerance = 1e-2)

  good_holdout <- purrr::map_lgl(
    rs4$splits,
    function(x) {
      length(intersect(x$in_ind, x$out_id)) == 0
    }
  )
  expect_true(all(good_holdout))
})


test_that("bad args", {
  expect_snapshot(error = TRUE, {
    bootstraps(warpbreaks, strata = warpbreaks$tension)
  })
  expect_snapshot(error = TRUE, {
    bootstraps(warpbreaks, strata = c("tension", "wool"))
  })
  set.seed(1)
  expect_snapshot(
    group_bootstraps(warpbreaks, tension)
  )
  expect_snapshot(
    bootstraps(mtcars[2, ])
  )
})


test_that("printing", {
  skip_if(new_rng_snapshots)
  expect_snapshot({
    set.seed(11)
    bootstraps(warpbreaks)
  })
})


test_that("rsplit labels", {
  rs <- bootstraps(warpbreaks)
  all_labs <- purrr::map(rs$splits, labels) |>
    list_rbind()
  original_id <- rs[, grepl("^id", names(rs))]
  expect_equal(all_labs, original_id)
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom lifecycle deprecated
## usethis namespace: end
NULL

#------------------------------------------------------------------------------#

#' @importFrom utils globalVariables
#' @importFrom purrr map map2 map_dbl pluck map_lgl list_rbind
#' @importFrom tibble tibble is_tibble as_tibble obj_sum
#' @importFrom pillar type_sum
#' @importFrom tidyr unnest
#' @importFrom dplyr select %>% bind_cols bind_rows arrange_ arrange full_join
#' @importFrom dplyr mutate last ungroup group_by inner_join summarize do case_when
#' @importFrom vctrs vec_unique_count vec_count vec_slice vec_in vec_ptype_abbr
#' @importFrom rlang !! is_call is_string enquo quos exec is_list abort warn .data %||%
#' @importFrom rlang check_dots_empty
#' @importFrom methods formalArgs
#' @importFrom stats quantile setNames qnorm pnorm dist
#' @importFrom tidyselect vars_select one_of
#' @importFrom furrr future_map

#------------------------------------------------------------------------------#

utils::globalVariables(
  c(
    "model", "splits", "statistic", "Data", "Row", "id", ".", ".estimate",
    ".lower", ".upper", "Z0", "Za", "Zl", "Zu", "a", "cubed", "estimate",
    "orig", "p0", "squared", "term", "theta_0", "loo", "n", "..index", "models",
    paste0("predictor_", gsub(" ", "0", format(1:20))), "linear_pred",
    "non_linear_1", "non_linear_2", "non_linear_3", "outcome", "true_prob",
    "two_factor_1", "two_factor_2", "rand"
  )
)

#' Create a Validation Set
#'
#' `validation_split()` takes a single random sample (without replacement) of
#'  the original data set to be used for analysis. All other data points are
#'  added to the assessment set (to be used as the validation set).
#'  `validation_time_split()` does the same, but takes the _first_ `prop` samples
#'  for training, instead of a random selection.
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param prop The proportion of data to be retained for modeling/analysis.
#' @export
#' @return An tibble with classes `validation_split`, `rset`, `tbl_df`, `tbl`,
#'  and `data.frame`. The results include a column for the data split objects
#'  and a column called `id` that has a character string with the resample
#'  identifier.
#' @examples
#' validation_split(mtcars, prop = .9)
#'
#' data(drinks, package = "modeldata")
#' validation_time_split(drinks)
#' @export
validation_split <- function(data, prop = 3 / 4,
                             strata = NULL, breaks = 4, pool = 0.1, ...) {
  if (!missing(strata)) {
    strata <- tidyselect::vars_select(names(data), !!enquo(strata))
    if (length(strata) == 0) {
      strata <- NULL
    }
  }

  strata_check(strata, data)

  split_objs <-
    mc_splits(
      data = data,
      prop = prop,
      times = 1,
      strata = strata,
      breaks = breaks,
      pool = pool
    )

  ## We remove the holdout indices since it will save space and we can
  ## derive them later when they are needed.

  split_objs$splits <- map(split_objs$splits, rm_out)
  class(split_objs$splits[[1]]) <- c("val_split", "rsplit")

  val_att <- list(
    prop = prop,
    strata = !is.null(strata)
  )

  new_rset(
    splits = split_objs$splits,
    ids = "validation",
    attrib = val_att,
    subclass = c("validation_split", "rset")
  )
}

#' @rdname validation_split
#' @inheritParams vfold_cv
#' @inheritParams initial_time_split
#' @export
validation_time_split <- function(data, prop = 3 / 4, lag = 0, ...) {
  if (!is.numeric(prop) | prop >= 1 | prop <= 0) {
    rlang::abort("`prop` must be a number on (0, 1).")
  }

  if (!is.numeric(lag) | !(lag %% 1 == 0)) {
    rlang::abort("`lag` must be a whole number.")
  }

  n_train <- floor(nrow(data) * prop)

  if (lag > n_train) {
    rlang::abort("`lag` must be less than or equal to the number of training observations.")
  }

  split <- rsplit(data, 1:n_train, (n_train + 1 - lag):nrow(data))
  split <- rm_out(split)
  class(split) <- c("val_split", "rsplit")
  splits <- list(split)

  val_att <- list(prop = prop, strata = FALSE)

  new_rset(
    splits = splits,
    ids = "validation",
    attrib = val_att,
    subclass = c("validation_split", "rset")
  )
}

#' Constructor for new rset objects
#' @param splits A list column of `rsplits` or a tibble with a single column
#'  called "splits" with a list column of `rsplits`.
#' @param ids A character vector or a tibble with one or more columns that
#' begin with "id".
#' @param attrib An optional named list of attributes to add to the object.
#' @param subclass A character vector of subclasses to add.
#' @return An `rset` object.
#' @keywords internal
#' @export
new_rset <-  function(splits, ids, attrib = NULL,
                      subclass = character()) {
  stopifnot(is.list(splits))
  if (!is_tibble(ids)) {
    ids <- tibble(id = ids)
  } else {
    if (!all(grepl("^id", names(ids)))) {
      rlang::abort("The `ids` tibble column names should start with 'id'.")
    }
  }
  either_type <- function(x)
    is.character(x) | is.factor(x)
  ch_check <- vapply(ids, either_type, c(logical = TRUE))
  if (!all(ch_check)) {
    rlang::abort("All ID columns should be character or factor vectors.")
  }

  if (!is_tibble(splits)) {
    splits <- tibble(splits = splits)
  } else {
    if (ncol(splits) > 1 | names(splits)[1] != "splits") {
      rlang::abort(
        "The `splits` tibble should have a single column named `splits`."
      )
    }
  }

  where_rsplits <- vapply(splits[["splits"]], is_rsplit, logical(1))

  if (!all(where_rsplits)) {
    rlang::abort("Each element of `splits` must be an `rsplit` object.")
  }

  if (nrow(ids) != nrow(splits)) {
    rlang::abort("Split and ID vectors have different lengths.")
  }

  # Create another element to the splits that is a tibble containing
  # an identifier for each id column so that, in isolation, the resample
  # id can be known just based on the `rsplit` object. This can then be
  # accessed using the `labels` method for `rsplits`

  splits$splits <- map2(
    splits$splits,
    split_unnamed(ids, rlang::seq2(1L, nrow(ids))),
    add_id
  )

  res <- bind_cols(splits, ids)

  if (!is.null(attrib)) {
    if (any(names(attrib) == "")) {
      rlang::abort("`attrib` should be a fully named list.")
    }
    for (i in names(attrib)) {
      attr(res, i) <- attrib[[i]]
    }
  }

  if (length(subclass) > 0) {
    res <- add_class(res, cls = subclass, at_end = FALSE)
  }

  res
}

add_id <- function(split, id) {
  split$id <- id
  split
}

# ------------------------------------------------------------------------------

#' @export
`[.rset` <- function(x, i, j, drop = FALSE, ...) {
  out <- NextMethod()
  rset_reconstruct(out, x)
}

# ------------------------------------------------------------------------------

#' @export
`names<-.rset` <- function(x, value) {
  out <- NextMethod()
  rset_reconstruct(out, x)
}

## The `pretty` methods below are good for when you need to
## textually describe the resampling procedure. Note that they
## can have more than one element (in the case of nesting)

#' @export
pretty.vfold_cv <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(details$v, "-fold cross-validation")
  if (details$repeats > 1) {
    res <- paste(res, "repeated", details$repeats, "times")
  }
  if (details$strata) {
    res <- paste(res, "using stratification")
  }
  res
}

#' @export
pretty.loo_cv <- function(x, ...) {
  "Leave-one-out cross-validation"
}

#' @export
pretty.apparent <- function(x, ...) {
  "Apparent sampling"
}

#' @export
pretty.rolling_origin <- function(x, ...) {
  "Rolling origin forecast resampling"
}

#' @export
pretty.sliding_window <- function(x, ...) {
  "Sliding window resampling"
}

#' @export
pretty.sliding_index <- function(x, ...) {
  "Sliding index resampling"
}

#' @export
pretty.sliding_period <- function(x, ...) {
  "Sliding period resampling"
}

#' @export
pretty.mc_cv <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Monte Carlo cross-validation (",
    signif(details$prop, 2),
    "/",
    signif(1 - details$prop, 2),
    ") with ",
    details$times,
    " resamples "
  )
  if (details$strata) {
    res <- paste(res, "using stratification")
  }
  res
}

#' @export
pretty.validation_split <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Validation Set Split (",
    signif(details$prop, 2),
    "/",
    signif(1 - details$prop, 2),
    ") "
  )
  if (details$strata) {
    res <- paste(res, "using stratification")
  }
  res
}

#' @export
pretty.nested_cv <- function(x, ...) {
  details <- attributes(x)

  if (is_call(details$outside)) {
    class(x) <- class(x)[!(class(x) == "nested_cv")]
    outer_label <- pretty(x)
  } else {
    outer_label <- paste0("`", deparse(details$outside), "`")
  }

  inner_label <- if (is_call(details$inside)) {
    pretty(x$inner_resamples[[1]])
  } else {
    paste0("`", deparse(details$inside), "`")
  }

  res <- c(
    "Nested resampling:",
    paste(" outer:", outer_label),
    paste(" inner:", inner_label)
  )
  res
}

#' @export
pretty.bootstraps <- function(x, ...) {
  details <- attributes(x)
  res <- "Bootstrap sampling"
  if (details$strata) {
    res <- paste(res, "using stratification")
  }
  if (details$apparent) {
    res <- paste(res, "with apparent sample")
  }
  res
}

#' @export
pretty.permutations <- function(x, ...) {
  details <- attributes(x)
  res <- "Permutation sampling"
  if (details$apparent) {
    res <- paste(res, "with apparent sample")
  }
  res
}

#' @export
pretty.group_vfold_cv <- function(x, ...) {
  details <- attributes(x)
  paste0("Group ", details$v, "-fold cross-validation")
}

#' @export
pretty.manual_rset <- function(x, ...) {
  "Manual resampling"
}

# The print methods below control the display of rset objects

#' @export
print.apparent <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("apparent", "rset"))]
  print(x, ...)
}

#' @export
print.bootstraps <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("bootstraps", "rset"))]
  print(x, ...)
}

#' @export
print.group_vfold_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("group_vfold_cv", "rset"))]
  print(x, ...)
}

#' @export
print.initial_split <- function(x, ...) {
  out_char <-
    if (is_missing_out_id(x)) {
      paste(length(complement(x)))
    } else {
      paste(length(x$out_id))
    }

  cat("<Training/Testing/Total>\n")
  cat("<",
      length(x$in_id), "/",
      out_char, "/",
      nrow(x$data), ">\n",
      sep = ""
  )
}

#' @export
print.loo_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("loo_cv", "rset"))]
  print(x, ...)
}

#' @export
print.manual_rset <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("manual_rset", "rset"))]
  print(x, ...)
}

#' @export
print.mc_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("mc_cv", "rset"))]
  print(x, ...)
}

#' @export
print.nested_cv <- function(x, ...) {
  char_x <- paste("#", pretty(x))
  cat(char_x, sep = "\n")
  class(x) <- class(tibble())
  print(x, ...)
}


#' @export
print.permutations <- function(x, ...) {
  shuffled_cols <- paste(names(attr(x, "col_id")), collapse = ", ")
  cat("#", pretty(x), "\n")
  cat("# Permuted columns: [", shuffled_cols, "] \n", sep = "")
  class(x) <- class(x)[!(class(x) %in% c("permutations", "rset"))]
  print(x, ...)
}

#' @export
print.rolling_origin <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("rolling_origin", "rset"))]
  print(x, ...)
}

#' @export
print.sliding_window <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("sliding_window", "rset"))]
  print(x, ...)
}

#' @export
print.sliding_index <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("sliding_index", "rset"))]
  print(x, ...)
}

#' @export
print.sliding_period <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("sliding_period", "rset"))]
  print(x, ...)
}

#' @export
print.validation_split <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("validation_split", "rset"))]
  print(x, ...)
}

#' @export
print.val_split <- function(x, ...) {
  if (is_missing_out_id(x)) {
    out_char <- paste(length(complement(x)))
  } else {
    out_char <- paste(length(x$out_id))
  }

  cat("<Training/Validation/Total>\n")
  cat("<",
      length(x$in_id), "/",
      out_char, "/",
      nrow(x$data), ">\n",
      sep = ""
  )
}

#' @export
print.vfold_cv <- function(x, ...) {
  cat("# ", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("vfold_cv", "rset"))]
  print(x, ...)
}

#' Bootstrap Sampling
#'
#' A bootstrap sample is a sample that is the same size as the original data
#'  set that is made using replacement. This results in analysis samples that
#'  have multiple replicates of some of the original rows of the data. The
#'  assessment set is defined as the rows of the original data that were not
#'  included in the bootstrap sample. This is often referred to as the
#'  "out-of-bag" (OOB) sample.
#' @details The argument `apparent` enables the option of an additional
#'  "resample" where the analysis and assessment data sets are the same as the
#'  original data set. This can be required for some types of analysis of the
#'  bootstrap results.
#'
#' @template strata_details
#' @inheritParams vfold_cv
#' @inheritParams make_strata
#' @param times The number of bootstrap samples.
#' @param apparent A logical. Should an extra resample be added where the
#'  analysis and holdout subset are the entire data set. This is required for
#'  some estimators used by the `summary` function that require the apparent
#'  error rate.
#' @export
#' @return An tibble with classes `bootstraps`, `rset`, `tbl_df`, `tbl`, and
#'  `data.frame`. The results include a column for the data split objects and a
#'  column called `id` that has a character string with the resample identifier.
#' @examples
#' bootstraps(mtcars, times = 2)
#' bootstraps(mtcars, times = 2, apparent = TRUE)
#'
#' library(purrr)
#' library(modeldata)
#' data(wa_churn)
#'
#' set.seed(13)
#' resample1 <- bootstraps(wa_churn, times = 3)
#' map_dbl(
#'   resample1$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' resample2 <- bootstraps(wa_churn, strata = churn, times = 3)
#' map_dbl(
#'   resample2$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#'
#' set.seed(13)
#' resample3 <- bootstraps(wa_churn, strata = tenure, breaks = 6, times = 3)
#' map_dbl(
#'   resample3$splits,
#'   function(x) {
#'     dat <- as.data.frame(x)$churn
#'     mean(dat == "Yes")
#'   }
#' )
#' @export
bootstraps <-
  function(data,
           times = 25,
           strata = NULL,
           breaks = 4,
           pool = 0.1,
           apparent = FALSE,
           ...) {
    if (!missing(strata)) {
      strata <- tidyselect::vars_select(names(data), !!enquo(strata))
      if (length(strata) == 0) strata <- NULL
    }

    strata_check(strata, data)

    split_objs <-
      boot_splits(
        data = data,
        times = times,
        strata = strata,
        breaks = breaks,
        pool = pool
      )
    if (apparent) {
      split_objs <- bind_rows(split_objs, apparent(data))
    }

    boot_att <- list(
      times = times,
      apparent = apparent,
      strata = !is.null(strata)
    )

    new_rset(
      splits = split_objs$splits,
      ids = split_objs$id,
      attrib = boot_att,
      subclass = c("bootstraps", "rset")
    )
  }

# Get the indices of the analysis set from the analysis set (= bootstrap sample)
boot_complement <- function(ind, n) {
  list(analysis = ind, assessment = NA)
}

boot_splits <-
  function(data,
           times = 25,
           strata = NULL,
           breaks = 4,
           pool = 0.1) {
    n <- nrow(data)

    if (is.null(strata)) {
      indices <- purrr::map(rep(n, times), sample, replace = TRUE)
    } else {
      stratas <- tibble::tibble(
        idx = 1:n,
        strata = make_strata(getElement(data, strata),
          breaks = breaks,
          pool = pool
        )
      )
      stratas <- split_unnamed(stratas, stratas$strata)
      stratas <-
        purrr::map_df(
          stratas,
          strat_sample,
          prop = 1,
          times = times,
          replace = TRUE
        )
      indices <- split_unnamed(stratas$idx, stratas$rs_id)
    }

    indices <- lapply(indices, boot_complement, n = n)

    split_objs <-
      purrr::map(indices, make_splits, data = data, class = "boot_split")
    list(
      splits = split_objs,
      id = names0(length(split_objs), "Bootstrap")
    )
  }

#' Find Labels from rset Object
#'
#' Produce a vector of resampling labels (e.g. "Fold1") from
#' an `rset` object. Currently, [nested_cv()] is not supported.
#'
#' @param object An `rset` object.
#' @param make_factor A logical for whether the results should be
#'  a character or a factor.
#' @param ... Not currently used.
#' @return A single character or factor vector.
#' @export
#' @examples
#' labels(vfold_cv(mtcars))
labels.rset <- function(object, make_factor = FALSE, ...) {
  if (inherits(object, "nested_cv")) {
    cli_abort("{.arg labels} not implemented for nested resampling")
  }
  if (make_factor) {
    as.factor(object$id)
  } else {
    as.character(object$id)
  }
}

#' @rdname labels.rset
#' @export
labels.vfold_cv <- function(object, make_factor = FALSE, ...) {
  if (inherits(object, "nested_cv")) {
    cli_abort("{.arg labels} not implemented for nested resampling")
  }
  is_repeated <- attr(object, "repeats") > 1
  if (is_repeated) {
    out <- as.character(paste(object$id, object$id2, sep = "."))
  } else {
    out <- as.character(object$id)
  }
  if (make_factor) {
    out <- as.factor(out)
  }
  out
}

#' Find Labels from rsplit Object
#'
#' Produce a tibble of identification variables so that single
#'  splits can be linked to a particular resample.
#'
#' @param object An `rsplit` object
#' @param ... Not currently used.
#' @return A tibble.
#' @seealso add_resample_id
#' @export
#' @examples
#' cv_splits <- vfold_cv(mtcars)
#' labels(cv_splits$splits[[1]])
labels.rsplit <- function(object, ...) {
  out <- if ("id" %in% names(object)) {
    object$id
  } else {
    tibble()
  }
  out
}

#' Augment a data set with resampling identifiers
#'
#' For a data set, `add_resample_id()` will add at least one new column that
#'  identifies which resample that the data came from. In most cases, a single
#'  column is added but for some resampling methods, two or more are added.
#' @param .data A data frame.
#' @param split A single `rset` object.
#' @param dots A single logical: should the id columns be prefixed with a "."
#'  to avoid name conflicts with `.data`?
#' @return An updated data frame.
#' @examples
#' library(dplyr)
#'
#' set.seed(363)
#' car_folds <- vfold_cv(mtcars, repeats = 3)
#'
#' analysis(car_folds$splits[[1]]) |>
#'   add_resample_id(car_folds$splits[[1]]) |>
#'   head()
#'
#' car_bt <- bootstraps(mtcars)
#'
#' analysis(car_bt$splits[[1]]) |>
#'   add_resample_id(car_bt$splits[[1]]) |>
#'   head()
#' @seealso labels.rsplit
#' @export
add_resample_id <- function(.data, split, dots = FALSE) {
  if (!inherits(dots, "logical") || length(dots) > 1) {
    cli_abort("{.arg dots} should be a single logical.")
  }
  if (!inherits(.data, "data.frame")) {
    cli_abort("{.arg .data} should be a data frame.")
  }
  if (!inherits(split, "rsplit")) {
    cli_abort("{.arg split} should be a single {.cls rset} object.")
  }
  labs <- labels(split)

  if (!tibble::is_tibble(labs) && nrow(labs) == 1) {
    cli_abort("{.arg split} should be a single {.cls rset} object.")
  }

  if (dots) {
    colnames(labs) <- paste0(".", colnames(labs))
  }

  cbind(.data, labs)
}

#' Constructor for new rset objects
#' @param splits A list column of `rsplits` or a tibble with a single column
#'  called "splits" with a list column of `rsplits`.
#' @param ids A character vector or a tibble with one or more columns that
#' begin with "id".
#' @param attrib An optional named list of attributes to add to the object.
#' @param subclass A character vector of subclasses to add.
#' @return An `rset` object.
#' @details Once the new `rset` is constructed, an additional attribute called
#' "fingerprint" is added that is a hash of the `rset`. This can be used to
#' make sure other objects have the exact same resamples.
#' @keywords internal
#' @export
new_rset <- function(splits, ids, attrib = NULL, subclass = character()) {
  stopifnot(is.list(splits))
  if (!is_tibble(ids)) {
    ids <- tibble(id = ids)
  } else {
    if (!all(grepl("^id", names(ids)))) {
      cli_abort("The {.code id} tibble column names should start with 'id'.")
    }
  }
  either_type <- function(x) {
    is.character(x) | is.factor(x)
  }
  ch_check <- vapply(ids, either_type, c(logical = TRUE))
  if (!all(ch_check)) {
    cli_abort("{.strong All} ID columns should be character or factor vectors.")
  }

  if (!is_tibble(splits)) {
    splits <- tibble(splits = splits)
  } else {
    if (ncol(splits) > 1 | names(splits)[1] != "splits") {
      cli_abort(
        "The {.var splits} tibble should have a single column named {.code splits}."
      )
    }
  }

  where_rsplits <- vapply(splits[["splits"]], is_rsplit, logical(1))

  if (!all(where_rsplits)) {
    cli_abort("Each element of {.arg splits} must be an {.cls rsplit} object.")
  }

  if (nrow(ids) != nrow(splits)) {
    cli_abort("Split and ID vectors have different lengths.")
  }

  # Create another element to the splits that is a tibble containing
  # an identifier for each id column so that, in isolation, the resample
  # id can be known just based on the `rsplit` object. This can then be
  # accessed using the `labels` method for `rsplits`

  splits$splits <- map2(
    splits$splits,
    split_unnamed(ids, rlang::seq2(1L, nrow(ids))),
    add_id
  )

  res <- bind_cols(splits, ids)

  if (!is.null(attrib)) {
    if (any(names(attrib) == "")) {
      cli_abort("{.arg attrib} should be a fully named list.")
    }
    for (i in names(attrib)) {
      attr(res, i) <- attrib[[i]]
    }
  }

  if (length(subclass) > 0) {
    res <- add_class(res, cls = subclass)
  }

  fingerprint <- map(res$splits, function(x) list(x$in_id, x$out_id))
  fingerprint <- rlang::hash(fingerprint)
  attr(res, "fingerprint") <- fingerprint

  res
}

add_id <- function(split, id) {
  split$id <- id
  split
}

# ------------------------------------------------------------------------------

#' @export
`[.rset` <- function(x, i, j, drop = FALSE, ...) {
  out <- NextMethod()
  rset_reconstruct(out, x)
}

# ------------------------------------------------------------------------------

#' @export
`names<-.rset` <- function(x, value) {
  out <- NextMethod()
  rset_reconstruct(out, x)
}

#' A convenience function for confidence intervals with linear-ish parametric models
#'
#' @param formula An R model formula with one outcome and at least one predictor.
#' @param data A data frame.
#' @param model_fn The model to fit. Allowable values are "lm", "glm",
#'  "survreg", and "coxph". The latter two require that the `survival` package
#'  be installed.
#' @param type The type of bootstrap confidence interval. Values of "student-t" and
#' "percentile" are allowed.
#' @param times A single integer for the number of bootstrap samples. If left
#' NULL, 1,001 are used for t-intervals and 2,001 for percentile intervals.
#' @param alpha Level of significance.
#' @param filter A logical expression used to remove rows from the final result, or `NULL` to keep all rows.
#' @param keep_reps Should the individual parameter estimates for each bootstrap
#' sample be retained?
#' @param ... Options to pass to the model function (such as `family` for `glm()`).
#' @return A tibble with columns "term", ".lower", ".estimate", ".upper",
#' ".alpha", and ".method". If `keep_reps = TRUE`, an additional list column
#' called ".replicates" is also returned.
#' @export
#' @seealso [int_pctl()], [int_t()]
#' @references
#' Davison, A., & Hinkley, D. (1997). _Bootstrap Methods and their
#'  Application_. Cambridge: Cambridge University Press.
#'  doi:10.1017/CBO9780511802843
#'
#' _Bootstrap Confidence Intervals_,
#' \url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}
#' @examples
#' \donttest{
#' set.seed(1)
#' reg_intervals(mpg ~ I(1 / sqrt(disp)), data = mtcars)
#'
#' set.seed(1)
#' reg_intervals(mpg ~ I(1 / sqrt(disp)), data = mtcars, keep_reps = TRUE)
#' }
reg_intervals <-
  function(formula, data, model_fn = "lm", type = "student-t", times = NULL,
           alpha = 0.05, filter = term != "(Intercept)", keep_reps = FALSE, ...) {
    model_fn <- rlang::arg_match(model_fn, c("lm", "glm", "survreg", "coxph"))
    type <- rlang::arg_match(type, c("student-t", "percentile"))

    filter <- rlang::enexpr(filter)

    if (is.null(times)) {
      if (type == "student-t") {
        times <- 1001
      } else {
        times <- 2001
      }
    } else {
      times <- times[1]
      if (!is.numeric(times)) {
        rlang::abort("'times' should be a single integer.")
      }
    }

    if (length(alpha) != 1 || !is.numeric(alpha)) {
      abort("`alpha` must be a single numeric value.")
    }

    if (model_fn %in% c("survreg", "coxph")) {
      pkg <- "survival"
      rlang::check_installed("survival")
    } else {
      pkg <- NULL
    }
    fn_call <- rlang::call2(model_fn,
      formula = formula,
      data = rlang::expr(data), .ns = pkg, ...
    )

    bt <- rsample::bootstraps(data, times = times, apparent = type %in% c("student-t"))
    bt <-
      dplyr::mutate(bt,
        models =
          purrr::map(
            splits,
            ~ model_results(rsample::analysis(.x), fn_call, filter)
          )
      )
    if (type == "student-t") {
      res <- int_t(bt, models, alpha = alpha)
    } else {
      res <- int_pctl(bt, models, alpha = alpha)
    }

    if (keep_reps) {
      bt <- bt[bt$id != "Apparent", ]
      reps <- purrr::map_dfr(bt$models, I)
      reps <- dplyr::group_nest(reps, term, .key = ".replicates")
      res <- dplyr::full_join(res, reps, by = "term")
    }

    res
  }

# TODO add handler for survival models to catch warnings? That seems to be the
# only way to know about convergence.
model_results <- function(data, cl, flt) {
  mod <- broom::tidy(rlang::eval_tidy(cl, data))
  mod <- mod[, c("term", "estimate", "std.error")]
  if (is.language(flt)) {
    mod <- dplyr::filter(mod, !!flt)
  }
  mod
}

test_that("regression intervals", {
  skip_if_not_installed("broom")
  skip_on_cran()

  expect_no_error({
    set.seed(1)
    int_1 <- reg_intervals(mpg ~ disp + wt, data = mtcars)
  })

  expect_equal(
    names(int_1),
    c("term", ".lower", ".estimate", ".upper", ".alpha", ".method")
  )

  expect_snapshot({
    skip_if(new_rng_snapshots)
    set.seed(123)
    int_2 <- reg_intervals(
      mpg ~ disp + wt,
      data = mtcars,
      filter = term == "wt",
      model_fn = "glm",
      keep_reps = TRUE
    )
    int_2
  })

  expect_equal(
    names(int_2),
    c(
      "term",
      ".lower",
      ".estimate",
      ".upper",
      ".alpha",
      ".method",
      ".replicates"
    )
  )
  expect_true(nrow(int_2) == 1)
  expect_true(all(int_2$term == "wt"))

  expect_snapshot(error = TRUE, {
    reg_intervals(mpg ~ disp + wt, data = mtcars, model_fn = "potato")
  })
  expect_snapshot(error = TRUE, {
    reg_intervals(mpg ~ disp + wt, data = mtcars, type = "random")
  })
  expect_snapshot(error = TRUE, {
    reg_intervals(mpg ~ disp + wt, data = mtcars, alpha = "a")
  })
})

#' Make groupings for grouped rsplits
#'
#' This function powers grouped resampling by splitting the data based upon
#' a grouping variable and returning the assessment set indices for each
#' split.
#'
#' @inheritParams vfold_cv
#' @param group A variable in `data` (single character or name) used for
#'  grouping observations with the same value to either the analysis or
#'  assessment set within a fold.
#' @param balance If `v` is less than the number of unique groups, how should
#'  groups be combined into folds? Should be one of
#'  `"groups"`, `"observations"`, `"prop"`.
#' @param ... Arguments passed to balance functions.
#'
#' @details
#' Not all `balance` options are accepted -- or make sense -- for all resampling
#'  functions. For instance, `balance = "prop"` assigns groups to folds at
#'  random, meaning that any given observation is not guaranteed to be in one
#'  (and only one) assessment set. That means `balance = "prop"` can't
#'  be used with [group_vfold_cv()], and so isn't an option available for that
#'  function.
#'
#' Similarly, [group_mc_cv()] and its derivatives don't assign data to one (and
#'  only one) assessment set, but rather allow each observation to be in an
#'  assessment set zero-or-more times. As a result, those functions don't have
#'  a `balance` argument, and under the hood always specify `balance = "prop"`
#'  when they call `make_groups()`.
#'
#' @keywords internal
make_groups <- function(
  data,
  group,
  v,
  balance = c("groups", "observations", "prop"),
  strata = NULL,
  ...
) {
  rlang::check_dots_used(call = rlang::caller_env())
  balance <- rlang::arg_match(balance, error_call = rlang::caller_env())

  data_ind <- tibble(
    ..index = seq_len(nrow(data)),
    ..group = group
  )
  data_ind$..group <- as.character(data_ind$..group)

  res <- switch(
    balance,
    "groups" = balance_groups(
      data_ind = data_ind,
      v = v,
      strata = strata,
      ...
    ),
    "observations" = balance_observations(
      data_ind = data_ind,
      v = v,
      strata = strata,
      ...
    ),
    "prop" = balance_prop(
      data_ind = data_ind,
      v = v,
      strata = strata,
      ...
    )
  )

  data_ind <- res$data_ind
  keys <- res$keys

  data_ind$..group <- as.character(data_ind$..group)
  keys$..group <- as.character(keys$..group)

  data_ind <- data_ind |>
    full_join(keys, by = "..group") |>
    arrange(..index)
  split_unnamed(data_ind$..index, data_ind$..folds)
}

balance_groups <- function(data_ind, v, strata = NULL, ...) {
  if (is.null(strata)) {
    balance_groups_normal(data_ind, v, ...)
  } else {
    balance_groups_strata(data_ind, v, strata, ...)
  }
}

balance_groups_normal <- function(data_ind, v, ...) {
  rlang::check_dots_empty()
  unique_groups <- unique(data_ind$..group)
  keys <- data.frame(
    ..group = unique_groups,
    ..folds = sample(
      rep(seq_len(v), length.out = length(unique_groups))
    )
  )
  list(
    data_ind = data_ind,
    keys = keys
  )
}

balance_groups_strata <- function(data_ind, v, strata, ...) {
  rlang::check_dots_empty()

  data_ind$..strata <- strata
  # Create a table that's all the unique group x strata combinations:
  keys <- vctrs::vec_unique(data_ind[c("..group", "..strata")])
  # Create as many fold IDs as there are group x strata,
  # in repeating order (1, 2, ..., n, 1, 2, ..., n)
  folds <- rep(1:v, length.out = nrow(keys))

  # Split the folds based on how many groups are within each strata
  # So if the first strata in sort is 3, and v is 2, that strata gets a
  # c(1, 2, 1) for fold IDs
  #
  # This means that, if nrow(keys) %% v == 0, each fold should have
  # the same number of groups from each strata
  #
  # We randomize "keys" here so that the function is stochastic even for
  # strata with only one group:
  unique_strata <- unique(keys$..strata)
  keys_order <- sample.int(length(unique_strata))

  # Re-order the keys data.frame based on the reshuffled strata variable:
  keys <- keys[
    order(match(keys$..strata, unique_strata[keys_order])),
  ]

  # And split both folds and keys with the reordered strata vector:
  folds <- split_unnamed(folds, keys$..strata)
  keys <- split_unnamed(keys, keys$..strata)

  # Randomly assign fold IDs to each group within each strata
  keys <- purrr::map2(
    keys,
    folds,
    function(x, y) {
      x$..folds <- sample(y)
      x
    }
  )

  keys <- dplyr::bind_rows(keys)
  keys <- keys[c("..group", "..folds")]
  list(
    data_ind = data_ind,
    keys = keys
  )
}

balance_observations <- function(data_ind, v, strata = NULL, ...) {
  rlang::check_dots_empty()
  n_obs <- nrow(data_ind)
  target_per_fold <- 1 / v

  # This is the core difference between stratification and not:
  #
  # Without stratification, data_ind is broken into v groups,
  # which are roughly balanced based on the number of observations
  #
  # With strata, data_ind is split up by strata, and then each _split_
  # is broken into v groups (which are then combined with the other strata);
  # the balancing for each fold is done separately inside each strata "split"
  data_splits <- if (is.null(strata)) {
    list(data_ind)
  } else {
    split_unnamed(data_ind, strata)
  }

  freq_table <- purrr::map(
    data_splits,
    balance_observations_helper,
    v = v,
    target_per_fold = target_per_fold
  ) |>
    list_rbind()

  collapse_groups(freq_table, data_ind, v)
}

balance_observations_helper <- function(data_split, v, target_per_fold) {
  n_obs <- nrow(data_split)
  # Create a frequency table counting how many of each group are in the data:
  freq_table <- vec_count(data_split$..group, sort = "location")
  # Randomly shuffle that table, then assign the first few rows to folds
  # (to ensure that each fold gets at least one group assigned):
  freq_table <- freq_table[sample.int(nrow(freq_table)), ]
  freq_table$assignment <- NA
  # Assign the first `v` rows to folds, so that each fold has _some_ data:
  freq_table$assignment[seq_len(v)] <- seq_len(v)

  # Each run of this loop assigns one "NA" assignment to a fold,
  # so we won't get caught in an endless loop here
  while (any(is.na(freq_table$assignment))) {
    # Get the index of the next row to be assigned, and its count:
    next_row <- which(is.na(freq_table$assignment))[[1]]
    next_size <- freq_table[next_row, ]$count

    # Calculate which fold to assign this new row into:
    group_breakdown <- freq_table |>
      # The only NA column in freq_table should be assignment
      # So this should only drop un-assigned groups:
      stats::na.omit() |>
      # Group by fold assignments and count data in each fold:
      dplyr::group_by(.data$assignment) |>
      dplyr::summarise(count = sum(.data$count), .groups = "drop") |>
      # Calculate...:
      dplyr::mutate(
        # The proportion of data in each fold so far,
        prop = .data$count / n_obs,
        # The amount off from the target proportion so far,
        pre_error = abs(.data$prop - target_per_fold),
        # The amount off from the target proportion if we add this new group,
        if_added_count = .data$count + next_size,
        if_added_prop = .data$if_added_count / n_obs,
        post_error = abs(.data$if_added_prop - target_per_fold),
        # And how much better or worse adding this new group would make things
        improvement = .data$post_error - .data$pre_error
      )

    # Assign the group in question to the best fold and move on to the next one:
    most_improved <- which.min(group_breakdown$improvement)
    freq_table[next_row, ]$assignment <-
      group_breakdown[most_improved, ]$assignment
  }
  freq_table
}

balance_prop <- function(
  prop,
  data_ind,
  v,
  replace = FALSE,
  strata = NULL,
  ...
) {
  rlang::check_dots_empty()

  # This is the core difference between stratification and not:
  #
  # Without stratification, `prop`% of `data_ind` is sampled `v` times;
  # the resampling is done with the entire set of groups
  #
  # With strata, data_ind is split up by strata, and then each _split_
  # has `prop`% of `data_ind` is sampled `v` times;
  # the resampling for each iteration is done inside each strata "split"
  data_splits <- if (is.null(strata)) {
    list(data_ind)
  } else {
    split_unnamed(data_ind, strata)
  }

  freq_table <- purrr::map(
    data_splits,
    balance_prop_helper,
    prop = prop,
    v = v,
    replace = replace
  ) |>
    list_rbind()

  collapse_groups(freq_table, data_ind, v)
}

balance_prop_helper <- function(prop, data_ind, v, replace) {
  freq_table <- vec_count(data_ind$..group, sort = "location")

  # Calculate how many groups to sample each iteration
  # If sampling with replacement,
  # set `n` to the number of resamples we'd need
  # if we somehow got the smallest group every time.
  # If sampling without replacement, just reshuffle all the groups.
  n <- nrow(freq_table)
  if (replace) {
    n <- n * prop * sum(freq_table$count) / min(freq_table$count)
  }
  n <- ceiling(n)

  purrr::map(
    seq_len(v),
    function(x) {
      row_idx <- sample.int(nrow(freq_table), n, replace = replace)
      work_table <- freq_table[row_idx, ]
      cumulative_proportion <- cumsum(work_table$count) / sum(freq_table$count)
      crosses_target <- which(cumulative_proportion > prop)[[1]]
      is_closest <- cumulative_proportion[c(crosses_target, crosses_target - 1)]
      is_closest <- which.min(abs(is_closest - prop)) - 1
      crosses_target <- crosses_target - is_closest
      out <- work_table[seq_len(crosses_target), ]
      out$assignment <- x
      out
    }
  ) |>
    list_rbind()
}

collapse_groups <- function(freq_table, data_ind, v) {
  data_ind <- dplyr::left_join(
    data_ind,
    freq_table,
    by = c("..group" = "key"),
    multiple = "all",
    relationship = "many-to-many"
  )
  data_ind$..group <- data_ind$assignment
  data_ind <- data_ind[c("..index", "..group")]

  # If a group was never assigned a fold, then its `..group` is NA
  #
  # If we leave that alone, it winds up messing up our fold assignments,
  # because it will be assigned some value in `seq_len(v)`
  #
  # So instead, we drop those groups here:
  data_ind <- stats::na.omit(data_ind)

  unique_groups <- unique(data_ind$..group)

  keys <- data.frame(
    ..group = unique_groups,
    ..folds = sample(rep(seq_len(v), length.out = length(unique_groups)))
  )

  list(
    data_ind = data_ind,
    keys = keys
  )
}

validate_group <- function(group, data, call = rlang::caller_env()) {
  if (!missing(group)) {
    group <- tidyselect::vars_select(names(data), !!enquo(group))
    if (length(group) == 0) {
      group <- NULL
    }
  }

  check_string(group, call = call)

  group
}

#' Nested or Double Resampling
#'
#' `nested_cv` can be used to take the results of one resampling procedure
#'   and conduct further resamples within each split. Any type of resampling
#'   used in `rsample` can be used.
#'
#' @details
#' It is a bad idea to use bootstrapping as the outer resampling procedure (see
#'   the example below)
#' @param data  A data frame.
#' @param outside The initial resampling specification. This can be an already
#'   created object or an expression of a new object (see the examples below).
#'   If the latter is used, the `data` argument does not need to be
#'   specified and, if it is given, will be ignored.
#' @param inside An expression for the type of resampling to be conducted
#'   within the initial procedure.
#' @return  An tibble with `nested_cv` class and any other classes that
#'   outer resampling process normally contains. The results include a
#'  column for the outer data split objects, one or more `id` columns,
#'  and a column of nested tibbles called `inner_resamples` with the
#'  additional resamples.
#' @examples
#' ## Using expressions for the resampling procedures:
#' nested_cv(mtcars, outside = vfold_cv(v = 3), inside = bootstraps(times = 5))
#'
#' ## Using an existing object:
#' folds <- vfold_cv(mtcars)
#' nested_cv(mtcars, folds, inside = bootstraps(times = 5))
#'
#' ## The dangers of outer bootstraps:
#' set.seed(2222)
#' bad_idea <- nested_cv(mtcars,
#'                       outside = bootstraps(times = 5),
#'                       inside = vfold_cv(v = 3))
#'
#' first_outer_split <- bad_idea$splits[[1]]
#' outer_analysis <- as.data.frame(first_outer_split)
#' sum(grepl("Volvo 142E", rownames(outer_analysis)))
#'
#' ## For the 3-fold CV used inside of each bootstrap, how are the replicated
#' ## `Volvo 142E` data partitioned?
#' first_inner_split <- bad_idea$inner_resamples[[1]]$splits[[1]]
#' inner_analysis <- as.data.frame(first_inner_split)
#' inner_assess   <- as.data.frame(first_inner_split, data = "assessment")
#'
#' sum(grepl("Volvo 142E", rownames(inner_analysis)))
#' sum(grepl("Volvo 142E", rownames(inner_assess)))
#' @export
nested_cv <- function(data, outside, inside)  {
  nest_args <- formalArgs(nested_cv)
  cl <- match.call()

  boot_msg <-
    paste0(
      "Using bootstrapping as the outer resample is dangerous ",
      "since the inner resample might have the same data ",
      "point in both the analysis and assessment set."
    )

  outer_cl <- cl[["outside"]]
  if (is_call(outer_cl)) {
    if (grepl("^bootstraps", deparse(outer_cl)))
      warning(boot_msg, call. = FALSE)
    outer_cl$data <- quote(data)
    outside <- eval(outer_cl)
  } else {
    if (inherits(outside, "bootstraps"))
      warning(boot_msg, call. = FALSE)
  }

  inner_cl <- cl[["inside"]]
  if (!is_call(inner_cl))
    stop(
      "`inside` should be a expression such as `vfold()` or ",
      "bootstraps(times = 10)` instead of a existing object.",
      call. = FALSE
    )
  inside <- map(outside$splits, inside_resample, cl = inner_cl)

  out <- dplyr::mutate(outside, inner_resamples = inside)

  out <- add_class(out, cls = "nested_cv", at_end = FALSE)

  attr(out, "outside") <- cl$outside
  attr(out, "inside") <- cl$inside

  out
}

inside_resample <- function(src, cl) {
  cl$data <- quote(as.data.frame(src))
  eval(cl)
}

#' @export
print.nested_cv <- function(x, ...) {
  char_x <- paste("#", pretty(x))
  cat(char_x, sep = "\n")
  class(x) <- class(tibble())
  print(x, ...)
}

#' Extract Predictor Names from Formula or Terms
#'
#' `all.vars` returns all variables used in a formula. This
#'  function only returns the variables explicitly used on the
#'  right-hand side (i.e., it will not resolve dots unless the
#'  object is terms with a data set specified).
#' @param object A model formula or [stats::terms()]
#'  object.
#' @param ... Arguments to pass to [all.vars()]
#' @return A character vector of names
#' @export
#' @examples
#' form_pred(y ~ x + z)
#' form_pred(terms(y ~ x + z))
#'
#' form_pred(y ~ x + log(z))
#' form_pred(log(y) ~ x + z)
#'
#' form_pred(y1 + y2 ~ x + z)
#' form_pred(log(y1) + y2 ~ x + z)
#'
#' # will fail:
#' # form_pred(y ~ .)
#'
#' form_pred(terms(mpg ~ (.)^2, data = mtcars))
#' form_pred(terms( ~ (.)^2, data = mtcars))
#' @importFrom stats terms

form_pred <- function(object, ...) {
  if(inherits(object, "formula")) {
    object <- terms(object)
  }
  y_index <- attr(object, "response")

  ## If there is something on the lhs of the formula,
  ## remove it and get vars
  if(y_index != 0) {
    object[[2]] <- NULL
    object <- terms(object)
  }
  all.vars(object, ...)
}

test_that("no dots", {
  expect_equal(form_pred(y ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(y ~ x + z)), c("x", "z"))

  expect_equal(form_pred(y ~ x + log(z)), c("x", "z"))
  expect_equal(form_pred(terms(y ~ x + log(z))), c("x", "z"))

  expect_equal(form_pred(log(y) ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(log(y) ~ x + z)), c("x", "z"))

  expect_equal(form_pred(y1 + y2 ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(y1 + y2 ~ x + z)), c("x", "z"))

  expect_equal(form_pred(log(y1) + y2 ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(log(y1) + y2 ~ x + z)), c("x", "z"))

  expect_equal(form_pred(~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(~ x + z)), c("x", "z"))

  expect_equal(form_pred(~x), "x")
  expect_equal(form_pred(terms(~x)), "x")

  expect_equal(form_pred(y ~ x), "x")
  expect_equal(form_pred(terms(y ~ x)), "x")
})

test_that("dots", {
  expect_error(form_pred(y ~ .))
  expect_error(form_pred(terms(y ~ .)))

  expect_error(form_pred(y ~ (.)^2))
  expect_error(form_pred(terms(y ~ (.)^2)))

  expect_equal(
    form_pred(terms(mpg ~ (.)^2, data = mtcars)),
    names(mtcars)[2:11]
  )
  expect_equal(
    form_pred(terms(~ (.)^2, data = mtcars)),
    names(mtcars)
  )
})

context("Predictor extraction")

library(testthat)

test_that('no dots', {
  expect_equal(form_pred(y ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(y ~ x + z)), c("x", "z"))

  expect_equal(form_pred(y ~ x + log(z)), c("x", "z"))
  expect_equal(form_pred(terms(y ~ x + log(z))), c("x", "z"))

  expect_equal(form_pred(log(y) ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(log(y) ~ x + z)), c("x", "z"))

  expect_equal(form_pred(y1 + y2 ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(y1 + y2 ~ x + z)), c("x", "z"))

  expect_equal(form_pred(log(y1) + y2 ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(log(y1) + y2 ~ x + z)), c("x", "z"))

  expect_equal(form_pred(~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(~ x + z)), c("x", "z"))

  expect_equal(form_pred(~ x), "x")
  expect_equal(form_pred(terms(~ x)), "x")

  expect_equal(form_pred(y ~ x), "x")
  expect_equal(form_pred(terms(y ~ x)), "x")

})

test_that('dots', {
  expect_error(form_pred(y ~ .))
  expect_error(form_pred(terms(y ~ .)))

  expect_error(form_pred(y ~ (.)^2))
  expect_error(form_pred(terms(y ~ (.)^2)))

  expect_equal(form_pred(terms(mpg ~ (.)^2, data = mtcars)),
               names(mtcars)[2:11])
  expect_equal(form_pred(terms(~ (.)^2, data = mtcars)),
               names(mtcars))
})

test_that("no dots", {
  expect_equal(form_pred(y ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(y ~ x + z)), c("x", "z"))

  expect_equal(form_pred(y ~ x + log(z)), c("x", "z"))
  expect_equal(form_pred(terms(y ~ x + log(z))), c("x", "z"))

  expect_equal(form_pred(log(y) ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(log(y) ~ x + z)), c("x", "z"))

  expect_equal(form_pred(y1 + y2 ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(y1 + y2 ~ x + z)), c("x", "z"))

  expect_equal(form_pred(log(y1) + y2 ~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(log(y1) + y2 ~ x + z)), c("x", "z"))

  expect_equal(form_pred(~ x + z), c("x", "z"))
  expect_equal(form_pred(terms(~ x + z)), c("x", "z"))

  expect_equal(form_pred(~x), "x")
  expect_equal(form_pred(terms(~x)), "x")

  expect_equal(form_pred(y ~ x), "x")
  expect_equal(form_pred(terms(y ~ x)), "x")
})

test_that("dots", {
  expect_snapshot(error = TRUE, {
    form_pred(y ~ .)
  })
  expect_snapshot(error = TRUE, {
    form_pred(terms(y ~ .))
  })

  expect_snapshot(error = TRUE, {
    form_pred(y ~ (.)^2)
  })
  expect_snapshot(error = TRUE, {
    form_pred(terms(y ~ (.)^2))
  })

  expect_equal(
    form_pred(terms(mpg ~ (.)^2, data = mtcars)),
    names(mtcars)[2:11]
  )
  expect_equal(
    form_pred(terms(~ (.)^2, data = mtcars)),
    names(mtcars)
  )
})

#' Convert Resampling Objects to Other Formats
#'
#' These functions can convert resampling objects between
#'  \pkg{rsample} and \pkg{caret}.
#'
#' @param object An `rset` object. Currently,
#'  `nested_cv` is not supported.
#' @return `rsample2caret` returns a list that mimics the
#'  `index` and `indexOut` elements of a
#'  `trainControl` object. `caret2rsample` returns an
#'  `rset` object of the appropriate class.
#' @export
rsample2caret <- function(object, data = c("analysis", "assessment")) {
  if (!inherits(object, "rset")) {
    rlang::abort("`object` must be an `rset`")
  }
  data <- rlang::arg_match(data)
  in_ind <- purrr::map(object$splits, as.integer, data = "analysis")
  names(in_ind) <- labels(object)
  out_ind <- purrr::map(object$splits, as.integer, data = "assessment")
  names(out_ind) <- names(in_ind)
  list(index = in_ind, indexOut = out_ind)
}

#' @rdname rsample2caret
#' @param ctrl An object produced by `trainControl` that has
#'  had the `index` and `indexOut` elements populated by
#'  integers. One method of getting this is to extract the
#'  `control` objects from an object produced by `train`.
#' @param data The data that was originally used to produce the
#'  `ctrl` object.
#' @export
caret2rsample <- function(ctrl, data = NULL) {
  if (is.null(data)) {
    rlang::abort("Must supply original data")
  }
  if (!any(names(ctrl) == "index")) {
    rlang::abort("`ctrl` should have an element `index`")
  }
  if (!any(names(ctrl) == "indexOut")) {
    rlang::abort("`ctrl` should have an element `indexOut`")
  }
  if (is.null(ctrl$index)) {
    rlang::abort("`ctrl$index` should be populated with integers")
  }
  if (is.null(ctrl$indexOut)) {
    rlang::abort("`ctrl$indexOut` should be populated with integers")
  }

  indices <- purrr::map2(ctrl$index, ctrl$indexOut, extract_int)
  id_data <- names(indices)
  indices <- unname(indices)
  indices <- purrr::map(indices, add_data, y = data)
  indices <-
    map(indices, add_rsplit_class, cl = map_rsplit_method(ctrl$method))
  indices <- tibble::tibble(splits = indices)
  if (ctrl$method %in% c("repeatedcv", "adaptive_cv")) {
    id_data <- strsplit(id_data, split = ".", fixed = TRUE)
    id_data <- tibble::tibble(
      id = vapply(id_data, function(x) {
        x[2]
      }, character(1)),
      id2 = vapply(id_data, function(x) {
        x[1]
      }, character(1))
    )
  } else {
    id_data <- tibble::tibble(id = id_data)
  }

  new_rset(
    splits = indices$splits,
    ids = id_data[, grepl("^id", names(id_data))],
    attrib = map_attr(ctrl),
    subclass = c(map_rset_method(ctrl$method), "rset")
  )
}

extract_int <- function(x, y) {
  list(in_id = x, out_id = y)
}

add_data <- function(x, y) {
  c(list(data = y), x)
}

add_rsplit_class <- function(x, cl) {
  class(x) <- c("rsplit", cl)
  x
}

add_rset_class <- function(x, cl) {
  class(x) <- c(cl, "rset", "tbl_df", "tbl", "data.frame")
  x
}

map_rsplit_method <- function(method) {
  out <- switch(
    method,
    cv = ,
    repeatedcv = ,
    adaptive_cv = "vfold_split",
    boot = ,
    boot_all = ,
    boot632 = ,
    optimism_boot = ,
    adaptive_boot = "boot_split",
    LOOCV = "loo_split",
    LGOCV = ,
    adaptive_LGOCV = "mc_split",
    timeSlice = "rof_split",
    "error"
  )
  if (out == "error") {
    rlang::abort(
      "Resampling method `",
      method,
      "` cannot be converted into an `rsplit` object"
    )
  }
  out
}

map_rset_method <- function(method) {
  out <- switch(
    method,
    cv = ,
    repeatedcv = ,
    adaptive_cv = "vfold_cv",
    boot = ,
    boot_all = ,
    boot632 = ,
    optimism_boot = ,
    adaptive_boot = "bootstraps",
    LOOCV = "loo_cv",
    LGOCV = ,
    adaptive_LGOCV = "mc_cv",
    timeSlice = "rolling_origin",
    "error"
  )
  if (out == "error") {
    rlang::abort("Resampling method `",
                 method,
                 "` cannot be converted into an `rset` object"
    )
  }
  out
}


map_attr <- function(object) {
  if (grepl("cv$", object$method)) {
    out <- list(
      v = object$number,
      repeats = ifelse(!is.na(object$repeats),
                       object$repeats, 1
      ),
      strata = TRUE
    )
  } else if (grepl("boot", object$method)) {
    out <- list(
      times = object$number,
      apparent = FALSE,
      strata = FALSE
    )
  } else if (grepl("LGOCV$", object$method)) {
    out <- list(
      times = object$number,
      prop = object$p,
      strata = FALSE
    )
  } else if (object$method == "LOOCV") {
    out <- list()
  } else if (object$method == "timeSlice") {
    out <- list(
      initial = object$initialWindow,
      assess = object$horizon,
      cumulative = !object$fixedWindow,
      skip = object$skip
    )
  } else {
    rlang::abort("Method", object$method, "cannot be converted")
  }
  out
}

context("Conversions for caret")

library(testthat)
library(rsample)

###################################################################
## Test cases for caret -> rsample that mimic `trainControl`

dat <- data.frame(y = 1:15, x = 15:1)

lgo1 <-
  structure(
    list(
      method = "LGOCV",
      index = structure(
        list(
          Resample1 = c(1L, 4L, 5L, 6L, 7L, 9L, 10L, 14L),
          Resample2 = c(2L, 4L, 5L, 6L, 9L, 10L, 14L, 15L),
          Resample3 = c(1L, 2L, 3L, 5L, 6L, 7L, 8L, 9L)
        ),
        .Names = c("Resample1", "Resample2", "Resample3")
      ),
      indexOut = structure(
        list(
          Resample1 = c(2L, 3L, 8L, 11L, 12L, 13L, 15L),
          Resample2 = c(1L, 3L, 7L, 8L, 11L, 12L, 13L),
          Resample3 = c(4L, 10L, 11L, 12L, 13L, 14L, 15L)
        ),
        .Names = c("Resample1", "Resample2", "Resample3")
      ),
      number = 3,
      p = 0.5
    ),
    .Names = c("method", "index", "indexOut", "number", "p")
  )

cv_1 <- structure(
  list(
    method = "cv",
    index = structure(
      list(
        Fold1 = c(2L, 3L, 4L, 5L, 6L, 7L, 8L, 10L, 11L, 13L),
        Fold2 = c(1L, 4L, 6L,
                  8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
        Fold3 = c(1L, 2L, 3L, 5L, 7L, 9L, 12L, 14L, 15L)
      ),
      .Names = c("Fold1", "Fold2", "Fold3")
    ),
    indexOut = structure(
      list(
        Resample1 = c(1L, 9L, 12L, 14L, 15L),
        Resample2 = c(2L, 3L, 5L, 7L),
        Resample3 = c(4L, 6L, 8L, 10L, 11L, 13L)
      ),
      .Names = c("Resample1", "Resample2", "Resample3")
    ),
    number = 3,
    repeats = NA
  ),
  .Names = c("method", "index", "indexOut", "number", "repeats")
)

cv_2 <-
  structure(
  list(
    method = "repeatedcv",
    index = structure(
      list(
        Fold1.Rep1 = c(1L, 3L, 4L, 6L, 9L, 10L, 12L, 13L, 14L, 15L),
        Fold2.Rep1 = c(2L, 5L, 7L, 8L, 10L, 11L, 13L, 14L, 15L),
        Fold3.Rep1 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 11L, 12L),
        Fold1.Rep2 = c(1L, 2L, 3L, 5L, 6L, 7L, 10L, 11L, 12L, 14L),
        Fold2.Rep2 = c(2L, 4L, 6L, 8L, 9L, 11L, 13L, 14L, 15L),
        Fold3.Rep2 = c(1L, 3L, 4L, 5L, 7L, 8L, 9L, 10L, 12L, 13L, 15L)
      ),
      .Names = c(
        "Fold1.Rep1", "Fold2.Rep1", "Fold3.Rep1",
        "Fold1.Rep2", "Fold2.Rep2", "Fold3.Rep2"
      )
    ),
    indexOut = structure(
      list(
        Resample1 = c(2L, 5L, 7L, 8L, 11L),
        Resample2 = c(1L, 3L, 4L, 6L, 9L, 12L),
        Resample3 = c(10L, 13L, 14L, 15L),
        Resample4 = c(4L, 8L, 9L, 13L, 15L),
        Resample5 = c(1L, 3L, 5L, 7L, 10L, 12L),
        Resample6 = c(2L, 6L, 11L, 14L)
      ),
      .Names = c(
        "Resample1", "Resample2", "Resample3",
        "Resample4", "Resample5", "Resample6"
      )
    ),
    number = 3,
    repeats = 2
  ),
  .Names = c("method", "index", "indexOut", "number", "repeats")
)

cv_3 <- cv_2
cv_3$method <- "adaptive_cv"

bt_1 <-
  structure(
    list(
      method = "boot",
      index = structure(
        list(
          Resample1 = c(1L, 1L, 4L, 4L, 5L, 7L, 8L, 10L, 11L, 11L, 12L, 13L, 15L, 15L, 15L),
          Resample2 = c(1L, 2L, 3L, 5L, 5L, 5L, 6L, 7L, 8L, 9L, 9L, 9L, 10L, 10L, 12L)
        ),
        .Names = c("Resample1", "Resample2")
      ),
      indexOut = structure(
        list(
          Resample1 = c(2L, 3L, 6L, 9L, 14L),
          Resample2 = c(4L, 11L, 13L, 14L, 15L)
        ),
        .Names = c("Resample1", "Resample2")
      ),
      number = 2
    ),
    .Names = c("method",
               "index", "indexOut", "number")
  )
bt_2 <- bt_1
bt_2$method <- "boot632"

bt_3 <- bt_1
bt_3$method <- "optimism_boot"

bt_4 <- bt_1
bt_4$method <- "boot_all"

bt_5 <- bt_1
bt_5$method <- "adaptive_boot"

loo_1 <-
  structure(
    list(
      method = "LOOCV",
      index = structure(
        list(
          Fold01 = 2:15,
          Fold02 = c(1L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold03 = c(1L, 2L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold04 = c(1L, 2L, 3L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold05 = c(1L, 2L, 3L, 4L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold06 = c(1L, 2L, 3L, 4L, 5L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold07 = c(1L, 2L, 3L, 4L, 5L, 6L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold08 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold09 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold10 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 11L, 12L, 13L, 14L, 15L),
          Fold11 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 12L, 13L, 14L, 15L),
          Fold12 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 13L, 14L, 15L),
          Fold13 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 14L, 15L),
          Fold14 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 15L),
          Fold15 = 1:14
        ),
        .Names = c(
          "Fold01", "Fold02", "Fold03",  "Fold04", "Fold05", "Fold06",
          "Fold07", "Fold08", "Fold09", "Fold10", "Fold11", "Fold12",
          "Fold13", "Fold14", "Fold15"
        )
      ),
      indexOut = structure(
        list(
          Resample01 = 1L, Resample02 = 2L,
          Resample03 = 3L, Resample04 = 4L,
          Resample05 = 5L, Resample06 = 6L,
          Resample07 = 7L, Resample08 = 8L,
          Resample09 = 9L, Resample10 = 10L,
          Resample11 = 11L, Resample12 = 12L,
          Resample13 = 13L, Resample14 = 14L,
          Resample15 = 15L
        ),
        .Names = c(
          "Resample01", "Resample02", "Resample03", "Resample04",
          "Resample05", "Resample06", "Resample07", "Resample08",
          "Resample09", "Resample10", "Resample11", "Resample12",
          "Resample13", "Resample14", "Resample15"
        )
      )
    ),
    .Names = c("method", "index", "indexOut")
  )

rof_1 <-
  structure(
    list(
      method = "timeSlice",
      index = structure(
        list(
          Training04 = 1:4, Training05 = 2:5, Training06 = 3:6,
          Training07 = 4:7, Training08 = 5:8, Training09 = 6:9,
          Training10 = 7:10
        ),
        .Names = c(
          "Training04", "Training05", "Training06",  "Training07",
          "Training08", "Training09", "Training10"
        )
      ),
      indexOut = structure(
        list(
          Testing04 = 5:9, Testing05 = 6:10, Testing06 = 7:11,
          Testing07 = 8:12, Testing08 = 9:13, Testing09 = 10:14,
          Testing10 = 11:15
        ),
        .Names = c(
          "Testing04", "Testing05", "Testing06", "Testing07",
          "Testing08", "Testing09", "Testing10"
        )
      ),
      initialWindow = 4, horizon = 5, fixedWindow = TRUE, skip = 0
    ),
    .Names = c(
      "method", "index", "indexOut", "initialWindow",
      "horizon", "fixedWindow", "skip"
    )
  )

###################################################################
##

check_indices <- function(newer, orig) {
  for (i in seq_along(newer$splits)) {
    expect_equal(as.integer(newer$splits[[i]]),
                 orig$index[[i]])
    expect_equal(as.integer(newer$splits[[i]], "assessment"),
                 orig$indexOut[[i]])
  }
  invisible(NULL)
}

###################################################################
## Tests

test_that('basic v-fold', {
  vfold_obj_1 <- caret2rsample(cv_1, data = dat)
  check_indices(vfold_obj_1, cv_1)
  for (i in seq_along(vfold_obj_1$splits))
    expect_equal(vfold_obj_1$id[[i]], names(cv_1$index)[i])
})

test_that('repeated v-fold', {
  vfold_obj_2 <- caret2rsample(cv_2, data = dat)
  check_indices(vfold_obj_2, cv_2)
  for (i in seq_along(vfold_obj_2$splits))
    expect_equal(paste(vfold_obj_2$id2[[i]], vfold_obj_2$id[[i]],
                       sep = "."),
                 names(cv_2$index)[i])

})

test_that('basic boot', {
  bt_obj_1 <- caret2rsample(bt_1, data = dat)
  check_indices(bt_obj_1, bt_1)
  for (i in seq_along(bt_obj_1$splits))
    expect_equal(bt_obj_1$id[[i]], names(bt_1$index)[i])
})

test_that('boot 632', {
  bt_obj_2 <- caret2rsample(bt_2, data = dat)
  check_indices(bt_obj_2, bt_2)
  for (i in seq_along(bt_obj_2$splits))
    expect_equal(bt_obj_2$id[[i]], names(bt_2$index)[i])
})

test_that('boot optim', {
  bt_obj_3 <- caret2rsample(bt_3, data = dat)
  check_indices(bt_obj_3, bt_3)
  for (i in seq_along(bt_obj_3$splits))
    expect_equal(bt_obj_3$id[[i]], names(bt_3$index)[i])
})

test_that('boot all', {
  bt_obj_4 <- caret2rsample(bt_4, data = dat)
  check_indices(bt_obj_4, bt_4)
  for (i in seq_along(bt_obj_4$splits))
    expect_equal(bt_obj_4$id[[i]], names(bt_4$index)[i])
})

test_that('adaptive boot', {
  bt_obj_5 <- caret2rsample(bt_5, data = dat)
  check_indices(bt_obj_5, bt_5)
  for (i in seq_along(bt_obj_5$splits))
    expect_equal(bt_obj_5$id[[i]], names(bt_5$index)[i])
})


test_that('loo', {
  loo_obj <- caret2rsample(loo_1, data = dat)
  check_indices(loo_obj, loo_1)
  for (i in seq_along(loo_obj$splits))
    expect_equal(loo_obj$id[[i]], names(loo_1$index)[i])
})

test_that('mcv', {
  mcv_obj <- caret2rsample(lgo1, data = dat)
  check_indices(mcv_obj, lgo1)
  for (i in seq_along(mcv_obj$splits))
    expect_equal(mcv_obj$id[[i]], names(lgo1$index)[i])
})

test_that('rolling origin', {
  rof_obj <- caret2rsample(rof_1, data = dat)
  check_indices(rof_obj, rof_1)
  for (i in seq_along(rof_obj$splits))
    expect_equal(rof_obj$id[[i]], names(rof_1$index)[i])
})

###################################################################
## Test cases for caret -> rsample that mimic `trainControl`

check_indices <- function(newer, orig) {
  for (i in seq_along(newer$splits)) {
    expect_equal(
      as.integer(newer$splits[[i]]),
      orig$index[[i]]
    )
    expect_equal(
      as.integer(newer$splits[[i]], "assessment"),
      orig$indexOut[[i]]
    )
  }
  invisible(NULL)
}

###################################################################

test_that("basic v-fold", {
  cv_1 <- structure(
    list(
      method = "cv",
      index = structure(
        list(
          Fold1 = c(2L, 3L, 4L, 5L, 6L, 7L, 8L, 10L, 11L, 13L),
          Fold2 = c(
            1L,
            4L,
            6L,
            8L,
            9L,
            10L,
            11L,
            12L,
            13L,
            14L,
            15L
          ),
          Fold3 = c(1L, 2L, 3L, 5L, 7L, 9L, 12L, 14L, 15L)
        ),
        .Names = c("Fold1", "Fold2", "Fold3")
      ),
      indexOut = structure(
        list(
          Resample1 = c(1L, 9L, 12L, 14L, 15L),
          Resample2 = c(2L, 3L, 5L, 7L),
          Resample3 = c(4L, 6L, 8L, 10L, 11L, 13L)
        ),
        .Names = c("Resample1", "Resample2", "Resample3")
      ),
      number = 3,
      repeats = NA
    ),
    .Names = c("method", "index", "indexOut", "number", "repeats")
  )
  dat <- data.frame(y = 1:15, x = 15:1)
  vfold_obj_1 <- caret2rsample(cv_1, data = dat)

  check_indices(vfold_obj_1, cv_1)
  for (i in seq_along(vfold_obj_1$splits)) {
    expect_equal(vfold_obj_1$id[[i]], names(cv_1$index)[i])
  }
})

test_that("repeated v-fold", {
  cv_2 <-
    structure(
      list(
        method = "repeatedcv",
        index = structure(
          list(
            Fold1.Rep1 = c(1L, 3L, 4L, 6L, 9L, 10L, 12L, 13L, 14L, 15L),
            Fold2.Rep1 = c(2L, 5L, 7L, 8L, 10L, 11L, 13L, 14L, 15L),
            Fold3.Rep1 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 11L, 12L),
            Fold1.Rep2 = c(1L, 2L, 3L, 5L, 6L, 7L, 10L, 11L, 12L, 14L),
            Fold2.Rep2 = c(2L, 4L, 6L, 8L, 9L, 11L, 13L, 14L, 15L),
            Fold3.Rep2 = c(1L, 3L, 4L, 5L, 7L, 8L, 9L, 10L, 12L, 13L, 15L)
          ),
          .Names = c(
            "Fold1.Rep1",
            "Fold2.Rep1",
            "Fold3.Rep1",
            "Fold1.Rep2",
            "Fold2.Rep2",
            "Fold3.Rep2"
          )
        ),
        indexOut = structure(
          list(
            Resample1 = c(2L, 5L, 7L, 8L, 11L),
            Resample2 = c(1L, 3L, 4L, 6L, 9L, 12L),
            Resample3 = c(10L, 13L, 14L, 15L),
            Resample4 = c(4L, 8L, 9L, 13L, 15L),
            Resample5 = c(1L, 3L, 5L, 7L, 10L, 12L),
            Resample6 = c(2L, 6L, 11L, 14L)
          ),
          .Names = c(
            "Resample1",
            "Resample2",
            "Resample3",
            "Resample4",
            "Resample5",
            "Resample6"
          )
        ),
        number = 3,
        repeats = 2
      ),
      .Names = c("method", "index", "indexOut", "number", "repeats")
    )
  dat <- data.frame(y = 1:15, x = 15:1)
  vfold_obj_2 <- caret2rsample(cv_2, data = dat)

  check_indices(vfold_obj_2, cv_2)
  for (i in seq_along(vfold_obj_2$splits)) {
    expect_equal(
      paste(vfold_obj_2$id2[[i]], vfold_obj_2$id[[i]], sep = "."),
      names(cv_2$index)[i]
    )
  }
})

test_that("basic boot", {
  bt_1 <-
    structure(
      list(
        method = "boot",
        index = structure(
          list(
            Resample1 = c(
              1L,
              1L,
              4L,
              4L,
              5L,
              7L,
              8L,
              10L,
              11L,
              11L,
              12L,
              13L,
              15L,
              15L,
              15L
            ),
            Resample2 = c(
              1L,
              2L,
              3L,
              5L,
              5L,
              5L,
              6L,
              7L,
              8L,
              9L,
              9L,
              9L,
              10L,
              10L,
              12L
            )
          ),
          .Names = c("Resample1", "Resample2")
        ),
        indexOut = structure(
          list(
            Resample1 = c(2L, 3L, 6L, 9L, 14L),
            Resample2 = c(4L, 11L, 13L, 14L, 15L)
          ),
          .Names = c("Resample1", "Resample2")
        ),
        number = 2
      ),
      .Names = c(
        "method",
        "index",
        "indexOut",
        "number"
      )
    )
  dat <- data.frame(y = 1:15, x = 15:1)
  bt_obj_1 <- caret2rsample(bt_1, data = dat)

  check_indices(bt_obj_1, bt_1)
  for (i in seq_along(bt_obj_1$splits)) {
    expect_equal(bt_obj_1$id[[i]], names(bt_1$index)[i])
  }
})

test_that("boot 632", {
  bt_2 <-
    structure(
      list(
        method = "boot",
        index = structure(
          list(
            Resample1 = c(
              1L,
              1L,
              4L,
              4L,
              5L,
              7L,
              8L,
              10L,
              11L,
              11L,
              12L,
              13L,
              15L,
              15L,
              15L
            ),
            Resample2 = c(
              1L,
              2L,
              3L,
              5L,
              5L,
              5L,
              6L,
              7L,
              8L,
              9L,
              9L,
              9L,
              10L,
              10L,
              12L
            )
          ),
          .Names = c("Resample1", "Resample2")
        ),
        indexOut = structure(
          list(
            Resample1 = c(2L, 3L, 6L, 9L, 14L),
            Resample2 = c(4L, 11L, 13L, 14L, 15L)
          ),
          .Names = c("Resample1", "Resample2")
        ),
        number = 2
      ),
      .Names = c(
        "method",
        "index",
        "indexOut",
        "number"
      )
    )
  bt_2$method <- "boot632"
  dat <- data.frame(y = 1:15, x = 15:1)
  bt_obj_2 <- caret2rsample(bt_2, data = dat)

  check_indices(bt_obj_2, bt_2)
  for (i in seq_along(bt_obj_2$splits)) {
    expect_equal(bt_obj_2$id[[i]], names(bt_2$index)[i])
  }
})

test_that("boot optim", {
  bt_3 <-
    structure(
      list(
        method = "boot",
        index = structure(
          list(
            Resample1 = c(
              1L,
              1L,
              4L,
              4L,
              5L,
              7L,
              8L,
              10L,
              11L,
              11L,
              12L,
              13L,
              15L,
              15L,
              15L
            ),
            Resample2 = c(
              1L,
              2L,
              3L,
              5L,
              5L,
              5L,
              6L,
              7L,
              8L,
              9L,
              9L,
              9L,
              10L,
              10L,
              12L
            )
          ),
          .Names = c("Resample1", "Resample2")
        ),
        indexOut = structure(
          list(
            Resample1 = c(2L, 3L, 6L, 9L, 14L),
            Resample2 = c(4L, 11L, 13L, 14L, 15L)
          ),
          .Names = c("Resample1", "Resample2")
        ),
        number = 2
      ),
      .Names = c(
        "method",
        "index",
        "indexOut",
        "number"
      )
    )
  bt_3$method <- "optimism_boot"
  dat <- data.frame(y = 1:15, x = 15:1)
  bt_obj_3 <- caret2rsample(bt_3, data = dat)

  check_indices(bt_obj_3, bt_3)
  for (i in seq_along(bt_obj_3$splits)) {
    expect_equal(bt_obj_3$id[[i]], names(bt_3$index)[i])
  }
})

test_that("boot all", {
  bt_4 <-
    structure(
      list(
        method = "boot",
        index = structure(
          list(
            Resample1 = c(
              1L,
              1L,
              4L,
              4L,
              5L,
              7L,
              8L,
              10L,
              11L,
              11L,
              12L,
              13L,
              15L,
              15L,
              15L
            ),
            Resample2 = c(
              1L,
              2L,
              3L,
              5L,
              5L,
              5L,
              6L,
              7L,
              8L,
              9L,
              9L,
              9L,
              10L,
              10L,
              12L
            )
          ),
          .Names = c("Resample1", "Resample2")
        ),
        indexOut = structure(
          list(
            Resample1 = c(2L, 3L, 6L, 9L, 14L),
            Resample2 = c(4L, 11L, 13L, 14L, 15L)
          ),
          .Names = c("Resample1", "Resample2")
        ),
        number = 2
      ),
      .Names = c(
        "method",
        "index",
        "indexOut",
        "number"
      )
    )
  bt_4$method <- "boot_all"
  dat <- data.frame(y = 1:15, x = 15:1)
  bt_obj_4 <- caret2rsample(bt_4, data = dat)

  check_indices(bt_obj_4, bt_4)
  for (i in seq_along(bt_obj_4$splits)) {
    expect_equal(bt_obj_4$id[[i]], names(bt_4$index)[i])
  }
})

test_that("adaptive boot", {
  bt_5 <-
    structure(
      list(
        method = "boot",
        index = structure(
          list(
            Resample1 = c(
              1L,
              1L,
              4L,
              4L,
              5L,
              7L,
              8L,
              10L,
              11L,
              11L,
              12L,
              13L,
              15L,
              15L,
              15L
            ),
            Resample2 = c(
              1L,
              2L,
              3L,
              5L,
              5L,
              5L,
              6L,
              7L,
              8L,
              9L,
              9L,
              9L,
              10L,
              10L,
              12L
            )
          ),
          .Names = c("Resample1", "Resample2")
        ),
        indexOut = structure(
          list(
            Resample1 = c(2L, 3L, 6L, 9L, 14L),
            Resample2 = c(4L, 11L, 13L, 14L, 15L)
          ),
          .Names = c("Resample1", "Resample2")
        ),
        number = 2
      ),
      .Names = c(
        "method",
        "index",
        "indexOut",
        "number"
      )
    )
  bt_5$method <- "adaptive_boot"
  dat <- data.frame(y = 1:15, x = 15:1)
  bt_obj_5 <- caret2rsample(bt_5, data = dat)

  check_indices(bt_obj_5, bt_5)
  for (i in seq_along(bt_obj_5$splits)) {
    expect_equal(bt_obj_5$id[[i]], names(bt_5$index)[i])
  }
})


test_that("loo", {
  loo_1 <-
    structure(
      list(
        method = "LOOCV",
        index = structure(
          list(
            Fold01 = 2:15,
            Fold02 = c(
              1L,
              3L,
              4L,
              5L,
              6L,
              7L,
              8L,
              9L,
              10L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold03 = c(
              1L,
              2L,
              4L,
              5L,
              6L,
              7L,
              8L,
              9L,
              10L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold04 = c(
              1L,
              2L,
              3L,
              5L,
              6L,
              7L,
              8L,
              9L,
              10L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold05 = c(
              1L,
              2L,
              3L,
              4L,
              6L,
              7L,
              8L,
              9L,
              10L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold06 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              7L,
              8L,
              9L,
              10L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold07 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              6L,
              8L,
              9L,
              10L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold08 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              6L,
              7L,
              9L,
              10L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold09 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              6L,
              7L,
              8L,
              10L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold10 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              6L,
              7L,
              8L,
              9L,
              11L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold11 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              6L,
              7L,
              8L,
              9L,
              10L,
              12L,
              13L,
              14L,
              15L
            ),
            Fold12 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              6L,
              7L,
              8L,
              9L,
              10L,
              11L,
              13L,
              14L,
              15L
            ),
            Fold13 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              6L,
              7L,
              8L,
              9L,
              10L,
              11L,
              12L,
              14L,
              15L
            ),
            Fold14 = c(
              1L,
              2L,
              3L,
              4L,
              5L,
              6L,
              7L,
              8L,
              9L,
              10L,
              11L,
              12L,
              13L,
              15L
            ),
            Fold15 = 1:14
          ),
          .Names = c(
            "Fold01",
            "Fold02",
            "Fold03",
            "Fold04",
            "Fold05",
            "Fold06",
            "Fold07",
            "Fold08",
            "Fold09",
            "Fold10",
            "Fold11",
            "Fold12",
            "Fold13",
            "Fold14",
            "Fold15"
          )
        ),
        indexOut = structure(
          list(
            Resample01 = 1L,
            Resample02 = 2L,
            Resample03 = 3L,
            Resample04 = 4L,
            Resample05 = 5L,
            Resample06 = 6L,
            Resample07 = 7L,
            Resample08 = 8L,
            Resample09 = 9L,
            Resample10 = 10L,
            Resample11 = 11L,
            Resample12 = 12L,
            Resample13 = 13L,
            Resample14 = 14L,
            Resample15 = 15L
          ),
          .Names = c(
            "Resample01",
            "Resample02",
            "Resample03",
            "Resample04",
            "Resample05",
            "Resample06",
            "Resample07",
            "Resample08",
            "Resample09",
            "Resample10",
            "Resample11",
            "Resample12",
            "Resample13",
            "Resample14",
            "Resample15"
          )
        )
      ),
      .Names = c("method", "index", "indexOut")
    )
  dat <- data.frame(y = 1:15, x = 15:1)
  loo_obj <- caret2rsample(loo_1, data = dat)

  check_indices(loo_obj, loo_1)
  for (i in seq_along(loo_obj$splits)) {
    expect_equal(loo_obj$id[[i]], names(loo_1$index)[i])
  }
})

test_that("mcv", {
  lgo1 <-
    structure(
      list(
        method = "LGOCV",
        index = structure(
          list(
            Resample1 = c(1L, 4L, 5L, 6L, 7L, 9L, 10L, 14L),
            Resample2 = c(2L, 4L, 5L, 6L, 9L, 10L, 14L, 15L),
            Resample3 = c(1L, 2L, 3L, 5L, 6L, 7L, 8L, 9L)
          ),
          .Names = c("Resample1", "Resample2", "Resample3")
        ),
        indexOut = structure(
          list(
            Resample1 = c(2L, 3L, 8L, 11L, 12L, 13L, 15L),
            Resample2 = c(1L, 3L, 7L, 8L, 11L, 12L, 13L),
            Resample3 = c(4L, 10L, 11L, 12L, 13L, 14L, 15L)
          ),
          .Names = c("Resample1", "Resample2", "Resample3")
        ),
        number = 3,
        p = 0.5
      ),
      .Names = c("method", "index", "indexOut", "number", "p")
    )
  dat <- data.frame(y = 1:15, x = 15:1)
  mcv_obj <- caret2rsample(lgo1, data = dat)

  check_indices(mcv_obj, lgo1)
  for (i in seq_along(mcv_obj$splits)) {
    expect_equal(mcv_obj$id[[i]], names(lgo1$index)[i])
  }
})

test_that("rolling origin", {
  rof_1 <-
    structure(
      list(
        method = "timeSlice",
        index = structure(
          list(
            Training04 = 1:4,
            Training05 = 2:5,
            Training06 = 3:6,
            Training07 = 4:7,
            Training08 = 5:8,
            Training09 = 6:9,
            Training10 = 7:10
          ),
          .Names = c(
            "Training04",
            "Training05",
            "Training06",
            "Training07",
            "Training08",
            "Training09",
            "Training10"
          )
        ),
        indexOut = structure(
          list(
            Testing04 = 5:9,
            Testing05 = 6:10,
            Testing06 = 7:11,
            Testing07 = 8:12,
            Testing08 = 9:13,
            Testing09 = 10:14,
            Testing10 = 11:15
          ),
          .Names = c(
            "Testing04",
            "Testing05",
            "Testing06",
            "Testing07",
            "Testing08",
            "Testing09",
            "Testing10"
          )
        ),
        initialWindow = 4,
        horizon = 5,
        fixedWindow = TRUE,
        skip = 0
      ),
      .Names = c(
        "method",
        "index",
        "indexOut",
        "initialWindow",
        "horizon",
        "fixedWindow",
        "skip"
      )
    )
  dat <- data.frame(y = 1:15, x = 15:1)
  rof_obj <- caret2rsample(rof_1, data = dat)

  check_indices(rof_obj, rof_1)
  for (i in seq_along(rof_obj$splits)) {
    expect_equal(rof_obj$id[[i]], names(rof_1$index)[i])
  }
})

###################################################################
## Test cases for caret -> rsample that mimic `trainControl`

dat <- data.frame(y = 1:15, x = 15:1)

lgo1 <-
  structure(
    list(
      method = "LGOCV",
      index = structure(
        list(
          Resample1 = c(1L, 4L, 5L, 6L, 7L, 9L, 10L, 14L),
          Resample2 = c(2L, 4L, 5L, 6L, 9L, 10L, 14L, 15L),
          Resample3 = c(1L, 2L, 3L, 5L, 6L, 7L, 8L, 9L)
        ),
        .Names = c("Resample1", "Resample2", "Resample3")
      ),
      indexOut = structure(
        list(
          Resample1 = c(2L, 3L, 8L, 11L, 12L, 13L, 15L),
          Resample2 = c(1L, 3L, 7L, 8L, 11L, 12L, 13L),
          Resample3 = c(4L, 10L, 11L, 12L, 13L, 14L, 15L)
        ),
        .Names = c("Resample1", "Resample2", "Resample3")
      ),
      number = 3,
      p = 0.5
    ),
    .Names = c("method", "index", "indexOut", "number", "p")
  )

cv_1 <- structure(
  list(
    method = "cv",
    index = structure(
      list(
        Fold1 = c(2L, 3L, 4L, 5L, 6L, 7L, 8L, 10L, 11L, 13L),
        Fold2 = c(
          1L, 4L, 6L,
          8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L
        ),
        Fold3 = c(1L, 2L, 3L, 5L, 7L, 9L, 12L, 14L, 15L)
      ),
      .Names = c("Fold1", "Fold2", "Fold3")
    ),
    indexOut = structure(
      list(
        Resample1 = c(1L, 9L, 12L, 14L, 15L),
        Resample2 = c(2L, 3L, 5L, 7L),
        Resample3 = c(4L, 6L, 8L, 10L, 11L, 13L)
      ),
      .Names = c("Resample1", "Resample2", "Resample3")
    ),
    number = 3,
    repeats = NA
  ),
  .Names = c("method", "index", "indexOut", "number", "repeats")
)

cv_2 <-
  structure(
    list(
      method = "repeatedcv",
      index = structure(
        list(
          Fold1.Rep1 = c(1L, 3L, 4L, 6L, 9L, 10L, 12L, 13L, 14L, 15L),
          Fold2.Rep1 = c(2L, 5L, 7L, 8L, 10L, 11L, 13L, 14L, 15L),
          Fold3.Rep1 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 11L, 12L),
          Fold1.Rep2 = c(1L, 2L, 3L, 5L, 6L, 7L, 10L, 11L, 12L, 14L),
          Fold2.Rep2 = c(2L, 4L, 6L, 8L, 9L, 11L, 13L, 14L, 15L),
          Fold3.Rep2 = c(1L, 3L, 4L, 5L, 7L, 8L, 9L, 10L, 12L, 13L, 15L)
        ),
        .Names = c(
          "Fold1.Rep1", "Fold2.Rep1", "Fold3.Rep1",
          "Fold1.Rep2", "Fold2.Rep2", "Fold3.Rep2"
        )
      ),
      indexOut = structure(
        list(
          Resample1 = c(2L, 5L, 7L, 8L, 11L),
          Resample2 = c(1L, 3L, 4L, 6L, 9L, 12L),
          Resample3 = c(10L, 13L, 14L, 15L),
          Resample4 = c(4L, 8L, 9L, 13L, 15L),
          Resample5 = c(1L, 3L, 5L, 7L, 10L, 12L),
          Resample6 = c(2L, 6L, 11L, 14L)
        ),
        .Names = c(
          "Resample1", "Resample2", "Resample3",
          "Resample4", "Resample5", "Resample6"
        )
      ),
      number = 3,
      repeats = 2
    ),
    .Names = c("method", "index", "indexOut", "number", "repeats")
  )

cv_3 <- cv_2
cv_3$method <- "adaptive_cv"

bt_1 <-
  structure(
    list(
      method = "boot",
      index = structure(
        list(
          Resample1 = c(1L, 1L, 4L, 4L, 5L, 7L, 8L, 10L, 11L, 11L, 12L, 13L, 15L, 15L, 15L),
          Resample2 = c(1L, 2L, 3L, 5L, 5L, 5L, 6L, 7L, 8L, 9L, 9L, 9L, 10L, 10L, 12L)
        ),
        .Names = c("Resample1", "Resample2")
      ),
      indexOut = structure(
        list(
          Resample1 = c(2L, 3L, 6L, 9L, 14L),
          Resample2 = c(4L, 11L, 13L, 14L, 15L)
        ),
        .Names = c("Resample1", "Resample2")
      ),
      number = 2
    ),
    .Names = c(
      "method",
      "index", "indexOut", "number"
    )
  )
bt_2 <- bt_1
bt_2$method <- "boot632"

bt_3 <- bt_1
bt_3$method <- "optimism_boot"

bt_4 <- bt_1
bt_4$method <- "boot_all"

bt_5 <- bt_1
bt_5$method <- "adaptive_boot"

loo_1 <-
  structure(
    list(
      method = "LOOCV",
      index = structure(
        list(
          Fold01 = 2:15,
          Fold02 = c(1L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold03 = c(1L, 2L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold04 = c(1L, 2L, 3L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold05 = c(1L, 2L, 3L, 4L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold06 = c(1L, 2L, 3L, 4L, 5L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold07 = c(1L, 2L, 3L, 4L, 5L, 6L, 8L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold08 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 9L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold09 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 10L, 11L, 12L, 13L, 14L, 15L),
          Fold10 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 11L, 12L, 13L, 14L, 15L),
          Fold11 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 12L, 13L, 14L, 15L),
          Fold12 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 13L, 14L, 15L),
          Fold13 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 14L, 15L),
          Fold14 = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L, 11L, 12L, 13L, 15L),
          Fold15 = 1:14
        ),
        .Names = c(
          "Fold01", "Fold02", "Fold03", "Fold04", "Fold05", "Fold06",
          "Fold07", "Fold08", "Fold09", "Fold10", "Fold11", "Fold12",
          "Fold13", "Fold14", "Fold15"
        )
      ),
      indexOut = structure(
        list(
          Resample01 = 1L, Resample02 = 2L,
          Resample03 = 3L, Resample04 = 4L,
          Resample05 = 5L, Resample06 = 6L,
          Resample07 = 7L, Resample08 = 8L,
          Resample09 = 9L, Resample10 = 10L,
          Resample11 = 11L, Resample12 = 12L,
          Resample13 = 13L, Resample14 = 14L,
          Resample15 = 15L
        ),
        .Names = c(
          "Resample01", "Resample02", "Resample03", "Resample04",
          "Resample05", "Resample06", "Resample07", "Resample08",
          "Resample09", "Resample10", "Resample11", "Resample12",
          "Resample13", "Resample14", "Resample15"
        )
      )
    ),
    .Names = c("method", "index", "indexOut")
  )

rof_1 <-
  structure(
    list(
      method = "timeSlice",
      index = structure(
        list(
          Training04 = 1:4, Training05 = 2:5, Training06 = 3:6,
          Training07 = 4:7, Training08 = 5:8, Training09 = 6:9,
          Training10 = 7:10
        ),
        .Names = c(
          "Training04", "Training05", "Training06",  "Training07",
          "Training08", "Training09", "Training10"
        )
      ),
      indexOut = structure(
        list(
          Testing04 = 5:9, Testing05 = 6:10, Testing06 = 7:11,
          Testing07 = 8:12, Testing08 = 9:13, Testing09 = 10:14,
          Testing10 = 11:15
        ),
        .Names = c(
          "Testing04", "Testing05", "Testing06", "Testing07",
          "Testing08", "Testing09", "Testing10"
        )
      ),
      initialWindow = 4, horizon = 5, fixedWindow = TRUE, skip = 0
    ),
    .Names = c(
      "method", "index", "indexOut", "initialWindow",
      "horizon", "fixedWindow", "skip"
    )
  )

###################################################################
##

check_indices <- function(newer, orig) {
  for (i in seq_along(newer$splits)) {
    expect_equal(
      as.integer(newer$splits[[i]]),
      orig$index[[i]]
    )
    expect_equal(
      as.integer(newer$splits[[i]], "assessment"),
      orig$indexOut[[i]]
    )
  }
  invisible(NULL)
}

###################################################################

test_that("basic v-fold", {
  vfold_obj_1 <- caret2rsample(cv_1, data = dat)
  check_indices(vfold_obj_1, cv_1)
  for (i in seq_along(vfold_obj_1$splits)) {
    expect_equal(vfold_obj_1$id[[i]], names(cv_1$index)[i])
  }
})

test_that("repeated v-fold", {
  vfold_obj_2 <- caret2rsample(cv_2, data = dat)
  check_indices(vfold_obj_2, cv_2)
  for (i in seq_along(vfold_obj_2$splits)) {
    expect_equal(
      paste(vfold_obj_2$id2[[i]], vfold_obj_2$id[[i]],
        sep = "."
      ),
      names(cv_2$index)[i]
    )
  }
})

test_that("basic boot", {
  bt_obj_1 <- caret2rsample(bt_1, data = dat)
  check_indices(bt_obj_1, bt_1)
  for (i in seq_along(bt_obj_1$splits)) {
    expect_equal(bt_obj_1$id[[i]], names(bt_1$index)[i])
  }
})

test_that("boot 632", {
  bt_obj_2 <- caret2rsample(bt_2, data = dat)
  check_indices(bt_obj_2, bt_2)
  for (i in seq_along(bt_obj_2$splits)) {
    expect_equal(bt_obj_2$id[[i]], names(bt_2$index)[i])
  }
})

test_that("boot optim", {
  bt_obj_3 <- caret2rsample(bt_3, data = dat)
  check_indices(bt_obj_3, bt_3)
  for (i in seq_along(bt_obj_3$splits)) {
    expect_equal(bt_obj_3$id[[i]], names(bt_3$index)[i])
  }
})

test_that("boot all", {
  bt_obj_4 <- caret2rsample(bt_4, data = dat)
  check_indices(bt_obj_4, bt_4)
  for (i in seq_along(bt_obj_4$splits)) {
    expect_equal(bt_obj_4$id[[i]], names(bt_4$index)[i])
  }
})

test_that("adaptive boot", {
  bt_obj_5 <- caret2rsample(bt_5, data = dat)
  check_indices(bt_obj_5, bt_5)
  for (i in seq_along(bt_obj_5$splits)) {
    expect_equal(bt_obj_5$id[[i]], names(bt_5$index)[i])
  }
})


test_that("loo", {
  loo_obj <- caret2rsample(loo_1, data = dat)
  check_indices(loo_obj, loo_1)
  for (i in seq_along(loo_obj$splits)) {
    expect_equal(loo_obj$id[[i]], names(loo_1$index)[i])
  }
})

test_that("mcv", {
  mcv_obj <- caret2rsample(lgo1, data = dat)
  check_indices(mcv_obj, lgo1)
  for (i in seq_along(mcv_obj$splits)) {
    expect_equal(mcv_obj$id[[i]], names(lgo1$index)[i])
  }
})

test_that("rolling origin", {
  rof_obj <- caret2rsample(rof_1, data = dat)
  check_indices(rof_obj, rof_1)
  for (i in seq_along(rof_obj$splits)) {
    expect_equal(rof_obj$id[[i]], names(rof_1$index)[i])
  }
})

#' Compatibility with dplyr
#'
#' @description
#' rsample should be fully compatible with dplyr 1.0.0.
#'
#' With older versions of dplyr, there is partial support for the following
#' verbs: `mutate()`, `arrange()`, `filter()`, `rename()`, `select()`, and
#' `slice()`. We strongly recommend updating to dplyr 1.0.0 if possible to
#' get more complete integration with dplyr.
#'
#' @section Version Specific Behavior:
#'
#' rsample performs somewhat differently depending on whether you have
#' dplyr >= 1.0.0 (new) or dplyr < 1.0.0 (old). Additionally, version
#' 0.0.7 of rsample (new) introduced some changes to how rsample objects
#' work with dplyr, even on old dplyr. Most of these changes influence the
#' return value of a dplyr verb and determine whether it will be a tibble
#' or an rsample rset subclass.
#'
#' The table below attempts to capture most of these changes. These examples
#' are not exhaustive and may not capture some edge-cases.
#'
#' ## Joins
#'
#' The following affect all of the dplyr joins, such as `left_join()`,
#' `right_join()`, `full_join()`, and `inner_join()`.
#'
#' Joins that alter the rows of the original rset object:
#'
#' | operation                  | old rsample + old dplyr | new rsample + old dplyr | new rsample + new dplyr
#' | :------------------------- | :---------------------: | :---------------------: | :---------------------:
#' | `join(rset, tbl)`          | error                   | error                   | tibble
#'
#' The idea here is that, if there are less rows in the result, the result should
#' not be an rset object. For example, you can't have a 10-fold CV object
#' without 10 rows.
#'
#' Joins that keep the rows of the original rset object:
#'
#' | operation                  | old rsample + old dplyr | new rsample + old dplyr | new rsample + new dplyr
#' | :------------------------- | :---------------------: | :---------------------: | :---------------------:
#' | `join(rset, tbl)`          | error                   | error                   | rset
#'
#' As with the logic above, if the original rset object (defined by the split
#' column and the id column(s)) is left intact, the results should be an rset.
#'
#' ## Row Subsetting
#'
#' As mentioned above, this should result in a tibble if any rows are removed
#' or added. Simply reordering rows still results in a valid rset with new
#' rsample.
#'
#' Cases where rows are removed or added:
#'
#' | operation       | old rsample + old dplyr | new rsample + old dplyr | new rsample + new dplyr
#' | :-------------- | :---------------------: | :---------------------: | :---------------------:
#' | `rset[ind,]`    | tibble                  | tibble                  | tibble
#' | `slice(rset)`   | rset                    | tibble                  | tibble
#' | `filter(rset)`  | rset                    | tibble                  | tibble
#'
#' Cases where all rows are kept, but are possibly reordered:
#'
#' | operation       | old rsample + old dplyr | new rsample + old dplyr | new rsample + new dplyr
#' | :-------------- | :---------------------: | :---------------------: | :---------------------:
#' | `rset[ind,]`    | tibble                  | rset                    | rset
#' | `slice(rset)`   | rset                    | rset                    | rset
#' | `filter(rset)`  | rset                    | rset                    | rset
#' | `arrange(rset)` | rset                    | rset                    | rset
#'
#' ## Column Subsetting
#'
#' When the `splits` column or any `id` columns are dropped or renamed,
#' the result should no longer be considered a valid rset.
#'
#' Cases when the required columns are removed or renamed:
#'
#' | operation       | old rsample + old dplyr | new rsample + old dplyr | new rsample + new dplyr
#' | :-------------- | :---------------------: | :---------------------: | :---------------------:
#' | `rset[,ind]`    | tibble                  | tibble                  | tibble
#' | `select(rset)`  | rset                    | tibble                  | tibble
#' | `rename(rset)`  | tibble                  | tibble                  | tibble
#'
#' Cases when no required columns are affected:
#'
#' | operation       | old rsample + old dplyr | new rsample + old dplyr | new rsample + new dplyr
#' | :-------------- | :---------------------: | :---------------------: | :---------------------:
#' | `rset[,ind]`    | tibble                  | rset                    | rset
#' | `select(rset)`  | rset                    | rset                    | rset
#' | `rename(rset)`  | rset                    | rset                    | rset
#'
#' ## Other Column Operations
#'
#' Cases when the required columns are altered:
#'
#' | operation       | old rsample + old dplyr | new rsample + old dplyr | new rsample + new dplyr
#' | :-------------- | :---------------------: | :---------------------: | :---------------------:
#' | `mutate(rset)`  | rset                    | tibble                  | tibble
#'
#' Cases when no required columns are affected:
#'
#' | operation       | old rsample + old dplyr | new rsample + old dplyr | new rsample + new dplyr
#' | :-------------- | :---------------------: | :---------------------: | :---------------------:
#' | `mutate(rset)`  | rset                    | rset                    | rset
#'
#' @name rsample-dplyr
NULL

# `dplyr_reconstruct()`
#
# `dplyr_reconstruct()` is called:
# - After a complex dplyr operation, like a `left_join()`, to restore to the
#   type of the first input, `x`.
# - At the end of a call to `dplyr_col_modify()`
# - At the end of a call to `dplyr_row_slice()`
# - See `?dplyr_reconstruct` for the full list.
#
# Because `dplyr_reconstruct()` is called at the end of `dplyr_col_modify()`
# and `dplyr_row_slice()`, we don't need methods for them. The default methods
# in dplyr do the right thing automatically, and then our reconstruction
# method decides whether or not the result should still be an rset.
#
# The implementation for rsample is the same as `vec_restore()`. Generally
# it will fall back to reconstructing a bare tibble, unless the rset structure
# is still completely intact. This happens when rset specific rows and columns
# (splits, id cols) are still exactly identical to how they were before the
# dplyr operation (with the exception of column reordering).

# Registered in `.onLoad()`
dplyr_reconstruct_rset <- function(data, template) {
  rset_reconstruct(data, template)
}

#' Time-based Resampling
#'
#' @description
#' These resampling functions are focused on various forms of _time series_
#' resampling.
#'
#' - `sliding_window()` uses the row number when computing the resampling
#'   indices. It is independent of any time index, but is useful with
#'   completely regular series.
#'
#' - `sliding_index()` computes resampling indices relative to the `index`
#'   column. This is often a Date or POSIXct column, but doesn't have to be.
#'   This is useful when resampling irregular series, or for using irregular
#'   lookback periods such as `lookback = lubridate::years(1)` with daily
#'   data (where the number of days in a year may vary).
#'
#' - `sliding_period()` first breaks up the `index` into less granular groups
#'   based on `period`, and then uses that to construct the resampling indices.
#'   This is extremely useful for constructing rolling monthly or yearly
#'   windows from daily data.
#'
#' @inheritParams rlang::args_dots_empty
#'
#' @param data A data frame.
#'
#' @param index The index to compute resampling indices relative to, specified
#'   as a bare column name. This must be an existing column in `data`.
#'
#'   - For `sliding_index()`, this is commonly a date vector, but is not
#'     required.
#'
#'   - For `sliding_period()`, it is required that this is a Date or POSIXct
#'     vector.
#'
#'   The `index` must be an _increasing_ vector, but duplicate values are
#'   allowed. Additionally, the index cannot contain any missing values.
#'
#' @param period The period to group the `index` by. This is specified as a
#'   single string, such as `"year"` or `"month"`. See the `.period` argument
#'   of [slider::slide_period()] for the full list of options and further
#'   explanation.
#'
#' @param lookback The number of elements to look back from the current element
#'   when computing the resampling indices of the analysis set. The current
#'   row is always included in the analysis set.
#'
#'   - For `sliding_window()`, a single integer defining the number of rows to
#'     look back from the current row.
#'
#'   - For `sliding_index()`, a single object that will be subtracted from the
#'     `index` as `index - lookback` to define the boundary of where to start
#'     searching for rows to include in the current resample. This is often
#'     an integer value corresponding to the number of days to look back,
#'     or a lubridate Period object.
#'
#'   - For `sliding_period()`, a single integer defining the number of groups
#'     to look back from the current group, where the groups were defined from
#'     breaking up the `index` according to the `period`.
#'
#'   In all cases, `Inf` is also allowed to force an expanding window.
#'
#' @param assess_start,assess_stop This combination of arguments determines
#'   how far into the future to look when constructing the assessment set.
#'   Together they construct a range of
#'   `[index + assess_start, index + assess_stop]` to search for rows to include
#'   in the assessment set.
#'
#'   Generally, `assess_start` will always be `1` to indicate that the first
#'   value to potentially include in the assessment set should start one element
#'   after the current row, but it can be increased to a larger value to
#'   create "gaps" between the analysis and assessment set if you are worried
#'   about high levels of correlation in short term forecasting.
#'
#'   - For `sliding_window()`, these are both single integers defining the
#'     number of rows to look forward from the current row.
#'
#'   - For `sliding_index()`, these are single objects that will be added
#'     to the `index` to compute the range to search for rows to include
#'     in the assessment set. This is often an integer value corresponding to
#'     the number of days to look forward, or a lubridate Period object.
#'
#'   - For `sliding_period()`, these are both single integers defining the
#'     number of groups to look forward from the current group, where the groups
#'     were defined from breaking up the `index` according to the `period`.
#'
#' @param complete A single logical. When using `lookback` to compute the
#'   analysis sets, should only complete windows be considered? If set to
#'   `FALSE`, partial windows will be used until it is possible to create
#'   a complete window (based on `lookback`). This is a way to use an
#'   expanding window up to a certain point, and then switch to a sliding
#'   window.
#'
#' @param step A single positive integer. After computing the resampling
#'   indices, `step` is used to thin out the results by selecting every
#'   `step`-th result by subsetting the indices with
#'   `seq(1L, n_indices, by = step)`. `step` is applied after `skip`.
#'   Note that `step` is independent of any time `index` used.
#'
#' @param skip A single positive integer, or zero. After computing the
#'   resampling indices, the first `skip` results will be dropped by subsetting
#'   the indices with `seq(skip + 1L, n_indices)`. This can be especially
#'   useful when combined with `lookback = Inf`, which creates an expanding
#'   window starting from the first row. By skipping forward, you can drop
#'   the first few windows that have very few data points. `skip` is
#'   applied before `step`. Note that `skip` is independent of any time
#'   `index` used.
#'
#' @param every A single positive integer. The number of periods to group
#'   together.
#'
#'   For example, if the `period` was set to `"year"` with an `every`
#'   value of 2, then the years 1970 and 1971 would be placed in the same
#'   group.
#'
#' @param origin The reference date time value. The default when left
#'   as `NULL` is the epoch time of `1970-01-01 00:00:00`,
#'   _in the time zone of the index_.
#'
#'   This is generally used to define the anchor time to count from,
#'   which is relevant when the `every` value is `> 1`.
#'
#' @seealso
#' [rolling_origin()]
#'
#' [slider::slide()], [slider::slide_index()], and [slider::slide_period()],
#' which power these resamplers.
#'
#' @name slide-resampling
#'
#' @examplesIf rlang::is_installed("modeldata")
#' library(vctrs)
#' library(tibble)
#' library(modeldata)
#' data("Chicago")
#'
#' index <- new_date(c(1, 3, 4, 7, 8, 9, 13, 15, 16, 17))
#' df <- tibble(x = 1:10, index = index)
#' df
#'
#' # Look back two rows beyond the current row, for a total of three rows
#' # in each analysis set. Each assessment set is composed of the two rows after
#' # the current row.
#' sliding_window(df, lookback = 2, assess_stop = 2)
#'
#' # Same as before, but step forward by 3 rows between each resampling slice,
#' # rather than just by 1.
#' rset <- sliding_window(df, lookback = 2, assess_stop = 2, step = 3)
#' rset
#'
#' analysis(rset$splits[[1]])
#' analysis(rset$splits[[2]])
#'
#' # Now slide relative to the `index` column in `df`. This time we look back
#' # 2 days from the current row's `index` value, and 2 days forward from
#' # it to construct the assessment set. Note that this series is irregular,
#' # so it produces different results than `sliding_window()`. Additionally,
#' # note that it is entirely possible for the assessment set to contain no
#' # data if you have a highly irregular series and "look forward" into a
#' # date range where no data points actually exist!
#' sliding_index(df, index, lookback = 2, assess_stop = 2)
#'
#' # With `sliding_period()`, we can break up our date index into less granular
#' # chunks, and slide over them instead of the index directly. Here we'll use
#' # the Chicago data, which contains daily data spanning 16 years, and we'll
#' # break it up into rolling yearly chunks. Three years worth of data will
#' # be used for the analysis set, and one years worth of data will be held out
#' # for performance assessment.
#' sliding_period(
#'   Chicago,
#'   date,
#'   "year",
#'   lookback = 2,
#'   assess_stop = 1
#' )
#'
#' # Because `lookback = 2`, three years are required to form a "complete"
#' # window of data. To allow partial windows, set `complete = FALSE`.
#' # Here that first constructs two expanding windows until a complete three
#' # year window can be formed, at which point we switch to a sliding window.
#' sliding_period(
#'   Chicago,
#'   date,
#'   "year",
#'   lookback = 2,
#'   assess_stop = 1,
#'   complete = FALSE
#' )
#'
#' # Alternatively, you could break the resamples up by month. Here we'll
#' # use an expanding monthly window by setting `lookback = Inf`, and each
#' # assessment set will contain two months of data. To ensure that we have
#' # enough data to fit our models, we'll `skip` the first 4 expanding windows.
#' # Finally, to thin out the results, we'll `step` forward by 2 between
#' # each resample.
#' sliding_period(
#'   Chicago,
#'   date,
#'   "month",
#'   lookback = Inf,
#'   assess_stop = 2,
#'   skip = 4,
#'   step = 2
#' )
NULL

#' @export
#' @rdname slide-resampling
sliding_window <- function(
  data,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L
) {
  rlang::check_dots_empty()

  if (!is.data.frame(data)) {
    cli_abort("{.arg data} must be a data frame.")
  }

  lookback <- check_lookback(lookback)
  assess_start <- check_assess(assess_start, "assess_start")
  assess_stop <- check_assess(assess_stop, "assess_stop")
  step <- check_step(step)
  skip <- check_skip(skip)

  if (assess_start > assess_stop) {
    cli_abort(
      "{.arg assess_start} must be less than or equal to {.arg assess_stop}."
    )
  }

  seq <- vctrs::vec_seq_along(data)

  id_in <- slider::slide(
    .x = seq,
    .f = identity,
    .before = lookback,
    .after = 0L,
    .step = 1L,
    .complete = complete
  )

  id_out <- slider::slide(
    .x = seq,
    .f = identity,
    .before = -assess_start,
    .after = assess_stop,
    .step = 1L,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (!identical(skip, 0L)) {
    indices <- slice_skip(indices, skip)
  }

  if (!identical(step, 1L)) {
    indices <- slice_step(indices, step)
  }

  splits <- purrr::map(
    indices,
    \(.x) make_splits(.x, data = data, class = "sliding_window_split")
  )

  ids <- names0(length(indices), prefix = "Slice")

  attrib <- list(
    lookback = lookback,
    assess_start = assess_start,
    assess_stop = assess_stop,
    complete = complete,
    step = step,
    skip = skip
  )

  new_rset(
    splits = splits,
    ids = ids,
    attrib = attrib,
    subclass = c("sliding_window", "rset")
  )
}

# ------------------------------------------------------------------------------

#' @export
#' @rdname slide-resampling
sliding_index <- function(
  data,
  index,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L
) {
  rlang::check_dots_empty()

  if (!is.data.frame(data)) {
    cli_abort("{.arg data} must be a data frame.")
  }

  step <- check_step(step)
  skip <- check_skip(skip)

  index <- rlang::enexpr(index)
  loc <- tidyselect::eval_select(index, data)

  if (length(loc) != 1L) {
    cli_abort("{.arg index} must specify exactly one column in {.arg data}.")
  }

  index_attrib <- index
  index <- data[[loc]]

  seq <- vctrs::vec_seq_along(data)

  id_in <- slider::slide_index(
    .x = seq,
    .i = index,
    .f = identity,
    .before = lookback,
    .after = 0L,
    .complete = complete
  )

  id_out <- slider::slide_index(
    .x = seq,
    .i = index,
    .f = identity,
    .before = -assess_start,
    .after = assess_stop,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (!identical(skip, 0L)) {
    indices <- slice_skip(indices, skip)
  }

  if (!identical(step, 1L)) {
    indices <- slice_step(indices, step)
  }

  splits <- purrr::map(
    indices,
    \(.x) make_splits(.x, data = data, class = "sliding_index_split")
  )

  ids <- names0(length(indices), prefix = "Slice")

  attrib <- list(
    index = index_attrib,
    lookback = lookback,
    assess_start = assess_start,
    assess_stop = assess_stop,
    complete = complete,
    step = step,
    skip = skip
  )

  new_rset(
    splits = splits,
    ids = ids,
    attrib = attrib,
    subclass = c("sliding_index", "rset")
  )
}

# ------------------------------------------------------------------------------

#' @export
#' @rdname slide-resampling
sliding_period <- function(
  data,
  index,
  period,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L,
  every = 1L,
  origin = NULL
) {
  rlang::check_dots_empty()

  if (!is.data.frame(data)) {
    cli_abort("{.arg data} must be a data frame.")
  }

  lookback <- check_lookback(lookback)
  assess_start <- check_assess(assess_start, "assess_start")
  assess_stop <- check_assess(assess_stop, "assess_stop")
  step <- check_step(step)

  if (assess_start > assess_stop) {
    cli_abort(
      "{.arg assess_start} must be less than or equal to {.arg assess_stop}."
    )
  }

  index <- rlang::enexpr(index)
  loc <- tidyselect::eval_select(index, data)

  if (length(loc) != 1L) {
    cli_abort("{.arg index} must specify exactly one column in {.arg data}.")
  }

  index_attrib <- index
  index <- data[[loc]]

  seq <- vctrs::vec_seq_along(data)

  id_in <- slider::slide_period(
    .x = seq,
    .i = index,
    .period = period,
    .f = identity,
    .every = every,
    .origin = origin,
    .before = lookback,
    .after = 0L,
    .complete = complete
  )

  id_out <- slider::slide_period(
    .x = seq,
    .i = index,
    .period = period,
    .f = identity,
    .every = every,
    .origin = origin,
    .before = -assess_start,
    .after = assess_stop,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (!identical(skip, 0L)) {
    indices <- slice_skip(indices, skip)
  }

  if (!identical(step, 1L)) {
    indices <- slice_step(indices, step)
  }

  splits <- purrr::map(
    indices,
    \(.x) make_splits(.x, data = data, class = "sliding_period_split")
  )

  ids <- names0(length(indices), prefix = "Slice")

  attrib <- list(
    index = index_attrib,
    period = period,
    lookback = lookback,
    assess_start = assess_start,
    assess_stop = assess_stop,
    complete = complete,
    step = step,
    skip = skip,
    every = every,
    origin = origin
  )

  new_rset(
    splits = splits,
    ids = ids,
    attrib = attrib,
    subclass = c("sliding_period", "rset")
  )
}

# ------------------------------------------------------------------------------

check_lookback <- function(x) {
  if (vctrs::vec_size(x) != 1L) {
    cli_abort("{.arg lookback} must have size 1.")
  }

  if (identical(x, Inf)) {
    return(x)
  }

  if (!rlang::is_integerish(x, finite = TRUE)) {
    cli_abort("{.arg lookback} must be an integer of size 1, or {.arg Inf}.")
  }

  if (x < 0L) {
    cli_abort("{.arg lookback} must be positive, or zero.")
  }

  vctrs::vec_cast(x, integer(), x_arg = "lookback")
}

check_assess <- function(x, arg) {
  if (vctrs::vec_size(x) != 1L) {
    cli_abort("{.arg {arg}} must have size 1.")
  }

  if (identical(x, Inf)) {
    return(x)
  }

  if (!rlang::is_integerish(x, finite = TRUE)) {
    cli_abort("{.arg {arg}} must be an integer of size 1, or {.code Inf}.")
  }

  if (x <= 0L) {
    cli_abort("{.arg {arg}} must be positive.")
  }

  vctrs::vec_cast(x, integer(), x_arg = arg)
}

check_step <- function(x) {
  if (vctrs::vec_size(x) != 1L) {
    cli_abort("{.arg step} must have size 1.")
  }

  if (!rlang::is_integerish(x, finite = TRUE)) {
    cli_abort("{.arg step} must be an integer of size 1.")
  }

  if (x <= 0L) {
    cli_abort("{.arg step} must be positive.")
  }

  vctrs::vec_cast(x, integer(), x_arg = "step")
}

check_skip <- function(x) {
  if (vctrs::vec_size(x) != 1L) {
    cli_abort("{.arg skip} must have size 1.")
  }

  if (!rlang::is_integerish(x, finite = TRUE)) {
    cli_abort("{.arg skip}` must be an integer of size 1.")
  }

  if (x < 0L) {
    cli_abort("{.arg skip} must be positive, or zero.")
  }

  vctrs::vec_cast(x, integer(), x_arg = "skip")
}

compute_complete_indices <- function(id_in, id_out) {
  # Remove where either list has a `NULL` element.
  # These are incomplete windows.
  id_in_na <- vctrs::vec_detect_missing(id_in)
  id_out_na <- vctrs::vec_detect_missing(id_out)

  id_either_na <- id_in_na | id_out_na

  id_in <- vctrs::vec_slice(id_in, !id_either_na)
  id_out <- vctrs::vec_slice(id_out, !id_either_na)

  purrr::map2(id_in, id_out, merge_lists)
}

slice_skip <- function(indices, skip) {
  n_indices <- length(indices)
  slicer <- rlang::seq2(skip + 1L, n_indices)
  vctrs::vec_slice(indices, slicer)
}

slice_step <- function(indices, step) {
  n_indices <- length(indices)
  slicer <- seq2_by(1L, n_indices, by = step)
  indices <- vctrs::vec_slice(indices, slicer)
}

seq2_by <- function(from, to, by) {
  if (length(from) != 1) {
    cli_abort("{.arg from} must be length one")
  }
  if (length(to) != 1) {
    cli_abort("{.arg to} must be length one")
  }

  by <- as.integer(by)

  if (length(by) != 1) {
    cli_abort("{.arg by} must be length one")
  }
  if (by <= 0L) {
    cli_abort("{.arg by} must be positive")
  }

  if (from > to) {
    integer()
  } else {
    seq.int(from, to, by = by)
  }
}

#' Make groupings for grouped rsplits
#'
#' This function powers [group_vfold_cv] by splitting the data based upon
#' a grouping variable and returning the assessment set indices for each
#' split.
#'
#' @inheritParams vfold_cv
#' @param group A variable in `data` (single character or name) used for
#'  grouping observations with the same value to either the analysis or
#'  assessment set within a fold.
#'
#' @keywords internal
make_groups <- function(data, group, v) {
  data_ind <- data.frame(..index = 1:nrow(data), ..group = group)
  data_ind$..group <- as.character(data_ind$..group)

  res <- balance_groups(data_ind, v)
  data_ind <- res$data_ind
  keys <- res$keys

  keys$..group <- as.character(keys$..group)

  data_ind <- data_ind %>%
    full_join(keys, by = "..group") %>%
    arrange(..index)
  split_unnamed(data_ind$..index, data_ind$..folds)

}

balance_groups <- function(data_ind, v) {
  unique_groups <- unique(data_ind$..group)
  keys <- data.frame(
    ..group = unique_groups,
    ..folds = sample(rep(seq_len(v), length.out = length(unique_groups)))
  )
  list(
    data_ind = data_ind,
    keys = keys
  )
}

#' "Reshuffle" an rset to re-generate a new rset with the same parameters
#'
#' This function re-generates an rset object, using the same arguments used
#' to generate the original.
#'
#' @param rset The `rset` object to be reshuffled
#'
#' @return An rset of the same class as `rset`.
#'
#' @examples
#' set.seed(123)
#' (starting_splits <- group_vfold_cv(mtcars, cyl, v = 3))
#' reshuffle_rset(starting_splits)
#'
#' @export
reshuffle_rset <- function(rset) {
  if (!inherits(rset, "rset")) {
    cli_abort("{.arg rset} must be an {.cls rset} object.")
  }

  if (inherits(rset, "manual_rset")) {
    cli_abort("{.arg manual_rset} objects cannot be reshuffled.")
  }

  # non-random classes is defined below
  if (any(non_random_classes %in% class(rset))) {
    cls <- class(rset)[[1]]
    cli::cli_warn(
      "{.fun reshuffle_rset} will return an identical {.cls rset} when called on {.cls {cls}} objects."
    )
    if ("validation_set" %in% class(rset)) {
      return(rset)
    }
  }

  rset_type <- class(rset)[[1]]
  split_arguments <- .get_split_args(rset)
  if (identical(split_arguments$strata, TRUE)) {
    cli_abort(c(
      "Cannot reshuffle this rset ({.code attr(rset, 'strata')} is {.val TRUE}, not a column identifier)",
      i = "If the original object was created with an older version of rsample, try recreating it with the newest version of the package."
    ))
  }

  do.call(
    rset_type,
    c(list(data = rset$splits[[1]]$data), split_arguments)
  )
}

non_random_classes <- c(
  "sliding_index",
  "sliding_period",
  "sliding_window",
  "rolling_origin",
  "validation_time_split",
  "validation_set"
)

#' Extending rsample with new rset subclasses
#'
#' `rset_reconstruct()` encapsulates the logic for allowing new rset
#' subclasses to work properly with vctrs (through `vctrs::vec_restore()`) and
#' dplyr (through `dplyr::dplyr_reconstruct()`). It is intended to be a
#' developer tool, and is not required for normal usage of rsample.
#'
#' rset objects are considered "reconstructable" after a vctrs/dplyr operation
#' if:
#'
#' - `x` and `to` both have an identical column named `"splits"` (column
#'   and row order do not matter).
#'
#' - `x` and `to` both have identical columns prefixed with `"id"` (column
#'   and row order do not matter).
#'
#' @param x A data frame to restore to an rset subclass.
#' @param to An rset subclass to restore to.
#'
#' @return `x` restored to the rset subclass of `to`.
#'
#' @export
#' @examples
#' to <- bootstraps(mtcars, times = 25)
#'
#' # Imitate a vctrs/dplyr operation,
#' # where the class might be lost along the way
#' x <- tibble::as_tibble(to)
#'
#' # Say we added a new column to `x`. Here we mock a `mutate()`.
#' x$foo <- "bar"
#'
#' # This is still reconstructable to `to`
#' rset_reconstruct(x, to)
#'
#' # Say we lose the first row
#' x <- x[-1, ]
#'
#' # This is no longer reconstructable to `to`, as `x` is no longer an rset
#' # bootstraps object with 25 bootstraps if one is lost!
#' rset_reconstruct(x, to)
rset_reconstruct <- function(x, to) {
  if (rset_reconstructable(x, to)) {
    df_reconstruct(x, to)
  } else {
    tib_upcast(x)
  }
}

# ------------------------------------------------------------------------------

# Two data frames are considered identical by `rset_reconstructable()` if the rset
# sub-data-frames are identical. This means that if we select out the rset
# specific columns, they should be exactly the same (modulo reordering).

# It is expected that `to` is an rset object already, but `x` can be a
# bare data frame, or even a named list.

rset_reconstructable <- function(x, to) {
  x_names <- names(x)
  to_names <- names(to)

  x_rset_indicator <- col_equals_splits(x_names) | col_starts_with_id(x_names)
  to_rset_indicator <- col_equals_splits(to_names) | col_starts_with_id(to_names)

  # Special casing of `nested_cv` to also look for `inner_resamples`
  if (inherits(to, "nested_cv")) {
    x_rset_indicator <- x_rset_indicator | col_equals_inner_resamples(x_names)
    to_rset_indicator <- to_rset_indicator | col_equals_inner_resamples(to_names)
  }

  x_rset_names <- x_names[x_rset_indicator]
  to_rset_names <- to_names[to_rset_indicator]

  # Ignore ordering
  x_rset_names <- sort(x_rset_names)
  to_rset_names <- sort(to_rset_names)

  # Early return if names aren't identical
  if (!identical(x_rset_names, to_rset_names)) {
    return(FALSE)
  }

  # Avoid all non-bare-data-frame S3 dispatch and
  # don't compare outer data frame attributes.
  # Only look at column names and actual column data.
  x <- new_data_frame(x)
  to <- new_data_frame(to)

  # Early return if number of rows doesn't match
  if (!identical(vec_size(x), vec_size(to))) {
    return(FALSE)
  }

  x_rset_cols <- x[x_rset_names]
  to_rset_cols <- to[to_rset_names]

  # Row order doesn't matter
  x_rset_cols <- vec_sort(x_rset_cols)
  to_rset_cols <- vec_sort(to_rset_cols)

  # Check identical structures
  identical(x_rset_cols, to_rset_cols)
}

# ------------------------------------------------------------------------------

test_data <- function() {
  data.frame(
    x = 1:50,
    y = rep(c(1, 2), each = 25),
    index = as.Date(0:49, origin = "1970-01-01")
  )
}

# Keep this list up to date with known rset subclasses for testing.
# Delay assignment because we are creating this directly in the R script
# and not all of the required helpers might have been sourced yet.
delayedAssign("rset_subclasses", {
  list(
    bootstraps       = bootstraps(test_data()),
    vfold_cv         = vfold_cv(test_data(), v = 10, repeats = 2),
    group_vfold_cv   = group_vfold_cv(test_data(), y),
    loo_cv           = loo_cv(test_data()),
    mc_cv            = mc_cv(test_data()),
    nested_cv        = nested_cv(test_data(), outside = vfold_cv(v = 3), inside = bootstraps(times = 5)),
    validation_split = validation_split(test_data()),
    rolling_origin   = rolling_origin(test_data()),
    sliding_window   = sliding_window(test_data()),
    sliding_index    = sliding_index(test_data(), index),
    sliding_period   = sliding_period(test_data(), index, "week"),
    manual_rset      = manual_rset(bootstraps(test_data())$splits[1:2], c("ID1", "ID2")),
    apparent         = apparent(test_data())
  )
})

# ------------------------------------------------------------------------------

col_equals_splits <- function(x) {
  vec_equal(x, "splits")
}

col_starts_with_id <- function(x) {
  grepl("(^id$)|(^id[1-9]$)", x)
}

col_equals_inner_resamples <- function(x) {
  vec_equal(x, "inner_resamples")
}

# ------------------------------------------------------------------------------

# Maybe this should live in vctrs?
# Fallback to a tibble from the current data frame subclass.
# Removes subclass specific attributes and additional ones added by the user.
tib_upcast <- function(x) {
  size <- df_size(x)

  # Strip all attributes except names to construct
  # a bare list to build the tibble back up from.
  attributes(x) <- list(names = names(x))

  tibble::new_tibble(x, nrow = size)
}

df_size <- function(x) {
  if (!is.list(x)) {
    rlang::abort("Cannot get the df size of a non-list.")
  }

  if (length(x) == 0L) {
    return(0L)
  }

  col <- x[[1L]]

  vec_size(col)
}

# ------------------------------------------------------------------------------

# Maybe this should live in vctrs?
df_reconstruct <- function(x, to) {
  attrs <- attributes(to)
  attrs$names <- names(x)
  attrs$row.names <- .row_names_info(x, type = 0L)
  attributes(x) <- attrs
  x
}

#' Sampling for the Apparent Error Rate
#'
#' When building a model on a data set and re-predicting the same data, the
#'   performance estimate from those predictions is often called the
#'   "apparent" performance of the model. This estimate can be wildly
#'   optimistic. "Apparent sampling" here means that the analysis and
#'   assessment samples are the same. These resamples are sometimes used in
#'   the analysis of bootstrap samples and should otherwise be
#'   avoided like old sushi.
#'
#' @inheritParams vfold_cv
#' @return  A tibble with a single row and classes `apparent`,
#'   `rset`, `tbl_df`, `tbl`, and `data.frame`. The
#'   results include a column for the data split objects and one column
#'   called `id` that has a character string with the resample identifier.
#' @examples
#' apparent(mtcars)
#' @export
apparent <- function(data, ...) {
  splits <- rsplit(data, in_id = 1:nrow(data), out_id = 1:nrow(data))
  # splits <- rm_out(splits)
  class(splits) <- c("rsplit", "apparent_split")
  split_objs <- tibble::tibble(splits = list(splits), id = "Apparent")

  split_objs <-
    add_class(split_objs,
              cls = c("apparent", "rset"),
              at_end = FALSE)

  split_objs
}

#' @export
print.apparent <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("apparent", "rset"))]
  print(x, ...)
}

#' Nested or Double Resampling
#'
#' `nested_cv` can be used to take the results of one resampling procedure
#'   and conduct further resamples within each split. Any type of resampling
#'   used in `rsample` can be used.
#'
#' @details
#' It is a bad idea to use bootstrapping as the outer resampling procedure (see
#'   the example below)
#' @param data  A data frame.
#' @param outside The initial resampling specification. This can be an already
#'   created object or an expression of a new object (see the examples below).
#'   If the latter is used, the `data` argument does not need to be
#'   specified and, if it is given, will be ignored.
#' @param inside An expression for the type of resampling to be conducted
#'   within the initial procedure.
#' @return  An tibble with `nested_cv` class and any other classes that
#'   outer resampling process normally contains. The results include a
#'  column for the outer data split objects, one or more `id` columns,
#'  and a column of nested tibbles called `inner_resamples` with the
#'  additional resamples.
#' @examples
#' ## Using expressions for the resampling procedures:
#' nested_cv(mtcars, outside = vfold_cv(v = 3), inside = bootstraps(times = 5))
#'
#' ## Using an existing object:
#' folds <- vfold_cv(mtcars)
#' nested_cv(mtcars, folds, inside = bootstraps(times = 5))
#'
#' ## The dangers of outer bootstraps:
#' set.seed(2222)
#' bad_idea <- nested_cv(mtcars,
#'   outside = bootstraps(times = 5),
#'   inside = vfold_cv(v = 3)
#' )
#'
#' first_outer_split <- bad_idea$splits[[1]]
#' outer_analysis <- as.data.frame(first_outer_split)
#' sum(grepl("Volvo 142E", rownames(outer_analysis)))
#'
#' ## For the 3-fold CV used inside of each bootstrap, how are the replicated
#' ## `Volvo 142E` data partitioned?
#' first_inner_split <- bad_idea$inner_resamples[[1]]$splits[[1]]
#' inner_analysis <- as.data.frame(first_inner_split)
#' inner_assess <- as.data.frame(first_inner_split, data = "assessment")
#'
#' sum(grepl("Volvo 142E", rownames(inner_analysis)))
#' sum(grepl("Volvo 142E", rownames(inner_assess)))
#' @export
nested_cv <- function(data, outside, inside) {
  cl <- match.call()
  env <- rlang::caller_env()

  boot_msg <-
    paste0(
      "Using bootstrapping as the outer resample is dangerous ",
      "since the inner resample might have the same data ",
      "point in both the analysis and assessment set."
    )

  outer_cl <- cl[["outside"]]
  if (is_call(outer_cl)) {
    if (grepl("^bootstraps", deparse(outer_cl))) {
      warn(boot_msg)
    }
    outer_cl <- rlang::call_modify(outer_cl, data = data)
    outside <- eval(outer_cl, env)
  } else {
    if (inherits(outside, "bootstraps")) {
      warn(boot_msg)
    }
  }

  inner_cl <- cl[["inside"]]
  if (!is_call(inner_cl)) {
    abort(
      "`inside` should be a expression such as `vfold()` or ",
      "bootstraps(times = 10)` instead of an existing object.",
    )
  }
  inside <- map(outside$splits, inside_resample, cl = inner_cl, env = env)

  out <- dplyr::mutate(outside, inner_resamples = inside)
  out <- add_class(out, cls = "nested_cv")
  attr(out, "outside") <- cl$outside
  attr(out, "inside") <- cl$inside

  out
}

inside_resample <- function(src, cl, env) {
  cl <- rlang::call_modify(cl, data = as.data.frame(src))
  eval(cl, envir = env)
}

#' Determine the Assessment Samples
#'
#' This method and function help find which data belong in the analysis and
#' assessment sets.
#'
#' Given an `rsplit` object, `complement()` will determine which
#'   of the data rows are contained in the assessment set. To save space,
#'   many of the `rsplit` objects will not contain indices for the
#'   assessment split.
#'
#' @param x An `rsplit` object.
#' @param ... Not currently used.
#' @return A integer vector.
#' @seealso [populate()]
#' @examples
#' set.seed(28432)
#' fold_rs <- vfold_cv(mtcars)
#' head(fold_rs$splits[[1]]$in_id)
#' fold_rs$splits[[1]]$out_id
#' complement(fold_rs$splits[[1]])
#' @export
complement <- function(x, ...) {
  UseMethod("complement")
}

#' @export
#' @rdname complement
complement.rsplit <- function(x, ...) {
  check_dots_empty()
  if (!is_missing_out_id(x)) {
    return(x$out_id)
  } else {
    seq_len(nrow(x$data))[-unique(x$in_id)]
  }
}
#' @export
#' @rdname complement
complement.rof_split <- function(x, ...) {
  check_dots_empty()
  get_stored_out_id(x)
}
#' @export
#' @rdname complement
complement.sliding_window_split <- function(x, ...) {
  check_dots_empty()
  get_stored_out_id(x)
}
#' @export
#' @rdname complement
complement.sliding_index_split <- function(x, ...) {
  check_dots_empty()
  get_stored_out_id(x)
}
#' @export
#' @rdname complement
complement.sliding_period_split <- function(x, ...) {
  check_dots_empty()
  get_stored_out_id(x)
}

get_stored_out_id <- function(x) {
  out_id <- x$out_id

  if (length(out_id) == 0L) {
    return(out_id)
  }

  if (all(is.na(out_id))) {
    rlang::abort("Cannot derive the assessment set for this type of resampling.")
  }

  out_id
}

#' @export
#' @rdname complement
complement.apparent_split <- function(x, ...) {
  check_dots_empty()
  if (!is_missing_out_id(x)) {
    return(x$out_id)
  } else {
    seq_len(nrow(x$data))
  }
}

#' @export
complement.default <- function(x, ...) {
  cls <- paste0("'", class(x), "'", collapse = ", ")
  rlang::abort(
    paste("No `complement()` method for this class(es)", cls)
  )
}

# Get the indices of the analysis set from the assessment set
default_complement <- function(ind, n) {
  list(
    analysis = setdiff(1:n, ind),
    assessment = unique(ind)
  )
}


#' Add Assessment Indices
#'
#' Many `rsplit` and `rset` objects do not contain indicators for
#'   the assessment samples. `populate()` can be used to fill the slot
#'   for the appropriate indices.
#' @param x A `rsplit` and `rset` object.
#' @param ... Not currently used.
#' @return An object of the same kind with the integer indices.
#' @examples
#' set.seed(28432)
#' fold_rs <- vfold_cv(mtcars)
#'
#' fold_rs$splits[[1]]$out_id
#' complement(fold_rs$splits[[1]])
#'
#' populate(fold_rs$splits[[1]])$out_id
#'
#' fold_rs_all <- populate(fold_rs)
#' fold_rs_all$splits[[1]]$out_id
#' @export
populate <- function(x, ...) UseMethod("populate")

#' @export
populate.rsplit <- function(x, ...) {
  check_dots_empty()
  x$out_id <- complement(x)
  x
}

#' @export
populate.rset <- function(x, ...) {
  check_dots_empty()
  x$splits <- map(x$splits, populate)
  x
}

## This will remove the assessment indices from an rsplit object
rm_out <- function(x) {
  x$out_id <- NA
  x
}

is_missing_out_id <- function(x) {
  identical(x$out_id, NA)
}

#' Permutation sampling
#'
#' @description
#' A permutation sample is the same size as the original data set and is made
#'   by permuting/shuffling one or more columns. This results in analysis
#'   samples where some columns are in their original order and some columns
#'   are permuted to a random order. Unlike other sampling functions in
#'   rsample, there is no assessment set and calling [assessment()] on a
#'   permutation split will throw an error.
#'
#' @param data A data frame.
#' @param permute One or more columns to shuffle. This argument supports
#'   tidyselect selectors. Multiple expressions can be combined with `c()`.
#'   Variable names can be used as if they were positions in the data frame, so
#'   expressions like `x:y` can be used to select a range of variables.
#'   See \code{\link[tidyselect]{language}} for more details.
#' @param times The number of permutation samples.
#' @param apparent A logical. Should an extra resample be added where the
#'   analysis is the standard data set.
#' @inheritParams rlang::args_dots_empty
#'
#' @details The argument `apparent` enables the option of an additional
#'   "resample" where the analysis data set is the same as the original data
#'   set. Permutation-based resampling can be especially helpful for computing
#'   a statistic under the null hypothesis (e.g. t-statistic). This forms the
#'   basis of a permutation test, which computes a test statistic under all
#'   possible permutations of the data.
#'
#' @return A `tibble` with classes `permutations`, `rset`, `tbl_df`, `tbl`, and
#'   `data.frame`. The results include a column for the data split objects and a
#'   column called `id` that has a character string with the resample
#'   identifier.
#'
#' @examples
#' permutations(mtcars, mpg, times = 2)
#' permutations(mtcars, mpg, times = 2, apparent = TRUE)
#'
#' library(purrr)
#' resample1 <- permutations(mtcars, starts_with("c"), times = 1)
#' resample1$splits[[1]] |> analysis()
#'
#' resample2 <- permutations(mtcars, hp, times = 10, apparent = TRUE)
#' map_dbl(resample2$splits, function(x) {
#'   t.test(hp ~ vs, data = analysis(x))$statistic
#' })
#' @export
permutations <- function(
  data,
  permute = NULL,
  times = 25,
  apparent = FALSE,
  ...
) {
  check_dots_empty()

  permute <- rlang::enquo(permute)
  if (is.null(permute)) {
    cli_abort("You must specify at least one column to permute.")
  }
  col_id <- tidyselect::eval_select(permute, data)

  if (identical(length(col_id), 0L)) {
    cli_abort("You must specify at least one column to permute.")
  } else if (identical(length(col_id), ncol(data))) {
    cli_abort(c(
      "You have selected all columns to permute.",
      "i" = "This effectively reorders the rows in the original data without 
      changing the data structure.",
      ">" = "Please select fewer columns to permute."
    ))
  }

  split_objs <- perm_splits(data, times)

  if (apparent) {
    split_objs <- dplyr::bind_rows(split_objs, apparent(data))
  }

  split_objs$splits <-
    purrr::map(split_objs$splits, function(x) {
      x$col_id <- col_id
      x
    })

  perm_att <- list(
    times = times,
    apparent = apparent,
    col_id = col_id,
    permute = names(col_id)
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = perm_att,
    subclass = c("permutations", "rset")
  )
}

perm_complement <- function(ind, n) {
  list(analysis = ind, assessment = NA)
}

perm_splits <- function(data, times = 25) {
  n <- nrow(data)
  indices <- purrr::map(rep(n, times), sample, replace = FALSE)

  indices <- lapply(indices, perm_complement, n = n)

  split_objs <-
    purrr::map(indices, make_splits, data = data, class = "perm_split")

  list(
    splits = split_objs,
    id = names0(length(split_objs), "Permutations")
  )
}

#' Extract Predictor Names from Formula or Terms
#'
#' While [all.vars()] returns all variables used in a formula, this
#'  function only returns the variables explicitly used on the
#'  right-hand side (i.e., it will not resolve dots unless the
#'  object is terms with a data set specified).
#' @param object A model formula or [stats::terms()]
#'  object.
#' @param ... Arguments to pass to [all.vars()]
#' @return A character vector of names
#' @export
#' @examples
#' form_pred(y ~ x + z)
#' form_pred(terms(y ~ x + z))
#'
#' form_pred(y ~ x + log(z))
#' form_pred(log(y) ~ x + z)
#'
#' form_pred(y1 + y2 ~ x + z)
#' form_pred(log(y1) + y2 ~ x + z)
#'
#' # will fail:
#' # form_pred(y ~ .)
#'
#' form_pred(terms(mpg ~ (.)^2, data = mtcars))
#' form_pred(terms(~ (.)^2, data = mtcars))
#' @importFrom stats terms

form_pred <- function(object, ...) {
  if (inherits(object, "formula")) {
    object <- terms(object)
  }
  y_index <- attr(object, "response")

  ## If there is something on the lhs of the formula,
  ## remove it and get vars
  if (y_index != 0) {
    object[[2]] <- NULL
    object <- terms(object)
  }
  all.vars(object, ...)
}

#' Internal calibration split of the analysis set for fitting a post-processor
#'
#' @param x An `rsplit` object.
#' @param split_args A list of arguments to be used for the internal calibration
#' split.
#' @param ... Not currently used.
#' @return An `rsplit` object.
#' @details
#' `rsplit` objects live most commonly inside of an `rset` object. The
#' `split_args` argument can be the output of [.get_split_args()] on that
#' corresponding `rset` object, even if some of the arguments used to create the
#' `rset` object are not needed for the internal calibration split.
#' * For `mc_split` and `group_mc_split` objects, `internal_calibration_split()`
#' will ignore `split_args$times`.
#' * For `vfold_split` and `group_vfold_split` objects, it will ignore
#' `split_args$times` and `split_args$repeats`. `split_args$v` will be used to
#' set `split_args$prop` to `1 - 1/v` if `prop` is not already set and otherwise
#' ignored. The method for `group_vfold_split` will always use
#' `split_args$balance = NULL`.
#' * For `boot_split` and `group_boot_split` objects, it will ignore
#' `split_args$times`.
#' * For `val_split`, `group_val_split`, and `time_val_split` objects, it will
#' interpret a length-2 `split_args$prop` as a ratio between the training and
#' validation sets and split into inner analysis and calibration set in
#' the same ratio. If `split_args$prop` is a single value, it will be used as
#' the proportion of the inner analysis set.
#' * For `clustering_split` objects, it will ignore `split_args$repeats`.
#'
#' @keywords internal
#' @export
internal_calibration_split <- function(x, ...) {
  UseMethod("internal_calibration_split")
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.default <- function(x, ...) {
  cls <- class(x)
  cli::cli_abort(
    "No method for objects of class{?es}: {cls}."
  )
}

# mc ---------------------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.mc_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  split_args$times <- 1

  split_cal <- internal_calibration_split_core(
    analysis_set,
    split_function = mc_splits,
    split_args = split_args,
    classes = c("mc_split_cal", "internal_calibration_split", class(x))
  )

  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.group_mc_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  split_args$times <- 1

  split_cal <- internal_calibration_split_core(
    analysis_set,
    split_function = group_mc_splits,
    split_args = split_args,
    classes = c("group_mc_split_cal", "internal_calibration_split", class(x))
  )

  split_cal
}


# vfold ------------------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.vfold_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  # TODO should this be done outside of rsample,
  # in workflows or tune?
  if (is.null(split_args$prop)) {
    split_args$prop <- 1 - 1 / split_args$v
  }
  # use mc_splits for a random split
  split_args$times <- 1
  split_args$v <- NULL
  split_args$repeats <- NULL

  split_cal <- internal_calibration_split_core(
    analysis_set,
    split_function = mc_splits,
    split_args = split_args,
    classes = c("vfold_split_cal", "internal_calibration_split", class(x))
  )

  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.group_vfold_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  # TODO should this be done outside of rsample,
  # in workflows or tune?
  if (is.null(split_args$prop)) {
    split_args$prop <- 1 - 1 / split_args$v
  }

  # use group_mc_splits for a random split
  split_args$times <- 1
  split_args$v <- NULL
  split_args$repeats <- NULL
  split_args$balance <- NULL

  split_cal <- internal_calibration_split_core(
    analysis_set,
    split_function = group_mc_splits,
    split_args = split_args,
    classes = c(
      "group_vfold_split_cal",
      "internal_calibration_split",
      class(x)
    )
  )

  split_cal
}


# bootstrap --------------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.boot_split <- function(x, split_args, ...) {
  check_dots_empty()

  # use unique rows to prevent the same information from entering
  # both the inner analysis and inner assessment set
  id_outer_analysis <- unique(x$in_id)
  analysis_set <- x$data[id_outer_analysis, , drop = FALSE]

  split_args$times <- 1
  split_cal <- rlang::try_fetch(
    {
      split_cal <- rlang::inject(
        bootstraps(analysis_set, !!!split_args)
      )
      split_cal <- split_cal$splits[[1]]
    },
    rsample_bootstrap_empty_assessment = function(cnd) {
      return("mock_needed")
    },
    error = function(cnd) {
      return("mock_needed")
    }
  )

  if (identical(split_cal, "mock_needed")) {
    cli::cli_warn(
      "Cannot create calibration split; creating an empty calibration set."
    )
    # with a 0-row calibration set, we can't end up with rows in both
    # calibration and analysis, thus use the full analysis set with duplicate
    # rows here
    analysis_set <- analysis(x)
    split_cal <- internal_calibration_split_mock(analysis_set)
  }

  class_cal <- "boot_split_cal"
  class(split_cal) <- c(class_cal, "internal_calibration_split", class(x))
  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.group_boot_split <- function(x, split_args, ...) {
  check_dots_empty()

  # use unique rows to prevent the same information from entering
  # both the inner analysis and inner assessment set
  id_outer_analysis <- unique(x$in_id)
  analysis_set <- x$data[id_outer_analysis, , drop = FALSE]

  split_args$times <- 1
  split_cal <- rlang::try_fetch(
    {
      split_cal <- rlang::inject(
        group_bootstraps(analysis_set, !!!split_args)
      )
      split_cal <- split_cal$splits[[1]]
    },
    rsample_bootstrap_empty_assessment = function(cnd) {
      return("mock_needed")
    },
    error = function(cnd) {
      return("mock_needed")
    }
  )

  if (identical(split_cal, "mock_needed")) {
    cli::cli_warn(
      "Cannot create calibration split; creating an empty calibration set."
    )
    # with a 0-row calibration set, we can't end up with rows in both
    # calibration and analysis, thus use the full analysis set with duplicate
    # rows here
    analysis_set <- analysis(x)
    split_cal <- internal_calibration_split_mock(analysis_set)
  }

  class_cal <- "group_boot_split_cal"
  class(split_cal) <- c(class_cal, "internal_calibration_split", class(x))
  split_cal
}


# validation set ---------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.val_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  if (length(split_args$prop) == 2) {
    # keep ratio between training and validation as ratio between
    # inner analysis and inner assessment
    split_args$prop <- split_args$prop[[1]] / sum(split_args$prop)
  } else {
    split_args$prop <- split_args$prop[[1]]
  }
  split_args$times <- 1

  split_cal <- internal_calibration_split_core(
    analysis_set,
    split_function = mc_splits,
    split_args = split_args,
    classes = c("val_split_cal", "internal_calibration_split", class(x))
  )

  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.group_val_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  if (length(split_args$prop) == 2) {
    # keep ratio between training and validation as ratio between
    # inner analysis and inner assessment
    split_args$prop <- split_args$prop[[1]] / sum(split_args$prop)
  } else {
    split_args$prop <- split_args$prop[[1]]
  }
  split_args$times <- 1

  split_cal <- internal_calibration_split_core(
    analysis_set,
    split_function = group_mc_splits,
    split_args = split_args,
    classes = c("group_val_split_cal", "internal_calibration_split", class(x))
  )

  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.time_val_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  if (length(split_args$prop) == 2) {
    # keep ratio between training and validation as ratio between
    # inner analysis and inner assessment
    split_args$prop <- split_args$prop[[1]] / sum(split_args$prop)
  } else {
    split_args$prop <- split_args$prop[[1]]
  }

  split_cal <- internal_calibration_split_core(
    analysis_set,
    split_function = initial_time_split,
    split_args = split_args,
    classes = c("time_val_split_cal", "internal_calibration_split", class(x))
  )

  split_cal
}


# clustering -------------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.clustering_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  # TODO: reduce the number of clusters by 1 in tune?
  split_args$repeats <- 1

  split_cal <- internal_calibration_split_core(
    analysis_set,
    split_function = clustering_cv,
    split_args = split_args,
    classes = c(
      "clustering_split_cal",
      "internal_calibration_split",
      class(x)
    )
  )

  split_cal
}


# apparent ---------------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.apparent_split <- function(x, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  split_cal <- apparent(analysis_set)
  split_cal <- split_cal$splits[[1]]

  class_cal <- "apparent_split_cal"
  class(split_cal) <- c(class_cal, "internal_calibration_split", class(x))
  split_cal
}


# slide ------------------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.sliding_window_split <- function(
  x,
  split_args,
  ...
) {
  check_dots_empty()

  analysis_set <- analysis(x)

  if (nrow(analysis_set) < 2) {
    cli::cli_warn(
      "This set cannot be split into an analysis and a calibration set as there 
      is only one row; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(
      analysis_set,
      class = c(
        "sliding_window_split_cal",
        "internal_calibration_split",
        "sliding_window_split"
      )
    )
    return(split_cal)
  }

  split_args_inner <- translate_window_definition(
    split_args$lookback,
    split_args$assess_start,
    split_args$assess_stop
  )

  lookback <- split_args_inner$lookback
  assess_start <- split_args_inner$assess_start
  assess_stop <- split_args_inner$assess_stop

  seq <- vctrs::vec_seq_along(analysis_set)

  id_in <- slider::slide(
    .x = seq,
    .f = identity,
    .before = lookback,
    .after = 0L,
    .step = 1L,
    .complete = split_args$complete
  )

  id_out <- slider::slide(
    .x = seq,
    .f = identity,
    .before = -assess_start,
    .after = assess_stop,
    .step = 1L,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (length(indices) < 1) {
    cli::cli_warn(
      "Cannot create calibration split; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(analysis_set)
  } else {
    indices <- indices[[length(indices)]]
    split_cal <- make_splits(
      indices,
      data = analysis_set,
      class = "sliding_window_split"
    )
  }

  # no need to use skip and step args since they don't apply to _within_ an rsplit

  class_cal <- "sliding_window_split_cal"
  class(split_cal) <- c(class_cal, "internal_calibration_split", class(x))
  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.sliding_index_split <- function(x, split_args, ...) {
  check_dots_empty()

  analysis_set <- analysis(x)

  if (nrow(analysis_set) < 2) {
    cli::cli_warn(
      "This set cannot be split into an analysis and a calibration set as there 
      is only one row; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(
      analysis_set,
      class = c(
        "sliding_index_split_cal",
        "internal_calibration_split",
        "sliding_index_split"
      )
    )
    return(split_cal)
  }

  split_args_inner <- translate_window_definition(
    split_args$lookback,
    split_args$assess_start,
    split_args$assess_stop
  )

  lookback <- split_args_inner$lookback
  assess_start <- split_args_inner$assess_start
  assess_stop <- split_args_inner$assess_stop

  loc <- tidyselect::eval_select(split_args$index, analysis_set)
  index <- analysis_set[[loc]]

  seq <- vctrs::vec_seq_along(analysis_set)

  id_in <- slider::slide_index(
    .x = seq,
    .i = index,
    .f = identity,
    .before = lookback,
    .after = 0L,
    .complete = split_args$complete
  )

  id_out <- slider::slide_index(
    .x = seq,
    .i = index,
    .f = identity,
    .before = -assess_start,
    .after = assess_stop,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (length(indices) < 1) {
    cli::cli_warn(
      "Cannot create calibration split; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(analysis_set)
  } else {
    indices <- indices[[length(indices)]]
    split_cal <- make_splits(
      indices,
      data = analysis_set,
      class = "sliding_index_split"
    )
  }

  # no need to use skip and step args since they don't apply to _within_ an rsplit

  class_cal <- "sliding_index_split_cal"
  class(split_cal) <- c(class_cal, "internal_calibration_split", class(x))
  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.sliding_period_split <- function(
  x,
  split_args,
  ...
) {
  check_dots_empty()

  analysis_set <- analysis(x)

  if (nrow(analysis_set) < 2) {
    cli::cli_warn(
      "This set cannot be split into an analysis and a calibration set as there 
      is only one row; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(
      analysis_set,
      class = c(
        "sliding_period_split_cal",
        "internal_calibration_split",
        "sliding_period_split"
      )
    )
    return(split_cal)
  }
  if (split_args$lookback < 1) {
    cli::cli_warn(
      "Cannot create calibration split; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(
      analysis_set,
      class = c(
        "sliding_period_split_cal",
        "internal_calibration_split",
        "sliding_period_split"
      )
    )
    return(split_cal)
  }

  split_args_inner <- translate_window_definition(
    split_args$lookback,
    split_args$assess_start,
    split_args$assess_stop
  )

  lookback <- split_args_inner$lookback
  assess_start <- split_args_inner$assess_start
  assess_stop <- split_args_inner$assess_stop

  loc <- tidyselect::eval_select(split_args$index, analysis_set)
  index <- analysis_set[[loc]]

  seq <- vctrs::vec_seq_along(analysis_set)

  id_in <- slider::slide_period(
    .x = seq,
    .i = index,
    .period = split_args$period,
    .f = identity,
    .every = split_args$every,
    .origin = split_args$origin,
    .before = lookback,
    .after = 0L,
    .complete = split_args$complete
  )

  id_out <- slider::slide_period(
    .x = seq,
    .i = index,
    .period = split_args$period,
    .f = identity,
    .every = split_args$every,
    .origin = split_args$origin,
    .before = -assess_start,
    .after = assess_stop,
    .complete = TRUE
  )

  indices <- compute_complete_indices(id_in, id_out)

  if (length(indices) < 1) {
    cli::cli_warn(
      "Cannot create calibration split; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(analysis_set)
  } else {
    indices <- indices[[length(indices)]]
    split_cal <- make_splits(
      indices,
      data = analysis_set,
      class = "sliding_period_split"
    )
  }

  # no need to use skip and step args since they don't apply to _within_ an rsplit

  class_cal <- "sliding_period_split_cal"
  class(split_cal) <- c(class_cal, "internal_calibration_split", class(x))
  split_cal
}

translate_window_definition <- function(lookback, assess_start, assess_stop) {
  length_window <- lookback + 1 + assess_stop
  length_analysis <- lookback + 1

  prop_analysis <- length_analysis / length_window
  prop_assess <- (assess_stop - assess_start + 1) /
    length_window

  length_inner_analysis <- ceiling(prop_analysis * length_analysis)
  length_calibration <- ceiling(prop_assess * length_analysis)
  if (length_inner_analysis + length_calibration > length_analysis) {
    if (length_calibration > 1) {
      length_calibration <- length_calibration - 1
    } else {
      length_inner_analysis <- length_inner_analysis - 1
    }
  }

  lookback <- length_inner_analysis - 1
  assess_stop <- length_analysis - length_inner_analysis
  assess_start <- assess_stop - length_calibration + 1

  lookback <- check_lookback(lookback)
  assess_start <- check_assess(assess_start, "assess_start")
  assess_stop <- check_assess(assess_stop, "assess_stop")
  if (assess_start > assess_stop) {
    cli_abort(
      "{.arg assess_start} must be less than or equal to {.arg assess_stop}."
    )
  }

  list(
    lookback = lookback,
    assess_start = assess_start,
    assess_stop = assess_stop
  )
}


# initial split ----------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.initial_time_split <- function(x, split_args, ...) {
  check_dots_empty()

  training_set <- training(x)

  # to avoid deprecation warning if lag is just the default of 0
  if (-1 < split_args$lag && split_args$lag < 1) {
    split_args$lag <- NULL
  }

  split_cal <- internal_calibration_split_core(
    training_set,
    split_function = initial_time_split,
    split_args = split_args,
    classes = c(
      "initial_time_split_cal",
      "internal_calibration_split",
      class(x)
    )
  )

  split_cal
}


# initial validation split -----------------------------------------------

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.initial_validation_split <- function(
  x,
  split_args,
  ...
) {
  check_dots_empty()

  training_set <- training(x)

  split_args$prop <- split_args$prop[1] / sum(split_args$prop)
  split_args$times <- 1

  split_cal <- internal_calibration_split_core(
    training_set,
    split_function = mc_splits,
    split_args = split_args,
    classes = c(
      "initial_validation_split_cal",
      "internal_calibration_split",
      "mc_split",
      "rsplit"
    )
  )

  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.group_initial_validation_split <- function(
  x,
  split_args,
  ...
) {
  check_dots_empty()

  training_set <- training(x)

  split_args$prop <- split_args$prop[1] / sum(split_args$prop)
  split_args$times <- 1

  split_cal <- internal_calibration_split_core(
    training_set,
    split_function = group_mc_splits,
    split_args = split_args,
    classes = c(
      "group_initial_validation_split_cal",
      "internal_calibration_split",
      "group_mc_split",
      "mc_split",
      "rsplit"
    )
  )

  split_cal
}

#' @rdname internal_calibration_split
#' @export
internal_calibration_split.initial_validation_time_split <- function(
  x,
  split_args,
  ...
) {
  check_dots_empty()

  training_set <- training(x)

  if (nrow(training_set) < 2) {
    cli::cli_warn(
      "This set cannot be split into a training and a calibration set as there 
      is only one row; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(training_set)
    class_cal <- "initial_validation_time_split_cal"
    class(split_cal) <- c(
      class_cal,
      "internal_calibration_split",
      class(split_cal)
    )
    return(split_cal)
  }

  prop_analysis <- split_args$prop[1] / sum(split_args$prop)
  n_analysis <- floor(nrow(training_set) * prop_analysis)

  analysis_id <- seq(1, n_analysis, by = 1)
  cal_id <- seq(n_analysis + 1, nrow(training_set), by = 1)

  split_cal <- make_splits(
    list(analysis = as.integer(analysis_id), assessment = as.integer(cal_id)),
    data = training_set
  )

  class_cal <- "initial_validation_time_split_cal"
  class(split_cal) <- c(
    class_cal,
    "internal_calibration_split",
    class(split_cal)
  )
  split_cal
}


# helpers ----------------------------------------------------------------

internal_calibration_split_core <- function(
  analysis_set,
  split_function,
  split_args,
  classes
) {
  split_cal <- try(
    {
      split_cal <- rlang::inject(
        split_function(analysis_set, !!!split_args)
      )
      if (!inherits(split_cal, "rsplit")) {
        split_cal <- split_cal$splits[[1]]
      }
      split_cal
    },
    silent = TRUE
  )

  if (inherits(split_cal, "try-error")) {
    cli::cli_warn(
      "Cannot create calibration split; creating an empty calibration set."
    )
    split_cal <- internal_calibration_split_mock(analysis_set)
  }

  class(split_cal) <- classes

  split_cal
}

internal_calibration_split_mock <- function(analysis_set, class = NULL) {
  calibration_set <- analysis_set[0, , drop = FALSE]
  mock_split <- make_splits(analysis_set, calibration_set, class = class)
  mock_split
}

#   ----------------------------------------------------------------------

#' @rdname internal_calibration_split
#' @export
calibration <- function(x, ...) {
  UseMethod("calibration")
}

#' @rdname internal_calibration_split
#' @export
calibration.default <- function(x, ...) {
  cls <- class(x)
  cli::cli_abort(
    "No method for objects of class{?es}: {cls}"
  )
}

#' @rdname internal_calibration_split
#' @export
calibration.internal_calibration_split <- function(x, ...) {
  as.data.frame(x, data = "assessment", ...)
}

#' @rdname internal_calibration_split
#' @export
assessment.internal_calibration_split <- function(x, ...) {
  cli_abort(
    "Internal calibration splits are designed to only return analysis and calibration sets."
  )
}

#' @rdname internal_calibration_split
#' @export
print.internal_calibration_split <- function(x, ...) {
  out_char <-
    if (is_missing_out_id(x)) {
      paste(length(complement(x)))
    } else {
      paste(length(x$out_id))
    }

  cat("<Analysis/Calibration/Total>\n")
  cat("<", length(x$in_id), "/", out_char, "/", nrow(x$data), ">\n", sep = "")
}

#' Compatibility with dplyr
#'
#' @description
#' This page lays out the compatibility between rsample and dplyr. The `rset`
#' objects from rsample are a specific subclass of tibbles, hence standard
#' dplyr operations like joins as well row or column modifications work.
#' However, whether the operation returns an rset or a tibble depends on the
#' details of the operation.
#'
#' The overarching principle is that any operation which leaves the specific
#' characteristics of an rset intact will return an rset. If an operation
#' modifies any of the following characteristics, the result will be a `tibble`
#' rather than an `rset`:
#'
#' * Rows: The number of rows needs to remain unchanged to retain the rset
#' property. For example, you can't have a 10-fold CV object without 10 rows.
#' The order of the rows can be changed though and the object remains an rset.
#'
#' * Columns: The `splits` column and the `id` column(s) are required for an
#' rset and need to remain untouched. They cannot be dropped, renamed, or
#' modified if the result should remain an rset.
#'
#' ## Joins
#'
#' The following affect all of the dplyr joins, such as `left_join()`,
#' `right_join()`, `full_join()`, and `inner_join()`.
#'
#' The resulting object is an `rset` if the number of rows is unaffected.
#' Rows can be reordered but not added or removed, otherwise the resulting object
#' is a `tibble`.
#'
#' | operation          | same rows, possibly reordered | add or remove rows
#' | :----------------- | :---------------------------: | :---------------------:
#' | `join(rset, tbl)`  | `rset`                        | `tibble`
#'
#' ## Row Operations
#'
#' The resulting object is an `rset` if the number of rows is unaffected.
#' Rows can be reordered but not added or removed, otherwise the resulting object
#' is a `tibble`.
#'
#' | operation          | same rows, possibly reordered | add or remove rows
#' | :----------------- | :---------------------------: | :---------------------:
#' | `rset[ind,]`       | `rset`                        | `tibble`
#' | `slice(rset)`      | `rset`                        | `tibble`
#' | `filter(rset)`     | `rset`                        | `tibble`
#' | `arrange(rset)`    | `rset`                        | `tibble`
#'
#' ## Column Operations
#'
#' The resulting object is an `rset` if the required `splits` and `id` columns
#' remain unaltered. Otherwise the resulting object is a `tibble`.
#'
#' | operation          | required columns unaltered    | required columns removed, renamed, or modified
#' | :----------------- | :---------------------------: | :---------------------:
#' | `rset[,ind]`       | `rset`                        | `tibble`
#' | `select(rset)`     | `rset`                        | `tibble`
#' | `rename(rset)`     | `rset`                        | `tibble`
#' | `mutate(rset)`     | `rset`                        | `tibble`
#'
#' @name rsample-dplyr
NULL

# `dplyr_reconstruct()`
#
# `dplyr_reconstruct()` is called:
# - After a complex dplyr operation, like a `left_join()`, to restore to the
#   type of the first input, `x`.
# - At the end of a call to `dplyr_col_modify()`
# - At the end of a call to `dplyr_row_slice()`
# - See `?dplyr_reconstruct` for the full list.
#
# Because `dplyr_reconstruct()` is called at the end of `dplyr_col_modify()`
# and `dplyr_row_slice()`, we don't need methods for them. The default methods
# in dplyr do the right thing automatically, and then our reconstruction
# method decides whether or not the result should still be an rset.
#
# The implementation for rsample is the same as `vec_restore()`. Generally
# it will fall back to reconstructing a bare tibble, unless the rset structure
# is still completely intact. This happens when rset specific rows and columns
# (splits, id cols) are still exactly identical to how they were before the
# dplyr operation (with the exception of column reordering).

# Registered in `.onLoad()`
dplyr_reconstruct_rset <- function(data, template) {
  rset_reconstruct(data, template)
}

#' Rolling Origin Forecast Resampling
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' This resampling method is useful when the data set has a strong time
#'  component. The resamples are not random and contain data points that are
#'  consecutive values. The function assumes that the original data set are
#'  sorted in time order.
#'
#' This function is superseded by [sliding_window()], [sliding_index()], and
#' [sliding_period()] which provide more flexibility and control. Superseded
#' functions will not go away, but active development will be focused on the new
#' functions.
#'
#' @details The main options, `initial` and `assess`, control the number of
#'  data points from the original data that are in the analysis and assessment
#'  set, respectively. When `cumulative = TRUE`, the analysis set will grow as
#'  resampling continues while the assessment set size will always remain
#'  static.
#' `skip` enables the function to not use every data point in the resamples.
#'  When `skip = 0`, the resampling data sets will increment by one position.
#'  Suppose that the rows of a data set are consecutive days. Using `skip = 6`
#'  will make the analysis data set to operate on *weeks* instead of days. The
#'  assessment set size is not affected by this option.
#' @seealso
#' [sliding_window()], [sliding_index()], and [sliding_period()] for additional
#' time based resampling functions.
#' @inheritParams vfold_cv
#' @param initial The number of samples used for analysis/modeling in the
#'  initial resample.
#' @param assess The number of samples used for each assessment resample.
#' @param cumulative A logical. Should the analysis resample grow beyond the
#'  size specified by `initial` at each resample?.
#' @param skip A integer indicating how many (if any) _additional_ resamples
#'  to skip to thin the total amount of data points in the analysis resample.
#' See the example below.
#' @param lag A value to include a lag between the assessment
#'  and analysis set. This is useful if lagged predictors will be used
#'  during training and testing.
#' @export
#' @return An tibble with classes `rolling_origin`, `rset`, `tbl_df`, `tbl`,
#'  and `data.frame`. The results include a column for the data split objects
#'  and a column called `id` that has a character string with the resample
#'  identifier.
#' @examplesIf rlang::is_installed("modeldata")
#' set.seed(1131)
#' ex_data <- data.frame(row = 1:20, some_var = rnorm(20))
#' dim(rolling_origin(ex_data))
#' dim(rolling_origin(ex_data, skip = 2))
#' dim(rolling_origin(ex_data, skip = 2, cumulative = FALSE))
#'
#' # You can also roll over calendar periods by first nesting by that period,
#' # which is especially useful for irregular series where a fixed window
#' # is not useful. This example slides over 5 years at a time.
#' library(dplyr)
#' library(tidyr)
#' data(drinks, package = "modeldata")
#'
#' drinks_annual <- drinks |>
#'   mutate(year = as.POSIXlt(date)$year + 1900) |>
#'   nest(data = c(-year))
#'
#' multi_year_roll <- rolling_origin(drinks_annual, cumulative = FALSE)
#'
#' analysis(multi_year_roll$splits[[1]])
#' assessment(multi_year_roll$splits[[1]])
#' @export
rolling_origin <- function(
  data,
  initial = 5,
  assess = 1,
  cumulative = TRUE,
  skip = 0,
  lag = 0,
  ...
) {
  lifecycle::signal_stage(
    stage = "superseded",
    what = "rolling_origin()",
    with = I("`sliding_window()`, `sliding_index()` and `sliding_period()`")
  )

  check_dots_empty()

  n <- nrow(data)

  if (n < initial + assess) {
    rlang::abort(
      "There should be at least ",
      initial + assess,
      " rows in `data`."
    )
  }

  if (!is.numeric(lag) | !(lag %% 1 == 0)) {
    rlang::abort("`lag` must be a whole number.")
  }

  if (lag > initial) {
    rlang::abort(
      "`lag` must be less than or equal to the number of training observations."
    )
  }

  stops <- seq(initial, (n - assess), by = skip + 1)
  starts <- if (!cumulative) {
    stops - initial + 1
  } else {
    starts <- rep(1, length(stops))
  }

  in_ind <- mapply(seq, starts, stops, SIMPLIFY = FALSE)
  out_ind <-
    mapply(seq, stops + 1 - lag, stops + assess, SIMPLIFY = FALSE)
  indices <- mapply(merge_lists, in_ind, out_ind, SIMPLIFY = FALSE)
  split_objs <-
    purrr::map(indices, make_splits, data = data, class = "rof_split")
  split_objs <- list(
    splits = split_objs,
    id = names0(length(split_objs), "Slice")
  )

  roll_att <- list(
    initial = initial,
    assess = assess,
    cumulative = cumulative,
    skip = skip,
    lag = lag
  )

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = roll_att,
    subclass = c("rolling_origin", "rset")
  )
}

#' Sampling for the Apparent Error Rate
#'
#' When building a model on a data set and re-predicting the same data, the
#'   performance estimate from those predictions is often called the
#'   "apparent" performance of the model. This estimate can be wildly
#'   optimistic. "Apparent sampling" here means that the analysis and
#'   assessment samples are the same. These resamples are sometimes used in
#'   the analysis of bootstrap samples and should otherwise be
#'   avoided like old sushi.
#'
#' @inheritParams vfold_cv
#' @return  A tibble with a single row and classes `apparent`,
#'   `rset`, `tbl_df`, `tbl`, and `data.frame`. The
#'   results include a column for the data split objects and one column
#'   called `id` that has a character string with the resample identifier.
#' @examples
#' apparent(mtcars)
#' @export
apparent <- function(data, ...) {
  check_dots_empty()

  splits <- rsplit(
    data,
    in_id = seq_len(nrow(data)),
    out_id = seq_len(nrow(data))
  )

  class(splits) <- c("apparent_split", "rsplit")
  split_objs <- tibble::tibble(splits = list(splits), id = "Apparent")

  new_rset(
    splits = split_objs$splits,
    ids = split_objs$id,
    attrib = NULL,
    subclass = c("apparent", "rset")
  )
}

#' Tidy Resampling Object
#'
#' The `tidy()` function from the \pkg{broom} package can be used on `rset` and
#'  `rsplit` objects to generate tibbles with which rows are in the analysis and
#'  assessment sets.
#' @param x A `rset` or `rsplit` object
#' @param unique_ind Should unique row identifiers be returned? For example,
#'  if `FALSE` then bootstrapping results will include multiple rows in the
#'  sample for the same row in the original data.
#' @inheritParams rlang::args_dots_empty
#' @return A tibble with columns `Row` and `Data`. The latter has possible
#'   values "Analysis" or "Assessment". For `rset` inputs, identification
#'   columns are also returned but their names and values depend on the type of
#'   resampling. For [vfold_cv()], contains a column "Fold" and, if repeats are
#'   used, another called "Repeats". [bootstraps()] and [mc_cv()] use the column
#'   "Resample".
#' @details Note that for nested resampling, the rows of the inner resample,
#'  named `inner_Row`, are *relative* row indices and do not correspond to the
#'  rows in the original data set.
#' @examplesIf rlang::is_installed("ggplot2")
#' library(ggplot2)
#' theme_set(theme_bw())
#'
#' set.seed(4121)
#' cv <- tidy(vfold_cv(mtcars, v = 5))
#' ggplot(cv, aes(x = Fold, y = Row, fill = Data)) +
#'   geom_tile() +
#'   scale_fill_brewer()
#'
#' set.seed(4121)
#' rcv <- tidy(vfold_cv(mtcars, v = 5, repeats = 2))
#' ggplot(rcv, aes(x = Fold, y = Row, fill = Data)) +
#'   geom_tile() +
#'   facet_wrap(~Repeat) +
#'   scale_fill_brewer()
#'
#' set.seed(4121)
#' mccv <- tidy(mc_cv(mtcars, times = 5))
#' ggplot(mccv, aes(x = Resample, y = Row, fill = Data)) +
#'   geom_tile() +
#'   scale_fill_brewer()
#'
#' set.seed(4121)
#' bt <- tidy(bootstraps(mtcars, time = 5))
#' ggplot(bt, aes(x = Resample, y = Row, fill = Data)) +
#'   geom_tile() +
#'   scale_fill_brewer()
#'
#' dat <- data.frame(day = 1:30)
#' # Resample by week instead of day
#' ts_cv <- rolling_origin(dat,
#'   initial = 7, assess = 7,
#'   skip = 6, cumulative = FALSE
#' )
#' ts_cv <- tidy(ts_cv)
#' ggplot(ts_cv, aes(x = Resample, y = factor(Row), fill = Data)) +
#'   geom_tile() +
#'   scale_fill_brewer()
#' @export
tidy.rsplit <- function(x, unique_ind = TRUE, ...) {
  check_dots_empty()
  if (unique_ind) {
    x$in_id <- unique(x$in_id)
  }
  out <- tibble(
    Row = c(x$in_id, complement(x)),
    Data = rep(
      c("Analysis", "Assessment"),
      c(length(x$in_id), length(complement(x)))
    )
  )
  out <- dplyr::arrange(.data = out, Data, Row)
  out
}

#' @rdname tidy.rsplit
#' @export
tidy.rset <- function(x, unique_ind = TRUE, ...) {
  check_dots_empty()
  stacked <- purrr::map(x$splits, tidy, unique_ind = unique_ind)
  for (i in seq_along(stacked)) {
    stacked[[i]]$Resample <- x$id[i]
  }
  stacked <- dplyr::bind_rows(stacked)
  stacked <- dplyr::arrange(.data = stacked, Data, Row)
  stacked
}

#' @rdname tidy.rsplit
#' @export
tidy.vfold_cv <- function(x, ...) {
  check_dots_empty()
  stacked <- purrr::map(x$splits, tidy)
  for (i in seq_along(stacked)) {
    if (attr(x, "repeats") > 1) {
      stacked[[i]]$Repeat <- x$id[i]
      stacked[[i]]$Fold <- x$id2[i]
    } else {
      stacked[[i]]$Fold <- x$id[i]
    }
  }
  stacked <- dplyr::bind_rows(stacked)
  stacked <- dplyr::arrange(.data = stacked, Data, Row)
  stacked
}

#' @rdname tidy.rsplit
#' @export
tidy.nested_cv <- function(x, unique_ind = TRUE, ...) {
  check_dots_empty()
  x$inner_tidy <- purrr::map(
    x$inner_resamples,
    tidy_wrap,
    unique_ind = unique_ind
  )
  inner_tidy <- tidyr::unnest(x, inner_tidy)
  class(x) <- class(x)[class(x) != "nested_cv"]
  outer_tidy <- tidy(x, unique_ind = unique_ind)
  id_cols <- names(outer_tidy)
  id_cols <- id_cols[!(id_cols %in% c("Row", "Data"))]

  inner_id <- grep("^id", names(inner_tidy))
  if (length(inner_id) != length(id_cols)) {
    cli_abort("Cannot merge tidy data sets.")
  }
  names(inner_tidy)[inner_id] <- id_cols
  full_join(outer_tidy, inner_tidy, by = id_cols)
}

tidy_wrap <- function(x, unique_ind = TRUE) {
  x <- tidy(x, unique_ind = unique_ind)
  names(x) <- paste0("inner_", names(x))
  x
}

## The `pretty()` methods below are good for when you need to
## textually describe the resampling procedure. Note that they
## can have more than one element (in the case of nesting)

#' @export
pretty.vfold_cv <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(details$v, "-fold cross-validation")
  if (details$repeats > 1) {
    res <- paste(res, "repeated", details$repeats, "times")
  }
  if (has_strata(details)) {
    res <- paste(res, "using stratification")
  }
  res
}

#' @export
pretty.loo_cv <- function(x, ...) {
  "Leave-one-out cross-validation"
}

#' @export
pretty.apparent <- function(x, ...) {
  "Apparent sampling"
}

#' @export
pretty.rolling_origin <- function(x, ...) {
  "Rolling origin forecast resampling"
}

#' @export
pretty.sliding_window <- function(x, ...) {
  "Sliding window resampling"
}

#' @export
pretty.sliding_index <- function(x, ...) {
  "Sliding index resampling"
}

#' @export
pretty.sliding_period <- function(x, ...) {
  "Sliding period resampling"
}

#' @export
pretty.mc_cv <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Monte Carlo cross-validation (",
    signif(details$prop, 2),
    "/",
    signif(1 - details$prop, 2),
    ") with ",
    details$times,
    " resamples"
  )
  if (has_strata(details)) {
    res <- paste(res, " using stratification")
  }
  res
}

#' @export
pretty.validation_set <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Validation Set (",
    signif(details$prop[1] / sum(details$prop), 2),
    "/",
    signif(details$prop[2] / sum(details$prop), 2),
    ")"
  )
  res
}

#' @export
pretty.validation_split <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Validation Set Split (",
    signif(details$prop, 2),
    "/",
    signif(1 - details$prop, 2),
    ") "
  )
  if (has_strata(details)) {
    res <- paste(res, "using stratification")
  }
  res
}

#' @export
pretty.group_validation_split <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Group Validation Set Split (",
    signif(details$prop, 2),
    "/",
    signif(1 - details$prop, 2),
    ") "
  )
  if (has_strata(details)) {
    res <- paste0(res, "using stratification")
  }
  res
}

#' @export
pretty.nested_cv <- function(x, ...) {
  details <- attributes(x)

  if (is_call(details$outside)) {
    class(x) <- class(x)[!(class(x) == "nested_cv")]
    outer_label <- pretty(x)
  } else {
    outer_label <- paste0("`", deparse(details$outside), "`")
  }

  inner_label <- if (is_call(details$inside)) {
    pretty(x$inner_resamples[[1]])
  } else {
    paste0("`", deparse(details$inside), "`")
  }

  res <- c(
    "Nested resampling:",
    paste(" outer:", outer_label),
    paste(" inner:", inner_label)
  )
  res
}

#' @export
pretty.bootstraps <- function(x, ...) {
  details <- attributes(x)
  res <- "Bootstrap sampling"
  if (has_strata(details)) {
    res <- paste(res, "using stratification")
  }
  if (details$apparent) {
    res <- paste(res, "with apparent sample")
  }
  res
}

#' @export
pretty.group_bootstraps <- function(x, ...) {
  details <- attributes(x)
  res <- "Group bootstrap sampling"
  if (has_strata(details)) {
    res <- paste(res, "using stratification")
  }
  if (details$apparent) {
    res <- paste(res, "with apparent sample")
  }
  res
}


#' @export
pretty.permutations <- function(x, ...) {
  details <- attributes(x)
  res <- "Permutation sampling"
  if (details$apparent) {
    res <- paste(res, "with apparent sample")
  }
  res
}

#' @export
pretty.group_vfold_cv <- function(x, ...) {
  details <- attributes(x)
  paste0("Group ", details$v, "-fold cross-validation")
}

#' @export
pretty.group_mc_cv <- function(x, ...) {
  details <- attributes(x)
  res <- paste0(
    "Group Monte Carlo cross-validation (",
    signif(details$prop, 2),
    "/",
    signif(1 - details$prop, 2),
    ") with ",
    details$times,
    " resamples "
  )

  if (has_strata(details)) {
    res <- paste0(res, "using stratification")
  }

  res
}

#' @export
pretty.manual_rset <- function(x, ...) {
  "Manual resampling"
}

#' @export
pretty.clustering_cv <- function(x, ...) {
  details <- attributes(x)
  paste0(details$v, "-cluster cross-validation")
}

# The print methods below control the display of rset objects

#' @export
print.apparent <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("apparent", "rset"))]
  print(x, ...)
}

#' @export
print.bootstraps <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("bootstraps", "rset"))]
  print(x, ...)
}

#' @export
print.group_bootstraps <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[
    !(class(x) %in% c("group_bootstraps", "bootstraps", "group_rset", "rset"))
  ]
  print(x, ...)
}

#' @export
print.group_vfold_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[
    !(class(x) %in% c("group_vfold_cv", "vfold_cv", "group_rset", "rset"))
  ]
  print(x, ...)
}

#' @export
print.initial_split <- function(x, ...) {
  out_char <-
    if (is_missing_out_id(x)) {
      paste(length(complement(x)))
    } else {
      paste(length(x$out_id))
    }

  cat("<Training/Testing/Total>\n")
  cat("<", length(x$in_id), "/", out_char, "/", nrow(x$data), ">\n", sep = "")
}

#' @export
print.loo_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("loo_cv", "rset"))]
  print(x, ...)
}

#' @export
print.manual_rset <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("manual_rset", "rset"))]
  print(x, ...)
}

#' @export
print.mc_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("mc_cv", "rset"))]
  print(x, ...)
}

#' @export
print.nested_cv <- function(x, ...) {
  char_x <- paste("#", pretty(x))
  cat(char_x, sep = "\n")
  class(x) <- class(tibble())
  print(x, ...)
}


#' @export
print.permutations <- function(x, ...) {
  shuffled_cols <- paste(names(attr(x, "col_id")), collapse = ", ")
  cat("#", pretty(x), "\n")
  cat("# Permuted columns: [", shuffled_cols, "] \n", sep = "")
  class(x) <- class(x)[!(class(x) %in% c("permutations", "rset"))]
  print(x, ...)
}

#' @export
print.rolling_origin <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("rolling_origin", "rset"))]
  print(x, ...)
}

#' @export
print.sliding_window <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("sliding_window", "rset"))]
  print(x, ...)
}

#' @export
print.sliding_index <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("sliding_index", "rset"))]
  print(x, ...)
}

#' @export
print.sliding_period <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("sliding_period", "rset"))]
  print(x, ...)
}

#' @export
print.validation_split <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("validation_split", "rset"))]
  print(x, ...)
}

#' @export
print.group_validation_split <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[
    !(class(x) %in%
      c("group_validation_split", "validation_split", "group_rset", "rset"))
  ]
  print(x, ...)
}


#' @export
print.val_split <- function(x, ...) {
  if (is_missing_out_id(x)) {
    out_char <- paste(length(complement(x)))
  } else {
    out_char <- paste(length(x$out_id))
  }

  cat("<Training/Validation/Total>\n")
  cat("<", length(x$in_id), "/", out_char, "/", nrow(x$data), ">\n", sep = "")
}

#' @export
print.group_mc_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[
    !(class(x) %in% c("group_mc_cv", "group_rset", "mc_cv", "rset"))
  ]
  print(x, ...)
}

#' @export
print.vfold_cv <- function(x, ...) {
  cat("# ", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("vfold_cv", "rset"))]
  print(x, ...)
}

#' @export
print.clustering_cv <- function(x, ...) {
  cat("#", pretty(x), "\n")
  class(x) <- class(x)[!(class(x) %in% c("clustering_cv", "rset"))]
  print(x, ...)
}

has_strata <- function(x) {
  !is.null(x$strata) && !identical(x$strata, FALSE)
}

#' A convenience function for confidence intervals with linear-ish parametric models
#'
#' @param formula An R model formula with one outcome and at least one predictor.
#' @param data A data frame.
#' @param model_fn The model to fit. Allowable values are `"lm"`, `"glm"`,
#'  `"survreg"`, and `"coxph"`. The latter two require that the survival package
#'  be installed.
#' @param type The type of bootstrap confidence interval. Values of `"student-t"` and
#' `"percentile"` are allowed.
#' @param times A single integer for the number of bootstrap samples. If left
#' `NULL`, 1,001 are used for t-intervals and 2,001 for percentile intervals.
#' @param alpha Level of significance.
#' @param filter A logical expression used to remove rows from the final result, or `NULL` to keep all rows.
#' @param keep_reps Should the individual parameter estimates for each bootstrap
#' sample be retained?
#' @param ... Options to pass to the model function (such as `family` for [stats::glm()]).
#' @return A tibble with columns "term", ".lower", ".estimate", ".upper",
#' ".alpha", and ".method". If `keep_reps = TRUE`, an additional list column
#' called ".replicates" is also returned.
#' @export
#' @seealso [int_pctl()], [int_t()]
#' @references
#' Davison, A., & Hinkley, D. (1997). _Bootstrap Methods and their
#'  Application_. Cambridge: Cambridge University Press.
#'  doi:10.1017/CBO9780511802843
#'
#' _Bootstrap Confidence Intervals_,
#' \url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}
#' @examplesIf rlang::is_installed("broom")
#' \donttest{
#' set.seed(1)
#' reg_intervals(mpg ~ I(1 / sqrt(disp)), data = mtcars)
#'
#' set.seed(1)
#' reg_intervals(mpg ~ I(1 / sqrt(disp)), data = mtcars, keep_reps = TRUE)
#' }
reg_intervals <-
  function(
    formula,
    data,
    model_fn = "lm",
    type = "student-t",
    times = NULL,
    alpha = 0.05,
    filter = term != "(Intercept)",
    keep_reps = FALSE,
    ...
  ) {
    rlang::check_installed("broom")

    model_fn <- rlang::arg_match(model_fn, c("lm", "glm", "survreg", "coxph"))
    type <- rlang::arg_match(type, c("student-t", "percentile"))

    filter <- rlang::enexpr(filter)

    if (is.null(times)) {
      if (type == "student-t") {
        times <- 1001
      } else {
        times <- 2001
      }
    } else {
      times <- times[1]
      if (!is.numeric(times)) {
        cli_abort("{.arg times} should be a single integer.")
      }
    }

    if (length(alpha) != 1 || !is.numeric(alpha)) {
      cli_abort("{.arg alpha} must be a single numeric value.")
    }

    if (model_fn %in% c("survreg", "coxph")) {
      pkg <- "survival"
      rlang::check_installed("survival")
    } else {
      pkg <- NULL
    }
    fn_call <- rlang::call2(
      model_fn,
      formula = formula,
      data = rlang::expr(data),
      .ns = pkg,
      ...
    )

    bt <- rsample::bootstraps(
      data,
      times = times,
      apparent = type %in% c("student-t")
    )
    bt <-
      dplyr::mutate(
        bt,
        models = purrr::map(
          splits,
          \(.x) model_results(rsample::analysis(.x), fn_call, filter)
        )
      )
    if (type == "student-t") {
      res <- int_t(bt, models, alpha = alpha)
    } else {
      res <- int_pctl(bt, models, alpha = alpha)
    }

    if (keep_reps) {
      bt <- bt[bt$id != "Apparent", ]
      reps <- purrr::map(bt$models, I) |> list_rbind()
      reps <- dplyr::group_nest(reps, term, .key = ".replicates")
      res <- dplyr::full_join(res, reps, by = "term")
    }

    res
  }

# TODO add handler for survival models to catch warnings? That seems to be the
# only way to know about convergence.
model_results <- function(data, cl, flt) {
  mod <- broom::tidy(rlang::eval_tidy(cl, data))
  mod <- mod[, c("term", "estimate", "std.error")]
  if (is.language(flt)) {
    mod <- dplyr::filter(mod, !!flt)
  }
  mod
}

#' Convert Resampling Objects to Other Formats
#'
#' These functions can convert resampling objects between
#'  \pkg{rsample} and \pkg{caret}.
#'
#' @param object An `rset` object. Currently,
#'  [nested_cv()] is not supported.
#' @return `rsample2caret()` returns a list that mimics the
#'  `index` and `indexOut` elements of a
#'  `trainControl` object. `caret2rsample()` returns an
#'  `rset` object of the appropriate class.
#' @export
rsample2caret <- function(object, data = c("analysis", "assessment")) {
  if (!inherits(object, "rset")) {
    cli_abort("{.arg object} must be an {.cls rset}.")
  }
  data <- rlang::arg_match(data)
  in_ind <- purrr::map(object$splits, as.integer, data = "analysis")
  names(in_ind) <- labels(object)
  out_ind <- purrr::map(object$splits, as.integer, data = "assessment")
  names(out_ind) <- names(in_ind)
  list(index = in_ind, indexOut = out_ind)
}

#' @rdname rsample2caret
#' @param ctrl An object produced by `caret::trainControl()` that has
#'  had the `index` and `indexOut` elements populated by
#'  integers. One method of getting this is to extract the
#'  `control` objects from an object produced by `train`.
#' @param data The data that was originally used to produce the
#'  `ctrl` object.
#' @export
caret2rsample <- function(ctrl, data = NULL) {
  if (is.null(data)) {
    cli_abort("Must supply original data")
  }
  if (!any(names(ctrl) == "index")) {
    cli_abort("{.arg ctrl} should have an element {.val index}")
  }
  if (!any(names(ctrl) == "indexOut")) {
    cli_abort("{.arg ctrl} should have an element {.val indexOut}")
  }
  if (is.null(ctrl$index)) {
    cli_abort("{.val ctrl$index} should be populated with integers")
  }
  if (is.null(ctrl$indexOut)) {
    cli_abort("{.val ctrl$indexOut} should be populated with integers")
  }

  indices <- purrr::map2(ctrl$index, ctrl$indexOut, extract_int)
  id_data <- names(indices)
  indices <- unname(indices)
  indices <- purrr::map(indices, add_data, y = data)
  indices <-
    map(indices, add_rsplit_class, cl = map_rsplit_method(ctrl$method))
  indices <- tibble::tibble(splits = indices)
  if (ctrl$method %in% c("repeatedcv", "adaptive_cv")) {
    id_data <- strsplit(id_data, split = ".", fixed = TRUE)
    id_data <- tibble::tibble(
      id = vapply(
        id_data,
        function(x) {
          x[2]
        },
        character(1)
      ),
      id2 = vapply(
        id_data,
        function(x) {
          x[1]
        },
        character(1)
      )
    )
  } else {
    id_data <- tibble::tibble(id = id_data)
  }

  new_rset(
    splits = indices$splits,
    ids = id_data[, grepl("^id", names(id_data))],
    attrib = map_attr(ctrl),
    subclass = c(map_rset_method(ctrl$method), "rset")
  )
}

extract_int <- function(x, y) {
  list(in_id = x, out_id = y)
}

add_data <- function(x, y) {
  c(list(data = y), x)
}

add_rsplit_class <- function(x, cl) {
  class(x) <- c("rsplit", cl)
  x
}

add_rset_class <- function(x, cl) {
  class(x) <- c(cl, "rset", "tbl_df", "tbl", "data.frame")
  x
}

map_rsplit_method <- function(method) {
  out <- switch(
    method,
    cv = ,
    repeatedcv = ,
    adaptive_cv = "vfold_split",
    boot = ,
    boot_all = ,
    boot632 = ,
    optimism_boot = ,
    adaptive_boot = "boot_split",
    LOOCV = "loo_split",
    LGOCV = ,
    adaptive_LGOCV = "mc_split",
    timeSlice = "rof_split",
    "error"
  )
  if (out == "error") {
    cli_abort(
      "Resampling method {.arg method} cannot be converted into an {.cls rsplit} object"
    )
  }
  out
}

map_rset_method <- function(method) {
  out <- switch(
    method,
    cv = ,
    repeatedcv = ,
    adaptive_cv = "vfold_cv",
    boot = ,
    boot_all = ,
    boot632 = ,
    optimism_boot = ,
    adaptive_boot = "bootstraps",
    LOOCV = "loo_cv",
    LGOCV = ,
    adaptive_LGOCV = "mc_cv",
    timeSlice = "rolling_origin",
    "error"
  )
  if (out == "error") {
    cli_abort(
      "Resampling method {.arg method} cannot be converted into an {.cls rset} object"
    )
  }
  out
}


map_attr <- function(object) {
  if (grepl("cv$", object$method)) {
    out <- list(
      v = object$number,
      repeats = ifelse(!is.na(object$repeats), object$repeats, 1),
      strata = TRUE
    )
  } else if (grepl("boot", object$method)) {
    out <- list(
      times = object$number,
      apparent = FALSE,
      strata = FALSE
    )
  } else if (grepl("LGOCV$", object$method)) {
    out <- list(
      times = object$number,
      prop = object$p,
      strata = FALSE
    )
  } else if (object$method == "LOOCV") {
    out <- list()
  } else if (object$method == "timeSlice") {
    out <- list(
      initial = object$initialWindow,
      assess = object$horizon,
      cumulative = !object$fixedWindow,
      skip = object$skip
    )
  } else {
    cli_abort("Method {.arg object$method} cannot be converted")
  }
  out
}

#' Extending rsample with new rset subclasses
#'
#' `rset_reconstruct()` encapsulates the logic for allowing new rset
#' subclasses to work properly with vctrs (through `vctrs::vec_restore()`) and
#' dplyr (through `dplyr::dplyr_reconstruct()`). It is intended to be a
#' developer tool, and is not required for normal usage of rsample.
#'
#' rset objects are considered "reconstructable" after a vctrs/dplyr operation
#' if:
#'
#' - `x` and `to` both have an identical column named `"splits"` (column
#'   and row order do not matter).
#'
#' - `x` and `to` both have identical columns prefixed with `"id"` (column
#'   and row order do not matter).
#'
#' @param x A data frame to restore to an rset subclass.
#' @param to An rset subclass to restore to.
#'
#' @return `x` restored to the rset subclass of `to`.
#'
#' @export
#' @examples
#' to <- bootstraps(mtcars, times = 25)
#'
#' # Imitate a vctrs/dplyr operation,
#' # where the class might be lost along the way
#' x <- tibble::as_tibble(to)
#'
#' # Say we added a new column to `x`. Here we mock a `mutate()`.
#' x$foo <- "bar"
#'
#' # This is still reconstructable to `to`
#' rset_reconstruct(x, to)
#'
#' # Say we lose the first row
#' x <- x[-1, ]
#'
#' # This is no longer reconstructable to `to`, as `x` is no longer an rset
#' # bootstraps object with 25 bootstraps if one is lost!
#' rset_reconstruct(x, to)
rset_reconstruct <- function(x, to) {
  if (rset_reconstructable(x, to)) {
    df_reconstruct(x, to)
  } else {
    tib_upcast(x)
  }
}

# ------------------------------------------------------------------------------

# Two data frames are considered identical by `rset_reconstructable()` if the rset
# sub-data-frames are identical. This means that if we select out the rset
# specific columns, they should be exactly the same (modulo reordering).

# It is expected that `to` is an rset object already, but `x` can be a
# bare data frame, or even a named list.

rset_reconstructable <- function(x, to) {
  x_names <- names(x)
  to_names <- names(to)

  x_rset_indicator <- col_equals_splits(x_names) | col_starts_with_id(x_names)
  to_rset_indicator <- col_equals_splits(to_names) |
    col_starts_with_id(to_names)

  # Special casing of `nested_cv` to also look for `inner_resamples`
  if (inherits(to, "nested_cv")) {
    x_rset_indicator <- x_rset_indicator | col_equals_inner_resamples(x_names)
    to_rset_indicator <- to_rset_indicator |
      col_equals_inner_resamples(to_names)
  }

  x_rset_names <- x_names[x_rset_indicator]
  to_rset_names <- to_names[to_rset_indicator]

  # Ignore ordering
  x_rset_names <- sort(x_rset_names)
  to_rset_names <- sort(to_rset_names)

  # Early return if names aren't identical
  if (!identical(x_rset_names, to_rset_names)) {
    return(FALSE)
  }

  # Avoid all non-bare-data-frame S3 dispatch and
  # don't compare outer data frame attributes.
  # Only look at column names and actual column data.
  x <- new_data_frame(x)
  to <- new_data_frame(to)

  # Early return if number of rows doesn't match
  if (!identical(vec_size(x), vec_size(to))) {
    return(FALSE)
  }

  x_rset_cols <- x[x_rset_names]
  to_rset_cols <- to[to_rset_names]

  # Row order doesn't matter
  x_rset_cols <- vec_sort(x_rset_cols)
  to_rset_cols <- vec_sort(to_rset_cols)

  # Check identical structures
  identical(x_rset_cols, to_rset_cols)
}

# ------------------------------------------------------------------------------

test_data <- function() {
  data.frame(
    x = 1:50,
    y = rep(seq.int(10), each = 5),
    index = as.Date(0:49, origin = "1970-01-01")
  )
}

# Keep this list up to date with known rset subclasses for testing.
# Delay assignment because we are creating this directly in the R script
# and not all of the required helpers might have been sourced yet.
delayedAssign("rset_subclasses", {
  if (rlang::is_installed("withr")) {
    withr::with_seed(
      123,
      list(
        bootstraps = bootstraps(test_data()),
        group_bootstraps = group_bootstraps(test_data(), y),
        vfold_cv = vfold_cv(test_data(), v = 10, repeats = 2),
        group_vfold_cv = group_vfold_cv(test_data(), y),
        loo_cv = loo_cv(test_data()),
        mc_cv = mc_cv(test_data()),
        group_mc_cv = group_mc_cv(test_data(), y),
        nested_cv = nested_cv(
          test_data(),
          outside = vfold_cv(v = 3),
          inside = bootstraps(times = 5)
        ),
        validation_split = suppressWarnings(validation_split(test_data())),
        validation_time_split = suppressWarnings(validation_time_split(test_data())),
        group_validation_split = suppressWarnings(group_validation_split(
          test_data(),
          y
        )),
        rolling_origin = rolling_origin(test_data()),
        sliding_window = sliding_window(test_data()),
        sliding_index = sliding_index(test_data(), index),
        sliding_period = sliding_period(test_data(), index, "week"),
        manual_rset = manual_rset(
          list(
            initial_time_split(test_data()),
            initial_time_split(test_data())
          ),
          c("ID1", "ID2")
        ),
        apparent = apparent(test_data()),
        permutations = permutations(test_data(), y),
        clustering_cv = clustering_cv(test_data(), y, repeats = 2),
        validation_set = validation_set(initial_validation_split(test_data()))
      )
    )
  } else {
    NULL
  }
})

# ------------------------------------------------------------------------------

col_equals_splits <- function(x) {
  vec_equal(x, "splits")
}

col_starts_with_id <- function(x) {
  grepl("(^id$)|(^id[1-9]$)", x)
}

col_equals_inner_resamples <- function(x) {
  vec_equal(x, "inner_resamples")
}

# ------------------------------------------------------------------------------

# Maybe this should live in vctrs?
# Fallback to a tibble from the current data frame subclass.
# Removes subclass specific attributes and additional ones added by the user.
tib_upcast <- function(x) {
  size <- df_size(x)

  # Strip all attributes except names to construct
  # a bare list to build the tibble back up from.
  attributes(x) <- list(names = names(x))

  tibble::new_tibble(x, nrow = size)
}

df_size <- function(x) {
  if (!is.list(x)) {
    rlang::abort("Cannot get the df size of a non-list.")
  }

  if (length(x) == 0L) {
    return(0L)
  }

  col <- x[[1L]]

  vec_size(col)
}

# ------------------------------------------------------------------------------

# Maybe this should live in vctrs?
df_reconstruct <- function(x, to) {
  attrs <- attributes(to)
  attrs$names <- names(x)
  attrs$row.names <- .row_names_info(x, type = 0L)
  attributes(x) <- attrs
  x
}

#' Nested or Double Resampling
#'
#' `nested_cv()` can be used to take the results of one resampling procedure
#'   and conduct further resamples within each split. Any type of resampling
#'   used in rsample can be used.
#'
#' @details
#' It is a bad idea to use bootstrapping as the outer resampling procedure (see
#'   the example below)
#' @param data  A data frame.
#' @param outside The initial resampling specification. This can be an already
#'   created object or an expression of a new object (see the examples below).
#'   If the latter is used, the `data` argument does not need to be
#'   specified and, if it is given, will be ignored.
#' @param inside An expression for the type of resampling to be conducted
#'   within the initial procedure.
#' @return  An tibble with `nested_cv` class and any other classes that
#'   outer resampling process normally contains. The results include a
#'  column for the outer data split objects, one or more `id` columns,
#'  and a column of nested tibbles called `inner_resamples` with the
#'  additional resamples.
#' @examples
#' ## Using expressions for the resampling procedures:
#' nested_cv(mtcars, outside = vfold_cv(v = 3), inside = bootstraps(times = 5))
#'
#' ## Using an existing object:
#' folds <- vfold_cv(mtcars)
#' nested_cv(mtcars, folds, inside = bootstraps(times = 5))
#'
#' ## The dangers of outer bootstraps:
#' set.seed(2222)
#' bad_idea <- nested_cv(mtcars,
#'   outside = bootstraps(times = 5),
#'   inside = vfold_cv(v = 3)
#' )
#'
#' first_outer_split <- get_rsplit(bad_idea, 1)
#' outer_analysis <- analysis(first_outer_split)
#' sum(grepl("Camaro Z28", rownames(outer_analysis)))
#'
#' ## For the 3-fold CV used inside of each bootstrap, how are the replicated
#' ## `Camaro Z28` data partitioned?
#' first_inner_split <- get_rsplit(bad_idea$inner_resamples[[1]], 1)
#' inner_analysis <- analysis(first_inner_split)
#' inner_assess <- assessment(first_inner_split)
#'
#' sum(grepl("Camaro Z28", rownames(inner_analysis)))
#' sum(grepl("Camaro Z28", rownames(inner_assess)))
#' @export
nested_cv <- function(data, outside, inside) {
  cl <- match.call()
  env <- rlang::caller_env()

  boot_msg <-
    paste0(
      "Using bootstrapping as the outer resample is dangerous ",
      "since the inner resample might have the same data ",
      "point in both the analysis and assessment set."
    )

  outer_cl <- cl[["outside"]]
  if (is_call(outer_cl)) {
    using_bootstraps <- grepl(
      "^bootstraps",
      paste(
        deparse(outer_cl, width.cutoff = 500L),
        collapse = " "
      )
    )
    if (using_bootstraps) {
      warn(boot_msg)
    }
    outer_cl <- rlang::call_modify(outer_cl, data = data)
    outside <- eval(outer_cl, env)
  } else {
    if (inherits(outside, "bootstraps")) {
      warn(boot_msg)
    }
  }

  inner_cl <- cl[["inside"]]
  if (!is_call(inner_cl)) {
    cli_abort(
      "{.arg inside} should be a expression such as {.code vfold()} or
      {.code bootstraps(times = 10)} instead of an existing object."
    )
  }
  inside <- map(outside$splits, inside_resample, cl = inner_cl, env = env)

  out <- dplyr::mutate(outside, inner_resamples = inside)
  out <- add_class(out, cls = "nested_cv")
  attr(out, "outside") <- cl$outside
  attr(out, "inside") <- cl$inside

  out
}

inside_resample <- function(src, cl, env) {
  cl <- rlang::call_modify(cl, data = as.data.frame(src))
  eval(cl, envir = env)
}

---
title: "Bootstrap confidence intervals"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Bootstrap confidence intervals}
output:
  knitr:::html_vignette:
    toc: yes
---


```{r setup, include=FALSE}
knitr::opts_chunk$set(
  eval = rlang::is_installed("ggplot2") && rlang::is_installed("modeldata")
)
```

```{r package_setup, include = FALSE}
library(tidymodels)
library(nlstools)
library(GGally)
theme_set(theme_bw())
```


The bootstrap was originally intended for estimating confidence intervals for complex statistics whose variance properties are difficult to analytically derive. Davison and Hinkley's [_Bootstrap Methods and Their Application_](https://www.cambridge.org/core/books/bootstrap-methods-and-their-application/ED2FD043579F27952363566DC09CBD6A) is a great resource for these methods. rsample contains a few functions to compute the most common types of intervals. 

## A nonlinear regression example

To demonstrate the computations for the different types of intervals, we'll use a nonlinear regression example from [Baty _et al_ (2015)](https://www.jstatsoft.org/article/view/v066i05). They showed data that monitored oxygen uptake in a patient with rest and exercise phases (in the data frame `O2K`). 

```{r O2K-dat, fig.alt = "A scatterplot with time on the x-axis and oxygen uptake on the y-axis. For time < 6, the oxygen uptake remains at around 400. Between time = 6 and time = 9, it rises to about 1600, then remains around that value."}
library(tidymodels)
library(nlstools)
library(GGally)

data(O2K)

ggplot(O2K, aes(x = t, y = VO2)) + 
  geom_point()
```

The authors fit a segmented regression model where the transition point was known (this is the time when exercise commenced). Their model was:

```{r O2K-fit}
nonlin_form <-  
  as.formula(
    VO2 ~ (t <= 5.883) * VO2rest + 
      (t > 5.883) * 
      (VO2rest + (VO2peak - VO2rest) * (1 - exp(-(t - 5.883) / mu)))
    )

# Starting values from visual inspection
start_vals <- list(VO2rest = 400, VO2peak = 1600, mu = 1)

res <- nls(nonlin_form, start = start_vals, data = O2K) 

tidy(res)
```


`broom::tidy()` returns our analysis object in a standardized way. The column names shown here are used for most types of objects and this allows us to use the results more easily. 

For rsample, we'll rely on the `tidy()` method to work with bootstrap estimates when we need confidence intervals. There's an example at the end of a univariate statistic that isn't automatically formatted with `tidy()`. 

To run our model over different bootstraps, we'll write a function that uses the `split` object as input and produces a tidy data frame:

```{r model-info}
# Will be used to fit the models to different bootstrap data sets:
fit_fun <- function(split, ...) {
  # We could check for convergence, make new parameters, etc.
  nls(nonlin_form, data = analysis(split), ...) |>
    tidy()
}
```

First, let's create a set of resamples and fit separate models to each. The options `apparent = TRUE` will be set. This creates a final resample that is a copy of the original (unsampled) data set. This is required for some of the interval methods. 

```{r resample}
set.seed(462)
nlin_bt <-
  bootstraps(O2K, times = 2000, apparent = TRUE) |>
  mutate(models = map(splits, \(.x) fit_fun(.x, start = start_vals)))
nlin_bt

nlin_bt$models[[1]]
```

Let's look at the data and see if there any outliers or aberrant results:

```{r extract}
library(tidyr)
nls_coef <- 
  nlin_bt |>
  dplyr::select(-splits) |>
  # Turn it into a tibble by stacking the `models` col
  unnest(cols = models) |>
  # Get rid of unneeded columns
  dplyr::select(id, term, estimate) 

head(nls_coef)
```
Now let's create a scatterplot matrix:

```{r splom, fig.alt = "The 3x3 scatter plot matrix for the variables VO2rest, VO2peak, and mu. We can see one value for VO2peak that is larger than the rest."} 
nls_coef |>
  # Put different parameters in columns
  tidyr::pivot_wider(names_from = term, values_from = estimate) |> 
  # Keep only numeric columns
  dplyr::select(-id) |> 
  ggscatmat(alpha = .25)
```

One potential outlier on the right for `VO2peak` but we'll leave it in. 

The univariate distributions are:

```{r hists, fig.alt = "Three histograms for mu, VO2peak, and VO2rest: all distributions are unimodal and symmetric."}
nls_coef |> 
  ggplot(aes(x = estimate)) + 
  geom_histogram(bins = 20, col = "white") + 
  facet_wrap(~ term, scales = "free_x")
```

### Percentile intervals

The most basic type of interval uses _percentiles_ of the resampling distribution. To get the percentile intervals, the `rset` object is passed as the first argument and the second argument is the list column of tidy results: 

```{r pctl}
p_ints <- int_pctl(nlin_bt, models)
p_ints
```

When overlaid with the univariate distributions: 

```{r pctl-plot, fig.alt = "The three histograms for mu, VO2peak, and VO2rest with added vertical lines for the estimated lower and upper bounds of the percentile intervals."}
nls_coef |> 
  ggplot(aes(x = estimate)) + 
  geom_histogram(bins = 20, col = "white") + 
  facet_wrap(~ term, scales = "free_x") + 
  geom_vline(data = p_ints, aes(xintercept = .lower), col = "red") + 
  geom_vline(data = p_ints, aes(xintercept = .upper), col = "red")
```

How do these intervals compare to the parametric asymptotic values?

```{r int-compare}
parametric <- 
  tidy(res, conf.int = TRUE) |> 
  dplyr::select(
    term,
    .lower = conf.low,
    .estimate = estimate,
    .upper = conf.high
  ) |> 
  mutate(
    .alpha = 0.05,
    .method = "parametric"
  )

intervals <- 
  bind_rows(parametric, p_ints) |> 
  arrange(term, .method)
intervals |> split(intervals$term)
```

The percentile intervals are wider than the parametric intervals (which assume asymptotic normality). Do the estimates appear to be normally distributed? We can look at quantile-quantile plots:  

```{r qqplot, fig.alt = "Three QQ plots for mu, VO2peak, and VO2rest which show that the empirical distributions deviate from a normal distribution in the tails."}
nls_coef |> 
  ggplot(aes(sample = estimate)) + 
  stat_qq() +
  stat_qq_line(alpha = .25) + 
  facet_wrap(~ term, scales = "free") 
```

### t-intervals

Bootstrap _t_-intervals are estimated by computing intermediate statistics that are _t_-like in structure. To use these, we require the estimated variance _for each individual resampled estimate_. In our example, this comes along with the fitted model object. We can extract the standard errors of the parameters. Luckily, most `tidy()` methods provide this in a column named `std.error`. 

The arguments for these intervals are the same:

```{r t-ints}
t_stats <- int_t(nlin_bt, models)
intervals <- 
  bind_rows(intervals, t_stats) |> 
  arrange(term, .method)
intervals |> split(intervals$term)
```


### Bias-corrected and accelerated intervals

For bias-corrected and accelerated (BCa) intervals, an additional argument is required. The `.fn` argument is a function that computes the statistic of interest. The first argument should be for the `rsplit` object and other arguments can be passed in using the ellipses. 

These intervals use an internal leave-one-out resample to compute the Jackknife statistic and will recompute the statistic for _every bootstrap resample_. If the statistic is expensive to compute, this may take some time. For those calculations, we use the furrr package so these can be computed in parallel if you have set up a parallel processing plan (see `?future::plan`). 

The user-facing function takes an argument for the function and the ellipses. 

```{r bca-comp}
bias_corr <- int_bca(nlin_bt, models, .fn = fit_fun, start = start_vals)
intervals <- 
  bind_rows(intervals, bias_corr) |> 
  arrange(term, .method)
intervals |> split(intervals$term)
```


## No existing tidy method

In this case, your function can emulate the minimum results: 

 * a character column called `term`, 
 * a numeric column called `estimate`, and, optionally, 
 * a numeric column called `std.error`. 

The last column is only needed for `int_t()`. 

Suppose we just want to estimate the fold-increase in the outcome between the 90th and 10th percentiles over the course of the experiment. Our function might look like:

```{r fold-foo}
fold_incr <- function(split, ...) {
  dat <- analysis(split)
  quants <- quantile(dat$VO2, probs = c(.1, .9))
  tibble(
    term = "fold increase",
    estimate = unname(quants[2]/quants[1]),
    # We don't know the analytical formula for this 
    std.error = NA_real_
  )
}
```

Everything else works the same as before:

```{r fold-ci}
nlin_bt <-
  nlin_bt |>
  mutate(folds = map(splits, fold_incr))

int_pctl(nlin_bt, folds)
int_bca(nlin_bt, folds, .fn = fold_incr)
```

## Intervals for linear(ish) parametric intervals

rsample also contains the `reg_intervals()` function that can be used for linear regression (via `lm()`), generalized linear models (`glm()`), or log-linear survival models (`survival::survreg()` or `survival::coxph()`). This function makes it easier to get intervals for these models. 

A simple example is a logistic regression using the dementia data from the `modeldata` package: 

```{r ad-data}
data(ad_data, package = "modeldata")
```

Let's fit a model with a few predictors: 

```{r ad-model}
lr_mod <- glm(Class ~ male + age + Ab_42 + tau, data = ad_data,
              family = binomial)
glance(lr_mod)
tidy(lr_mod)
```

Let's use this model with student-t intervals: 

```{r ad-t-int}
set.seed(29832)
lr_int <- 
  reg_intervals(Class ~ male + age + Ab_42 + tau, 
                data = ad_data,
                model_fn = "glm",
                family = binomial)
lr_int
```

We can also save the resamples for plotting: 

```{r ad-t-int-plot}
set.seed(29832)
lr_int <- 
  reg_intervals(Class ~ male + age + Ab_42 + tau, 
                data = ad_data,
                keep_reps = TRUE,
                model_fn = "glm",
                family = binomial)
lr_int
```

Now we can unnest the data to use in a ggplot: 

```{r ad-plot, fig.alt = "Four histograms for the distribution of the regression model coefficients, with vertical lines indicating the estimated lower and upper interval bounds, as well as vertical lines for a coefficient value of 0. For age, the interval encloses 0, for the other variables it does not."}
lr_int |> 
  select(term, .replicates) |> 
  unnest(cols = .replicates) |> 
  ggplot(aes(x = estimate)) + 
  geom_histogram(bins = 30) + 
  facet_wrap(~ term, scales = "free_x") + 
  geom_vline(data = lr_int, aes(xintercept = .lower), col = "red") + 
  geom_vline(data = lr_int, aes(xintercept = .upper), col = "red") + 
  geom_vline(xintercept = 0, col = "green")
```

---
title: "Calibration sets"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{calibration-sets}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(rsample)
```


## Context

We are adding post-processing to tidymodels. One possible post-processing operation is calibration. 

Calibration is a transformation of the predictions from our primary model, based on a separate calibration model. That calibration model is used to model the relationship between the predictions based on the primary model and the true outcomes.

Crucial: To avoid overfitting, we cannot use the same data to fit our calibration model as we use to assess the combination of primary and calibration model. Using the same data to fit the primary model and the calibration model means the predictions used to fit the calibration model are re-predictions of the same observations used to fit the primary model. Hence they are closer to the true values than predictions on new data would be and the calibration model doesn't have accurate information to estimate the right trends (so that they then can be removed).

rsample provides a collection of functions to make resamples for empirical validation of prediction models. So far, the assumption was that the prediction model is the only model that needs fitting, i.e., a resample consists of an analysis set and an assessment set.

If we include calibration into our workflow (bundeling preprocessing, (primary) model, and post-processing), we want an analysis set, a calibration set, and an assessment set.

## General principles

First principle: we leave the assessment set purely for performance assessment and rather treat the analysis set as "the data to fit" - except now we split it to fit two different models instead of just one. Or in other words, we take the data for the calibration set from the analysis set, not from the assessment set. 

If you compare a model with calibration to one without, and you use the same resamples, you are also using the same assessment sets.

"Taking data from the analysis set" means splitting up the analysis set to end up with ... an analysis set and a calibration set. Now we have two sets called analysis set, that's confusing. If we need to distinguish them, we'll refer to them as "outer" and "inner" analysis set for "before" and "after" the split for a calibration set.


```{r}
#| label: calibration
#| echo: false
#| out.height: 350
#| fig.align: "center"
knitr::include_graphics("images/analysis-calibration-assessment.jpg")
```

It is important to note that including calibration requires more data than fitting a workflow without calibration. For small datasets, you might run into situations where fitting either model (primary model or calibration model) is not particularly stable - or you might not even be able to construct a calibration set properly. In this case, our policy is to avoid erroring. We treat this situation as if no calibration model was used.

Second principle: we try to mimick the data splitting process used to create the (outer) analysis set and the assessment set.

Let's look at v-fold cross-validation as an example.


### Example: v-fold cross-validation

For v-fold cross-validation, we split the data into `v` folds and create `v` resamples by always using one of the folds as the assessment set while combining the other folds into the analysis set. 

Since we want to have a calibration split similar in spirit, we are splitting the (outer) analysis set into (inner) analysis set and calibration set with the same proportion of observations going into the analysis set: 1 - 1/v.

Note: the `internal_calibration_split()` methods are exported to be used by tune but not intended to be user-facing. 

```{r}
#| label: vfold

set.seed(11)
r_set <- vfold_cv(warpbreaks, v = 5)
r_split <- get_rsplit(r_set, 1)
r_split
# proportion of observations allocated for fitting the model in the outer split
nrow(analysis(r_split)) / nrow(warpbreaks)

split_cal <- internal_calibration_split(r_split, .get_split_args(r_set))
split_cal
# similar proportion of observations allocated for fitting the model in the inner split
nrow(analysis(split_cal)) / nrow(analysis(r_split))
```


## Splits for ordered observations

rsample's functions for creating resamples of ordered observations are the `sliding_*()` family: `sliding_window()`, `sliding_index()`, and `sliding_period()`.

- `sliding_window()` operates on the ordered rows, allowing us to specify how far to look back to create an analysis set, how far to look ahead to create an assessment set and then sliding that across the data to create a set of resamples.
- `sliding_index()` works much the same with the key difference being that what we are sliding over aren't the rows directly but rather an index. This is useful when dealing with irregular series of data: while the number of rows which fall into a window can vary, the window length is the same across all resamples.
- `sliding_period()` takes the idea of the `sliding_index()` function one step further and aggregates the index into periods before sliding across. This is useful for aggregating, e.g., daily data into months and then defining the analysis and assessment set in terms of months rather than the daily index directly.

### Row-based splitting

Let's start with the row-based splitting done by `sliding_window()`. We'll use a very small example dataset. This will make it easier to illustrate how the different subsets of the data are created but note that it is too small for real-world purposes. Let's use a data frame with 11 rows and say we want to use 5 for the analysis set, 3 for the assessment set, and leave a gap of 2 in between those two sets. We can make two such resamples from our data frame.

![](images/calibration-split-window.jpg)

Or, in code:
```{r}
#| label: sliding_window_rset

df <- data.frame(x = 1:11)
r_set <- sliding_window(df, lookback = 4, assess_start = 3, assess_stop = 5)
r_set
```

Let's take that first resample and split it into an (inner) analysis set and a calibration set. 

```{r}
#| label: sliding_window_rsplit

r_split <- get_rsplit(r_set, 1)
analysis(r_split)
```

To mimick the split into (outer) analysis and assessment set, we'll calculate the proportion of the analysis set and apply it the observations available for fitting both the primary model and the calibration model, creating our (inner) analysis set. Since the remaining observations could be for the calibration model or a gap, we also calculate the proportion of the assessment set and use it to construct the calibration set. 

```{r}
#| label: sliding_window_props

lookback <- 4
assess_start <- 3
assess_stop <- 5

# absolute sizes in the  outer split
length_window <- lookback + 1 + assess_stop
length_analysis_outer <- lookback + 1
length_assess <- assess_stop - assess_start + 1

# relative sizes
prop_analysis <- length_analysis_outer / length_window
prop_assess <- length_assess / length_window

# absolute sizes in the inner split
length_analysis_inner <- ceiling(prop_analysis * length_analysis_outer)
length_calibration <- ceiling(prop_assess * length_analysis_outer)

c(length_analysis_inner, length_calibration)
```

Calculating the length of the calibration set rather than the gap, together with rounding up when translating proportions to new lengths within the outer analysis set means that we prioritize allocating observations to the (inner) analysis and calibration set over allocating them to the gap. In this example here, this means that we are not leaving a gap between the analysis set and the calibration set.

However, rounding up for both (inner) analysis and calibration set when we don't have a gap could mean we end up allocating more observations than we actually have. So in that case, we try to take from the calibration set if possible and thus prioritzing fitting the prediction model over the calibration model.

Keep in mind though that this example dataset is pathalogically small for illustration purposes. Of these details, the prioritization of analysis and calibration set over any gap in between them is most likely to be relevant. But even this should not be massively influential for amounts of data typically required for calibration.

Note that we cannot create a calibration split if `lookback = 0`, as this means that we only have a single row to work with.

```{r}
#| label: sliding_window_adjustments

if (length_analysis_inner + length_calibration > length_analysis_outer) {
  if (length_calibration > 1) {
    length_calibration <- length_calibration - 1
  } else {
    length_analysis_inner <- length_analysis_inner - 1
  }
}
```

From the `length_*`s, we can calculate the new `lookback`, `assess_start`, and `assess_stop` to be applied to the outer analysis set.

![](images/calibration-split-window-inner.jpg)

```{r}
#| label: sliding_window_internal_calibration_split

split_cal <- internal_calibration_split(r_split, .get_split_args(r_set))
split_cal
analysis(split_cal)
calibration(split_cal)
```

This is the basic principle. 

Apart from the arguments to define the look forwards and backwards, `sliding_window()`has a few more options: `step`, `skip`, `complete`. `step` and `skip` apply on the level of resamples, so we are not applying them to the (inner) split.  With `complete = FALSE`, `sliding_window()` allows you to create incomplete analysis sets (but not assessment sets).

![](images/incomplete-window.jpg)

Or, in code:

```{r}
#| label: sliding_window_incomplete

df <- data.frame(x = 1:11)
r_set <- sliding_window(df, lookback = 4, assess_start = 3, assess_stop = 5, complete = FALSE)
r_set
```

When creating the calibration split, we apply the `complete` option to the (now inner) analysis set and ensure that the calibration set is complete. In this example, it means that, until we have at least three rows (two for the complete calibration set, and at least one for the analysis set), we can't do a calibration split according to the rules laid out above. This is the case for the second resample, where we only have two rows to work with in the outer analysis set. Instead of erroring here, we fall back to not using calibration at all and return the (outer) analysis set as the (inner) analsysis set, along with an empty calibration set.

```{r}
#| label: sliding_window_split_incomplete_2

r_split <- get_rsplit(r_set, 2) 
split_cal <- internal_calibration_split(r_split, .get_split_args(r_set))

split_cal
```

A calibration split on the third resample, which contains three rows, succeeds with the required two rows in the calibration set and the remaining one in the incomplete analysis set.

```{r}
#| label: sliding_window_split_incomplete_3

r_split <- get_rsplit(r_set, 3) 
split_cal <- internal_calibration_split(r_split, .get_split_args(r_set))
split_cal
```

If we can fit a complete (inner) analysis set in the available rows, we do so, even if incomplete sets are allowed.

```{r}
#| label: sliding_window_split_incomplete_5

r_split <- get_rsplit(r_set, 5)
split_cal <- internal_calibration_split(r_split, .get_split_args(r_set))
split_cal
```


### Index-based splitting

When working with irregular series, it can be helpful to define resamples based on sliding over an index, rather than the rows directly. That index often is a form of time but it does not have to be. Let's take our previous data frame and use `x` as the index, with a small modification. The index doesn't go straight from 1 to 11 but rather is missing an observation at the index value of 2. We use the same  `lookback`, `assess_start`, and `assess_stop`. The difference is that now these define the resamples on the _index_ rather than the _rows_ directly.

![](images/calibration-split-index.jpg)

We still get two resamples, however, the analysis set contains only 4 rows because only those fall into the window defined by the index.

```{r}
#| label: sliding_index_rset

df <- data.frame(x = c(1, 3:11))
r_set <- sliding_index(df, index = x, lookback = 4, assess_start = 3, assess_stop = 5)
r_set
```

For the first resample, this is all well. We just skip over the missing row, both when constructing the split into analysis and assessment set and the split into analysis and calibration set.

```{r}
#| label: sliding_index_rsplit_1

r_split <- get_rsplit(r_set, 1)
analysis(r_split)

split_cal <- internal_calibration_split(r_split, .get_split_args(r_set))
split_cal
analysis(split_cal)
calibration(split_cal)
```

#### When we can't split

Abstracting away from observed rows to a theoretical index does pose two challenges for us though.

For the second resample, the theoretical window for the (outer) analysis set is 2-6. Because it fits within the boundaries of the observed rows (which is 1 to 11), we construct this set, even though we don't observe anything at an index of 2. 

Based on `lookback`, `assess_start`, and `assess_stop`, we calculate the same allocation for the (inner) analysis set and the calibration set as in the previous example working on the rows directly with `sliding_window()`: three for the inner analysis set and two for the calibration set. However, the dataset we want to split only covers the range [3, 6].

```{r}
#| label: sliding_index_rsplit_2

r_split <- get_rsplit(r_set, 2)
analysis(r_split)
```

The sliding splits slide over _the data_, meaning they slide over observed values of the index and they slide only within the boundaries of the observed index values. So here, we can only slide within [3, 6] and thus cannot fit an inner analysis set of three and a calibration set of two into it.

So while we do not observe the lower boundary of our theoretical [2, 6] window, we always observed the upper boundary. That is because "sliding over the observed instances" also means that we will skip over potential splits if we do not observe the value relative to which a split is defined.

In our example so far, we've been able to generate two resamples on an index running from 1 to 11. The first resample is defined relative to an index of 5, the second one on 6. If there is no observed index of 6, the windows can't be defined and there is no corresponding split and resample.

![](images/calibration-split-index-6.jpg)

Or, in code:

```{r}
#| label: sliding_index_rset_anchor

# no index 6
df <- data.frame(x = c(1:5, 7:11))
r_set <- sliding_index(df, index = x, lookback = 4, assess_start = 3, assess_stop = 5)
r_set
```

The one resample we get is relative to an index value of 5, the last observation in the analysis set.

```{r}
#| label: sliding_index_rsplit_anchor

r_split <- get_rsplit(r_set, 1)
analysis(r_split)
```

This principle of sliding across _observations_ makes it impossible to construct a calibration split if we happen to define it relative to an index value that we do not observe. Let's modify our example so that now we don't observe anything at index 3.

![](images/calibration-split-index-3.jpg)

This is the index value relative to which we can define an inner analysis set of 3 and a calibration set of 2. Thus we cannot actually make that split and fall back on an empty calibration set.

```{r}
#| label: sliding_index_split_anchor

# missing 3 is the anchor for the inner split of the first resample
df <- data.frame(x = c(1:2, 4:11))
r_set <- sliding_index(df, index = x, lookback = 4, assess_start = 3, assess_stop = 5)
r_split <- get_rsplit(r_set, 1)

internal_calibration_split(r_split, .get_split_args(r_set))
```

We expect this to not be relevant in practice too often, as chances of this happening are lower with more observations and we assume you are only attempting calibration with sufficiently large datasets.


### Period-based splitting

For the third sliding function in the family, we first aggregate the index to periods and then slide across those, e.g., aggregate daily data to monthly periods. 

![](images/calibration-split-period.jpg)

The principle of how to contruct a calibration split on the (outer) analysis set remains the same. The challenges of abstracting away from the rows, i.e., the observations directly, as illustrated for sliding over an index also remain. We observe a period, if we observe an index within that period. 

For `lookback = 0`, we remain unable to make a calibration split. However, it is possible to choose a smaller period when defining the resamples, e.g., using weeks instead of months when aggregating daily data. In that case, the calibration split is also defined in terms of weeks, rather than failing to split a single month on the scale of months.

Now that we've looked at the principles at play here on very small datasets, let's look at a dataset with a sufficient amount of observations to consider using calibration. The `Chicago` dataset has daily ridership numbers for over 15 years. Here we are constructing resamples which use 2 weeks worth of data for the assessment set, and roughly 15 years (52*15 weeks) for the analysis set. That's enough to use some of those observations for a calibration set!

```{r}
#| label: chicago

data(Chicago, package = "modeldata")

chicago_split <- initial_time_split(Chicago, prop = 1 - (14/nrow(Chicago)))
chicago_train <- training(chicago_split)

chicago_r_set <- sliding_period(
  chicago_train,
  index = "date",  
  period = "week",
  lookback = 52 * 15,
  assess_stop = 2,
  step = 2
)
chicago_r_set
```

Taking the first resample, we can see the two weeks for the assessment set.
```{r}
#| label: chicago_rsplit

chicago_r_split <- get_rsplit(chicago_r_set, 1)
chicago_assess <- assessment(chicago_r_split)
range(chicago_assess$date)
```

The calibration set consists of one week from the (outer) analysis set. 
```{r}
#| label: chicago_isplit

split_cal <- internal_calibration_split(chicago_r_split, .get_split_args(chicago_r_set))
chicago_cal <- calibration(split_cal)
range(chicago_cal$date)
```

## Splits for bootstrap samples

For bootstrap samples, our goal of avoiding overfitting means we need to refine our approach to making calibration splits a little further. 

If our (outer) analysis set is a bootstrap sample of the training data, it likely contains several replications of an observation. 

![](images/bootstrap-resample.jpg)

We don't want those to be split up into the inner analysis and the calibration set. Simply creating another bootstrap resample from the (outer) analysis set makes it likely for this to happen. In the example illustration below, the third row (`2-1`) of the (outer) analysis set is sampled twice into the (inner) analysis set. The fourth (`2-2`) and fifth (`2-3`) row are not sampled and thus are put into the calibration set. However, all those rows are actually the same row, hence we have the same row in both the (inner) analysis set and the calibration set.

![](images/bootstrap-regular.jpg)

To avoid this, we could do a grouped resampling with the row ID as the group. Then all replications of a row would end up in the same set, either (inner) analysis or calibration. However, this would mean that rows in the calibration set are potentially not unique, unlike the typical bootstrap OOB sample.

![](images/bootstrap-grouped.jpg)

To prevent both these potential issues, we create the calibration split by sampling the (inner) analysis set, with replacement, from the pool of unique rows in the (outer) analysis set. 

![](images/bootstrap-modified.jpg)

---
title: "Common Resampling Patterns"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Common Resampling Patterns}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = rlang::is_installed("modeldata")
)
```

The rsample package provides a number of resampling methods which are broadly applicable to a wide variety of modeling applications. This vignette walks through the most popular methods in the package, with brief descriptions of how they can be applied. For a more in-depth overview of resampling, check out the matching chapters in [Tidy Modeling with R](https://www.tmwr.org/resampling.html) and [Feature Engineering and Selection](http://www.feat.engineering/resampling.html).

Let's go ahead and load rsample now:

```{r}
#| label: setup
library(rsample)
```

We'll also load in a few data sets from the modeldata package. First, the Ames housing data, containing the sale prices of homes in Ames, Iowa:

```{r}
data(ames, package = "modeldata")
head(ames, 2)
```

Secondly, data on Chicago transit ridership numbers:

```{r}
data(Chicago, package = "modeldata")
head(Chicago, 2)
```

In addition to these data sets from the modeldata package, we'll also make use of the Orange data set in base R, containing repeated measurements of 5 orange trees over time:

```{r}
head(Orange, 2)
```

And last but not least, we'll set a seed so our results are reproducible:

```{r}
set.seed(123)
```


## Random Resampling

By far and away, the most common use for rsample is to generate simple random resamples of your data. The rsample package includes a number of functions specifically for this purpose.

### Initial Splits

To split your data into two sets -- often referred to as the "training" and "testing" sets -- rsample provides the `initial_split()` function:

```{r}
initial_split(ames)
```

The output of this is [an rsplit object](./rsample.html) with each observation assigned to one of the two sets. You can control the proportion of data assigned to the "training" set through the `prop` argument:

```{r}
initial_split(ames, prop = 0.8)
```

To get the actual data assigned to either set, use the `training()` and `testing()` functions:

```{r}
resample <- initial_split(ames, prop = 0.6)

head(training(resample), 2)
head(testing(resample), 2)
```

### V-Fold Cross-Validation

You should only evaluate models against your test set once, when you've completely finished tuning and training your models. To estimate performance of model candidates, you typically split your training data into one part used for model fitting and one part used for measuring performance. To distinguish those set from training and test set, we refer to them as analysis and assessment set, respectively. Typically, you split your training data into analysis and assessment sets multiple times to get stable estimates of model performance.

Perhaps the most common cross-validation method is [V-fold cross-validation](https://www.tmwr.org/resampling.html#cv). Also known as "k-fold cross-validation", this method creates V resamples by splitting your data into V groups (also known as "folds") of roughly equal size. The analysis set of each resample is made up of V-1 folds, with the remaining fold being used as the assessment set. This way, each observation in your data is used in exactly one assessment set.

To use V-fold cross-validation in rsample, use the `vfold_cv()` function:

```{r}
vfold_cv(ames, v = 2)
```

One downside to V-fold cross validation is that it tends to produce "noisy", or high-variance, estimates [when compared to other resampling methods](https://bookdown.org/max/FES/resampling.html#resample-var-bias). To try and reduce that variance, it's often helpful to perform what's known as [repeated cross-validation](https://www.tmwr.org/resampling.html#repeated-cross-validation), effectively running the V-fold resampling procedure multiple times for your data. To perform repeated V-fold cross-validation in rsample, you can use the repeats argument inside of `vfold_cv()`:

```{r}
vfold_cv(ames, v = 2, repeats = 2)
```


### Monte-Carlo Cross-Validation

An alternative to V-fold cross-validation is Monte-Carlo cross-validation. Where V-fold assigns each observation in your data to one (and exactly one) assessment set, Monte-Carlo cross-validation takes a random subset of your data for each assessment set, meaning each observation can be used in 0, 1, or many assessment sets. The analysis set is then made up of all the observations that weren't selected. Because each assessment set is sampled independently, you can repeat this as many times as you want.

To use Monte-Carlo cross-validation in rsample, use the `mc_cv()` function:

```{r}
mc_cv(ames, prop = 0.8, times = 2)
```

Similar to `initial_split()`, you can control the proportion of your data assigned to the analysis fold using `prop`. You can also control the number of resamples you create using the `times` argument.

Monte-Carlo cross-validation tends to produce more biased estimates than V-fold. As such, when computationally feasible we typically recommend using [five or so repeats of 10-fold cross-validation](https://bookdown.org/max/FES/resampling.html#resample-var-bias) for model assessment.

### Bootstrap Resampling

The last primary technique in rsample for creating resamples from the training data is bootstrap resampling. A "bootstrap sample" is a sample of your data set, the same size as your data set, taken with replacement so that a single observation might be sampled multiple times. The assessment set is then made up of all the observations that weren't selected for the analysis set. Generally, bootstrap resampling produces pessimistic estimates of model accuracy.

You can create bootstrap resamples in rsample using the `bootstraps()` function. While you can't control the proportion of data in each set -- the assessment set of a bootstrap resample is always the same size as the training data -- the function otherwise works exactly like `mc_cv()`:

```{r}
bootstraps(ames, times = 2)
```

### Validation Set

If your data is vast enough for a reliable performance estimate from just one assessment set, you can do a three-way split of your data into a training, validation and test set right at the start. (The validation set has the role of the single assessment set.) Instead of using `initial_split()` to create a binary split, you can use `initial_validation_split()` to create that three-way split:

```{r}
three_way_split <- initial_validation_split(ames, prop = c(0.6, 0.2))
three_way_split
```

The `prop` argument here has two elements, specifying the proportion of the data assigned to the training and the validation set.

To create an `rset` object for tuning, `validation_set()` bundles together the training and validation set, read for use with the [tune](https://tune.tidymodels.org/)) package.

```{r}
validation_set(three_way_split)
```

## Stratified Resampling

If your data is heavily imbalanced (that is, if the distribution of an important continuous variable is skewed, or some classes of a categorical variable are much more common than others), simple random resampling may accidentally skew your data even further by allocating more "rare" observations disproportionately into the analysis or assessment fold. In these situations, it can be useful to instead use [stratified resampling](https://www.tmwr.org/splitting.html#splitting-methods) to ensure the analysis and assessment folds have a similar distribution as your overall data.

All of the functions discussed so far support stratified resampling through their `strata` argument. This argument takes a single column identifier and uses it to stratify the resampling procedure:

```{r}
vfold_cv(ames, v = 2, strata = Sale_Price)
```

By default, rsample will cut continuous variables into four bins, and ensure that each bin is proportionally represented in each set. If desired, this behavior can be changed using the `breaks` argument:

```{r}
vfold_cv(ames, v = 2, strata = Sale_Price, breaks = 100)
```

## Grouped Resampling

Often, some observations in your data will be "more related" to each other than would be probable under random chance, for instance because they represent repeated measurements of the same subject or were all collected at a single location. In these situations, you often want to assign all related observations to either the analysis or assessment fold as a group, to avoid having assessment data that's closely related to the data used to fit a model.

All of the functions discussed so far have a "grouped resampling" variation to handle these situations. These functions all start with the `group_` prefix, and use the argument `group` to specify which column should be used to group observations. Other than respecting these groups, these functions all work like their ungrouped variants:

```{r}
resample <- group_initial_split(Orange, group = Tree)

unique(training(resample)$Tree)
unique(testing(resample)$Tree)
```

It's important to note that, while functions like `group_mc_cv()` still let you specify what proportion of your data should be in the analysis set (and `group_bootstraps()` still attempts to create analysis sets the same size as your original data), rsample won't "split" groups in order to exactly meet that proportion. These functions start out by assigning one group at random to each set (or, for `group_vfold_cv()`, to each fold) and then assign each of the remaining groups, in a random order, to whichever set brings the relative sizes of each set closest to the target proportion. That means that resamples are randomized, and you can safely use repeated cross-validation just as you would with ungrouped resampling, but also means you can wind up with very differently sized analysis and assessment sets than anticipated if your groups are unbalanced:

```{r}
set.seed(1)
group_bootstraps(ames, Neighborhood, times = 2)
```

While most of the grouped resampling functions are always focused on balancing the proportion of data in the analysis set, by default `group_vfold_cv()` will attempt to balance the number of groups assigned to each fold. If instead you'd like to balance the number of observations in each fold (meaning your assessment sets will be of similar sizes, but smaller groups will be more likely to be assigned to the same folds than would happen under random chance), you can use the argument `balance = "observations"`:

```{r}
group_vfold_cv(ames, Neighborhood, balance = "observations", v = 2)
```

If you're working with spatial data, your observations will often be more related to their neighbors than to the rest of the data set; as [Tobler's first law of geography](https://en.wikipedia.org/wiki/Tobler%27s_first_law_of_geography) puts it, "everything is related to everything else, but near things are more related than distant things." However, you often won't have a pre-defined "location" variable that you can use to group related observations. The [spatialsample](https://spatialsample.tidymodels.org/) package provides functions for spatial cross-validation using rsample syntax and classes, and is often useful for these situations.

## Time-Based Resampling

When working with time-based data, it usually doesn't make sense to randomly resample your data: random resampling will likely result in your analysis set having observations from later than your assessment set, which isn't a realistic way to assess model performance.

As such, rsample provides a few different functions to make sure that all data in your assessment sets are after that in the analysis set.

First off, two variants on `initial_split()` and `initial_validation_split()`, `initial_time_split()` and `initial_validation_time_split()`, will assign the _first_ rows of your data to the training set (with the number of rows assigned determined by `prop`):

```{r}
initial_time_split(Chicago)

initial_validation_time_split(Chicago)
```

There are also several functions in rsample to help you construct multiple analysis and assessment sets from time-based data. For instance, the `sliding_window()` will create "windows" of your data, moving down through the rows of the data frame:

```{r}
sliding_window(Chicago) |>
  head(2)
```

If you want to create sliding windows of your data based on a specific variable, you can use the `sliding_index()` function:

```{r}
sliding_index(Chicago, date) |>
  head(2)
```

And if you want to set the size of windows based on units of time, for instance to have each window contain a year of data, you can use `sliding_period()`:

```{r}
sliding_period(Chicago, date, "year") |>
  head(2)
```

All of these functions produce analysis sets of the same size, with the start and end of the analysis set "sliding" down your data frame. If you'd rather have your analysis set get progressively larger, so that you're predicting new data based upon a growing set of older observations, you can use the `sliding_window()` function with `lookback = -Inf`:

```{r}
sliding_window(Chicago, lookback = Inf) |>
  head(2)
```

This is commonly referred to as "evaluation on a rolling forecasting origin", or more colloquially, "rolling origin cross-validation".

Note that all of these time-based resampling functions are deterministic: unlike the rest of the package, running these functions repeatedly under different random seeds will always return the same results. 

---
title: "Recipes with rsample"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Recipes with rsample}
output:
  knitr:::html_vignette:
    toc: yes
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  eval = rlang::is_installed("ggplot2") && rlang::is_installed("modeldata")
)
options(digits = 3)
library(rsample)
library(recipes)
library(purrr)
```

The [recipes](https://recipes.tidymodels.org/) package contains a data preprocessor that can be used to avoid the potentially expensive formula methods as well as providing a richer set of data manipulation tools than base R can provide. This document uses version `r packageDescription("recipes")$Version` of recipes. 

In many cases, the preprocessing steps might contain quantities that require statistical estimation of parameters, such as

* signal extraction using principal component analysis

* imputation of missing values

* transformations of individual variables (e.g. Box-Cox transformations)

It is critical that any complex preprocessing steps be contained _inside_ of resampling so that the model performance estimates take into account the variability of these steps. 

Before discussing how rsample can use recipes, let's look at an example recipe for the Ames housing data. 

## An Example Recipe

For illustration, the Ames housing data will be used. There are sale prices of homes along with various other descriptors for the property:

```{r ames-data, message=FALSE}
data(ames, package = "modeldata")
```

Suppose that we will again fit a simple regression model with the formula:

```{r form, eval = FALSE}
log10(Sale_Price) ~ Neighborhood + House_Style + Year_Sold + Lot_Area
```

The distribution of the lot size is right-skewed:

```{r build, fig.alt = "The histogram of the lot size, showing a long right tail of the empirical distribution."}
library(ggplot2)
theme_set(theme_bw())
ggplot(ames, aes(x = Lot_Area)) + 
  geom_histogram(binwidth = 5000, col = "red", fill ="red", alpha = .5)
```

It might benefit the model if we estimate a transformation of the data using the Box-Cox procedure. 

Also, note that the frequencies of the neighborhoods can vary:

```{r hood, fig.alt = "A bar plot of the counts of different neighborhoods, with the frequencies ranging from over 400 observations of North_Ames to only a handful for Green_Hills and Landmark."}
ggplot(ames, aes(x = Neighborhood)) + geom_bar() + coord_flip() + xlab("")
```

When these are resampled, some neighborhoods will not be included in the test set and this will result in a column of dummy variables with zero entries. The same is true for the `House_Style` variable. We might want to collapse rarely occurring values into "other" categories. 

To define the design matrix, an initial recipe is created: 

```{r rec_setup, message=FALSE, warning=FALSE}
library(recipes)

# Apply log10 transformation outside the recipe
# https://www.tmwr.org/recipes.html#skip-equals-true
ames <-
  ames |>
  mutate(Sale_Price = log10(Sale_Price))

rec <- 
  recipe(Sale_Price ~ Neighborhood + House_Style + Year_Sold + Lot_Area, 
         data = ames) |>
  # Collapse rarely occurring jobs into "other"
  step_other(Neighborhood, House_Style, threshold = 0.05) |>
  # Dummy variables on the qualitative predictors
  step_dummy(all_nominal()) |>
  # Unskew a predictor
  step_BoxCox(Lot_Area) |>
  # Normalize
  step_center(all_predictors()) |>
  step_scale(all_predictors()) 
rec
```
This recreates the work that the formula method traditionally uses with the additional steps. 

While the original data object `ames` is used in the call, it is only used to define the variables and their characteristics so a single recipe is valid across all resampled versions of the data. The recipe can be estimated on the analysis component of the resample. 

If we execute the recipe on the entire data set:

```{r recipe-all}
rec_training_set <- prep(rec, training = ames)
rec_training_set
```
To get the values of the data, the `bake` function can be used: 

```{r baked}
# By default, the selector `everything()` is used to 
# return all the variables. Other selectors can be used too. 
bake(rec_training_set, new_data = head(ames))
```
Note that there are fewer dummy variables for `Neighborhood` and `House_Style` than in the data. 

Also, the above code using `prep()` benefits from the default argument of `retain = TRUE`, which keeps the processed version of the data set so that we don't have to reapply the steps to extract the processed values. For the data used to train the recipe, we would have used:

```{r juiced}
bake(rec_training_set, new_data = NULL) |> head()
```

The next section will explore recipes and bootstrap resampling for modeling:

```{r boot}
library(rsample)
set.seed(7712)
bt_samples <- bootstraps(ames)
bt_samples
bt_samples$splits[[1]]
```

## Working with Resamples

We can add a recipe column to the tibble. recipes has a convenience function called `prepper()` that can be used to call `prep()` but has the split object as the first argument (for easier purrring):

```{r col-pred}
library(purrr)

bt_samples$recipes <- map(bt_samples$splits, prepper, recipe = rec)
bt_samples
bt_samples$recipes[[1]]
```

Now, to fit the model, the fit function only needs the recipe as input. This is because the above code implicitly used the `retain = TRUE` option in `prep()`. Otherwise, the split objects would also be needed to `bake()` the recipe (as it will in the prediction function below). 

```{r cols-fit}
fit_lm <- function(rec_obj, ...) 
  lm(..., data = bake(rec_obj, new_data = NULL, everything()))

bt_samples$lm_mod <- 
  map(
    bt_samples$recipes, 
    fit_lm, 
    Sale_Price ~ .
  )
bt_samples
```

To get predictions, the function needs three arguments: the splits (to get the assessment data), the recipe (to process them), and the model. To iterate over these, the function `purrr::pmap()` is used: 

```{r cols-pred}
pred_lm <- function(split_obj, rec_obj, model_obj, ...) {
  mod_data <- bake(
    rec_obj, 
    new_data = assessment(split_obj),
    all_predictors(),
    all_outcomes()
  ) 
  
  out <- mod_data |> select(Sale_Price)
  out$predicted <- predict(model_obj, newdata = mod_data |> select(-Sale_Price))
  out
}

bt_samples$pred <- 
  pmap(
    lst(
      split_obj = bt_samples$splits, 
      rec_obj = bt_samples$recipes, 
      model_obj = bt_samples$lm_mod
    ),
    pred_lm 
  )
bt_samples
```

Calculating the RMSE:

```{r cols-rmse}
library(yardstick)
results <- map(bt_samples$pred, rmse, Sale_Price, predicted) |> list_rbind()
results
mean(results$.estimate)
```

# reverse_splits is working

    Code
      reverse_splits(1)
    Condition
      Error in `reverse_splits()`:
      ! `x` must be either an <rsplit> or an <rset> object.

---

    Code
      reverse_splits(permutes)
    Condition
      Error in `reverse_splits()`:
      ! Permutations cannot have their splits reversed.

---

    Code
      reverse_splits(permutes$splits[[1]])
    Condition
      Error in `reverse_splits()`:
      ! Permutations cannot have their splits reversed.

---
title: "Time Series Analysis Example"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Time Series Analysis Example}
output:
  knitr:::html_vignette:
    toc: yes
---

This vignette has been moved to [`tidymodels.org`](https://www.tidymodels.org/learn/models/time-series/). 

# bad args

    Code
      new_rset(car_folds$splits[1:2], car_folds$id)
    Condition
      Error in `new_rset()`:
      ! Split and ID vectors have different lengths.

---

    Code
      new_rset(car_folds$splits, car_folds["splits"])
    Condition
      Error in `new_rset()`:
      ! The `id` tibble column names should start with 'id'.

---

    Code
      new_rset(car_folds$splits, car_folds$splits)
    Condition
      Error in `new_rset()`:
      ! All ID columns should be character or factor vectors.

---

    Code
      new_rset(list(1), "x")
    Condition
      Error in `new_rset()`:
      ! Each element of `splits` must be an <rsplit> object.

---

    Code
      new_rset(car_folds$splits, car_folds$id, attrib = args)
    Condition
      Error in `new_rset()`:
      ! `attrib` should be a fully named list.

# not an rsplit

    Code
      analysis(folds$splits[1])
    Condition
      Error in `analysis()`:
      ! No method for objects of class: list

---

    Code
      assessment(folds$splits[1])
    Condition
      Error in `assessment()`:
      ! No method for objects of class: list

# lag

    Code
      rolling_origin(drinks, initial = 5, lag = 6)
    Condition
      Error in `rolling_origin()`:
      ! `lag` must be less than or equal to the number of training observations.

---

    Code
      rolling_origin(drinks, lag = 2.1)
    Condition
      Error in `rolling_origin()`:
      ! `lag` must be a whole number.

# grouped variants are consistent across R sessions

    Code
      analysis(rs$splits[[1]])
    Output
      # A tibble: 87 x 14
         name     height  mass hair_color skin_color eye_color birth_year sex   gender
         <chr>     <int> <dbl> <chr>      <chr>      <chr>          <dbl> <chr> <chr> 
       1 Luke Sk~    172    77 blond      fair       blue            19   male  mascu~
       2 C-3PO       167    75 <NA>       gold       yellow         112   none  mascu~
       3 Darth V~    202   136 none       white      yellow          41.9 male  mascu~
       4 Leia Or~    150    49 brown      light      brown           19   fema~ femin~
       5 Leia Or~    150    49 brown      light      brown           19   fema~ femin~
       6 Leia Or~    150    49 brown      light      brown           19   fema~ femin~
       7 Beru Wh~    165    75 brown      light      blue            47   fema~ femin~
       8 Biggs D~    183    84 black      light      brown           24   male  mascu~
       9 Biggs D~    183    84 black      light      brown           24   male  mascu~
      10 Biggs D~    183    84 black      light      brown           24   male  mascu~
      # i 77 more rows
      # i 5 more variables: homeworld <chr>, species <chr>, films <list>,
      #   vehicles <list>, starships <list>

---

    Code
      analysis(rs$splits[[1]])
    Output
      # A tibble: 86 x 14
         name     height  mass hair_color skin_color eye_color birth_year sex   gender
         <chr>     <int> <dbl> <chr>      <chr>      <chr>          <dbl> <chr> <chr> 
       1 Luke Sk~    172    77 blond      fair       blue            19   male  mascu~
       2 C-3PO       167    75 <NA>       gold       yellow         112   none  mascu~
       3 R2-D2        96    32 <NA>       white, bl~ red             33   none  mascu~
       4 Darth V~    202   136 none       white      yellow          41.9 male  mascu~
       5 Leia Or~    150    49 brown      light      brown           19   fema~ femin~
       6 Owen La~    178   120 brown, gr~ light      blue            52   male  mascu~
       7 Beru Wh~    165    75 brown      light      blue            47   fema~ femin~
       8 R5-D4        97    32 <NA>       white, red red             NA   none  mascu~
       9 Biggs D~    183    84 black      light      brown           24   male  mascu~
      10 Obi-Wan~    182    77 auburn, w~ fair       blue-gray       57   male  mascu~
      # i 76 more rows
      # i 5 more variables: homeworld <chr>, species <chr>, films <list>,
      #   vehicles <list>, starships <list>

---

    Code
      analysis(rs$splits[[1]])
    Output
      # A tibble: 65 x 14
         name     height  mass hair_color skin_color eye_color birth_year sex   gender
         <chr>     <int> <dbl> <chr>      <chr>      <chr>          <dbl> <chr> <chr> 
       1 C-3PO       167    75 <NA>       gold       yellow         112   none  mascu~
       2 R2-D2        96    32 <NA>       white, bl~ red             33   none  mascu~
       3 Leia Or~    150    49 brown      light      brown           19   fema~ femin~
       4 Owen La~    178   120 brown, gr~ light      blue            52   male  mascu~
       5 Beru Wh~    165    75 brown      light      blue            47   fema~ femin~
       6 R5-D4        97    32 <NA>       white, red red             NA   none  mascu~
       7 Biggs D~    183    84 black      light      brown           24   male  mascu~
       8 Obi-Wan~    182    77 auburn, w~ fair       blue-gray       57   male  mascu~
       9 Anakin ~    188    84 blond      fair       blue            41.9 male  mascu~
      10 Greedo      173    74 <NA>       green      black           44   male  mascu~
      # i 55 more rows
      # i 5 more variables: homeworld <chr>, species <chr>, films <list>,
      #   vehicles <list>, starships <list>

---

    Code
      analysis(rs$splits[[1]])
    Output
      # A tibble: 65 x 14
         name     height  mass hair_color skin_color eye_color birth_year sex   gender
         <chr>     <int> <dbl> <chr>      <chr>      <chr>          <dbl> <chr> <chr> 
       1 C-3PO       167    75 <NA>       gold       yellow         112   none  mascu~
       2 Darth V~    202   136 none       white      yellow          41.9 male  mascu~
       3 Leia Or~    150    49 brown      light      brown           19   fema~ femin~
       4 Owen La~    178   120 brown, gr~ light      blue            52   male  mascu~
       5 Beru Wh~    165    75 brown      light      blue            47   fema~ femin~
       6 R5-D4        97    32 <NA>       white, red red             NA   none  mascu~
       7 Biggs D~    183    84 black      light      brown           24   male  mascu~
       8 Obi-Wan~    182    77 auburn, w~ fair       blue-gray       57   male  mascu~
       9 Wilhuff~    180    NA auburn, g~ fair       blue            64   male  mascu~
      10 Chewbac~    228   112 brown      unknown    blue           200   male  mascu~
      # i 55 more rows
      # i 5 more variables: homeworld <chr>, species <chr>, films <list>,
      #   vehicles <list>, starships <list>

# informative error messages

    Code
      internal_calibration_split(NULL)
    Condition
      Error in `internal_calibration_split()`:
      ! No method for objects of class: NULL.

---

    Code
      internal_calibration_split(m_rset)
    Condition
      Error in `internal_calibration_split()`:
      ! No method for objects of classes: manual_rset, rset, tbl_df, tbl, and data.frame.

# mc_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# group_mc_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# vfold_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# group_vfold_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# boot_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# group_boot_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# val_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# group_val_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# time_val_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# clustering_split can create mock split

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# sliding_window_split needs at least 2 observations

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      This set cannot be split into an analysis and a calibration set as there is only one row; creating an empty calibration set.

# sliding_window_split with incomplete sets

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# sliding_index_split needs at least 2 observations

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      This set cannot be split into an analysis and a calibration set as there is only one row; creating an empty calibration set.

# sliding_index_split with incomplete sets

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# sliding_period_split needs at least 2 observations

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      This set cannot be split into an analysis and a calibration set as there is only one row; creating an empty calibration set.

# sliding_period_split needs observations in at least 2 periods

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

---

    Code
      split_cal <- internal_calibration_split(r_split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# initial_time_split can create mock split

    Code
      split_cal <- internal_calibration_split(split, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# initial_validation_split can create mock split

    Code
      split_cal <- internal_calibration_split(initial_vsplit, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# group_initial_validation_split can create mock split

    Code
      split_cal <- internal_calibration_split(initial_vsplit, split_args)
    Condition
      Warning:
      Cannot create calibration split; creating an empty calibration set.

# initial_validation_time_split can create mock split

    Code
      split_cal <- internal_calibration_split(initial_vsplit, split_args)
    Condition
      Warning:
      This set cannot be split into a training and a calibration set as there is only one row; creating an empty calibration set.

# assessment() fails

    Code
      assessment(internal_calibration_split)
    Condition
      Error in `assessment()`:
      ! Internal calibration splits are designed to only return analysis and calibration sets.

# print()

    Code
      print(internal_calibration_split)
    Output
      <Analysis/Calibration/Total>
      <32/0/32>

# reshuffle_rset is working

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical <rset> when called on <sliding_index> objects.
    Output
      # Sliding index resampling 
      # A tibble: 49 x 2
         splits        id     
         <list>        <chr>  
       1 <split [1/1]> Slice01
       2 <split [1/1]> Slice02
       3 <split [1/1]> Slice03
       4 <split [1/1]> Slice04
       5 <split [1/1]> Slice05
       6 <split [1/1]> Slice06
       7 <split [1/1]> Slice07
       8 <split [1/1]> Slice08
       9 <split [1/1]> Slice09
      10 <split [1/1]> Slice10
      # i 39 more rows

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical <rset> when called on <sliding_period> objects.
    Output
      # Sliding period resampling 
      # A tibble: 7 x 2
        splits        id    
        <list>        <chr> 
      1 <split [7/7]> Slice1
      2 <split [7/7]> Slice2
      3 <split [7/7]> Slice3
      4 <split [7/7]> Slice4
      5 <split [7/7]> Slice5
      6 <split [7/7]> Slice6
      7 <split [7/1]> Slice7

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical <rset> when called on <sliding_window> objects.
    Output
      # Sliding window resampling 
      # A tibble: 49 x 2
         splits        id     
         <list>        <chr>  
       1 <split [1/1]> Slice01
       2 <split [1/1]> Slice02
       3 <split [1/1]> Slice03
       4 <split [1/1]> Slice04
       5 <split [1/1]> Slice05
       6 <split [1/1]> Slice06
       7 <split [1/1]> Slice07
       8 <split [1/1]> Slice08
       9 <split [1/1]> Slice09
      10 <split [1/1]> Slice10
      # i 39 more rows

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical <rset> when called on <rolling_origin> objects.
    Output
      # Rolling origin forecast resampling 
      # A tibble: 45 x 2
         splits         id     
         <list>         <chr>  
       1 <split [5/1]>  Slice01
       2 <split [6/1]>  Slice02
       3 <split [7/1]>  Slice03
       4 <split [8/1]>  Slice04
       5 <split [9/1]>  Slice05
       6 <split [10/1]> Slice06
       7 <split [11/1]> Slice07
       8 <split [12/1]> Slice08
       9 <split [13/1]> Slice09
      10 <split [14/1]> Slice10
      # i 35 more rows

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical <rset> when called on <validation_time_split> objects.
    Output
      # Validation Set Split (0.75/0.25)  
      # A tibble: 1 x 2
        splits          id        
        <list>          <chr>     
      1 <split [37/13]> validation

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical <rset> when called on <validation_set> objects.
    Output
      # A tibble: 1 x 2
        splits          id        
        <list>          <chr>     
      1 <split [30/10]> validation

---

    Code
      reshuffle_rset(resample)
    Condition
      Error in `reshuffle_rset()`:
      ! Cannot reshuffle this rset (`attr(rset, 'strata')` is "TRUE", not a column identifier)
      i If the original object was created with an older version of rsample, try recreating it with the newest version of the package.

---

    Code
      reshuffle_rset(rset_subclasses[["manual_rset"]])
    Condition
      Error in `reshuffle_rset()`:
      ! `manual_rset` objects cannot be reshuffled.

---

    Code
      reshuffle_rset(rset_subclasses[["manual_rset"]]$splits[[1]])
    Condition
      Error in `reshuffle_rset()`:
      ! `rset` must be an <rset> object.

---
title: "Bootstrap Confidence Intervals"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Bootstrap Confidence Intervals}
output:
  knitr:::html_vignette:
    toc: yes
---


```{r setup, include=FALSE}
library(tidymodels)
library(nlstools)
library(GGally)
theme_set(theme_bw())
```

The bootstrap was originally intended for estimating confidence intervals for complex statistics whose variance properties are difficult to analytically derive. Davison and Hinkley's [_Bootstrap Methods and Their Application_](https://www.cambridge.org/core/books/bootstrap-methods-and-their-application/ED2FD043579F27952363566DC09CBD6A) is a great resource for these methods. `rsample` contains a few function to compute the most common types of intervals. 

To demonstrate the computations for the different types of intervals, we'll use a nonlinear regression example from [Baty _et al_ (2015)](https://www.jstatsoft.org/article/view/v066i05). They showed data that monitored oxygen uptake in a patient with rest and exercise phases (in the data frame `O2K`). 

```{r O2K-dat}
library(tidymodels)
library(nlstools)
library(GGally)

data(O2K)

ggplot(O2K, aes(x = t, y = VO2)) + 
  geom_point()
```

The authors fit a segmented regression model where the transition point was known (this is the time when exercise commenced). Their model was:

```{r O2K-fit}
nonlin_form <-  
  as.formula(
    VO2 ~ (t <= 5.883) * VO2rest + 
      (t > 5.883) * 
      (VO2rest + (VO2peak - VO2rest) * (1 - exp(-(t - 5.883) / mu)))
    )

# Starting values from visual inspection
start_vals <- list(VO2rest = 400, VO2peak = 1600, mu = 1)

res <- nls(nonlin_form, start = start_vals, data = O2K) 

tidy(res)
```


`broom::tidy()` returns our analysis object in a standardized way. The column names shown here are used for most types of objects and this allows us to use the results more easily. 

For `rsample`, we'll rely on the `tidy()` method to work with bootstrap estimates when we need confidence intervals. There's an example at the end of a univariate statistic that isn't automatically formatted with `tidy()`. 

To run our model over different bootstraps, we'll write a function that uses the `split` object as input and produces a tidy data frame:

```{r model-info}
# Will be used to fit the models to different bootstrap data sets:
fit_fun <- function(split, ...) {
  # We could check for convergence, make new parameters, etc.
  nls(nonlin_form, data = analysis(split), ...) %>%
    tidy()
}
```

First, let's create a set of resamples and fit separate models to each. The options `apparent = TRUE` will be set. This creates a final resample that is a copy of the original (unsampled) data set. This is required for some of the interval methods. 

```{r resample}
set.seed(462)
nlin_bt <-
  bootstraps(O2K, times = 2000, apparent = TRUE) %>%
  mutate(models = map(splits, ~ fit_fun(.x, start = start_vals)))
nlin_bt

nlin_bt$models[[1]]
```

Let's look at the data and see if there any outliers or aberrant results:

```{r extract}
library(tidyr)
nls_coef <- 
  nlin_bt %>%
  dplyr::select(-splits) %>%
  # Turn it into a tibble by stacking the `models` col
  unnest() %>%
  # Get rid of unneeded columns
  dplyr::select(id, term, estimate) 

head(nls_coef)
```
Now let's create a scatterplot matrix:

```{r splom} 
nls_coef %>%
  # Put different parameters in columns
  tidyr::spread(term, estimate) %>% 
  # Keep only numeric columns
  dplyr::select(-id) %>% 
  ggscatmat(alpha = .25)
```

One potential outlier on the right for `VO2peak` but we'll leave it in. 

The univariate distributions are:

```{r hists}
nls_coef %>% 
  ggplot(aes(x = estimate)) + 
  geom_histogram(bins = 20, col = "white") + 
  facet_wrap(~ term, scales = "free_x")
```

## Percentile intervals

The most basic type of interval uses _percentiles_ of the resampling distribution. To get the percentile intervals, the `rset` object is passed as the first argument and the second argument is the list column of tidy results: 

```{r pctl}
p_ints <- int_pctl(nlin_bt, models)
p_ints
```

When overlaid with the univariate distributions: 

```{r pctl-plot}
nls_coef %>% 
  ggplot(aes(x = estimate)) + 
  geom_histogram(bins = 20, col = "white") + 
  facet_wrap(~ term, scales = "free_x") + 
  geom_vline(data = p_ints, aes(xintercept = .lower), col = "red") + 
  geom_vline(data = p_ints, aes(xintercept = .upper), col = "red")
```

How do these intervals compare to the parametric asymptotic values?

```{r int-compare}
parametric <- 
  tidy(res, conf.int = TRUE) %>% 
  dplyr::select(
    term,
    .lower = conf.low,
    .estimate = estimate,
    .upper = conf.high
  ) %>% 
  mutate(
    .alpha = 0.05,
    .method = "parametric"
  )

intervals <- 
  bind_rows(parametric, p_ints) %>% 
  arrange(term, .method)
intervals %>% split(intervals$term)
```

The percentile intervals are wider than the parametric intervals (which assume asymptotic normality). Do the estimates appear to be normally distributed? We can look at quantile-quantile plots:  

```{r qqplot}
nls_coef %>% 
  ggplot(aes(sample = estimate)) + 
  stat_qq() +
  stat_qq_line(alpha = .25) + 
  facet_wrap(~ term, scales = "free") 
```

## t-intervals

Bootstrap _t_-intervals are estimated by computing intermediate statistics that are _t_-like in structure. To use these, we require the estimated variance _for each individual resampled estimate_. In our example, this comes along with the fitted model object. We can extract the standard errors of the parameters. Luckily, most `tidy()` provide this in a column named `std.error`. 

The arguments for these intervals are the same:

```{r t-ints}
t_stats <- int_t(nlin_bt, models)
intervals <- 
  bind_rows(intervals, t_stats) %>% 
  arrange(term, .method)
intervals %>% split(intervals$term)
```


## Bias-corrected and accelerated intervals

For bias-corrected and accelerated (BCa) intervals, an additional argument is required. The `.fn` argument is a function that computes the statistic of interest. The first argument should be for the `rsplit` object and other arguments can be passed in using the ellipses. 

These intervals use an internal leave-one-out resample to compute the Jackknife statistic and will recompute the statistic for _every bootstrap resample_. If the statistic is expensive to compute, this may take some time. For those calculations, we use the `furrr` package so these can be computed in parallel if you have set up a parallel processing plan (see `?future::plan`). 

The user-facing function takes an argument for the function and the ellipses. 

```{r bca-comp}
bias_corr <- int_bca(nlin_bt, models, .fn = fit_fun, start = start_vals)
intervals <- 
  bind_rows(intervals, bias_corr) %>% 
  arrange(term, .method)
intervals %>% split(intervals$term)
```


## No existing tidy method

In this case, your function can emulate the minimum results: 

 * a character column called `term`, 
 * a numeric column called `estimate`, and, optionally, 
 * a numeric column called `std.error`. 

The last column is only needed for `int_t`. 

Suppose we just want to estimate the fold-increase in the outcome between the 90th and 10th percentiles over the course of the experiment. Our function might look like:

```{r fold-foo}
fold_incr <- function(split, ...) {
  dat <- analysis(split)
  quants <- quantile(dat$VO2, probs = c(.1, .9))
  tibble(
    term = "fold increase",
    estimate = unname(quants[2]/quants[1]),
    # We don't know the analytical formula for this 
    std.error = NA_real_
  )
}
```

Everything else works the same as before:

```{r fold-ci}
nlin_bt <-
  nlin_bt %>%
  mutate(folds = map(splits, fold_incr))

int_pctl(nlin_bt, folds)
int_bca(nlin_bt, folds, .fn = fold_incr)
```

 

---
title: "Bootstrap confidence intervals"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Bootstrap confidence intervals}
output:
  knitr:::html_vignette:
    toc: yes
---


```{r setup, include=FALSE}
library(tidymodels)
library(nlstools)
library(GGally)
theme_set(theme_bw())
```

The bootstrap was originally intended for estimating confidence intervals for complex statistics whose variance properties are difficult to analytically derive. Davison and Hinkley's [_Bootstrap Methods and Their Application_](https://www.cambridge.org/core/books/bootstrap-methods-and-their-application/ED2FD043579F27952363566DC09CBD6A) is a great resource for these methods. `rsample` contains a few function to compute the most common types of intervals. 

## A nonlinear regression example

To demonstrate the computations for the different types of intervals, we'll use a nonlinear regression example from [Baty _et al_ (2015)](https://www.jstatsoft.org/article/view/v066i05). They showed data that monitored oxygen uptake in a patient with rest and exercise phases (in the data frame `O2K`). 

```{r O2K-dat}
library(tidymodels)
library(nlstools)
library(GGally)

data(O2K)

ggplot(O2K, aes(x = t, y = VO2)) + 
  geom_point()
```

The authors fit a segmented regression model where the transition point was known (this is the time when exercise commenced). Their model was:

```{r O2K-fit}
nonlin_form <-  
  as.formula(
    VO2 ~ (t <= 5.883) * VO2rest + 
      (t > 5.883) * 
      (VO2rest + (VO2peak - VO2rest) * (1 - exp(-(t - 5.883) / mu)))
    )

# Starting values from visual inspection
start_vals <- list(VO2rest = 400, VO2peak = 1600, mu = 1)

res <- nls(nonlin_form, start = start_vals, data = O2K) 

tidy(res)
```


`broom::tidy()` returns our analysis object in a standardized way. The column names shown here are used for most types of objects and this allows us to use the results more easily. 

For `rsample`, we'll rely on the `tidy()` method to work with bootstrap estimates when we need confidence intervals. There's an example at the end of a univariate statistic that isn't automatically formatted with `tidy()`. 

To run our model over different bootstraps, we'll write a function that uses the `split` object as input and produces a tidy data frame:

```{r model-info}
# Will be used to fit the models to different bootstrap data sets:
fit_fun <- function(split, ...) {
  # We could check for convergence, make new parameters, etc.
  nls(nonlin_form, data = analysis(split), ...) %>%
    tidy()
}
```

First, let's create a set of resamples and fit separate models to each. The options `apparent = TRUE` will be set. This creates a final resample that is a copy of the original (unsampled) data set. This is required for some of the interval methods. 

```{r resample}
set.seed(462)
nlin_bt <-
  bootstraps(O2K, times = 2000, apparent = TRUE) %>%
  mutate(models = map(splits, ~ fit_fun(.x, start = start_vals)))
nlin_bt

nlin_bt$models[[1]]
```

Let's look at the data and see if there any outliers or aberrant results:

```{r extract}
library(tidyr)
nls_coef <- 
  nlin_bt %>%
  dplyr::select(-splits) %>%
  # Turn it into a tibble by stacking the `models` col
  unnest() %>%
  # Get rid of unneeded columns
  dplyr::select(id, term, estimate) 

head(nls_coef)
```
Now let's create a scatterplot matrix:

```{r splom} 
nls_coef %>%
  # Put different parameters in columns
  tidyr::spread(term, estimate) %>% 
  # Keep only numeric columns
  dplyr::select(-id) %>% 
  ggscatmat(alpha = .25)
```

One potential outlier on the right for `VO2peak` but we'll leave it in. 

The univariate distributions are:

```{r hists}
nls_coef %>% 
  ggplot(aes(x = estimate)) + 
  geom_histogram(bins = 20, col = "white") + 
  facet_wrap(~ term, scales = "free_x")
```

### Percentile intervals

The most basic type of interval uses _percentiles_ of the resampling distribution. To get the percentile intervals, the `rset` object is passed as the first argument and the second argument is the list column of tidy results: 

```{r pctl}
p_ints <- int_pctl(nlin_bt, models)
p_ints
```

When overlaid with the univariate distributions: 

```{r pctl-plot}
nls_coef %>% 
  ggplot(aes(x = estimate)) + 
  geom_histogram(bins = 20, col = "white") + 
  facet_wrap(~ term, scales = "free_x") + 
  geom_vline(data = p_ints, aes(xintercept = .lower), col = "red") + 
  geom_vline(data = p_ints, aes(xintercept = .upper), col = "red")
```

How do these intervals compare to the parametric asymptotic values?

```{r int-compare}
parametric <- 
  tidy(res, conf.int = TRUE) %>% 
  dplyr::select(
    term,
    .lower = conf.low,
    .estimate = estimate,
    .upper = conf.high
  ) %>% 
  mutate(
    .alpha = 0.05,
    .method = "parametric"
  )

intervals <- 
  bind_rows(parametric, p_ints) %>% 
  arrange(term, .method)
intervals %>% split(intervals$term)
```

The percentile intervals are wider than the parametric intervals (which assume asymptotic normality). Do the estimates appear to be normally distributed? We can look at quantile-quantile plots:  

```{r qqplot}
nls_coef %>% 
  ggplot(aes(sample = estimate)) + 
  stat_qq() +
  stat_qq_line(alpha = .25) + 
  facet_wrap(~ term, scales = "free") 
```

### t-intervals

Bootstrap _t_-intervals are estimated by computing intermediate statistics that are _t_-like in structure. To use these, we require the estimated variance _for each individual resampled estimate_. In our example, this comes along with the fitted model object. We can extract the standard errors of the parameters. Luckily, most `tidy()` provide this in a column named `std.error`. 

The arguments for these intervals are the same:

```{r t-ints}
t_stats <- int_t(nlin_bt, models)
intervals <- 
  bind_rows(intervals, t_stats) %>% 
  arrange(term, .method)
intervals %>% split(intervals$term)
```


### Bias-corrected and accelerated intervals

For bias-corrected and accelerated (BCa) intervals, an additional argument is required. The `.fn` argument is a function that computes the statistic of interest. The first argument should be for the `rsplit` object and other arguments can be passed in using the ellipses. 

These intervals use an internal leave-one-out resample to compute the Jackknife statistic and will recompute the statistic for _every bootstrap resample_. If the statistic is expensive to compute, this may take some time. For those calculations, we use the `furrr` package so these can be computed in parallel if you have set up a parallel processing plan (see `?future::plan`). 

The user-facing function takes an argument for the function and the ellipses. 

```{r bca-comp}
bias_corr <- int_bca(nlin_bt, models, .fn = fit_fun, start = start_vals)
intervals <- 
  bind_rows(intervals, bias_corr) %>% 
  arrange(term, .method)
intervals %>% split(intervals$term)
```


## No existing tidy method

In this case, your function can emulate the minimum results: 

 * a character column called `term`, 
 * a numeric column called `estimate`, and, optionally, 
 * a numeric column called `std.error`. 

The last column is only needed for `int_t`. 

Suppose we just want to estimate the fold-increase in the outcome between the 90th and 10th percentiles over the course of the experiment. Our function might look like:

```{r fold-foo}
fold_incr <- function(split, ...) {
  dat <- analysis(split)
  quants <- quantile(dat$VO2, probs = c(.1, .9))
  tibble(
    term = "fold increase",
    estimate = unname(quants[2]/quants[1]),
    # We don't know the analytical formula for this 
    std.error = NA_real_
  )
}
```

Everything else works the same as before:

```{r fold-ci}
nlin_bt <-
  nlin_bt %>%
  mutate(folds = map(splits, fold_incr))

int_pctl(nlin_bt, folds)
int_bca(nlin_bt, folds, .fn = fold_incr)
```

## Intervals for linear(ish) parametric intervals

`rsample` also contains the `reg_intervals()` function that can be used for linear regression (via `lm()`), generalized linear models (`glm()`), or log-linear survival models (`survival::survreg()` or `survival::coxph()`). This function makes it easier to get intervals for these models. 

A simple example is a logistic regression using the dementia data from the `modeldata` package: 

```{r ad-data}
data(ad_data, package = "modeldata")
```

Let's fit a model with a few predictors: 

```{r ad-model}
lr_mod <- glm(Class ~ male + age + Ab_42 + tau, data = ad_data,
              family = binomial)
glance(lr_mod)
tidy(lr_mod)
```

Let's use this model with student-t intervals: 

```{r ad-t-int}
set.seed(29832)
lr_int <- 
  reg_intervals(Class ~ male + age + Ab_42 + tau, 
                data = ad_data,
                model_fn = "glm",
                family = binomial)
lr_int
```

We can also save the resamples for plotting: 

```{r ad-t-int-plot}
set.seed(29832)
lr_int <- 
  reg_intervals(Class ~ male + age + Ab_42 + tau, 
                data = ad_data,
                keep_reps = TRUE,
                model_fn = "glm",
                family = binomial)
lr_int
```

Now we can unnest the data to use in a ggplot: 

```{r ad-plot}
lr_int %>% 
  select(term, .replicates) %>% 
  unnest(cols = .replicates) %>% 
  ggplot(aes(x = estimate)) + 
  geom_histogram(bins = 30) + 
  facet_wrap(~ term, scales = "free_x") + 
  geom_vline(data = lr_int, aes(xintercept = .lower), col = "red") + 
  geom_vline(data = lr_int, aes(xintercept = .upper), col = "red") + 
  geom_vline(xintercept = 0, col = "green")
```

# `data` is validated

    Code
      sliding_window(1)
    Condition
      Error in `sliding_window()`:
      ! `data` must be a data frame.

---

    Code
      sliding_index(1)
    Condition
      Error in `sliding_index()`:
      ! `data` must be a data frame.

---

    Code
      sliding_period(1)
    Condition
      Error in `sliding_period()`:
      ! `data` must be a data frame.

# `lookback` is validated

    Code
      sliding_window(data.frame(), lookback = -1)
    Condition
      Error in `check_lookback()`:
      ! `lookback` must be positive, or zero.

---

    Code
      sliding_window(data.frame(), lookback = "a")
    Condition
      Error in `check_lookback()`:
      ! `lookback` must be an integer of size 1, or `Inf`.

---

    Code
      sliding_window(data.frame(), lookback = c(1, 2))
    Condition
      Error in `check_lookback()`:
      ! `lookback` must have size 1.

---

    Code
      sliding_window(data.frame(), lookback = NA)
    Condition
      Error in `check_lookback()`:
      ! `lookback` must be an integer of size 1, or `Inf`.

# `assess_start` is validated

    Code
      sliding_window(data.frame(), assess_start = -1)
    Condition
      Error in `check_assess()`:
      ! `assess_start` must be positive.

---

    Code
      sliding_window(data.frame(), assess_start = "a")
    Condition
      Error in `check_assess()`:
      ! `assess_start` must be an integer of size 1, or `Inf`.

---

    Code
      sliding_window(data.frame(), assess_start = c(1, 2))
    Condition
      Error in `check_assess()`:
      ! `assess_start` must have size 1.

---

    Code
      sliding_window(data.frame(), assess_start = NA)
    Condition
      Error in `check_assess()`:
      ! `assess_start` must be an integer of size 1, or `Inf`.

# `assess_stop` is validated

    Code
      sliding_window(data.frame(), assess_stop = -1)
    Condition
      Error in `check_assess()`:
      ! `assess_stop` must be positive.

---

    Code
      sliding_window(data.frame(), assess_stop = "a")
    Condition
      Error in `check_assess()`:
      ! `assess_stop` must be an integer of size 1, or `Inf`.

---

    Code
      sliding_window(data.frame(), assess_stop = c(1, 2))
    Condition
      Error in `check_assess()`:
      ! `assess_stop` must have size 1.

---

    Code
      sliding_window(data.frame(), assess_stop = NA)
    Condition
      Error in `check_assess()`:
      ! `assess_stop` must be an integer of size 1, or `Inf`.

# `assess_start` must be before or equal to `assess_stop`

    Code
      sliding_window(data.frame(), assess_start = 2, assess_stop = 1)
    Condition
      Error in `sliding_window()`:
      ! `assess_start` must be less than or equal to `assess_stop`.

# `index` is validated

    Code
      sliding_index(df, y)
    Condition
      Error in `sliding_index()`:
      ! Can't select columns that don't exist.
      x Column `y` doesn't exist.

---

    Code
      sliding_period(df, y)
    Condition
      Error in `sliding_period()`:
      ! Can't select columns that don't exist.
      x Column `y` doesn't exist.

# cannot create a split with an empty analysis set

    Code
      make_splits(indices, df)
    Condition
      Error in `rsplit()`:
      ! At least one row should be selected for the analysis set.

# cannot create a split from empty training dataframe

    Code
      make_splits(training, testing)
    Condition
      Error in `make_splits()`:
      ! The analysis set must contain at least one row.

# cannot create a split from dataframes with different columns

    Code
      make_splits(training, testing)
    Condition
      Error in `make_splits()`:
      ! The analysis and assessment sets must have the same columns.

# improper argument

    Code
      make_splits("potato")
    Condition
      Error in `make_splits()`:
      ! No method for objects of class: character

# get_rsplit()

    Code
      get_rsplit(val, 3)
    Condition
      Error in `get_rsplit()`:
      ! `index` must be a length-1 integer between 1 and 1.
      * A value of 3 was provided.

---

    Code
      get_rsplit(val, c(1, 2))
    Condition
      Error in `get_rsplit()`:
      ! `index` must be a length-1 integer between 1 and 1.
      * Index was of length 2.

---

    Code
      get_rsplit(val, 1.5)
    Condition
      Error in `get_rsplit()`:
      ! `index` must be a length-1 integer between 1 and 1.
      * A value of 1.5 was provided.

---

    Code
      get_rsplit(warpbreaks, 1)
    Condition
      Error in `get_rsplit()`:
      ! No method for objects of class: data.frame

# strata

    Code
      rs4 <- vfold_cv(mlc_churn, strata = state, pool = 0.01)
    Condition
      Warning:
      Stratifying groups that make up 1% of the data may be statistically risky.
      i Consider increasing `pool` to at least 0.1.

# strata arg is checked

    Code
      vfold_cv(iris, strata = iris$Species)
    Condition
      Error in `vfold_cv()`:
      ! Can't select columns that don't exist.
      x Columns `setosa`, `setosa`, `setosa`, `setosa`, `setosa`, etc. don't exist.

---

    Code
      vfold_cv(iris, strata = c("Species", "Sepal.Width"))
    Condition
      Error in `vfold_cv()`:
      ! `strata` must be a single string or `NULL`, not a character vector.

---

    Code
      vfold_cv(iris, strata = NA)
    Condition
      Error in `vfold_cv()`:
      ! Selections can't have missing values.

---

    Code
      vfold_cv(dat, strata = b)
    Condition
      Error in `vfold_cv()`:
      ! strata cannot be a <Surv> object.
      i Use the time or event variable directly.

# v arg is checked

    Code
      vfold_cv(iris, v = -500)
    Condition
      Error in `vfold_cv()`:
      ! `v` must be a whole number larger than or equal to 2, not the number -500.

---

    Code
      vfold_cv(iris, v = 1)
    Condition
      Error in `vfold_cv()`:
      ! `v` must be a whole number larger than or equal to 2, not the number 1.

---

    Code
      vfold_cv(iris, v = NULL)
    Condition
      Error in `vfold_cv()`:
      ! `v` must be a whole number, not `NULL`.

---

    Code
      vfold_cv(iris, v = 500)
    Condition
      Error in `vfold_cv()`:
      ! The number of rows is less than `v` = 500.

---

    Code
      vfold_cv(mtcars, v = nrow(mtcars))
    Condition
      Error in `vfold_cv()`:
      ! Leave-one-out cross-validation is not supported by this function.
      x You set `v` to `nrow(data)`, which would result in a leave-one-out cross-validation.
      i Use `loo_cv()` in this case.

# repeats arg is checked

    Code
      vfold_cv(iris, v = 150, repeats = 2)
    Condition
      Error in `vfold_cv()`:
      ! Repeated resampling when `v` is 150 would create identical resamples.

---

    Code
      vfold_cv(Orange, repeats = 0)
    Condition
      Error in `vfold_cv()`:
      ! `repeats` must be a whole number larger than or equal to 1, not the number 0.

---

    Code
      vfold_cv(Orange, repeats = NULL)
    Condition
      Error in `vfold_cv()`:
      ! `repeats` must be a whole number, not `NULL`.

# printing

    Code
      vfold_cv(mtcars)
    Output
      #  10-fold cross-validation 
      # A tibble: 10 x 2
         splits         id    
         <list>         <chr> 
       1 <split [28/4]> Fold01
       2 <split [28/4]> Fold02
       3 <split [29/3]> Fold03
       4 <split [29/3]> Fold04
       5 <split [29/3]> Fold05
       6 <split [29/3]> Fold06
       7 <split [29/3]> Fold07
       8 <split [29/3]> Fold08
       9 <split [29/3]> Fold09
      10 <split [29/3]> Fold10

# grouping -- bad args

    Code
      group_vfold_cv(warpbreaks, group = warpbreaks$tension)
    Condition
      Error in `validate_group()`:
      ! Can't select columns that don't exist.
      x Columns `L`, `L`, `L`, `L`, `L`, etc. don't exist.

---

    Code
      group_vfold_cv(warpbreaks, group = c("tension", "wool"))
    Condition
      Error in `group_vfold_cv()`:
      ! `group` must be a single string, not a character vector.

---

    Code
      group_vfold_cv(warpbreaks, group = "tensio")
    Condition
      Error in `validate_group()`:
      ! Can't select columns that don't exist.
      x Column `tensio` doesn't exist.

---

    Code
      group_vfold_cv(warpbreaks)
    Condition
      Error in `group_vfold_cv()`:
      ! `group` must be a single string, not `NULL`.

---

    Code
      group_vfold_cv(warpbreaks, group = "tension", v = 10)
    Condition
      Error in `group_vfold_cv()`:
      ! The number of groups is less than `v` = 10.

---

    Code
      group_vfold_cv(dat1, c, v = 4, repeats = 4)
    Condition
      Error in `group_vfold_cv()`:
      ! Repeated resampling when `v` is 4 would create identical resamples.

---

    Code
      group_vfold_cv(dat1, c, repeats = 4)
    Condition
      Error in `group_vfold_cv()`:
      ! Repeated resampling when `v` is "NULL" would create identical resamples.

---

    Code
      group_vfold_cv(Orange, v = 1, group = "Tree")
    Condition
      Error in `group_vfold_cv()`:
      ! `v` must be a whole number larger than or equal to 2, not the number 1.

# grouping -- other balance methods

    Code
      rs1
    Output
      # Group 5-fold cross-validation 
      # A tibble: 5 x 2
        splits             id       
        <list>             <chr>    
      1 <split [2364/566]> Resample1
      2 <split [2371/559]> Resample2
      3 <split [2360/570]> Resample3
      4 <split [2278/652]> Resample4
      5 <split [2347/583]> Resample5

# grouping -- strata

    Code
      sizes4
    Output
      # A tibble: 5 x 5
        analysis assessment      n     p id       
           <int>      <int>  <int> <int> <chr>    
      1    80004      19996 100000     3 Resample1
      2    79850      20150 100000     3 Resample2
      3    79912      20088 100000     3 Resample3
      4    80131      19869 100000     3 Resample4
      5    80103      19897 100000     3 Resample5

---

    Code
      group_vfold_cv(sample_data, group, strata = outcome)
    Condition
      Warning in `group_vfold_cv()`:
      Leaving `v = NULL` while using stratification will set `v` to the number of groups present in the least common stratum.
      i Set `v` explicitly to override this warning.
    Output
      # Group 30-fold cross-validation 
      # A tibble: 30 x 2
         splits               id        
         <list>               <chr>     
       1 <split [96070/3930]> Resample01
       2 <split [95898/4102]> Resample02
       3 <split [96079/3921]> Resample03
       4 <split [96008/3992]> Resample04
       5 <split [95982/4018]> Resample05
       6 <split [95955/4045]> Resample06
       7 <split [96025/3975]> Resample07
       8 <split [96053/3947]> Resample08
       9 <split [96030/3970]> Resample09
      10 <split [96069/3931]> Resample10
      # i 20 more rows

---

    Code
      sizes5
    Output
      # A tibble: 5 x 5
        analysis assessment      n     p id       
           <int>      <int>  <int> <int> <chr>    
      1    80096      19904 100000     3 Resample1
      2    79962      20038 100000     3 Resample2
      3    79928      20072 100000     3 Resample3
      4    80058      19942 100000     3 Resample4
      5    79956      20044 100000     3 Resample5

---

    Code
      group_vfold_cv(sample_data, group, strata = outcome)
    Condition
      Warning in `group_vfold_cv()`:
      Leaving `v = NULL` while using stratification will set `v` to the number of groups present in the least common stratum.
      i Set `v` explicitly to override this warning.
    Output
      # Group 30-fold cross-validation 
      # A tibble: 30 x 2
         splits               id        
         <list>               <chr>     
       1 <split [95985/4015]> Resample01
       2 <split [95983/4017]> Resample02
       3 <split [96052/3948]> Resample03
       4 <split [95867/4133]> Resample04
       5 <split [96056/3944]> Resample05
       6 <split [95956/4044]> Resample06
       7 <split [95975/4025]> Resample07
       8 <split [96062/3938]> Resample08
       9 <split [95932/4068]> Resample09
      10 <split [96051/3949]> Resample10
      # i 20 more rows

# grouping fails for strata not constant across group members

    Code
      group_vfold_cv(sample_data, group, v = 5, strata = outcome)
    Condition
      Error in `group_vfold_cv()`:
      ! strata must be constant across all members of each group.

# grouping -- printing

    Code
      group_vfold_cv(warpbreaks, "tension")
    Output
      # Group 3-fold cross-validation 
      # A tibble: 3 x 2
        splits          id       
        <list>          <chr>    
      1 <split [36/18]> Resample1
      2 <split [36/18]> Resample2
      3 <split [36/18]> Resample3

# grouping -- printing with ...

    Code
      print(group_vfold_cv(warpbreaks, "tension"), n = 2)
    Output
      # Group 3-fold cross-validation 
      # A tibble: 3 x 2
        splits          id       
        <list>          <chr>    
      1 <split [36/18]> Resample1
      2 <split [36/18]> Resample2
      # i 1 more row

# bad args

    Code
      clustering_cv(dat1)
    Condition
      Error in `clustering_cv()`:
      ! `vars` is required and must contain at least one variable in `data`.

---

    Code
      clustering_cv(iris, Sepal.Length, v = -500)
    Condition
      Error in `clustering_cv()`:
      ! `v` must be a whole number larger than or equal to 2, not the number -500.

---

    Code
      clustering_cv(iris, Sepal.Length, v = 500)
    Condition
      Error in `clustering_cv()`:
      ! The number of rows is less than `v` = 500.

---

    Code
      clustering_cv(iris, Sepal.Length, cluster_function = "not an option")
    Condition
      Error in `clustering_cv()`:
      ! `cluster_function` must be one of "kmeans" or "hclust", not "not an option".

---

    Code
      clustering_cv(Orange, v = 1, vars = "Tree")
    Condition
      Error in `clustering_cv()`:
      ! `v` must be a whole number larger than or equal to 2, not the number 1.

---

    Code
      clustering_cv(Orange, repeats = 0)
    Condition
      Error in `clustering_cv()`:
      ! `repeats` must be a whole number larger than or equal to 1, not the number 0.

---

    Code
      clustering_cv(Orange, repeats = NULL)
    Condition
      Error in `clustering_cv()`:
      ! `repeats` must be a whole number, not `NULL`.

---

    Code
      clustering_cv(mtcars, mpg, v = nrow(mtcars))
    Condition
      Error in `clustering_cv()`:
      ! Leave-one-out cross-validation is not supported by this function.
      x You set `v` to `nrow(data)`, which would result in a leave-one-out cross-validation.
      i Use `loo_cv()` in this case.

# printing

    Code
      clustering_cv(dat1, c, v = 2)
    Output
      # 2-cluster cross-validation 
      # A tibble: 2 x 2
        splits         id   
        <list>         <chr>
      1 <split [5/15]> Fold1
      2 <split [15/5]> Fold2

# Upper & lower confidence interval does not contain NA

    Code
      int_pctl(bt_resamples, res)
    Condition
      Warning in `int_pctl()`:
      Recommend at least 1000 non-missing bootstrap resamples for term `mean`.
      Error in `int_pctl()`:
      ! All statistics have missing values.

---

    Code
      int_t(bt_resamples, res)
    Condition
      Warning in `int_t()`:
      Recommend at least 500 non-missing bootstrap resamples for term `mean`.
      Error in `int_t()`:
      ! All statistics have missing values.

---

    Code
      int_bca(bt_resamples, res, .fn = bad_stats)
    Condition
      Warning in `int_bca()`:
      Recommend at least 1000 non-missing bootstrap resamples for term `mean`.
      Error in `int_bca()`:
      ! All statistics have missing values.

# Sufficient replications needed to sufficiently reduce Monte Carlo sampling Error for BCa method

    Code
      int_pctl(bt_small, stats)
    Condition
      Warning in `int_pctl()`:
      Recommend at least 1000 non-missing bootstrap resamples for term `mean`.
    Output
      # A tibble: 1 x 6
        term  .lower .estimate .upper .alpha .method   
        <chr>  <dbl>     <dbl>  <dbl>  <dbl> <chr>     
      1 mean    9.97      10.0   10.1   0.05 percentile

---

    Code
      int_t(bt_small, stats)
    Condition
      Warning in `int_t()`:
      Recommend at least 500 non-missing bootstrap resamples for term `mean`.
    Output
      # A tibble: 1 x 6
        term  .lower .estimate .upper .alpha .method  
        <chr>  <dbl>     <dbl>  <dbl>  <dbl> <chr>    
      1 mean    9.96      10.0   10.1   0.05 student-t

# bad input

    Code
      int_pctl(bt_small, id)
    Condition
      Error in `int_pctl()`:
      ! `statistics` should select a list column of tidy results.

---

    Code
      int_pctl(bt_small, junk)
    Condition
      Error in `int_pctl()`:
      ! `statistics` should select a list column of tidy results.

---

    Code
      int_pctl(bt_small, stats, alpha = c(0.05, 0.2))
    Condition
      Error in `int_pctl()`:
      ! `alpha` must be a number, not a double vector.

---

    Code
      int_t(bt_small, stats, alpha = "potato")
    Condition
      Error in `int_t()`:
      ! `alpha` must be a number, not the string "potato".

---

    Code
      int_bca(bt_small, stats, alpha = 1:2, .fn = get_stats)
    Condition
      Error in `int_bca()`:
      ! `alpha` must be a number, not an integer vector.

---

    Code
      int_pctl(vfold_cv(mtcars))
    Condition
      Error in `int_pctl()`:
      ! No method for objects of classes: <vfold_cv/rset/tbl_df/tbl/data.frame>

---

    Code
      int_t(vfold_cv(mtcars))
    Condition
      Error in `int_t()`:
      ! No method for objects of classes: <vfold_cv/rset/tbl_df/tbl/data.frame>

---

    Code
      int_bca(vfold_cv(mtcars))
    Condition
      Error in `int_bca()`:
      ! No method for objects of classes: <vfold_cv/rset/tbl_df/tbl/data.frame>

---

    Code
      int_t(bad_bt_norm, stats)
    Condition
      Error in `int_t()`:
      ! `statistics` should select a single column for the standard error.

---

    Code
      int_bca(bt_norm, stats, .fn = no_dots)
    Condition
      Error in `int_bca()`:
      ! `.fn` must have an argument `...`.

---

    Code
      int_pctl(as.data.frame(bt_norm), stats)
    Condition
      Error in `int_pctl()`:
      ! No method for objects of class: <data.frame>

---

    Code
      int_t(as.data.frame(bt_norm), stats)
    Condition
      Error in `int_t()`:
      ! No method for objects of class: <data.frame>

---

    Code
      int_bca(as.data.frame(bt_norm), stats, .fn = get_stats)
    Condition
      Error in `int_bca()`:
      ! No method for objects of class: <data.frame>

---

    Code
      int_pctl(badder_bt_norm, bad_term)
    Condition
      Error in `int_pctl()`:
      ! The tibble in `statistics` must have a column for 'term'.

---

    Code
      int_t(badder_bt_norm, bad_err)
    Condition
      Error in `int_t()`:
      ! `statistics` should select a single column for the standard error.

---

    Code
      int_bca(badder_bt_norm, bad_est, .fn = get_stats)
    Condition
      Error in `int_bca()`:
      ! The tibble in `statistics` must have a column for 'estimate'.

---

    Code
      int_pctl(badder_bt_norm, bad_num)
    Condition
      Error in `int_pctl()`:
      ! All statistics must be numeric.

# checks for apparent bootstrap

    Code
      int_t(rs_boot)
    Condition
      Error in `int_t()`:
      ! The bootstrap resamples must include an apparent sample.
      i Please set `apparent = TRUE` in the `bootstraps()` function.

---

    Code
      int_bca(rs_boot)
    Condition
      Error in `int_bca()`:
      ! The bootstrap resamples must include an apparent sample.
      i Please set `apparent = TRUE` in the `bootstraps()` function.

# checks input for statistics

    Code
      int_t(rs_boot_missing_term, stats)
    Condition
      Error in `int_t()`:
      ! The tibble in `statistics` must have a column for 'term'.

---

    Code
      int_t(rs_boot_missing_estimate, stats)
    Condition
      Error in `int_t()`:
      ! The tibble in `statistics` must have a column for 'estimate'.

---

    Code
      int_t(rs_boot_missing_std_err, stats)
    Condition
      Error in `int_t()`:
      ! `statistics` should select a single column for the standard error.

# no assessment set

    Code
      assessment(xx$splits[[1]])
    Condition
      Error in `as.data.frame()`:
      ! There is no assessment data set for an `rsplit` object with class <perm_split>.

# bad args

    Code
      permutations(mtcars)
    Condition
      Error in `permutations()`:
      ! You must specify at least one column to permute.

---

    Code
      permutations(mtcars, foo)
    Condition
      Error in `permutations()`:
      ! Can't select columns that don't exist.
      x Column `foo` doesn't exist.

---

    Code
      permutations(mtcars, start_with("z"))
    Condition
      Error in `permutations()`:
      i In argument: `start_with("z")`.
      Caused by error in `start_with()`:
      ! could not find function "start_with"

---

    Code
      permutations(mtcars, everything())
    Condition
      Error in `permutations()`:
      ! You have selected all columns to permute.
      i This effectively reorders the rows in the original data without changing the data structure.
      > Please select fewer columns to permute.

# printing

    Code
      permutations(mtcars, 1)
    Output
      # Permutation sampling 
      # Permuted columns: [mpg] 
      # A tibble: 25 x 2
         splits         id            
         <list>         <chr>         
       1 <split [32/0]> Permutations01
       2 <split [32/0]> Permutations02
       3 <split [32/0]> Permutations03
       4 <split [32/0]> Permutations04
       5 <split [32/0]> Permutations05
       6 <split [32/0]> Permutations06
       7 <split [32/0]> Permutations07
       8 <split [32/0]> Permutations08
       9 <split [32/0]> Permutations09
      10 <split [32/0]> Permutations10
      # i 15 more rows

# simple numerics

    Code
      str1b <- make_strata(x1, depth = 500)
    Condition
      Warning:
      The number of observations in each quantile is below the recommended threshold of 500.
      * Stratification will use 2 breaks instead.

# simple character

    Code
      str2a <- make_strata(x2, pool = 0.05)
    Condition
      Warning:
      Stratifying groups that make up 5% of the data may be statistically risky.
      i Consider increasing `pool` to at least 0.1.

---

    Code
      str2b <- make_strata(x2, pool = 0.05)
    Condition
      Warning:
      Stratifying groups that make up 5% of the data may be statistically risky.
      i Consider increasing `pool` to at least 0.1.

# bad data

    Code
      s0 <- make_strata(x3)
    Condition
      Warning:
      Too little data to stratify.
      * Resampling will be unstratified.

---

    Code
      s1 <- make_strata(x3, pool = 0.06)
    Condition
      Warning:
      Stratifying groups that make up 6% of the data may be statistically risky.
      i Consider increasing `pool` to at least 0.1.

---

    Code
      s2 <- make_strata(mtcars$mpg)
    Condition
      Warning:
      The number of observations in each quantile is below the recommended threshold of 20.
      * Stratification will use 1 breaks instead.
      Warning:
      Too little data to stratify.
      * Resampling will be unstratified.

---

    Code
      s3 <- make_strata(seq_len(50), breaks = -1)
    Condition
      Warning:
      The bins specified by `breaks` must be >=2.
      * Resampling will be unstratified.

# don't stratify on Surv objects

    Code
      check_strata("surv", df)
    Condition
      Error:
      ! strata cannot be a <Surv> object.
      i Use the time or event variable directly.

# bad args

    Code
      mc_cv(mtcars, prop = -1)
    Condition
      Error in `mc_cv()`:
      ! `prop` must be greater than 0.

---

    Code
      mc_cv(mtcars, prop = 1)
    Condition
      Error in `mc_cv()`:
      ! `prop` must be less than 1.

---

    Code
      mc_cv(warpbreaks, strata = warpbreaks$tension)
    Condition
      Error in `mc_cv()`:
      ! Can't select columns that don't exist.
      x Columns `L`, `L`, `L`, `L`, `L`, etc. don't exist.

---

    Code
      mc_cv(warpbreaks, strata = c("tension", "wool"))
    Condition
      Error in `mc_cv()`:
      ! `strata` must be a single string or `NULL`, not a character vector.

# printing

    Code
      mc_cv(warpbreaks)
    Output
      # Monte Carlo cross-validation (0.75/0.25) with 25 resamples 
      # A tibble: 25 x 2
         splits          id        
         <list>          <chr>     
       1 <split [40/14]> Resample01
       2 <split [40/14]> Resample02
       3 <split [40/14]> Resample03
       4 <split [40/14]> Resample04
       5 <split [40/14]> Resample05
       6 <split [40/14]> Resample06
       7 <split [40/14]> Resample07
       8 <split [40/14]> Resample08
       9 <split [40/14]> Resample09
      10 <split [40/14]> Resample10
      # i 15 more rows

# grouping - bad args

    Code
      group_mc_cv(warpbreaks, group = warpbreaks$tension)
    Condition
      Error in `validate_group()`:
      ! Can't select columns that don't exist.
      x Columns `L`, `L`, `L`, `L`, `L`, etc. don't exist.

---

    Code
      group_mc_cv(warpbreaks, group = c("tension", "wool"))
    Condition
      Error in `group_mc_cv()`:
      ! `group` must be a single string, not a character vector.

---

    Code
      group_mc_cv(warpbreaks, group = "tensio")
    Condition
      Error in `validate_group()`:
      ! Can't select columns that don't exist.
      x Column `tensio` doesn't exist.

---

    Code
      group_mc_cv(warpbreaks)
    Condition
      Error in `group_mc_cv()`:
      ! `group` must be a single string, not `NULL`.

---

    Code
      group_mc_cv(mtcars, group = "cyl", prop = 1)
    Condition
      Error in `group_mc_cv()`:
      ! `prop` must be less than 1.

---

    Code
      group_mc_cv(warpbreaks, group = "tension", balance = "groups")
    Condition
      Error in `group_mc_cv()`:
      ! `...` must be empty.
      x Problematic argument:
      * balance = "groups"

---

    Code
      group_mc_cv(warpbreaks, group = "tension", prop = 0.99)
    Condition
      Error in `group_mc_cv()`:
      ! Some assessment sets contained zero rows
      i Consider using a non-grouped resampling method

# grouping -- strata

    Code
      sizes4
    Output
      # A tibble: 5 x 5
        analysis assessment     n     p id       
           <int>      <int> <int> <int> <chr>    
      1    37939      12061 50000     3 Resample1
      2    37063      12937 50000     3 Resample2
      3    37178      12822 50000     3 Resample3
      4    37950      12050 50000     3 Resample4
      5    37585      12415 50000     3 Resample5

# grouping - printing

    Code
      group_mc_cv(warpbreaks, "tension")
    Output
      # Group Monte Carlo cross-validation (0.75/0.25) with 25 resamples  
      # A tibble: 25 x 2
         splits          id        
         <list>          <chr>     
       1 <split [36/18]> Resample01
       2 <split [36/18]> Resample02
       3 <split [36/18]> Resample03
       4 <split [36/18]> Resample04
       5 <split [36/18]> Resample05
       6 <split [36/18]> Resample06
       7 <split [36/18]> Resample07
       8 <split [36/18]> Resample08
       9 <split [36/18]> Resample09
      10 <split [36/18]> Resample10
      # i 15 more rows

# grouping - printing with ...

    Code
      print(group_mc_cv(warpbreaks, "tension"), n = 2)
    Output
      # Group Monte Carlo cross-validation (0.75/0.25) with 25 resamples  
      # A tibble: 25 x 2
        splits          id        
        <list>          <chr>     
      1 <split [36/18]> Resample01
      2 <split [36/18]> Resample02
      # i 23 more rows

# vec_rbind() returns a bare tibble

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `inner_resamples` -> `inner_resamples...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `inner_resamples` -> `inner_resamples...6`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

# vec_cbind() returns a bare tibble

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `inner_resamples` -> `inner_resamples...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `inner_resamples` -> `inner_resamples...6`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `inner_resamples` -> `inner_resamples...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `inner_resamples` -> `inner_resamples...6`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `inner_resamples` -> `inner_resamples...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `inner_resamples` -> `inner_resamples...6`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `inner_resamples` -> `inner_resamples...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `inner_resamples` -> `inner_resamples...6`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `inner_resamples` -> `inner_resamples...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `inner_resamples` -> `inner_resamples...6`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `id2` -> `id2...3`
      * `splits` -> `splits...4`
      * `id` -> `id...5`
      * `id2` -> `id2...6`

---

    Code
      expect_identical(vec_cbind(x, x), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_identical(vec_cbind(x, tbl), vec_cbind(tbl, tbl))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

---

    Code
      expect_s3_class_bare_tibble(vec_cbind(x, x))
    Message
      New names:
      * `splits` -> `splits...1`
      * `id` -> `id...2`
      * `splits` -> `splits...3`
      * `id` -> `id...4`

# regression intervals

    Code
      skip_if(new_rng_snapshots)
      set.seed(123)
      int_2 <- reg_intervals(mpg ~ disp + wt, data = mtcars, filter = term == "wt",
      model_fn = "glm", keep_reps = TRUE)
      int_2
    Output
      # A tibble: 1 x 7
        term  .lower .estimate .upper .alpha .method          .replicates
        <chr>  <dbl>     <dbl>  <dbl>  <dbl> <chr>     <list<tibble[,2]>>
      1 wt     -5.62     -3.46 -0.955   0.05 student-t        [1,001 x 2]

---

    Code
      reg_intervals(mpg ~ disp + wt, data = mtcars, model_fn = "potato")
    Condition
      Error in `reg_intervals()`:
      ! `model_fn` must be one of "lm", "glm", "survreg", or "coxph", not "potato".

---

    Code
      reg_intervals(mpg ~ disp + wt, data = mtcars, type = "random")
    Condition
      Error in `reg_intervals()`:
      ! `type` must be one of "student-t" or "percentile", not "random".

---

    Code
      reg_intervals(mpg ~ disp + wt, data = mtcars, alpha = "a")
    Condition
      Error in `reg_intervals()`:
      ! `alpha` must be a single numeric value.

# grouping -- strata

    Code
      sizes4
    Output
      # A tibble: 5 x 5
        analysis assessment     n     p id        
           <int>      <int> <int> <int> <chr>     
      1    49669      18552 50000     3 Bootstrap1
      2    49943      17429 50000     3 Bootstrap2
      3    50133      16350 50000     3 Bootstrap3
      4    50314      18071 50000     3 Bootstrap4
      5    50253      17830 50000     3 Bootstrap5

# bad args

    Code
      bootstraps(warpbreaks, strata = warpbreaks$tension)
    Condition
      Error in `bootstraps()`:
      ! Can't select columns that don't exist.
      x Columns `L`, `L`, `L`, `L`, `L`, etc. don't exist.

---

    Code
      bootstraps(warpbreaks, strata = c("tension", "wool"))
    Condition
      Error in `bootstraps()`:
      ! `strata` must be a single string or `NULL`, not a character vector.

---

    Code
      group_bootstraps(warpbreaks, tension)
    Condition
      Warning in `group_bootstraps()`:
      Some assessment sets contained zero rows.
      i Consider using a non-grouped resampling method.
    Output
      # Group bootstrap sampling 
      # A tibble: 25 x 2
         splits          id         
         <list>          <chr>      
       1 <split [54/18]> Bootstrap01
       2 <split [54/36]> Bootstrap02
       3 <split [54/18]> Bootstrap03
       4 <split [54/18]> Bootstrap04
       5 <split [54/36]> Bootstrap05
       6 <split [54/18]> Bootstrap06
       7 <split [54/18]> Bootstrap07
       8 <split [54/18]> Bootstrap08
       9 <split [54/18]> Bootstrap09
      10 <split [54/18]> Bootstrap10
      # i 15 more rows

---

    Code
      bootstraps(mtcars[2, ])
    Condition
      Warning in `bootstraps()`:
      Some assessment sets contained zero rows.
    Output
      # Bootstrap sampling 
      # A tibble: 25 x 2
         splits        id         
         <list>        <chr>      
       1 <split [1/0]> Bootstrap01
       2 <split [1/0]> Bootstrap02
       3 <split [1/0]> Bootstrap03
       4 <split [1/0]> Bootstrap04
       5 <split [1/0]> Bootstrap05
       6 <split [1/0]> Bootstrap06
       7 <split [1/0]> Bootstrap07
       8 <split [1/0]> Bootstrap08
       9 <split [1/0]> Bootstrap09
      10 <split [1/0]> Bootstrap10
      # i 15 more rows

# printing

    Code
      set.seed(11)
      bootstraps(warpbreaks)
    Output
      # Bootstrap sampling 
      # A tibble: 25 x 2
         splits          id         
         <list>          <chr>      
       1 <split [54/20]> Bootstrap01
       2 <split [54/22]> Bootstrap02
       3 <split [54/17]> Bootstrap03
       4 <split [54/22]> Bootstrap04
       5 <split [54/17]> Bootstrap05
       6 <split [54/19]> Bootstrap06
       7 <split [54/18]> Bootstrap07
       8 <split [54/17]> Bootstrap08
       9 <split [54/23]> Bootstrap09
      10 <split [54/22]> Bootstrap10
      # i 15 more rows

# bad inputs

    Code
      rsplit(as.list(dat1), 1:2, 4:5)
    Condition
      Error in `rsplit()`:
      ! `data` must be a data frame.

---

    Code
      rsplit(dat1, letters[1:2], 4:5)
    Condition
      Error in `rsplit()`:
      ! `in_id` must be a positive integer vector.

---

    Code
      rsplit(as.list(dat1), 1:2, letters[4:5])
    Condition
      Error in `rsplit()`:
      ! `data` must be a data frame.

---

    Code
      rsplit(as.list(dat1), -1:2, 4:5)
    Condition
      Error in `rsplit()`:
      ! `data` must be a data frame.

---

    Code
      rsplit(as.list(dat1), 1:2, -4:5)
    Condition
      Error in `rsplit()`:
      ! `data` must be a data frame.

---

    Code
      rsplit(as.list(dat1), integer(0), 4:5)
    Condition
      Error in `rsplit()`:
      ! `data` must be a data frame.

# print methods

    Code
      set.seed(233)
      vfold_cv(mtcars)$splits[[1]]
    Output
      <Analysis/Assess/Total>
      <28/4/32>

---

    Code
      set.seed(233)
      validation_split(mtcars)$splits[[1]]
    Output
      <Training/Validation/Total>
      <24/8/32>

---

    Code
      set.seed(233)
      validation_split(mtcars)
    Output
      # Validation Set Split (0.75/0.25)  
      # A tibble: 1 x 2
        splits         id        
        <list>         <chr>     
      1 <split [24/8]> validation

# `complement()` error messages

    Code
      complement(fake_rsplit)
    Condition
      Error in `complement()`:
      ! No `complement()` method for objects of class: <not_an_rsplit>

---

    Code
      complement(fake_rsplit)
    Condition
      Error in `complement()`:
      ! No `complement()` method for objects of classes: <not_an_rsplit/really_not_an_rsplit>

---

    Code
      get_stored_out_id(list(out_id = NA))
    Condition
      Error in `get_stored_out_id()`:
      ! Cannot derive the assessment set for this type of resampling with class: <list>.

# `lag` arg to `initial_time_split()` is deprecated

    Code
      initial_time_split(mtcars, lag = 2)
    Condition
      Warning:
      The `lag` argument of `initial_time_split()` is deprecated as of rsample 1.3.1.9000.
      i Please lag your predictors prior to splitting the dataset.
    Output
      <Training/Testing/Total>
      <24/10/32>

# `initial_time_split()` error messages

    Code
      initial_time_split(drinks, prop = 2)
    Condition
      Error in `initial_time_split()`:
      ! `prop` must be less than 1.

---

    Code
      initial_time_split(drinks, lag = 12.5)
    Condition
      Error in `initial_time_split()`:
      ! `lag` must be a whole number.

---

    Code
      initial_time_split(drinks, lag = nrow(drinks) + 1)
    Condition
      Error in `initial_time_split()`:
      ! `lag` must be less than or equal to the number of training observations.

# printing initial split objects

    Code
      initial_split(mtcars)
    Output
      <Training/Testing/Total>
      <24/8/32>

---

    Code
      initial_time_split(mtcars)
    Output
      <Training/Testing/Total>
      <24/8/32>

# prop is checked

    Code
      initial_split(mtcars, prop = 1)
    Condition
      Error in `initial_split()`:
      ! `prop` must be less than 1.

---

    Code
      initial_time_split(mtcars, prop = 1)
    Condition
      Error in `initial_time_split()`:
      ! `prop` must be less than 1.

---

    Code
      group_initial_split(mtcars, group = "cyl", prop = 1)
    Condition
      Error in `group_initial_split()`:
      ! `prop` must be less than 1.

---
title: "Introduction to rsample"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Introduction to rsample}
output:
  knitr:::html_vignette:
    toc: yes
---

```{r ex_setup, include=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  digits = 3,
  collapse = TRUE,
  comment = "#>"
  )
options(digits = 3)
```

## Terminology

We define a _resample_ as the result of a two-way split of a data set. For example, when bootstrapping, one part of the resample is a sample with replacement of the original data. The other part of the split contains the instances that were not contained in the bootstrap sample. Cross-validation is another type of resampling. 

## `rset` Objects Contain Many Resamples

The main class in the package (`rset`) is for a _set_ or _collection_ of resamples. In 10-fold cross-validation, the set would consist of the 10 different resamples of the original data. 

Like [`modelr`](https://cran.r-project.org/package=modelr), the resamples are stored in data-frame-like `tibble` object. As a simple example, here is a small set of bootstraps of the `mtcars` data:

```{r mtcars_bt, message=FALSE}
library(rsample)
set.seed(8584)
bt_resamples <- bootstraps(mtcars, times = 3)
bt_resamples
```


## Individual Resamples are `rsplit` Objects

The resamples are stored in the `splits` column in an object that has class `rsplit`. 

In this package we use the following terminology for the two partitions that comprise a resample: 

 * The _analysis_ data are those that we selected in the resample. For a bootstrap, this is the sample with replacement. For 10-fold cross-validation, this is the 90% of the data. These data are often used to fit a model or calculate a statistic in traditional bootstrapping. 
 * The _assessment_ data are usually the section of the original data not covered by the analysis set. Again, in 10-fold CV, this is the 10% held out. These data are often used to evaluate the performance of a model that was fit to the analysis data. 

(Aside: While some might use the term "training" and "testing" for these data sets, we avoid them since those labels often conflict with the data that result from an initial partition of the data that is typically done _before_ resampling. The training/test split can be conducted using the `initial_split` function in this package.)

Let's look at one of the `rsplit` objects
```{r rsplit}
first_resample <- bt_resamples$splits[[1]]
first_resample
```
This indicates that there were `r dim(bt_resamples$splits[[1]])["analysis"]` data points in the analysis set, `r dim(bt_resamples$splits[[1]])["assessment"]` instances were in the assessment set, and that the original data contained `r dim(bt_resamples$splits[[1]])["n"]` data points. These results can also be determined using the `dim` function on an `rsplit` object. 

To obtain either of these data sets from an `rsplit`, the `as.data.frame` function can be used. By default, the analysis set is returned but the `data` option can be used to return the assessment data: 
```{r rsplit_df}
head(as.data.frame(first_resample))
as.data.frame(first_resample, data = "assessment")
```
Alternatively, you can use the shortcuts  `analysis(first_resample)` and `assessment(first_resample)`.

# accessor functions for `val_split`

    Code
      testing(val_split)
    Condition
      Error in `testing()`:
      ! The testing data is not part of the validation set object.
      i It is part of the result of the initial 3-way split, e.g., with `initial_validation_split()`.

# basic split - accessor functions

    Code
      analysis(val_split)
    Condition
      Error in `analysis()`:
      ! The initial validation split does not contain an analysis set.
      i You can access the training data with `training()`.

---

    Code
      assessment(val_split)
    Condition
      Error in `assessment()`:
      ! The initial validation split does not contain an assessment set.
      i You can access the testing data with `testing()`.

# basic split stratified

    Code
      initial_validation_split(dat, strata = does_not_exist)
    Condition
      Error in `initial_validation_split()`:
      ! Can't select columns that don't exist.
      x Column `does_not_exist` doesn't exist.

---

    Code
      initial_validation_split(dat, strata = c(x, f))
    Condition
      Error in `initial_validation_split()`:
      ! Can't select columns that don't exist.
      x Column `x` doesn't exist.

---

    Code
      initial_validation_split(dat, strata = rep(1:3, times = c(50, 25, 25)))
    Condition
      Error in `initial_validation_split()`:
      ! Can't select columns past the end.
      i Location 3 doesn't exist.
      i There are only 2 columns.

# grouped split - accessor functions

    Code
      analysis(val_split)
    Condition
      Error in `analysis()`:
      ! The initial validation split does not contain an analysis set.
      i You can access the training data with `training()`.

---

    Code
      assessment(val_split)
    Condition
      Error in `assessment()`:
      ! The initial validation split does not contain an assessment set.
      i You can access the testing data with `testing()`.

# check_prop_3() works

    Code
      check_prop_3(0.3)
    Condition
      Error:
      ! `prop` needs to contain the proportions for training and validation.

---

    Code
      check_prop_3("zero")
    Condition
      Error:
      ! `prop` needs to be numeric.

---

    Code
      check_prop_3(NULL)
    Condition
      Error:
      ! `prop` needs to be numeric.

---

    Code
      check_prop_3(NA)
    Condition
      Error:
      ! `prop` needs to be numeric.

---

    Code
      check_prop_3(0)
    Condition
      Error:
      ! `prop` needs to contain the proportions for training and validation.

---

    Code
      check_prop_3(c(0.3, NA))
    Condition
      Error:
      ! `prop` cannot contain `NA`.

---

    Code
      check_prop_3(c(0.3, NULL))
    Condition
      Error:
      ! `prop` needs to contain the proportions for training and validation.

---

    Code
      check_prop_3(c(0.3, 1))
    Condition
      Error:
      ! Elements of `prop` need to be in (0, 1).

---

    Code
      check_prop_3(c(0.3, 0.7))
    Condition
      Error:
      ! The sum of the proportions in `prop` needs to be in (0, 1).

---
title: "Grid Search Tuning of Keras Models"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Grid Search Tuning of Keras Models}
output:
  knitr:::html_vignette:
    toc: yes
---

Since `keras` models can be tuning using the `tune` package, this vignette is deprecated. The last version of it can be found [here](https://github.com/tidymodels/rsample/blob/8333b7bb441e0097418205dd735bd3f0cc418916/vignettes/Applications/Keras.Rmd). 

# bad args

    Code
      skip_if(new_rng_snapshots)
      set.seed(123)
      nested_cv(mtcars, outside = bootstraps(times = 5), inside = vfold_cv(v = 3))
    Condition
      Warning:
      Using bootstrapping as the outer resample is dangerous since the inner resample might have the same data point in both the analysis and assessment set.
    Output
      # Nested resampling:
      #  outer: Bootstrap sampling
      #  inner: 3-fold cross-validation
      # A tibble: 5 x 3
        splits          id         inner_resamples
        <list>          <chr>      <list>         
      1 <split [32/11]> Bootstrap1 <vfold [3 x 2]>
      2 <split [32/9]>  Bootstrap2 <vfold [3 x 2]>
      3 <split [32/10]> Bootstrap3 <vfold [3 x 2]>
      4 <split [32/14]> Bootstrap4 <vfold [3 x 2]>
      5 <split [32/11]> Bootstrap5 <vfold [3 x 2]>

---

    Code
      nested_cv(mtcars, outside = vfold_cv(), inside = folds)
    Condition
      Error in `nested_cv()`:
      ! `inside` should be a expression such as `vfold()` or `bootstraps(times = 10)` instead of an existing object.

# printing

    Code
      rs1
    Output
      # Nested resampling:
      #  outer: 10-fold cross-validation
      #  inner: 3-fold cross-validation
      # A tibble: 10 x 3
         splits         id     inner_resamples
         <list>         <chr>  <list>         
       1 <split [27/3]> Fold01 <vfold [3 x 2]>
       2 <split [27/3]> Fold02 <vfold [3 x 2]>
       3 <split [27/3]> Fold03 <vfold [3 x 2]>
       4 <split [27/3]> Fold04 <vfold [3 x 2]>
       5 <split [27/3]> Fold05 <vfold [3 x 2]>
       6 <split [27/3]> Fold06 <vfold [3 x 2]>
       7 <split [27/3]> Fold07 <vfold [3 x 2]>
       8 <split [27/3]> Fold08 <vfold [3 x 2]>
       9 <split [27/3]> Fold09 <vfold [3 x 2]>
      10 <split [27/3]> Fold10 <vfold [3 x 2]>

---
title: "Nested Resampling"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Nested Resampling}
output:
  knitr:::html_vignette:
    toc: yes
---


```{r setup, include=FALSE}
library(rsample)   
library(purrr)
library(dplyr)
library(ggplot2)
library(scales)
library(mlbench)
library(kernlab)
theme_set(theme_bw())
```


(A version of this article was originally published in the [_Applied Predictive Modeling_ blog](http://appliedpredictivemodeling.com/blog/2017/9/2/njdc83d01pzysvvlgik02t5qnaljnd))

A typical scheme for splitting the data when developing a predictive model is to create an initial split of the data into a training and test set. If resampling is used, it is executed on the training set. A series of binary splits is created. In `rsample`, we use the term _analysis set_ for the data that are used to fit the model and the _assessment set_ is used to compute performance: 

![](diagram.png)

A common method for tuning models is grid search where a candidate set of tuning parameters is created. The full set of models for every combination of the tuning parameter grid and the resamples is created. Each time, the assessment data are used to measure performance and the average value is determined for each tuning parameter. 

The potential problem is, once we pick the tuning parameter associated with the best performance, this performance value is usually quoted as the performance of the model. There is serious potential for _optimization bias_ since we uses the same data to tune the model and quote performance. This would result in an optimistic estimate of performance. 

Nested resampling does an additional layer of resampling that separates the tuning activities from the process used to estimate the efficacy of the model. An _outer_ resampling scheme is used and, for every split in the outer resample, another full set of resampling splits are created on the original analysis set. For example, if 10-fold cross-validation is used on the outside and 5-fold cross-validation on the inside, a total of 500 models will be fit. The parameter tuning will be conducted 10 times and the best parameters are determined from the average of the 5 assessment sets. This process occurs 10 times. 

Once the tuning results are complete, a model is fit to each of the outer resampling splits using the best parameter associated with that resample. The average of the outer method's assessment sets are a unbiased estimate of the model.  

We will simulate some regression data to illustrate the methods. The `mlbench` function `mlbench::mlbench.friedman1` can simulate a complex regression data structure from the [original MARS publication](https://scholar.google.com/scholar?hl=en&q=%22Multivariate+adaptive+regression+splines%22&btnG=&as_sdt=1%2C7&as_sdtp=). A training set size of 100 data points are generated as well as a large set that will be used to characterize how well the resampling procedure performed.  

```{r sim-data}
library(mlbench)
sim_data <- function(n) {
  tmp <- mlbench.friedman1(n, sd=1)
  tmp <- cbind(tmp$x, tmp$y)
  tmp <- as.data.frame(tmp)
  names(tmp)[ncol(tmp)] <- "y"
  tmp
}

set.seed(9815)
train_dat <- sim_data(100)
large_dat <- sim_data(10^5)
```

To get started, the types of resampling methods need to be specified. This isn't a large data set, so 5 repeats of 10-fold cross validation will be used as the _outer_ resampling method that will be used to generate the estimate of overall performance. To tune the model, it would be good to have precise estimates for each of the values of the tuning parameter so 25 iterations of the bootstrap will be used. This means that there will eventually be `5 * 10 * 25 = 1250` models that are fit to the data _per tuning parameter_. These will be discarded once the performance of the model has been quantified. 

To create the tibble with the resampling specifications: 

```{r tibble-gen}
library(rsample) 
results <- nested_cv(train_dat, 
                     outside = vfold_cv(repeats = 5), 
                     inside = bootstraps(times = 25))
results
```

The splitting information for each resample is contained in the `split` objects. Focusing on the second fold of the first repeat:

```{r split-example}
results$splits[[2]]
```

`<90/10/100>` indicates the number of data in the analysis set, assessment set, and the original data. 

Each element of `inner_resamples` has its own tibble with the bootstrapping splits. 

```{r inner-splits}
results$inner_resamples[[5]]
```

These are self-contained, meaning that the bootstrap sample is aware that it is a sample of a specific 90% of the data:

```{r inner-boot-split}
results$inner_resamples[[5]]$splits[[1]]
```

To start, we need to define how the model will be created and measured. For our example, a radial basis support vector machine model will be created using the function `kernlab::ksvm`. This model is generally thought of as having _two_ tuning parameters: the SVM cost value and the kernel parameter `sigma`. For illustration, only the cost value will be tuned and the function `kernlab::sigest` will be used to estimate `sigma` during each model fit. This is automatically done by `ksvm`. 

After the model is fit to the analysis set, the root-mean squared error (RMSE) is computed on the assessment set. One important note: for this model, it is critical to center and scale the predictors before computing dot products. We don't do this operation here because `mlbench.friedman1` simulates all of the predictors to be standard uniform random variables. 

Our function to fit the model and compute the RMSE is:

```{r rmse-func}
library(kernlab)

# `object` will be an `rsplit` object from our `results` tibble
# `cost` is the tuning parameter
svm_rmse <- function(object, cost = 1) {
  y_col <- ncol(object$data)
  mod <- ksvm(y ~ ., data = analysis(object),  C = cost)
  holdout_pred <- predict(mod, assessment(object)[-y_col])
  rmse <- sqrt(mean((assessment(object)$y - holdout_pred) ^ 2, na.rm = TRUE))
  rmse
}

# In some case, we want to parameterize the function over the tuning parameter:
rmse_wrapper <- function(cost, object) svm_rmse(object, cost)
```

For the nested resampling, a model needs to be fit for each tuning parameter and each bootstrap split. To do this, a wrapper can be created: 
```{r inner-tune-func}
library(purrr)
library(dplyr)

# `object` will be an `rsplit` object for the bootstrap samples
tune_over_cost <- function(object) {
  results <- tibble(cost = 2 ^ seq(-2, 8, by = 1))
  results$RMSE <- map_dbl(results$cost, 
                          rmse_wrapper,
                          object = object)
  results
}
```

Since this will be called across the set of outer cross-validation splits, another wrapper is required: 
```{r inner-func}
# `object` is an `rsplit` object in `results$inner_resamples` 
summarize_tune_results <- function(object) {
  # Return row-bound tibble that has the 25 bootstrap results
  map_df(object$splits, tune_over_cost) %>%
    # For each value of the tuning parameter, compute the 
    # average RMSE which is the inner bootstrap estimate. 
    group_by(cost) %>%
    summarize(mean_RMSE = mean(RMSE, na.rm = TRUE),
              n = length(RMSE))
}
```

Now that those functions are defined, we can execute all the inner resampling loops:

```{r inner-runs}
tuning_results <- map(results$inner_resamples, summarize_tune_results) 
```

`tuning_results` is a list of data frames for each of the 50 outer resamples. 

Let's make a plot of the averaged results to see what the relationship is between the RMSE and the tuning parameters for each of the inner bootstrapping operations:

```{r rmse-plot, fig.height=4}
library(ggplot2)
library(scales)

pooled_inner <- tuning_results %>% bind_rows

best_cost <- function(dat) dat[which.min(dat$mean_RMSE),]

p <- ggplot(pooled_inner, aes(x = cost, y = mean_RMSE)) + 
  scale_x_continuous(trans = 'log2') +
  xlab("SVM Cost") + ylab("Inner RMSE")

for (i in 1:length(tuning_results))
  p <- p  +
  geom_line(data = tuning_results[[i]], alpha = .2) +
  geom_point(data = best_cost(tuning_results[[i]]), pch = 16)

p <- p + geom_smooth(data = pooled_inner, se = FALSE)
p
```

Each grey line is a separate bootstrap resampling curve created from a different 90% of the data. The blue line is a loess smooth of all the results pooled together. 

To determine the best parameter estimate for each of the outer resampling iterations:

```{r choose, fig.height=4}
cost_vals <- tuning_results %>% map_df(best_cost) %>% select(cost) 
results <- bind_cols(results, cost_vals)
results$cost <- factor(results$cost, levels = paste(2 ^ seq(-2, 8, by = 1)))

ggplot(results, aes(x = cost)) + 
  geom_bar() + 
  xlab("SVM Cost") + 
  scale_x_discrete(drop = FALSE)
```

Most of the resamples produced and optimal cost values of 2.0 but the distribution is right-skewed due to the flat trend in the resampling profile once the cost value becomes 10 or larger.

Now that we have these estimates, we can compute the outer resampling results for each of the `r nrow(results)` splits using the corresponding tuning parameter value:

```{r run-outer}
results$RMSE <- map2_dbl(results$splits, results$cost, svm_rmse)
summary(results$RMSE)
```

The estimated RMSE for the model tuning process is `r round(mean(results$RMSE), 2)`. 

What is the RMSE estimate for the non-nested procedure when only the outer resampling method is used? For each cost value in the tuning grid, `r nrow(results)` SVM models are fit and their RMSE values are averaged. The table of cost values and mean RMSE estimates is used to determine the best cost value. The associated RMSE is the biased estimate. 

```{r not-nested, fig.height=4}
not_nested <- map(results$splits, tune_over_cost) %>%
  bind_rows

outer_summary <- not_nested %>% 
  group_by(cost) %>% 
  summarize(outer_RMSE = mean(RMSE),
            n = length(RMSE))
outer_summary

ggplot(outer_summary, aes(x = cost, y = outer_RMSE)) + 
  geom_point() + 
  geom_line() + 
  scale_x_continuous(trans = 'log2') +
  xlab("SVM Cost") + ylab("RMSE")
```

The non-nested procedure estimates the RMSE to be `r round(min(outer_summary$outer_RMSE), 2)`. Both estimates are fairly close. 

The approximately true RMSE for an SVM model with a cost value of 2.0 and be approximated with the large sample that was simulated at the beginning. 

```{r large-sample-estimate}
finalModel <- ksvm(y ~ ., data = train_dat, C = 2)
large_pred <- predict(finalModel, large_dat[, -ncol(large_dat)])
sqrt(mean((large_dat$y - large_pred) ^ 2, na.rm = TRUE))
```

The nested procedure produces a closer estimate to the approximate truth but the non-nested estimate is very similar.

# dots

    Code
      form_pred(y ~ .)
    Condition
      Error in `terms.formula()`:
      ! '.' in formula and no 'data' argument

---

    Code
      form_pred(terms(y ~ .))
    Condition
      Error in `terms.formula()`:
      ! '.' in formula and no 'data' argument

---

    Code
      form_pred(y ~ (.)^2)
    Condition
      Error in `terms.formula()`:
      ! '.' in formula and no 'data' argument

---

    Code
      form_pred(terms(y ~ (.)^2))
    Condition
      Error in `terms.formula()`:
      ! '.' in formula and no 'data' argument

# nested cv

    Code
      labels(nested_cv(mtcars, outside = vfold_cv(v = 3), inside = bootstraps(times = 5)))
    Condition
      Error in `labels()`:
      ! `labels` not implemented for nested resampling

# adding labels

    Code
      add_resample_id(analysis(car_folds$splits[[1]]), car_folds$splits[[1]], 7)
    Condition
      Error in `add_resample_id()`:
      ! `dots` should be a single logical.

---

    Code
      add_resample_id(analysis(car_folds$splits[[1]]), car_folds$splits[[1]], c(TRUE,
        TRUE))
    Condition
      Error in `add_resample_id()`:
      ! `dots` should be a single logical.

---

    Code
      add_resample_id(analysis(car_folds$splits[[1]]), car_folds$splits)
    Condition
      Error in `add_resample_id()`:
      ! `split` should be a single <rset> object.

---

    Code
      add_resample_id(as.matrix(analysis(car_folds$splits[[1]])), car_folds$splits[[
        1]])
    Condition
      Error in `add_resample_id()`:
      ! `.data` should be a data frame.

# `validation_split()` is deprecated

    Code
      set.seed(11)
      rs1 <- validation_split(dat1)
    Condition
      Warning:
      `validation_split()` was deprecated in rsample 1.2.0.
      i Please use `initial_validation_split()` instead.
    Code
      sizes1 <- dim_rset(rs1)
      expect_true(all(sizes1$analysis == 15))
      expect_true(all(sizes1$assessment == 5))

# `validation_time_split()` is deprecated

    Code
      set.seed(11)
      rs1 <- validation_time_split(dat1)
    Condition
      Warning:
      `validation_time_split()` was deprecated in rsample 1.2.0.
      i Please use `initial_validation_time_split()` instead.
    Code
      sizes1 <- dim_rset(rs1)
      expect_true(all(sizes1$analysis == 15))
      expect_true(all(sizes1$assessment == 5))

# default time param with lag

    Code
      validation_time_split(drinks)
    Output
      # Validation Set Split (0.75/0.25)  
      # A tibble: 1 x 2
        splits           id        
        <list>           <chr>     
      1 <split [231/78]> validation

---

    Code
      validation_time_split(drinks, lag = 12.5)
    Condition
      Error in `validation_time_split()`:
      ! `lag` must be a whole number.

---

    Code
      validation_time_split(drinks, lag = 500)
    Condition
      Error in `validation_time_split()`:
      ! `lag` must be less than or equal to the number of training observations.

# `group_validation_split()` is deprecated

    Code
      set.seed(11)
      rs1 <- group_validation_split(dat1, c)
    Condition
      Warning:
      `group_validation_split()` was deprecated in rsample 1.2.0.
      i Please use `group_initial_validation_split()` instead.
    Code
      sizes1 <- dim_rset(rs1)
      expect_true(all(sizes1$analysis == 15))
      expect_true(all(sizes1$assessment == 5))

# grouping -- strata

    Code
      sizes4
    Output
      # A tibble: 1 x 5
        analysis assessment     n     p id        
           <int>      <int> <int> <int> <chr>     
      1    37074      12926 50000     3 validation

# bad args

    Code
      validation_split(mtcars, prop = 1)
    Condition
      Error in `validation_split()`:
      ! `prop` must be less than 1.

---

    Code
      validation_time_split(mtcars, prop = 1)
    Condition
      Error in `validation_time_split()`:
      ! `prop` must be less than 1.

---

    Code
      group_validation_split(mtcars, group = "cyl", prop = 1)
    Condition
      Error in `group_validation_split()`:
      ! `prop` must be less than 1.

---

    Code
      validation_split(warpbreaks, strata = warpbreaks$tension)
    Condition
      Error in `validation_split()`:
      ! Can't select columns that don't exist.
      x Columns `L`, `L`, `L`, `L`, `L`, etc. don't exist.

---

    Code
      validation_split(warpbreaks, strata = c("tension", "wool"))
    Condition
      Error in `validation_split()`:
      ! `strata` must be a single string or `NULL`, not a character vector.

# printing

    Code
      validation_split(warpbreaks)
    Output
      # Validation Set Split (0.75/0.25)  
      # A tibble: 1 x 2
        splits          id        
        <list>          <chr>     
      1 <split [40/14]> validation

---

    Code
      validation_split(warpbreaks)$splits[[1]]
    Output
      <Training/Validation/Total>
      <40/14/54>

---
title: "Survival Analysis Example"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Survival Analysis Example}
output:
  knitr:::html_vignette:
    toc: yes
---

```{r setup, include = FALSE}
options(digits = 3)
library(survival)
library(purrr)
library(rsample)
library(dplyr)
library(tidyposterior)
library(ggplot2)
library(tidyr)
```

In this article, a parametric analysis of censored data is conducted and `rsample` is used to measure the importance of predictors in the model. The data that will be used is the NCCTG lung cancer data contained in the `survival` package:

```{r lung}
library(survival)
str(lung)
```

`status` is an indicator for which patients are censored (`status = 1`) or an actual event (`status = 2`). The help file `?survreg` has the following model fit:

```{r example-model}
lung_mod <- survreg(Surv(time, status) ~ ph.ecog + age + strata(sex), data = lung)
summary(lung_mod)
```

Note that the stratification on gender only affects the scale parameter; the estimates above are from a log-linear model for the scale parameter even though they are listed with the regression variables for the other parameter. `coef` gives results that are more clear:

```{r coef}
coef(lung_mod)
```

To resample these data, it would be a good idea to try to maintain the same censoring rate across the splits. To do this, stratified resampling can be used where each analysis/assessment split is conducted within each value of the status indicator. To demonstrate, Monte Carlo resampling is used where 75% of the data are in the analysis set. A total of 100 splits are created.   

```{r splits}
library(rsample)
set.seed(9666)
mc_samp <- mc_cv(lung, strata = "status", times = 100)

library(purrr)
cens_rate <- function(x) mean(analysis(x)$status == 1)
summary(map_dbl(mc_samp$splits, cens_rate))
```

To demonstrate the use of resampling with censored data, the parametric model shown above will be fit with different variable sets to characterize how important each predictor is to the outcome. 

To do this, a set of formulas are created for the different variable sets:

```{r forms}
three_fact <- as.formula(Surv(time, status) ~ ph.ecog + age + strata(sex))
rm_ph.ecog <- as.formula(Surv(time, status) ~           age + strata(sex))
rm_age     <- as.formula(Surv(time, status) ~ ph.ecog +       strata(sex))
rm_sex     <- as.formula(Surv(time, status) ~ ph.ecog + age              )
```

The model fitting function will take the formula as an argument:

```{r fit-func}
mod_fit <- function(x, form, ...) 
  survreg(form, data = analysis(x), ...)
```

To calculate the efficacy of the model, the concordance statistic is used (see `?survConcordance`):

```{r concord}
get_concord <- function(split, mod, ...) {
  pred_dat <- assessment(split)
  pred_dat$pred <- predict(mod, newdata = pred_dat)
  concordance(Surv(time, status) ~ pred, pred_dat, ...)$concordance
}
```

With these functions, a series of models are created for each variable set.

```{r models}
mc_samp$mod_full    <- map(mc_samp$splits, mod_fit, form = three_fact)
mc_samp$mod_ph.ecog <- map(mc_samp$splits, mod_fit, form = rm_ph.ecog)
mc_samp$mod_age     <- map(mc_samp$splits, mod_fit, form = rm_age)
mc_samp$mod_sex     <- map(mc_samp$splits, mod_fit, form = rm_sex)
```

Similarly, the concordance values are computed for each model:

```{r concord-est}
mc_samp$full    <- map2_dbl(mc_samp$splits, mc_samp$mod_full, get_concord)
mc_samp$ph.ecog <- map2_dbl(mc_samp$splits, mc_samp$mod_ph.ecog, get_concord)
mc_samp$age     <- map2_dbl(mc_samp$splits, mc_samp$mod_age, get_concord)
mc_samp$sex     <- map2_dbl(mc_samp$splits, mc_samp$mod_sex, get_concord)
```

The distributions of the resampling estimates

```{r concord-df}
library(dplyr)
concord_est <- mc_samp %>%
  dplyr::select(-matches("^mod"))

library(tidyr)
library(ggplot2)
concord_est %>%
  gather() %>%
  ggplot(aes(x = statistic, col = model)) + 
  geom_line(stat = "density") + 
  theme_bw() + 
  theme(legend.position = "top")
```

It looks as though the model missing `ph.ecog` has larger concordance values than the other models. As one might expect, the full model and the model absent `sex` are very similar; the difference in these models should only be the scale parameters estimates. 

To more formally test this, the `tidyposterior` package is used to create a Bayesian model for the concordance statistics. 

```{r perf-mod, warning = FALSE, message = FALSE}
library(tidyposterior)
concord_est <- perf_mod(concord_est, seed = 6507, iter = 5000)
concord_est$stan
```

To summarize the posteriors for each model:

```{r post}
ggplot(tidy(concord_est)) + 
  theme_bw()
```

While this seems clear-cut, let's assume that a difference in the concordance statistic of 0.1 is a real effect. To compute the posteriors for the difference in models, the full model will be contrasted with the others:

```{r diffs}
comparisons <- contrast_models(
  concord_est, 
  list_1 = rep("full", 3),
  list_2 = c("ph.ecog", "age", "sex"),
  seed = 4654
  )
```

The posterior distributions show that, statistically, `ph.ecog` has real importance ot the model. However, since these distributions are mostly with +/- 0.05, they are unlikely to be real differences. 

```{r diff-post}
ggplot(comparisons, size = 0.05) + 
  theme_bw()
```

The ROPE statistics quantify the practical effects:

```{r diff-sum}
summary(comparisons, size = 0.05) %>%
  dplyr::select(contrast, starts_with("pract"))
```

# default time param with lag

    Code
      initial_time_split(drinks, lag = 12.5)
    Condition
      Error in `initial_time_split()`:
      ! `lag` must be a whole number.

---

    Code
      initial_time_split(drinks, lag = 500)
    Condition
      Error in `initial_time_split()`:
      ! `lag` must be less than or equal to the number of training observations.

# printing initial split objects

    Code
      initial_split(mtcars)
    Output
      <Training/Testing/Total>
      <24/8/32>

---

    Code
      initial_time_split(mtcars)
    Output
      <Training/Testing/Total>
      <24/8/32>

# printing

    Code
      group_vfold_cv(warpbreaks, "tension")
    Output
      # Group 3-fold cross-validation 
      # A tibble: 3 x 2
        splits          id       
        <list>          <chr>    
      1 <split [36/18]> Resample1
      2 <split [36/18]> Resample2
      3 <split [36/18]> Resample3

# printing with ...

    Code
      print(group_vfold_cv(warpbreaks, "tension"), n = 2)
    Output
      # Group 3-fold cross-validation 
      # A tibble: 3 x 2
        splits          id       
        <list>          <chr>    
      1 <split [36/18]> Resample1
      2 <split [36/18]> Resample2
      # ... with 1 more row

# bad data

    Code
      s1 <- make_strata(x3, pool = 0.06)
    Condition
      Warning:
      Stratifying groups that make up 6% of the data may be statistically risky.
      * Consider increasing `pool` to at least 0.1

---

    Code
      s2 <- make_strata(mtcars$mpg)
    Condition
      Warning:
      The number of observations in each quantile is below the recommended threshold of 20.
      * Stratification will use 1 breaks instead.
      Warning:
      Too little data to stratify.
      * Resampling will be unstratified.

---

    Code
      s3 <- make_strata(seq_len(50), breaks = -1)
    Condition
      Warning:
      The bins specified by `breaks` must be >=2.
      * Resampling will be unstratified.

# printing

    Code
      loo_cv(dat1)
    Output
      # Leave-one-out cross-validation 
      # A tibble: 20 x 2
         splits         id        
         <list>         <chr>     
       1 <split [19/1]> Resample1 
       2 <split [19/1]> Resample2 
       3 <split [19/1]> Resample3 
       4 <split [19/1]> Resample4 
       5 <split [19/1]> Resample5 
       6 <split [19/1]> Resample6 
       7 <split [19/1]> Resample7 
       8 <split [19/1]> Resample8 
       9 <split [19/1]> Resample9 
      10 <split [19/1]> Resample10
      11 <split [19/1]> Resample11
      12 <split [19/1]> Resample12
      13 <split [19/1]> Resample13
      14 <split [19/1]> Resample14
      15 <split [19/1]> Resample15
      16 <split [19/1]> Resample16
      17 <split [19/1]> Resample17
      18 <split [19/1]> Resample18
      19 <split [19/1]> Resample19
      20 <split [19/1]> Resample20

# strata

    Code
      rs4 <- vfold_cv(mlc_churn, strata = state, pool = 0.01)
    Condition
      Warning:
      Stratifying groups that make up 1% of the data may be statistically risky.
      * Consider increasing `pool` to at least 0.1

# bad args

    `v` must be a single positive integer

---

    The number of rows is less than `v = 500`

# printing

    Code
      vfold_cv(mtcars)
    Output
      #  10-fold cross-validation 
      # A tibble: 10 x 2
         splits         id    
         <list>         <chr> 
       1 <split [28/4]> Fold01
       2 <split [28/4]> Fold02
       3 <split [29/3]> Fold03
       4 <split [29/3]> Fold04
       5 <split [29/3]> Fold05
       6 <split [29/3]> Fold06
       7 <split [29/3]> Fold07
       8 <split [29/3]> Fold08
       9 <split [29/3]> Fold09
      10 <split [29/3]> Fold10

# default time param with lag

    Code
      validation_time_split(drinks)
    Output
      # Validation Set Split (0.75/0.25)  
      # A tibble: 1 x 2
        splits           id        
        <list>           <chr>     
      1 <split [231/78]> validation

---

    Code
      validation_time_split(drinks, lag = 12.5)
    Condition
      Error in `validation_time_split()`:
      ! `lag` must be a whole number.

---

    Code
      validation_time_split(drinks, lag = 500)
    Condition
      Error in `validation_time_split()`:
      ! `lag` must be less than or equal to the number of training observations.

# printing

    Code
      validation_split(warpbreaks)
    Output
      # Validation Set Split (0.75/0.25)  
      # A tibble: 1 x 2
        splits          id        
        <list>          <chr>     
      1 <split [40/14]> validation

---

    Code
      validation_split(warpbreaks)$splits[[1]]
    Output
      <Training/Validation/Total>
      <40/14/54>

# Upper & lower confidence interval does not contain NA

    Code
      int_pctl(bt_resamples, res)
    Condition
      Warning:
      Recommend at least 1000 non-missing bootstrap resamples for term `mean`.
      Error in `pctl_single()`:
      ! All statistics have missing values..

---

    Code
      int_t(bt_resamples, res)
    Condition
      Warning:
      Recommend at least 500 non-missing bootstrap resamples for term `mean`.
      Error in `t_single()`:
      ! All statistics have missing values.

---

    Code
      int_bca(bt_resamples, res, .fn = bad_stats)
    Condition
      Warning:
      Recommend at least 1000 non-missing bootstrap resamples for term `mean`.
      Error in `bca_calc()`:
      ! All statistics have missing values.

# regression intervals

    Code
      skip_if(new_rng_snapshots)
      set.seed(123)
      int_2 <- reg_intervals(mpg ~ disp + wt, data = mtcars, filter = term == "wt",
      model_fn = "glm", keep_reps = TRUE)
      int_2
    Output
      # A tibble: 1 x 7
        term  .lower .estimate .upper .alpha .method          .replicates
        <chr>  <dbl>     <dbl>  <dbl>  <dbl> <chr>     <list<tibble[,2]>>
      1 wt     -5.62     -3.46 -0.955   0.05 student-t        [1,001 x 2]

# no assessment set

    Code
      assessment(xx$splits[[1]])
    Condition
      Error in `as.data.frame()`:
      ! There is no assessment data set for an `rsplit` object with class `rsplit`.

# printing

    Code
      permutations(mtcars, 1)
    Output
      # Permutation sampling 
      # Permuted columns: [mpg] 
      # A tibble: 25 x 2
         splits         id            
         <list>         <chr>         
       1 <split [32/0]> Permutations01
       2 <split [32/0]> Permutations02
       3 <split [32/0]> Permutations03
       4 <split [32/0]> Permutations04
       5 <split [32/0]> Permutations05
       6 <split [32/0]> Permutations06
       7 <split [32/0]> Permutations07
       8 <split [32/0]> Permutations08
       9 <split [32/0]> Permutations09
      10 <split [32/0]> Permutations10
      # ... with 15 more rows

# grouped variants are consistent across R sessions

    Code
      analysis(rs$splits[[1]])
    Output
      # A tibble: 87 x 14
         name        height  mass hair_~1 skin_~2 eye_c~3 birth~4 sex   gender homew~5
         <chr>        <int> <dbl> <chr>   <chr>   <chr>     <dbl> <chr> <chr>  <chr>  
       1 Luke Skywa~    172    77 blond   fair    blue       19   male  mascu~ Tatooi~
       2 C-3PO          167    75 <NA>    gold    yellow    112   none  mascu~ Tatooi~
       3 Darth Vader    202   136 none    white   yellow     41.9 male  mascu~ Tatooi~
       4 Leia Organa    150    49 brown   light   brown      19   fema~ femin~ Aldera~
       5 Leia Organa    150    49 brown   light   brown      19   fema~ femin~ Aldera~
       6 Leia Organa    150    49 brown   light   brown      19   fema~ femin~ Aldera~
       7 Beru White~    165    75 brown   light   blue       47   fema~ femin~ Tatooi~
       8 Biggs Dark~    183    84 black   light   brown      24   male  mascu~ Tatooi~
       9 Biggs Dark~    183    84 black   light   brown      24   male  mascu~ Tatooi~
      10 Biggs Dark~    183    84 black   light   brown      24   male  mascu~ Tatooi~
      # ... with 77 more rows, 4 more variables: species <chr>, films <list>,
      #   vehicles <list>, starships <list>, and abbreviated variable names
      #   1: hair_color, 2: skin_color, 3: eye_color, 4: birth_year, 5: homeworld

---

    Code
      analysis(rs$splits[[1]])
    Output
      # A tibble: 86 x 14
         name        height  mass hair_~1 skin_~2 eye_c~3 birth~4 sex   gender homew~5
         <chr>        <int> <dbl> <chr>   <chr>   <chr>     <dbl> <chr> <chr>  <chr>  
       1 Luke Skywa~    172    77 blond   fair    blue       19   male  mascu~ Tatooi~
       2 C-3PO          167    75 <NA>    gold    yellow    112   none  mascu~ Tatooi~
       3 R2-D2           96    32 <NA>    white,~ red        33   none  mascu~ Naboo  
       4 Darth Vader    202   136 none    white   yellow     41.9 male  mascu~ Tatooi~
       5 Leia Organa    150    49 brown   light   brown      19   fema~ femin~ Aldera~
       6 Owen Lars      178   120 brown,~ light   blue       52   male  mascu~ Tatooi~
       7 Beru White~    165    75 brown   light   blue       47   fema~ femin~ Tatooi~
       8 R5-D4           97    32 <NA>    white,~ red        NA   none  mascu~ Tatooi~
       9 Biggs Dark~    183    84 black   light   brown      24   male  mascu~ Tatooi~
      10 Obi-Wan Ke~    182    77 auburn~ fair    blue-g~    57   male  mascu~ Stewjon
      # ... with 76 more rows, 4 more variables: species <chr>, films <list>,
      #   vehicles <list>, starships <list>, and abbreviated variable names
      #   1: hair_color, 2: skin_color, 3: eye_color, 4: birth_year, 5: homeworld

---

    Code
      analysis(rs$splits[[1]])
    Output
      # A tibble: 65 x 14
         name        height  mass hair_~1 skin_~2 eye_c~3 birth~4 sex   gender homew~5
         <chr>        <int> <dbl> <chr>   <chr>   <chr>     <dbl> <chr> <chr>  <chr>  
       1 C-3PO          167    75 <NA>    gold    yellow    112   none  mascu~ Tatooi~
       2 R2-D2           96    32 <NA>    white,~ red        33   none  mascu~ Naboo  
       3 Leia Organa    150    49 brown   light   brown      19   fema~ femin~ Aldera~
       4 Owen Lars      178   120 brown,~ light   blue       52   male  mascu~ Tatooi~
       5 Beru White~    165    75 brown   light   blue       47   fema~ femin~ Tatooi~
       6 R5-D4           97    32 <NA>    white,~ red        NA   none  mascu~ Tatooi~
       7 Biggs Dark~    183    84 black   light   brown      24   male  mascu~ Tatooi~
       8 Obi-Wan Ke~    182    77 auburn~ fair    blue-g~    57   male  mascu~ Stewjon
       9 Anakin Sky~    188    84 blond   fair    blue       41.9 male  mascu~ Tatooi~
      10 Greedo         173    74 <NA>    green   black      44   male  mascu~ Rodia  
      # ... with 55 more rows, 4 more variables: species <chr>, films <list>,
      #   vehicles <list>, starships <list>, and abbreviated variable names
      #   1: hair_color, 2: skin_color, 3: eye_color, 4: birth_year, 5: homeworld

---

    Code
      analysis(rs$splits[[1]])
    Output
      # A tibble: 65 x 14
         name        height  mass hair_~1 skin_~2 eye_c~3 birth~4 sex   gender homew~5
         <chr>        <int> <dbl> <chr>   <chr>   <chr>     <dbl> <chr> <chr>  <chr>  
       1 C-3PO          167    75 <NA>    gold    yellow    112   none  mascu~ Tatooi~
       2 Darth Vader    202   136 none    white   yellow     41.9 male  mascu~ Tatooi~
       3 Leia Organa    150    49 brown   light   brown      19   fema~ femin~ Aldera~
       4 Owen Lars      178   120 brown,~ light   blue       52   male  mascu~ Tatooi~
       5 Beru White~    165    75 brown   light   blue       47   fema~ femin~ Tatooi~
       6 R5-D4           97    32 <NA>    white,~ red        NA   none  mascu~ Tatooi~
       7 Biggs Dark~    183    84 black   light   brown      24   male  mascu~ Tatooi~
       8 Obi-Wan Ke~    182    77 auburn~ fair    blue-g~    57   male  mascu~ Stewjon
       9 Wilhuff Ta~    180    NA auburn~ fair    blue       64   male  mascu~ Eriadu 
      10 Chewbacca      228   112 brown   unknown blue      200   male  mascu~ Kashyy~
      # ... with 55 more rows, 4 more variables: species <chr>, films <list>,
      #   vehicles <list>, starships <list>, and abbreviated variable names
      #   1: hair_color, 2: skin_color, 3: eye_color, 4: birth_year, 5: homeworld

# printing

    Code
      mc_cv(warpbreaks)
    Output
      # Monte Carlo cross-validation (0.75/0.25) with 25 resamples  
      # A tibble: 25 x 2
         splits          id        
         <list>          <chr>     
       1 <split [40/14]> Resample01
       2 <split [40/14]> Resample02
       3 <split [40/14]> Resample03
       4 <split [40/14]> Resample04
       5 <split [40/14]> Resample05
       6 <split [40/14]> Resample06
       7 <split [40/14]> Resample07
       8 <split [40/14]> Resample08
       9 <split [40/14]> Resample09
      10 <split [40/14]> Resample10
      # ... with 15 more rows

# printing

    Code
      set.seed(11)
      bootstraps(warpbreaks)
    Output
      # Bootstrap sampling 
      # A tibble: 25 x 2
         splits          id         
         <list>          <chr>      
       1 <split [54/20]> Bootstrap01
       2 <split [54/22]> Bootstrap02
       3 <split [54/17]> Bootstrap03
       4 <split [54/22]> Bootstrap04
       5 <split [54/17]> Bootstrap05
       6 <split [54/19]> Bootstrap06
       7 <split [54/18]> Bootstrap07
       8 <split [54/17]> Bootstrap08
       9 <split [54/23]> Bootstrap09
      10 <split [54/22]> Bootstrap10
      # ... with 15 more rows

# print methods

    Code
      set.seed(233)
      vfold_cv(mtcars)$splits[[1]]
    Output
      <Analysis/Assess/Total>
      <28/4/32>

---

    Code
      set.seed(233)
      validation_split(mtcars)$splits[[1]]
    Output
      <Training/Validation/Total>
      <24/8/32>

---

    Code
      set.seed(233)
      validation_split(mtcars)
    Output
      # Validation Set Split (0.75/0.25)  
      # A tibble: 1 x 2
        splits         id        
        <list>         <chr>     
      1 <split [24/8]> validation

# default complement method errors

    Code
      complement("a string")
    Condition
      Error in `complement()`:
      ! No `complement()` method for this class(es) 'character'

# nested cv

    Code
      labels(nested_cv(mtcars, outside = vfold_cv(v = 3), inside = bootstraps(times = 5)))
    Condition
      Error in `labels()`:
      ! `labels` not implemented for nested resampling

# reverse_splits is working

    Code
      reverse_splits(1)
    Condition
      Error in `reverse_splits()`:
      ! `x` must be either an `rsplit` or an `rset` object

---

    Code
      reverse_splits(permutes)
    Condition
      Error in `reverse_splits()`:
      ! Permutations cannot have their splits reversed

---

    Code
      reverse_splits(permutes$splits[[1]])
    Condition
      Error in `reverse_splits()`:
      ! Permutations cannot have their splits reversed

# reshuffle_rset is working

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical rset when called on sliding_index objects
    Output
      # Sliding index resampling 
      # A tibble: 49 x 2
         splits        id     
         <list>        <chr>  
       1 <split [1/1]> Slice01
       2 <split [1/1]> Slice02
       3 <split [1/1]> Slice03
       4 <split [1/1]> Slice04
       5 <split [1/1]> Slice05
       6 <split [1/1]> Slice06
       7 <split [1/1]> Slice07
       8 <split [1/1]> Slice08
       9 <split [1/1]> Slice09
      10 <split [1/1]> Slice10
      # i 39 more rows

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical rset when called on sliding_period objects
    Output
      # Sliding period resampling 
      # A tibble: 7 x 2
        splits        id    
        <list>        <chr> 
      1 <split [7/7]> Slice1
      2 <split [7/7]> Slice2
      3 <split [7/7]> Slice3
      4 <split [7/7]> Slice4
      5 <split [7/7]> Slice5
      6 <split [7/7]> Slice6
      7 <split [7/1]> Slice7

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical rset when called on sliding_window objects
    Output
      # Sliding window resampling 
      # A tibble: 49 x 2
         splits        id     
         <list>        <chr>  
       1 <split [1/1]> Slice01
       2 <split [1/1]> Slice02
       3 <split [1/1]> Slice03
       4 <split [1/1]> Slice04
       5 <split [1/1]> Slice05
       6 <split [1/1]> Slice06
       7 <split [1/1]> Slice07
       8 <split [1/1]> Slice08
       9 <split [1/1]> Slice09
      10 <split [1/1]> Slice10
      # i 39 more rows

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical rset when called on rolling_origin objects
    Output
      # Rolling origin forecast resampling 
      # A tibble: 45 x 2
         splits         id     
         <list>         <chr>  
       1 <split [5/1]>  Slice01
       2 <split [6/1]>  Slice02
       3 <split [7/1]>  Slice03
       4 <split [8/1]>  Slice04
       5 <split [9/1]>  Slice05
       6 <split [10/1]> Slice06
       7 <split [11/1]> Slice07
       8 <split [12/1]> Slice08
       9 <split [13/1]> Slice09
      10 <split [14/1]> Slice10
      # i 35 more rows

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical rset when called on validation_time_split objects
    Output
      # Validation Set Split (0.75/0.25)  
      # A tibble: 1 x 2
        splits          id        
        <list>          <chr>     
      1 <split [37/13]> validation

---

    Code
      reshuffle_rset(rset_subclasses[[non_random_classes[[i]]]])
    Condition
      Warning:
      `reshuffle_rset()` will return an identical rset when called on validation_set objects
    Output
      # A tibble: 1 x 2
        splits          id        
        <list>          <chr>     
      1 <split [30/10]> validation

---

    Cannot reshuffle this rset (`attr(rset, 'strata')` is `TRUE`, not a column identifier)

---

    `manual_rset` objects cannot be reshuffled

---

    `rset` must be an rset object

# get_rsplit()

    Code
      get_rsplit(val, 3)
    Condition
      Error in `get_rsplit()`:
      ! `index` must be a length-1 integer between 1 and 1.
      x A value of 3 was provided.

---

    Code
      get_rsplit(val, c(1, 2))
    Condition
      Error in `get_rsplit()`:
      ! `index` must be a length-1 integer between 1 and 1.
      x Index was of length 2.

---

    Code
      get_rsplit(val, 1.5)
    Condition
      Error in `get_rsplit()`:
      ! `index` must be a length-1 integer between 1 and 1.
      x A value of 1.5 was provided.

---

    Code
      get_rsplit(warpbreaks, 1)
    Condition
      Error in `get_rsplit()`:
      ! No `get_rsplit()` method for this class(es) 'data.frame'

# strata

    Code
      rs4 <- vfold_cv(mlc_churn, strata = state, pool = 0.01)
    Condition
      Warning:
      Stratifying groups that make up 1% of the data may be statistically risky.
      * Consider increasing `pool` to at least 0.1

# bad args

    `v` must be a single positive integer greater than 1

---

    `v` must be a single positive integer greater than 1

---

    `v` must be a single positive integer greater than 1

---

    The number of rows is less than `v = 500`

---

    Repeated resampling when `v` is 150 would create identical resamples

---

    `repeats` must be a single positive integer

---

    `repeats` must be a single positive integer

# printing

    Code
      vfold_cv(mtcars)
    Output
      #  10-fold cross-validation 
      # A tibble: 10 x 2
         splits         id    
         <list>         <chr> 
       1 <split [28/4]> Fold01
       2 <split [28/4]> Fold02
       3 <split [29/3]> Fold03
       4 <split [29/3]> Fold04
       5 <split [29/3]> Fold05
       6 <split [29/3]> Fold06
       7 <split [29/3]> Fold07
       8 <split [29/3]> Fold08
       9 <split [29/3]> Fold09
      10 <split [29/3]> Fold10

# grouping -- bad args

    Repeated resampling when `v` is 4 would create identical resamples

---

    Repeated resampling when `v` is `NULL` would create identical resamples

---

    Code
      group_vfold_cv(Orange, v = 1, group = "Tree")
    Condition
      Error in `group_vfold_cv()`:
      ! `v` must be a single positive integer greater than 1

# grouping -- other balance methods

    Code
      rs1
    Output
      # Group 5-fold cross-validation 
      # A tibble: 5 x 2
        splits             id       
        <list>             <chr>    
      1 <split [2364/566]> Resample1
      2 <split [2371/559]> Resample2
      3 <split [2360/570]> Resample3
      4 <split [2278/652]> Resample4
      5 <split [2347/583]> Resample5

# grouping -- strata

    Code
      sizes4
    Output
      # A tibble: 5 x 5
        analysis assessment      n     p id       
           <int>      <int>  <int> <int> <chr>    
      1    80004      19996 100000     3 Resample1
      2    79850      20150 100000     3 Resample2
      3    79912      20088 100000     3 Resample3
      4    80131      19869 100000     3 Resample4
      5    80103      19897 100000     3 Resample5

---

    Leaving `v = NULL` while using stratification will set `v` to the number of groups present in the least common stratum.
    i Set `v` explicitly to override this warning.

---

    Code
      sizes5
    Output
      # A tibble: 5 x 5
        analysis assessment      n     p id       
           <int>      <int>  <int> <int> <chr>    
      1    80096      19904 100000     3 Resample1
      2    79962      20038 100000     3 Resample2
      3    79928      20072 100000     3 Resample3
      4    80058      19942 100000     3 Resample4
      5    79956      20044 100000     3 Resample5

---

    Leaving `v = NULL` while using stratification will set `v` to the number of groups present in the least common stratum.
    i Set `v` explicitly to override this warning.

# grouping -- printing

    Code
      group_vfold_cv(warpbreaks, "tension")
    Output
      # Group 3-fold cross-validation 
      # A tibble: 3 x 2
        splits          id       
        <list>          <chr>    
      1 <split [36/18]> Resample1
      2 <split [36/18]> Resample2
      3 <split [36/18]> Resample3

# grouping -- printing with ...

    Code
      print(group_vfold_cv(warpbreaks, "tension"), n = 2)
    Output
      # Group 3-fold cross-validation 
      # A tibble: 3 x 2
        splits          id       
        <list>          <chr>    
      1 <split [36/18]> Resample1
      2 <split [36/18]> Resample2
      # i 1 more row

# bad args

    `vars` are required and must be variables in `data`.

---

    `v` must be a single positive integer greater than 1

---

    The number of rows is less than `v = 500`

---

    `cluster_function` must be one of "kmeans" or "hclust", not "not an option".

---

    Code
      clustering_cv(Orange, v = 1, vars = "Tree")
    Condition
      Error in `clustering_cv()`:
      ! `v` must be a single positive integer greater than 1

---

    `repeats` must be a single positive integer

---

    `repeats` must be a single positive integer

# printing

    Code
      clustering_cv(dat1, c, v = 2)
    Output
      # 2-cluster cross-validation 
      # A tibble: 2 x 2
        splits         id   
        <list>         <chr>
      1 <split [5/15]> Fold1
      2 <split [15/5]> Fold2

# no assessment set

    Code
      assessment(xx$splits[[1]])
    Condition
      Error in `as.data.frame()`:
      ! There is no assessment data set for an `rsplit` object with class `perm_split`.

# printing

    Code
      permutations(mtcars, 1)
    Output
      # Permutation sampling 
      # Permuted columns: [mpg] 
      # A tibble: 25 x 2
         splits         id            
         <list>         <chr>         
       1 <split [32/0]> Permutations01
       2 <split [32/0]> Permutations02
       3 <split [32/0]> Permutations03
       4 <split [32/0]> Permutations04
       5 <split [32/0]> Permutations05
       6 <split [32/0]> Permutations06
       7 <split [32/0]> Permutations07
       8 <split [32/0]> Permutations08
       9 <split [32/0]> Permutations09
      10 <split [32/0]> Permutations10
      # i 15 more rows

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lables.R
\name{pretty.vfold_cv}
\alias{pretty.vfold_cv}
\alias{pretty.loo_cv}
\alias{pretty.apparent}
\alias{pretty.rolling_origin}
\alias{pretty.sliding_window}
\alias{pretty.sliding_index}
\alias{pretty.sliding_period}
\alias{pretty.mc_cv}
\alias{pretty.validation_split}
\alias{pretty.nested_cv}
\alias{pretty.bootstraps}
\alias{pretty.group_vfold_cv}
\alias{pretty.manual_rset}
\title{Short Descriptions of rsets}
\usage{
\method{pretty}{vfold_cv}(x, ...)

\method{pretty}{loo_cv}(x, ...)

\method{pretty}{apparent}(x, ...)

\method{pretty}{rolling_origin}(x, ...)

\method{pretty}{sliding_window}(x, ...)

\method{pretty}{sliding_index}(x, ...)

\method{pretty}{sliding_period}(x, ...)

\method{pretty}{mc_cv}(x, ...)

\method{pretty}{validation_split}(x, ...)

\method{pretty}{nested_cv}(x, ...)

\method{pretty}{bootstraps}(x, ...)

\method{pretty}{group_vfold_cv}(x, ...)

\method{pretty}{manual_rset}(x, ...)
}
\arguments{
\item{x}{An \code{rset} object}

\item{...}{Not currently used.}
}
\value{
A character vector.
}
\description{
Produce a character vector describing the resampling method.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reg_intervals.R
\name{reg_intervals}
\alias{reg_intervals}
\title{A convenience function for confidence intervals with linear-ish parametric models}
\usage{
reg_intervals(
  formula,
  data,
  model_fn = "lm",
  type = "student-t",
  times = NULL,
  alpha = 0.05,
  filter = term != "(Intercept)",
  keep_reps = FALSE,
  ...
)
}
\arguments{
\item{formula}{An R model formula with one outcome and at least one predictor.}

\item{data}{A data frame.}

\item{model_fn}{The model to fit. Allowable values are "lm", "glm",
"survreg", and "coxph". The latter two require that the \code{survival} package
be installed.}

\item{type}{The type of bootstrap confidence interval. Values of "student-t" and
"percentile" are allowed.}

\item{times}{A single integer for the number of bootstrap samples. If left
NULL, 1,001 are used for t-intervals and 2,001 for percentile intervals.}

\item{alpha}{Level of significance.}

\item{filter}{A logical expression used to remove rows from the final result, or \code{NULL} to keep all rows.}

\item{keep_reps}{Should the individual parameter estimates for each bootstrap
sample be retained?}

\item{...}{Options to pass to the model function (such as \code{family} for \code{glm()}).}
}
\value{
A tibble with columns "term", ".lower", ".estimate", ".upper",
".alpha", and ".method". If \code{keep_reps = TRUE}, an additional list column
called ".replicates" is also returned.
}
\description{
A convenience function for confidence intervals with linear-ish parametric models
}
\examples{
\donttest{
set.seed(1)
reg_intervals(mpg ~ I(1 / sqrt(disp)), data = mtcars)

set.seed(1)
reg_intervals(mpg ~ I(1 / sqrt(disp)), data = mtcars, keep_reps = TRUE)
}
}
\references{
Davison, A., & Hinkley, D. (1997). \emph{Bootstrap Methods and their
Application}. Cambridge: Cambridge University Press.
doi:10.1017/CBO9780511802843

\emph{Bootstrap Confidence Intervals},
\url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}
}
\seealso{
\code{\link[=int_pctl]{int_pctl()}}, \code{\link[=int_t]{int_t()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rset.R
\name{new_rset}
\alias{new_rset}
\title{Constructor for new rset objects}
\usage{
new_rset(splits, ids, attrib = NULL, subclass = character())
}
\arguments{
\item{splits}{A list column of \code{rsplits} or a tibble with a single column
called "splits" with a list column of \code{rsplits}.}

\item{ids}{A character vector or a tibble with one or more columns that
begin with "id".}

\item{attrib}{An optional named list of attributes to add to the object.}

\item{subclass}{A character vector of subclasses to add.}
}
\value{
An \code{rset} object.
}
\description{
Constructor for new rset objects
}
\details{
Once the new \code{rset} is constructed, an additional attribute called
"fingerprint" is added that is a hash of the \code{rset}. This can be used to
make sure other objects have the exact same resamples.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rset.R
\name{new_rset}
\alias{new_rset}
\title{Constructor for new rset objects}
\usage{
new_rset(splits, ids, attrib = NULL, subclass = character())
}
\arguments{
\item{splits}{A list column of \code{rsplits} or a tibble with a single column
called "splits" with a list column of \code{rsplits}.}

\item{ids}{A character vector or a tibble with one or more columns that
begin with "id".}

\item{attrib}{An optional named list of attributes to add to the object.}

\item{subclass}{A character vector of subclasses to add.}
}
\value{
An \code{rset} object.
}
\description{
Constructor for new rset objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reg_intervals.R
\name{reg_intervals}
\alias{reg_intervals}
\title{A convenience function for confidence intervals with linear-ish parametric models}
\usage{
reg_intervals(
  formula,
  data,
  model_fn = "lm",
  type = "student-t",
  times = NULL,
  alpha = 0.05,
  filter = term != "(Intercept)",
  keep_reps = FALSE,
  ...
)
}
\arguments{
\item{formula}{An R model formula with one outcome and at least one predictor.}

\item{data}{A data frame.}

\item{model_fn}{The model to fit. Allowable values are \code{"lm"}, \code{"glm"},
\code{"survreg"}, and \code{"coxph"}. The latter two require that the survival package
be installed.}

\item{type}{The type of bootstrap confidence interval. Values of \code{"student-t"} and
\code{"percentile"} are allowed.}

\item{times}{A single integer for the number of bootstrap samples. If left
\code{NULL}, 1,001 are used for t-intervals and 2,001 for percentile intervals.}

\item{alpha}{Level of significance.}

\item{filter}{A logical expression used to remove rows from the final result, or \code{NULL} to keep all rows.}

\item{keep_reps}{Should the individual parameter estimates for each bootstrap
sample be retained?}

\item{...}{Options to pass to the model function (such as \code{family} for \code{\link[stats:glm]{stats::glm()}}).}
}
\value{
A tibble with columns "term", ".lower", ".estimate", ".upper",
".alpha", and ".method". If \code{keep_reps = TRUE}, an additional list column
called ".replicates" is also returned.
}
\description{
A convenience function for confidence intervals with linear-ish parametric models
}
\examples{
\dontshow{if (rlang::is_installed("broom")) withAutoprint(\{ # examplesIf}
\donttest{
set.seed(1)
reg_intervals(mpg ~ I(1 / sqrt(disp)), data = mtcars)

set.seed(1)
reg_intervals(mpg ~ I(1 / sqrt(disp)), data = mtcars, keep_reps = TRUE)
}
\dontshow{\}) # examplesIf}
}
\references{
Davison, A., & Hinkley, D. (1997). \emph{Bootstrap Methods and their
Application}. Cambridge: Cambridge University Press.
doi:10.1017/CBO9780511802843

\emph{Bootstrap Confidence Intervals},
\url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}
}
\seealso{
\code{\link[=int_pctl]{int_pctl()}}, \code{\link[=int_t]{int_t()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complement.R
\name{populate}
\alias{populate}
\title{Add Assessment Indices}
\usage{
populate(x, ...)
}
\arguments{
\item{x}{A \code{rsplit} and \code{rset} object.}

\item{...}{Not currently used}
}
\value{
An object of the same kind with the integer indices.
}
\description{
Many \code{rsplit} and \code{rset} objects do not contain indicators for
the assessment samples. \code{populate()} can be used to fill the slot
for the appropriate indices.
}
\examples{
set.seed(28432)
fold_rs <- vfold_cv(mtcars)

fold_rs$splits[[1]]$out_id
complement(fold_rs$splits[[1]])

populate(fold_rs$splits[[1]])$out_id

fold_rs_all <- populate(fold_rs)
fold_rs_all$splits[[1]]$out_id
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_strata.R
\name{make_strata}
\alias{make_strata}
\title{Create or Modify Stratification Variables}
\usage{
make_strata(x, breaks = 4, nunique = 5, pool = 0.1, depth = 20)
}
\arguments{
\item{x}{An input vector.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{nunique}{An integer for the number of unique value threshold in the
algorithm.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{depth}{An integer that is used to determine the best number of
percentiles that should be used. The number of bins are based on
\code{min(5, floor(n / depth))} where \code{n = length(x)}.
If \code{x} is numeric, there must be at least 40 rows in the data set
(when \code{depth = 20}) to conduct stratified sampling.}
}
\value{
A factor vector.
}
\description{
This function can create strata from numeric data and make non-numeric data
more conducive for stratification.
}
\details{
For numeric data, if the number of unique levels is less than
\code{nunique}, the data are treated as categorical data.

For categorical inputs, the function will find levels of \code{x} than
occur in the data with percentage less than \code{pool}. The values from
these groups will be randomly assigned to the remaining strata (as will
data points that have missing values in \code{x}).

For numeric data with more unique values than \code{nunique}, the data
will be converted to being categorical based on percentiles of the data.
The percentile groups will have no more than 20 percent of the data in
each group. Again, missing values in \code{x} are randomly assigned
to groups.
}
\examples{
set.seed(61)
x1 <- rpois(100, lambda = 5)
table(x1)
table(make_strata(x1))

set.seed(554)
x2 <- rpois(100, lambda = 1)
table(x2)
table(make_strata(x2))

# small groups are randomly assigned
x3 <- factor(x2)
table(x3)
table(make_strata(x3))

# `oilType` data from `caret`
x4 <- rep(LETTERS[1:7], c(37, 26, 3, 7, 11, 10, 2))
table(x4)
table(make_strata(x4))
table(make_strata(x4, pool = 0.1))
table(make_strata(x4, pool = 0.0))

# not enough data to stratify
x5 <- rnorm(20)
table(make_strata(x5))

set.seed(483)
x6 <- rnorm(200)
quantile(x6, probs = (0:10) / 10)
table(make_strata(x6, breaks = 10))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_strata.R
\name{make_strata}
\alias{make_strata}
\title{Create or Modify Stratification Variables}
\usage{
make_strata(x, breaks = 4, nunique = 5, pool = 0.1, depth = 20)
}
\arguments{
\item{x}{An input vector.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{nunique}{An integer for the number of unique value threshold in the
algorithm.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group.}

\item{depth}{An integer that is used to determine the best number of
percentiles that should be used. The number of bins are based on
\code{min(5, floor(n / depth))} where \code{n = length(x)}.
If \code{x} is numeric, there must be at least 40 rows in the data set
(when \code{depth = 20}) to conduct stratified sampling.}
}
\value{
A factor vector.
}
\description{
For stratified resampling, this function can create strata from numeric data
and also make non-numeric data more conducive to be used for
stratification.
}
\details{
For numeric data, if the number of unique levels is less than
\code{nunique}, the data are treated as categorical data.

For categorical inputs, the function will find levels of \code{x} than
occur in the data with percentage less than \code{pool}. The values from
these groups will be randomly assigned to the remaining strata (as will
data points that have missing values in \code{x}).

For numeric data with more unique values than \code{nunique}, the data
will be converted to being categorical based on percentiles of the data.
The percentile groups will have no more than 20 percent of the data in
each group. Again, missing values in \code{x} are randomly assigned
to groups.
}
\examples{
set.seed(61)
x1 <- rpois(100, lambda = 5)
table(x1)
table(make_strata(x1))

set.seed(554)
x2 <- rpois(100, lambda = 1)
table(x2)
table(make_strata(x2))

# small groups are randomly assigned
x3 <- factor(x2)
table(x3)
table(make_strata(x3))

# `oilType` data from `caret`
x4 <- rep(LETTERS[1:7], c(37, 26, 3, 7, 11, 10, 2))
table(x4)
table(make_strata(x4))
table(make_strata(x4, pool = 0.1))
table(make_strata(x4, pool = 0.0))

# not enough data to stratify
x5 <- rnorm(20)
table(make_strata(x5))

set.seed(483)
x6 <- rnorm(200)
quantile(x6, probs = (0:10)/10)
table(make_strata(x6, breaks = 10))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{.get_fingerprint}
\alias{.get_fingerprint}
\alias{.get_fingerprint.default}
\alias{.get_fingerprint.rset}
\title{Obtain a identifier for the resamples}
\usage{
.get_fingerprint(x, ...)

\method{.get_fingerprint}{default}(x, ...)

\method{.get_fingerprint}{rset}(x, ...)
}
\arguments{
\item{x}{An \code{rset} or \code{tune_results} object.}

\item{...}{Not currently used.}
}
\value{
A character value or \code{NA_character_} if the object was created prior
to \code{rsample} version 0.1.0.
}
\description{
This function returns a hash (or NA) for an attribute that is created when
the \code{rset} was initially constructed. This can be used to compare with other
resampling objects to see if they are the same.
}
\examples{
set.seed(1)
.get_fingerprint(vfold_cv(mtcars))

set.seed(1)
.get_fingerprint(vfold_cv(mtcars))

set.seed(2)
.get_fingerprint(vfold_cv(mtcars))

set.seed(1)
.get_fingerprint(vfold_cv(mtcars, repeats = 2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/caret.R
\name{rsample2caret}
\alias{rsample2caret}
\alias{caret2rsample}
\title{Convert Resampling Objects to Other Formats}
\usage{
rsample2caret(object, data = c("analysis", "assessment"))

caret2rsample(ctrl, data = NULL)
}
\arguments{
\item{object}{An \code{rset} object. Currently,
\code{nested_cv} is not supported.}

\item{data}{The data that was originally used to produce the
\code{ctrl} object.}

\item{ctrl}{An object produced by \code{trainControl} that has
had the \code{index} and \code{indexOut} elements populated by
integers. One method of getting this is to extract the
\code{control} objects from an object produced by \code{train}.}
}
\value{
\code{rsample2caret} returns a list that mimics the
\code{index} and \code{indexOut} elements of a
\code{trainControl} object. \code{caret2rsample} returns an
\code{rset} object of the appropriate class.
}
\description{
These functions can convert resampling objects between
\pkg{rsample} and \pkg{caret}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fingerprint.R
\name{.get_fingerprint}
\alias{.get_fingerprint}
\alias{.get_fingerprint.default}
\alias{.get_fingerprint.rset}
\title{Obtain a identifier for the resamples}
\usage{
.get_fingerprint(x, ...)

\method{.get_fingerprint}{default}(x, ...)

\method{.get_fingerprint}{rset}(x, ...)
}
\arguments{
\item{x}{An \code{rset} or \code{tune_results} object.}

\item{...}{Not currently used.}
}
\value{
A character value or \code{NA_character_} if the object was created prior
to rsample version 0.1.0.
}
\description{
This function returns a hash (or NA) for an attribute that is created when
the \code{rset} was initially constructed. This can be used to compare with other
resampling objects to see if they are the same.
}
\examples{
set.seed(1)
.get_fingerprint(vfold_cv(mtcars))

set.seed(1)
.get_fingerprint(vfold_cv(mtcars))

set.seed(2)
.get_fingerprint(vfold_cv(mtcars))

set.seed(1)
.get_fingerprint(vfold_cv(mtcars, repeats = 2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/manual.R
\name{manual_rset}
\alias{manual_rset}
\title{Manual resampling}
\usage{
manual_rset(splits, ids)
}
\arguments{
\item{splits}{A list of \code{"rsplit"} objects. It is easiest to create these
using \code{\link[=make_splits]{make_splits()}}.}

\item{ids}{A character vector of ids. The length of \code{ids} must be the same
as the length of \code{splits}.}
}
\description{
\code{manual_rset()} is used for constructing the most minimal rset possible. It
can be useful when you have custom rsplit objects built from
\code{\link[=make_splits]{make_splits()}}, or when you want to create a new rset from splits
contained within an existing rset.
}
\examples{
df <- data.frame(x = c(1, 2, 3, 4, 5, 6))

# Create an rset from custom indices
indices <- list(
  list(analysis = c(1L, 2L), assessment = 3L),
  list(analysis = c(4L, 5L), assessment = 6L)
)

splits <- lapply(indices, make_splits, data = df)

manual_rset(splits, c("Split 1", "Split 2"))

# You can also use this to create an rset from a subset of an
# existing rset
resamples <- vfold_cv(mtcars)
best_split <- resamples[5,]
manual_rset(best_split$splits, best_split$id)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/manual.R
\name{manual_rset}
\alias{manual_rset}
\title{Manual resampling}
\usage{
manual_rset(splits, ids)
}
\arguments{
\item{splits}{A list of \code{"rsplit"} objects. It is easiest to create these
using \code{\link[=make_splits]{make_splits()}}.}

\item{ids}{A character vector of ids. The length of \code{ids} must be the same
as the length of \code{splits}.}
}
\description{
\code{manual_rset()} is used for constructing the most minimal rset possible. It
can be useful when you have custom rsplit objects built from
\code{\link[=make_splits]{make_splits()}}, or when you want to create a new rset from splits
contained within an existing rset.
}
\examples{
df <- data.frame(x = c(1, 2, 3, 4, 5, 6))

# Create an rset from custom indices
indices <- list(
  list(analysis = c(1L, 2L), assessment = 3L),
  list(analysis = c(4L, 5L), assessment = 6L)
)

splits <- lapply(indices, make_splits, data = df)

manual_rset(splits, c("Split 1", "Split 2"))

# You can also use this to create an rset from a subset of an
# existing rset
resamples <- vfold_cv(mtcars)
best_split <- resamples[5, ]
manual_rset(best_split$splits, best_split$id)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reshuffle_rset.R
\name{reshuffle_rset}
\alias{reshuffle_rset}
\title{"Reshuffle" an rset to re-generate a new rset with the same parameters}
\usage{
reshuffle_rset(rset)
}
\arguments{
\item{rset}{The \code{rset} object to be reshuffled}
}
\value{
An rset of the same class as \code{rset}.
}
\description{
This function re-generates an rset object, using the same arguments used
to generate the original.
}
\examples{
set.seed(123)
(starting_splits <- group_vfold_cv(mtcars, cyl, v = 3))
reshuffle_rset(starting_splits)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lables.R
\name{labels.rset}
\alias{labels.rset}
\alias{labels.vfold_cv}
\title{Find Labels from rset Object}
\usage{
\method{labels}{rset}(object, make_factor = FALSE, ...)

\method{labels}{vfold_cv}(object, make_factor = FALSE, ...)
}
\arguments{
\item{object}{An \code{rset} object}

\item{make_factor}{A logical for whether the results should be
a character or a factor.}

\item{...}{Not currently used.}
}
\value{
A single character or factor vector.
}
\description{
Produce a vector of resampling labels (e.g. "Fold1") from
an \code{rset} object. Currently, \code{nested_cv}
is not supported.
}
\examples{
labels(vfold_cv(mtcars))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{drinks}
\alias{drinks}
\title{Sample Time Series Data}
\description{
Sample Time Series Data
}
\details{
Drink sales. The exact name of the series from FRED is:
"Merchant Wholesalers, Except Manufacturers' Sales Branches and Offices
Sales: Nondurable Goods: Beer, Wine, and Distilled Alcoholic Beverages Sales"

These data are now in the \code{modeldata} package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{labels.rset}
\alias{labels.rset}
\alias{labels.vfold_cv}
\title{Find Labels from rset Object}
\usage{
\method{labels}{rset}(object, make_factor = FALSE, ...)

\method{labels}{vfold_cv}(object, make_factor = FALSE, ...)
}
\arguments{
\item{object}{An \code{rset} object}

\item{make_factor}{A logical for whether the results should be
a character or a factor.}

\item{...}{Not currently used.}
}
\value{
A single character or factor vector.
}
\description{
Produce a vector of resampling labels (e.g. "Fold1") from
an \code{rset} object. Currently, \code{nested_cv}
is not supported.
}
\examples{
labels(vfold_cv(mtcars))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rsplit.R
\name{as.data.frame.rsplit}
\alias{as.data.frame.rsplit}
\alias{analysis}
\alias{assessment}
\title{Convert an \code{rsplit} object to a data frame}
\usage{
\method{as.data.frame}{rsplit}(x, row.names = NULL, optional = FALSE, data = "analysis", ...)

analysis(x, ...)

assessment(x, ...)
}
\arguments{
\item{x}{An \code{rsplit} object.}

\item{row.names}{\code{NULL} or a character vector giving the row names for the data frame. Missing values are not allowed.}

\item{optional}{A logical: should the column names of the data be checked for legality?}

\item{data}{Either "analysis" or "assessment" to specify which data are returned.}

\item{...}{Additional arguments to be passed to or from methods. Not currently used.}
}
\description{
The analysis or assessment code can be returned as a data
frame (as dictated by the \code{data} argument) using
\code{as.data.frame.rsplit}. \code{analysis} and
\code{assessment} are shortcuts.
}
\examples{
library(dplyr)
set.seed(104)
folds <- vfold_cv(mtcars)

model_data_1 <- folds$splits[[1]] \%>\% analysis()
holdout_data_1 <- folds$splits[[1]] \%>\% assessment()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rsample-package.R
\docType{package}
\name{rsample-package}
\alias{rsample}
\alias{rsample-package}
\title{rsample: General Resampling Infrastructure}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Classes and functions to create and summarize different types of resampling objects (e.g. bootstrap, cross-validation).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rsample.tidymodels.org}
  \item \url{https://github.com/tidymodels/rsample}
  \item Report bugs at \url{https://github.com/tidymodels/rsample/issues}
}

}
\author{
\strong{Maintainer}: Julia Silge \email{julia.silge@rstudio.com} (\href{https://orcid.org/0000-0002-3671-836X}{ORCID})

Authors:
\itemize{
  \item Fanny Chow \email{fannybchow@gmail.com}
  \item Max Kuhn \email{max@rstudio.com}
  \item Hadley Wickham \email{hadley@rstudio.com}
}

Other contributors:
\itemize{
  \item RStudio [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rsample-package.R
\docType{package}
\name{rsample-package}
\alias{rsample}
\alias{rsample-package}
\title{rsample: General Resampling Infrastructure}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Classes and functions to create and summarize different types of resampling objects (e.g. bootstrap, cross-validation).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rsample.tidymodels.org}
  \item \url{https://github.com/tidymodels/rsample}
  \item Report bugs at \url{https://github.com/tidymodels/rsample/issues}
}

}
\author{
\strong{Maintainer}: Hannah Frick \email{hannah@posit.co} (\href{https://orcid.org/0000-0002-6049-5258}{ORCID})

Authors:
\itemize{
  \item Fanny Chow \email{fannybchow@gmail.com}
  \item Max Kuhn \email{max@posit.co}
  \item Michael Mahoney \email{mike.mahoney.218@gmail.com} (\href{https://orcid.org/0000-0003-2402-304X}{ORCID})
  \item Julia Silge \email{julia.silge@posit.co} (\href{https://orcid.org/0000-0002-3671-836X}{ORCID})
  \item Hadley Wickham \email{hadley@posit.co}
}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootci.R
\name{int_pctl}
\alias{int_pctl}
\alias{int_t}
\alias{int_bca}
\title{Bootstrap confidence intervals}
\usage{
int_pctl(.data, statistics, alpha = 0.05)

int_t(.data, statistics, alpha = 0.05)

int_bca(.data, statistics, alpha = 0.05, .fn, ...)
}
\arguments{
\item{.data}{A data frame containing the bootstrap resamples created using
\code{bootstraps()}. For t- and BCa-intervals, the \code{apparent} argument
should be set to \code{TRUE}. Even if the \code{apparent} argument is set to
\code{TRUE} for the percentile method, the apparent data is never used in calculating
the percentile confidence interval.}

\item{statistics}{An unquoted column name or \code{dplyr} selector that identifies
a single column in the data set containing the individual bootstrap
estimates. This must be a list column of tidy tibbles (with columns
\code{term} and \code{estimate}). For t-intervals, a
standard tidy column (usually called \code{std.err}) is required.
See the examples below.}

\item{alpha}{Level of significance}

\item{.fn}{A function to calculate statistic of interest. The
function should take an \code{rsplit} as the first argument and the \code{...} are
required.}

\item{...}{Arguments to pass to \code{.fn}.}
}
\value{
Each function returns a tibble with columns \code{.lower},
\code{.estimate}, \code{.upper}, \code{.alpha}, \code{.method}, and \code{term}.
\code{.method} is the type of interval (eg. "percentile",
"student-t", or "BCa"). \code{term} is the name of the estimate. Note
the \code{.estimate} returned from \code{int_pctl()}
is the mean of the estimates from the bootstrap resamples
and not the estimate from the apparent model.
}
\description{
Calculate bootstrap confidence intervals using various methods.
}
\details{
Percentile intervals are the standard method of
obtaining confidence intervals but require thousands of
resamples to be accurate. T-intervals may need fewer
resamples but require a corresponding variance estimate.
Bias-corrected and accelerated intervals require the original function
that was used to create the statistics of interest and are
computationally taxing.
}
\examples{
\donttest{
library(broom)
library(dplyr)
library(purrr)
library(tibble)

lm_est <- function(split, ...) {
  lm(mpg ~ disp + hp, data = analysis(split)) \%>\%
    tidy()
}

set.seed(52156)
car_rs <-
  bootstraps(mtcars, 500, apparent = TRUE) \%>\%
  mutate(results = map(splits, lm_est))

int_pctl(car_rs, results)
int_t(car_rs, results)
int_bca(car_rs, results, .fn = lm_est)

# putting results into a tidy format
rank_corr <- function(split) {
  dat <- analysis(split)
  tibble(
    term = "corr",
    estimate = cor(dat$sqft, dat$price, method = "spearman"),
    # don't know the analytical std.err so no t-intervals
    std.err = NA_real_
  )
}

set.seed(69325)
data(Sacramento, package = "modeldata")
bootstraps(Sacramento, 1000, apparent = TRUE) \%>\%
  mutate(correlations = map(splits, rank_corr)) \%>\%
  int_pctl(correlations)
}
}
\references{
Davison, A., & Hinkley, D. (1997). \emph{Bootstrap Methods and their
Application}. Cambridge: Cambridge University Press.
doi:10.1017/CBO9780511802843

\url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}
}
\seealso{
\code{\link[=reg_intervals]{reg_intervals()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initial_split.R
\name{initial_split}
\alias{initial_split}
\alias{initial_time_split}
\alias{training}
\alias{testing}
\title{Simple Training/Test Set Splitting}
\usage{
initial_split(data, prop = 3/4, strata = NULL, breaks = 4, pool = 0.1, ...)

initial_time_split(data, prop = 3/4, lag = 0, ...)

training(x)

testing(x)
}
\arguments{
\item{data}{A data frame.}

\item{prop}{The proportion of data to be retained for modeling/analysis.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{...}{Not currently used.}

\item{lag}{A value to include a lag between the assessment
and analysis set. This is useful if lagged predictors will be used
during training and testing.}

\item{x}{An \code{rsplit} object produced by \code{initial_split()} or
\code{initial_time_split()}.}
}
\value{
An \code{rsplit} object that can be used with the \code{training} and \code{testing}
functions to extract the data in each split.
}
\description{
\code{initial_split} creates a single binary split of the data into a training
set and testing set. \code{initial_time_split} does the same, but takes the
\emph{first} \code{prop} samples for training, instead of a random selection.
\code{training} and \code{testing} are used to extract the resulting data.
}
\details{
With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
set.seed(1353)
car_split <- initial_split(mtcars)
train_data <- training(car_split)
test_data <- testing(car_split)

data(drinks, package = "modeldata")
drinks_split <- initial_time_split(drinks)
train_data <- training(drinks_split)
test_data <- testing(drinks_split)
c(max(train_data$date), min(test_data$date)) # no lag

# With 12 period lag
drinks_lag_split <- initial_time_split(drinks, lag = 12)
train_data <- training(drinks_lag_split)
test_data <- testing(drinks_lag_split)
c(max(train_data$date), min(test_data$date)) # 12 period lag
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clustering.R
\name{clustering_cv}
\alias{clustering_cv}
\title{Cluster Cross-Validation}
\usage{
clustering_cv(
  data,
  vars,
  v = 10,
  repeats = 1,
  distance_function = "dist",
  cluster_function = c("kmeans", "hclust"),
  ...
)
}
\arguments{
\item{data}{A data frame.}

\item{vars}{A vector of bare variable names to use to cluster the data.}

\item{v}{The number of partitions of the data set.}

\item{repeats}{The number of times to repeat the clustered partitioning.}

\item{distance_function}{Which function should be used for distance calculations?
Defaults to \code{\link[stats:dist]{stats::dist()}}. You can also provide your own
function; see \code{Details}.}

\item{cluster_function}{Which function should be used for clustering?
Options are either \code{"kmeans"} (to use \code{\link[stats:kmeans]{stats::kmeans()}})
or \code{"hclust"} (to use \code{\link[stats:hclust]{stats::hclust()}}). You can also provide your own
function; see \code{Details}.}

\item{...}{Extra arguments passed on to \code{cluster_function}.}
}
\value{
A tibble with classes \code{rset}, \code{tbl_df}, \code{tbl}, and \code{data.frame}.
The results include a column for the data split objects and
an identification variable \code{id}.
}
\description{
Cluster cross-validation splits the data into V groups of
disjointed sets using k-means clustering of some variables.
A resample of the analysis data consists of V-1 of the
folds/clusters while the assessment set contains the final fold/cluster. In
basic cross-validation (i.e. no repeats), the number of resamples
is equal to V.
}
\details{
The variables in the \code{vars} argument are used for k-means clustering of
the data into disjointed sets or for hierarchical clustering of the data.
These clusters are used as the folds for cross-validation. Depending on how
the data are distributed, there may not be an equal number of points
in each fold.

You can optionally provide a custom function to \code{distance_function}. The
function should take a data frame (as created via \code{data[vars]}) and return
a \code{\link[stats:dist]{stats::dist()}} object with distances between data points.

You can optionally provide a custom function to \code{cluster_function}. The
function must take three arguments:
\itemize{
\item \code{dists}, a \code{\link[stats:dist]{stats::dist()}} object with distances between data points
\item \code{v}, a length-1 numeric for the number of folds to create
\item \code{...}, to pass any additional named arguments to your function
}

The function should return a vector of cluster assignments of length
\code{nrow(data)}, with each element of the vector corresponding to the matching
row of the data frame.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
data(ames, package = "modeldata")
clustering_cv(ames, vars = c(Sale_Price, First_Flr_SF, Second_Flr_SF), v = 2)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{.get_split_args}
\alias{.get_split_args}
\title{Get the split arguments from an rset}
\usage{
.get_split_args(x, allow_strata_false = FALSE)
}
\arguments{
\item{x}{An \code{rset} or \code{initial_split} object.}

\item{allow_strata_false}{A logical to specify which value to use if no
stratification was specified. The default is to use \code{strata = NULL}, the
alternative is \code{strata = FALSE}.}
}
\value{
A list of arguments used to create the rset.
}
\description{
Get the split arguments from an rset
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bootci.R
\name{int_pctl}
\alias{int_pctl}
\alias{int_pctl.default}
\alias{int_pctl.bootstraps}
\alias{int_t}
\alias{int_t.default}
\alias{int_t.bootstraps}
\alias{int_bca}
\alias{int_bca.default}
\alias{int_bca.bootstraps}
\title{Bootstrap confidence intervals}
\usage{
int_pctl(.data, ...)

\method{int_pctl}{default}(.data, ...)

\method{int_pctl}{bootstraps}(.data, statistics, alpha = 0.05, ...)

int_t(.data, ...)

\method{int_t}{default}(.data, ...)

\method{int_t}{bootstraps}(.data, statistics, alpha = 0.05, ...)

int_bca(.data, ...)

\method{int_bca}{default}(.data, ...)

\method{int_bca}{bootstraps}(.data, statistics, alpha = 0.05, .fn, ...)
}
\arguments{
\item{.data}{A object containing the bootstrap resamples, created using
\code{\link[=bootstraps]{bootstraps()}}. For t- and BCa-intervals, the \code{apparent} argument
should be set to \code{TRUE}. Even if the \code{apparent} argument is set to
\code{TRUE} for the percentile method, the apparent data is never used in calculating
the percentile confidence interval.}

\item{...}{Arguments to pass to \code{.fn} (\code{int_bca()} only).}

\item{statistics}{An unquoted column name or \code{dplyr} selector that identifies
a single column in the data set containing the individual bootstrap
estimates. This must be a list column of tidy tibbles (with columns
\code{term} and \code{estimate}). Optionally, users can include columns whose names
begin with a period and the intervals will be created for each combination
of these variables and the \code{term} column. For t-intervals, a standard tidy
column (usually called \code{std.error}) is required. See the examples below.}

\item{alpha}{Level of significance.}

\item{.fn}{A function to calculate statistic of interest. The
function should take an \code{rsplit} as the first argument and the \code{...} are
required.}
}
\value{
Each function returns a tibble with columns \code{.lower},
\code{.estimate}, \code{.upper}, \code{.alpha}, \code{.method}, and \code{term}.
\code{.method} is the type of interval (eg. "percentile",
"student-t", or "BCa"). \code{term} is the name of the estimate. Note
the \code{.estimate} returned from \code{int_pctl()}
is the mean of the estimates from the bootstrap resamples
and not the estimate from the apparent model.
}
\description{
Calculate bootstrap confidence intervals using various methods.
}
\details{
Percentile intervals are the standard method of
obtaining confidence intervals but require thousands of
resamples to be accurate. T-intervals may need fewer
resamples but require a corresponding variance estimate.
Bias-corrected and accelerated intervals require the original function
that was used to create the statistics of interest and are
computationally taxing.
}
\examples{
\dontshow{if (rlang::is_installed("broom") & rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
\donttest{
library(broom)
library(dplyr)
library(purrr)
library(tibble)
library(tidyr)

# ------------------------------------------------------------------------------

lm_est <- function(split, ...) {
  lm(mpg ~ disp + hp, data = analysis(split)) |>
    tidy()
}

set.seed(52156)
car_rs <-
  bootstraps(mtcars, 500, apparent = TRUE) |>
  mutate(results = map(splits, lm_est))

int_pctl(car_rs, results)
int_t(car_rs, results)
int_bca(car_rs, results, .fn = lm_est)

# ------------------------------------------------------------------------------

# putting results into a tidy format
rank_corr <- function(split) {
  dat <- analysis(split)
  tibble(
    term = "corr",
    estimate = cor(dat$sqft, dat$price, method = "spearman"),
    # don't know the analytical std.error so no t-intervals
    std.error = NA_real_
  )
}

set.seed(69325)
data(Sacramento, package = "modeldata")
bootstraps(Sacramento, 1000, apparent = TRUE) |>
  mutate(correlations = map(splits, rank_corr)) |>
  int_pctl(correlations)

# ------------------------------------------------------------------------------
# An example of computing the interval for each value of a custom grouping
# factor (type of house in this example)

# Get regression estimates for each house type
lm_est <- function(split, ...) {
  analysis(split) |>
    tidyr::nest(.by = c(type)) |>
    # Compute regression estimates for each house type
    mutate(
      betas = purrr::map(data, \(.x) lm(log10(price) ~ sqft, data = .x) |> tidy())
    ) |>
    # Convert the column name to begin with a period
    rename(.type = type) |>
    select(.type, betas) |>
    unnest(cols = betas)
}

set.seed(52156)
house_rs <-
  bootstraps(Sacramento, 1000, apparent = TRUE) |>
  mutate(results = map(splits, lm_est))

int_pctl(house_rs, results)
}
\dontshow{\}) # examplesIf}
}
\references{
\url{https://rsample.tidymodels.org/articles/Applications/Intervals.html}

Davison, A., & Hinkley, D. (1997). \emph{Bootstrap Methods and their
Application}. Cambridge: Cambridge University Press.
doi:10.1017/CBO9780511802843
}
\seealso{
\code{\link[=reg_intervals]{reg_intervals()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complement.R
\name{complement}
\alias{complement}
\alias{complement.rsplit}
\alias{complement.rof_split}
\alias{complement.sliding_window_split}
\alias{complement.sliding_index_split}
\alias{complement.sliding_period_split}
\alias{complement.apparent_split}
\title{Determine the Assessment Samples}
\usage{
complement(x, ...)

\method{complement}{rsplit}(x, ...)

\method{complement}{rof_split}(x, ...)

\method{complement}{sliding_window_split}(x, ...)

\method{complement}{sliding_index_split}(x, ...)

\method{complement}{sliding_period_split}(x, ...)

\method{complement}{apparent_split}(x, ...)
}
\arguments{
\item{x}{An \code{rsplit} object}

\item{...}{Not currently used}
}
\value{
A integer vector.
}
\description{
This method and function help find which data belong in the analysis and
assessment sets.
}
\details{
Given an \code{rsplit} object, \code{complement()} will determine which
of the data rows are contained in the assessment set. To save space,
many of the \code{rsplit} objects will not contain indices for the
assessment split.
}
\examples{
set.seed(28432)
fold_rs <- vfold_cv(mtcars)
head(fold_rs$splits[[1]]$in_id)
fold_rs$splits[[1]]$out_id
complement(fold_rs$splits[[1]])
}
\seealso{
\code{\link[=populate]{populate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{make_splits}
\alias{make_splits}
\alias{make_splits.default}
\alias{make_splits.list}
\alias{make_splits.data.frame}
\title{Constructors for split objects}
\usage{
make_splits(x, ...)

\method{make_splits}{default}(x, ...)

\method{make_splits}{list}(x, data, class = NULL, ...)

\method{make_splits}{data.frame}(x, assessment, ...)
}
\arguments{
\item{x}{A list of integers with names "analysis" and "assessment", or a
data frame of analysis or training data.}

\item{...}{Further arguments passed to or from other methods (not currently
used).}

\item{data}{A data frame.}

\item{class}{An optional class to give the object.}

\item{assessment}{A data frame of assessment or testing data, which can be empty.}
}
\description{
Constructors for split objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initial_split.R
\name{initial_split}
\alias{initial_split}
\alias{initial_time_split}
\alias{training}
\alias{training.default}
\alias{training.rsplit}
\alias{testing}
\alias{testing.default}
\alias{testing.rsplit}
\alias{group_initial_split}
\title{Simple Training/Test Set Splitting}
\usage{
initial_split(data, prop = 3/4, strata = NULL, breaks = 4, pool = 0.1, ...)

initial_time_split(data, prop = 3/4, lag = lifecycle::deprecated(), ...)

training(x, ...)

\method{training}{default}(x, ...)

\method{training}{rsplit}(x, ...)

testing(x, ...)

\method{testing}{default}(x, ...)

\method{testing}{rsplit}(x, ...)

group_initial_split(data, group, prop = 3/4, ..., strata = NULL, pool = 0.1)
}
\arguments{
\item{data}{A data frame.}

\item{prop}{The proportion of data to be retained for modeling/analysis.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{...}{These dots are for future extensions and must be empty.}

\item{lag}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} This is deprecated, please lag
your predictors prior to splitting the dataset.}

\item{x}{An \code{rsplit} object produced by \code{initial_split()} or
\code{initial_time_split()}.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}
}
\value{
An \code{rsplit} object that can be used with the \code{training()} and \code{testing()}
functions to extract the data in each split.
}
\description{
\code{initial_split()} creates a single binary split of the data into a training
set and testing set. \code{initial_time_split()} does the same, but takes the
\emph{first} \code{prop} samples for training, instead of a random selection.
\code{group_initial_split()} creates splits of the data based
on some grouping variable, so that all data in a "group" is assigned to
the same split.
}
\details{
\code{training()} and \code{testing()} are used to extract the resulting data.

To avoid data leakage when using lagged variables, lag the predictors before
the initial split.

With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
set.seed(1353)
car_split <- initial_split(mtcars)
train_data <- training(car_split)
test_data <- testing(car_split)

data(drinks, package = "modeldata")
drinks_split <- initial_time_split(drinks)
train_data <- training(drinks_split)
test_data <- testing(drinks_split)
c(max(train_data$date), min(test_data$date))

set.seed(1353)
car_split <- group_initial_split(mtcars, cyl)
train_data <- training(car_split)
test_data <- testing(car_split)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rolling_origin.R
\name{rolling_origin}
\alias{rolling_origin}
\title{Rolling Origin Forecast Resampling}
\usage{
rolling_origin(
  data,
  initial = 5,
  assess = 1,
  cumulative = TRUE,
  skip = 0,
  lag = 0,
  ...
)
}
\arguments{
\item{data}{A data frame.}

\item{initial}{The number of samples used for analysis/modeling in the
initial resample.}

\item{assess}{The number of samples used for each assessment resample.}

\item{cumulative}{A logical. Should the analysis resample grow beyond the
size specified by \code{initial} at each resample?.}

\item{skip}{A integer indicating how many (if any) \emph{additional} resamples
to skip to thin the total amount of data points in the analysis resample.
See the example below.}

\item{lag}{A value to include a lag between the assessment
and analysis set. This is useful if lagged predictors will be used
during training and testing.}

\item{...}{Not currently used.}
}
\value{
An tibble with classes \code{rolling_origin}, \code{rset}, \code{tbl_df}, \code{tbl},
and \code{data.frame}. The results include a column for the data split objects
and a column called \code{id} that has a character string with the resample
identifier.
}
\description{
This resampling method is useful when the data set has a strong time
component. The resamples are not random and contain data points that are
consecutive values. The function assumes that the original data set are
sorted in time order.
}
\details{
The main options, \code{initial} and \code{assess}, control the number of
data points from the original data that are in the analysis and assessment
set, respectively. When \code{cumulative = TRUE}, the analysis set will grow as
resampling continues while the assessment set size will always remain
static.
\code{skip} enables the function to not use every data point in the resamples.
When \code{skip = 0}, the resampling data sets will increment by one position.
Suppose that the rows of a data set are consecutive days. Using \code{skip = 6}
will make the analysis data set to operate on \emph{weeks} instead of days. The
assessment set size is not affected by this option.
}
\examples{
set.seed(1131)
ex_data <- data.frame(row = 1:20, some_var = rnorm(20))
dim(rolling_origin(ex_data))
dim(rolling_origin(ex_data, skip = 2))
dim(rolling_origin(ex_data, skip = 2, cumulative = FALSE))

# You can also roll over calendar periods by first nesting by that period,
# which is especially useful for irregular series where a fixed window
# is not useful. This example slides over 5 years at a time.
library(dplyr)
library(tidyr)
data(drinks, package = "modeldata")

drinks_annual <- drinks \%>\%
  mutate(year = as.POSIXlt(date)$year + 1900) \%>\%
  nest(data = c(-year))

multi_year_roll <- rolling_origin(drinks_annual, cumulative = FALSE)

analysis(multi_year_roll$splits[[1]])
assessment(multi_year_roll$splits[[1]])
}
\seealso{
\code{\link[=sliding_window]{sliding_window()}}, \code{\link[=sliding_index]{sliding_index()}}, and \code{\link[=sliding_period]{sliding_period()}} for additional
time based resampling functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initial_validation_split.R
\name{initial_validation_split}
\alias{initial_validation_split}
\alias{initial_validation_time_split}
\alias{group_initial_validation_split}
\alias{training.initial_validation_split}
\alias{testing.initial_validation_split}
\alias{validation}
\alias{validation.default}
\alias{validation.initial_validation_split}
\title{Create an Initial Train/Validation/Test Split}
\usage{
initial_validation_split(
  data,
  prop = c(0.6, 0.2),
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  ...
)

initial_validation_time_split(data, prop = c(0.6, 0.2), ...)

group_initial_validation_split(
  data,
  group,
  prop = c(0.6, 0.2),
  ...,
  strata = NULL,
  pool = 0.1
)

\method{training}{initial_validation_split}(x, ...)

\method{testing}{initial_validation_split}(x, ...)

validation(x, ...)

\method{validation}{default}(x, ...)

\method{validation}{initial_validation_split}(x, ...)
}
\arguments{
\item{data}{A data frame.}

\item{prop}{A length-2 vector of proportions of data to be retained for training and
validation data, respectively.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{...}{These dots are for future extensions and must be empty.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}

\item{x}{An object of class \code{initial_validation_split}.}
}
\value{
An \code{initial_validation_split} object that can be used with the
\code{\link[=training]{training()}}, \code{\link[=validation]{validation()}}, and \code{\link[=testing]{testing()}} functions to extract the data
in each split.
}
\description{
\code{initial_validation_split()} creates a random three-way split of the data
into a training set, a validation set, and a testing set.
\code{initial_validation_time_split()} does the same, but instead of a random
selection the training, validation, and testing set are in order of the full
data set, with the first observations being put into the training set.
\code{group_initial_validation_split()} creates similar random splits of the data
based on some grouping variable, so that all data in a "group" are assigned
to the same partition.
}
\details{
\code{\link[=training]{training()}}, \code{\link[=validation]{validation()}}, and \code{\link[=testing]{testing()}} can be used to extract the
resulting data sets.
Use \code{\link[=validation_set]{validation_set()}} to create an \code{rset} object for use with functions from
the tune package such as \code{tune::tune_grid()}.

With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
set.seed(1353)
car_split <- initial_validation_split(mtcars)
train_data <- training(car_split)
validation_data <- validation(car_split)
test_data <- testing(car_split)

data(drinks, package = "modeldata")
drinks_split <- initial_validation_time_split(drinks)
train_data <- training(drinks_split)
validation_data <- validation(drinks_split)
c(max(train_data$date), min(validation_data$date))

data(ames, package = "modeldata")
set.seed(1353)
ames_split <- group_initial_validation_split(ames, group = Neighborhood)
train_data <- training(ames_split)
validation_data <- validation(ames_split)
test_data <- testing(ames_split)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=validation_set]{validation_set()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/permutations.R
\name{permutations}
\alias{permutations}
\title{Permutation sampling}
\usage{
permutations(data, permute = NULL, times = 25, apparent = FALSE, ...)
}
\arguments{
\item{data}{A data frame.}

\item{permute}{One or more columns to shuffle. This argument supports
\code{tidyselect} selectors. Multiple expressions can be combined with \code{c()}.
Variable names can be used as if they were positions in the data frame, so
expressions like \code{x:y} can be used to select a range of variables.
See \code{\link[tidyselect]{language}} for more details.}

\item{times}{The number of permutation samples.}

\item{apparent}{A logical. Should an extra resample be added where the
analysis is the standard data set.}

\item{...}{Not currently used.}
}
\value{
A \code{tibble} with classes \code{permutations}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and a
column called \code{id} that has a character string with the resample
identifier.
}
\description{
A permutation sample is the same size as the original data set and is made
by permuting/shuffling one or more columns. This results in analysis
samples where some columns are in their original order and some columns
are permuted to a random order. Unlike other sampling functions in
\code{rsample}, there is no assessment set and calling \code{assessment()} on a
permutation split will throw an error.
}
\details{
The argument \code{apparent} enables the option of an additional
"resample" where the analysis data set is the same as the original data
set. Permutation-based resampling can be especially helpful for computing
a statistic under the null hypothesis (e.g. t-statistic). This forms the
basis of a permutation test, which computes a test statistic under all
possible permutations of the data.
}
\examples{
permutations(mtcars, mpg, times = 2)
permutations(mtcars, mpg, times = 2, apparent = TRUE)

library(purrr)
resample1 <- permutations(mtcars, starts_with("c"), times = 1)
resample1$splits[[1]] \%>\% analysis()

resample2 <- permutations(mtcars, hp, times = 10, apparent = TRUE)
map_dbl(resample2$splits, function(x) {
  t.test(hp ~ vs, data = analysis(x))$statistic
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{get_rsplit}
\alias{get_rsplit}
\alias{get_rsplit.rset}
\alias{get_rsplit.default}
\title{Retrieve individual rsplits objects from an rset}
\usage{
get_rsplit(x, index, ...)

\method{get_rsplit}{rset}(x, index, ...)

\method{get_rsplit}{default}(x, index, ...)
}
\arguments{
\item{x}{The \code{rset} object to retrieve an rsplit from.}

\item{index}{An integer indicating which rsplit to retrieve: \code{1} for the
rsplit in the first row of the rset, \code{2} for the second, and so on.}

\item{...}{Not currently used.}
}
\value{
The rsplit object in row \code{index} of \code{rset}
}
\description{
Retrieve individual rsplits objects from an rset
}
\examples{
set.seed(123)
(starting_splits <- group_vfold_cv(mtcars, cyl, v = 3))
get_rsplit(starting_splits, 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{attrition}
\alias{attrition}
\title{Job Attrition}
\description{
Job Attrition
}
\details{
These data are from the IBM Watson Analytics Lab.
The website describes the data with \dQuote{Uncover the
factors that lead to employee attrition and explore important
questions such as \sQuote{show me a breakdown of distance
from home by job role and attrition} or \sQuote{compare
average monthly income by education and attrition}. This is a
fictional data set created by IBM data scientists.}. There
are 1470 rows.

These data are now in the \code{modeldata} package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_groups.R
\name{make_groups}
\alias{make_groups}
\title{Make groupings for grouped rsplits}
\usage{
make_groups(data, group, v)
}
\arguments{
\item{data}{A data frame.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}

\item{v}{The number of partitions of the data set.}
}
\description{
This function powers \link{group_vfold_cv} by splitting the data based upon
a grouping variable and returning the assessment set indices for each
split.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boot.R
\name{group_bootstraps}
\alias{group_bootstraps}
\title{Group Bootstraps}
\usage{
group_bootstraps(
  data,
  group,
  times = 25,
  apparent = FALSE,
  ...,
  strata = NULL,
  pool = 0.1
)
}
\arguments{
\item{data}{A data frame.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}

\item{times}{The number of bootstrap samples.}

\item{apparent}{A logical. Should an extra resample be added where the
analysis and holdout subset are the entire data set. This is required for
some estimators used by the \code{\link[=summary]{summary()}} function that require the apparent
error rate.}

\item{...}{These dots are for future extensions and must be empty.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}
}
\value{
An tibble with classes \code{group_bootstraps} \code{bootstraps}, \code{rset},
\code{tbl_df}, \code{tbl}, and \code{data.frame}. The results include a column for the data
split objects and a column called \code{id} that has a character string with the
resample identifier.
}
\description{
Group bootstrapping creates splits of the data based
on some grouping variable (which may have more than a single row
associated with it). A common use of this kind of resampling is when you
have repeated measures of the same subject.
A bootstrap sample is a sample that is the same size as the original data
set that is made using replacement. This results in analysis samples that
have multiple replicates of some of the original rows of the data. The
assessment set is defined as the rows of the original data that were not
included in the bootstrap sample. This is often referred to as the
"out-of-bag" (OOB) sample.
}
\details{
The argument \code{apparent} enables the option of an additional
"resample" where the analysis and assessment data sets are the same as the
original data set. This can be required for some types of analysis of the
bootstrap results.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
data(ames, package = "modeldata")

set.seed(13)
group_bootstraps(ames, Neighborhood, times = 3)
group_bootstraps(ames, Neighborhood, times = 3, apparent = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg.R
\docType{package}
\name{rsample}
\alias{rsample}
\title{rsample: General Resampling Infrastructure for R}
\description{
\pkg{rsample} has functions to create variations of a data set
that can be used to evaluate models or to estimate the
sampling distribution of some statistic.
}
\section{Terminology}{

\itemize{
\item A \strong{resample} is the result of a two-way split of a
data set. For example, when bootstrapping, one part of the
resample is a sample with replacement of the original data.
The other part of the split contains the instances that were
not contained in the bootstrap sample. The data structure
\code{rsplit} is used to store a single resample.
\item When the data are split in two, the portion that is
used to estimate the model or calculate the statistic is
called the \strong{analysis} set here. In machine learning this
is sometimes called the "training set" but this would be
poorly named since it might conflict with any initial split
of the original data.
\item Conversely, the other data in the split are called the
\strong{assessment} data. In bootstrapping, these data are
often called the "out-of-bag" samples.
\item A collection of resamples is contained in an
\code{rset} object.
}
}

\section{Basic Functions}{

The main resampling functions are: \code{\link[=vfold_cv]{vfold_cv()}},
\code{\link[=bootstraps]{bootstraps()}}, \code{\link[=mc_cv]{mc_cv()}},
\code{\link[=rolling_origin]{rolling_origin()}}, and \code{\link[=nested_cv]{nested_cv()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mc.R
\name{mc_cv}
\alias{mc_cv}
\title{Monte Carlo Cross-Validation}
\usage{
mc_cv(data, prop = 3/4, times = 25, strata = NULL, breaks = 4, pool = 0.1, ...)
}
\arguments{
\item{data}{A data frame.}

\item{prop}{The proportion of data to be retained for modeling/analysis.}

\item{times}{The number of times to repeat the sampling.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{...}{Not currently used.}
}
\value{
An tibble with classes \code{mc_cv}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and a
column called \code{id} that has a character string with the resample identifier.
}
\description{
One resample of Monte Carlo cross-validation takes a random sample (without
replacement) of the original data set to be used for analysis. All other
data points are added to the assessment set.
}
\details{
With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
mc_cv(mtcars, times = 2)
mc_cv(mtcars, prop = .5, times = 2)

library(purrr)
data(wa_churn, package = "modeldata")

set.seed(13)
resample1 <- mc_cv(wa_churn, times = 3, prop = .5)
map_dbl(
  resample1$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
resample2 <- mc_cv(wa_churn, strata = churn, times = 3, prop = .5)
map_dbl(
  resample2$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
resample3 <- mc_cv(wa_churn, strata = tenure, breaks = 6, times = 3, prop = .5)
map_dbl(
  resample3$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complement.R
\name{complement}
\alias{complement}
\alias{complement.rsplit}
\alias{complement.rof_split}
\alias{complement.sliding_window_split}
\alias{complement.sliding_index_split}
\alias{complement.sliding_period_split}
\alias{complement.apparent_split}
\title{Determine the Assessment Samples}
\usage{
complement(x, ...)

\method{complement}{rsplit}(x, ...)

\method{complement}{rof_split}(x, ...)

\method{complement}{sliding_window_split}(x, ...)

\method{complement}{sliding_index_split}(x, ...)

\method{complement}{sliding_period_split}(x, ...)

\method{complement}{apparent_split}(x, ...)
}
\arguments{
\item{x}{An \code{rsplit} object.}

\item{...}{Not currently used.}
}
\value{
A integer vector.
}
\description{
This method and function help find which data belong in the analysis and
assessment sets.
}
\details{
Given an \code{rsplit} object, \code{complement()} will determine which
of the data rows are contained in the assessment set. To save space,
many of the \code{rsplit} objects will not contain indices for the
assessment split.
}
\examples{
set.seed(28432)
fold_rs <- vfold_cv(mtcars)
head(fold_rs$splits[[1]]$in_id)
fold_rs$splits[[1]]$out_id
complement(fold_rs$splits[[1]])
}
\seealso{
\code{\link[=populate]{populate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{make_splits}
\alias{make_splits}
\alias{make_splits.default}
\alias{make_splits.list}
\alias{make_splits.data.frame}
\title{Constructors for split objects}
\usage{
make_splits(x, ...)

\method{make_splits}{default}(x, ...)

\method{make_splits}{list}(x, data, class = NULL, ...)

\method{make_splits}{data.frame}(x, assessment, class = NULL, ...)
}
\arguments{
\item{x}{A list of integers with names "analysis" and "assessment", or a
data frame of analysis or training data.}

\item{...}{Not currently used.}

\item{data}{A data frame.}

\item{class}{An optional class to give the object.}

\item{assessment}{A data frame of assessment or testing data, which can be empty.}
}
\description{
Constructors for split objects
}
\examples{
df <- data.frame(
  year = 1900:1999,
  value = 10 + 8*1900:1999 + runif(100L, 0, 100)
)
split_from_indices <- make_splits(
  x = list(analysis = which(df$year <= 1980),
           assessment = which(df$year > 1980)),
  data = df
)
split_from_data_frame <- make_splits(
  x = df[df$year <= 1980,],
  assessment = df[df$year > 1980,]
)
identical(split_from_indices, split_from_data_frame)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/form_pred.R
\name{form_pred}
\alias{form_pred}
\title{Extract Predictor Names from Formula or Terms}
\usage{
form_pred(object, ...)
}
\arguments{
\item{object}{A model formula or \code{\link[stats:terms]{stats::terms()}}
object.}

\item{...}{Arguments to pass to \code{\link[=all.vars]{all.vars()}}}
}
\value{
A character vector of names
}
\description{
\code{all.vars} returns all variables used in a formula. This
function only returns the variables explicitly used on the
right-hand side (i.e., it will not resolve dots unless the
object is terms with a data set specified).
}
\examples{
form_pred(y ~ x + z)
form_pred(terms(y ~ x + z))

form_pred(y ~ x + log(z))
form_pred(log(y) ~ x + z)

form_pred(y1 + y2 ~ x + z)
form_pred(log(y1) + y2 ~ x + z)

# will fail:
# form_pred(y ~ .)

form_pred(terms(mpg ~ (.)^2, data = mtcars))
form_pred(terms(~ (.)^2, data = mtcars))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/form_pred.R
\name{form_pred}
\alias{form_pred}
\title{Extract Predictor Names from Formula or Terms}
\usage{
form_pred(object, ...)
}
\arguments{
\item{object}{A model formula or \code{\link[stats:terms]{stats::terms()}}
object.}

\item{...}{Arguments to pass to \code{\link[=all.vars]{all.vars()}}}
}
\value{
A character vector of names
}
\description{
\code{all.vars} returns all variables used in a formula. This
function only returns the variables explicitly used on the
right-hand side (i.e., it will not resolve dots unless the
object is terms with a data set specified).
}
\examples{
form_pred(y ~ x + z)
form_pred(terms(y ~ x + z))

form_pred(y ~ x + log(z))
form_pred(log(y) ~ x + z)

form_pred(y1 + y2 ~ x + z)
form_pred(log(y1) + y2 ~ x + z)

# will fail:
# form_pred(y ~ .)

form_pred(terms(mpg ~ (.)^2, data = mtcars))
form_pred(terms( ~ (.)^2, data = mtcars))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apparent.R
\name{apparent}
\alias{apparent}
\title{Sampling for the Apparent Error Rate}
\usage{
apparent(data, ...)
}
\arguments{
\item{data}{A data frame.}

\item{...}{Not currently used.}
}
\value{
A tibble with a single row and classes \code{apparent},
\code{rset}, \code{tbl_df}, \code{tbl}, and \code{data.frame}. The
results include a column for the data split objects and one column
called \code{id} that has a character string with the resample identifier.
}
\description{
When building a model on a data set and re-predicting the same data, the
performance estimate from those predictions is often called the
"apparent" performance of the model. This estimate can be wildly
optimistic. "Apparent sampling" here means that the analysis and
assessment samples are the same. These resamples are sometimes used in
the analysis of bootstrap samples and should otherwise be
avoided like old sushi.
}
\examples{
apparent(mtcars)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lables.R
\name{add_resample_id}
\alias{add_resample_id}
\title{Augment a data set with resampling identifiers}
\usage{
add_resample_id(.data, split, dots = FALSE)
}
\arguments{
\item{.data}{A data frame}

\item{split}{A single \code{rset} object.}

\item{dots}{A single logical: should the id columns be prefixed with a "."
to avoid name conflicts with \code{.data}?}
}
\value{
An updated data frame.
}
\description{
For a data set, \code{add_resample_id()} will add at least one new column that
identifies which resample that the data came from. In most cases, a single
column is added but for some resampling methods, two or more are added.
}
\examples{
library(dplyr)

set.seed(363)
car_folds <- vfold_cv(mtcars, repeats = 3)

analysis(car_folds$splits[[1]]) \%>\%
  add_resample_id(car_folds$splits[[1]]) \%>\%
  head()

car_bt <- bootstraps(mtcars)

analysis(car_bt$splits[[1]]) \%>\%
  add_resample_id(car_bt$splits[[1]]) \%>\%
  head()
}
\seealso{
labels.rsplit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{add_resample_id}
\alias{add_resample_id}
\title{Augment a data set with resampling identifiers}
\usage{
add_resample_id(.data, split, dots = FALSE)
}
\arguments{
\item{.data}{A data frame}

\item{split}{A single \code{rset} object.}

\item{dots}{A single logical: should the id columns be prefixed with a "."
to avoid name conflicts with \code{.data}?}
}
\value{
An updated data frame.
}
\description{
For a data set, \code{add_resample_id()} will add at least one new column that
identifies which resample that the data came from. In most cases, a single
column is added but for some resampling methods, two or more are added.
}
\examples{
library(dplyr)

set.seed(363)
car_folds <- vfold_cv(mtcars, repeats = 3)

analysis(car_folds$splits[[1]]) \%>\%
  add_resample_id(car_folds$splits[[1]]) \%>\%
  head()

car_bt <- bootstraps(mtcars)

analysis(car_bt$splits[[1]]) \%>\%
  add_resample_id(car_bt$splits[[1]]) \%>\%
  head()
}
\seealso{
labels.rsplit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rolling_origin.R
\name{rolling_origin}
\alias{rolling_origin}
\title{Rolling Origin Forecast Resampling}
\usage{
rolling_origin(
  data,
  initial = 5,
  assess = 1,
  cumulative = TRUE,
  skip = 0,
  lag = 0,
  ...
)
}
\arguments{
\item{data}{A data frame.}

\item{initial}{The number of samples used for analysis/modeling in the
initial resample.}

\item{assess}{The number of samples used for each assessment resample.}

\item{cumulative}{A logical. Should the analysis resample grow beyond the
size specified by \code{initial} at each resample?.}

\item{skip}{A integer indicating how many (if any) \emph{additional} resamples
to skip to thin the total amount of data points in the analysis resample.
See the example below.}

\item{lag}{A value to include a lag between the assessment
and analysis set. This is useful if lagged predictors will be used
during training and testing.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
An tibble with classes \code{rolling_origin}, \code{rset}, \code{tbl_df}, \code{tbl},
and \code{data.frame}. The results include a column for the data split objects
and a column called \code{id} that has a character string with the resample
identifier.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

This resampling method is useful when the data set has a strong time
component. The resamples are not random and contain data points that are
consecutive values. The function assumes that the original data set are
sorted in time order.

This function is superseded by \code{\link[=sliding_window]{sliding_window()}}, \code{\link[=sliding_index]{sliding_index()}}, and
\code{\link[=sliding_period]{sliding_period()}} which provide more flexibility and control. Superseded
functions will not go away, but active development will be focused on the new
functions.
}
\details{
The main options, \code{initial} and \code{assess}, control the number of
data points from the original data that are in the analysis and assessment
set, respectively. When \code{cumulative = TRUE}, the analysis set will grow as
resampling continues while the assessment set size will always remain
static.
\code{skip} enables the function to not use every data point in the resamples.
When \code{skip = 0}, the resampling data sets will increment by one position.
Suppose that the rows of a data set are consecutive days. Using \code{skip = 6}
will make the analysis data set to operate on \emph{weeks} instead of days. The
assessment set size is not affected by this option.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
set.seed(1131)
ex_data <- data.frame(row = 1:20, some_var = rnorm(20))
dim(rolling_origin(ex_data))
dim(rolling_origin(ex_data, skip = 2))
dim(rolling_origin(ex_data, skip = 2, cumulative = FALSE))

# You can also roll over calendar periods by first nesting by that period,
# which is especially useful for irregular series where a fixed window
# is not useful. This example slides over 5 years at a time.
library(dplyr)
library(tidyr)
data(drinks, package = "modeldata")

drinks_annual <- drinks |>
  mutate(year = as.POSIXlt(date)$year + 1900) |>
  nest(data = c(-year))

multi_year_roll <- rolling_origin(drinks_annual, cumulative = FALSE)

analysis(multi_year_roll$splits[[1]])
assessment(multi_year_roll$splits[[1]])
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=sliding_window]{sliding_window()}}, \code{\link[=sliding_index]{sliding_index()}}, and \code{\link[=sliding_period]{sliding_period()}} for additional
time based resampling functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy.R
\name{tidy.rsplit}
\alias{tidy.rsplit}
\alias{tidy.rset}
\alias{tidy.vfold_cv}
\alias{tidy.nested_cv}
\title{Tidy Resampling Object}
\usage{
\method{tidy}{rsplit}(x, unique_ind = TRUE, ...)

\method{tidy}{rset}(x, ...)

\method{tidy}{vfold_cv}(x, ...)

\method{tidy}{nested_cv}(x, ...)
}
\arguments{
\item{x}{A  \code{rset} or  \code{rsplit} object}

\item{unique_ind}{Should unique row identifiers be returned? For example,
if \code{FALSE} then bootstrapping results will include multiple rows in the
sample for the same row in the original data.}

\item{...}{Not currently used.}
}
\value{
A tibble with columns \code{Row} and \code{Data}. The latter has possible
values "Analysis" or "Assessment". For \code{rset} inputs, identification columns
are also returned but their names and values depend on the type of
resampling. \code{vfold_cv} contains a column "Fold" and, if repeats are used,
another called "Repeats". \code{bootstraps} and \code{mc_cv} use the column
"Resample".
}
\description{
The \code{tidy} function from the \pkg{broom} package can be used on \code{rset} and
\code{rsplit} objects to generate tibbles with which rows are in the analysis and
assessment sets.
}
\details{
Note that for nested resampling, the rows of the inner resample,
named \code{inner_Row}, are \emph{relative} row indices and do not correspond to the
rows in the original data set.
}
\examples{
library(ggplot2)
theme_set(theme_bw())

set.seed(4121)
cv <- tidy(vfold_cv(mtcars, v = 5))
ggplot(cv, aes(x = Fold, y = Row, fill = Data)) +
  geom_tile() +
  scale_fill_brewer()

set.seed(4121)
rcv <- tidy(vfold_cv(mtcars, v = 5, repeats = 2))
ggplot(rcv, aes(x = Fold, y = Row, fill = Data)) +
  geom_tile() +
  facet_wrap(~Repeat) +
  scale_fill_brewer()

set.seed(4121)
mccv <- tidy(mc_cv(mtcars, times = 5))
ggplot(mccv, aes(x = Resample, y = Row, fill = Data)) +
  geom_tile() +
  scale_fill_brewer()

set.seed(4121)
bt <- tidy(bootstraps(mtcars, time = 5))
ggplot(bt, aes(x = Resample, y = Row, fill = Data)) +
  geom_tile() +
  scale_fill_brewer()

dat <- data.frame(day = 1:30)
# Resample by week instead of day
ts_cv <- rolling_origin(dat,
  initial = 7, assess = 7,
  skip = 6, cumulative = FALSE
)
ts_cv <- tidy(ts_cv)
ggplot(ts_cv, aes(x = Resample, y = factor(Row), fill = Data)) +
  geom_tile() +
  scale_fill_brewer()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation_set.R
\name{validation_set}
\alias{validation_set}
\alias{analysis.val_split}
\alias{assessment.val_split}
\alias{training.val_split}
\alias{validation.val_split}
\alias{testing.val_split}
\title{Create a Validation Split for Tuning}
\usage{
validation_set(split, ...)

\method{analysis}{val_split}(x, ...)

\method{assessment}{val_split}(x, ...)

\method{training}{val_split}(x, ...)

\method{validation}{val_split}(x, ...)

\method{testing}{val_split}(x, ...)
}
\arguments{
\item{split}{An object of class \code{initial_validation_split}, such as resulting
from \code{\link[=initial_validation_split]{initial_validation_split()}} or \code{\link[=group_initial_validation_split]{group_initial_validation_split()}}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{x}{An \code{rsplit} object produced by \code{validation_set()}.}
}
\value{
An tibble with classes \code{validation_set}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split object and a
column called \code{id} that has a character string with the resample identifier.
}
\description{
\code{validation_set()} creates a the validation split for model tuning.
}
\examples{
set.seed(1353)
car_split <- initial_validation_split(mtcars)
car_set <- validation_set(car_split)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/permutations.R
\name{permutations}
\alias{permutations}
\title{Permutation sampling}
\usage{
permutations(data, permute = NULL, times = 25, apparent = FALSE, ...)
}
\arguments{
\item{data}{A data frame.}

\item{permute}{One or more columns to shuffle. This argument supports
tidyselect selectors. Multiple expressions can be combined with \code{c()}.
Variable names can be used as if they were positions in the data frame, so
expressions like \code{x:y} can be used to select a range of variables.
See \code{\link[tidyselect]{language}} for more details.}

\item{times}{The number of permutation samples.}

\item{apparent}{A logical. Should an extra resample be added where the
analysis is the standard data set.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A \code{tibble} with classes \code{permutations}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and a
column called \code{id} that has a character string with the resample
identifier.
}
\description{
A permutation sample is the same size as the original data set and is made
by permuting/shuffling one or more columns. This results in analysis
samples where some columns are in their original order and some columns
are permuted to a random order. Unlike other sampling functions in
rsample, there is no assessment set and calling \code{\link[=assessment]{assessment()}} on a
permutation split will throw an error.
}
\details{
The argument \code{apparent} enables the option of an additional
"resample" where the analysis data set is the same as the original data
set. Permutation-based resampling can be especially helpful for computing
a statistic under the null hypothesis (e.g. t-statistic). This forms the
basis of a permutation test, which computes a test statistic under all
possible permutations of the data.
}
\examples{
permutations(mtcars, mpg, times = 2)
permutations(mtcars, mpg, times = 2, apparent = TRUE)

library(purrr)
resample1 <- permutations(mtcars, starts_with("c"), times = 1)
resample1$splits[[1]] |> analysis()

resample2 <- permutations(mtcars, hp, times = 10, apparent = TRUE)
map_dbl(resample2$splits, function(x) {
  t.test(hp ~ vs, data = analysis(x))$statistic
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation_split.R
\name{validation_split}
\alias{validation_split}
\alias{validation_time_split}
\alias{group_validation_split}
\title{Create a Validation Set}
\usage{
validation_split(data, prop = 3/4, strata = NULL, breaks = 4, pool = 0.1, ...)

validation_time_split(data, prop = 3/4, lag = 0, ...)

group_validation_split(data, group, prop = 3/4, ..., strata = NULL, pool = 0.1)
}
\arguments{
\item{data}{A data frame.}

\item{prop}{The proportion of data to be retained for modeling/analysis.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{...}{These dots are for future extensions and must be empty.}

\item{lag}{A value to include a lag between the assessment and analysis set.
This is useful if lagged predictors will be used during training and
testing.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}
}
\value{
An tibble with classes \code{validation_split}, \code{rset}, \code{tbl_df}, \code{tbl},
and \code{data.frame}. The results include a column for the data split objects
and a column called \code{id} that has a character string with the resample
identifier.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

This function is deprecated because it's part of an approach to constructing
a training, validation, and testing set by doing a sequence of two binary
splits: testing / not-testing (with \code{\link[=initial_split]{initial_split()}} or one of its variants)
and then not-testing split into training/validation with \code{validation_split()}.
Instead, now use \code{\link[=initial_validation_split]{initial_validation_split()}} or one if its variants to
construct the three sets via one 3-way split.

\code{validation_split()} takes a single random sample (without replacement) of
the original data set to be used for analysis. All other data points are
added to the assessment set (to be used as the validation set).
\code{validation_time_split()} does the same, but takes the \emph{first} \code{prop} samples
for training, instead of a random selection.
\code{group_validation_split()} creates splits of the data based
on some grouping variable, so that all data in a "group" is assigned to
the same split.

Note that the input \code{data} to \code{validation_split()}, \code{validation_time_split()},
and \code{group_validation_split()} should \emph{not} contain the testing data. To
create a three-way split directly of the entire data set, use
\code{\link[=initial_validation_split]{initial_validation_split()}}.
}
\details{
With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
cars_split <- initial_split(mtcars)
cars_not_testing <- training(cars_split)
validation_split(cars_not_testing, prop = .9)
group_validation_split(cars_not_testing, cyl)

data(drinks, package = "modeldata")
validation_time_split(drinks[1:200,])

# Alternative
cars_split_3 <- initial_validation_split(mtcars)
validation_set(cars_split_3)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=initial_validation_split]{initial_validation_split()}}, \code{\link[=group_initial_validation_split]{group_initial_validation_split()}},
\code{\link[=validation_set]{validation_set()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{loo_cv}
\alias{loo_cv}
\title{Leave-One-Out Cross-Validation}
\usage{
loo_cv(data, ...)
}
\arguments{
\item{data}{A data frame.}

\item{...}{Not currently used.}
}
\value{
An tibble with classes \code{loo_cv}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and
one column called \code{id} that has a character string with the resample
identifier.
}
\description{
Leave-one-out (LOO) cross-validation uses one data point in the original
set as the assessment data and all other data points as the analysis set. A
LOO resampling set has as many resamples as rows in the original data set.
}
\examples{
loo_cv(mtcars)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_groups.R
\name{make_groups}
\alias{make_groups}
\title{Make groupings for grouped rsplits}
\usage{
make_groups(
  data,
  group,
  v,
  balance = c("groups", "observations", "prop"),
  strata = NULL,
  ...
)
}
\arguments{
\item{data}{A data frame.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}

\item{v}{The number of partitions of the data set.}

\item{balance}{If \code{v} is less than the number of unique groups, how should
groups be combined into folds? Should be one of
\code{"groups"}, \code{"observations"}, \code{"prop"}.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{...}{Arguments passed to balance functions.}
}
\description{
This function powers grouped resampling by splitting the data based upon
a grouping variable and returning the assessment set indices for each
split.
}
\details{
Not all \code{balance} options are accepted -- or make sense -- for all resampling
functions. For instance, \code{balance = "prop"} assigns groups to folds at
random, meaning that any given observation is not guaranteed to be in one
(and only one) assessment set. That means \code{balance = "prop"} can't
be used with \code{\link[=group_vfold_cv]{group_vfold_cv()}}, and so isn't an option available for that
function.

Similarly, \code{\link[=group_mc_cv]{group_mc_cv()}} and its derivatives don't assign data to one (and
only one) assessment set, but rather allow each observation to be in an
assessment set zero-or-more times. As a result, those functions don't have
a \code{balance} argument, and under the hood always specify \code{balance = "prop"}
when they call \code{make_groups()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groups.R
\name{group_vfold_cv}
\alias{group_vfold_cv}
\title{Group V-Fold Cross-Validation}
\usage{
group_vfold_cv(data, group = NULL, v = NULL, ...)
}
\arguments{
\item{data}{A data frame.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}

\item{v}{The number of partitions of the data set. If left as \code{NULL}, \code{v}
will be set to the number of unique values in the group.}

\item{...}{Not currently used.}
}
\value{
A tibble with classes \code{group_vfold_cv},
\code{rset}, \code{tbl_df}, \code{tbl}, and \code{data.frame}.
The results include a column for the data split objects and an
identification variable.
}
\description{
Group V-fold cross-validation creates splits of the data based
on some grouping variable (which may have more than a single row
associated with it). The function can create as many splits as
there are unique values of the grouping variable or it can
create a smaller set of splits where more than one value is left
out at a time. A common use of this kind of resampling is when you have
repeated measures of the same subject.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
data(Sacramento, package = "modeldata")

set.seed(123)
group_vfold_cv(Sacramento, group = city, v = 5)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mc.R
\name{mc_cv}
\alias{mc_cv}
\title{Monte Carlo Cross-Validation}
\usage{
mc_cv(data, prop = 3/4, times = 25, strata = NULL, breaks = 4, pool = 0.1, ...)
}
\arguments{
\item{data}{A data frame.}

\item{prop}{The proportion of data to be retained for modeling/analysis.}

\item{times}{The number of times to repeat the sampling.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
An tibble with classes \code{mc_cv}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and a
column called \code{id} that has a character string with the resample identifier.
}
\description{
One resample of Monte Carlo cross-validation takes a random sample (without
replacement) of the original data set to be used for analysis. All other
data points are added to the assessment set.
}
\details{
With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
mc_cv(mtcars, times = 2)
mc_cv(mtcars, prop = .5, times = 2)

library(purrr)
data(wa_churn, package = "modeldata")

set.seed(13)
resample1 <- mc_cv(wa_churn, times = 3, prop = .5)
map_dbl(
  resample1$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
resample2 <- mc_cv(wa_churn, strata = churn, times = 3, prop = .5)
map_dbl(
  resample2$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
resample3 <- mc_cv(wa_churn, strata = tenure, breaks = 6, times = 3, prop = .5)
map_dbl(
  resample3$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slide.R
\name{slide-resampling}
\alias{slide-resampling}
\alias{sliding_window}
\alias{sliding_index}
\alias{sliding_period}
\title{Time-based Resampling}
\usage{
sliding_window(
  data,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L
)

sliding_index(
  data,
  index,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L
)

sliding_period(
  data,
  index,
  period,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L,
  every = 1L,
  origin = NULL
)
}
\arguments{
\item{data}{A data frame.}

\item{...}{These dots are for future extensions and must be empty.}

\item{lookback}{The number of elements to look back from the current element
when computing the resampling indices of the analysis set. The current
row is always included in the analysis set.
\itemize{
\item For \code{sliding_window()}, a single integer defining the number of rows to
look back from the current row.
\item For \code{sliding_index()}, a single object that will be subtracted from the
\code{index} as \code{index - lookback} to define the boundary of where to start
searching for rows to include in the current resample. This is often
an integer value corresponding to the number of days to look back,
or a lubridate Period object.
\item For \code{sliding_period()}, a single integer defining the number of groups
to look back from the current group, where the groups were defined from
breaking up the \code{index} according to the \code{period}.
}

In all cases, \code{Inf} is also allowed to force an expanding window.}

\item{assess_start, assess_stop}{This combination of arguments determines
how far into the future to look when constructing the assessment set.
Together they construct a range of
\verb{[index + assess_start, index + assess_stop]} to search for rows to include
in the assessment set.

Generally, \code{assess_start} will always be \code{1} to indicate that the first
value to potentially include in the assessment set should start one element
after the current row, but it can be increased to a larger value to
create "gaps" between the analysis and assessment set if you are worried
about high levels of correlation in short term forecasting.
\itemize{
\item For \code{sliding_window()}, these are both single integers defining the
number of rows to look forward from the current row.
\item For \code{sliding_index()}, these are single objects that will be added
to the \code{index} to compute the range to search for rows to include
in the assessment set. This is often an integer value corresponding to
the number of days to look forward, or a lubridate Period object.
\item For \code{sliding_period()}, these are both single integers defining the
number of groups to look forward from the current group, where the groups
were defined from breaking up the \code{index} according to the \code{period}.
}}

\item{complete}{A single logical. When using \code{lookback} to compute the
analysis sets, should only complete windows be considered? If set to
\code{FALSE}, partial windows will be used until it is possible to create
a complete window (based on \code{lookback}). This is a way to use an
expanding window up to a certain point, and then switch to a sliding
window.}

\item{step}{A single positive integer. After computing the resampling
indices, \code{step} is used to thin out the results by selecting every
\code{step}-th result by subsetting the indices with
\code{seq(1L, n_indices, by = step)}. \code{step} is applied after \code{skip}.
Note that \code{step} is independent of any time \code{index} used.}

\item{skip}{A single positive integer, or zero. After computing the
resampling indices, the first \code{skip} results will be dropped by subsetting
the indices with \code{seq(skip + 1L, n_indices)}. This can be especially
useful when combined with \code{lookback = Inf}, which creates an expanding
window starting from the first row. By skipping forward, you can drop
the first few windows that have very few data points. \code{skip} is
applied before \code{step}. Note that \code{skip} is independent of any time
\code{index} used.}

\item{index}{The index to compute resampling indices relative to, specified
as a bare column name. This must be an existing column in \code{data}.
\itemize{
\item For \code{sliding_index()}, this is commonly a date vector, but is not
required.
\item For \code{sliding_period()}, it is required that this is a Date or POSIXct
vector.
}

The \code{index} must be an \emph{increasing} vector, but duplicate values are
allowed. Additionally, the index cannot contain any missing values.}

\item{period}{The period to group the \code{index} by. This is specified as a
single string, such as \code{"year"} or \code{"month"}. See the \code{.period} argument
of \code{\link[slider:slide_index]{slider::slide_index()}} for the full list of options and further
explanation.}

\item{every}{A single positive integer. The number of periods to group
together.

For example, if the \code{period} was set to \code{"year"} with an \code{every}
value of 2, then the years 1970 and 1971 would be placed in the same
group.}

\item{origin}{The reference date time value. The default when left
as \code{NULL} is the epoch time of \verb{1970-01-01 00:00:00},
\emph{in the time zone of the index}.

This is generally used to define the anchor time to count from,
which is relevant when the \code{every} value is \verb{> 1}.}
}
\description{
These resampling functions are focused on various forms of \emph{time series}
resampling.
\itemize{
\item \code{sliding_window()} uses the row number when computing the resampling
indices. It is independent of any time index, but is useful with
completely regular series.
\item \code{sliding_index()} computes resampling indices relative to the \code{index}
column. This is often a Date or POSIXct column, but doesn't have to be.
This is useful when resampling irregular series, or for using irregular
lookback periods such as \code{lookback = lubridate::years(1)} with daily
data (where the number of days in a year may vary).
\item \code{sliding_period()} first breaks up the \code{index} into less granular groups
based on \code{period}, and then uses that to construct the resampling indices.
This is extremely useful for constructing rolling monthly or yearly
windows from daily data.
}
}
\examples{
library(vctrs)
library(tibble)
library(modeldata)
data("Chicago")

index <- new_date(c(1, 3, 4, 7, 8, 9, 13, 15, 16, 17))
df <- tibble(x = 1:10, index = index)
df

# Look back two rows beyond the current row, for a total of three rows
# in each analysis set. Each assessment set is composed of the two rows after
# the current row.
sliding_window(df, lookback = 2, assess_stop = 2)

# Same as before, but step forward by 3 rows between each resampling slice,
# rather than just by 1.
rset <- sliding_window(df, lookback = 2, assess_stop = 2, step = 3)
rset

analysis(rset$splits[[1]])
analysis(rset$splits[[2]])

# Now slide relative to the `index` column in `df`. This time we look back
# 2 days from the current row's `index` value, and 2 days forward from
# it to construct the assessment set. Note that this series is irregular,
# so it produces different results than `sliding_window()`. Additionally,
# note that it is entirely possible for the assessment set to contain no
# data if you have a highly irregular series and "look forward" into a
# date range where no data points actually exist!
sliding_index(df, index, lookback = 2, assess_stop = 2)

# With `sliding_period()`, we can break up our date index into less granular
# chunks, and slide over them instead of the index directly. Here we'll use
# the Chicago data, which contains daily data spanning 16 years, and we'll
# break it up into rolling yearly chunks. Three years worth of data will
# be used for the analysis set, and one years worth of data will be held out
# for performance assessment.
sliding_period(
  Chicago,
  date,
  "year",
  lookback = 2,
  assess_stop = 1
)

# Because `lookback = 2`, three years are required to form a "complete"
# window of data. To allow partial windows, set `complete = FALSE`.
# Here that first constructs two expanding windows until a complete three
# year window can be formed, at which point we switch to a sliding window.
sliding_period(
  Chicago,
  date,
  "year",
  lookback = 2,
  assess_stop = 1,
  complete = FALSE
)

# Alternatively, you could break the resamples up by month. Here we'll
# use an expanding monthly window by setting `lookback = Inf`, and each
# assessment set will contain two months of data. To ensure that we have
# enough data to fit our models, we'll `skip` the first 4 expanding windows.
# Finally, to thin out the results, we'll `step` forward by 2 between
# each resample.
sliding_period(
  Chicago,
  date,
  "month",
  lookback = Inf,
  assess_stop = 2,
  skip = 4,
  step = 2
)
}
\seealso{
\code{\link[=rolling_origin]{rolling_origin()}}

\code{\link[slider:slide]{slider::slide()}}, \code{\link[slider:slide_index]{slider::slide_index()}}, and \code{\link[slider:slide_period]{slider::slide_period()}},
which power these resamplers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slide.R
\name{slide-resampling}
\alias{slide-resampling}
\alias{sliding_window}
\alias{sliding_index}
\alias{sliding_period}
\title{Time-based Resampling}
\usage{
sliding_window(
  data,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L
)

sliding_index(
  data,
  index,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L
)

sliding_period(
  data,
  index,
  period,
  ...,
  lookback = 0L,
  assess_start = 1L,
  assess_stop = 1L,
  complete = TRUE,
  step = 1L,
  skip = 0L,
  every = 1L,
  origin = NULL
)
}
\arguments{
\item{data}{A data frame.}

\item{...}{These dots are for future extensions and must be empty.}

\item{lookback}{The number of elements to look back from the current element
when computing the resampling indices of the analysis set. The current
row is always included in the analysis set.
\itemize{
\item For \code{sliding_window()}, a single integer defining the number of rows to
look back from the current row.
\item For \code{sliding_index()}, a single object that will be subtracted from the
\code{index} as \code{index - lookback} to define the boundary of where to start
searching for rows to include in the current resample. This is often
an integer value corresponding to the number of days to look back,
or a lubridate Period object.
\item For \code{sliding_period()}, a single integer defining the number of groups
to look back from the current group, where the groups were defined from
breaking up the \code{index} according to the \code{period}.
}

In all cases, \code{Inf} is also allowed to force an expanding window.}

\item{assess_start, assess_stop}{This combination of arguments determines
how far into the future to look when constructing the assessment set.
Together they construct a range of
\verb{[index + assess_start, index + assess_stop]} to search for rows to include
in the assessment set.

Generally, \code{assess_start} will always be \code{1} to indicate that the first
value to potentially include in the assessment set should start one element
after the current row, but it can be increased to a larger value to
create "gaps" between the analysis and assessment set if you are worried
about high levels of correlation in short term forecasting.
\itemize{
\item For \code{sliding_window()}, these are both single integers defining the
number of rows to look forward from the current row.
\item For \code{sliding_index()}, these are single objects that will be added
to the \code{index} to compute the range to search for rows to include
in the assessment set. This is often an integer value corresponding to
the number of days to look forward, or a lubridate Period object.
\item For \code{sliding_period()}, these are both single integers defining the
number of groups to look forward from the current group, where the groups
were defined from breaking up the \code{index} according to the \code{period}.
}}

\item{complete}{A single logical. When using \code{lookback} to compute the
analysis sets, should only complete windows be considered? If set to
\code{FALSE}, partial windows will be used until it is possible to create
a complete window (based on \code{lookback}). This is a way to use an
expanding window up to a certain point, and then switch to a sliding
window.}

\item{step}{A single positive integer. After computing the resampling
indices, \code{step} is used to thin out the results by selecting every
\code{step}-th result by subsetting the indices with
\code{seq(1L, n_indices, by = step)}. \code{step} is applied after \code{skip}.
Note that \code{step} is independent of any time \code{index} used.}

\item{skip}{A single positive integer, or zero. After computing the
resampling indices, the first \code{skip} results will be dropped by subsetting
the indices with \code{seq(skip + 1L, n_indices)}. This can be especially
useful when combined with \code{lookback = Inf}, which creates an expanding
window starting from the first row. By skipping forward, you can drop
the first few windows that have very few data points. \code{skip} is
applied before \code{step}. Note that \code{skip} is independent of any time
\code{index} used.}

\item{index}{The index to compute resampling indices relative to, specified
as a bare column name. This must be an existing column in \code{data}.
\itemize{
\item For \code{sliding_index()}, this is commonly a date vector, but is not
required.
\item For \code{sliding_period()}, it is required that this is a Date or POSIXct
vector.
}

The \code{index} must be an \emph{increasing} vector, but duplicate values are
allowed. Additionally, the index cannot contain any missing values.}

\item{period}{The period to group the \code{index} by. This is specified as a
single string, such as \code{"year"} or \code{"month"}. See the \code{.period} argument
of \code{\link[slider:slide_index]{slider::slide_index()}} for the full list of options and further
explanation.}

\item{every}{A single positive integer. The number of periods to group
together.

For example, if the \code{period} was set to \code{"year"} with an \code{every}
value of 2, then the years 1970 and 1971 would be placed in the same
group.}

\item{origin}{The reference date time value. The default when left
as \code{NULL} is the epoch time of \verb{1970-01-01 00:00:00},
\emph{in the time zone of the index}.

This is generally used to define the anchor time to count from,
which is relevant when the \code{every} value is \verb{> 1}.}
}
\description{
These resampling functions are focused on various forms of \emph{time series}
resampling.
\itemize{
\item \code{sliding_window()} uses the row number when computing the resampling
indices. It is independent of any time index, but is useful with
completely regular series.
\item \code{sliding_index()} computes resampling indices relative to the \code{index}
column. This is often a Date or POSIXct column, but doesn't have to be.
This is useful when resampling irregular series, or for using irregular
lookback periods such as \code{lookback = lubridate::years(1)} with daily
data (where the number of days in a year may vary).
\item \code{sliding_period()} first breaks up the \code{index} into more granular groups
based on \code{period}, and then uses that to construct the resampling indices.
This is extremely useful for constructing rolling monthly or yearly
windows from daily data.
}
}
\examples{
library(vctrs)
library(tibble)
library(modeldata)
data("Chicago")

index <- new_date(c(1, 3, 4, 7, 8, 9, 13, 15, 16, 17))
df <- tibble(x = 1:10, index = index)
df

# Look back two rows beyond the current row, for a total of three rows
# in each analysis set. Each assessment set is composed of the two rows after
# the current row.
sliding_window(df, lookback = 2, assess_stop = 2)

# Same as before, but step forward by 3 rows between each resampling slice,
# rather than just by 1.
rset <- sliding_window(df, lookback = 2, assess_stop = 2, step = 3)
rset

analysis(rset$splits[[1]])
analysis(rset$splits[[2]])

# Now slide relative to the `index` column in `df`. This time we look back
# 2 days from the current row's `index` value, and 2 days forward from
# it to construct the assessment set. Note that this series is irregular,
# so it produces different results than `sliding_window()`. Additionally,
# note that it is entirely possible for the assessment set to contain no
# data if you have a highly irregular series and "look forward" into a
# date range where no data points actually exist!
sliding_index(df, index, lookback = 2, assess_stop = 2)

# With `sliding_period()`, we can break up our date index into more granular
# chunks, and slide over them instead of the index directly. Here we'll use
# the Chicago data, which contains daily data spanning 16 years, and we'll
# break it up into rolling yearly chunks. Three years worth of data will
# be used for the analysis set, and one years worth of data will be held out
# for performance assessment.
sliding_period(
  Chicago,
  date,
  "year",
  lookback = 2,
  assess_stop = 1
)

# Because `lookback = 2`, three years are required to form a "complete"
# window of data. To allow partial windows, set `complete = FALSE`.
# Here that first constructs two expanding windows until a complete three
# year window can be formed, at which point we switch to a sliding window.
sliding_period(
  Chicago,
  date,
  "year",
  lookback = 2,
  assess_stop = 1,
  complete = FALSE
)

# Alternatively, you could break the resamples up by month. Here we'll
# use an expanding monthly window by setting `lookback = Inf`, and each
# assessment set will contain two months of data. To ensure that we have
# enough data to fit our models, we'll `skip` the first 4 expanding windows.
# Finally, to thin out the results, we'll `step` forward by 2 between
# each resample.
sliding_period(
  Chicago,
  date,
  "month",
  lookback = Inf,
  assess_stop = 2,
  skip = 4,
  step = 2
)
}
\seealso{
\code{\link[=rolling_origin]{rolling_origin()}}

\code{\link[slider:slide]{slider::slide()}}, \code{\link[slider:slide_index]{slider::slide_index()}}, and \code{\link[slider:slide_period]{slider::slide_period()}},
which power these resamplers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mc.R
\name{group_mc_cv}
\alias{group_mc_cv}
\title{Group Monte Carlo Cross-Validation}
\usage{
group_mc_cv(
  data,
  group,
  prop = 3/4,
  times = 25,
  ...,
  strata = NULL,
  pool = 0.1
)
}
\arguments{
\item{data}{A data frame.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}

\item{prop}{The proportion of data to be retained for modeling/analysis.}

\item{times}{The number of times to repeat the sampling.}

\item{...}{These dots are for future extensions and must be empty.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}
}
\value{
A tibble with classes \code{group_mc_cv},
\code{rset}, \code{tbl_df}, \code{tbl}, and \code{data.frame}.
The results include a column for the data split objects and an
identification variable.
}
\description{
Group Monte Carlo cross-validation creates splits of the data based
on some grouping variable (which may have more than a single row
associated with it). One resample of Monte Carlo cross-validation takes a
random sample (without replacement) of groups in the original data set to be
used for analysis. All other data points are added to the assessment set.
A common use of this kind of resampling is when you have
repeated measures of the same subject.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
data(ames, package = "modeldata")

set.seed(123)
group_mc_cv(ames, group = Neighborhood, times = 5)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vfold.R
\name{vfold_cv}
\alias{vfold_cv}
\title{V-Fold Cross-Validation}
\usage{
vfold_cv(data, v = 10, repeats = 1, strata = NULL, breaks = 4, pool = 0.1, ...)
}
\arguments{
\item{data}{A data frame.}

\item{v}{The number of partitions of the data set.}

\item{repeats}{The number of times to repeat the V-fold partitioning.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{...}{Not currently used.}
}
\value{
A tibble with classes \code{vfold_cv}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and
one or more identification variables. For a single repeat, there will be
one column called \code{id} that has a character string with the fold identifier.
For repeats, \code{id} is the repeat number and an additional column called \code{id2}
that contains the fold information (within repeat).
}
\description{
V-fold cross-validation (also known as k-fold cross-validation) randomly
splits the data into V groups of roughly equal size (called "folds"). A
resample of the analysis data consists of V-1 of the folds while the
assessment set contains the final fold. In basic V-fold cross-validation
(i.e. no repeats), the number of resamples is equal to V.
}
\details{
With more than one repeat, the basic V-fold cross-validation is
conducted each time. For example, if three repeats are used with \code{v = 10},
there are a total of 30 splits: three groups of 10 that are generated
separately.

With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
vfold_cv(mtcars, v = 10)
vfold_cv(mtcars, v = 10, repeats = 2)

library(purrr)
data(wa_churn, package = "modeldata")

set.seed(13)
folds1 <- vfold_cv(wa_churn, v = 5)
map_dbl(
  folds1$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
folds2 <- vfold_cv(wa_churn, strata = churn, v = 5)
map_dbl(
  folds2$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
folds3 <- vfold_cv(wa_churn, strata = tenure, breaks = 6, v = 5)
map_dbl(
  folds3$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compat-vctrs-helpers.R
\name{rset_reconstruct}
\alias{rset_reconstruct}
\title{Extending rsample with new rset subclasses}
\usage{
rset_reconstruct(x, to)
}
\arguments{
\item{x}{A data frame to restore to an rset subclass.}

\item{to}{An rset subclass to restore to.}
}
\value{
\code{x} restored to the rset subclass of \code{to}.
}
\description{
\code{rset_reconstruct()} encapsulates the logic for allowing new rset
subclasses to work properly with vctrs (through \code{vctrs::vec_restore()}) and
dplyr (through \code{dplyr::dplyr_reconstruct()}). It is intended to be a
developer tool, and is not required for normal usage of rsample.
}
\details{
rset objects are considered "reconstructable" after a vctrs/dplyr operation
if:
\itemize{
\item \code{x} and \code{to} both have an identical column named \code{"splits"} (column
and row order do not matter).
\item \code{x} and \code{to} both have identical columns prefixed with \code{"id"} (column
and row order do not matter).
}
}
\examples{
to <- bootstraps(mtcars, times = 25)

# Imitate a vctrs/dplyr operation,
# where the class might be lost along the way
x <- tibble::as_tibble(to)

# Say we added a new column to `x`. Here we mock a `mutate()`.
x$foo <- "bar"

# This is still reconstructable to `to`
rset_reconstruct(x, to)

# Say we lose the first row
x <- x[-1, ]

# This is no longer reconstructable to `to`, as `x` is no longer an rset
# bootstraps object with 25 bootstraps if one is lost!
rset_reconstruct(x, to)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nest.R
\name{nested_cv}
\alias{nested_cv}
\title{Nested or Double Resampling}
\usage{
nested_cv(data, outside, inside)
}
\arguments{
\item{data}{A data frame.}

\item{outside}{The initial resampling specification. This can be an already
created object or an expression of a new object (see the examples below).
If the latter is used, the \code{data} argument does not need to be
specified and, if it is given, will be ignored.}

\item{inside}{An expression for the type of resampling to be conducted
within the initial procedure.}
}
\value{
An tibble with \code{nested_cv} class and any other classes that
outer resampling process normally contains. The results include a
column for the outer data split objects, one or more \code{id} columns,
and a column of nested tibbles called \code{inner_resamples} with the
additional resamples.
}
\description{
\code{nested_cv} can be used to take the results of one resampling procedure
and conduct further resamples within each split. Any type of resampling
used in \code{rsample} can be used.
}
\details{
It is a bad idea to use bootstrapping as the outer resampling procedure (see
the example below)
}
\examples{
## Using expressions for the resampling procedures:
nested_cv(mtcars, outside = vfold_cv(v = 3), inside = bootstraps(times = 5))

## Using an existing object:
folds <- vfold_cv(mtcars)
nested_cv(mtcars, folds, inside = bootstraps(times = 5))

## The dangers of outer bootstraps:
set.seed(2222)
bad_idea <- nested_cv(mtcars,
  outside = bootstraps(times = 5),
  inside = vfold_cv(v = 3)
)

first_outer_split <- bad_idea$splits[[1]]
outer_analysis <- as.data.frame(first_outer_split)
sum(grepl("Volvo 142E", rownames(outer_analysis)))

## For the 3-fold CV used inside of each bootstrap, how are the replicated
## `Volvo 142E` data partitioned?
first_inner_split <- bad_idea$inner_resamples[[1]]$splits[[1]]
inner_analysis <- as.data.frame(first_inner_split)
inner_assess <- as.data.frame(first_inner_split, data = "assessment")

sum(grepl("Volvo 142E", rownames(inner_analysis)))
sum(grepl("Volvo 142E", rownames(inner_assess)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{labels.rsplit}
\alias{labels.rsplit}
\title{Find Labels from rsplit Object}
\usage{
\method{labels}{rsplit}(object, ...)
}
\arguments{
\item{object}{An \code{rsplit} object}

\item{...}{Not currently used.}
}
\value{
A tibble.
}
\description{
Produce a tibble of identification variables so that single
splits can be linked to a particular resample.
}
\examples{
cv_splits <- vfold_cv(mtcars)
labels(cv_splits$splits[[1]])
}
\seealso{
add_resample_id
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy.R
\name{tidy.rsplit}
\alias{tidy.rsplit}
\alias{tidy.rset}
\alias{tidy.vfold_cv}
\alias{tidy.nested_cv}
\title{Tidy Resampling Object}
\usage{
\method{tidy}{rsplit}(x, unique_ind = TRUE, ...)

\method{tidy}{rset}(x, unique_ind = TRUE, ...)

\method{tidy}{vfold_cv}(x, ...)

\method{tidy}{nested_cv}(x, unique_ind = TRUE, ...)
}
\arguments{
\item{x}{A \code{rset} or \code{rsplit} object}

\item{unique_ind}{Should unique row identifiers be returned? For example,
if \code{FALSE} then bootstrapping results will include multiple rows in the
sample for the same row in the original data.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A tibble with columns \code{Row} and \code{Data}. The latter has possible
values "Analysis" or "Assessment". For \code{rset} inputs, identification
columns are also returned but their names and values depend on the type of
resampling. For \code{\link[=vfold_cv]{vfold_cv()}}, contains a column "Fold" and, if repeats are
used, another called "Repeats". \code{\link[=bootstraps]{bootstraps()}} and \code{\link[=mc_cv]{mc_cv()}} use the column
"Resample".
}
\description{
The \code{tidy()} function from the \pkg{broom} package can be used on \code{rset} and
\code{rsplit} objects to generate tibbles with which rows are in the analysis and
assessment sets.
}
\details{
Note that for nested resampling, the rows of the inner resample,
named \code{inner_Row}, are \emph{relative} row indices and do not correspond to the
rows in the original data set.
}
\examples{
\dontshow{if (rlang::is_installed("ggplot2")) withAutoprint(\{ # examplesIf}
library(ggplot2)
theme_set(theme_bw())

set.seed(4121)
cv <- tidy(vfold_cv(mtcars, v = 5))
ggplot(cv, aes(x = Fold, y = Row, fill = Data)) +
  geom_tile() +
  scale_fill_brewer()

set.seed(4121)
rcv <- tidy(vfold_cv(mtcars, v = 5, repeats = 2))
ggplot(rcv, aes(x = Fold, y = Row, fill = Data)) +
  geom_tile() +
  facet_wrap(~Repeat) +
  scale_fill_brewer()

set.seed(4121)
mccv <- tidy(mc_cv(mtcars, times = 5))
ggplot(mccv, aes(x = Resample, y = Row, fill = Data)) +
  geom_tile() +
  scale_fill_brewer()

set.seed(4121)
bt <- tidy(bootstraps(mtcars, time = 5))
ggplot(bt, aes(x = Resample, y = Row, fill = Data)) +
  geom_tile() +
  scale_fill_brewer()

dat <- data.frame(day = 1:30)
# Resample by week instead of day
ts_cv <- rolling_origin(dat,
  initial = 7, assess = 7,
  skip = 6, cumulative = FALSE
)
ts_cv <- tidy(ts_cv)
ggplot(ts_cv, aes(x = Resample, y = factor(Row), fill = Data)) +
  geom_tile() +
  scale_fill_brewer()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lables.R
\name{labels.rsplit}
\alias{labels.rsplit}
\title{Find Labels from rsplit Object}
\usage{
\method{labels}{rsplit}(object, ...)
}
\arguments{
\item{object}{An \code{rsplit} object}

\item{...}{Not currently used.}
}
\value{
A tibble.
}
\description{
Produce a tibble of identification variables so that single
splits can be linked to a particular resample.
}
\examples{
cv_splits <- vfold_cv(mtcars)
labels(cv_splits$splits[[1]])
}
\seealso{
add_resample_id
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compat-dplyr.R
\name{rsample-dplyr}
\alias{rsample-dplyr}
\title{Compatibility with dplyr}
\description{
rsample should be fully compatible with dplyr 1.0.0.

With older versions of dplyr, there is partial support for the following
verbs: \code{mutate()}, \code{arrange()}, \code{filter()}, \code{rename()}, \code{select()}, and
\code{slice()}. We strongly recommend updating to dplyr 1.0.0 if possible to
get more complete integration with dplyr.
}
\section{Version Specific Behavior}{


rsample performs somewhat differently depending on whether you have
dplyr >= 1.0.0 (new) or dplyr < 1.0.0 (old). Additionally, version
0.0.7 of rsample (new) introduced some changes to how rsample objects
work with dplyr, even on old dplyr. Most of these changes influence the
return value of a dplyr verb and determine whether it will be a tibble
or an rsample rset subclass.

The table below attempts to capture most of these changes. These examples
are not exhaustive and may not capture some edge-cases.
\subsection{Joins}{

The following affect all of the dplyr joins, such as \code{left_join()},
\code{right_join()}, \code{full_join()}, and \code{inner_join()}.

Joins that alter the rows of the original rset object:\tabular{lccc}{
   operation \tab old rsample + old dplyr \tab new rsample + old dplyr \tab new rsample + new dplyr \cr
   \code{join(rset, tbl)} \tab error \tab error \tab tibble \cr
}


The idea here is that, if there are less rows in the result, the result should
not be an rset object. For example, you can't have a 10-fold CV object
without 10 rows.

Joins that keep the rows of the original rset object:\tabular{lccc}{
   operation \tab old rsample + old dplyr \tab new rsample + old dplyr \tab new rsample + new dplyr \cr
   \code{join(rset, tbl)} \tab error \tab error \tab rset \cr
}


As with the logic above, if the original rset object (defined by the split
column and the id column(s)) is left intact, the results should be an rset.
}

\subsection{Row Subsetting}{

As mentioned above, this should result in a tibble if any rows are removed
or added. Simply reordering rows still results in a valid rset with new
rsample.

Cases where rows are removed or added:\tabular{lccc}{
   operation \tab old rsample + old dplyr \tab new rsample + old dplyr \tab new rsample + new dplyr \cr
   \code{rset[ind,]} \tab tibble \tab tibble \tab tibble \cr
   \code{slice(rset)} \tab rset \tab tibble \tab tibble \cr
   \code{filter(rset)} \tab rset \tab tibble \tab tibble \cr
}


Cases where all rows are kept, but are possibly reordered:\tabular{lccc}{
   operation \tab old rsample + old dplyr \tab new rsample + old dplyr \tab new rsample + new dplyr \cr
   \code{rset[ind,]} \tab tibble \tab rset \tab rset \cr
   \code{slice(rset)} \tab rset \tab rset \tab rset \cr
   \code{filter(rset)} \tab rset \tab rset \tab rset \cr
   \code{arrange(rset)} \tab rset \tab rset \tab rset \cr
}

}

\subsection{Column Subsetting}{

When the \code{splits} column or any \code{id} columns are dropped or renamed,
the result should no longer be considered a valid rset.

Cases when the required columns are removed or renamed:\tabular{lccc}{
   operation \tab old rsample + old dplyr \tab new rsample + old dplyr \tab new rsample + new dplyr \cr
   \code{rset[,ind]} \tab tibble \tab tibble \tab tibble \cr
   \code{select(rset)} \tab rset \tab tibble \tab tibble \cr
   \code{rename(rset)} \tab tibble \tab tibble \tab tibble \cr
}


Cases when no required columns are affected:\tabular{lccc}{
   operation \tab old rsample + old dplyr \tab new rsample + old dplyr \tab new rsample + new dplyr \cr
   \code{rset[,ind]} \tab tibble \tab rset \tab rset \cr
   \code{select(rset)} \tab rset \tab rset \tab rset \cr
   \code{rename(rset)} \tab rset \tab rset \tab rset \cr
}

}

\subsection{Other Column Operations}{

Cases when the required columns are altered:\tabular{lccc}{
   operation \tab old rsample + old dplyr \tab new rsample + old dplyr \tab new rsample + new dplyr \cr
   \code{mutate(rset)} \tab rset \tab tibble \tab tibble \cr
}


Cases when no required columns are affected:\tabular{lccc}{
   operation \tab old rsample + old dplyr \tab new rsample + old dplyr \tab new rsample + new dplyr \cr
   \code{mutate(rset)} \tab rset \tab rset \tab rset \cr
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{loo_cv}
\alias{loo_cv}
\title{Leave-One-Out Cross-Validation}
\usage{
loo_cv(data, ...)
}
\arguments{
\item{data}{A data frame.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
An tibble with classes \code{loo_cv}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and
one column called \code{id} that has a character string with the resample
identifier.
}
\description{
Leave-one-out (LOO) cross-validation uses one data point in the original
set as the assessment data and all other data points as the analysis set. A
LOO resampling set has as many resamples as rows in the original data set.
}
\examples{
loo_cv(mtcars)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boot.R
\name{bootstraps}
\alias{bootstraps}
\title{Bootstrap Sampling}
\usage{
bootstraps(
  data,
  times = 25,
  strata = NULL,
  breaks = 4,
  pool = 0.1,
  apparent = FALSE,
  ...
)
}
\arguments{
\item{data}{A data frame.}

\item{times}{The number of bootstrap samples.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{apparent}{A logical. Should an extra resample be added where the
analysis and holdout subset are the entire data set. This is required for
some estimators used by the \code{summary} function that require the apparent
error rate.}

\item{...}{Not currently used.}
}
\value{
An tibble with classes \code{bootstraps}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and a
column called \code{id} that has a character string with the resample identifier.
}
\description{
A bootstrap sample is a sample that is the same size as the original data
set that is made using replacement. This results in analysis samples that
have multiple replicates of some of the original rows of the data. The
assessment set is defined as the rows of the original data that were not
included in the bootstrap sample. This is often referred to as the
"out-of-bag" (OOB) sample.
}
\details{
The argument \code{apparent} enables the option of an additional
"resample" where the analysis and assessment data sets are the same as the
original data set. This can be required for some types of analysis of the
bootstrap results.

With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
bootstraps(mtcars, times = 2)
bootstraps(mtcars, times = 2, apparent = TRUE)

library(purrr)
library(modeldata)
data(wa_churn)

set.seed(13)
resample1 <- bootstraps(wa_churn, times = 3)
map_dbl(
  resample1$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
resample2 <- bootstraps(wa_churn, strata = churn, times = 3)
map_dbl(
  resample2$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
resample3 <- bootstraps(wa_churn, strata = tenure, breaks = 6, times = 3)
map_dbl(
  resample3$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reverse_splits.R
\name{reverse_splits}
\alias{reverse_splits}
\alias{reverse_splits.default}
\alias{reverse_splits.permutations}
\alias{reverse_splits.perm_split}
\alias{reverse_splits.rsplit}
\alias{reverse_splits.rset}
\title{Reverse the analysis and assessment sets}
\usage{
reverse_splits(x, ...)

\method{reverse_splits}{default}(x, ...)

\method{reverse_splits}{permutations}(x, ...)

\method{reverse_splits}{perm_split}(x, ...)

\method{reverse_splits}{rsplit}(x, ...)

\method{reverse_splits}{rset}(x, ...)
}
\arguments{
\item{x}{An \code{rset} or \code{rsplit} object.}

\item{...}{Not currently used.}
}
\value{
An object of the same class as \code{x}
}
\description{
This functions "swaps" the analysis and assessment sets of either a single
\code{rsplit} or all \code{rsplit}s in the \code{splits} column of an \code{rset} object.
}
\examples{
set.seed(123)
starting_splits <- vfold_cv(mtcars, v = 3)
reverse_splits(starting_splits)
reverse_splits(starting_splits$splits[[1]])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vfold.R
\name{group_vfold_cv}
\alias{group_vfold_cv}
\title{Group V-Fold Cross-Validation}
\usage{
group_vfold_cv(
  data,
  group = NULL,
  v = NULL,
  repeats = 1,
  balance = c("groups", "observations"),
  ...,
  strata = NULL,
  pool = 0.1
)
}
\arguments{
\item{data}{A data frame.}

\item{group}{A variable in \code{data} (single character or name) used for
grouping observations with the same value to either the analysis or
assessment set within a fold.}

\item{v}{The number of partitions of the data set. If left as \code{NULL} (the
default), \code{v} will be set to the number of unique values in the grouping
variable, creating "leave-one-group-out" splits.}

\item{repeats}{The number of times to repeat the V-fold partitioning.}

\item{balance}{If \code{v} is less than the number of unique groups, how should
groups be combined into folds? Should be one of
\code{"groups"}, which will assign roughly the same number of groups to each
fold, or \code{"observations"}, which will assign roughly the same number of
observations to each fold.}

\item{...}{These dots are for future extensions and must be empty.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}
}
\value{
A tibble with classes \code{group_vfold_cv},
\code{rset}, \code{tbl_df}, \code{tbl}, and \code{data.frame}.
The results include a column for the data split objects and an
identification variable.
}
\description{
Group V-fold cross-validation creates splits of the data based
on some grouping variable (which may have more than a single row
associated with it). The function can create as many splits as
there are unique values of the grouping variable or it can
create a smaller set of splits where more than one group is left
out at a time. A common use of this kind of resampling is when you have
repeated measures of the same subject.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
data(ames, package = "modeldata")

set.seed(123)
group_vfold_cv(ames, group = Neighborhood, v = 5)
group_vfold_cv(
  ames,
  group = Neighborhood,
  v = 5,
  balance = "observations"
)
group_vfold_cv(ames, group = Neighborhood, v = 5, repeats = 2)

# Leave-one-group-out CV
group_vfold_cv(ames, group = Neighborhood)

library(dplyr)
data(Sacramento, package = "modeldata")

city_strata <- Sacramento |>
  group_by(city) |>
  summarize(strata = mean(price)) |>
  reframe(city = city,
          strata = cut(strata, quantile(strata), include.lowest = TRUE))

sacramento_data <- Sacramento |>
  full_join(city_strata, by = "city")

group_vfold_cv(sacramento_data, city, strata = strata)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boot.R
\name{bootstraps}
\alias{bootstraps}
\title{Bootstrap Sampling}
\usage{
bootstraps(data, times = 25, strata = NULL, breaks = 4, apparent = FALSE, ...)
}
\arguments{
\item{data}{A data frame.}

\item{times}{The number of bootstrap samples.}

\item{strata}{A variable that is used to conduct stratified sampling. When
not \code{NULL}, each bootstrap sample is created within the stratification
variable. This could be a single character value or a variable name that
corresponds to a variable that exists in the data frame.}

\item{breaks}{A single number giving the number of bins desired to stratify
a numeric stratification variable.}

\item{apparent}{A logical. Should an extra resample be added where the
analysis and holdout subset are the entire data set. This is required for
some estimators used by the \code{summary} function that require the apparent
error rate.}

\item{...}{Not currently used.}
}
\value{
An tibble with classes \code{bootstraps}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and a
column called \code{id} that has a character string with the resample identifier.
}
\description{
A bootstrap sample is a sample that is the same size as the original data
set that is made using replacement. This results in analysis samples that
have multiple replicates of some of the original rows of the data. The
assessment set is defined as the rows of the original data that were not
included in the bootstrap sample. This is often referred to as the
"out-of-bag" (OOB) sample.
}
\details{
The argument \code{apparent} enables the option of an additional
"resample" where the analysis and assessment data sets are the same as the
original data set. This can be required for some types of analysis of the
bootstrap results.
The \code{strata} argument is based on a similar argument in the random forest
package were the bootstrap samples are conducted \emph{within the stratification
variable}. This can help ensure that the number of data points in the
bootstrap sample is equivalent to the proportions in the original data set.
(Strata below 10\% of the total are pooled together.)
}
\examples{
bootstraps(mtcars, times = 2)
bootstraps(mtcars, times = 2, apparent = TRUE)

library(purrr)
data(wa_churn, package = "modeldata")

set.seed(13)
resample1 <- bootstraps(wa_churn, times = 3)
map_dbl(resample1$splits,
        function(x) {
          dat <- as.data.frame(x)$churn
          mean(dat == "Yes")
        })

set.seed(13)
resample2 <- bootstraps(wa_churn, strata = "churn", times = 3)
map_dbl(resample2$splits,
        function(x) {
          dat <- as.data.frame(x)$churn
          mean(dat == "Yes")
        })

set.seed(13)
resample3 <- bootstraps(wa_churn, strata = "tenure", breaks = 6, times = 3)
map_dbl(resample3$splits,
        function(x) {
          dat <- as.data.frame(x)$churn
          mean(dat == "Yes")
        })
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{two_class_dat}
\alias{two_class_dat}
\title{Two Class Data}
\description{
Two Class Data
}
\details{
There are artificial data with two predictors (\code{A} and \code{B}) and
a factor outcome variable (\code{Class}).

These data are now in the \code{modeldata} package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gather.R
\name{gather.rset}
\alias{gather.rset}
\title{Gather an \code{rset} Object}
\usage{
\method{gather}{rset}(
  data,
  key = NULL,
  value = NULL,
  ...,
  na.rm = TRUE,
  convert = FALSE,
  factor_key = TRUE
)
}
\arguments{
\item{data}{An \code{rset} object.}

\item{key, value, ...}{Not specified in this method and will be
ignored. Note that this means that selectors are ignored if
they are passed to the function.}

\item{na.rm}{If \code{TRUE}, will remove rows from output where the
value column in \code{NA}.}

\item{convert}{If \code{TRUE} will automatically run
\code{type.convert()} on the key column. This is useful if the column
names are actually numeric, integer, or logical.}

\item{factor_key}{If FALSE, the default, the key values will be
stored as a character vector. If \code{TRUE}, will be stored as a
factor, which preserves the original ordering of the columns.}
}
\value{
A data frame with the ID columns, a column called
\code{model} (with the previous column names), and a column called
\code{statistic} (with the values).
}
\description{
This method uses \code{gather} on an \code{rset} object to stack all of
the non-ID or split columns in the data and is useful for
stacking model evaluation statistics. The resulting data frame
has a column based on the column names of \code{data} and another for
the values.
}
\examples{
library(rsample)
cv_obj <- vfold_cv(mtcars, v = 10)
cv_obj$lm_rmse <- rnorm(10, mean = 2)
cv_obj$nnet_rmse <- rnorm(10, mean = 1)
gather(cv_obj)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vfold.R
\name{vfold_cv}
\alias{vfold_cv}
\title{V-Fold Cross-Validation}
\usage{
vfold_cv(data, v = 10, repeats = 1, strata = NULL, breaks = 4, pool = 0.1, ...)
}
\arguments{
\item{data}{A data frame.}

\item{v}{The number of partitions of the data set.}

\item{repeats}{The number of times to repeat the V-fold partitioning.}

\item{strata}{A variable in \code{data} (single character or name) used to conduct
stratified sampling. When not \code{NULL}, each resample is created within the
stratification variable. Numeric \code{strata} are binned into quartiles.}

\item{breaks}{A single number giving the number of bins desired to stratify a
numeric stratification variable.}

\item{pool}{A proportion of data used to determine if a particular group is
too small and should be pooled into another group. We do not recommend
decreasing this argument below its default of 0.1 because of the dangers
of stratifying groups that are too small.}

\item{...}{These dots are for future extensions and must be empty.}
}
\value{
A tibble with classes \code{vfold_cv}, \code{rset}, \code{tbl_df}, \code{tbl}, and
\code{data.frame}. The results include a column for the data split objects and
one or more identification variables. For a single repeat, there will be
one column called \code{id} that has a character string with the fold identifier.
For repeats, \code{id} is the repeat number and an additional column called \code{id2}
that contains the fold information (within repeat).
}
\description{
V-fold cross-validation (also known as k-fold cross-validation) randomly
splits the data into V groups of roughly equal size (called "folds"). A
resample of the analysis data consists of V-1 of the folds while the
assessment set contains the final fold. In basic V-fold cross-validation
(i.e. no repeats), the number of resamples is equal to V.
}
\details{
With more than one repeat, the basic V-fold cross-validation is
conducted each time. For example, if three repeats are used with \code{v = 10},
there are a total of 30 splits: three groups of 10 that are generated
separately.

With a \code{strata} argument, the random sampling is conducted
\emph{within the stratification variable}. This can help ensure that the
resamples have equivalent proportions as the original data set. For
a categorical variable, sampling is conducted separately within each class.
For a numeric stratification variable, \code{strata} is binned into quartiles,
which are then used to stratify. Strata below 10\% of the total are
pooled together; see \code{\link[=make_strata]{make_strata()}} for more details.
}
\examples{
\dontshow{if (rlang::is_installed("modeldata")) withAutoprint(\{ # examplesIf}
vfold_cv(mtcars, v = 10)
vfold_cv(mtcars, v = 10, repeats = 2)

library(purrr)
data(wa_churn, package = "modeldata")

set.seed(13)
folds1 <- vfold_cv(wa_churn, v = 5)
map_dbl(
  folds1$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
folds2 <- vfold_cv(wa_churn, strata = churn, v = 5)
map_dbl(
  folds2$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)

set.seed(13)
folds3 <- vfold_cv(wa_churn, strata = tenure, breaks = 6, v = 5)
map_dbl(
  folds3$splits,
  function(x) {
    dat <- as.data.frame(x)$churn
    mean(dat == "Yes")
  }
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal_calibration_split.R
\name{internal_calibration_split}
\alias{internal_calibration_split}
\alias{internal_calibration_split.default}
\alias{internal_calibration_split.mc_split}
\alias{internal_calibration_split.group_mc_split}
\alias{internal_calibration_split.vfold_split}
\alias{internal_calibration_split.group_vfold_split}
\alias{internal_calibration_split.boot_split}
\alias{internal_calibration_split.group_boot_split}
\alias{internal_calibration_split.val_split}
\alias{internal_calibration_split.group_val_split}
\alias{internal_calibration_split.time_val_split}
\alias{internal_calibration_split.clustering_split}
\alias{internal_calibration_split.apparent_split}
\alias{internal_calibration_split.sliding_window_split}
\alias{internal_calibration_split.sliding_index_split}
\alias{internal_calibration_split.sliding_period_split}
\alias{internal_calibration_split.initial_time_split}
\alias{internal_calibration_split.initial_validation_split}
\alias{internal_calibration_split.group_initial_validation_split}
\alias{internal_calibration_split.initial_validation_time_split}
\alias{calibration}
\alias{calibration.default}
\alias{calibration.internal_calibration_split}
\alias{assessment.internal_calibration_split}
\alias{print.internal_calibration_split}
\title{Internal calibration split of the analysis set for fitting a post-processor}
\usage{
internal_calibration_split(x, ...)

\method{internal_calibration_split}{default}(x, ...)

\method{internal_calibration_split}{mc_split}(x, split_args, ...)

\method{internal_calibration_split}{group_mc_split}(x, split_args, ...)

\method{internal_calibration_split}{vfold_split}(x, split_args, ...)

\method{internal_calibration_split}{group_vfold_split}(x, split_args, ...)

\method{internal_calibration_split}{boot_split}(x, split_args, ...)

\method{internal_calibration_split}{group_boot_split}(x, split_args, ...)

\method{internal_calibration_split}{val_split}(x, split_args, ...)

\method{internal_calibration_split}{group_val_split}(x, split_args, ...)

\method{internal_calibration_split}{time_val_split}(x, split_args, ...)

\method{internal_calibration_split}{clustering_split}(x, split_args, ...)

\method{internal_calibration_split}{apparent_split}(x, ...)

\method{internal_calibration_split}{sliding_window_split}(x, split_args, ...)

\method{internal_calibration_split}{sliding_index_split}(x, split_args, ...)

\method{internal_calibration_split}{sliding_period_split}(x, split_args, ...)

\method{internal_calibration_split}{initial_time_split}(x, split_args, ...)

\method{internal_calibration_split}{initial_validation_split}(x, split_args, ...)

\method{internal_calibration_split}{group_initial_validation_split}(x, split_args, ...)

\method{internal_calibration_split}{initial_validation_time_split}(x, split_args, ...)

calibration(x, ...)

\method{calibration}{default}(x, ...)

\method{calibration}{internal_calibration_split}(x, ...)

\method{assessment}{internal_calibration_split}(x, ...)

\method{print}{internal_calibration_split}(x, ...)
}
\arguments{
\item{x}{An \code{rsplit} object.}

\item{...}{Not currently used.}

\item{split_args}{A list of arguments to be used for the internal calibration
split.}
}
\value{
An \code{rsplit} object.
}
\description{
Internal calibration split of the analysis set for fitting a post-processor
}
\details{
\code{rsplit} objects live most commonly inside of an \code{rset} object. The
\code{split_args} argument can be the output of \code{\link[=.get_split_args]{.get_split_args()}} on that
corresponding \code{rset} object, even if some of the arguments used to create the
\code{rset} object are not needed for the internal calibration split.
\itemize{
\item For \code{mc_split} and \code{group_mc_split} objects, \code{internal_calibration_split()}
will ignore \code{split_args$times}.
\item For \code{vfold_split} and \code{group_vfold_split} objects, it will ignore
\code{split_args$times} and \code{split_args$repeats}. \code{split_args$v} will be used to
set \code{split_args$prop} to \code{1 - 1/v} if \code{prop} is not already set and otherwise
ignored. The method for \code{group_vfold_split} will always use
\code{split_args$balance = NULL}.
\item For \code{boot_split} and \code{group_boot_split} objects, it will ignore
\code{split_args$times}.
\item For \code{val_split}, \code{group_val_split}, and \code{time_val_split} objects, it will
interpret a length-2 \code{split_args$prop} as a ratio between the training and
validation sets and split into inner analysis and calibration set in
the same ratio. If \code{split_args$prop} is a single value, it will be used as
the proportion of the inner analysis set.
\item For \code{clustering_split} objects, it will ignore \code{split_args$repeats}.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nested_cv.R
\name{nested_cv}
\alias{nested_cv}
\title{Nested or Double Resampling}
\usage{
nested_cv(data, outside, inside)
}
\arguments{
\item{data}{A data frame.}

\item{outside}{The initial resampling specification. This can be an already
created object or an expression of a new object (see the examples below).
If the latter is used, the \code{data} argument does not need to be
specified and, if it is given, will be ignored.}

\item{inside}{An expression for the type of resampling to be conducted
within the initial procedure.}
}
\value{
An tibble with \code{nested_cv} class and any other classes that
outer resampling process normally contains. The results include a
column for the outer data split objects, one or more \code{id} columns,
and a column of nested tibbles called \code{inner_resamples} with the
additional resamples.
}
\description{
\code{nested_cv()} can be used to take the results of one resampling procedure
and conduct further resamples within each split. Any type of resampling
used in rsample can be used.
}
\details{
It is a bad idea to use bootstrapping as the outer resampling procedure (see
the example below)
}
\examples{
## Using expressions for the resampling procedures:
nested_cv(mtcars, outside = vfold_cv(v = 3), inside = bootstraps(times = 5))

## Using an existing object:
folds <- vfold_cv(mtcars)
nested_cv(mtcars, folds, inside = bootstraps(times = 5))

## The dangers of outer bootstraps:
set.seed(2222)
bad_idea <- nested_cv(mtcars,
  outside = bootstraps(times = 5),
  inside = vfold_cv(v = 3)
)

first_outer_split <- get_rsplit(bad_idea, 1)
outer_analysis <- analysis(first_outer_split)
sum(grepl("Camaro Z28", rownames(outer_analysis)))

## For the 3-fold CV used inside of each bootstrap, how are the replicated
## `Camaro Z28` data partitioned?
first_inner_split <- get_rsplit(bad_idea$inner_resamples[[1]], 1)
inner_analysis <- analysis(first_inner_split)
inner_assess <- assessment(first_inner_split)

sum(grepl("Camaro Z28", rownames(inner_analysis)))
sum(grepl("Camaro Z28", rownames(inner_assess)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reexports.R, R/tidyselect.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{tidy}
\alias{contains}
\alias{select_helpers}
\alias{ends_with}
\alias{everything}
\alias{matches}
\alias{num_range}
\alias{starts_with}
\alias{last_col}
\alias{any_of}
\alias{all_of}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{generics}{\code{\link[generics]{tidy}}}

  \item{tidyselect}{\code{\link[tidyselect]{all_of}}, \code{\link[tidyselect:all_of]{any_of}}, \code{\link[tidyselect:starts_with]{contains}}, \code{\link[tidyselect:starts_with]{ends_with}}, \code{\link[tidyselect]{everything}}, \code{\link[tidyselect:everything]{last_col}}, \code{\link[tidyselect:starts_with]{matches}}, \code{\link[tidyselect:starts_with]{num_range}}, \code{\link[tidyselect]{starts_with}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compat-dplyr.R
\name{rsample-dplyr}
\alias{rsample-dplyr}
\title{Compatibility with dplyr}
\description{
This page lays out the compatibility between rsample and dplyr. The \code{rset}
objects from rsample are a specific subclass of tibbles, hence standard
dplyr operations like joins as well row or column modifications work.
However, whether the operation returns an rset or a tibble depends on the
details of the operation.

The overarching principle is that any operation which leaves the specific
characteristics of an rset intact will return an rset. If an operation
modifies any of the following characteristics, the result will be a \code{tibble}
rather than an \code{rset}:
\itemize{
\item Rows: The number of rows needs to remain unchanged to retain the rset
property. For example, you can't have a 10-fold CV object without 10 rows.
The order of the rows can be changed though and the object remains an rset.
\item Columns: The \code{splits} column and the \code{id} column(s) are required for an
rset and need to remain untouched. They cannot be dropped, renamed, or
modified if the result should remain an rset.
}
\subsection{Joins}{

The following affect all of the dplyr joins, such as \code{left_join()},
\code{right_join()}, \code{full_join()}, and \code{inner_join()}.

The resulting object is an \code{rset} if the number of rows is unaffected.
Rows can be reordered but not added or removed, otherwise the resulting object
is a \code{tibble}.\tabular{lcc}{
   operation \tab same rows, possibly reordered \tab add or remove rows \cr
   \code{join(rset, tbl)} \tab \code{rset} \tab \code{tibble} \cr
}

}

\subsection{Row Operations}{

The resulting object is an \code{rset} if the number of rows is unaffected.
Rows can be reordered but not added or removed, otherwise the resulting object
is a \code{tibble}.\tabular{lcc}{
   operation \tab same rows, possibly reordered \tab add or remove rows \cr
   \code{rset[ind,]} \tab \code{rset} \tab \code{tibble} \cr
   \code{slice(rset)} \tab \code{rset} \tab \code{tibble} \cr
   \code{filter(rset)} \tab \code{rset} \tab \code{tibble} \cr
   \code{arrange(rset)} \tab \code{rset} \tab \code{tibble} \cr
}

}

\subsection{Column Operations}{

The resulting object is an \code{rset} if the required \code{splits} and \code{id} columns
remain unaltered. Otherwise the resulting object is a \code{tibble}.\tabular{lcc}{
   operation \tab required columns unaltered \tab required columns removed, renamed, or modified \cr
   \code{rset[,ind]} \tab \code{rset} \tab \code{tibble} \cr
   \code{select(rset)} \tab \code{rset} \tab \code{tibble} \cr
   \code{rename(rset)} \tab \code{rset} \tab \code{tibble} \cr
   \code{mutate(rset)} \tab \code{rset} \tab \code{tibble} \cr
}

}
}
